/*=============================================================================
 
  Name: bsac.c

  Abs: This file contains all subroutine support for the beam synchronous
       acquisition and control subroutine records.
       
  Rem: All functions called by the subroutine record get passed one argument:

         psub                       Pointer to the subroutine record data.
          Use:  pointer
          Type: struct subRecord *
          Acc:  read/write
          Mech: reference

         All functions return a long integer.  0 = OK, -1 = ERROR.
         The subroutine record ignores the status returned by the Init
         routines.  For the calculation routines, the record status (STAT) is
         set to SOFT_ALARM (unless it is already set to LINK_ALARM due to
         severity maximization) and the severity (SEVR) is set to psub->brsv
         (BRSV - set by the user in the database though it is expected to
          be invalid).

  Auth:  
  Rev:  

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
   20-jul-07 dfairley changed CTRL to ACCESS, changed offline from 0 to 3
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int bsacSubFlag = DP_ERROR;
#endif

/* c includes */
#include <string.h>        /* memcmp */
#include <math.h>          /* sqrt   */

#include <subRecord.h>        /* for struct subRecord      */
#include <sSubRecord.h>       /* for struct sSubRecord in site area */
#include <registryFunction.h> /* for epicsExport           */
#include <epicsExport.h>      /* for epicsRegisterFunction */
#include <epicsTime.h>        /* epicsTimeStamp */
#include <dbAccess.h>         /* dbGetTimeStamp */
#include <alarm.h>            /* NO_ALARM, INVALID_ALARM */

#ifdef SLCAWARE
#include "slcMsg.h"           /* BSAC_DONE, BSAC_DATA function codes */
#include "slcTypes.h"         /* BSAC_HDLR_NAME, bsacHdlr */
#include "msgTypes.h"         /* msgmail_ts */
#include "bsacTypes.h"        /* bsacDevIdHdr_ts, BSACPUID_INDEX */
#include "bpmstruc.h"         /* bpm_pulseid_data_ts */
#include "msgQUtil.h"         /* msgQGetBuffer, msgQMsgReply */
#include "bsacData.h"         /* debug- bsacDataDebugPrintBPM() */
#include "bpmstat.h"          /* STAT_GOOD */
#endif


static int bsacInit(subRecord *psub)
{
  /*
   * General purpose initialization required since all subroutine records
   * require a non-NULL init routine even if no initialization is required.
    * Note that most subroutines in this file use this routine as an init
   * routine.  If init logic is needed for a specific subroutine, create a
   * new routine for it - don't modify this one.
   */
  return 0;
}


/*=============================================================================

  Name: bsacAvgCount

  Abs:  Device Average Counter; 120Hz Processing
		
  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        sSubRecord *        psub        read       point to subroutine record

  Rem:  Subroutine for PRIM:LOCA:UNIT:AVGCNT$MDID

  Side: INPA - MDEF:IOC:MD:AVGCNTMAX
        INPB - PRIM:LOCA:UNIT:STATUS   (EPICS STAT for device)
        INPC - PRIM:LOCA:UNIT:STATUS.L (EPICS SEVR for device)
        INPD - PRIM:LOCA:UNIT:STATUS.K (SLC status for device)
        INPE - MDEF:IOC:MD:PATTERNCHECK 1= triggered, PP, YY,INCL, EXCL match
		OUT
           F - EPICS sevr for average
           G - First time flag (0=first time after a reset, !=1 not first time)
	   H - Enable for SECN$MDID based on PATTERNCHECK, time valid
	   I - reset counters, values (avcnt, goodmeas, stat, timestamp check)
	   J - goodmeas count
           K - SLC stat for average
	   L - timestamp mismatch count
           M - EPICS stat for average
	   VAL = AVGCNT
      
  Ret:  0

==============================================================================*/
static long bsacAvgCount(sSubRecord *psub)
{ 
#ifndef linux
  epicsTimeStamp  timeEVR; 
  epicsTimeStamp  timeSecn;  /* for comparison */
#endif

  psub->h = 0; /* default to disable fanout */

  /* Do we care about this pulse ? */
  if (psub->e==0) {
        return 0;  /* don't care about this pulse */
  }
#ifndef linux
  /*EVR timestamp:*/
  if (dbGetTimeStamp(&psub->inpe, &timeEVR)){
        DEBUGPRINT(DP_ERROR, bsacSubFlag, ("bascAvgCount for %s: Unable to determine EVR timestamp.\n",psub->name));
        psub->l++;
        return -1;
  }
  /*Secondary timestamp:*/
  if (dbGetTimeStamp(&psub->inpb, &timeSecn)){
        DEBUGPRINT(DP_ERROR, bsacSubFlag, ("bascAvgCount for %s: Unable to determine device timestamp.\n",psub->name));
        psub->l++;
        return -1;
  }
  /* if timestamps are different, then processing is not synched */
  if (memcmp(&timeEVR,&timeSecn,sizeof(epicsTimeStamp))) {
	DEBUGPRINT(DP_ERROR, bsacSubFlag, ("bascAvgCount for %s: Timestamp MISMATCH.\n",psub->name));
	psub->l++;
	return -1;
  }
#endif
  psub->h = 1; /* enable fanout */
  if (psub->g < 2) psub->g++; /* increment first-time flag */

  /* Reinit for a new average */
  if (psub->i) {
    psub->i = 0;
    /*  reset avgcnt, goodmeas, stat   */
    psub->val = 0;
    psub->j = 0; /* goodmeas */
    psub->f = psub->c;
    psub->m = psub->b;
    psub->k = psub->d;
    /*DEBUGPRINT(DP_DEBUG, bsacSubFlag, ("bascAvgCount for %s: Reset counters; psub->i = %ld\n",psub->name, (unsigned long)psub->i));*/
  }
  /* No averaging requested? */
  if (psub->a <= 0) {
    psub->i = 1;
  } else {
    if (psub->c < INVALID_ALARM) {
      psub->j++;
      /* EPICS - check for worst severity */
      if (psub->f < psub->c) {
        psub->f = psub->c;
        psub->m = psub->b;
      }
#ifdef SLCAWARE
      /* SLC status - dead (not "online" overrides all else,
         otherwise take the biggest (SICK > OK > GOOD) */
      if (((unsigned short)psub->d | STAT_DEAD) ||
          (((unsigned short)psub->d & 0xF) >
           ((unsigned short)psub->k & 0xF))) {
        psub->k = psub->d;
      }
      DEBUGPRINT(DP_DEBUG, bsacSubFlag, ("bascAvgCount for %s: SLC stat=%ld\n",
                                        psub->name,(unsigned long) psub->k));
#endif
    }
    /* always incr avgcnt */
    ++psub->val;
    /* if val >= MDEF AVGCNTMAX */
    if (psub->val >= psub->a) {
      /*  then averaging is done */
      psub->i = 1;
      DEBUGPRINT(DP_DEBUG, bsacSubFlag, ("bascAvgCount done for %s:count=%ld\n",
                                        psub->name,(unsigned long) psub->val));
    } else {
      /*  else averaging is NOT done */
      psub->i = 0;
    }
  }  
  DEBUGPRINT(DP_DEBUG, bsacSubFlag, ("bascAvgCount for %s: VAL = %ld, goodmeas = %hd, psub->d = %ld; psub->i = %ld\n", psub->name,(unsigned long) psub->val, (unsigned short)psub->j, (unsigned long)psub->d, (unsigned long)psub->i));
  return 0;
}

/*=============================================================================

  Name: bsacStatus

  Abs:  Device EPICS Severity and Status, SLC Status - 120Hz Processing

		
  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        subRecord *        psub        read       point to subroutine record

  Rem:  Subroutine for PRIM:LOCA:UNIT:STATUS

  Side: INPA - PRIM:LOCA:UNIT:SECN1.STAT
        INPB - PRIM:LOCA:UNIT:SECN2.STAT 
        INPC - PRIM:LOCA:UNIT:SECN3.STAT 
        INPD - PRIM:LOCA:UNIT:SECN1.SEVR
        INPE - PRIM:LOCA:UNIT:SECN2.SEVR 
        INPF - PRIM:LOCA:UNIT:SECN3.SEVR 
        ***INPG - PRIM:LOCA:UNIT:CTRL***NOT ANY MORE
        INPG - PRIM:LOCA:UNIT:ACCESS
        INPH - PRIM:LOCA:UNIT:CALBSTATUS.SEVR
		OUT
	   K - SLC STAT
	   L - EPICS severity
	   VAL = EPICS STAT
      
  Ret:  0

==============================================================================*/
static long bsacStatus(subRecord *psub)
{
  psub->val = psub->a;
  psub->l   = psub->d;
  if (psub->e > psub->l) {
    psub->val = psub->b;
    psub->l   = psub->e;
  }
  if (psub->f > psub->l) {
    psub->val = psub->c;
    psub->l   = psub->f;
  }
#ifdef SLCAWARE  
  if      (psub->g == 3) /* offline? */
    psub->k = STAT_DEAD;
  else if (psub->g == 1) /* maintenance? */
    psub->k = STAT_MAINTENANCE;
  else if (psub->g == 2) /* checkout? */
    psub->k = STAT_CHECKOUT;
  else if (psub->h>NO_ALARM) /* calibration bad or in-progress */
    psub->k = STAT_BADCAL;
  else if (psub->l == NO_ALARM)
    psub->k = STAT_GOOD;
  else if (psub->l <= MAJOR_ALARM)
    psub->k = STAT_OK;
  else
    psub->k = STAT_SICK;
#endif
  return 0;
}

/*=============================================================================

  Name: bsacAvg

  Abs:  Device Secondary Processing
        Computes running average and RMS values of Secondary
        for PRIM:LOCA:UNIT:$SECN$MDID.VAL 

		
  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        subRecord *        psub        read       point to subroutine record

  Rem:  Subroutine for PRIM:LOCA:UNIT:MEASCNT$MDID

  Side: INPA - PRIM:LOCA:UNIT:SECN.VAL
        INPB - PRIM:LOCA:UNIT:AVGCNT$MDID.J - goodmeas count
        INPC - PRIM:LOCA:UNIT:AVGCNT$MDID.I - average done (0=not done, 1=done)
		OUT 
           H = Variance value used in RMS calc
           I = B - 1
           J = B - 2
           K = A - previous VAL
           L = RMS value when average is done
	   VAL = Running average 
      
  Ret:  0
==============================================================================*/
static long bsacAvg(subRecord *psub)
{  
  /* first time thru for new cycle; reset previous avg, variance */
  if (psub->b<=1) {
	psub->val = psub->a;
        psub->h = 0;
        psub->i = 0;
        psub->j = 0;
        psub->k = 0;
        psub->l = 0;
	DEBUGPRINT(DP_DEBUG, bsacSubFlag, ("bascAvg: First time thru for %s; psub->b=%ld\n", psub->name,(unsigned long) psub->b) );
  /* compute running avg and variance                      */
  /*        This is translated from REF_RMX_BPM:CUM.F86.   */
  /*                                                       */
  /*        CUM computes VAR as the sample variance:       */ 
  /*          VAR = (SUMSQ - SUM*SUM/N)/(N-1)              */
  /*          where SUM = sum of the N values, and         */
  /*           SUMSQ = sum of the squares of the N values. */
  /*                                                       */
  /*        Note that CUM's method of computing VAR avoids */
  /*        possible loss of significance.                 */
  } else {
        psub->i = psub->b-1.0;
        psub->j = psub->b-2.0;
        psub->k = psub->a-psub->val;
        psub->val += psub->k/psub->b;
        psub->k /= psub->i;
        psub->h = (psub->j*(psub->h/psub->i)) + (psub->b*psub->k*psub->k);
        DEBUGPRINT(DP_DEBUG, bsacSubFlag, ("bascAvg for %s: Avg: %f; Var: %f \n", psub->name, psub->val, psub->h) );
        if (psub->c) {
          psub->l = psub->h/psub->b;
          psub->l = sqrt(psub->l);
        }
  }
  return 0;
}

/*=============================================================================

  Name: bsacMeasCount

  Abs:  Device and Pattern Measurement Counter; 120Hz Processing
		
        After being enabled by bsacAvgCnt or bsacEVRCheck,
		send data to slc threads

		
  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        sSubRecord *        psub        read       point to subroutine record

  Rem:  Subroutine for PRIM:LOCA:UNIT:MEASCNT$MDID.VAL (devices)
	           and MDEF:IOC:MD:CNT.VAL (Pattern)

  $GDEV: ARRY:LOCA:UNIT
  $DEV:  PRIM:LOCA:UNIT, (DUGADC) where PRIM = BPMS,TORO,PMT,BLEN,WIRE,KICK
         or MDEF:IOC:MD  (Pattern)
  Side: INPA - Primary ID: 0 = BPMS, 1 = TORO, 2 = DUGADC, 4 = Pattern
        INPB - $GDEV:UNITID - unit index identification, Pattern: 1
        INPC - $SIG, channel of GDEV (0 for BPMS, TORO, Pattern; else ARRY channel)  
        INPD - $MD - measdef # 1-20 identification
        INPE - $DEV:$(SECN)$(MDID).VAL - Running average value of secondary
        INPF - $DEV:$(SECN)$(MDID).L  -  RMS value of secondary, Pattern: 0
the following 4 inputs are only valid for bpms:
        INPG - $DEV:Y$(MDID).VAL - Running average value of Y
        INPH - $DEV:Y$(MDID).L   - RMS value of Y
        INPI - $DEV:TMIT$(MDID).VAL - Running average value of TMIT
	INPJ - $DEV:TMIT$(MDID).L   - RMS value of TMIT
        INPK - $DEV:AVGCNT$MDID.I or MDEF:$(IOC):MD:PATTERNCHECK
                                    - Done averaging = 1
        INPL - $DEV:AVGCNT$MDID.G or MDEF:$(IOC):MD:PATTERNCHECK.Z
                                    - 1st pulse of new YY cycle
		                    (ensure counters are reset )
	INPM - MDEF:IOC:MD:MEASCNTMAX, MDEF:IOC:MD:CNTMAX for Pattern
        INPN - MDEF:IOC:MD:MODE (0=USER, 1=FEEDBACK, 2=SLCSINGLE, 3=SLCBUFFER)
	INPO - $DEV:AVGCNT$(MDID).J - goodmeas, Pattern: 1
	INPP - $DEV:AVGCNT$(MDID).K - stat, Pattern: 1
		OUT 
	   Z - slc threads flag; 1=slc threads; 0=no slc threads
	   Y - counter of failed message send
           X - function code to send to slc
	   Q - done buffering flag, 0=not done, 1=done
           R - enable history buffering, 0=disable, 1=enable
	   VAL = MEASCNT for devices, CNT for pattern

  Ret:  0

==============================================================================*/
static long bsacMeasCount(sSubRecord *psub)
{
#ifdef SLCAWARE
  msgptr_ts         mSndptr_s;
  bsacMsg_ts*       mSnd_ps;
#endif

  psub->r = psub->k; /* enable/disable history buffering */
  psub->z = 0;       /* no slc threads, or not a slc meas def */
  /* Check abort case - MEASCNTMAX set to zero */
  if (psub->m==0) {
	psub->q   = 1; /* done buffering */
	psub->r   = 0;
	psub->val = 0;        
  }
  /* first pulse of brand new cycle ?  */
  else if (psub->l==1) {
	psub->q   = 0;
	psub->val = 0;
    DEBUGPRINT(DP_DEBUG,bsacSubFlag,("bsacMeasCount: First pulse found for %s\n",psub->name));
  }  
  /* If buffering is done, check if reset is needed */
  else if (psub->q) {
	if ((psub->m < 0) || (psub->val < psub->m)
#ifdef SLCAWARE
		|| (psub->n == MODE_SLCSINGLE) ||
		(psub->n == MODE_SLCSINGLE_PUID)
#endif
		) {
	  psub->q   = 0; /* reset buffering */
	  psub->val = 0;
	}
	else {
	  psub->r = 0; /* disable history buffering */
	}
  }
  /* Done averaging? Check with AVGCNT.I record; 1 means done */
  if (psub->r) {
    if (psub->m > 0) {
      psub->val++;
      /* Check if buffering is done */
      if (psub->val >= psub->m) {
		psub->q = 1;
 DEBUGPRINT(DP_DEBUG,bsacSubFlag,("bsacMeasCount for %s: q set to 1 \n",psub->name));
	  }
    }
    /* if MODE > 1 (not 0=USER or 1=FEEDBACK) then SLC-Aware */
#ifdef SLCAWARE
    if (psub->n > MODE_FEEDBACK) {
      /* if not PUID or the acquisition is a PUID acquisition continue on */
      if ((psub->a != BSACPUID_INDEX) || (psub->n > MODE_SLCBUFFER)) {
        DEBUGPRINT(DP_DEBUG, bsacSubFlag,("bascMeasCount for %s: STAT = %hd\n",
                                         psub->name,(unsigned short) psub->p));
        /* if slc threads are up */
        if (slcThreads_as[bsacHdlr].mqId_ps) {
          psub->z = 1;
          mSndptr_s.msgbytes = sizeof(bsacMsg_ts);
          if (!(SUCCESS( msgQGetBuffer( mSndptr_s.msgbytes,
                                        (void*)&mSndptr_s.msgptr_p ) ))) {
            ++psub->y;	/* counter of failed message sends to slc threads */
          } else {
            mSnd_ps = (bsacMsg_ts *)mSndptr_s.msgptr_p;
            mSnd_ps->bsacMsgHdr_s.md        = (unsigned short) psub->d;
            mSnd_ps->bsacMsgHdr_s.primIndex = (unsigned short) psub->a;
            mSnd_ps->bsacMsgHdr_s.unitIndex = (unsigned short) psub->b;
            mSnd_ps->bsacMsgHdr_s.chanIndex = (unsigned short) psub->c;
            
            if (psub->val >= psub->m) {
              psub->x = BSAC_DONE;
              DEBUGPRINT(DP_DEBUG, bsacSubFlag,
                         ("bascMeasCount for %s: sending md %hd, primIndex %hd, unitIndex %hd, chanIndex %hd to bsacHdlr; func=BSAC_DONE\n",
                          psub->name, mSnd_ps->bsacMsgHdr_s.md,
                          mSnd_ps->bsacMsgHdr_s.primIndex,
                          mSnd_ps->bsacMsgHdr_s.unitIndex,  mSnd_ps->bsacMsgHdr_s.chanIndex) );
            } else {
              psub->x = BSAC_DATA;
              DEBUGPRINT(DP_DEBUG, bsacSubFlag,
                         ("bascMeasCount for %s: sending md %hd, primIndex %hd, unitIndex %hd, chanIndex %hd to bsacHdlr; func=BSAC_DATA\n",
                          psub->name, mSnd_ps->bsacMsgHdr_s.md,
                          mSnd_ps->bsacMsgHdr_s.primIndex,
                          mSnd_ps->bsacMsgHdr_s.unitIndex,  mSnd_ps->bsacMsgHdr_s.chanIndex) );
            }
            mSnd_ps->msgheader_s.func = psub->x;

            /* stuff output data into structure particular to primary for slc */
            switch ( mSnd_ps->bsacMsgHdr_s.primIndex) {
              case BSACBPMS_INDEX:
                if ((psub->n == MODE_SLCBUFFER) ||
                    (psub->n == MODE_SLCBUFFER_PUID)) {
                  /* buffered acquisition case */
                  mSnd_ps->data_u.bufacq_bpm_s.x    = (float) psub->e; /* x */
                  mSnd_ps->data_u.bufacq_bpm_s.y    = (float) psub->g;
                  mSnd_ps->data_u.bufacq_bpm_s.tmit = (float) psub->i;
                  mSnd_ps->data_u.bufacq_bpm_s.goodmeas = (unsigned short)psub->o;
                  mSnd_ps->data_u.bufacq_bpm_s.stat = (unsigned short)psub->p;
                } else { /* single acquisition case */
                  mSnd_ps->data_u.bpm_s.x        = (float) psub->e; /* x */
                  mSnd_ps->data_u.bpm_s.x_rms    = (float) psub->f; /* x rms */
                  mSnd_ps->data_u.bpm_s.y        = (float) psub->g;
                  mSnd_ps->data_u.bpm_s.y_rms    = (float) psub->h;
                  mSnd_ps->data_u.bpm_s.tmit     = (float) psub->i;
                  mSnd_ps->data_u.bpm_s.tmit_rms = (float) psub->j;
                  mSnd_ps->data_u.bpm_s.goodmeas = (unsigned short)psub->o;
                  mSnd_ps->data_u.bpm_s.stat     = (unsigned short)psub->p;
                  mSnd_ps->data_u.bpm_s.qraw[0]  = 0;
                  mSnd_ps->data_u.bpm_s.qraw[1]  = 0;
                  mSnd_ps->data_u.bpm_s.qraw[2]  = 0;
                  mSnd_ps->data_u.bpm_s.qraw[3]  = 0;
                  /*bsacDataDebugPrintBPM ((const bpms_data_ts *)&mSnd_ps->data_u.bpm_s);*/
                }
                break;
              case BSACTORO_INDEX:
                if ((psub->n == MODE_SLCBUFFER) ||
                    (psub->n == MODE_SLCBUFFER_PUID)){
                  /* buffered acquisition case */
                  mSnd_ps->data_u.bufacq_toro_s.tmit     = (float) psub->e;
                  mSnd_ps->data_u.bufacq_toro_s.goodmeas = (unsigned short)psub->o;
                  mSnd_ps->data_u.bufacq_toro_s.stat = (unsigned short)psub->p;
                  DEBUGPRINT(DP_DEBUG, bsacSubFlag,
                             ("bascMeasCount: ws_toro_data tmit=%f, goodmeas=%hd, stat=%hd\n",
                              mSnd_ps->data_u.bufacq_toro_s.tmit, mSnd_ps->data_u.bufacq_toro_s.goodmeas,
                              mSnd_ps->data_u.bufacq_toro_s.stat));
                } else { /* single acquisition case */
                  mSnd_ps->data_u.toro_s.tmit     = (float) psub->e;
                  mSnd_ps->data_u.toro_s.tmit_rms = (float) psub->f;
                  mSnd_ps->data_u.toro_s.qraw     = 0;
                  mSnd_ps->data_u.toro_s.goodmeas = (unsigned short)psub->o;
                  mSnd_ps->data_u.toro_s.stat     = (unsigned short)psub->p;
                  DEBUGPRINT(DP_DEBUG, bsacSubFlag,
                             ("bascMeasCount: toro_data tmit=%f, goodmeas=%hd, stat=%hd\n",
                              mSnd_ps->data_u.toro_s.tmit,
                              mSnd_ps->data_u.toro_s.goodmeas,
                              mSnd_ps->data_u.toro_s.stat));
                }
                break;
              case BSACARRY_INDEX:
                mSnd_ps->data_u.arryChan_s.data      = (float) psub->e;
                mSnd_ps->data_u.arryChan_s.drms      = (float) psub->f;
                mSnd_ps->data_u.arryChan_s.raw       = 0;
                mSnd_ps->data_u.arryChan_s.goodmeas  = (unsigned short)psub->o;
                mSnd_ps->data_u.arryChan_s.stat      = (unsigned short)psub->p;
                break;
              case BSACPUID_INDEX:
                mSnd_ps->data_u.puid_s.pulseid  = (unsigned long) psub->e;
                DEBUGPRINT(DP_DEBUG, bsacSubFlag,
                             ("bascMeasCount: ws_pulseid_data pulseid=%ld\n",
                              mSnd_ps->data_u.puid_s.pulseid));
                break;
              default:
                DEBUGPRINT(DP_ERROR, bsacSubFlag,
                      ("bascMeasCount: UNKNOWN md %hd, primIndex %hd, unitIndex %hd, chanIndex %hd DATA received\n",
                            mSnd_ps->bsacMsgHdr_s.md,mSnd_ps->bsacMsgHdr_s.primIndex,
                            mSnd_ps->bsacMsgHdr_s.unitIndex, mSnd_ps->bsacMsgHdr_s.chanIndex) );
                break;
            } /* switch */

            if (!(SUCCESS(messageQSend(bsacHdlr, epicsFalse,
                                       sizeof(mSndptr_s), &mSndptr_s)))) {
              ++psub->y;	/* counter of failed message sends to slc threads */
			  msgQFreeBuffer(sizeof(msgmail_ts), (void*)mSnd_ps);
            } /* end messageQSend */
          } /* end msgQGetBuffer   */
        } /* end SLC thread check */
      } /* end if message send required */  
    } /* end MODE check */
#endif
  } /* end averaging check */
  return 0;
}

/*=============================================================================

  Name: bsacEVRCheck

  Abs:  360Hz Processing
        Check to see if current beam pulse is to be used in any
		current measurement definition.
		

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        sSubRecord *        psub        read       point to sSub subroutine record

  Rem:   Subroutine for MDEF:$IOC:$MD:PATTERNCHECK

  side: INPA - $(DEV):YY
        INPB - $(DEV):BEAMCODE
        INPC - $(DEV):MODIFIER1 
        INPD - $(DEV):MODIFIER2 
        INPE - $(DEV):MODIFIER3 
        INPF - $(DEV):MODIFIER4 
        INPG - $(DEV):MODIFIER5
        INPH - MDEF:$(IOC):$(MD):YY  
        INPI - MDEF:$(IOC):$(MD):PP1
        INPJ - MDEF:$(IOC):$(MD):PP2
        INPK - MDEF:$(IOC):$(MD):PP3
        INPL - MDEF:$(IOC):$(MD):PP4
        INPM - MDEF:$(IOC):$(MD):INCLUSION1
        INPN - MDEF:$(IOC):$(MD):INCLUSION2
        INPO - MDEF:$(IOC):$(MD):INCLUSION3
        INPP - MDEF:$(IOC):$(MD):INCLUSION4
        INPQ - MDEF:$(IOC):$(MD):INCLUSION5
        INPR - MDEF:$(IOC):$(MD):EXCLUSION1
        INPS - MDEF:$(IOC):$(MD):EXCLUSION2
        INPT - MDEF:$(IOC):$(MD):EXCLUSION3
        INPU - MDEF:$(IOC):$(MD):EXCLUSION4
        INPV - MDEF:$(IOC):$(MD):EXCLUSION5
        INPW - $(DEV):BEAMCODE.SEVR
        INPX - MDEF:$(IOC):$(MD):CNT.Q
                OUT
	   Y - YY trigger
           Z - First time flag (0=first time after a reset, !=1 not first time)
	   VAL = Pattern match = 1
		 no match = 0; enable/disable for bsacMeasCount

           All three outputs are initialized to 0 before first-time processing
           when MDEF CTRL changes to ON.

  Ret:  none

==============================================================================*/
static long bsacEVRCheck(sSubRecord *psub)
{
  psub->val = 0;

  if (psub->w>MAJOR_ALARM) {
    /* bad data - do nothing this pulse and return bad status */
    return(-1);
  }
  /* If current meas is done, restart YY check.  For non-SLC meas,
     this will stop all checking and end the measurement. */
  if (psub->z) {
    if (psub->x) {
      if (psub->y) {
		psub->y = 0;
		DEBUGPRINT(DP_INFO,bsacSubFlag,("bsacEVRCheck: this YY DONE \n"));
        psub->z = 0; /* reset first time flag when YY done */
	  }
	}
  }
  /* If we've initialized, set Y flag if we're not checking YY */
  else if (psub->h < 0) {
    psub->y = 1;
  } 
  /* if YY hasn't been seen yet, check for it now */
  if ((psub->y==0) && (psub->a==psub->h)) {
    psub->y = 1;
    DEBUGPRINT(DP_INFO,bsacSubFlag,("bsacEVRCheck: YY match for %s, z=%f, x=%f\n", 
									psub->name, psub->z, psub->x));
  }
  /* if YY has already been seen, check this pulse */
  if (psub->y) {
	
	/* if Pattern PP == any MDEF PP1-4.. */
	if ( (psub->b==psub->i)||(psub->b==psub->j)||
	     (psub->b==psub->k)||(psub->b==psub->l)) {
	  /* check inclusion mask */
/*	  if (    (((unsigned long)psub->c & (unsigned long)psub->m) ==
				 (unsigned long)psub->m) &&
                  (((unsigned long)psub->d & (unsigned long)psub->n) ==
                   (unsigned long)psub->n) &&
                  (((unsigned long)psub->e & (unsigned long)psub->o) ==
                   (unsigned long)psub->o) &&
                  (((unsigned long)psub->f & (unsigned long)psub->p) ==
                   (unsigned long)psub->p) &&
                  (((unsigned long)psub->g & (unsigned long)psub->q) ==
				  (unsigned long)psub->q) ) {
*/
	  DEBUGPRINT(DP_DEBUG, bsacSubFlag, ("bsacEVRCheck:\t PPs match \n"));
	  if (    (((unsigned long)psub->d & (unsigned long)psub->m) ==
                   (unsigned long)psub->m) &&
                  (((unsigned long)psub->e & (unsigned long)psub->n) ==
                   (unsigned long)psub->n) &&
                  (((unsigned long)psub->f & (unsigned long)psub->o) ==
                   (unsigned long)psub->o)  ) {
            DEBUGPRINT(DP_DEBUG, bsacSubFlag,
                       ("bsacEVRCheck: inclusion match\n"));

/*            if (    (((unsigned long)psub->r & (unsigned long)psub->c) == 0) &&
                    (((unsigned long)psub->s & (unsigned long)psub->d) == 0) &&
                    (((unsigned long)psub->t & (unsigned long)psub->e) == 0) &&
                    (((unsigned long)psub->u & (unsigned long)psub->f) == 0) &&
                    (((unsigned long)psub->v & (unsigned long)psub->g) == 0) ) {
*/
            if (    (((unsigned long)psub->r & (unsigned long)psub->d) == 0) &&
                    (((unsigned long)psub->s & (unsigned long)psub->e) == 0) &&
                    (((unsigned long)psub->t & (unsigned long)psub->f) == 0) ) {
              psub->val = 1;
              if (psub->z < 2) psub->z++;
              DEBUGPRINT(DP_INFO, bsacSubFlag,
                         ("bsacEVRCheck: incl&excl match - pulse is good\n"));
            }
          }
        }
  }  	
  return 0;
}

epicsRegisterFunction(bsacInit);
epicsRegisterFunction(bsacAvg);
epicsRegisterFunction(bsacStatus);
epicsRegisterFunction(bsacAvgCount);
epicsRegisterFunction(bsacMeasCount);
epicsRegisterFunction(bsacEVRCheck);
