/*=============================================================================
 
  Name: bsacAbort.c


  Abs:  bsacAbort contains functions pertaining to aborting a measurement
        Reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | BSAC Service Design

  Auth: 8   aug 2005, dfairley    created 
  Rev:  

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest) 
       27-apr-2007, dfairley started LOCA2 fixes
        08 aug 2005, Dfairley    created 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int bsacAbortFlag = DP_DEBUG;
#endif


/* c includes */
#include <string.h>        /* memcpy */

/* epics */
#include <dbAccess.h>      /* EPICS dbAddr struct */

/* shared slc micro includes */
#include "bpmfunc.hc"      /* bpmo_xxx func codes */
#include "micrdef.hc"      /* MICR_EXNET_ codes */
#include "bpmodef.hc"      /* BPMO_OKOK etc */
#include "msgfunc.hc"      /* RESPONSE for func code */

/* slc global vars */
#include "slcGlob.h"       /* slcThreads_as */

/* slcIOC utilities */
#include "slcUtil.h"       /* slcPrintMsgheader */
#include "msgQUtil.h"      /* messageQSend, messageQRecv */
#include "cvtUtil.h"       /* slcSSwap, slcLSwap, etc */
#include "slcCmlog.h"      /* slcCmlogLogMsg */
#include "timeUtil.h"      /* timeGetVMS, timeDiffVMS10usec */

/* slcMain function prototypes */
/*#include "slcMain.h"        slcStop(), slcRestart() */

/* bsac includes */
#include    "bsacGlob.h"
#include    "bsacAbort.h"
#include    "bpmstruc.h"
#include    "bsacPrepTypes.h"

extern mdefAddr_ts mdefAddr_as[MDEF_MAX];
extern primAddrStruct_ts primAddr_as[BSACPRIM_MAX]; /* contains epics db addrs for */
                  /* records PRIM:LOCA:UNIT:UNITID */
                  /*  and PRIM:LOCA:UNIT:CTRLM */

/* prototypes */
LOCAL vmsstat_t abortMeas (const short md);
LOCAL vmsstat_t bsacAbortActiveGOs (const short md);

/*-----------------------------------------------------------------------
  Name: abortMeas

  Abs:  abort the measdef indicated by md

  Args: Type		  Name  Access	    Description
	    ------------- ----- ---------- ----------------------------
		const short   md    read only   number of the measdef to be aborted

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status 
----------------------------------------------------------------------*/
LOCAL vmsstat_t abortMeas (const short md)
{
  vmsstat_t iss = BPMO_OKOK;
  unsigned short ctrl = 0;
  char username[MAX_STRING_SIZE];
  
  /* init the username to 0*/
  memset(&username[0], 0, MAX_STRING_SIZE);

  memset(&bsacMD_as[md].username[0],0,5);
  bsacMD_as[md].abort = epicsTrue;
  bsacMD_as[md].timeoutCntr = 0;

  DEBUGPRINT(DP_INFO, bsacAbortFlag, 
	("bsacAbort abortMeas:aborting md=%d\n", md)); 

  /* reset device control for the MDEF */
  iss=bsacAbortActiveGOs (md);

  /* Next - set ctrl EPICS record to 0 */
  /* This will also disable the EVR GO record */
  if (mdefAddr_as[md].CTRL_s.precord){
	if (dbPutField(&mdefAddr_as[md].CTRL_s, DBR_USHORT, &ctrl, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacAbortFlag, 
			("abortMeas: dbPutField of MDEF:%s:%d:CTRL = %d failed; CTRL_s = %p\n", 
			 slcGetIOCName(),md+1, ctrl,(void *) mdefAddr_as[md].CTRL_s.pfield));
	} 
  }
  
  /* clear USERNAME */
  if (dbPutField(&mdefAddr_as[md].USERNAME_s, DBR_STRING, 
				 username, 1)!=0) {
	iss=BPMO_FAIL;
	DEBUGPRINT(DP_ERROR, bsacAbortFlag, 
			   ("bsacAbort abortMeas: dbPutField of MDEF:%s:%d:USERNAME failed\n", 
				slcGetIOCName(),md+1));
  }	else DEBUGPRINT(DP_DEBUG, bsacAbortFlag, 
					("bsacAbort abortMeas: of MDEF:%s:%d:USERNAME to blank\n", 
					 slcGetIOCName(),md+1));
  /* set MODE to 0 */
  if (dbPutField(&mdefAddr_as[md].MODE_s, DBR_USHORT, 
				 &ctrl, 1)) {
	iss=BPMO_FAIL;
	DEBUGPRINT(DP_ERROR, bsacAbortFlag, 
			   ("bsacAbort abortMeas: dbPutField of MDEF:%s:%d:MODE failed\n", 
				slcGetIOCName(),md+1));
  }	else DEBUGPRINT(DP_DEBUG, bsacAbortFlag, 
					("bsacAbort abortMeas of MDEF:%s:%d:MODE = 0\n", 
					 slcGetIOCName(),md+1));
  /* delete existing data for this md */
  iss = bsacAbortClearMeasBufs(md);

  return(iss);
}

/*-----------------------------------------------------------------------
  Name: bsacAbortCancelMeas

  Abs:  abort the measdef indicated by md

  Args: Type		  Name  Access	    Description
	    ------------- ----- ---------- ----------------------------
		const short   md    read only   number of the measdef to be cancelled

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacAbortCancelMeas (const short md)
{
  vmsstat_t iss;
  DEBUGPRINT(DP_INFO, bsacAbortFlag, 
	("bsacAbort cancelMeas:cancelling md=%d\n", md)); 
  iss = abortMeas(md);

  /* update diagnostics for cancelled measurement */
  return(iss);
}

/*-----------------------------------------------------------------------
  Name: bsacTimeoutMeas

  Abs:  abort the measdef indicated by md

  Args: Type		  Name  Access	    Description
	    ------------- ----- ---------- ----------------------------
		const short   md    read only   number of the measdef to be aborted

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacAbortTimeoutMeas (const short md)
{
  vmsstat_t iss, err_iss;

  DEBUGPRINT(DP_INFO, bsacAbortFlag, 
	("bsacAbortTimeoutMeas:Received timeout abort for md=%d\n", md)); 

  iss = abortMeas(md);
  /* update diagnostics */
  /* log a "timed out" message  */
  err_iss = BPMO_LARGBUFTMOUT;
	  slcCmlogLogMsg(err_iss, "LARGBUFTMOUT",
					 "measurement request timed out, unable to send all data.\n"); 
  return (iss);
}

/*=============================================================================

  Name: bsacAbortFindScp()

  Abs: clear all measurement definitions assigned by this scp

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const char *        scpName     read only   name of scp that requested CLR SCP 
		short *             md          read/write  md number assigned to this SCP

  Rem:  

  side: 
   
	 NOTE:  !!!bsacAbortFindScp - the call used to use dbGetFields to read epics
	 records CTRL and USERNAME to find the md.  This caused a segmentation error
	 in LINUX only, which to this point is unexplained.  A workaround was created 
	 by changing the bsacAbortFindScp to check a bsacMD_as[md].username stored 
	 during the measurement preparation 
 
  Ret: iss vmsstat_t vms style status

===============================================================================*/
vmsstat_t bsacAbortFindScp(const char* scpName, short *md)
{
  short mdef;
  vmsstat_t iss = MICR_EEXIST;

  DEBUGPRINT(DP_INFO, bsacAbortFlag, 
	("bsacAbortFindScp: for scp=%s\n", scpName)); 

  for (mdef = 0; mdef < MDEF_MAX; mdef++) {
    /* check each measdef for username==scpname and abort=false */
     /* get the USERNAME */
	if ( (strncmp(bsacMD_as[mdef].username, scpName, 4) == 0) &&
		 (bsacMD_as[mdef].abort==epicsFalse) ) {
	  DEBUGPRINT(DP_INFO, bsacAbortFlag, 
				 ("bsacAbortFindScp: MDEF:LOCA:%hd:USERNAME=%s, matches scpName=%s\n", 
				  mdef+1, bsacMD_as[mdef].username, scpName));
	  *md = mdef;
	  iss = BPMO_OKOK;
	  break; /* there is only one per scp */
	}
  }

  if (!SUCCESS(iss)) {
	*md = 0;
	DEBUGPRINT(DP_DEBUG, bsacAbortFlag, 
			   ("bsacAbortFindScp: NO match for scpName=%s\n", scpName));
  }
  return (iss);
}

/*=============================================================================

  Name: bsacAbortClrScp()

  Abs: clear all measurement definitions assigned by this scp

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const char *        scpName     read only   name of scp that requested CLR SCP   

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
vmsstat_t bsacAbortClrScp(const char* scpName)
{
  short md=0;
  vmsstat_t iss = BPMO_OKOK;
  
  if (SUCCESS(iss = (bsacAbortFindScp(scpName, &md)))) {
	iss = bsacAbortCancelMeas(md);
  }

  return (iss);
}

/*=============================================================================

  Name: bsacAbortMeasStop()

  Abs: clear the measdef using the i_seg number

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const char *        scpName     read only   name of scp that requested MEASS TOP 
        const unsigned short i_seg      read only   meas id that must be stopped  

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
vmsstat_t bsacAbortMeasStop(const char* scpName, const unsigned short i_seg)
{
  vmsstat_t iss = BPMO_OKOK;
  short md;
  long  nRequest=1;      /* these are for fetching Epics PV's */
  long  options=0;
  unsigned short ctrl=0;
  unsigned short measid=0;
  char  username[MAX_STRING_SIZE];
  int comp;

  memset(&username[0], 0, MAX_STRING_SIZE);

  DEBUGPRINT(DP_INFO, bsacAbortFlag, 
	("bsacAbortMeasStop:Received Meas Stop request for scp=%s, i_seg=%d\n", scpName, i_seg)); 

  for (md = 0; md < MDEF_MAX; md++) {
    /* check each measdef for ctrl=ON and abort=false */
	ctrl = 0; /* need to set each loop ?*/
	if ( (dbGetField(&mdefAddr_as[md].CTRL_s, DBR_ENUM, 
			   &ctrl, &options, &nRequest, NULL))) {
	  DEBUGPRINT(DP_INFO, bsacAbortFlag, 
		 ("bsacAbortMeasStop: Error for dbGetField of MDEF:LOCA:%d:CTRL\n", md+1)); 
	}
	if ( (ctrl != 0) && (bsacMD_as[md].abort == epicsFalse) ) {
      
	  if ((dbGetField(&mdefAddr_as[md].ID_s, DBR_USHORT, 
			   &measid, &options, &nRequest, NULL))==0) {
        /* if the i_seg matches check the USERNAME */
		if (measid == i_seg){
	 
		  /* get the USERNAME */
		  if ( (dbGetField(&mdefAddr_as[md].USERNAME_s, DBR_STRING, 
			   username, &options, &nRequest, NULL)) == 0) {
			/* compare to scpname in request*/
			comp = strncmp(username, scpName,4);
			if (strncmp(username, scpName, 4) == 0){
			  iss = bsacAbortCancelMeas(md);
			  break;  /* there should be only one, so stop now */
			}
		  }
		  else {
			DEBUGPRINT(DP_INFO, bsacAbortFlag, 
			("bsacAbortMeasStop: Error for dbGetField of MDEF:LOCA:%d:USERNAME\n", md+1));
		  }
		} 
	  }
	  else {
	    DEBUGPRINT(DP_INFO, bsacAbortFlag, 
				   ("bsacAbortMeasStop: Error for dbGetField of MDEF:LOCA:%d:ID\n", md+1));
	  }
	}
  }/* for */
  DEBUGPRINT(DP_INFO, bsacAbortFlag, ("bsacAbortMeasStop: Leaving\n"));
  return (iss);
}

/*=============================================================================

  Name: bsacAbortClearMeasBufs()

  Abs: clear all measurement data buffers assigned to this mdef

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const short         md          read only   mdef number of buffer set to clear
        const unsigned short i_seg      read only   meas id that must be stopped  

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
vmsstat_t bsacAbortClearMeasBufs(const short md)
{
  vmsstat_t iss = BPMO_OKOK;
  unsigned short buf = 0;
  lrgmail_ts * msg_ps = NULL;
  bsacDataBufs_ts * mdBuf_ps = &bsacMDBufs_as[md]; 

  for (buf = 0; buf < mdBuf_ps->bufCount; buf++) {
	msg_ps = mdBuf_ps->bufs_as[buf].buf_ps;
    if (msg_ps) {
	  if ( ((mdBuf_ps->bufs_as[buf].partwc*2)+sizeof(msgheader_ts)+sizeof(fwd_hdr_ts)) <
		   sizeof(msgmail_ts) ) {
		msgQFreeBuffer(sizeof(msgmail_ts), (void *)msg_ps);
		DEBUGPRINT(DP_INFO, bsacAbortFlag, 
				 ("bsacAbortClearMeasBuffs: free a SMALL buffer for md=%d\n", md));
	  } else {
		msgQFreeBuffer(sizeof(lrgmail_ts), (void *)msg_ps);
		DEBUGPRINT(DP_INFO, bsacAbortFlag, 
				 ("bsacAbortClearMeasBuffs: free a LARGE buffer for md=%d\n", md));
	  }
	}

	mdBuf_ps->bufs_as[buf].buf_ps = NULL;
    mdBuf_ps->bufs_as[buf].data_p = NULL;
	mdBuf_ps->bufs_as[buf].totwc  = 0;
	mdBuf_ps->bufs_as[buf].partwc = 0;
	mdBuf_ps->bufs_as[buf].cumpartwc = 0;
    
  }
  mdBuf_ps->bufCount = 0;
  mdBuf_ps->nextBuf = 0;
  return iss;
}

/*=============================================================================

  Name: bsacAbortAllActiveMDefs()

  Abs: abort all active MDefs - called when shutting down

  Args: none 

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
vmsstat_t bsacAbortAllActiveMDefs(void)
{
  short md = 0;
  long  nRequest=1;      /* these are for fetching Epics PV's */
  long  options=0;
  unsigned short ctrl=0;
  vmsstat_t iss = BPMO_OKOK;

  for (md = 0; md < MDEF_MAX; md++) {
	ctrl = 0;  /* need to initialize this each time ?*/
    /* check each measdef for ctrl=ON and abort=false */
	if ( (dbGetField(&mdefAddr_as[md].CTRL_s, DBR_ENUM, 
			   &ctrl, &options, &nRequest, NULL))) {
	  DEBUGPRINT(DP_INFO, bsacAbortFlag, 
		 ("bsacAbortMeasStop: Error for dbGetField of MDEF:LOCA:%d:CTRL\n", md+1)); 
	}
	else {
	  if ( (ctrl != 0) && (bsacMD_as[md].abort == epicsFalse) ) 
		iss = bsacAbortCancelMeas(md);
	}
  }
  return iss;
}
/*=============================================================================

  Name: bsacAbortActiveGOs()

  Abs:  reset the CTRL records associated with devices involved in aborted
        MDEF.

  Args: Type		        Name	    Access	    Description
	------------------- -----------	---------- ----------------------------
	const short         md          read only   mdef number of buffer set to clear

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status: BPMO_FAIL, or BPMO_OKOK

===============================================================================*/
LOCAL vmsstat_t bsacAbortActiveGOs (const short md) {
  int prim, unit;
  unsigned short ctrl = 0;
  vmsstat_t iss = BPMO_OKOK;

  DEBUGPRINT(DP_DEBUG, bsacAbortFlag,("bsacAbortActiveGOs: starting\n"));
  /* for now check devices BPMS, TOROs, ARRYs */
  for (prim=0; prim<BSACARRY_INDEX; prim++) {
	DEBUGPRINT(DP_DEBUG, bsacAbortFlag,("bsacAbortActiveGOs: checking prim #%d\n", prim));
	for (unit=0; unit<primCollect_as[prim].unitCnt; unit++) {
	  DEBUGPRINT(DP_DEBUG, bsacAbortFlag,("bsacAbortActiveGOs: checking unit #%d\n", unit+1));
	  if (primCollect_as[prim].unit_as[unit].mdef_as[md].included){
		DEBUGPRINT(DP_DEBUG, bsacAbortFlag,
				   ("bsacAbortActiveGOs: unit #%d included precord=%p \n", unit+1,
					primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord));
		if (primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord){
		  if (dbPutField(&primAddr_as[prim].unit_as[unit].mdefCTRL_as[md], 
						 DBR_USHORT, &ctrl, 1)!=0) 
		  {
			iss=BPMO_FAIL;
            if (slcGetLOCA2Name()!=NULL) {
			  DEBUGPRINT(DP_ERROR, bsacAbortFlag, 
					   ("bsacAbortActiveGOs: dbPutField of %.4s:%s(%s):%s:CTRLM%d failed\n", 
						(char *)primCollect_as[prim].slcPrim_a,
						slcGetLOCAName(),slcGetLOCA2Name(),
						(char *)primCollect_as[prim].unit_as[unit].slcUnit_a,md+1));
			} else {
			  DEBUGPRINT(DP_ERROR, bsacAbortFlag, 
					   ("bsacAbortActiveGOs: dbPutField of %.4s:%s:%s:CTRLM%d failed\n", 
						(char *)primCollect_as[prim].slcPrim_a,
						slcGetLOCAName(),
						(char *)primCollect_as[prim].unit_as[unit].slcUnit_a,md+1));
			}
		  }	else {
			if (slcGetLOCA2Name()!=NULL) {
			      DEBUGPRINT(DP_DEBUG, bsacAbortFlag, 
					   ("bsacAbortActiveGOs: dbPutField of %.4s:%s(%s):%s:CTRLM%d to 0\n", 
						(char *)primCollect_as[prim].slcPrim_a,
						slcGetLOCAName(),slcGetLOCA2Name(),
						(char *)primCollect_as[prim].unit_as[unit].slcUnit_a,md+1));
			} else {
			      DEBUGPRINT(DP_DEBUG, bsacAbortFlag, 
					   ("bsacAbortActiveGOs: dbPutField of %.4s:%s:%s:CTRLM%d to 0\n", 
						(char *)primCollect_as[prim].slcPrim_a,
						slcGetLOCAName(),
						(char *)primCollect_as[prim].unit_as[unit].slcUnit_a,md+1));
			}
		  }
		}
	  }
	}
  }
  /* do the ARRYs separately, they use the IOC name */
  prim=BSACARRY_INDEX;
  for (unit=0; unit<primCollect_as[prim].unitCnt; unit++) {
	if (primCollect_as[prim].unit_as[unit].mdef_as[md].included){
  	  if (primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord){
  	    if (dbPutField(&primAddr_as[prim].unit_as[unit].mdefCTRL_as[md], 
						 DBR_USHORT, &ctrl, 1)!=0) 
	    {
			iss=BPMO_FAIL;
			DEBUGPRINT(DP_ERROR, bsacAbortFlag, 
					   ("bsacAbortStop: dbPutField of %.4s:%s:%d:CTRLM%d failed\n", 
						(char *)primCollect_as[prim].slcPrim_a,
						slcGetIOCName(),unit+1,md+1));
	    } else DEBUGPRINT(DP_DEBUG, bsacAbortFlag, 
					   ("bsacAbortStop: dbPutField of %.4s:%s:%d:CTRLM%d to 0, p=%p\n", 
						(char *)primCollect_as[prim].slcPrim_a,
						slcGetIOCName(),unit+1,md+1,
						&primAddr_as[prim].unit_as[unit].mdefCTRL_as[md]));
	  }
	}
  }
  return iss;
}
