/*=============================================================================
 
  Name: bsacBufAcqData.c


  Abs:  bsacBufAcqData contains functions that handle data received from the EPICS
        bsac database. 
        Reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | BSAC Service Design

  Auth: 31 aug 2005, drogind    created from diane's bsacData.c 

  Rev:  

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest) 
        18 aug 2005, Dfairley    created 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int bsacBufAcqDataFlag = DP_DEBUG;
#endif


/* c includes */
#include <string.h>        /* memcpy */

/* epics */
#include <dbAccess.h>      /* EPICS dbAddr struct */

/* shared slc micro includes */
#include "bpmfunc.hc"      /* bpmo_xxx func codes */
#include "micrdef.hc"      /* MICR_EXNET_ codes */
#include "bpmodef.hc"      /* BPMO_OKOK etc */
#include "msgfunc.hc"      /* RESPONSE for func code */

/* slc global vars */
#include "slcGlob.h"       /* slcThreads_as */

/* slcIOC utilities */
#include "slcUtil.h"       /* slcPrintMsgheader */
#include "msgQUtil.h"      /* messageQSend, messageQRecv */
#include "cvtUtil.h"       /* slcSSwap, slcLSwap, etc */
#include "slcCmlog.h"      /* slcCmlogLogMsg */
#include "timeUtil.h"      /* timeGetVMS, timeDiffVMS10usec */


/* bsac includes */
#include    "bsacGlob.h"
#include    "bsacPrepTypes.h"
#include    "bsacPrep.h"
#include    "bsacData.h"
#include    "bsacBufAcqData.h"
#include    "bpmstruc.h"


/* static bsacBufAcqData file scoped vars */
LOCAL ws_arry_data_ts   tmpArryData_s;
/*debug */
LOCAL ws_puid_ts * puidTop_p = NULL;

/* local prototypes */
LOCAL void bsacBufAcqDataCvtMoveBPMSData( const ws_bpm_data_ts *bpms_ps,
										  unsigned char * buf_p);
LOCAL void bsacBufAcqDataCvtMoveTOROData( const ws_toro_data_ts *toro_ps,
										  unsigned char * buf_p);
LOCAL void bsacBufAcqDataSaveARRYChan(const unsigned short chanIndex, 
									  const arry_chan_data_ts *chan_ps,
									  ws_arry_data_ts *arry_ps);
LOCAL void bsacBufAcqDataCvtMoveARRYData( const ws_arry_data_ts *arry_ps,
										  unsigned char * buf_p);
LOCAL void bsacBufAcqDataCvtMovePUIDData( const bsacMD_ts *bsacMD_ps,
										  const ws_puid_data_ts *puid_ps,
										  unsigned char * buf_p);
LOCAL void bsacBufAcqDataCvtMoveScanblkPrefix (const unsigned short pI, 
											   const unsigned short uI, 
											   const unsigned short md,
											   unsigned char * buf_p);
/* debug routines */
LOCAL void bsacBufAcqDebugPrintBPM (const ws_bpm_data_ts * bpm_p, void * buf_p);
LOCAL void bsacBufAcqDebugPrintARRY (const ws_arry_data_ts * arry_p);
LOCAL void bsacBufAcqDebugPrintScanblk (const scanblk_prefix_ts * sb_p, void * buf_p);
LOCAL void bsacBufAcqDebugPrintAllPUIDs (const int npuid);
/*-----------------------------------------------------------------------
  Name: bsacBufAcqData

  Abs: move the data from this message into its proper place in the msg buffer

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const msgmail_ts* rMsg_ps read only  incoming data message

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacBufAcqData (const msgmail_ts* rMsg_ps)
{
  unsigned short pI, uI, cI, md; 
  unsigned long datasize=0;
  short bufIndex = 0;
  bsacMsgHdr_ts * hdr = NULL;
  unsigned char * indata_p = NULL;
  unsigned char * bufdata_p = NULL;
  vmsstat_t iss = BPMO_OKOK;
  /* debug */
  /*unsigned char * debbufdata_p = NULL;*/
  /* debug */
   
  /* interpret the header */
  hdr = (bsacMsgHdr_ts*)&rMsg_ps->reqdata;
  pI = hdr->primIndex;
  uI = hdr->unitIndex;
  cI = hdr->chanIndex;
  if (cI > 0) cI -= 1; 
  /* EPICS bsacXL*.substitutions has ARRY channels 1-12*/
  /* incoming cI channel number of 0 means no ARRY is involved with this device */
  /*  so subtract 1 only if there are channels involved w this device */

  md = hdr->md-1; /* the EPICS md is one more than slc threads md */

  /* keep this data if we are not aborting this md */
  if (bsacMD_as[md].abort!=epicsTrue)
	{

	  /* special case - ARRY data is stored in a structure of 
		 12 channels, but not all of the channels are used in a 
		 particular IOC/md.  primCollect....chanCntr will count down
		 the expected number of channel data messages for each measured
		 pulse, and the channel data is stored in a local static temp 
		 structure until all channels data is sent for a pulse. 
		 HERE - we keep track of channel count for arry data
		 if chanCntr=-1 we start the counter again by resetting to chanCnt;
		 when resetting the counter, we also initialize the tmpArryData_s
	  */
	  if ( primCollect_as[pI].unit_as[uI].chanCntr==-1) {
		primCollect_as[pI].unit_as[uI].chanCntr = primCollect_as[pI].unit_as[uI].chanCnt; 
		memset(&tmpArryData_s, 0, sizeof(tmpArryData_s));
	  }
	  
	  /* get pointer to outgoing bufdata_p */
	  bufIndex = primCollect_as[pI].unit_as[uI].mdef_as[md].bufIndex;
	  bufdata_p = bsacMDBufs_as[md].bufs_as[bufIndex].data_p;
	  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag,
				   ("bsacBufAcqData: Buffer ptr =%p for pI=%d, uI=%d; offset = %ld, bufIndex = %hd\n", 
					bufdata_p, pI, uI,
					primCollect_as[pI].unit_as[uI].mdef_as[md].offset, bufIndex )); 
	  if (bufdata_p){
		bufdata_p+= primCollect_as[pI].unit_as[uI].mdef_as[md].offset;
	  }
	  /* if first piece of data for this unit, output debug statement*/
	  else {
		DEBUGPRINT(DP_ERROR, bsacBufAcqDataFlag, 
			  ("bsacBufAcqData:received data from pI=%d, md=%d, uI=%d, cI=%d but it has no lrg msg buffer!\n",
					pI, md, uI, cI ));
		return iss;
	  }
	  
	  /* get a pointer to the data part of the incoming data msg */
	  indata_p = (unsigned char *)hdr + sizeof(bsacMsgHdr_ts);
	  
	  /* if first piece of data for this unit, put in prefix header */
	  if (primCollect_as[pI].unit_as[uI].mdef_as[md].counter <= 0) {
		DEBUGPRINT(DP_INFO , bsacBufAcqDataFlag,
				   ("bsacBufAcqData: Insert scanblk hdr. Buf ptr = %p, offset = %ld pI=%d, uI=%d; count = %ld;\n",  bsacMDBufs_as[md].bufs_as[bufIndex].data_p, primCollect_as[pI].unit_as[uI].mdef_as[md].offset, pI, uI,primCollect_as[pI].unit_as[uI].mdef_as[md].counter)); 
		bsacBufAcqDataCvtMoveScanblkPrefix (pI, uI, md, (void *)bufdata_p);
		datasize = sizeof(scanblk_prefix_ts);
		primCollect_as[pI].unit_as[uI].mdef_as[md].offset+=datasize;
		bufdata_p += datasize;
	  }	else DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag,
				   ("bsacBufAcqData: COUNTER for scanblk header for pI=%d, uI=%d, = %ld\n", pI, uI, primCollect_as[pI].unit_as[uI].mdef_as[md].counter));
  	
	  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag,
				 ("bsacBufAcqData: Buffer ptr for pI=%d, uI=%d = %p; count = %ld; offset = %ld\n", 
				  pI, uI, bufdata_p,primCollect_as[pI].unit_as[uI].mdef_as[md].counter,
				  primCollect_as[pI].unit_as[uI].mdef_as[md].offset )); 	  
	  /* switch on primary and cvt/move data into lrg msg buffer */
	  switch (pI){
	  case BSACBPMS_INDEX: 
		DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, 
				   ("bsacBufAcqData:received BPMS data from md=%d where uI=%d, cI=%d; offset = %ld \n",
					md, uI, cI, primCollect_as[pI].unit_as[uI].mdef_as[md].offset )); 
		/* convert and move the data to outgoing large msg buffer */
		/* scale simulated tmit data 1. E+9; simulated data is in nC  */
	    /*bsacBufAcqDebugPrintBPM ((const ws_bpm_data_ts *)indata_p, bufdata_p);*/
		  bsacBufAcqDataCvtMoveBPMSData((const ws_bpm_data_ts *)indata_p, (void *)bufdata_p);
		datasize = WS_BPM_DATA_SIZE;
		break;
	  case BSACTORO_INDEX: 
		DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
				   ("bsacBufAcqData:received TORO tmit data=%f; stat=%hd; goodmeas=%hd, from md=%d where uI=%d, cI=%d\n",
					( (ws_toro_data_ts *)indata_p)->tmit, 
					( (ws_toro_data_ts *)indata_p)->stat,
					( (ws_toro_data_ts *)indata_p)->goodmeas,  
					md, uI, cI )); 
		datasize = WS_TORO_DATA_SIZE;
		
		/* convert and move the data to outgoing large msg buffer */
		/* scale simulated tmit data 1. E+9; simulated data is in nC  */
/*		( (ws_toro_data_ts *)indata_p)->tmit *=1000000000.0;*/
		bsacBufAcqDataCvtMoveTOROData((const ws_toro_data_ts *)indata_p, (void *)bufdata_p);
		break;
	  case BSACARRY_INDEX: 
		/* special case - ARRY data is stored in a structure of 12 channels, but not
		   all of the channels are used in a particular IOC/md.  primCollect....chanCntr
		   will count down the expected number of channel data messages for each measured
		   pulse, and the channel data is stored in a local static temp structure, until
		   all channels data is sent for the pulse */
		DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
				   ("bsacBufAcqData:received ARRY data from md=%d where uI=%d, cI=%d\n",
					md, uI, cI ));    
		if ( primCollect_as[pI].unit_as[uI].chanCntr>0) {
		  /* collect up ALL the channels for this md/navg before moving to lrg buffer */
		  /* chanCntr counts down the expected number of chans for this pulse measurment*/
		  primCollect_as[pI].unit_as[uI].chanCntr--;
		  /* now copy this measurement to the temporary structure */
		  bsacBufAcqDataSaveARRYChan(cI, (const arry_chan_data_ts *)indata_p, &tmpArryData_s);
		  /* this data was not moved into the buffer, do not add a datasize to the offset */
		  datasize=0;
		  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
					 ("bsacBufAcqData: stored ARRY data from md=%d where uI=%d, cI=%d\n",
					  md, uI, cI ));    
		  if ( primCollect_as[pI].unit_as[uI].chanCntr==0) {	
			/* copy this data to the temp structure */
			bsacBufAcqDataSaveARRYChan(cI, (const arry_chan_data_ts *)indata_p, &tmpArryData_s);
		    /*bsacBufAcqDebugPrintARRY ((const ws_arry_data_ts *) &tmpArryData_s);*/
			/* convert and move the temp structure data to outgoing large msg buffer */
			bsacBufAcqDataCvtMoveARRYData((const ws_arry_data_ts *)&tmpArryData_s, (void *)bufdata_p);

			datasize = WS_ARRY_DATA_SIZE;
			DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
					   ("bsacBufAcqData: stored ARRY data from md=%d where uI=%d, on last cI=%d\n",
						md, uI, cI ));
			/* reset unit's chanCntr for next time */
			primCollect_as[pI].unit_as[uI].chanCntr=-1;
			DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
					   ("bsacData: stored ARRY data from md=%d where uI=%d, on last cI=%d\n",
						md, uI, cI ));
		  }    
		} else {
		  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
					 ("bsacBufAcqData:ERROR!! received too many ARRY data from md=%d where uI=%d, cI=%d\n",
					  md, uI, cI ));    
		} 
		break;
	  case BSACPUID_INDEX: 
		DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
				   ("bsacBufAcqData:received PUID data from md=%d where  uI=%d, cI=%d\n",
					md, uI, cI ));    

		/* convert and move the data to outgoing large msg buffer */
		/* special case - we need to know when the first pulse of a meas. has
		   occurred, so that we can insert the correct PUID hdr data*/
		bsacBufAcqDataCvtMovePUIDData( &bsacMD_as[md], 
									   (const ws_puid_data_ts *)indata_p, 
									   (void *)bufdata_p);
		/* if this is the first pulse of this measurement we must reset the flag */
		if (bsacMD_as[md].firstPulse) {
		  bsacMD_as[md].firstPulse=epicsFalse;
		  datasize = WS_PUID_DATA_SIZE;
		}
		else datasize = sizeof (unsigned long);
		break;
	  case BSACWIRE_INDEX: 
		DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
				   ("bsacBufAcqData:ERROR!!!received WIRE data from md=%d where  uI=%d, cI=%d\n",
					md, uI, cI ));    
		datasize = 0;
		break;
	  default:
		DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
				   ("bsacBufAcqData:ERROR!!!received UNKOWN data from md=%d where pI=%d, uI=%d, cI=%d\n",
					md, pI, uI, cI ));
		datasize = 0;
		break;
	  }
	  
	  /* update how many messages this individual unit has sent from epics */
	  /* this counter (when zero) indicates to store scanblk header */
	  primCollect_as[pI].unit_as[uI].mdef_as[md].counter++;	  
	  /* update the msg counter and offset */
	  primCollect_as[pI].unit_as[uI].mdef_as[md].offset+=datasize;
	  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag,
				 ("bsacBufAcqData: Exiting offset for pI=%d, uI=%d = %ld; count = %ld\n", 
				  pI, uI,primCollect_as[pI].unit_as[uI].mdef_as[md].offset,
				  primCollect_as[pI].unit_as[uI].mdef_as[md].counter )); 
	  if (bsacMD_as[md].msgCounter>0) bsacMD_as[md].msgCounter--;
	  else DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
					  ("bsacBufAcqData:ERROR!!!too many data messages for  md=%d \n",hdr->md-1));
	  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
				 ("bsacBufAcqData: msgCounter = %ld after handling %s data\n",  
				  bsacMD_as[md].msgCounter, (char *)primCollect_as[pI].slcPrim_a));
	}
  
  /* update diagnostics for data measurement */
  return(BPMO_OKOK);
}


/*-----------------------------------------------------------------------
  Name: bsacBufAcqDone

  Abs: move this last data from this message into its proper
       place in the msg buffer. Measurement is done, prepare to send.

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const msgmail_ts* rMsg_ps read only  incoming data message

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacBufAcqDone (const msgmail_ts* rMsg_ps)
{
  unsigned short md;
  bsacMsgHdr_ts * hdr = NULL;
  vmsstat_t iss = BPMO_OKOK;
  
  /* interpret the header */
  hdr = (bsacMsgHdr_ts*)&rMsg_ps->reqdata;
  md = hdr->md-1; /* epics records use md+1 */
  
  /* store the incoming data */
  iss = bsacBufAcqData(rMsg_ps);

  /* handle done stuff */
  if (bsacMD_as[md].msgCounter == 0 ){
	/*DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
	("bsacBufAcqDone: printing PUIDS for md=%d, navg=%hd, nrpos =%hd\n", 
	md, bsacMD_as[md].navg, bsacMD_as[md].nrpos));*/
	/*bsacBufAcqDebugPrintAllPUIDs (bsacMD_as[md].nrpos * bsacMD_as[md].navg);*/

	/* send the first lrgbuffer to the scp */
	DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
	("bsacBufAcqDone: sending data message for md=%d\n", md));
	iss = bsacDataSendFirst(md);
  }
  else
	DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
	   ("bsacBufAcqData:not enough data messages yet for  md=%d \n",md));

     
  /* update diagnostics for done measurement */
  return(iss);
}

/*-----------------------------------------------------------------------
  Name: bsacBufAcqDataCvtMoveBPMSData

  Abs: store this packet of BPMS data into the large data buffer

  Args: Type		         Name    Access	    Description
	    -------------------- ------- ---------- ----------------------------
		const bpms_data_ts*  bpms_ps read only  incoming data message
        unsigned char *      buf_p   read/write pointer to lrgmsg buffer area for bpms

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacBufAcqDataCvtMoveBPMSData( const ws_bpm_data_ts *bpms_ps,
										  unsigned char * buf_p)
{
  /* bpms float values */
  cvtFloatIEEEToVMS((const void *)&bpms_ps->x, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->x);
  cvtFloatIEEEToVMS((const void *)&bpms_ps->y, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->y);
  cvtFloatIEEEToVMS((const void *)&bpms_ps->tmit, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->tmit);
  cvtSSwapMove((const void *)&bpms_ps->stat, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->stat);
  cvtSSwapMove((const void *)&bpms_ps->goodmeas, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->goodmeas);
}

/*-----------------------------------------------------------------------
  Name: bsacBufAcqDataCvtMoveTOROData

  Abs: store this packet of TORO data into the large data buffer

  Args: Type		         Name    Access	    Description
	    -------------------- ------- ---------- ----------------------------
		const toro_data_ts*  toro_ps read only  incoming data message
        unsigned char *      buf_p   read/write pointer to lrgmsg buffer area for bpms


  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacBufAcqDataCvtMoveTOROData( const ws_toro_data_ts *toro_ps,
					  unsigned char * buf_p)
{
  /* toro float values */
  cvtFloatIEEEToVMS((const void *)&toro_ps->tmit, (void *)buf_p);
  buf_p+=sizeof(toro_ps->tmit);
  /* toro short values */
  cvtSSwapMove((const void *)&toro_ps->stat, (void *)buf_p);
  buf_p+=sizeof(toro_ps->stat);
  cvtSSwapMove((const void *)&toro_ps->goodmeas, (void *)buf_p);
  buf_p+=sizeof(toro_ps->goodmeas);
}

/*-----------------------------------------------------------------------
  Name: bsacBufAcqDataSaveARRYChan

  Abs: store this packet of arry channel data into a temp arry channel struct

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const msgmail_ts* rMsg_ps read only  incoming data message

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacBufAcqDataSaveARRYChan( const unsigned short chanIndex, 
								 const arry_chan_data_ts *chan_ps,
								 ws_arry_data_ts *arry_ps)
{
  arry_ps->data[chanIndex] = chan_ps->data;
  arry_ps->stat[chanIndex] = chan_ps->stat;
  arry_ps->goodmeas[chanIndex] = chan_ps->goodmeas;
}

/*-----------------------------------------------------------------------
  Name: bsacBufAcqDataCvtMoveARRYData

  Abs: store this packet of arry data into the large data buffer

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const arry_data_ts*  arry_ps read only  incoming data message
        unsigned char *      buf_p   read/write pointer to lrgmsg buffer area for bpms

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacBufAcqDataCvtMoveARRYData( const ws_arry_data_ts *arry_ps,
										 unsigned char * buf_p)
{
  unsigned short cnt = 0;

  /* arry float values */
  for (cnt = 0; cnt<MAX_ARRY_CHAN; cnt++) {
	cvtFloatIEEEToVMS((const void *)&arry_ps->data[cnt], (void *)buf_p);
	buf_p+=sizeof(float);  
  }
  for (cnt = 0; cnt<MAX_ARRY_CHAN; cnt++) {
	cvtSSwapMove((const void *)&arry_ps->stat[cnt], (void *)buf_p);
	buf_p+=sizeof(unsigned short);  
  }
  for (cnt = 0; cnt<MAX_ARRY_CHAN; cnt++) {
	cvtSSwapMove((const void *)&arry_ps->goodmeas[cnt], (void *)buf_p);
	buf_p+=sizeof(unsigned short);  
  }
}

/*-----------------------------------------------------------------------
  Name: bsacBufAcqDataCvtMovePUIDData

  Abs: store this packet of pulse id data into the large data buffer

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const bpm_pulseid_data_ts*  puid_ps read only  incoming data message
        unsigned char *      buf_p   read/write pointer to lrgmsg buffer area for bpms

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacBufAcqDataCvtMovePUIDData(const bsacMD_ts *bsacMD_ps, 
										 const ws_puid_data_ts *puid_ps,
										 unsigned char * buf_p)
{
  vmsTimeStamp_ta vmsTime;
  unsigned short nmuxmin = 1;

  /* puid values */
  /* if this is the very first pulse we have to insert a header */
  if (bsacMD_ps->firstPulse==epicsTrue) {
    /* debug - grab ptr to buffer at which ws_puid_ts starts (lk at on linux) */
	puidTop_p =  (ws_puid_ts *)buf_p;
	/*  setting navg */
	DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
	 ("bsacBufAcqDataCvtMovePUIDData: first pulse with navg=%d, nmuxmin=%d, buf_p = %p\n",
	  bsacMD_ps->navg, nmuxmin, buf_p ) )  ;
	cvtSSwapMove((const void *)&bsacMD_ps->navg, (void *)buf_p);
	buf_p+=sizeof(unsigned short);
	/*setting nmuxmin to one */
	cvtSSwapMove((const void *)&nmuxmin, (void *)buf_p);
	buf_p+=sizeof(unsigned short);
	/* set the timestamp */
    if ((timeGetVMS(vmsTime))!= epicsTimeERROR) {
	  memcpy((void *)buf_p, &vmsTime,8);
	}
	else{ 
	  memset((void *)buf_p, 0, 8);
	  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, 
				 ("bsacBufAcqDataCvtMovePUIDData: continuing pulse with ass'ed navg=%d, nmuxmin=%d, buf_p = %p \n",
	  bsacMD_ps->navg, nmuxmin, buf_p ) )  ;
	}
	buf_p+=sizeof(vmsTime);
  }
  /* now copy in the pulseid */
  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, ("bsacBufAcqDataCvtMovePUIDData:\n"));  
  DEBUGPRINT(DP_DEBUG, bsacBufAcqDataFlag, ("\tPulse id = %ld, buf_p =%p \n", puid_ps->pulseid, buf_p));  
  cvtLSwapMove((const void *)&puid_ps->pulseid, (void *)buf_p);
  buf_p+=sizeof(unsigned long);
}

/*-----------------------------------------------------------------------
  Name: bsacBufAcqDataCvtMoveScanblkPrefix

  Abs: store scanblk_prefix_ts for this primary, unit into the large data buffer

  Args: Type		      Name    Access	    Description
  ----------------- ------- ---------- ----------------------------
  const unsigned short        pI     read only    primary index into primCollect
  const unsigned short        uI     read only    unit index into primCollect 
  const unsigned short        md     read only    measdef number
  unsigned char *            buf_p   read/write   pointer to lrgmsg buffer area for bpms

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/

LOCAL void bsacBufAcqDataCvtMoveScanblkPrefix (const unsigned short pI, 
					       const unsigned short uI, 
					       const unsigned short md,
					       unsigned char * buf_p)
/*
based on units involved in mdef, calculates wc for this unit and places into scanblk_prefix_ts
Each unit has its own buffer pointer (index) and .offset per measdef
   it starts w scanblk_prefix, then occupies nrpos * sizeof(datasize)
The scanblk_prefix is copied into the large msg buffer starting at the unit's initial offset location
 then advance offset : .offset += sizeof (scanblk_prefix_ts)
  wc = prim data size * nrpos

*/
{
  unsigned long datasize, bc =0;
  scanblk_prefix_ts prefix_s;
  char prim_c4[4];

  memcpy(prim_c4, primCollect_as[pI].slcPrim_a, 4);
  prefix_s.prim = *(long *)prim_c4;
  prefix_s.unit = primCollect_as[pI].unit_as[uI].unitnum;
  /* calculate wc */
  switch (pI){
  case BSACBPMS_INDEX: 
    datasize = WS_BPM_DATA_SIZE;
    break;
  case BSACTORO_INDEX: 
    datasize = WS_TORO_DATA_SIZE;
    break;
  case BSACARRY_INDEX: 
    datasize = WS_ARRY_DATA_SIZE;
    break;
  case BSACPUID_INDEX: 
    datasize = sizeof(unsigned long);
	prefix_s.unit=1;
    break;
  default:
    datasize = 0;
    break;
  }
  bc = datasize * bsacMD_as[md].nrpos;
  if (pI == BSACPUID_INDEX){
    /* adjust for additional pulse id header */
    bc -= datasize;
    bc += WS_PUID_DATA_SIZE;
  }
  prefix_s.wc = (unsigned short) (bc/2);

  /*bsacBufAcqDebugPrintScanblk (&prefix_s, buf_p);*/

  /*cvtLSwapMove((const void *)&prefix_s.prim, (void *)buf_p);*/
  memcpy ((void *)buf_p, &prefix_s.prim,  4);
  buf_p+=sizeof(prefix_s.prim);
  cvtSSwapMove((const void *)&prefix_s.unit, (void *)buf_p);
  buf_p+=sizeof(prefix_s.unit);
  cvtSSwapMove((const void *)&prefix_s.wc, (void *)buf_p);
  buf_p+=sizeof(prefix_s.wc);

  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("bsacBufAcqDataCvtMoveScanblkPrefix exiting pI=%hd, uI=%hd, bc = %lu; wc=%u\n", pI, uI, bc, prefix_s.wc)); 
  return;
}

/* debug routines */
LOCAL void bsacBufAcqDebugPrintBPM (const ws_bpm_data_ts * bpm_p, void * buf_p)
{
  ws_bpm_data_ts bpmData_s;
  memcpy (&bpmData_s, bpm_p, sizeof(bpmData_s));

  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("bsacBufAcqDataDebugPrintBPM at buf_p = %p: \n", buf_p));  
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tX: %f \n", bpmData_s.x)); 
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tY: %f \n", bpmData_s.y));   
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tTMIT: %f \n", bpmData_s.tmit)); 
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tgoodmeas: %hd \n", bpmData_s.goodmeas)); 
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tstat: 0x%hx \n", bpmData_s.stat));  
  return;
}  

LOCAL void bsacBufAcqDebugPrintARRY (const ws_arry_data_ts * arry_p)
{
  ws_arry_data_ts arryData_s;
  int cnt;

  memcpy (&arryData_s, arry_p, sizeof(arryData_s));

  for (cnt=0; cnt<MAX_ARRY_CHAN; cnt++){
	DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("bsacBufAcqDataDebugPrintARRY: \n"));  
	DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tdata[%d]: %f \n", cnt, arryData_s.data[cnt]));  
	DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tgoodmeas[%d]: %hd \n", cnt, arryData_s.goodmeas[cnt])); 
	DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tstat[%d]: 0x%hx \n", cnt, arryData_s.stat[cnt])); 
  }  
  return;
}

LOCAL void bsacBufAcqDebugPrintScanblk (const scanblk_prefix_ts * sb_p, void * buf_p)
			
{
  char primName[5];

  memcpy (&primName, &sb_p->prim, 4);
  primName[4]='\0';
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("bsacBufAcqDataDebugPrintScanblk at buf_p = %p:\n", buf_p));  
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tprim: %s \n", (char *)primName));  
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tunit: %u \n", sb_p->unit)); 
  DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\twc: %u \n", sb_p->wc)); 
  
  return;
}

LOCAL void bsacBufAcqDebugPrintAllPUIDs (const int npuid)
	 /* uses debug global puidTop_p, which poitns to top of ws_puid_ts in outgoing buffer */
     /*  thus, look at on linux */
{
 int cnt;
 unsigned long * id_p = NULL;

 id_p = &puidTop_p->pulseid[0];
 DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("bsacBufAcqDataDebugPrintAllPUID (%d): \n", npuid));
 DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tnavg: %hd \n", puidTop_p->navg));  
 DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tnmuxmin: %hd \n", puidTop_p->nmuxmin));
 DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\ttstamp[0]: 0x%lx \n", puidTop_p->tstamp[0]));
 DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\ttstamp[1]: 0x%lx \n", puidTop_p->tstamp[1]));
 for (cnt=0;cnt<npuid;cnt++) {
   DEBUGPRINT(DP_INFO, bsacBufAcqDataFlag, ("\tpulseid[%d]: %ld\n", cnt, *id_p));
   id_p++;
 }
}
