/*=============================================================================
 
  Name: bsacCal.c


  Abs:  bsacCal performs the message responses for BPMO_CALB_PREP, namely
        a)(BPMO_CALB_PREP | RESPONSE immed reply) and b) BPMO_CALB_PROC data reply.
		Background: If a cal table is not present for a particular DGRP prior 
		to a measurement request, the SCP complains and will not perform the meas. 		
		Thus, for purposes of making the SCP content enough to perform the meas,
		the slc aware ioc sends dummy calibration data back. 
		(Calibration is not to be controlled by the SCP for LCLS.)
		The SCP needs a calibration table present for each DGRP.  
		Even though the micro is the only one who normally uses cal table data,
        (SCP downloads it to micro from disk when necessary), the SCP  looks 
		at the stat data on file for BPMP to see if it was good. 
		Since BPMP and GADC devices are handled by the same (BPMO_CALB_PREP) 
		command, data for these devices must be handled & faked.

        Reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | BSAC Service Design

  Auth: 28 jul 2005, dROGIND    created 
  Rev:  10 Aug 2005, drogind
        found when cal'ing multiple iocs, a delay between immed response and
        dummy data reply was necessary. Mike thought 2 sec would be sufficient.

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest) 
        20 jul 2005, DROGIND    created 
=============================================================================*/
/* Notes on SCP BPM Cal Operations */

/*
 It is necessary for the SCP to have cal tables (either private to that SCP
 or public for all SCPS)associated with each LCLS BPMD before it allows 
 any type of meas def. Since cal will not be controlled via the SCP for LCLS,
 the slc ioc sends dummy cal tables to SCP to fake it out and allow 
 measurements to proceed. Since the SCP uses itself the tables for bpms (looks
 at the stat bits returned during cal), the slc ioc sends back stat of GOOD in
 the data reply for bpms for the BPMO_CALB_PREP; all other data is set to "1".
 GADCs are also calibrated via the BPMO_CALB_PREP, so dummy data is sent for them.
 Toroids are calibrated via a completely different message, BPMO_TORCAL_PREP?, 
 which the SCP itself doesn't consult, so dummy tables were not necessary in 
 response to this message.

 To Calibrate:
 1) Make sure you have the desired iocs in the LCLS DGRP running.
 Currently the LCLS DGRP is XL01-XL05. See below to change the DGRP.
 2) For each ioc, ensure that all of its GADC, DUGADC, BPMP and BPMS
 units are online; or for BPMS, in MAINTENANCE.
 To do this, go to BPMS DIAG Panel
 Select BPMP, type in XL0x unit return
 BPMP HSTA ; type in reason, ie "LCLS testing", OK
 Select BPMS, type in XL0x unit (1) return
   select MAINTENANCE (was off-line); reason, ie "LCLS testing", OK
 Cycle through all the BPMS belonging to that BPMP, if more than 1:
 Select BPMS, next (return) repeat MAINTENANCE, reason (up arrow repeats previous)
 Cycle through all the BPMPs/associated BPMS for this ioc
 Select GADC, type in XL0x unit (1) return
 GADC HSTA; type in reason (up arrow) return
 Select DUGADC, select ARRY return; enter MICRO XL0x, unit return, etc
 Select ARRY HSTA... 

 3)To initiate cal for all running iocs: 
 Go to LCLS Index | BSAC Panel |:
  Choose BPMD LCLS on MCCDEV
  Create Meas Def, click on empty button; look at MESSAGES_DEV panel 

 4) To view (this private) calibration on GRAPHICS_DEV :
  Index | Pep II Index | BPM | BPM Cal | Cal Dsply; Disply Next Page
 
 5) To make public, ACCEPT PUBLIC
 This makes possible future meas defs for all iocs involved in cal proceed without
 first calibrating (for every new SCP).

 To remove public file on mccdev, at mccdev prompt, type Delete slcbpm:calb$52.*;*
 To remove public file on mcc, at mcc prompt, type Delete slcbpm:calb$51.*;*

 To change the DGRP:
To make XL01 the only micro in the DGRP
MCCDEV> dbedit xl01$dgrp.dbs
DATABASE EDIT

MIKE, Why are You doing this?
Data:   LCLS testing
EDIT FILE: USER_DISK_SLC:[MIKE]XL01$DGRP.DBS; (Y/N)? [Y]:
DEFAULTS FILE: REF_:[DBSFILE]DEFAULTS.DBS; (Y/N)? [Y]:

    Checking the input file for forbidden database edits!

DBEDIT which database (MCCDEV,MCC)? [MCCDEV,MCC]: MCCDEV

[snip]

To restore the full DGRP
MCCDEV> dbedit xlall$dgrp.dbs
DATABASE EDIT

MIKE, Why are You doing this?
Data:   LCLS testing
EDIT FILE: USER_DISK_SLC:[MIKE]XLALL$DGRP.DBS; (Y/N)? [Y]:
DEFAULTS FILE: REF_:[DBSFILE]DEFAULTS.DBS; (Y/N)? [Y]:

    Checking the input file for forbidden database edits!

DBEDIT which database (MCCDEV,MCC)? [MCCDEV,MCC]: MCCDEV

 [snip]

NOTE: The files xl01$dgrp.dbs and xlall$dgrp.dbs exist in your home directory on VMS.

 To turn micro hstas online/offline ie Mike:
   at mccdev prompt:
   sethsta
   Primary: CSTR
   Secondary: HSTA
   Micro: XL0...
   set mask "1" to turn online (happy?)
   clear mask "1" to turn offline (?)

*/
 
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int bsacCalFlag = DP_ERROR;
#endif

/* c includes */
#include <string.h>        /* memcpy */

/* shared slc micro includes */
#include "bpmfunc.hc"      /* bpmo_xxx func codes */
#include "bpmodef.hc"      /* BPMO_OKOK */
#include "micrdef.hc"      /* MICR_EXNET_ codes */
#include "slcMsg.h"        /* IOC_STOP, BSAC_* */
#include "msgfunc.hc"      /* RESPONSE for func code */

/* slc global vars */
#include "slcGlob.h"       /* slcThreads_as */

/* slcIOC types */
#include "msgTypes.h"      /* msgmail_ts etc */
/* slcIOC utilities */
#include "slcUtil.h"       /* slcPrintMsgheader */
#include "msgQUtil.h"      /* messageQSend, messageQRecv */
#include "cvtUtil.h"       /* slcSSwap, slcLSwap, etc */
#include "slcCmlog.h"      /* slcCmlogLogMsg */
#include "timeUtil.h"      /* timeGetVMS, timeDiffVMS10usec */

/* slcMain function prototypes */
#include "slcMain.h"       /* slcStop(), slcRestart() */

/* bsac includes */
#include    "bsacGlob.h"
#include    "bsacPrep.h"
#include    "bsacHdlr.h"   /* bsacHdlrReplyPrologue */
#include    "bpm_reqreply_struc.h"
#include    "bpmstruc.h"
#include    "bpmstat.h"

/* local structures */
typedef struct {
  unsigned short calbbits;                 /* calbbits from prep msg */
  unsigned short umin, umax, umin2, umax2; /* unit ranges, if not using ulist */
} calPrep_ts;
/* prototypes */

LOCAL vmsstat_t bsacCalPrepParseMsg (const msgmail_ts * rMsg_ps);
LOCAL vmsstat_t bsacCalPrepImmedReply(const msgmail_ts * rMsg_ps,
								   const vmsstat_t iss,
								   const unsigned short npulstep);
LOCAL vmsstat_t bsacCalProc (const msgmail_ts * rMsg_ps);
LOCAL void bsacCalFillData (msgmail_ts * sMsg_ps);
void bsacComputeChecksum (msgmail_ts * sMsg_ps); /*deprecated*/

/* file scoped globals  */

/* keeps data response header for PROC data reply */
/* this structure is longword aligned like VMS */
LOCAL bpm_alldata_replybeg_ts dataReply_s;
LOCAL calPrep_ts calPrep_s;
/*=============================================================================

  Name: bsacCalPrep

  Abs:  First calls a procedure to parse incoming BPMO_CALB_PREP message.
        Then Responds to  BPMO_CALB_PREP message by calling procedures to:
        1)(BPMO_CALB_PREP | RESPONSE) immediate reply - bsacCalPrepImmedReply
		and 2) (BPMO_CALB_PROC | RESPONSE) data reply - bsacCalProc.


  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		msgmail_ts          rMsg_ps     read      pointer to the received message


  Rem:  First parse through the message to see what primaries (calbbits)
        are of interest. It turns out all units configured in this ioc for 
		BPMPs and GADCs are of intereset if their correpsonding Primary bit is 
		set in calbbits field.

  side:

  Ret:  none

==============================================================================*/
vmsstat_t bsacCalPrep (const msgmail_ts * rMsg_ps)
{
  vmsstat_t iss = BPMO_OKOK; 

  /* fill in data reply header for data reply (comes after immed reply) */
  iss = bsacCalPrepParseMsg (rMsg_ps);

  /* create and send immediate data reply */
  /* first try to send back totwc= 0 and see if this satisfies SCP */
  iss = bsacCalPrepImmedReply(rMsg_ps, iss, 2 );
  /* note - in calbprep.c86, comment beg "NPULSTEP=", set =2 to be safe */

  /* data response */
  iss = bsacCalProc(rMsg_ps);

  return (iss);
}
/*=============================================================================

  Name: bsacCalPrepParseMsg

  Abs:  This routine unpacks the BPMO_CALB_PREP message by writing into
        global variables calPrep_s and dataReply_s.

		It turns out the only field of use to us at present time is 
		the calbbits field.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		msgmail_ts          rMsg_ps     read      pointer to the received message
  Rem: 

  side: Sets up global static var dataReply_s and calPrep_s (which contains
        umin-umax, umin2-umax2 if ever required).
		dataReply_s is set up with field .stat = BPMO_OKOK and stores the
		.calbbits field appropriately

  Ret:  none

==============================================================================*/
LOCAL vmsstat_t bsacCalPrepParseMsg (const msgmail_ts * rMsg_ps)
{         
  unsigned long calreplybc = 0; 

  /* parse incoming msg for calbbits field */
  /*  and store in data reply hdr structure for later data response */
  vmsstat_t iss = MICR_OKOK;
  unsigned char * reqdata_ps; /* byte ptr to data portion of incoming rMsg */

  DEBUGPRINT(DP_INFO, bsacCalFlag, ("bsacCalPrepParseMsg: Entering..\n")); 
  /* slog through the reqdata & extract calbbits from */
  /*   bpm_calbprep_req_ts; */
  /* point to top of data */  
  reqdata_ps = (unsigned char *)&rMsg_ps->reqdata;
  /* point to calbbits */
  reqdata_ps += sizeof(unsigned short) * 3;
  reqdata_ps += sizeof(unsigned char) * (BPM_MAXPP + 8) ;
   
  /* convert(source, dest) & copy to calbbits*/
  cvtLSwapMove((const void *)reqdata_ps, (void *)&(dataReply_s.bt.calbbits));
  reqdata_ps += sizeof (unsigned long);
  DEBUGPRINT(DP_INFO, bsacCalFlag, ("bsacCalPrepParseMsg: calbbits = 0x%lx\n", 
									dataReply_s.bt.calbbits));
  calPrep_s.calbbits = dataReply_s.bt.calbbits;
  
  /* convert(source, dest) & copy to umin */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&(calPrep_s.umin));
  reqdata_ps += sizeof (unsigned short);
  /* convert(source, dest) & copy to umax */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&(calPrep_s.umax));
  reqdata_ps += sizeof (unsigned short);
  /* convert(source, dest) & copy to umin2 */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&(calPrep_s.umin2));
  reqdata_ps += sizeof (unsigned short);
  /* convert(source, dest) & copy to umax2 */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&(calPrep_s.umax2));
  reqdata_ps += sizeof (unsigned short); 

  dataReply_s.iss = BPMO_OKOK;

  /* Now calculate wc reply size */
  /*  see calbproc.c86 - Calbp_compute_replysize() for format of returned data */
  /* assume non self-ided data */
  /*calreplybc = sizeof(bpm_alldata_replybeg_ts) + sizeof(bpmp_calbreply_ts) +
	sizeof(gadc_calbreply_ts); - the bpmp and gadc strucs are not packed like VMS*/
  calreplybc = sizeof(bpm_alldata_replybeg_ts);
  DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("bsacCalPrep - (alldata)reply bc = %ld\n", calreplybc) );
  if (dataReply_s.bt.calbbits & CALB_BPMP_BIT){
	/* cycle through slc ioc units to see if included in umin-umax or umin2-umax2 */
	calreplybc += CALB_BPMPUNION_REPLY_BC * calPrimCnt[BSACBPMS_INDEX];
	DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("bsacCalPrep - (+bpms)reply bc = %ld\n", calreplybc) );
  }
  if (dataReply_s.bt.calbbits & CALB_GADC_BIT){
	calreplybc += CALB_GADC_REPLY_BC * calPrimCnt[BSACGADC_INDEX];
	DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("bsacCalPrep - (+gadc)reply bc = %ld\n", calreplybc) );
  }
  /* toroid cal data is totally different and doesn'tneed to be faked out */
  /* for an ioc w only toroids- do we send back datalen = 0? */
  /* this didn't work for bpms and gadc case */
  dataReply_s.datrply_totwc = calreplybc/2;

  return (iss);
}
/*=============================================================================

  Name: bsacCalPrepImmedReply

  Abs:  set up  bpm_allprep_reply_ts structure (with help from static global
        dataReply_s, set up by bsacCalParseMsg()) after setting up message
		prologue. Then sends off the immediate reply.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		msgmail_ts          rMsg_ps     read       pointer to the received message
        vmsstat_t           iss         read       field of immed reply
		unsigned short      npulstep    read       field of immed reply

  Rem:  iss should be BPMO_OKOK for the dummy file

  Side: 

  Ret:  vmsstat_t the status code, BPM0_OKOK 
   
==============================================================================*/
LOCAL vmsstat_t bsacCalPrepImmedReply(const msgmail_ts * rMsg_ps,
									  const vmsstat_t iss,
									  const unsigned short npulstep)
{
  vmsstat_t thisiss = BPMO_OKOK;
  msgmail_ts *   sMsg_ps = NULL;  /* immed reply msg to send */
  bpm_allprep_reply_ts immedReply_s;
  int i;

  /* get a buffer for the reply data */
  thisiss = msgQGetBuffer( sizeof(msgmail_ts),(void*) &sMsg_ps);
  if ( SUCCESS(thisiss) )
  {
	/* add in the immed reply in the reqdata part of message*/
	/* stuff data part of immed response message & convert to VMS if nec */
	immedReply_s.iss = iss; /* vmsstat_t = int4u */
	immedReply_s.nmuxmin = 1; /* unsigned char */
	for (i=0; i<3; i++) {
	  immedReply_s.spareb[i]=0;}  /* 3 unsigned chars */
	immedReply_s.datrply_totwc = dataReply_s.datrply_totwc; /* unsigned long */
	immedReply_s.sparel = 0;              /* unsigned long */
	immedReply_s.npulstep = npulstep;     /* unsigned short */
	immedReply_s.sparew = 0;              /* unsigned short */
	cvtLSwap((unsigned long *)&immedReply_s.iss);
	cvtLSwap((unsigned long *)&immedReply_s.datrply_totwc);
	cvtSSwap((unsigned short *)&immedReply_s.npulstep);
    memcpy(&sMsg_ps->reqdata, &immedReply_s, sizeof(immedReply_s));

    /* debug - print the message */
	DEBUGPRINT(DP_DEBUG, bsacCalFlag, 
			   ("The message being sent by bsacCal is:\n"));
	slcPrintMsgheader(&sMsg_ps->msgheader, bsacCalFlag );
	for (i=0; i<sMsg_ps->msgheader.datalen; i++) {
	  DEBUGPRINT(DP_DEBUG, bsacCalFlag, 
				 ("\tthe data[%d]: 0x%.4x\n", i, sMsg_ps->reqdata[i]) );
	}

	/* calc some diagnostics ? see msgHdlr sendOKReply */

    /* send the msg to the msgSend thread */
	DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("Sending CALB_PREP immed response msg\n") );
    thisiss = msgQSlcReply(rMsg_ps->msgheader.func, sizeof(immedReply_s), rMsg_ps, sMsg_ps);
  
    /* calc some diagnostics */
    /*if (!SUCCESS(iss))
	  msgHdlrDiags_s.rplyFails++;
	else
	msgHdlrDiags_s.rplyCnt++;*/

  }
  
 return (thisiss);
}
/*=============================================================================

  Name: bsacCalProc

  Abs:  Sends data response message for BPMO_CALB_PREP message.
        This message is sent immediately after the immediate response
		with function code BPMO_CALB_PREP | RESPONSE.
		The function code for this data response is BPMO_CALB_PROC | RESPONSE.
		The dummy data is filled by calling bsacCalFillData.
		The alldata reply header is helped to be filled by using
		static global dataReply_s, which was filled during bsacCalPrepParseMsg
	    (when original BPMO_CALB_PREP message was received)
		Then sends the data reply to msgSend thread.
        
  Args: none

  Rem: 

  side:

  Ret:  none

==============================================================================*/
LOCAL vmsstat_t bsacCalProc (const msgmail_ts * rMsg_ps)
{
  vmsstat_t iss = BPMO_OKOK;
  msgmail_ts * sMsg_ps;
  unsigned long totbc;
  unsigned long totwc;
  char            *microname = slcGetIOCName();
  /* create and send back dat reply */
  /* first try to send back totwc= 0 and see if this satisfies SCP */

  /* create and send data reply */
  /* first try to send back totwc= 0 and see if this satisfies SCP */
  /* dataReply_s was filled by bsacCalPrep () */

  DEBUGPRINT(DP_INFO, bsacCalFlag, ("bsacCalProc: Entering..\n")); 
  /* get a buffer for the reply data */
  /* total bytes = job_id + fwdhdr + msghdr + data */
  totbc = (dataReply_s.datrply_totwc)*2 + sizeof(unsigned long) +
	sizeof(fwd_hdr_ts) + sizeof(msgheader_ts);
  iss = msgQGetBuffer(totbc, (void*) &sMsg_ps);
  if ( SUCCESS(iss) )
  {
	/* save the word count for later */
	totwc = dataReply_s.datrply_totwc;

    /* now fill cal data before alldata header */
   	bsacCalFillData(sMsg_ps); 

	/* Removed this requirement - compute checksum */
    /*bsacComputeChecksum(sMsg_ps);*/

	/* convert and fill the alldata reply part of message*/
	/*dataReply_s.iss, .calbbits was set already during parsing of PREP */
	dataReply_s.datrply_partwc = dataReply_s.datrply_totwc; 
	dataReply_s.ndatblks = 0;
	dataReply_s.nrpos = 0;
	/* put microname in checksum */
	memcpy(&dataReply_s.checksum,microname ,4);
	cvtLSwap((unsigned long *)&dataReply_s.iss);
	cvtLSwap((unsigned long *)&dataReply_s.bt.calbbits);
	cvtLSwap((unsigned long *)&dataReply_s.datrply_totwc);
	cvtLSwap((unsigned long *)&dataReply_s.datrply_partwc);
	cvtSSwap((unsigned short *)&dataReply_s.ndatblks);
	cvtSSwap((unsigned short *)&dataReply_s.nrpos);
	cvtLSwap((unsigned long *)&dataReply_s.checksum);
    memcpy(sMsg_ps->reqdata, &dataReply_s, sizeof(dataReply_s));

	/* Tony was worried about sending the cal data prior to the SCP sending */
	/* the ppYY request to the mpg. Thankfully, it doesn't seem to matter */
	DEBUGPRINT(DP_DEBUG, bsacCalFlag, 
	  ("Delay 2 sec before sending data response \n") );
	  epicsThreadSleep(2); 

	DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("Sending CALB_PROC data response msg\n") );
    /* send the msg to the msgSend thread */
    iss = msgQSlcReply(BPMO_CALB_PROC, (totwc*2), rMsg_ps, sMsg_ps);

	/* calc some diagnostics */
	/*if (!SUCCESS(iss))
	  msgHdlrDiags_s.rplyFails++;
	  else
	  msgHdlrDiags_s.rplyCnt++;*/
  }
  return (iss);
}

/*=============================================================================

  Name: bsacCalFillData

  Abs:  Responds to  BPMO_CALB_PREP message with:
        a)(BPMO_CALB_PREP | RESPONSE immed reply) and b) BPMO_CALB_PROC data reply.

  Args: Type	     Name        Access	    Description
        -----------  ----------  ---------- ----------------------------
		msgmail_ts   rMsg_ps     read/write ptr to the received message header

  Rem:  Fills the data part of the data response message with 
        bpmcalstruc.hc - bpmp_calbreply_ts for bpms and 
        bpmcalstruc.hs - gadc_calbreply_ts for gadcs
		Returns one of these packed structures for each unit of BPMP and GADC,
		respectively, that is found in the slc db downloaded.


  Side: NOTE: It may be that the BPMP and GADC unit numbers should be checked
		against the umin-umax, umin2-umax2 ranges sent down in the  
		bpm_calbprep_req_ts message. However, Tony said to send as many 
		bpmp_calbreply_ts's as there are units of primary BPMP, 
		if (calbbits & CALB_BPMP_BIT), and
		the same is for gadcs if (calbbits & CALB_GADC_BIT).  

  Ret:  none

==============================================================================*/
LOCAL void bsacCalFillData (msgmail_ts * sMsg_ps)
{
  int i, unit;
  unsigned char * reqdata_ps;
  /* values with which to stuff msg w dummy data */
  unsigned short calStat = 1;
  unsigned short calShort = 0;  /*STAT_GOOD not in bpmstat.h*/
  unsigned long calLong = 0;

  /* init reqdata_ps ptr to point to data part of send message */
  reqdata_ps = (unsigned char *) sMsg_ps->reqdata;
  reqdata_ps += sizeof (dataReply_s);   
  DEBUGPRINT(DP_ERROR,bsacCalFlag, 
				 ("bsacCalFillData: calPrimCnt=%hd\n", (unsigned int)calPrimCnt[BSACBPMS_INDEX]));  
  if (dataReply_s.bt.calbbits & CALB_BPMP_BIT)
	{ /* bpmp wanted */
	  /* NOTE: slc db has one for one bpmp to bpms */
	  DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("Filling data for BPMPs\n") );     
	  /* the only field of cal data that SCP cares about(reads) during meas */
      /*  are the stat bits */  
	  for (unit=0; unit < calPrimCnt[BSACBPMS_INDEX]; unit++) {
		/* stuff  equiv. of (bpmcalstruc.hc) bpmp_creply_unit_beg_ts: 
		     unsigned long  attn; unsigned short stat; into data response,  
           and then bpmp_creply_ep_ts:
		     unsigned short epmonunit; 
		     float m0[3],epeds[4],egcal[4][3],epvar[4],ecvar[4][3]; 
		   into data response message: */ 
		cvtLSwapMove((const void *)&calLong, reqdata_ps);  /* attn */
	    reqdata_ps += sizeof(unsigned long);
		cvtSSwapMove((const void *)&calStat, reqdata_ps); /* stat */
	    reqdata_ps += sizeof(unsigned short);
		cvtSSwapMove((const void *)&calShort, reqdata_ps); /*epmonunit*/
	    reqdata_ps += sizeof(unsigned short);
        for (i=0; i<35; i++) {
		  /* 35 floats: m0[3], epeds[4], egcal[4][3], epvar[4], ecvar[4][3] */
		  cvtLSwapMove((const void *)&calLong, reqdata_ps);
		  reqdata_ps += sizeof(unsigned long);	
		}		
		DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("\t unit = %d\n", unit) );
	  }
	} /* bpmp wanted */
  else {	  
	DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("no BPMPs specified for CAL!!!\n") );
  }
  if (dataReply_s.bt.calbbits & CALB_GADC_BIT){
	/* gadc wanted */
	DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("Filling data for GADCs..\n") );
	for (unit=0; unit< calPrimCnt[BSACGADC_INDEX]; unit++) {
	  /* stuff  equiv. of (bpmcalstruc.hc) gadc_calbreply_ts:
		   unsigned short stat[MGADCH]; float peds[MGADCH],pvar[MGADCH];
		 into data response message: */ 
	  for (i=0; i< MGADCH; i++) {
		/* stuff unsigned short stat[i]*/
		cvtSSwapMove((const void *)&calStat, reqdata_ps); 
		reqdata_ps += sizeof(unsigned short);
	  }
	  for (i=0; i< MGADCH; i++) {
		/* stuff float peds[i] */
		cvtLSwapMove((const void *)&calLong, reqdata_ps);
		reqdata_ps += sizeof(float);
	  }
	  for (i=0; i< MGADCH; i++) {
		/* stuff float pvar[i] */
		cvtLSwapMove((const void *)&calLong, reqdata_ps);
		reqdata_ps += sizeof(float);
	  }
	  DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("\t unit = %d\n", unit) );
	}
  }
  else {	  
	DEBUGPRINT(DP_DEBUG, bsacCalFlag, ("no GADCs specified for CAL!!!\n") );
  }
  return;
}
/*=============================================================================

  Name: bsacComputeChecksum

  The following routine is not begin used, but left for posterity 
  8/3/2005 - lcls controls group decided not to do checksumming on data 
  Note - do the following until Mike Z changes the SCP to defeat checking 
  checksums on data from slc-aware iocs: 
  on mccdev term that scp wt drogind is started 
  Ctrl Y 
  debug 
  set image bpmutil 
  set scope bpmread 
  deposit compute_checksum = .false.

  Args: Type	     Name        Access	    Description
        -----------  ----------  ---------- ----------------------------
		msgmail_ts   sMsg_ps     read/write ptr to the send  message

  Rem: Algorithm :
       Note that datrply_totwc does not include msgheader_ts,
       and longword checksum does not include msgheader_ts.
       If datrply_totwc is odd then store a 16-bit zero as padding
       immediately after the last 16-bit word in the message.
       Copy datrply_totwc into datrply_partwc.
       Store 1 (little-endian) in checksum (in bpm_alldata_replybeg_ts).
       Consider the entire message (starting at the beginning of
       bpm_alldata_replybeg_ts) to be a longword array (including
       possible zero padding at end).  Start an accumulator at 0.
       For each longword in that array, fetch that longword,
       byteswap it to make it big-endian, and subtract it from the
       accumulator.  After bottom of that loop, byteswap that
       accumulator to make it little-endian and store it into
       checksum (in bpm_alldata_replybeg_ts).

  Side: 

  Ret:  none

==============================================================================*/

     /* Checksum is such that if entire message excluding message service    */
     /* header is summed longwordwise (temporarily appending zeroes to the   */
     /* end of the message to make its total length be 0 mod 4 bytes, and    */
     /* using the final value of datrply_partwc), then the 32-bit sum is -1, */
     /* assuming checksum itself begins on a longword boundary within        */
     /* bpm_alldata_replybeg_ts.                                             */
void bsacComputeChecksum (msgmail_ts * sMsg_ps)
{
#define DATRPLY (*(bpm_alldata_replybegh_ts *)reply_tok)
  unsigned long checksum, j, reply_totdwc;
  unsigned long * reply_tok;
  unsigned short reply_datwc;

  reply_tok = (unsigned long *)&sMsg_ps->msgheader;
  *(long *) ((char *) reply_tok + ((DATRPLY.dat.datrply_totwc + NETHEADERLEN) << 1)) = 0;  
  reply_totdwc = (DATRPLY.dat.datrply_totwc + 1) >> 1;
  DATRPLY.dat.datrply_partwc = DATRPLY.dat.datrply_totwc;
  DATRPLY.dat.checksum = 1;
  checksum = 0;
  for (j = 0;  j < reply_totdwc;  ++j){
	/* extra step - endianize back to native */
	checksum -= ((long *) &DATRPLY.dat)[j];
  }
  DATRPLY.dat.checksum = checksum;
  reply_datwc = DATRPLY.dat.datrply_totwc <= NETMICLONGMSGLEN - NETHEADERLEN ?
	DATRPLY.dat.datrply_totwc : NETMICLONGMSGLEN - NETHEADERLEN;
  DATRPLY.dat.datrply_partwc = reply_datwc;
  /* don't forget to endianize before sending */
  return;
}
