/*=============================================================================
 
  Name: bsacData.c


  Abs:  bsacData contains functions that handle data received from the EPICS
        bsac database. 
        Reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | BSAC Service Design

  Auth: 18   aug 2005, dfairley    created 
  Rev:  

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest) 
        18 aug 2005, Dfairley    created 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int bsacDataFlag = DP_INFO;
#endif


/* c includes */
#include <string.h>        /* memcpy */

/* epics */
#include <dbAccess.h>      /* EPICS dbAddr struct */

/* shared slc micro includes */
#include "bpmfunc.hc"      /* bpmo_xxx func codes */
#include "micrdef.hc"      /* MICR_EXNET_ codes */
#include "bpmodef.hc"      /* BPMO_OKOK etc */
#include "msgfunc.hc"      /* RESPONSE for func code */

/* slc global vars */
#include "slcGlob.h"       /* slcThreads_as */

/* slcIOC utilities */
#include "slcUtil.h"       /* slcPrintMsgheader */
#include "msgQUtil.h"      /* messageQSend, messageQRecv */
#include "cvtUtil.h"       /* slcSSwap, slcLSwap, etc */
#include "slcCmlog.h"      /* slcCmlogLogMsg */
#include "timeUtil.h"      /* timeGetVMS, timeDiffVMS10usec */


/* bsac includes */
#include    "bsacGlob.h"
#include    "bsacData.h"
#include    "bpmstruc.h"
#include    "bpm_reqreply_struc.h"
#include    "bsacPrep.h"   /* bsacPrepForNewCycle */
#include    "bsacAbort.h"  /* bsacAbortFindScp */


/* static bsacData file scoped vars */
LOCAL arry_data_ts   tmpArryData_s;

/* local prototypes */
LOCAL void bsacDataCvtMoveBPMSData( bpms_data_ts *bpms_ps,
									unsigned char * buf_p);
LOCAL void bsacDataCvtMoveTOROData( toro_data_ts *toro_ps,
									unsigned char * buf_p);
LOCAL void bsacDataSaveARRYChan(const unsigned short chanIndex, 
								const arry_chan_data_ts *chan_ps,
								 arry_data_ts *arry_ps);
LOCAL void bsacDataCvtMoveARRYData( const arry_data_ts *arry_ps,
									unsigned char * buf_p);
LOCAL void bsacDataCvtMovePUIDData( const bsacMD_ts *bsacMD_ps,
									const ws_puid_data_ts *puid_ps,
									unsigned char * buf_p);
LOCAL vmsstat_t bsacDataSendNext(const unsigned short md);

/*-----------------------------------------------------------------------
  Name: bsacData

  Abs: move the data from this message into its proper place in the msg buffer

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const msgmail_ts* rMsg_ps read only  incoming data message

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacData (const msgmail_ts* rMsg_ps)
{
  unsigned short pI, uI, cI;
  short md; 
  unsigned short datasize=0;
  short bufIndex = 0;
  bsacMsgHdr_ts * hdr = NULL;
  unsigned char * indata_p = NULL;
  unsigned char * bufdata_p = NULL;
  vmsstat_t iss = BPMO_OKOK;
  /* debug 
	 unsigned char * debbufdata_p = NULL;*/

  /* interpret the header */
  hdr = (bsacMsgHdr_ts*)&rMsg_ps->reqdata;
  pI = hdr->primIndex;
  uI = hdr->unitIndex;
  cI = hdr->chanIndex; 
  if (cI > 0) cI -= 1; 
  /* EPICS bsacXL*.substitutions has ARRY channels 1-12*/
  /* incoming cI channel number of 0 means no ARRY is involved with this device */
  /*  so subtract 1- only if there are channels involved w this device to use as index */

  md = hdr->md-1; /* the EPICS md is one more than slc threads md */
  
  /* keep this data if we are not aborting this md */
  if (bsacMD_as[md].abort!=epicsTrue)
  {

	/* special case - ARRY data is stored in a structure of 
	   12 channels, but not all of the channels are used in a 
	   particular IOC/md.  primCollect....chanCntr will count down
	   the expected number of channel data messages for each measured
	   pulse, and the channel data is stored in a local static temp 
	   structure until all channels data is sent for a pulse. 
	   HERE - we keep track of channel count for arry data
	   if chanCntr=-1 we start the counter again by resetting to chanCnt;
	   when resetting the counter, we also initialize the tmpArryData_s
	*/
	if ( primCollect_as[pI].unit_as[uI].chanCntr==-1) {
	  primCollect_as[pI].unit_as[uI].chanCntr = primCollect_as[pI].unit_as[uI].chanCnt; 
	  memset(&tmpArryData_s, 0, sizeof(tmpArryData_s));
	}

	/* get pointer to outgoing buf data_p */
	bufIndex = primCollect_as[pI].unit_as[uI].mdef_as[md].bufIndex;
	bufdata_p = bsacMDBufs_as[md].bufs_as[bufIndex].data_p;
	DEBUGPRINT(DP_INFO, bsacDataFlag, 
				 ("bsacData: got data, storing in data_p=%p + offset=%ld\n",
				 bufdata_p, primCollect_as[pI].unit_as[uI].mdef_as[md].offset ));    
	if (bufdata_p) 
	  bufdata_p+= primCollect_as[pI].unit_as[uI].mdef_as[md].offset;
	else {
	  DEBUGPRINT(DP_ERROR, bsacDataFlag, 
	   ("bsacData:received data from pI=%d,md=%d,uI=%d,cI=%d but it has no lrg msg buffer!\n",
		pI, md, uI, cI ));
	  return iss;
	}

	/* get a pointer to the data part of the incoming data msg */
	indata_p = (unsigned char *)hdr + sizeof(bsacMsgHdr_ts);

	/* switch on primary and cvt/move data into lrg msg buffer */
	switch (pI){
	case BSACBPMS_INDEX: 
	  DEBUGPRINT(DP_INFO, bsacDataFlag, 
				 ("bsacData:received BPMS data from md=%d where uI=%d, cI=%d\n",
				  md, uI, cI ));    
	  /* convert and move the data to outgoing large msg buffer */
	  /* scale simulated tmit data 1. E+9; simulated data is in nC  */
	  /*( (bpms_data_ts *)indata_p)->tmit *=1000000000.0;*/
	 /*bsacDataDebugPrintBPM ((const bpms_data_ts *)indata_p);*/

	  bsacDataCvtMoveBPMSData((bpms_data_ts *)indata_p, (void *)bufdata_p);
	  /*debbufdata_p = bufdata_p;
		bsacDataDebugPrintBPM ((const bpms_data_ts *)debbufdata_p);*/
	  datasize = BPMS_DATA_SIZE;
	  break;
	case BSACTORO_INDEX: 
	  DEBUGPRINT(DP_INFO, bsacDataFlag, 
				 ("bsacData:received TORO data from md=%d where uI=%d, cI=%d\n",
				  md, uI, cI ));    
	  /* convert and move the data to outgoing large msg buffer */
	  /* scale simulated tmit data 1. E+9 for message; simulated data is in nC  */
	  /*     ( (toro_data_ts *)indata_p)->tmit *=1000000000.0;*/
	  bsacDataCvtMoveTOROData((toro_data_ts *)indata_p, (void *)bufdata_p);
	  datasize = TORO_DATA_SIZE;
	  break;
	case BSACARRY_INDEX: 
	  /* special case - ARRY data is stored in a structure of 12 channels, but not
		 all of the channels are used in a particular IOC/md.  primCollect....chanCntr
		 will count down the expected number of channel data messages for each measured
		 pulse, and the channel data is stored in a local static temp structure, until
		 all channels data is sent for the pulse */
	  DEBUGPRINT(DP_DEBUG, bsacDataFlag, 
				 ("bsacData:received ARRY data from md=%d where uI=%d, cI=%d; entering unit's chanCntr = %d \n",
				  md, uI, cI, primCollect_as[pI].unit_as[uI].chanCntr  ));    

	  if ( primCollect_as[pI].unit_as[uI].chanCntr>0) {
		/* collect up ALL the channels for this md/navg before moving to lrg buffer */
        /* chanCntr counts down the expected number of chans for this pulse measurment*/
		primCollect_as[pI].unit_as[uI].chanCntr--;
		/* now copy this measurement to the temporary structure */
		bsacDataDebugPrintChan ((const arry_chan_data_ts *)indata_p);
		bsacDataSaveARRYChan(cI, (const arry_chan_data_ts *)indata_p, &tmpArryData_s);
		/* this data was not moved into the buffer, do not add a datasize to the offset */
		datasize=0;
		DEBUGPRINT(DP_INFO, bsacDataFlag, 
				 ("bsacData: stored ARRY data from md=%d where uI=%d, cI=%d\n",
				  md, uI, cI ));    
		if ( primCollect_as[pI].unit_as[uI].chanCntr==0) {	
		  /* copy this data to the temp structure */
		  bsacDataSaveARRYChan(cI, (const arry_chan_data_ts *)indata_p, &tmpArryData_s);
		  /* convert and move the temp structure data to outgoing large msg buffer */
		  bsacDataDebugPrintARRY ((const arry_data_ts *) &tmpArryData_s);
		  bsacDataCvtMoveARRYData((const arry_data_ts *)&tmpArryData_s, (void *)bufdata_p);
		  bsacDataDebugPrintARRY ((const arry_data_ts *) bufdata_p);
		  datasize = ARRY_DATA_SIZE;
		  /* reset unit's chanCntr for next time */
		  primCollect_as[pI].unit_as[uI].chanCntr=-1;
		  DEBUGPRINT(DP_DEBUG, bsacDataFlag, 
					 ("bsacData: stored ARRY data from md=%d where uI=%d, on last cI=%d\n",
					  md, uI, cI ));
		}    

	  } else {

		DEBUGPRINT(DP_DEBUG, bsacDataFlag, 
		  ("bsacData:ERROR!! received too many ARRY data from md=%d where uI=%d, cI=%d\n",
		   md, uI, cI ));    
	  } 
	  break;
	case BSACPUID_INDEX: 
	  if (bsacMD_as[md].measbits & MEASPULSID_BIT) {
		DEBUGPRINT(DP_INFO, bsacDataFlag, 
				   ("bsacData:received PUID data from md=%d where  uI=%d, cI=%d\n",
					md, uI, cI ));    
		/* convert and move the data to outgoing large msg buffer */
		/* special case - we need to know when the first pulse of a meas. has
		   occurred, so that we can insert the correct PUID hdr data*/
		bsacDataCvtMovePUIDData( &bsacMD_as[md], 
								 (const ws_puid_data_ts *)indata_p, 
								 (void *)bufdata_p);
		/* if this is the first pulse of this measurement we must reset the flag */
		if (bsacMD_as[md].firstPulse) {
		  bsacMD_as[md].firstPulse=epicsFalse; 
		  datasize = PUID_DATA_SIZE;
		}else {
		  /* add size of a single pulse id */
		  datasize = sizeof(unsigned long);
		}
	  }
	  else {
		DEBUGPRINT(DP_INFO, bsacDataFlag, 
				   ("bsacData: Throwing away unnecessary PUID data from md=%d where  uI=%d, cI=%d\n",
					md, uI, cI )); 
		return iss;
	  }
	  break;
	case BSACWIRE_INDEX: 
	  DEBUGPRINT(DP_ERROR, bsacDataFlag, 
		   ("bsacData:ERROR!!!received WIRE data from md=%d where  uI=%d, cI=%d\n",
		   md, uI, cI ));    
	  datasize = 0;
	  break;
	default:
	  DEBUGPRINT(DP_ERROR, bsacDataFlag, 
	   ("bsacData:ERROR!!!received UNKOWN data from md=%d where pI=%d, uI=%d, cI=%d\n",
		md, pI, uI, cI ));
	  datasize = 0;
	  break;
	}

	/* update the msg counter and offset */
	primCollect_as[pI].unit_as[uI].mdef_as[md].counter++;
	primCollect_as[pI].unit_as[uI].mdef_as[md].offset+=datasize;
    if (bsacMD_as[md].msgCounter>0) bsacMD_as[md].msgCounter--;
	else DEBUGPRINT(DP_ERROR, bsacDataFlag, 
	   ("bsacData:ERROR!!!too many data messages for  md=%d \n",hdr->md-1));

  }
   
  /* update diagnostics for data measurement */
  return(BPMO_OKOK);
}


/*-----------------------------------------------------------------------
  Name: bsacDone

  Abs: move this last data from this message into its proper
       place in the msg buffer. Measurement is done, prepare to send.

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const msgmail_ts* rMsg_ps read only  incoming data message

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacDone (const msgmail_ts* rMsg_ps)
{
  short md;
  bsacMsgHdr_ts * hdr = NULL;
  vmsstat_t iss = BPMO_OKOK;
  
  /* interpret the header */
  hdr = (bsacMsgHdr_ts*)&rMsg_ps->reqdata;
  md = hdr->md-1; /* epics records use md+1 */
  
  /* store the incoming data */
  iss = bsacData(rMsg_ps);

  /* handle done stuff */
  if (bsacMD_as[md].msgCounter == 0 ){
	/* send the first lrgbuffer to the scp */
	DEBUGPRINT(DP_INFO, bsacDataFlag, 
	("bsacDone: sending data message for md=%d\n", md));
	iss = bsacDataSendFirst(md);
	/* if this is the last buffer, setup for the next yy cycle */
	/*  This is for Start/Stop */
	if (bsacMDBufs_as[md].nextBuf >= bsacMDBufs_as[md].bufCount) {
	  iss = bsacPrepForNewCycle(md);
	}
  }
  else
	DEBUGPRINT(DP_INFO, bsacDataFlag, 
	("bsacData:not enough data messages yet for md=%d, need %ld more \n",md, bsacMD_as[md].msgCounter));

     
  /* update diagnostics for done measurement */
  return(iss);
}

/*-----------------------------------------------------------------------
  Name: bsacDataCvtMoveBPMSData

  Abs: store this packet of BPMS data into the large data buffer

  Args: Type		         Name    Access	    Description
	    -------------------- ------- ---------- ----------------------------
		const bpms_data_ts*  bpms_ps read only  incoming data message
        unsigned char *      buf_p   read/write pointer to lrgmsg buffer area for bpms

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacDataCvtMoveBPMSData( bpms_data_ts *bpms_ps,
									unsigned char *buf_p)
{
  unsigned short cnt = 0;

  /* bpms float values */
  cvtFloatIEEEToVMS((const void *)&bpms_ps->x, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->x);
  DEBUGPRINT(DP_INFO, bsacDataFlag, 
	   ("bsacDataCvtMoveBPMSData X: %f\n",bpms_ps->x));
  cvtFloatIEEEToVMS((const void *)&bpms_ps->y, (void *)buf_p);
  DEBUGPRINT(DP_DEBUG, bsacDataFlag, 
	   ("bsacDataCvtMoveBPMSData Y: %f\n",bpms_ps->y));
  buf_p+=sizeof(bpms_ps->y);

  cvtFloatIEEEToVMS((const void *)&bpms_ps->tmit, (void *)buf_p);
  DEBUGPRINT(DP_DEBUG, bsacDataFlag, 
	   ("bsacDataCvtMoveBPMSData TMIT: %f\n",bpms_ps->tmit));
  buf_p+=sizeof(bpms_ps->tmit);
  cvtFloatIEEEToVMS((const void *)&bpms_ps->x_rms, (void *)buf_p);
  DEBUGPRINT(DP_DEBUG, bsacDataFlag, 
	   ("bsacDataCvtMoveBPMSData X_rms: %f\n",bpms_ps->x_rms));
  buf_p+=sizeof(bpms_ps->x_rms);
  cvtFloatIEEEToVMS((const void *)&bpms_ps->y_rms, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->y_rms);
  cvtFloatIEEEToVMS((const void *)&bpms_ps->tmit_rms, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->tmit_rms);
  /* bpms short values */
  for (cnt = 0; cnt<4; cnt++) {
	cvtSSwapMove((const void *)&bpms_ps->qraw[cnt], (void *)buf_p);
	buf_p+=sizeof(bpms_ps->qraw[cnt]);
  }
  cvtSSwapMove((const void *)&bpms_ps->stat, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->stat);
  cvtSSwapMove((const void *)&bpms_ps->goodmeas, (void *)buf_p);
  buf_p+=sizeof(bpms_ps->goodmeas);
}

/*-----------------------------------------------------------------------
  Name: bsacDataCvtMoveTOROData

  Abs: store this packet of TORO data into the large data buffer

  Args: Type		         Name    Access	    Description
	    -------------------- ------- ---------- ----------------------------
		const toro_data_ts*  toro_ps read only  incoming data message
        unsigned char *      buf_p   read/write pointer to lrgmsg buffer area for bpms


  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacDataCvtMoveTOROData( toro_data_ts *toro_ps,
									unsigned char *buf_p)
{
  /* toro float values */
  cvtFloatIEEEToVMS((const void *)&toro_ps->tmit, (void *)buf_p);
  buf_p+=sizeof(toro_ps->tmit);
  cvtFloatIEEEToVMS((const void *)&toro_ps->tmit_rms, (void *)buf_p);
  buf_p+=sizeof(toro_ps->tmit_rms);
  /* toro short values */
  cvtSSwapMove((const void *)&toro_ps->qraw, (void *)buf_p);
  buf_p+=sizeof(toro_ps->qraw);
  cvtSSwapMove((const void *)&toro_ps->stat, (void *)buf_p);
  buf_p+=sizeof(toro_ps->stat);
  cvtSSwapMove((const void *)&toro_ps->goodmeas, (void *)buf_p);
  buf_p+=sizeof(toro_ps->goodmeas);
}

/*-----------------------------------------------------------------------
  Name: bsacDataSaveARRYChan

  Abs: store this packet of arry channel data into a temp arry channel struct

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const msgmail_ts* rMsg_ps read only  incoming data message

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacDataSaveARRYChan( const unsigned short chanIndex, 
								 const arry_chan_data_ts *chan_ps,
								 arry_data_ts *arry_ps)
{

  arry_ps->data[chanIndex] = chan_ps->data;
  arry_ps->drms[chanIndex] = chan_ps->drms;
  arry_ps->raw[chanIndex]  = chan_ps->raw;
  arry_ps->stat[chanIndex] = chan_ps->stat;
  arry_ps->goodmeas[chanIndex] = chan_ps->goodmeas;
}

/*-----------------------------------------------------------------------
  Name: bsacDataCvtMoveARRYData

  Abs: store this packet of arry data into the large data buffer

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const arry_data_ts*  arry_ps read only  incoming data message
        unsigned char *      buf_p   read/write pointer to lrgmsg buffer area for bpms

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacDataCvtMoveARRYData( const arry_data_ts *arry_ps,
										 unsigned char * buf_p)
{
  unsigned short cnt = 0;

  /* arry float values */
  for (cnt = 0; cnt<MAX_ARRY_CHAN; cnt++) {
	cvtFloatIEEEToVMS((const void *)&arry_ps->data[cnt], (void *)buf_p);
	buf_p+=sizeof(float);  
  }
  for (cnt = 0; cnt<MAX_ARRY_CHAN; cnt++) {
	cvtFloatIEEEToVMS((const void *)&arry_ps->drms[cnt], (void *)buf_p);
	buf_p+=sizeof(float);  
  }
  for (cnt = 0; cnt<MAX_ARRY_CHAN; cnt++) {
	cvtSSwapMove((const void *)&arry_ps->raw[cnt], (void *)buf_p);
	buf_p+=sizeof(unsigned short);  
  }
  for (cnt = 0; cnt<MAX_ARRY_CHAN; cnt++) {
	cvtSSwapMove((const void *)&arry_ps->stat[cnt], (void *)buf_p);
	buf_p+=sizeof(unsigned short);  
  }
  for (cnt = 0; cnt<MAX_ARRY_CHAN; cnt++) {
	cvtSSwapMove((const void *)&arry_ps->goodmeas[cnt], (void *)buf_p);
	buf_p+=sizeof(unsigned short);  
  }
}

/*-----------------------------------------------------------------------
  Name: bsacDataCvtMovePUIDData

  Abs: store this packet of pulse id data into the large data buffer

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
		const bpm_pulseid_data_ts*  puid_ps read only  incoming data message
        unsigned char *      buf_p   read/write pointer to lrgmsg buffer area for bpms

  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL void bsacDataCvtMovePUIDData(const bsacMD_ts *bsacMD_ps, 
								   const ws_puid_data_ts *puid_ps,
									unsigned char * buf_p)
{
  vmsTimeStamp_ta vmsTime;
  unsigned short nmuxmin = 1;

  /* puid values */
  /* if this is the very first pulse we have to insert a header */
  if (bsacMD_ps->firstPulse==epicsTrue) {
	/*  setting navg */
	DEBUGPRINT(DP_INFO, bsacDataFlag, 
	 ("bsacDataCvtMovePUIDData: first pulse with navg=%d, nmuxmin=%d \n",
	  bsacMD_ps->navg, nmuxmin ) )  ;
	cvtSSwapMove((const void *)&bsacMD_ps->navg, (void *)buf_p);
	buf_p+=sizeof(unsigned short);
	/*setting nmuxmin to one */
	cvtSSwapMove((const void *)&nmuxmin, (void *)buf_p);
	buf_p+=sizeof(unsigned short);
	/* set the timestamp */
    if ((timeGetVMS(vmsTime))!= epicsTimeERROR) {
	  memcpy((void *)buf_p, &vmsTime,8);
	}
	else{ 
	  memset((void *)buf_p, 0, 8);
	  buf_p+=sizeof(vmsTime);
	}
  }
  /* now copy in the pulseid */
  cvtLSwapMove((const void *)&puid_ps->pulseid, (void *)buf_p);
  buf_p+=sizeof(unsigned long);
}

/*-----------------------------------------------------------------------
  Name: bsacDataSend

  Abs: send the first large message buffer to the scp in response to a
       meas request

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
        const unsigned short md   read only  mdef whose buffer must be sent
  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacDataSendFirst(const short md)
{
  vmsstat_t iss = BPMO_OKOK;
  lrgmail_ts * msg_ps = NULL;
  bsacDataBufs_ts * mdBuf_ps = &bsacMDBufs_as[md]; 

  if (mdBuf_ps->bufCount>0) {
	/* get the pointer to the first message buffer */
	msg_ps = mdBuf_ps->bufs_as[0].buf_ps;
    if (msg_ps) {
	  DEBUGPRINT(DP_INFO, bsacDataFlag, 
	   ("bsacDataSendFirst:sending first buffer for md=%d \n",md) )  ;
	  bsacDataDebugPrintRplybeg((const bpm_alldata_replybeg_ts *)&msg_ps->reqdata); 	  
	  /* send the msg to the msgSend thread */
	  if (SUCCESS(iss = (msgQMsgReply(msgSend, msg_ps))) ) {
		mdBuf_ps->bufs_as[0].buf_ps = NULL;
		mdBuf_ps->bufs_as[0].data_p = NULL; 
		mdBuf_ps->nextBuf++;
	  }else {
		DEBUGPRINT(DP_ERROR, bsacDataFlag, 
		 ("bsacDataSendFirst:ERROR!!Could not send first buffer for  md=%d \n",md) )  ;

	  }
	} else {
		DEBUGPRINT(DP_ERROR, bsacDataFlag, 
		 ("bsacDataSendFirst:ERROR!!not a valid pointer to first buffer for  md=%d \n",md) )  ;
	  
	}
  }
  return iss;
}

/*-----------------------------------------------------------------------
  Name: bsacDataSendNext

  Abs: send the first large message buffer to the scp in response to a
       meas request

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
        const unsigned short md   read only  mdef whose buffer must be sent
  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
LOCAL vmsstat_t bsacDataSendNext(const unsigned short md)
{
  vmsstat_t iss = BPMO_OKOK;
  lrgmail_ts * msg_ps = NULL;
  bsacDataBufs_ts * mdBuf_ps = &bsacMDBufs_as[md]; 

  DEBUGPRINT(DP_INFO, bsacDataFlag, 
	   ("bsacDataSendNext: bufCount=%d for md=%d, nextBuf=%d \n",
		mdBuf_ps->bufCount, md, mdBuf_ps->nextBuf ) );
  if ( (mdBuf_ps->bufCount>0) && (mdBuf_ps->nextBuf<mdBuf_ps->bufCount) ) {
	/* get the pointer to the first message buffer */
	msg_ps = mdBuf_ps->bufs_as[mdBuf_ps->nextBuf].buf_ps;
    if (msg_ps) {
	  DEBUGPRINT(DP_INFO, bsacDataFlag, 
	   ("bsacDataSendNext:sending next buffer for md=%d buf=%d \n", 
		md, mdBuf_ps->nextBuf ) );
      /* set the response function code */
	  msg_ps->msgheader.func = BPMO_GETREMDAT | RESPONSE;
	  bsacDataDebugPrintRplybeg((const bpm_alldata_replybeg_ts *)&msg_ps->reqdata); 
	  /* send the msg to the msgSend thread */
	  if (SUCCESS(iss = (msgQMsgReply(msgSend, msg_ps))) ) {
		mdBuf_ps->bufs_as[mdBuf_ps->nextBuf].buf_ps = NULL;
		mdBuf_ps->bufs_as[mdBuf_ps->nextBuf].data_p = NULL;
		(mdBuf_ps->nextBuf)++;
		DEBUGPRINT(DP_INFO, bsacDataFlag, 
				   ("bsacDataSendNext:completed. Next buffer for md=%d buf=%d \n",
					md,mdBuf_ps->nextBuf ) );
	  }else {
		DEBUGPRINT(DP_ERROR, bsacDataFlag, 
		 ("bsacDataSendNext:ERROR!!Could not send next buffer for  md=%d \n",md) )  ;

	  }
	} else {
	  DEBUGPRINT(DP_ERROR, bsacDataFlag, 
	   ("bsacDataSendNext:ERROR!!not a valid pointer to first buffer for  md=%d \n",md) );
	  
	}
  }
  return iss;
}


/*-----------------------------------------------------------------------
  Name: bsacDataGetRemDat

  Abs: send the first large message buffer to the scp in response to a
       meas request

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
        const unsigned short md   read only  mdef whose buffer must be sent
  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacDataGetRemDat(const char* scpName)
{
  vmsstat_t iss = BPMO_OKOK;
  short md = 0;


  /* find the md with the same scpname */
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("bsacDataGetRemDat: md = %hd\n", md));

  /* 
	 bsacAbortFindScp - the call used to use dbGetFields to read epics
	 records CTRL and USERNAME to find the md.  This caused a segmentation error
	 in LINUX only, which to this point is unexplained.  A workaround was created 
	 by changing the bsacAbortFindScp to check a bsacMD_as[md].username stored 
	 during the measurement preparation 
  */
  if (SUCCESS(iss = (bsacAbortFindScp(scpName, &md)))) {

	/* send the next buffer for this md */
	iss = bsacDataSendNext(md);	

	if (bsacMD_as[md].measprep){
	  /*if this is the last buffer, setup for the next yy cycle */
   	  if (bsacMDBufs_as[md].nextBuf>=bsacMDBufs_as[md].bufCount) {
		iss = bsacPrepForNewCycle(md); 
	  }  else {
		DEBUGPRINT(DP_DEBUG, bsacDataFlag, 
		("bsacDataGetRemDat: no measdef for scp=%s\n",scpName));  
	  }
	}
  }
  return (iss);
}


/* global debug routine */
void bsacDataDebugPrintRplybeg (const bpm_alldata_replybeg_ts * rb)
{
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("bsacDataDebugPrintRplybeg: \n"));  
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tiss: %ld \n", rb->iss)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tmeasbits: %ld \n", rb->bt.measbits));   
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\ttotwc: %ld \n", rb->datrply_totwc)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tpartwc: %ld \n", rb->datrply_partwc)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tndatblks: %d \n", rb->ndatblks));
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tnrpos: %d \n", rb->nrpos)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tchecksum: %ld \n", rb->checksum));   
  return;
}

/* global debug routine */
void bsacDataDebugPrintBPM (const bpms_data_ts * bpm_p)
{
  bpms_data_ts bpmData_s;
  int cnt;

  memcpy (&bpmData_s, bpm_p, sizeof(bpmData_s));
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("bsacDataDebugPrintBPM: \n"));  
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tX: %f \n", bpmData_s.x)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tY: %f \n", bpmData_s.y));   
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tTMIT: %f \n", bpmData_s.tmit)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tgoodmeas: %hd \n", bpmData_s.goodmeas)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tstat: 0x%hx \n", bpmData_s.stat));
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tX_rms: %f \n", bpmData_s.x_rms)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tY_rms: %f \n", bpmData_s.y_rms));   
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tTMIT_rms: %f \n", bpmData_s.tmit_rms)); 
  for (cnt=0; cnt<4; cnt++){
	DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tqraw[%d]: %hd \n", cnt, bpmData_s.qraw[cnt])); 
  }
  return;
}
void bsacDataDebugPrintChan (const arry_chan_data_ts * chan_p)
{
  arry_chan_data_ts chanData_s;
  memcpy (&chanData_s, chan_p, sizeof(chanData_s));

  DEBUGPRINT(DP_INFO, bsacDataFlag, ("bsacDataDebugPrintChan: \n"));  
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tdata: %f \n", chanData_s.data)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tdrms: %f \n", chanData_s.drms));   
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\traw: 0x%hx \n", chanData_s.raw)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tgoodmeas: %hd \n", chanData_s.goodmeas)); 
  DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tstat: 0x%hx \n", chanData_s.stat));   
  return;
}
void bsacDataDebugPrintARRY (const arry_data_ts * arry_p)
{
  arry_data_ts arryData_s;
  int cnt;

  memcpy (&arryData_s, arry_p, sizeof(arryData_s));

  for (cnt=0; cnt<MAX_ARRY_CHAN; cnt++){
	DEBUGPRINT(DP_INFO, bsacDataFlag, ("bsacDataDebugPrintARRY: \n"));  
	DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tdata[%d]: %f \n", cnt, arryData_s.data[cnt])); 
	DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tdrms[%d]: %f \n", cnt,arryData_s.drms[cnt]));   
	DEBUGPRINT(DP_INFO, bsacDataFlag, ("\traw[%d]: 0x%hx \n", cnt, arryData_s.raw[cnt])); 
	DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tgoodmeas[%d]: %hd \n", cnt, arryData_s.goodmeas[cnt])); 
	DEBUGPRINT(DP_INFO, bsacDataFlag, ("\tstat[%d]: 0x%hx \n", cnt, arryData_s.stat[cnt])); 
  }  
	return;
}
