/*=============================================================================
 
  Name: bsacGlob.c

  Abs:  Initializes BSAC Global variables

  Auth: 22-jun-2005, debbie rogind (DROGIND): created
  Rev: 

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  

        
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int bsacGlobFlag = DP_ERROR;
#endif


/* c library includes */
#include    <stdio.h>
#include	<string.h>

/* epics includes */
#include    <dbDefs.h>        /* NELEMENTS */
#include    <epicsEvent.h>    /* epicsEventId */
#include    <freeList.h>      /* freeListItemsAvail */
#include    <dbAccess.h>      /* Epics Runtime Database */

/* slc common includes */
#include    "slcGlob.h"       /* globals, types, */
#include    "msgTypes.h"      /* msgmail_ts */
#include    "slcMsg.h"        /* slcIOC internal msg function codes */
#include    "msgfunc.hc"      /* FUNC_TEST job code */
#include    "micrdef.hc"      /* MICR_OKOK */

/* slc utility prototypes */
#include    "msgQUtil.h"      /* message queue utilities */
#include    "slcCmlog.h"      /* slcCmlogLogMsg */
#include    "cvtMacr.h"       /* macro used for endian-ness */

/* thread prototypes  */
#include    "slcMain.h"    /* slcStart, slcRestart, slcStop */
#include    "slcExec.h"    /* slcExec thread */

/* dbs utility prototypes (for db free list commands) */
#include    "dbGlob.h"
#include    "dblistUtil.h"
#include    "dbl.h"
#include    "dbdict.h"
#include    "dbdef.hc"

/* bsac global */
#include    "bsacTypes.h"
#include    "bsacAbort.h"
#include    "bsacData.h"
#include    "bsacHsta.h"
#include    "bsacPrepTypes.h"
#include    "bsacPrepInit.h"  /*  bsacPrepInitMDEFAddr() */
#include    "bsacGlob.h" 

/* function prototypes */
vmsstat_t     bsacInitGlobals(void);

/* global variables */

/* primary collection structure to track collection of data from epics */
/*  into large message buffer(s) */
bsacMD_ts       bsacMD_as[MDEF_MAX];
/* keeps unit count for primaries BPMP, GADC, TORO */
/* The cal tables desire these primaries (not BPMS, ARRY, etc ) */ 	
unsigned short  calPrimCnt[BSACPRIM_MAX];
bsacDataBufs_ts bsacMDBufs_as[MDEF_MAX];
primStruct_ts   primCollect_as[BSACPRIM_MAX];
 
/* local function prototypes */
LOCAL vmsstat_t bsacGetSLCPrims (void);
LOCAL vmsstat_t bsacInitPrim (const unsigned short primIndex,
							  const char * name_c);
LOCAL vmsstat_t bsacGetARRYChans (void);
LOCAL vmsstat_t bsacGetSLCCalPrims (void);
LOCAL vmsstat_t bsacInitCalPrim (const unsigned short primIndex,
								 const char * name_c);
LOCAL void bsacInitMD (void);
LOCAL void bsacInitMDBufs (void);
/*=============================================================================

  Name: bsacInitGlobals

  Abs:  initialize the global variables and structures

  Args: none

  Rem:  this function initializes all the global vars 

  Side: 

  Ret: none

==============================================================================*/
/**procedure**/
vmsstat_t  bsacInitGlobals(void)
{
  vmsstat_t iss =MICR_OKOK; 

  bsacInitMD();
  iss=bsacGetSLCPrims();    /* get all prims from SLC db for this ioc */
  iss=bsacGetSLCCalPrims(); /* get all prims from SLC db for cal (GADC, BPMP) */ 
  bsacPrepInitDevIDs();     /* get epics addr for all unit primid, unitid */
                            /* bsacPrepInitDevIDs must follow bsacGetSLCPrims() */
  bsacPrepInitMDEFAddr();     /* get epics addr for all epics MDEF records */ 
  /*bsacDataInitmdAddr();     init epics addr for bsacData functions */ 
  bsacPrepInitGoAddr(); 
  bsacInitMDBufs();
 
  return (iss);
}

/*-----------------------------------------------------------------------
  Name: bsacInitPrim

  Abs:  Determines if this IOC contains any primaries of variety passed in
        by argument, and if so, initializes 

  Args: None

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/
LOCAL vmsstat_t bsacInitPrim (const unsigned short primIndex,
							  const char * name_c)
{

  vmsstat_t iss =MICR_OKOK; 
  dbheader_ts * primdbunits_ps = NULL;
  unsigned short unitIndex;
  unsigned short unitCnt=0;
  unsigned char * nextUnit_p; /* byte pointer into dbunits_ps list */
  char * char_p;
  epicsBoolean no_units = epicsFalse;
  /*  size_t cnt = 0;*/

  /* init unitCnt to zero to indicate no prim units present */
  primCollect_as[primIndex].unitCnt = 0;

  /* init primary ASCII name */
  memcpy ( primCollect_as[primIndex].slcPrim_a, name_c, 5);

  /* get list to hold data for Prim units */
  if (!SUCCESS(iss = dblistalloc(4,20,(const dbheader_ts ** const)&primdbunits_ps)))
	{
	  DEBUGPRINT(DP_ERROR,bsacGlobFlag, ("bsacInitPrim: Error alloc'ing %s dbunits list\n", name_c));
	  iss=DB_BADPTR;		  
	}
  else {  
	/* get Prim units */
	if (SUCCESS(iss = dblunits(name_c,(dbheader_ts ** const)&primdbunits_ps)))
	  {
		if (primdbunits_ps->cur == 0)
		  { iss=DB_NOTALLUNIT; }
		else 
		  { unitCnt = primdbunits_ps->cur/sizeof (unsigned short); }
	  }
	else {
	  DEBUGPRINT(DP_DEBUG,bsacGlobFlag,
	  ("bsacInitPrim: The PRIM %s has no units\n", name_c ) );
      no_units = epicsTrue; /* this is not actually an error condition, keep this knowledge, 
							   but do not allow this routine to continue */
	}	
  }
  if (SUCCESS(iss)) { /* we have some Prim... */ 
	/* point nextUnit_p to first char of first unit */
    nextUnit_p = (unsigned char *)primdbunits_ps; 
    nextUnit_p += sizeof(dbheader_ts);
    /* store total # of units */ 
    primCollect_as[primIndex].unitCnt = unitCnt;
    /* now get the unit name and init chancnt */
    for (unitIndex=0; unitIndex<unitCnt; unitIndex++) {
	  DEBUGPRINT(DP_DEBUG,bsacGlobFlag,
	  ("bsacInitPrim: processing %s UNITINDEX %d out of %d\n", name_c, unitIndex, unitCnt));
	  /* store slc unit name in primCollect structure */
      memcpy(&primCollect_as[primIndex].unit_as[unitIndex].slcUnit4_a[0],
             nextUnit_p, 4);
      memcpy(&primCollect_as[primIndex].unit_as[unitIndex].slcUnit_a[0],
             nextUnit_p, 4);

	  /* null terminate for sscanf DEBUGGING*/
   	  char_p=strchr(&(primCollect_as[primIndex].unit_as[unitIndex].slcUnit_a[0]),32);
	  if (char_p !=0)  *char_p='\0';
	  else primCollect_as[primIndex].unit_as[unitIndex].slcUnit_a[4]= '\0';
	  DEBUGPRINT(DP_DEBUG,bsacGlobFlag,
	  ("bsacInitPrim: PRIM %s UNITSTR is %s-\n", name_c,
	    &(primCollect_as[primIndex].unit_as[unitIndex].slcUnit_a[0]) ));

	  /* store the unsigned short unit number for comparison to measdef msgs */
	  sscanf (primCollect_as[primIndex].unit_as[unitIndex].slcUnit_a, "%hd",
		&primCollect_as[primIndex].unit_as[unitIndex].unitnum);
	  /* init the chanCnt to 0; only chanNums for ARRYs will be filled later */
	  primCollect_as[primIndex].unit_as[unitIndex].chanCnt = 0;
	  /* init the chanCntr (used when recieving data) to 0 for now */
	  primCollect_as[primIndex].unit_as[unitIndex].chanCntr = 0;
      nextUnit_p += 4;
	}
  }else {
	/* now, if no_units make sure it returns OKOK */
	if (no_units==epicsTrue) iss = MICR_OKOK;
    /* be sure we release any memory in this case */
	if (primdbunits_ps)
	  iss = dblistfree(primdbunits_ps);

	return (iss);
  }

  /* read the SLC HSTA values for each unit */ 
  bsacHstaGetHSTAs(primIndex, name_c );

  /* now handle pulseids -which won't be in slc db */
  
  if (primdbunits_ps)
	iss = dblistfree(primdbunits_ps);

  return (iss);
  }

/*-----------------------------------------------------------------------
  Name: bsacGetARRYChans

  Abs:  Get Number of channels ("ADCN") for all ARRY units

  Args: None

  Rem: 

  Side: Place channel count into 
        primCollect_as[primIndex].unit_as[unitIndex].chanCnt

  Ret:  void
----------------------------------------------------------------------*/
LOCAL vmsstat_t bsacGetARRYChans (void)
{
  vmsstat_t iss =MICR_OKOK;
  dbheader_ts * dblist_ps = NULL;
  dbheader_ts * dbdata_ps = NULL;
  unsigned short unitIndex; 
  unsigned char * next_p;


  if (primCollect_as[BSACARRY_INDEX].unitCnt>0) {
	/* if ARRY units configured, see how many channels in ea unit */
	if (!SUCCESS(iss = dblistalloc(4,20,(const dbheader_ts ** const)&dblist_ps)))
	  {
		DEBUGPRINT(DP_ERROR,bsacGlobFlag, ("bsacGetARRYChans: Error alloc'ing dblist list\n"));
		return (iss=DB_BADPTR);		  
	  }
	if (!SUCCESS(iss = dblistalloc(4,20,(const dbheader_ts ** const)&dbdata_ps)))
	  {
		DEBUGPRINT(DP_ERROR,bsacGlobFlag, ("bsacGetARRYChans: Error alloc'ing dbdata list\n"));
		return (iss=DB_BADPTR);		  
	  }
	/* get all addrs for all ARRY:unit:ADCN slc entries */
    for (unitIndex=0; unitIndex< primCollect_as[BSACARRY_INDEX].unitCnt; unitIndex++) {
	  DEBUGPRINT(DP_DEBUG,bsacGlobFlag,
	  ("bsacGetARRYChans: processing dblist ARRY unit %d for ADCN Channel\n", unitIndex));
	  if (!SUCCESS(iss=dblist ((const char *)"ARRY", 
							   (const char *)primCollect_as[BSACARRY_INDEX].unit_as[unitIndex].slcUnit4_a,
							   (const char *)"ADCN",
							   (const dbheader_ts **const)&dblist_ps))){
		return(iss=DB_BADPTR);
	  }
	}
	if (!SUCCESS(iss=dblget ((const dbheader_ts * const)dblist_ps,
							 (dbheader_ts **const)&dbdata_ps))){
	  return(iss=DB_BADPTR);
	}

	next_p = (unsigned char *)dbdata_ps + sizeof(dbheader_ts);

	for (unitIndex=0; unitIndex< primCollect_as[BSACARRY_INDEX].unitCnt; unitIndex++) {
	  primCollect_as[BSACARRY_INDEX].unit_as[unitIndex].chanCnt= *(unsigned short *)next_p;
	  DEBUGPRINT(DP_DEBUG,bsacGlobFlag,
				 ("bsacGetARRYChans: ARRY UNITINDEX %hd has ADCN =%hd \n", 
				  unitIndex,
				  primCollect_as[BSACARRY_INDEX].unit_as[unitIndex].chanCnt ));
	  next_p += sizeof(unsigned short);
	}  		
  }

  if (dblist_ps)
	iss = dblistfree(dblist_ps);
  if (dbdata_ps)
	iss = dblistfree(dbdata_ps);

  return (iss);
}
/*-----------------------------------------------------------------------
  Name: bsacGetSLCPrims

  Abs:  Determines which primaries and units are configured for this IOC 

  Args: None

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/
LOCAL vmsstat_t bsacGetSLCPrims (void)
{
  unsigned short prim;
  vmsstat_t iss =MICR_OKOK; 

  if (!(SUCCESS(iss = bsacInitPrim(BSACBPMS_INDEX, "BPMS")))) return iss;
  if (!(SUCCESS(iss = bsacInitPrim(BSACTORO_INDEX, "TORO")))) return iss;
  if (!(SUCCESS(iss = bsacInitPrim(BSACARRY_INDEX, "ARRY")))) return iss;

  for (prim=0; prim<BSACPRIM_MAX; prim++) {
	DEBUGPRINT(DP_DEBUG,bsacGlobFlag,
			   ("bsacGetSLCPrims: PRIMINDEX %hd has dblistHSTAs_ps =%lu \n", 
				prim, (unsigned long)primCollect_as[prim].dblistHSTAs_ps ));
  }

  /* now fill in the channel count for each ARRY unit */
  if (!(SUCCESS(iss = bsacGetARRYChans()))) return iss;
  /* handle pulse ids - not in slc db */
  /* init unitCnt to zero to indicate no prim units present */
  primCollect_as[BSACPUID_INDEX].unitCnt = 1;
  primCollect_as[BSACPUID_INDEX].unit_as[0].unitnum = 1;
  /* init primary ASCII name */
  memcpy ( primCollect_as[BSACPUID_INDEX].slcPrim_a, "PUID", 5);

  return (iss);
}
/*-----------------------------------------------------------------------
  Name: bsacInitCalPrim

  Abs:  Determines if this IOC contains any primaries of variety passed in
        by argument, and if so, initializes bsacCalPrimCnt[BSACPRIM_MAX];

  Args: None

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/
LOCAL vmsstat_t bsacInitCalPrim (const unsigned short primIndex,
								 const char * name_c)
{

  vmsstat_t iss =MICR_OKOK; 
  dbheader_ts * primdbunits_ps = NULL;
  unsigned short unitCnt=0;

  /* init unitCnt to zero to indicate no prim units present */
  calPrimCnt[primIndex] = 0;

  /* get list to hold data for Prim units */
  if (!SUCCESS(iss = dblistalloc(4,20,(const dbheader_ts ** const)&primdbunits_ps)))
	{
	  DEBUGPRINT(DP_ERROR,bsacGlobFlag, 
				 ("bsacInitPrim: Error alloc'ing %s dbunits list\n", name_c));
	  iss=DB_BADPTR;		  
	}
  else {  
	/* get Prim units */
	if (SUCCESS(iss = dblunits(name_c,(dbheader_ts ** const)&primdbunits_ps)))
	  {
		if (primdbunits_ps->cur == 0)
		  { iss=DB_NOTALLUNIT; }
		else 
		  { unitCnt = primdbunits_ps->cur/sizeof(unsigned short); }
	  }
  }
  if (SUCCESS(iss)) { /* we have some Prim... */ 
    /* store total # of units */ 
    calPrimCnt[primIndex] = unitCnt;
	DEBUGPRINT(DP_ERROR,bsacGlobFlag, 
				 ("bsacInitPrim: calPrimCnt=%hd for pindex = %hd\n", 
				  (unsigned int)calPrimCnt[primIndex], (unsigned int) primIndex));
  }

  if (primdbunits_ps)
	iss = dblistfree(primdbunits_ps);

  return (iss);
  }

/*-----------------------------------------------------------------------
  Name: bsacGetSLCCalPrims

  Abs:  Determines which primaries and units are configured for this IOC 

  Args: None

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/
LOCAL vmsstat_t bsacGetSLCCalPrims (void)
{
  vmsstat_t iss =MICR_OKOK; 

  iss = bsacInitCalPrim(BSACBPMS_INDEX, "BPMP");
  iss = bsacInitCalPrim(BSACTORO_INDEX, "TORO");
  iss = bsacInitCalPrim(BSACARRY_INDEX, "GADC");
  return (iss);
}
/*-----------------------------------------------------------------------
  Name: bsacInitMD

  Abs: initializes the bsacMD_as structure

  Args: None

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/
LOCAL void bsacInitMD (void)
{
  unsigned short md;

  for (md = 0; md < MDEF_MAX; md++) {
	bsacMD_as[md].timeout      = NO_TIMEOUT;
    bsacMD_as[md].timeoutCntr  = 0;
    bsacMD_as[md].totalByteCnt = 0;
    bsacMD_as[md].func         = 0;
    bsacMD_as[md].abort        = epicsFalse;
    bsacMD_as[md].measbits     = 0;
    bsacMD_as[md].measprep     = epicsFalse;             
    bsacMD_as[md].useUlist     = epicsFalse;            
    bsacMD_as[md].firstPulse   = epicsTrue;           
    bsacMD_as[md].ndatblks     = 0;          
    memset(&bsacMD_as[md].username,0,5);                 

  }
}
/*-----------------------------------------------------------------------
  Name: bsacInitMDBufs

  Abs: initializes the bsacMDBuf_as structure. This structure keeps track
       of the current replydata buffers for prep and wire request

  Args: None

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/
LOCAL void bsacInitMDBufs (void)
{
  unsigned short md;
  unsigned short bufs;

  for (md = 0; md < MDEF_MAX; md++) {
	bsacMDBufs_as[md].bufCount= 0;
	bsacMDBufs_as[md].nextBuf = 0;
	for (bufs = 0; bufs < MD_DATABUFS_MAX; bufs++) {
	  bsacMDBufs_as[md].bufs_as[bufs].totwc   = 0;
	  bsacMDBufs_as[md].bufs_as[bufs].partwc  = 0;
	  bsacMDBufs_as[md].bufs_as[bufs].buf_ps  = NULL;
      bsacMDBufs_as[md].bufs_as[bufs].data_p = NULL;
	}
  }
}
