/*=============================================================================
 
  Name: bsacHdlr.c
        bsacHdlrThread   - main routine for bsacHdlr thread
        cleanup         - reset globals, release resources before terminating

  Abs:  bsacHdlr contains the bsacHdlrThread() thread that receives BPMO_
        function code messages from the msgRecv thread, acts on them,
		then sends a reply to the msgSend thread.
        reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | BSAC Service Design

  Auth: 10 jul 2005, dfairley    created 
  Rev:  

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest) 
        10 jul 2005, dfairley    created 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int bsacHdlrFlag = DP_DEBUG;
#endif


/* c includes */
#include <string.h>        /* memcpy */

/* shared slc micro includes */
#include "bpmfunc.hc"      /* bpmo_xxx func codes */
#include "micrdef.hc"      /* MICR_EXNET_ codes */
#include "slcMsg.h"        /* IOC_STOP, BSAC_* */
#include "msgfunc.hc"      /* RESPONSE for func code */

/* slcIOC utilities */
#include "slcGlob.h"       /* slc global vars, structs, types */
#include "slcUtil.h"       /* slcPrintMsgheader*/
#include "msgQUtil.h"      /* messageQSend, messageQRecv */
#include "cvtUtil.h"       /* slcSSwap, slcLSwap, etc */
#include "slcCmlog.h"      /* slcCmlogLogMsg */
#include "timeUtil.h"      /* timeGetVMS, timeDiffVMS10usec */
#include "slcAsync.h"      /* for ioc async msg asyncmail_ts */

/* bsacHdlr include */
#include     "bsacGlob.h" /* bsac globals */
#include     "bsacHdlr.h" /* bsacHdlrDiags_ts */
#include     "bsacHsta.h" /* bsacHsta utils */
#include     "bsacPrep.h" /* bsacPrep() */
#include     "bsacCal.h"  /* bsacCal */
#include     "bsacAbort.h" /* bsacAbortMeas() */
#include     "bsacData.h" /* bsacData, bsacDone */
#include     "bsacBufAcqData.h" /* bsacBufAcqData, bsacBufAcqDone */
#include     "bpmodef.hc" /* BPMO_OKOK */

/* local function prototypes */
LOCAL void       cleanup(void);
LOCAL void       bsacHdlrStop(void);
LOCAL void       bsacHdlrTimeouts(void);
LOCAL vmsstat_t  bsacHdlrClrScp(const msgmail_ts* rMsg_ps);
LOCAL vmsstat_t  bsacHdlrMeasStop(const msgmail_ts* rMsg_ps);
LOCAL vmsstat_t  bsacHdlrGetRemDat(const msgmail_ts* rMsg_ps);
LOCAL vmsstat_t  bsacHdlrBSACMsg(const msgmail_ts* rMsg_ps,
								 const epicsBoolean dataMsg);
LOCAL vmsstat_t  bsacHldrSyncCTRLVal(const msgmail_ts* rMsg_ps);
LOCAL vmsstat_t  bsacHdlrIocAsync(const msgmail_ts* rMsg_ps);
/* bsacHdlr diagnostics global variable */
bsacHdlrDiags_ts bsacHdlrDiags_s = { {0, 0}, 0, {0, 0}, 0, 0, 0, 0, 0};

/*=============================================================================

  Name: bsacHdlrThread

  Abs:  The bsac handler thread

  Args: none

  Rem: This thread accepts and acts on FUNC_TEST messages from the SLC control system

  side: 

  Ret:  none

==============================================================================*/
/**procedure**/
void  bsacHdlrThread(void)
{
  vmsstat_t     iss;
  msgmail_ts*   rMsg_ps = NULL;   /* read msg buffer pointer */
  slcThread_ts* bsacHdlr_ps = &slcThreads_as[bsacHdlr]; /* pointer to my thread struct */ 

  /* init the diagnostics structure */
  memset(&bsacHdlrDiags_s, 0, sizeof(bsacHdlrDiags_s) );

  /* initialize all the bsac globals */
  if (!SUCCESS(bsacInitGlobals()) ) return;

  /* create the message queue id */
  if ( !SUCCESS(messageQCreate(bsacHdlr, MAX_QUEUED_MSGS, MAX_MSGPTR_BYTES)) ) {
	cleanup();
	return;
  }

  /* set up the CA channels for CTRL records after the message queue is created */
  if (!SUCCESS(bsacHstaInitChannels())) {
    DEBUGPRINT (DP_FATAL, bsacHdlrFlag, ("BSAC_H:  bsacHstaInitChannels failed\n"));
    cleanup();
	return;
  }

  bsacHdlr_ps->active = epicsTrue;
  DEBUGPRINT(DP_INFO, bsacHdlrFlag, ("BSAC_H: becoming active\n"));
  epicsTimeGetCurrent(&bsacHdlrDiags_s.lastStartTime);

  /* main loop of the bsac hdlr thread, wait at message queue for incoming 
	 messages, handle it, then send a reply to the msgSend thread
  */
  while(bsacHdlr_ps->active)
  {
	/* read a message at the message queue.*/
	if (SUCCESS(iss = msgQMsgRecv(bsacHdlr, 
								  &bsacHdlrDiags_s.lastReqTime, 
								  (void*)&rMsg_ps) ))
	{
	  bsacHdlrDiags_s.reqsSinceStart++;
      /* debug print the message header */
	  /*slcPrintMsgheader(&rMsg_ps->msgheader, bsacHdlrFlag);*/
	  /*DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, 
		("BSAC_H:incoming msg func code = %x\n", rMsg_ps->msgheader.func));*/
	  switch(rMsg_ps->msgheader.func){
	  case IOC_STOP:
		bsacHdlrStop();
		break;
	  case IOC_ASYNC:
        iss = bsacHdlrIocAsync(rMsg_ps);
		break;
	  case BPMO_CALB_PREP:
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received BPMO_CALB_PREP\n"));
		iss = bsacCalPrep(rMsg_ps);
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Handled BPMO_CALB_PREP\n"));
		break;
	  case BPMO_CALB_PROC:
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received BPMO_CALB_PROC....\n"));
		/*iss = bsacCalProc(rMsg_ps);*/
		/* BPMO_CALB_PROCs are sent unsolicited */
		break;
	  case BPMO_MEAS_PREP:
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received BPMO_MEAS_PREP\n"));
		if (SUCCESS(iss = bsacPrep(rMsg_ps, epicsTrue))){
		}
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Handled BPMO_MEAS_PREP\n"));
		break;
	  case BPMO_MEAS_PROC:
		/* shouldn't get this msg; we send after data is collected */
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received BPMO_MEAS_PROC...\n"));
		break;
	  case BPMO_WIRE_PREP:
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received BPMO_WIRE_PREP\n"));
		if (SUCCESS(iss = bsacPrep(rMsg_ps, epicsFalse))){
		}
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Handled BPMO_WIRE_PREP\n"));
		break;
	  case BPMO_CLR_SCP:
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received BPMO_CLR_SCP\n"));
	    iss = bsacHdlrClrScp(rMsg_ps); 
		/* msgQSlcIssReply (rMsg_ps, iss)  no reply to CLR SCP */
		break;
	  case BPMO_MEAS_STOP:
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received BPMO_MEAS_STOP\n"));
		iss = bsacHdlrMeasStop(rMsg_ps);
		DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, 
				 ("BSAC_H: Sending mgQSlcIssReply for BPMO_MEAS_STOP\n")); 
		msgQSlcIssReply (rMsg_ps, iss);
		break;
	  case BPMO_GETREMDAT:
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received BPMO_GETREMDAT\n"));
		iss = bsacHdlrGetRemDat(rMsg_ps);
		break;
	  case BSAC_DATA:
		DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, 
				 ("BSAC_H: Received BSAC_DATA\n"));
		/* go determine if measprep or wire scan */
		if (SUCCESS(iss = bsacHdlrBSACMsg(rMsg_ps, epicsTrue))){
		}
		DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, 
				 ("BSAC_H: Handled BSAC_DATA\n"));
		break;
	  case BSAC_DONE:
		DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, 
				 ("BSAC_H: Received BSAC_DONE\n"));
		/* go determine if measprep or wire scan */
		if (SUCCESS(iss = bsacHdlrBSACMsg(rMsg_ps, epicsFalse))){
		}
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Handled BSAC_DONE\n"));
		break;
	  case BSAC_SYNC_CTRL:
		DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, 
				 ("BSAC_H: Received BSAC_SYNC_CTRL\n"));
		/* synchronize CTRL with HSTA*/
		bsacHldrSyncCTRLVal(rMsg_ps);
		break;
	  case BPMO_ACCEPT_PUBLIC:
		/* just send good status back */
		msgQSlcIssReply ((void *)rMsg_ps, BPMO_OKOK);
		break;
	  default:
		/* unexpected message */
		DEBUGPRINT(DP_INFO, bsacHdlrFlag, 
				 ("BSAC_H: Received unexpected msg func code = %x\n", 
				  rMsg_ps->msgheader.func));
		break;
	  } /* switch */

	  /* release the receive buffer */
	  msgQFreeBuffer(sizeof(msgmail_ts), rMsg_ps);
	} /* if successful read */

    /* check the stop flag just in case */
	if (bsacHdlr_ps->stop)
	  bsacHdlr_ps->active = epicsFalse;

  } /* while active */

  /* active not set, time to shut down */
  cleanup();
  DEBUGPRINT(DP_INFO, bsacHdlrFlag, ("BSAC_H: shutdown\n"));
}
/*=============================================================================

  Name: bsacHdlrBSACMsg()

  Abs:  looks at incoming msg header for md #, indexes into bsacMD_as[md].measprep
        to determine whether this is a measprep or wireprep in order to call the
		correct data handling routine. Input arg dataMsg boolean is checked to see
		whether this is a BSAC_DATA or BSAC_DONE message

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming BSAC_** message
		const epicsBoolean  dataMsg     read only   T if BSAC_DATA; F if BSAC_DONE

  Rem:  

  side: Call appropriate measprep/wire data routines

  Ret: vmsstat_t iss the vms style status code

===============================================================================*/
LOCAL vmsstat_t bsacHdlrBSACMsg(const msgmail_ts* rMsg_ps,
								const epicsBoolean dataMsg)
{
  vmsstat_t iss = BPMO_OKOK;
  bsacMsgHdr_ts * hdr = NULL;
  unsigned short md;

    /* interpret the header */
  hdr = (bsacMsgHdr_ts*)&rMsg_ps->reqdata;
  md = hdr->md-1; /* the EPICS md is one more than slc threads md */

  if (dataMsg==epicsTrue) {
	if (bsacMD_as[md].measprep==epicsTrue){	  
	iss = bsacData(rMsg_ps);
	}else {
	  iss = bsacBufAcqData(rMsg_ps);
	}
  }else {
	if (bsacMD_as[md].measprep==epicsTrue){	  
	iss = bsacDone(rMsg_ps);
	}else {
	  iss = bsacBufAcqDone(rMsg_ps);
	}
  }

  return (iss);
}


/*=============================================================================

  Name: bsacHdlrTimeouts()

  Abs:  decrement timeout counters of all active mdefs.  Abort mdefs that have
        timed out on this decrement.

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		none

  Rem:  

  side: the bsacMD_as[].timeoutCntr is decremented.

  Ret: none

===============================================================================*/
LOCAL void bsacHdlrTimeouts(void)
{
  short md;

  for (md = 0; md < MDEF_MAX; md++) {
	if ( (bsacMD_as[md].timeout!=NO_TIMEOUT) && 
		 (bsacMD_as[md].timeoutCntr>0) &&
		 (bsacMD_as[md].abort == epicsFalse) ) {

	  DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, ("BSAC_H: timing down md=%d, Cntr=%d\n", 
										 md, bsacMD_as[md].timeoutCntr));

	  bsacMD_as[md].timeoutCntr--;
      if (bsacMD_as[md].timeoutCntr<=0) bsacAbortTimeoutMeas(md);
	}
  }
}

/*=============================================================================

  Name: bsacHdlrMeasStop(const short measid)

  Abs:  cancel the measurement associated with this measid
`
  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming MEAS_STOP message    

  Rem:  

  side: if it's already aborted or being aborted, then return OK

  Ret: vmsstat_t iss the vms style status code

===============================================================================*/
LOCAL vmsstat_t bsacHdlrMeasStop(const msgmail_ts* rMsg_ps)
{
  vmsstat_t iss = BPMO_OKOK;
  char scpName[5];

  /* get the name of the scp */
  memset (&scpName[0], 0, 5);
  memcpy (&scpName[0], &rMsg_ps->msgheader.source[0],4);

  /* convert the i_seg */
  cvtSSwap(&rMsg_ps->reqdata[0]);

  /* now stop the measurement */
  iss = bsacAbortMeasStop(scpName, rMsg_ps->reqdata[0]);
  DEBUGPRINT(DP_INFO, bsacHdlrFlag, ("bsacHdlrMeasStop: Leaving\n"));
  return (iss);
}

/*=============================================================================

  Name: bsacHdlrClrScp()

  Abs: clear all measurement definitions assigned by this scp

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL vmsstat_t bsacHdlrClrScp(const msgmail_ts* rMsg_ps)
{
  vmsstat_t iss = BPMO_OKOK;
  char scpName[5];

  memset (&scpName[0], 0, 5);
  memcpy (&scpName[0], &rMsg_ps->msgheader.source[0],4);

  iss = bsacAbortClrScp(scpName);

  return (iss);
}


/*=============================================================================

  Name: bsacHdlrSyncCTRLVal()

  Abs: clear all measurement definitions assigned by this scp

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL vmsstat_t bsacHldrSyncCTRLVal(const msgmail_ts* rMsg_ps)
{
  vmsstat_t iss = MICR_OKOK;
  bsacUnits_ts *unit_ps = NULL;
  unsigned short ctrlVal = rMsg_ps->reqdata[2];
  memcpy(&unit_ps, &rMsg_ps->reqdata[0],4);
  DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, 
			 ("BSAC_H syncCTRL:synchronizing unit_ps=%lu, ctrlVal = %d\n",
			  (unsigned long)unit_ps, ctrlVal) );
  if (!SUCCESS(bsacHstaSyncCTRLVal(unit_ps,ctrlVal) ) )
	  DEBUGPRINT(DP_DEBUG, bsacHdlrFlag, 
			   ("BSAC_H syncCTRL: cannot sync unit_ps=%lu, ctrlVal = %d\n",
			   (unsigned long)unit_ps, ctrlVal) );
  return (iss ); 
}


/*=============================================================================

  Name: bsacHdlrIocAsync()

  Abs:

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL vmsstat_t bsacHdlrIocAsync(const msgmail_ts* rMsg_ps)
{
  vmsstat_t iss = MICR_OKOK;
  asyncmail_ts asyncmail_s;

  memcpy(&asyncmail_s, rMsg_ps->reqdata, sizeof(asyncmail_s)); 
  if ( strncmp(asyncmail_s.cyclName, "TIME", 4)==0 ) {
	bsacHdlrTimeouts();
  } else {
	if (!SUCCESS(iss = (bsacHstaCheckHSTAs())) )
	  DEBUGPRINT(DP_ERROR, bsacHdlrFlag, 
				 ("BSAC_H: failed checking HSTAs\n"));
  }
  return(iss);
}

/*=============================================================================

  Name: bsacHdlrGetRemDat()

  Abs: clear all measurement definitions assigned by this scp

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  

  side: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL vmsstat_t bsacHdlrGetRemDat(const msgmail_ts* rMsg_ps)
{
  vmsstat_t iss = BPMO_OKOK;
  char scpName[5];

  memset (&scpName[0], 0, 5);
  memcpy (&scpName[0], &rMsg_ps->msgheader.source[0],4);

  iss = bsacDataGetRemDat(scpName);
  return (iss);
}

/*=============================================================================

  Name: bsacHdlrStop()

  Abs:  terminate this thread by setting the active flag to false

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		none

  Rem:  setting active to false will break the thread out of it's loop and cause 
        it to begin the termination process.

  side: the slcThreads_as[bsacHdlr].active flag is set to false.

  Ret: none

===============================================================================*/
LOCAL void bsacHdlrStop(void)
{
  slcThreads_as[bsacHdlr].active = epicsFalse;  
}


/*==============================================================================

  Name: cleanup()

  Abs:  cleanup the globals for this thread before shutdown

  Args: none

  Rem: clean up the slcThreads_as struct for this thread
       and release any resources 

  side: 

  Ret: none

==============================================================================*/
LOCAL void cleanup(void)
{
  bsacHstaCleanup();
  messageQRelease(bsacHdlr);
  bsacAbortAllActiveMDefs();
  slcThreads_as[bsacHdlr].active = epicsFalse;
  slcThreads_as[bsacHdlr].tid_ps = NO_THREAD;
}

