/*=============================================================================
 
  Name: bsacHsta.c
         
  Abs:  The bsacHsta module contains utilities to keep 
        SLC and Epics database HSTA / dev CTRL (now ACCESS!)  information 
		sychronized by using EPICS channel access to read/write/monitor 
		the EPICS device CTRL records. 

		Upon initialization, the bsacHdlr thread compares the CTRL values 
		to the SLC database HSTA values.
		The EPICS values are used as 'master' upon initialization.  From then
		on the HSTA and CTRL values reflect the last control command from either
		the SCP or from an EPICS client. 

		The epics CTRL value is monitored.  When the CTRL value is changed by an
		epics user, a callback routine is executed that will send an CTRL updated
		message	to the bsacHdlr thread. When the message is received the bsacHdlr
		will immediately update the hsta_ctrl global, and update the SLC database
		with a dbupdate request.

		The SLC HSTA value is polled.  The cstrAsync thread sends a message to the
		bsacHdlr periodically to check all the unit HSTAs.  If any one has changed
		from the hsta_ctrl global value, then that hsta_ctrl value, and the epics
		CTRL record are both updated immediately.

		The bsacHdlrThread is essentially a channel access client, but we
		rely on the assumption that when a channel access client and server are 
		in the same 'address space and host' the client calls bypass the normal 
		server and work directly with the server tool component.

  Auth: 10 jul 2005, dfairley    created 
  Rev:  

NOTE: the CTRL PV IS RENAMED TO ACCESS, but variable names, comments will still
refer the CTRL.  Now when you read CTRL, think ACCESS!!
-----------------------------------------------------------------------------*/


#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
       13-jul-2007, dfairley renamed CTRL pv to ACCESS but will keep the variable
	     names as CTRL - may be confusing several years from now...oh well.
       27-apr-2007, dfairley started LOCA2 fixes

=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
int bsacHstaFlag = DP_ERROR;           /* Set to DP_INFO for production */ 
#endif

/* c includes */
#include <string.h>         /* memcpy */

#include    "cadef.h"      /* channel access */
#include    "dbDefs.h"
#include    "epicsMutex.h" /* mutex utilities */

/* slcIOC utilities */
#include    "slcCmlog.h"   /* slcCmlogLogMsg */
#include    "msgQUtil.h"   /* messageQSend  */
#include    "slcMsg.h"     /* BSAC_ABORT */
#include    "micrdef.hc"   /* MICR_OKOK */
#include    "cvtUtil.h"    /* conversion utils */

/* SLC database utilities */
#include    "dbdef.hc"     /* db error codes */
#include    "dblist.h"     /* dblist macros and utils */
#include    "dbl.h"
#include    "dbupdate.h"   


/* bsac includes */
#include    "bpmodef.hc"
#include    "bsacGlob.h"   
#include    "bsacHsta.h" 
#include    "bpmstat.h"   /* HSTA values */

/* bsachsta typedefs */

/* channel access channel information for caMonitors */
typedef struct {
  chid chId;
  evid evId;
  int  chStatus;
} hstaUnit_ts; 

/* hsta specific info for units */
typedef struct {
  hstaUnit_ts   unit_as[UNITS_MAX]; /* array of primary units */
} hstaPrimStruct_ts;


/* static vars */
LOCAL hstaPrimStruct_ts hstaCollect_as[BSACPRIM_MAX];

/* local function prototypes */
LOCAL vmsstat_t initCaChannel(const unsigned short pIndex, 
							  const unsigned short uIndex,
							  const char * unit_a);  
LOCAL vmsstat_t getChannel(const char * name_c,
						   const unsigned short pIndex, 
						   const unsigned short uIndex);
LOCAL vmsstat_t setCaMonitor(const unsigned short pIndex,
							 const unsigned short uIndex);

LOCAL vmsstat_t bsacCheckUnitHSTAs (const unsigned short pIndex);
LOCAL vmsstat_t syncHSTAVal(const unsigned short pIndex,
							const unsigned short uIndex,
							const unsigned long  hstaVal);
LOCAL vmsstat_t updateEpicsCTRL(const unsigned short pIndex,
								const unsigned short uIndex,
								const unsigned short newCtrl);

LOCAL void bsacHstaCTRLCallback(struct event_handler_args eha);
LOCAL void cvtHSTAToCTRL(const unsigned short pIndex,
						 const unsigned long hsta,
						 unsigned short *ctrl_p);
LOCAL void cvtCTRLToHSTA(const unsigned short pIndex,
						 const unsigned short ctrl,
						 const unsigned long oldhsta,
						 unsigned long *newhsta_p);

/*-----------------------------------------------------------------------
  Name: bsacHstaGetHSTAs

  Abs:  initialize the dblists for synchronizing SLC HSTAs with epics CTRL
        records.  We initialize a list of ALL unit HSTAs per primary 
		in the primCollect_as[p].dblistHSTAs_ps,pointer, 
		and a separate list for each individual unit, in 
		primCollect_as[p].unit_as[u].dblistHSTA_ps 
		The primary index is also stored in the unit structure so that it
		can be used when converting HSTA to CTRL and vice versa. 

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        unsigned short      pIndex      read only  primary device index
        char *              name_c      read only  primary name

  Rem:  initializes dblists and does the first dblget for HSTA values

  Side:  assigns the following globals:
         primCollect_as[primIndex].unit_as[unitIndex].dblistHSTA_ps
         primCollect_as[primIndex].unit_as[unitIndex].primIndex
		 primCollect_as[primIndex].dblistHSTAs_ps

  Ret:  vmsstat_t vms style return code
----------------------------------------------------------------------*/
vmsstat_t bsacHstaGetHSTAs (const unsigned short primIndex,
							const char * name_c)
{
  vmsstat_t iss =MICR_OKOK;
  dbheader_ts * dblist_ps = NULL;  /* list for all units */
  dbheader_ts * dbdata_ps = NULL;  /* list for all unit data */
  dbheader_ts * dbunitl_ps = NULL; /* list for individual unit */

  unsigned short unitIndex; 
  unsigned char * next_p;

  /* init the dblistHSTA_ps */
  primCollect_as[primIndex].dblistHSTAs_ps = NULL;

  if (primCollect_as[primIndex].unitCnt>0) {
	/* if  units configured, see how many channels in ea unit */
	if (!SUCCESS(iss = dblistalloc(4,20,(const dbheader_ts ** const)&dblist_ps)))
	  {
		DEBUGPRINT(DP_ERROR,bsacHstaFlag, 
				   ("bsacGetUnitHSTA: Error alloc'ing dblist list\n"));
		return (iss=DB_BADPTR);		  
	  }
	if (!SUCCESS(iss = dblistalloc(4,20,(const dbheader_ts ** const)&dbdata_ps)))
	  {
		DEBUGPRINT(DP_ERROR,bsacHstaFlag, 
				   ("bsacGetUnitHSTA: Error alloc'ing dbdata list\n"));
		return (iss=DB_BADPTR);		  
	  }
	/* collect all addrs for all prim:unit:HSTA slc entries into dblist_ps */
    for (unitIndex=0; unitIndex< primCollect_as[primIndex].unitCnt; unitIndex++) {
	  if (!SUCCESS(iss=dblist ((const char *)name_c, 
				 (const char *)primCollect_as[primIndex].unit_as[unitIndex].slcUnit4_a,
							   (const char *)"HSTA",
							   (const dbheader_ts **const)&dblist_ps))){
		DEBUGPRINT(DP_ERROR,bsacHstaFlag, 
				   ("bsacGetUnitHSTA: Error dblist for HSTA\n"));
		return(iss=DB_BADPTR);
	  }

	  /* get the lists for this individual unit */
	  primCollect_as[primIndex].unit_as[unitIndex].dblistHSTA_ps = NULL;
	  dbunitl_ps = NULL;
	  /* allocate the list */
	  if (!SUCCESS(iss = dblistalloc(4,20,(const dbheader_ts ** const)&dbunitl_ps)))
	  {
		DEBUGPRINT(DP_ERROR,bsacHstaFlag, 
				   ("bsacGetUnitHSTA: Error alloc'ing dblist list\n"));
		return (iss=DB_BADPTR);		  
	  }
	
	  /* now fill the list for this unit */
	  if (!SUCCESS(iss=dblist (
			   (const char *)primCollect_as[primIndex].slcPrim_a, 
			   (const char *)primCollect_as[primIndex].unit_as[unitIndex].slcUnit4_a,
			   (const char *)"HSTA",
			   (const dbheader_ts **const)&dbunitl_ps))){
		return(iss=DB_BADPTR);
	  }
	  /* store the dblist for this unit */
	  primCollect_as[primIndex].unit_as[unitIndex].dblistHSTA_ps = dbunitl_ps;
	  /* store the primIndex into the unit structure, need this for HSTA synch */
	  primCollect_as[primIndex].unit_as[unitIndex].primIndex = primIndex;
	}



    /* now get all the HSTAs for this primary and store in hsta_ctrl */
	if (!SUCCESS(iss=dblget ((const dbheader_ts * const)dblist_ps,
							 (dbheader_ts **const)&dbdata_ps))){
	  return(iss=DB_BADPTR);
	}
    /* convert all the HSTAs into CTRL values and store them in hsta_ctrl */
	next_p = (unsigned char *)dbdata_ps + sizeof(dbheader_ts);

	for (unitIndex=0; unitIndex< primCollect_as[primIndex].unitCnt; unitIndex++) {
	  bsacHstaStoreHSTAAsCTRL( primIndex, unitIndex, *(unsigned long *)next_p );
	  DEBUGPRINT(DP_INFO,bsacHstaFlag,
				 ("bsacGetUnitHSTA: PRIMINDEX %hd UNITINDEX %hd has HSTA = %lu CTRL =%hd \n", 
				  primIndex, unitIndex,
				  *(unsigned long *)next_p,
				  primCollect_as[primIndex].unit_as[unitIndex].hsta_ctrl ));
	  next_p += sizeof(unsigned long);
	}  		
  }
  
  /* store the dblist for this primary */
  primCollect_as[primIndex].dblistHSTAs_ps = dblist_ps;

  /* free the dbdata list */
  if (dbdata_ps)
	iss = dblistfree(dbdata_ps);

  return (iss);
}

/*==============================================================================

  Name: bsacHstaInitChannels()

  Abs:  create a channel access context, and create the channel to 
        each unit's CTRL record. assign a monitor to each.

  Args: none

  Rem:  this routine uses the previously initialized primCollect_as global
        to search through all units and create a channel to each unit's 
		epics CTRL record.  It then assigns a monitor to each channel.

  side: the bsacInitGlobals must return successful before this routine is called.

  Ret: vmsstat_t iss the vms style return code

  ==============================================================================*/
vmsstat_t bsacHstaInitChannels(void)    
{
  unsigned short pIndex;
  unsigned short uIndex;
  vmsstat_t iss = MICR_OKOK;
  int status = ECA_NORMAL;

  /* create a channel access context */
  if ((status = (ca_context_create (ca_enable_preemptive_callback )))!=ECA_NORMAL )
    DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
				("BSAC_H:ca_create_context failure status = %d\n", status));
  
  for (pIndex=0; pIndex<BSACPRIM_MAX; pIndex++) {

	/* create channels to all the CTRL records for this primary's  units */
	for (uIndex=0; uIndex<primCollect_as[pIndex].unitCnt; uIndex++) {
      /* create a channel to epics CTRL record for this unit*/
	  hstaCollect_as[pIndex].unit_as[uIndex].chId = NULL;
	  hstaCollect_as[pIndex].unit_as[uIndex].evId = NULL;
	  iss = (initCaChannel(pIndex, uIndex,
						  (const char *)primCollect_as[pIndex].unit_as[uIndex].slcUnit_a));
	}
  }
  return (iss);
}

/*=============================================================================

  Name: initCaChannel

  Abs:  create a CA channel to the EPICS PV named, and set a CA Monitor on 
        then channel

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        unsigned short      pIndex      read only  primary device index
        unsigned short      uIndex      read only  unit index

  Rem:  dfairley added code to check for a second LOCA name

  Side: the primCollect_as[p].unit_as[u].chId is assigned

  Ret:  the vms style status code
   

  MOD: 13-jul-2007, dfairley   The CTRL pv is now called ACCESS
==============================================================================*/
vmsstat_t initCaChannel(const unsigned short pIndex, 
						const unsigned short uIndex,
						const char * unit_a)  
{
  char pvname_ca[80];   /* Epics PV channel name */
  char pvname2_ca[80];   /* Epics PV channel name */
  vmsstat_t iss = MICR_OKOK;
  
  switch (pIndex) {
  case BSACBPMS_INDEX:
	sprintf (pvname_ca, "BPMS:%s:%s:ACCESS.VAL", slcGetLOCAName(), unit_a);
	iss = getChannel(pvname_ca, pIndex, uIndex);
    if ( !(SUCCESS(iss)) ) {
	  if (slcGetLOCA2Name()!=NULL) {
		sprintf (pvname2_ca, "BPMS:%s:%s:ACCESS.VAL", slcGetLOCA2Name(), unit_a);
		iss = getChannel(pvname2_ca, pIndex, uIndex);
		if (SUCCESS(iss)) 
		  setCaMonitor(pIndex, uIndex);
		else
		  DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
			 ("BSAC_H:initCaChannel failure on connect for %s and %s\n", pvname_ca, pvname2_ca));
	  } 
	}
	else setCaMonitor(pIndex, uIndex);
	break;
  case BSACTORO_INDEX:
	sprintf (pvname_ca, "TORO:%s:%s:ACCESS.VAL", slcGetLOCAName(), unit_a);
	iss = getChannel(pvname_ca, pIndex, uIndex);
    if (!(SUCCESS(iss)) ) {
	  if (slcGetLOCA2Name()!=NULL)  {
		sprintf (pvname2_ca, "TORO:%s:%s:ACCESS.VAL", slcGetLOCA2Name(), unit_a);
		iss = getChannel(pvname2_ca, pIndex, uIndex);
		if (SUCCESS(iss)) setCaMonitor(pIndex, uIndex);
		else
		  DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
			 ("BSAC_H:initCaChannel failure on connect for %s and %s\n", pvname_ca, pvname2_ca));
	  }
	} 
	else setCaMonitor(pIndex, uIndex);
	break;
	/* special case, ARRY uses the slc IOC name */
  case BSACARRY_INDEX:
	sprintf (pvname_ca, "ARRY:%s:%s:ACCESS.VAL", slcGetIOCName(), unit_a);
	iss = getChannel(pvname_ca, pIndex, uIndex);
    if ( SUCCESS(iss)) setCaMonitor(pIndex, uIndex);
		else
		  DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
			 ("BSAC_H:initCaChannel failure on connect for %s \n", pvname_ca));
	break;
  case BSACWIRE_INDEX:
	sprintf (pvname_ca, "WIRE:%s:%s:ACCESS.VAL", slcGetLOCAName(), unit_a);
	iss = getChannel(pvname_ca, pIndex, uIndex);
    if (!(SUCCESS(iss)) ) {
      if (slcGetLOCA2Name()!=NULL ) {
		sprintf (pvname2_ca, "WIRE:%s:%s:ACCESS.VAL", slcGetLOCA2Name(), unit_a);
		iss = getChannel(pvname2_ca, pIndex, uIndex);
		if (SUCCESS(iss)) setCaMonitor(pIndex, uIndex);
		else
		  DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
			 ("BSAC_H:initCaChannel failure on connect for %s and %s\n", pvname_ca, pvname2_ca));
	  }
	} 
	else setCaMonitor(pIndex, uIndex);
	break;
  case BSACPUID_INDEX:
	break;
  default:
	iss = MICR_EMEM;
	break;
  }

  return(iss); 
}

/*=============================================================================

  Name: getChannel

  Abs:  create a CA channel to the EPICS PV named

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		char *              name_c      read only  pointer to name of EPICS pv
        unsigned short      pIndex      read only  primary device index
        unsigned short      uIndex      read only  unit index

  Rem:   

  Side: 

  Ret:  int channel access error code
==============================================================================*/
LOCAL vmsstat_t getChannel(const char * name_c,
						   const unsigned short pIndex, 
						   const unsigned short uIndex)
{ 
  vmsstat_t iss = MICR_OKOK;
  int status = ECA_NORMAL;
  chid chId;

  if ((status = (ca_create_channel(name_c,NULL,NULL,10, &chId)) )!=ECA_NORMAL ) {
    DEBUGPRINT (DP_INFO, bsacHstaFlag, 
				("BSAC_H:ca_create_channel failure, status = %d\n", status));
	hstaCollect_as[pIndex].unit_as[uIndex].chStatus = status; 
	 iss = BPMO_DBERR;
	 slcCmlogLogMsg(BPMO_DBERR, "DBERR", 
					"CA Create Channel failure for %s.\n", (char *)name_c);	
	 return(iss); 
  }
  else {
	hstaCollect_as[pIndex].unit_as[uIndex].chId = chId;
	hstaCollect_as[pIndex].unit_as[uIndex].chStatus = status; 
    DEBUGPRINT (DP_INFO, bsacHstaFlag, 
				("BSAC_H:created ca channel for %s\n", name_c));
  }

  if ((status = (ca_pend_io(1.0)))!=ECA_NORMAL ) {
    DEBUGPRINT (DP_INFO, bsacHstaFlag, ("BSAC_H:ca_pend_io failure on connect for %s\n", name_c));
	 iss = BPMO_DBERR;
     /* failure, release the channel */
	 if ( (chId) && (hstaCollect_as[pIndex].unit_as[uIndex].chStatus==ECA_NORMAL) ) {
		status = ca_clear_channel(chId);
		status = ca_pend_io(0.1);
	 }
  }

  return(iss); 
}

/*=============================================================================

  Name: bsacHstaCheckHSTAs

  Abs:  For each primary, compare the HSTA of each unit with the stored hsta_ctrl
        to determine of the HSTA has been changed.  If changed, update the 
		hsta_ctrl and the corresponding epics CTRL record.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------

  Rem:  This routine is called periodically by bsacHdlr to keep slc HSTAs 
        synchronized with their corresponding epics CTRL records.

  Side: the hsta_ctrl global value is changed to reflect the current SLC db.

  Ret: vmsstat_t iss the vmsstyle return code

==============================================================================*/
vmsstat_t bsacHstaCheckHSTAs(void)
{
  vmsstat_t iss = MICR_OKOK;
  

  if (!(SUCCESS(iss = bsacCheckUnitHSTAs(BSACBPMS_INDEX)))) return iss;
  if (!(SUCCESS(iss = bsacCheckUnitHSTAs(BSACTORO_INDEX)))) return iss;
  if (!(SUCCESS(iss = bsacCheckUnitHSTAs(BSACARRY_INDEX)))) return iss;
  return(iss);
}
 
/*=============================================================================

  Name: bsacHstaCheckUnitHSTAs

  Abs:  checks the units of this primary for new HSTA values in the SLC database.
        for each unit, if the HSTA value has changed (compared to hsta_ctrl 
		stored value) then the new value is stored in hsta_ctrl and Channel Access 
		is used to update the corresponding epics CTRL record.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        unsigned short      pIndex      read only  primary device index

  Rem:  this routine uses the dblists per primary, so that we can do one dblget
        per primary.

  Side: 

  Ret:  vmsstat_t iss the vms style return code

==============================================================================*/
LOCAL vmsstat_t bsacCheckUnitHSTAs (const unsigned short pIndex)
{
  vmsstat_t iss = MICR_OKOK;
  dbheader_ts * dblist_ps = NULL;
  dbheader_ts * dbdata_ps = NULL;
  unsigned short uIndex; 
  unsigned char * next_p;

  dblist_ps = primCollect_as[pIndex].dblistHSTAs_ps;

  if (primCollect_as[pIndex].unitCnt>0) {

	/* get the current HSTA values for these units from SLC database */
	if (!SUCCESS(iss=(dblget ((const dbheader_ts * const)dblist_ps,
							 (dbheader_ts **const)&dbdata_ps))) ){
	  DEBUGPRINT(DP_ERROR,bsacHstaFlag,
		("bsacCheckUnitHSTA: Failed dblget PRIMINDEX %hd\n",
		 pIndex));
	  return(iss=DB_BADPTR);
	}
	/* now compare new and old hstas*/
	next_p = (unsigned char *)dbdata_ps + sizeof(dbheader_ts);

	for (uIndex=0; uIndex< primCollect_as[pIndex].unitCnt; uIndex++) {
	  if (!SUCCESS(iss = (syncHSTAVal( pIndex, uIndex, 
									   *(unsigned long *)next_p ))) ) {
		DEBUGPRINT(DP_ERROR,bsacHstaFlag,
		("bsacCheckUnitHSTA: Failed syncing HSTA for PRIMINDEX %hd UNITINDEX %hd\n",
		pIndex, uIndex));
		} else { 

		  /*DEBUGPRINT(DP_DEBUG,bsacHstaFlag,
				 ("bsacCheckUnitHSTA: PRIMINDEX %hd UNITINDEX %hd has HSTA =%ld \n", 
				  pIndex, uIndex,
				  *(unsigned long *)next_p ));*/
		/*DEBUGPRINT(DP_DEBUG,bsacHstaFlag,
				 ("bsacCheckUnitHSTA: PRIMINDEX %hd UNITINDEX %hd has hsta_ctrl =%hd \n", 
				  pIndex, uIndex,
				  primCollect_as[pIndex].unit_as[uIndex].hsta_ctrl ));*/
		}
	  next_p += sizeof(unsigned long);
	}  		
  }

  /* release the data list */
  if (dbdata_ps)
	iss = dblistfree(dbdata_ps);
	
  return (iss);
}

/*=============================================================================

  Name: syncHSTAVal

  Abs:  synchronize the stored hsta value with the current ctrl epics record val.
        if the HSTA value has changed (compared to hsta_ctrl stored value) then
		the new value is stored in hsta_ctrl and Channel Access is used to update
		the corresponding epics CTRL record.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		const unsigned short pIndex       read only  primary index
		const unsigned short uIndex       read only  uIndex index
		const unsigned long  hstaVal    read only  the new hsta value to synch

  Rem:  

  Side: hsta_ctrl global may be changed, epics CTRL record may be changed

  Ret: iss vms style status code 
==============================================================================*/
LOCAL vmsstat_t syncHSTAVal(const unsigned short pIndex,
							const unsigned short uIndex,
							const unsigned long  hstaVal)
{
  vmsstat_t iss = MICR_OKOK;
  unsigned short ctrl;
  epicsBoolean update = epicsFalse;
  unsigned short ctrlVal = CTRL_OFFLINE;
  
  ctrl = primCollect_as[pIndex].unit_as[uIndex].hsta_ctrl;
  cvtHSTAToCTRL(pIndex, hstaVal, &ctrlVal );

  if  (ctrl != ctrlVal) {
	if (pIndex == BSACBPMS_INDEX) {
	  update = epicsTrue;
	}
	else {
	  if ( (ctrl == CTRL_OFFLINE)||(ctrlVal==CTRL_OFFLINE) )
		update = epicsTrue;
	}
  }
  if (update)
  {
	DEBUGPRINT(DP_DEBUG,bsacHstaFlag,
     ("syncHSTAVal: pINDEX %hd uINDEX %hd checked new HSTA = 0x%.8lx cur ACCESS = %d \n", 
	  pIndex, uIndex, hstaVal, ctrl ));
	DEBUGPRINT(DP_DEBUG,bsacHstaFlag,
			   ("syncHSTAVal: updating ctrl to  =%d \n",ctrlVal ));
	/* update the global hsta_ctrl */
	primCollect_as[pIndex].unit_as[uIndex].hsta_ctrl = ctrlVal;    
	/* and update epis CTRL record */
	iss = updateEpicsCTRL(pIndex, uIndex, ctrlVal);
  }
  return (iss);
}

/*=============================================================================

  Name: updateEpicsCTRL

  Abs:  caput the new CTRL value to the corresponding epics CTRL record

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		const unsigned short pIndex     read only  primary index
		const unsigned short uIndex     read only  uIndex index
		const unsigned short newCtrl    read only  new CTRL value

  Rem:  incoming newCTrl is mapped to a corresponding unit by pIndex, uIndex. The
        newCtrl val is ca_put to the CTRL record.

  Side: 

  Ret:  iss vmsstat_t vms style status code
   
==============================================================================*/
LOCAL vmsstat_t updateEpicsCTRL(const unsigned short pIndex,
								const unsigned short uIndex,
								const unsigned short newCtrl)
{
  vmsstat_t iss = MICR_OKOK;
  int status = ECA_NORMAL;
  chid chId;

  chId = hstaCollect_as[pIndex].unit_as[uIndex].chId;

  if ( (status = (ca_put(DBR_SHORT, chId,(const void *)&newCtrl)))!=ECA_NORMAL ) {
    DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
				("BSAC_H:updateEpicsCTRL failure on ca_put, status = %d\n", status));
  }
  if ( (status = (ca_pend_io(1.0)))!=ECA_NORMAL ) {
    DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
				("BSAC_H:updateEpicsCTRL ca_pend_io failure on get, status=%d\n", status));
	}
  if (status != ECA_NORMAL) iss = MICR_EMEM;

  return (iss);
}

/*=============================================================================

  Name: setCaMonitor

  Abs:  create a CA subscriiption to the EPICS PV named.  Pass in a pointer
        to the global unit structure needed by the subscription callback.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		const unsigned short pIndex       read only  primary index
		const unsigned short uIndex       read only  uIndex index

  Rem:  this routine will assign the subscription callback and pass in the
        pointer to the corresponding primCollect_as[p].unit_as[u] structure.

  Side: 

  Ret:  vmsstat_t vms style return code
   
==============================================================================*/
LOCAL vmsstat_t setCaMonitor(const unsigned short pIndex,
							 const unsigned short uIndex)
{
  vmsstat_t iss = MICR_OKOK;
  int status = ECA_NORMAL;
  hstaUnit_ts * unit_ps = NULL;
  bsacUnits_ts * usr_ps = NULL;

  unit_ps = &hstaCollect_as[pIndex].unit_as[uIndex];
  usr_ps = &primCollect_as[pIndex].unit_as[uIndex];
  DEBUGPRINT (DP_DEBUG, bsacHstaFlag, 
			  ("BSAC_H:setCaMonitor on pIndex =%d, uIndex=%d unit_ps= %ld\n",
			   pIndex, uIndex, (unsigned long)unit_ps));
  if ((status = (ca_create_subscription(DBR_SHORT, 1, unit_ps->chId, DBE_VALUE,  
			  bsacHstaCTRLCallback, (void *)usr_ps, &unit_ps->evId)) )!=ECA_NORMAL ) {
	DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
				("BSAC_H:ca_create_subscription failure, status = %d\n", status));
	iss = BPMO_DBERR;
	 slcCmlogLogMsg(BPMO_DBERR, "DBERR", 
					"CA Create Subscription failure for %s unit %u.\n", 
					(char *)&primCollect_as[pIndex].slcPrim_a, uIndex);	
  }
  if ((status = (ca_pend_io(1.0)))!=ECA_NORMAL ) {
    DEBUGPRINT (DP_FATAL, bsacHstaFlag, 
				("BSAC_H:ca_pend_io failure on create_subscr, status=%d\n", status));
	iss = BPMO_DBERR;
  }
  return (iss);
}



/*=============================================================================

  Name: bsacHstaCTRLCallback

  Abs:  callback for the caMonitor channel access subscription request.
        Send a BSAC_SYNC_CTRL message to the bsacHdlr.  
		The first two words of reqdata are the pointer to the unit_as[u] 
		struct needed to identify the unit whose CTRL val must be synched 
		with a HSTA.  
		The next word is the new CTRL value.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        struct event_handler_args eha   read only channel access event callback arguments

  Rem:  when a user changes the CTRL value, this callback is called. it will send
        a message to the slc bsac thread to immediately change the HSTA value to synch
		with the new CTRL value.

  Side: a message is sent to the bsacHdlr message queue 

  Ret: 
   
==============================================================================*/
LOCAL void bsacHstaCTRLCallback(struct event_handler_args eha)
{
  msgptr_ts  mSndptr_s;
  msgmail_ts * mSnd_ps = NULL;
  unsigned short *ctrlVal_p = NULL;
  bsacUnits_ts *usr_ps = NULL;

  if(eha.status==ECA_NORMAL) {
	usr_ps = (bsacUnits_ts *)eha.usr;
	ctrlVal_p = (unsigned short *)eha.dbr;
	DEBUGPRINT(DP_DEBUG, bsacHstaFlag,
			   ("bsacHstaCTRLCallback: got event %s = %d, unit_ps=%ld\n",
				ca_name(eha.chid), *ctrlVal_p, 
				(unsigned long)usr_ps) );

	mSndptr_s.msgbytes = sizeof(msgmail_ts);
	if (SUCCESS( msgQGetBuffer(mSndptr_s.msgbytes, (void*)&mSndptr_s.msgptr_p ) )) {
	  mSnd_ps = (msgmail_ts *)mSndptr_s.msgptr_p;

	  mSnd_ps->msgheader.func = BSAC_SYNC_CTRL;
	  memcpy( &mSnd_ps->reqdata[0], &usr_ps, 4);
	  mSnd_ps->reqdata[2] = *ctrlVal_p;
	  if (!(SUCCESS(messageQSend(bsacHdlr, epicsFalse, sizeof(mSndptr_s), &mSndptr_s))))
	  {
		DEBUGPRINT(DP_ERROR, bsacHstaFlag,
			   ("bsacHstaCTRLCallback: msg failed sending event %s = %d\n",
				ca_name(eha.chid),*ctrlVal_p) );
		msgQFreeBuffer(sizeof(msgmail_ts), (void*)mSnd_ps);
	  }
	}
  }
}

/*=============================================================================

  Name: bsacHstaSyncCTRLVal

  Abs:  This routine is called in response to the BSAC_SYNC_CTRL message.
        synchronize the SLC HSTA value with the current epics CTRL record val.
        this utility is called whenever a CTRL record is changed via an epics EDM 
		display.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        bsacUnits_ts*       unit_ps     read/write pointer to unit structure 
		const usigned short ctrlVal     read only  new CTRL val

  Rem:  when a user changes the CTRL value,a caMonitor callback is called. it will send
        a message to the slc bsac thread to immediately change the HSTA value to synch
		with the new CTRL value.

  Side:the hsta_ctrl global is updated for this unit 

  Ret: iss vms style status code 
   
==============================================================================*/
vmsstat_t bsacHstaSyncCTRLVal(bsacUnits_ts *unit_ps,
							  const unsigned short ctrlVal)
{
  vmsstat_t iss = MICR_OKOK;
  dbheader_ts * dbunitl_ps = NULL; /* list for individual unit */
  dbheader_ts * dbdata_ps = NULL;/* list for indiv. unit data */
  unsigned short ctrl = 0;
  unsigned char *hsta_p = NULL;
  unsigned long newHsta = 0;

  /* DO A DBLGET */
  dbunitl_ps =  unit_ps->dblistHSTA_ps;

  /* get the current HSTA value for this unit from SLC database */
  if (!SUCCESS(iss=dblget ((const dbheader_ts * const)dbunitl_ps,
							 (dbheader_ts **const)&dbdata_ps))){
	DEBUGPRINT(DP_ERROR, bsacHstaFlag,
			   ("bsacHstaSyncCTRLVal: failed dblget on dblistHSTA_ps = %ld\n",
				(unsigned long)unit_ps->dblistHSTA_ps) );
	return(iss=DB_BADPTR);
  }
  hsta_p = (unsigned char *)dbdata_ps + sizeof(dbheader_ts);


  cvtHSTAToCTRL(unit_ps->primIndex, *(unsigned long *)hsta_p, &ctrl);

  /* now update the SLC Database */
  unit_ps->hsta_ctrl = ctrl;   
 
  if  (ctrl != ctrlVal) {

	unit_ps->hsta_ctrl = ctrlVal;

	/* convert the ctrl value to a HSTA value */
	cvtCTRLToHSTA(unit_ps->primIndex, ctrlVal, *(unsigned long *)hsta_p, &newHsta);
  
	DEBUGPRINT(DP_DEBUG, bsacHstaFlag,
	 ("bsacHstaSyncCTRLVal: updating cur hsta= 0x%.8lx to newHsta = 0x%.8lx, pIndex=%d\n",
			   *(unsigned long *)hsta_p, newHsta, unit_ps->primIndex) );

	/* copy the new hsta val into the dbdata list */
	memcpy(hsta_p, &newHsta, sizeof(newHsta));

	/* now dblput the data */
	iss = dblput(dbunitl_ps, dbdata_ps, bpmoJob);
	iss = dbupdateNoWait(bpmoJob);
  } 

  /* release the data list */
  if (dbdata_ps)
	iss = dblistfree(dbdata_ps);

  return (iss);
}

/*=============================================================================

  Name: bsacHstaStoreHSTAAsCTRL

  Abs:  store this new HSTA value in the primCollect global, as a CTRL value
        Called when the slcIOC is first initialized

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		const unsigned short pIndex       read only  primary index
		const unsigned short uIndex       read only  unit index
		const unsigned long  newHsta    read only  new Hsta value in SLC format

  Rem:  incoming newHsta is mapped to a corresponding CTRL_  value. 

  Side: primCollect_as[pIndex].unit_as[uIndex].hsta_ctrl is stored

  Ret:  iss vmsstat_t vms style status code
   
==============================================================================*/
vmsstat_t bsacHstaStoreHSTAAsCTRL(const unsigned short pIndex,
								   const unsigned short uIndex,
								   const unsigned long newHsta)
{
  vmsstat_t iss = MICR_OKOK;
  unsigned short newCtrl = CTRL_ONLINE;

  cvtHSTAToCTRL(pIndex, newHsta, &newCtrl);
  primCollect_as[pIndex].unit_as[uIndex].hsta_ctrl = newCtrl;
  return (iss);
}

/*=============================================================================

  Name: cvtHSTAToCTRL

  Abs:  convert a HSTA value (unsigned long) to a 
        corresponding CTRL value (unsigned short)

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
	    const unsigned long hsta        read only  a SLC style HSTA value
        unsigned short *    ctrl_p      read/write a pointer to the new CTRL style value

  Rem:  incoming SLC style HSTA is mapped to a corresponding unit CTRL record value.
        The mapping is dependent on the primary (kind of device).  BPMS devices
		have an ONLINE, OFFLINE, MAINTENANCE and CHECKOUT mode in their HSTAs. All
		other devices have only ONLINE or OFFLINE for HSTA in the SLC Database.

		For devices other than BPMS, an ONLINE HSTA will only change CTRL if 
		CTRL=OFFLINE, otherwise, it is left alone.

		for devices other BPMS, an OFFLINE HSTA will change CTRL to OFFLINE if
		CTRL is not already OFFLINE.

  Side: 

  Ret:  iss vmsstat_t vms style status code
   
==============================================================================*/
LOCAL void cvtHSTAToCTRL(const unsigned short pIndex,
						 const unsigned long hsta,
						 unsigned short *ctrl_p)
{
  
  if (pIndex==BSACBPMS_INDEX) {
	if ((hsta & HSTA_CHECKOUT) == HSTA_CHECKOUT) {
	  *ctrl_p = CTRL_CHECKOUT;
	  return;
	}
    else if ( (hsta & HSTA_MAINTENANCE) == HSTA_MAINTENANCE) {
	  *ctrl_p = CTRL_MAINT;
	  return;
	}
  }
  if ( (hsta & HSTA_ONLINE) == HSTA_ONLINE) *ctrl_p = CTRL_ONLINE;
  else if ( (hsta & HSTA_OFFLINE) == HSTA_OFFLINE) *ctrl_p = CTRL_OFFLINE;
  
  return; 
}

/*=============================================================================

  Name: cvtCTRLToHSTA

  Abs:  convert a HSTA value (unsigned long) to a 
        corresponding CTRL value (unsigned short)

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        unsigned short      ctrl        read only  a epics CTRL record style value
	    const unsigned long *hsta       read/write a pointer to a new SLC style HSTA value

  Rem:  incoming unit CTRL record value is mapped to a corresponding slc style HSTA value. 
        The conversion is dependent on the primary (device type) and must be
		or-ed with the current HSTA value so as not to loose any other bits not
		associated with ONLINE/OFFLINE/CHECKOUT/MAINTENANCE.
		BPMS values map directly
		devices other than BPMS or set to ONLINE for ONLINE, CHECKOUT and MAINTENANCE

  Side: 

  Ret:  iss vmsstat_t vms style status code
   
==============================================================================*/
LOCAL void cvtCTRLToHSTA(const unsigned short pIndex,
						 const unsigned short ctrl,
						 const unsigned long oldhsta,
						 unsigned long *newhsta_p)
{
  switch (ctrl) {
  case CTRL_ONLINE:
	if (pIndex==BSACBPMS_INDEX)
	  *newhsta_p = (oldhsta |HSTA_ONLINE) & 0xfff3fff1;
	else
	  *newhsta_p = (oldhsta |HSTA_ONLINE) & 0xfffffff1;	  
	break;
  case CTRL_OFFLINE:
	if (pIndex==BSACBPMS_INDEX)
	  *newhsta_p = (oldhsta |HSTA_OFFLINE) & 0xfff3fff4;
	else
	  *newhsta_p = (oldhsta |HSTA_OFFLINE) & 0xfffffff4;
	break;
  case CTRL_MAINT:
	if (pIndex==BSACBPMS_INDEX)
	  *newhsta_p = (oldhsta | HSTA_MAINT) & 0xfff4fff1;
	else
	  *newhsta_p = (oldhsta | HSTA_ONLINE) & 0xfffffff1;
	break;
  case CTRL_CHECKOUT:
	if (pIndex==BSACBPMS_INDEX)
	  *newhsta_p = (oldhsta | HSTA_CHECK) & 0xfff8fff1;
	else
	  *newhsta_p = (oldhsta | HSTA_ONLINE) & 0xfffffff1;	  
	break;
  default:
	*newhsta_p = 0x00040001;
	break;
  }

  return; 
}


/*==============================================================================

  Name: cleanup()

  Abs:  cleanup the globals for this thread before shutdown

  Args: none

  Rem: Clean up the dblists used for synchronizing hstas / ctrl
  side:

  Ret: none

  ==============================================================================*/
void bsacHstaCleanup(void)    
{
  vmsstat_t iss = MICR_OKOK;
  int status = ECA_NORMAL;
  hstaUnit_ts * unit_ps = NULL;
  unsigned short pIndex;
  unsigned short uIndex;

  /* ca_client_status(5);*/
 

  for (pIndex=0; pIndex<BSACPRIM_MAX; pIndex++) {

	/* release the dblist */
	if (primCollect_as[pIndex].dblistHSTAs_ps)
	  iss = dblistfree(primCollect_as[pIndex].dblistHSTAs_ps);
	primCollect_as[pIndex].dblistHSTAs_ps = NULL;

	
	/* now do the units */
	for (uIndex=0; uIndex<UNITS_MAX; uIndex++) {
      /* release the dblists */
	  if (primCollect_as[pIndex].unit_as[uIndex].dblistHSTA_ps)
		iss = dblistfree(primCollect_as[pIndex].unit_as[uIndex].dblistHSTA_ps);
	  primCollect_as[pIndex].unit_as[uIndex].dblistHSTA_ps = NULL;
	  
	  /* now release the channels */
	  unit_ps = &hstaCollect_as[pIndex].unit_as[uIndex];
	  if ( (unit_ps->chId) && (unit_ps->chStatus==ECA_NORMAL) ) {
		status = ca_clear_subscription(unit_ps->evId);
		status = ca_clear_channel(unit_ps->chId);
		status = ca_pend_io(0.1);
	  }	
	}
  }
  /* destroy the channel access context */
  ca_context_destroy();

  /*ca_client_status(5);*/

  return;
}


