/*=============================================================================
 
  Name: bsacPrep.c


  Abs:  bsacPrep performs the measurement preparation for BPMO_MEAS_PREP
        and BPMO_WIRE_PREP.
        Reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | BSAC Service Design

  Auth: 20 jul 2005, dROGIND    created 
  Rev:  

NOTE: the device CTRL PVs (hsta equiv, not MDEF) are RENAMED TO ACCESS, but 
variable names, comments will still refer the CTRL.  Now when you read 
device CTRL, think ACCESS!!
-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest) 
        20 jul 2005, DROGIND    created 
		30-OCT 2006 DROGIND modified cvtReply
=============================================================================*/
/* SCP Operations notes
   BPMSs and TOROs are allowed to be collected together from the SCP
   If, on BSAC panel, Device btn says BPM, the measbits will be set for
     BPMS, TORO, and pulseid.
   For ARRYs, click Device and choose GADC, (then One Shot, etc)
 */
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int bsacPrepFlag = DP_DEBUG;
#endif


/* c includes */
#include <string.h>        /* memcpy */

/* epics */
#include <dbAccess.h>      /* EPICS dbAddr struct */

/* shared slc micro includes */
#include "bpmfunc.hc"      /* bpmo_xxx func codes */
#include "micrdef.hc"      /* MICR_EXNET_ codes */
#include "slcMsg.h"        /* IOC_STOP, BSAC_* */
#include "msgfunc.hc"      /* RESPONSE for func code */
#include "bpmodef.hc"      /* BPMO_OKOK */

/* slc global vars */
#include "slcGlob.h"       /* slcThreads_as */

/* slcIOC types */
#include "msgTypes.h"      /* msgmail_ts etc */
/* slcIOC utilities */
#include "slcUtil.h"       /* slcPrintMsgheader */
#include "msgQUtil.h"      /* messageQSend, messageQRecv */
#include "cvtUtil.h"       /* slcSSwap, slcLSwap, cvtULongtoUShorts etc */
#include "slcCmlog.h"      /* slcCmlogLogMsg */
#include "timeUtil.h"      /* timeGetVMS, timeDiffVMS10usec */

/* slcMain function prototypes */
#include "slcMain.h"       /* slcStop(), slcRestart() */

/* bsac includes */
#include    "bsacGlob.h"
#include    "bsacPrepTypes.h"
#include    "bsacPrep.h"
#include    "bpm_reqreply_struc.h"
#include    "bpmstruc.h"
#include    "bsacAbort.h"


/* vars global only to bsacPrep functions, declared in bsacPrepInit.c */
extern mdefAddr_ts mdefAddr_as[MDEF_MAX];
extern primAddrStruct_ts primAddr_as[BSACPRIM_MAX]; /* contains epics db addrs for */
                  /* records PRIM:LOCA:UNIT:UNITID */
                  /*  and PRIM:LOCA:UNIT:GO */

/* static structures */
LOCAL bsacMDEF_ts measdef_s;        /* unpack current request into this local */
LOCAL bsacMD_ts   slcmeasdef_s;     /* and this one - gets xferred into bsacMD_as */

/* prototypes */
LOCAL vmsstat_t bsacPrepCollect (const msgmail_ts * rMsg_ps,
								 const short md,
								 unsigned long * totbc_p);


LOCAL vmsstat_t bsacPrepCollectUnits ( const short md,
				 unsigned long * totbc_p);

LOCAL vmsstat_t bsacPrepCheckPrim(const unsigned short prim,
				  const unsigned short measprim,
				  const unsigned long datasize,
				  const short md,
				  unsigned long * totbc_p,
				  unsigned long * offset_p,
				  short *bufIndex,
				  unsigned short * ulistI_p);

LOCAL vmsstat_t bsacPrepAddUnit(const unsigned short prim,
			  const unsigned short unit,
			  const short md,
			  const unsigned long datasize,
			  unsigned long * totbc,
			  unsigned long * offset,
			  short * bufIndex);

LOCAL void bsacPrepNextDataReply( const short md, 
								  unsigned long * offset_p,
								  unsigned long * totbc_p,
								  short * bufIndex_p);

LOCAL vmsstat_t bsacPrepDataReplyMsg(const short md, 
			       const short bufIndex);

LOCAL void bsacPrepLastDataReply( const short md, 
			    const unsigned long totbc,
			    const unsigned long offset,
			    const short bufIndex);

LOCAL vmsstat_t bsacPrepParseMeasMsg (const msgmail_ts * rMsg_ps, 
								epicsBoolean measprep);

LOCAL vmsstat_t bsacPrepGo (const short md );

LOCAL vmsstat_t bsacPrepMDEF (const short md, const char * scpName );

LOCAL vmsstat_t bsacPrepFindMD (short * md);

LOCAL vmsstat_t bsacPrepAllocateBuffers( const short md);

/*=============================================================================

  Name: bsacPrep

  Abs:  

  Args: Type		      Name        Access	    Description
	    ----------------- -------    ----------   ----------------------------
		const msgmail_ts* rMsg_ps    read only     incoming data message
		const epoicsBoolean measprep read only     T for BPMO_MEAS_PREP
		                                           F for BPMO_WIRE_PREP
  Rem:  Called from bsacHdlr upon receipt of BPMO_*_PREP request messages
        from SCP 

  Side: First parses the incoming message into temp local slcmeasdef_s
        Then calls bsacPrepFindMD to see if there is a free measdef;
		if we have a measdef avail, slcmeasdef_s is copied into bsacMD_as[md] and md
        is returned here. For a valid md, an immediate reply is sent back to scp.
		bsacMD_as and primCollect_as vars are initilaized, then the collection
		areas are established.

  Ret:  none

==============================================================================*/
vmsstat_t bsacPrep (const msgmail_ts * rMsg_ps, const epicsBoolean measprep)
{
  vmsstat_t iss = BPMO_OKOK;
  unsigned long totbc;    /* filled in bsacPrepCollect: */
  short md = 0;           /* set in bsacPrepFindMD */
  char scpName[MAX_STRING_SIZE];
  int prim, unit, bufs;

  /* we start with a byte count that includes the data reply header */
  /* we will add to this as we go through the measdef */
  /* before: totbc = sizeof(bpm_alldata_replybeg_ts);  */
  /* stuff bc for immed reply first */
  totbc = sizeof(bpm_allprep_reply_ts);

  memcpy (&scpName[0], &rMsg_ps->msgheader.source[0],4);
  scpName[5] = '\0';

  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrep: Entering..\n"));
  /* fill in measdef_s local var from rMsg_ps */
  iss = bsacPrepParseMeasMsg (rMsg_ps, measprep);

  /* handle/ store incoming measdef into global array */
  if (!SUCCESS(iss = bsacPrepFindMD(&md))) {
	/* send i reply with iss=BPMO_RQSTID, meaning (in our case) no more measdefs avail */
	return (iss = bsacPrepImmedReply(rMsg_ps, iss, totbc, 0));	  
	/* possibly send err code TOOMNYDEFS */
  }
  /* reset device meas def globals for this new meas def # md */ 

  for (prim=0; prim<=BSACPUID_INDEX; prim++) {
	for (unit=0; unit<UNITS_MAX; unit++) {
	  primCollect_as[prim].unit_as[unit].mdef_as[md].included = epicsFalse;
	  primCollect_as[prim].unit_as[unit].mdef_as[md].offset = 0;
	  primCollect_as[prim].unit_as[unit].mdef_as[md].counter = 0;
	  primCollect_as[prim].unit_as[unit].mdef_as[md].bufIndex = 0;
	  primCollect_as[prim].unit_as[unit].chanCntr = 0;
	}
  }
  /* init message buffer variables - */
  bsacMDBufs_as[md].bufCount= 0;
  bsacMDBufs_as[md].nextBuf = 0;
  for (bufs = 0; bufs < MD_DATABUFS_MAX; bufs++) {
	bsacMDBufs_as[md].bufs_as[bufs].totwc   = 0;
	bsacMDBufs_as[md].bufs_as[bufs].partwc  = 0;
	bsacMDBufs_as[md].bufs_as[bufs].cumpartwc  = 0;
	bsacMDBufs_as[md].bufs_as[bufs].buf_ps  = NULL;
	bsacMDBufs_as[md].bufs_as[bufs].data_p = NULL;
  }

  bsacMD_as[md].ndatblks = 0;
  /* prepare primCollect_a area for collected data & prepare data */
  /*  for reply large message buffer (headers, etc) */
  /* quite possible there is no memory for this */
  if (SUCCESS(iss = (bsacPrepCollect (rMsg_ps, md, &totbc))))
  {
	/* output to EPICS MDEF records;may need to combine w above */
	iss = bsacPrepMDEF (md, scpName); 
	/* bias mdef # +1 for output to epics (range 1-20) */

	/* set MDEF device CTRL records */
	iss = bsacPrepGo (md);
  }
  /* create and send immediate data reply */
  iss = bsacPrepImmedReply(rMsg_ps, iss, totbc, 0);
  /* Tony says to leave npulstep = 0 for MEAS_PREP and WIRE_PREP */

  return (iss);
}
/*=============================================================================

  Name: bsacPrepParseMeasMsg

  Abs:  This routine unpacks the MEAS_PREP message by writing into
        local variables measdef_s and slcmeasdef_s to determine whether 
		the request is valid.
		Parses down structure bpm_measprep_req_ts; 

  Args: Type		      Name        Access	    Description
	    ----------------- -------    ----------   ----------------------------
		const msgmail_ts* rMsg_ps    read only     incoming data message
		const epoicsBoolean measprep read only     T for BPMO_MEAS_PREP
		                                           F for BPMO_WIRE_PREP

  Rem:  Parses incoming BPMO_*_PREP request messages
        from SCP. Stores temporarirly in slcmeasdef_s until a call can be made
        to see if we have a free measdef. 

  side: 

  Ret:  none

==============================================================================*/
LOCAL vmsstat_t bsacPrepParseMeasMsg (const msgmail_ts * rMsg_ps, epicsBoolean measprep)
{
  vmsstat_t iss = BPMO_OKOK;
  unsigned char * reqdata_ps; /* byte ptr to data portion of incoming meas */
  unsigned short cnt;
  unsigned long temp; /* debug*/

  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: Entering..\n")); 
  /* slog through the reqdata & extract pertinent fields ... */
  if (measprep) {
	slcmeasdef_s.measprep = epicsTrue;
	slcmeasdef_s.nrpos = 1;
  }
  else {
	slcmeasdef_s.measprep = epicsFalse;}
	
  slcmeasdef_s.abort = epicsFalse;

  /* get the scpName  into slcmeasdef_s.username */
  memset(&slcmeasdef_s.username[0], 0, 5);
  memcpy (&slcmeasdef_s.username[0], &rMsg_ps->msgheader.source[0],4);

  /* point to top of data - totwc*/  
  reqdata_ps = (unsigned char *)&rMsg_ps->reqdata;
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepParseMeasMsg: reqdata ptr init to = %p\n",reqdata_ps)); 
  cvtLSwapMove((const void *)reqdata_ps, (void *)&temp);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepParseMeasMsg: reqmsgtotwc = %ld\n",temp)); 
  /* point to partwc */
  reqdata_ps += sizeof(unsigned long);
  cvtLSwapMove((const void *)reqdata_ps, (void *)&temp);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepParseMeasMsg: reqmsgpartwc = %ld\n",temp)); 
 
  /* point to i_seg */
  reqdata_ps += sizeof(unsigned long); /*sizeof(bpm_largreq_hdrr_ts);*/
  /* store i_seg */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&measdef_s.measid);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: i_seg = %hd\n", 
  									 measdef_s.measid));
  /* point to pp array - chars */
  reqdata_ps += sizeof(short) * (BPM_RING_MAXCAL + 3);
  for (cnt=0; cnt<BPM_MAXPP; cnt++) {
	measdef_s.pp[cnt]=*reqdata_ps;
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: pp[%d] = %hd\n", 
									   cnt, measdef_s.pp[cnt]));
    reqdata_ps++;
  } 
  /* hard code pp to 1 for now, as scp is set for lcls bc pp = 10 */
  /*measdef_s.pp[0]=1;*/
  /* debug !!*/

  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: long_pulse_..ptr = %p\n", 
									 reqdata_ps));
  /* point to navg; left off long_pulse_bpm_phas_type */
  reqdata_ps += sizeof (unsigned char) * 5;
  DEBUGPRINT(DP_INFO, bsacPrepFlag, 
			 ("bsacPrepParseMeasMsg: navg ptr = %p\n",reqdata_ps)); 
  slcmeasdef_s.navg = *reqdata_ps;
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: navg = %hd\n", 
									 slcmeasdef_s.navg));
  /* point to yy */
  reqdata_ps += sizeof (unsigned char) * 2;
  measdef_s.yy = *reqdata_ps;
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: yy= %hd\n", 
									 measdef_s.yy)); 
  /* point to  measbits */
  reqdata_ps += sizeof (unsigned char) * 5;
  /*convert measbits */
  cvtLSwapMove((const void *)reqdata_ps, (void *)&slcmeasdef_s.measbits);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: measbits= 0x%lx\n", 
									 slcmeasdef_s.measbits)); 
  /* point to umin */
  reqdata_ps += sizeof(long);
  /* convert umin */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&slcmeasdef_s.umin);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: umin = %hd\n", 
									 slcmeasdef_s.umin)); 
  reqdata_ps += sizeof(short);      /* point to umax */
  /* convert umax */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&slcmeasdef_s.umax);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: umax = %hd\n", 
									 slcmeasdef_s.umax));
  reqdata_ps += sizeof(short);      /* point to umin2 */

  /* convert umin2 */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&slcmeasdef_s.umin2);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: umin2 = %hd\n", 
									 slcmeasdef_s.umin2)); 
  reqdata_ps += sizeof(short);      /* point to umax2 */
  /* convert umax2 */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&slcmeasdef_s.umax2);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: umax2 = %hd\n", 
									 slcmeasdef_s.umax2)); 
  /* point to exclmask[0]... */
  reqdata_ps += sizeof(short) * 5;
  reqdata_ps += sizeof(float) * BPM_RING_MAXCAL * 2;
  reqdata_ps += sizeof(long) * 2;
  reqdata_ps += sizeof(float)* 2;
  /* now we're pointing to exclmask[0] -convert exclmask[cnt] */
  for (cnt=0; cnt<MAX_SLCMODFR_MASK; cnt++) { 
	cvtLSwapMove((const void *)reqdata_ps, (void *)&measdef_s.exclmask[cnt]);
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: exclmask[%d] = 0x%ld\n", 
									   cnt, measdef_s.exclmask[cnt]));
	reqdata_ps += sizeof (unsigned long);   
  } 
  /* convert inclmask[cnt] */
  for (cnt=0; cnt<MAX_SLCMODFR_MASK; cnt++) { 
	cvtLSwapMove((const void *)reqdata_ps, (void *)&measdef_s.inclmask[cnt]);
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: inclmask[%d] = 0x%ld\n", 
									   cnt, measdef_s.inclmask[cnt]));
	reqdata_ps += sizeof (unsigned long);   
  } 
  reqdata_ps += sizeof (unsigned long) * MAX_SLCMODFR_MASK;

  reqdata_ps += sizeof (unsigned short) * 2;
  /* convert timeout (unsigned short nfwd_sec_max; store into int timeout) */
  cvtSSwapMove((const void *)reqdata_ps, (void *)&cnt);
  slcmeasdef_s.timeout = (int)cnt;
  /* debug !!!! */
  /*slcmeasdef_s.timeout = 100;*/
  slcmeasdef_s.timeoutCntr = slcmeasdef_s.timeout;


  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: timeout= %d\n", 
									 slcmeasdef_s.timeout)); 
  reqdata_ps += sizeof (unsigned short) * 6;

  reqdata_ps += sizeof (unsigned long) * 2;
  reqdata_ps += sizeof (unsigned short) * BPM_RING_MAXCAL;
  reqdata_ps += sizeof (unsigned long) * 2;
  reqdata_ps += sizeof (unsigned short) * 8;
  /* convert ulist */
  for (cnt=0; cnt<BPM_ULIST_PRIMS; cnt++) {
	cvtSSwapMove((const void *)reqdata_ps, (void *)&slcmeasdef_s.ulist.nunits[cnt]);
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: nunits[%d]= %hd\n", 
									   cnt, slcmeasdef_s.ulist.nunits[cnt])); 
    reqdata_ps += sizeof(unsigned short);
  }

  /* convert the explicit unit list */
  for (cnt=0; cnt<BPM_ULIST_MAX; cnt++) {
	cvtSSwapMove((const void *)reqdata_ps, (void *)&slcmeasdef_s.ulist.ulist[cnt]);
	/* mask off the 2 hi order bits, which are cal index */
	slcmeasdef_s.ulist.ulist[cnt]=slcmeasdef_s.ulist.ulist[cnt]&BPMULIST_UNIT_MASK;
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: ulist[%d]= %hd\n", 
									   cnt, slcmeasdef_s.ulist.ulist[cnt])); 
	reqdata_ps += sizeof(unsigned short);
  }
 
  if (!measprep){
	/* get nrpos */
	reqdata_ps += sizeof(unsigned short);
	cvtSSwapMove((const void *)reqdata_ps, (void *)&slcmeasdef_s.nrpos);
	
  }
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepParseMeasMsg: nrpos= %hd\n", 
									 slcmeasdef_s.nrpos));   
  return (iss);
}
/*=============================================================================

  Name: bsacPrepFindMD

  Abs:  Assign md # to this new measprep request. If the i_seg measdef id
        is assigned to an existing measdef, then terminate it and reject this one.
        (see measprep.c,  meas_prologue())
		
  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        short * md          md          write      mdef # 0 to (max-1) (PV's go 1-max)

  Rem: 

  side: The md discovered here is passed to many other routines during the scope
        of measurement preparation. Alterations occur to bsac_as[md]

  Ret:  none

==============================================================================*/
LOCAL vmsstat_t bsacPrepFindMD (short * md)
{
  vmsstat_t iss = BPMO_OKOK;
  short cnt = 0;
  long  nRequest = 1;      /* these are for fetching Epics PV's */
  long  options  = 0;
  unsigned short measid;
  unsigned short ctrl;

  *md = -1;
  /* see if this iseg is already running */
  for (cnt=0; cnt<MDEF_MAX; cnt++) {
	if ((dbGet(&mdefAddr_as[cnt].ID_s, DBR_USHORT, 
			   &measid, &options, &nRequest, NULL))) {
	    DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
				   ("bsacPrepFindMD: Error for dbGet of MDEF:LOCA:%hd:ID\n", cnt+1)); 
	}
	else {
	  if (measid == measdef_s.measid){
	    DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
				   ("bsacPrepFindMD: Comparing MDEF:LOCA:%hd:ID=%hd with measid=%hd\n", 
					cnt+1, measid, measdef_s.measid)); 	 
		if ((dbGet(&mdefAddr_as[cnt].CTRL_s, DBR_USHORT, 
			   &ctrl, &options, &nRequest, NULL))) {
		  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
				   ("bsacPrepFindMD: Error for dbGet of MDEF:LOCA:%hd:CTRL\n", cnt+1));
		}else {
		  if (ctrl==1){	
			/* TODO: First Terminate this bsacMD_as[md] if active  */
			bsacAbortCancelMeas(cnt);
			DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
				   ("bsacPrepFindMD: Terminating MDEF:LOCA:%hd:CTRL, which has requested segment # and CTRL is ON\n", cnt+1));
			break;
		  }
		}
	  }
	}
  }
  /* look at CTRL to find the first available mdef (mdef with ctrl = off) */
  for (cnt=0; cnt<MDEF_MAX; cnt++) {
	if ((dbGet(&mdefAddr_as[cnt].CTRL_s, DBR_USHORT, 
			   &ctrl, &options, &nRequest, NULL))) {
	    DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
				   ("bsacPrepFindMD: Error for dbGet of MDEF:LOCA:%hd:CTRL\n", *md)); 
	}
	else {
	  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
				 ("bsacPrepFindMD: MDEF:LOCA:%hd:CTRL = %hd\n", cnt+1, ctrl));
	  if (ctrl==0){
		*md = cnt;
		DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
				   ("bsacPrepFindMD: Found MDEF = %hd with CTRL Off (0)\n", *md+1));
		break;
	  }
	}
  }

  if (*md >=0){
	/* copy parsed slc measdef fields into permanent array storage */
	memcpy(&bsacMD_as[*md], &slcmeasdef_s, sizeof(slcmeasdef_s));
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			   ("bsacPrepFindMD: bsacMD_as[%d].measprep = %c\n", *md,(unsigned char) bsacMD_as[*md].measprep));  
  } else {
	DEBUGPRINT(DP_INFO, bsacPrepFlag, 
			   ("bsacPrepFindMD: All MDEFs are being used!\n"));
	return (iss = BPMO_NO_MEASDEF);
  }
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepFindMD: Exiting with md = %hd\n", *md)); 
  return (iss);
}


/*=============================================================================

  Name: bsacPrepCollect

  Abs:  

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		const msgmail_ts*  rMsg_ps     read         incoming data message
        const short * md   md          read         mdef # 0 to (max-1) (PV's go 1-max)
		unsigned long *    totbc_p     write        ptr to total byte count

  Rem: Routine stores data required for measurements, calculates required number of
       data buffers and the offsets for each devices data location within the buffer
	   Once all calculates are stored it will allocate the buffers. Data is stored into
	   the actual message buffers to be sent as the data is reported from EPICS.

  side:
          

  Ret:  none

==============================================================================*/
LOCAL vmsstat_t bsacPrepCollect (const msgmail_ts * rMsg_ps,
								 const short md,
								 unsigned long * totbc_p)
{
  vmsstat_t iss = BPMO_OKOK;
  long mcnt=0;  

  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepCollect: Entering..\n")); 
  /* store the request headers for initializing reply buffers */
  bsacMD_as[md].reqHdrs.job_id = rMsg_ps->job_id;
  memcpy(&bsacMD_as[md].reqHdrs.tcpfwdhdr, &rMsg_ps->tcpfwdhdr, sizeof(fwd_hdr_ts));
  memcpy(&bsacMD_as[md].reqHdrs.msgheader, &rMsg_ps->msgheader, sizeof(msgheader_ts));

  /* initialize byte count and msg counter */
  bsacMD_as[md].totalByteCnt = 0;
  bsacMD_as[md].msgCount = 0;	
  bsacMD_as[md].msgCounter = 0;

  /* if MEAS_PREP data response... */
  if (bsacMD_as[md].measprep) {
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			   ("bsacPrepCollect: Processing MEAS_PREP request\n") );
  } else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			   ("bsacPrepCollect: Processing WIRE_PREP request\n") );
    
  /* now calc all the buffer indeces and offsets for the units in this md */
  iss = bsacPrepCollectUnits(md, totbc_p);
  /* store total bytecount for this md */
  bsacMD_as[md].totalByteCnt = *totbc_p;
  /* counting expected BSAC_DATA messages, now add msg count from the units*nrpos */
  /* remember to discount the offline devices */
  mcnt = bsacMD_as[md].nunits - bsacMD_as[md].noffline;
  bsacMD_as[md].msgCount = (mcnt * bsacMD_as[md].nrpos); 
  DEBUGPRINT(DP_INFO, bsacPrepFlag, 
			 ("bsacPrepCollectUnits: TOTAL # MESSAGES EXPECTED = %ld\n",bsacMD_as[md].msgCount ) );  
  /* now allocate buffers */
  iss = bsacPrepAllocateBuffers(md);
  
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepCollect: Exiting..\n")); 

  return iss;
}

/*=============================================================================

  Name: bsacPrepCollectUnits

  Abs:  

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        short * md         md          write        mdef # 0 to (max-1) (PV's go 1-max)
		unsigned long *    totbc_p     write        ptr to total byte count


  Rem: Routine checks the measbits to decide whether to store data for each primary.
       If the measbits indicate a primary is desired, bsacPrepCheckPrim is called
	   along with the proper return data size for meas versus buf acq data.


        bc - keeps track of the total byte count of all messages
		offset - is the offset from the drMsg_ps->reqdata for the primary data
        bufIndex - keeps track of the number/index of data reply buffers needed
		ulistI - index into scp message ulist

  side: Reply data may be broken up into more than one large message buffer. Running
        indices into the scp message ulist, into outgoing buffer index and offset, 
		total bytes counts, etc are maintained as primaries and units are added.
          

  Ret:  vmsstat_t

==============================================================================*/
LOCAL vmsstat_t bsacPrepCollectUnits ( const short md,
									   unsigned long * totbc_p)
{
  vmsstat_t iss = BPMO_OKOK;  
  unsigned long offset = 0;    /* offset into the current large msg buff */
  unsigned long bc = 0;        /* total byte count of all data reply messages */
  short bufIndex = 0;          /* holds the index to the current data reply msg */
  unsigned short ulistI = 0;   /* running index into meas prep's ulist */
  int puidCnt = 0;
  unsigned long datasize = 0; /* datasize, variable depending on wire or meas prep*/
   
  offset = sizeof(bpm_alldata_replybeg_ts);
  /* add in alldata_reply to totbc for 1st buffer only */
  bc = offset;
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepCollectUnits: initial bc = %ld\n", bc)); 
  /* we always start with one buffer */
  bsacMDBufs_as[md].bufCount = 1;

  /* accumulate nunits starting at zero */
  bsacMD_as[md].nunits = 0;
  bsacMD_as[md].noffline = 0;

  /* Note: the following code is order dependent */
  /* Check primaries involved in data response order */
  /* as offsets & bytecount into the data response message 
	 buffer are calculated along the way */
  if ( (bsacMD_as[md].measbits &  MEASBPM_BIT) && (SUCCESS(iss)) ) {
	/* and we want some BPMS units ... */
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepCollectUnits: BPMS wanted\n"));

	if (bsacMD_as[md].measprep) 
	  datasize = BPMS_DATA_SIZE;
    else 
	  datasize = WS_BPM_DATA_SIZE;

	iss = bsacPrepCheckPrim(BSACBPMS_INDEX, MEASBPM_INDEX, 
							datasize, md, &bc, &offset, &bufIndex, &ulistI); 
  }	  
  if ( (bsacMD_as[md].measbits & MEASTOR_BIT) && (SUCCESS(iss)) ) {
	/* and we want some TORO units ... */
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepCollectUnits: TOROs wanted\n"));

	if (bsacMD_as[md].measprep) 
	  datasize = TORO_DATA_SIZE;
    else 
	  datasize = WS_TORO_DATA_SIZE;

	iss = bsacPrepCheckPrim(BSACTORO_INDEX, MEASTOR_INDEX, 
							datasize, md, &bc, &offset, &bufIndex, &ulistI); 
  }
  if ( (bsacMD_as[md].measbits & MEASARRY_BIT) && (SUCCESS(iss))) {
	/* and we want some ARRY units ... */
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepCollectUnits: ARRYs wanted\n"));

	if (bsacMD_as[md].measprep) 
	  datasize = ARRY_DATA_SIZE;
    else 
	  datasize = WS_ARRY_DATA_SIZE;

	iss = bsacPrepCheckPrim(BSACARRY_INDEX, MEASARRY_INDEX, 
							datasize, md, &bc, &offset, &bufIndex, &ulistI); 
  }
  if ( (bsacMD_as[md].measbits & MEASPULSID_BIT) && (SUCCESS(iss))) {
	/* and we want pulse ids ... */
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepCollectUnits: PULSEIDs wanted\n"));

	puidCnt = ((bsacMD_as[md].nrpos * bsacMD_as[md].navg) -1) * sizeof(unsigned long);
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag,("bsacPrepCollect: puid bc = %d\n", puidCnt) );
	if (bsacMD_as[md].measprep) 
	  puidCnt += PUID_DATA_SIZE;
    else 
	  puidCnt += sizeof(scanblk_prefix_ts) + WS_PUID_DATA_SIZE;
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag,("bsacPrepCollect: puid bc+hdrs = %d\n", puidCnt) );
	
	/* 1st call just has puid scanblk_prefix header */
	iss = bsacPrepAddUnit(BSACPUID_INDEX, 1, md, puidCnt,    
						  &bc, &offset, &bufIndex);
  }
  /* to be compatible with SCP, reduce each continuing buffer by alldata_reply size */
  /* the totwc and partwcs pretend that only the first buffer contains an alldata_reply */
  /* the msgheader.datalen field reflects the additional bytes for the alldata_reply */
  /* subract off here, cuz offsets above need to start at the correct place in the buffer, */
  /*   that is, afer the bpm_alldata_reply_ts */
  /* DR alldata_reply size is no longer added to totbc in  PrepNextDataReply for continuing buffs */
  /*bc -= sizeof (bpm_alldata_replybeg_ts); */
  /* end of debug */

  DEBUGPRINT(DP_DEBUG, bsacPrepFlag,("bsacPrepCollect: final bc = %lu\n", bc) );
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag,("bsacPrepCollect: final offset = %lu\n",offset));

  /* update total bytecount for all message buffers for this md */
  bsacPrepLastDataReply(md, bc, offset, bufIndex);
  /* fill in totbc */
  *totbc_p = bc;

  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepCollectUnits: Exiting..\n")); 

  return iss;
}

/*=============================================================================

  Name: cvtRplyBeg

  Abs:  Routine converts (for endianness) the data reply message header data

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        bpm_alldata_replybeg_ts * rb    read./write pointer to alldata reply header

  Rem:                                            

  Ret:  none


==============================================================================*/
LOCAL void cvtRplyBeg( bpm_alldata_replybeg_ts *rb)
{
  unsigned short i;

  /* init unused elements to 0 */
  rb->nmicountv = 0;
  rb->pulseidv = 0;
 
  /* now convert */
  cvtLSwap(&rb->iss);
  cvtLSwap(&rb->bt.measbits);
  cvtLSwap(&rb->datrply_totwc);
  cvtLSwap(&rb->datrply_partwc);
  cvtSSwap(&rb->ndatblks);
  cvtSSwap(&rb->nrpos);
  cvtLSwap(&rb->nmicountv);
  cvtLSwap(&rb->pulseidv);  
  for (i=0;i<(16-RNQABO_MXINJTSTMP); i++) {
	rb->rinqabo_precinj[i]=0;
	cvtLSwap(&rb->rinqabo_precinj[i]);
  }
  cvtLSwap(&rb->checksum);
}


/*=============================================================================

  Name: bsacPrepDataReplyMsg

  Abs:  Routine allocates a small or large message queue buffer, then
        pre-fills the data reply message tcpfwdhdr, msgmail and 
        alldata_rplybeg headers. The alldata_rplybeg is converted
		Called from a loop for all allocated buffers

 Args: Type	                Name        Access	   Description
       ------------------- -----------	---------- ----------------------------
       const short *         md          read        mdef # 0 to (max-1) (PV's go 1-max)
	   const short           bufIndex    read        buffer index from which to get data


  Rem:  The tcpfwdhdr and msgmail headers are and must remain unconverted.  
        Other msgSend utilities will convert those headers

  Ret:  vmsstat_t


==============================================================================*/
LOCAL vmsstat_t bsacPrepDataReplyMsg(const short md,  
									 const short bufIndex)
{
  vmsstat_t iss;
  lrgmail_ts *drMsg_ps = NULL;
  bpm_alldata_replybeg_ts rplybeg;
  char  *microname = slcGetIOCName();
  bsacDataBufs_ts *mdBuf_ps=NULL;
  unsigned long bufsize =0;

  /* get a pointer to the headers */
  bsacHdrs_ts *bsacMDReq_ps = &bsacMD_as[md].reqHdrs;
  mdBuf_ps = &bsacMDBufs_as[md];

  /* figure out the required size of buffer */
  bufsize = mdBuf_ps->bufs_as[bufIndex].partwc*2 + 
	        sizeof(msgheader_ts) + sizeof(fwd_hdr_ts)+ 
	        sizeof(unsigned long);

  /* now allocate */
  if ( bufsize  < sizeof(msgmail_ts)) {
	/* allocate a small buffer */
	iss=msgQGetBuffer(sizeof(msgmail_ts), (void *)&drMsg_ps);
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			   ("bsacPrepDataReplyMsg:allocated SMALL buf #%d of size= %ld; partwc = %ld\n",
			  bufIndex, bufsize,  mdBuf_ps->bufs_as[bufIndex].partwc ) );
  }else {
	/* allocate a large one */
	iss=msgQGetBuffer(sizeof(lrgmail_ts), (void *)&drMsg_ps);
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepDataReplyMsg:allocated LARGE buf #%d of size = %ld; partwc = %ld\n",
			  bufIndex, bufsize,  mdBuf_ps->bufs_as[bufIndex].partwc) );
  }
  

  /* insert buffer into the buffer array*/
  if (SUCCESS(iss))
  { 
	/* Copy tcpfwdhdr from the incoming msg into the outgoing msg */
	memcpy(&drMsg_ps-> tcpfwdhdr, &bsacMDReq_ps->tcpfwdhdr, sizeof(fwd_hdr_ts));

	/* fill in the reply message headers */
	drMsg_ps->job_id = bsacMDReq_ps->job_id;
	/* copy incoming time here, msgSend will put in outgoing time */
	drMsg_ps->msgheader.timestamp[0] = bsacMDReq_ps->msgheader.timestamp[0];
	drMsg_ps->msgheader.timestamp[1] = bsacMDReq_ps->msgheader.timestamp[1];
	drMsg_ps->msgheader.func         = (bsacMDReq_ps->msgheader.func + 1) | RESPONSE;
	memcpy(drMsg_ps->msgheader.dest, bsacMDReq_ps->msgheader.source, 4);
	memcpy(drMsg_ps->msgheader.source, microname, 4);

	/* now set msgheader.datalen*/
	drMsg_ps->msgheader.datalen =  mdBuf_ps->bufs_as[bufIndex].partwc;

	/*init the data reply header and convert */
    memset(&rplybeg, 0, sizeof(rplybeg));
	rplybeg.iss = BPMO_OKOK;
	rplybeg.bt.measbits = bsacMD_as[md].measbits;
	rplybeg.datrply_totwc =  mdBuf_ps->bufs_as[bufIndex].totwc;
    mdBuf_ps->bufs_as[bufIndex].cumpartwc = mdBuf_ps->bufs_as[bufIndex].partwc;
	if (bufIndex >0){
	  mdBuf_ps->bufs_as[bufIndex].cumpartwc += mdBuf_ps->bufs_as[bufIndex-1].cumpartwc;
	  /* DR - leave this; debug */ 
	  mdBuf_ps->bufs_as[bufIndex].cumpartwc -= (sizeof(bpm_alldata_replybeg_ts)/2);
	  /* end debug */
	}

	rplybeg.datrply_partwc =  mdBuf_ps->bufs_as[bufIndex].cumpartwc;
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			   ("\t bsacPrepDataReplyMsg:buf#%d reply cumpartwc/partwc = %ld; partwc = %ld\n", 
				bufIndex, rplybeg.datrply_partwc,mdBuf_ps->bufs_as[bufIndex].partwc ));
	if (bsacMD_as[md].measprep)rplybeg.ndatblks = 1; 
	else rplybeg.ndatblks = bsacMD_as[md].ndatblks;
    rplybeg.nrpos = bsacMD_as[md].nrpos;

	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			   ("\t bsacPrepDataReplyMsg:buf#%d datalen = %u\n", 
				bufIndex, drMsg_ps->msgheader.datalen));

	memcpy(&rplybeg.checksum, microname, 4);
    cvtRplyBeg(&rplybeg);

	/* now copy the temp var into the reqdata area */
	memcpy(&drMsg_ps->reqdata, &rplybeg, sizeof(rplybeg)); 

  	/* place this message buffer into the bsacMDBufs_as array for this md */
    mdBuf_ps = &bsacMDBufs_as[md];
	/* do not touch the wc fields now * - just handle the buf pointer */
    mdBuf_ps->bufs_as[bufIndex].buf_ps = drMsg_ps;
	mdBuf_ps->bufs_as[bufIndex].data_p = (unsigned char *)&drMsg_ps->reqdata;
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepDataReplyMsg:\t buf #%d has data_p=%p\n",
			  bufIndex, mdBuf_ps->bufs_as[bufIndex].data_p) );
  }
  return iss;
}


/*=============================================================================

  Name: bsacPrepNextDataReply

  Abs:  Prepare current buffer with final partwc and initialize next buffer's counts

  Args: Type	              Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        const short *        md          read        mdef # 0 to (max-1) (PV's go 1-max)
		unsigned long *      offset_p    write       ptr to buffer offset count
        short *              bufIndex_p  read/write  ptr to buffer index

  Rem:  Routine tidies up partwc for input bufIndex_p, then sets up
        *offset_p, bufCount for next buffer and increments (*bufIndex_p).                                      

  Ret:  vmsstat_t


==============================================================================*/
LOCAL void bsacPrepNextDataReply( const short md, 
								  unsigned long *offset_p,
								  unsigned long *totbc_p,
								  short *bufIndex_p)
{
  bsacDataBufs_ts *mdBuf_ps = &bsacMDBufs_as[md];

  /* update the partwc for this old buffer with the current offset */
  mdBuf_ps->bufs_as[*bufIndex_p].partwc = *offset_p/2;
  /* inc bufCount */
  mdBuf_ps->bufCount++;
  /* reset offset for the next buffer */
  *offset_p = sizeof(bpm_alldata_replybeg_ts);
  /* DR keep running tab of totbc, however don't add in any more alldata_reply sizes */
  /*  .. only reflect these in datalen field */
  /*(*totbc_p) += *offset_p;*/
  (*bufIndex_p)++;
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepNextDataReply: New buf#%d has offset=%ld\n", 
				*bufIndex_p, *offset_p ) );
}

/*=============================================================================

  Name: bsacPrepLastDataReply

  Abs:  Routine writes final bytecount info into ALL data reply messages
        We don't get another buffer here, just write byte counts into all
		current buffers.

 Args: Type	               Name          Access	   Description
       ------------------- -----------	---------- ----------------------------
       const short         md            read      mdef # 0 to (max-1) (PV's go 1-max)
	   const unsigned long totbc         read      total byte count
	   const unsigned long offset        read      buffer offset count
	   const short         bufIndex      read      buffer index

  Rem:                                            

  Ret:  vmsstat_t


==============================================================================*/
LOCAL void bsacPrepLastDataReply( const short md,
								  const unsigned long totbc,
								  const unsigned long offset,
								  const short bufIndex)
{
  unsigned short buf = 0;

  /* update this alldata header with the correct partwc */
   bsacDataBufs_ts *mdBuf_ps = &bsacMDBufs_as[md];
  /* insert the partwc for this buffer */
  mdBuf_ps->bufs_as[bufIndex].partwc = offset/2;

  DEBUGPRINT(DP_INFO, bsacPrepFlag, 
			 ("bsacPrepLastDataReply:buf #%d has partwc=%ld, totwc=%ld \n", 
				bufIndex, offset/2, totbc/2 ) );

  /* now go thru all data reply message buffers and finalize totwc*/
  for (buf = 0; buf < mdBuf_ps->bufCount; buf++) {
	mdBuf_ps->bufs_as[buf].totwc  = totbc/2;
  }

}

/*=============================================================================

  Name: bsacPrepAllocateBuffers

  Abs:  Routine allocates the message buffers required, checking the size and
        allocating only size necessary.  It 
		writes final bytecount info into ALL data reply messages.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        const short         md          read        mdef # 0 to (max-1) (PV's go 1-max)

  Rem:                                            

  Ret:  vmsstat_t


==============================================================================*/
LOCAL vmsstat_t bsacPrepAllocateBuffers( const short md)
{
  unsigned short buf=0;
  vmsstat_t iss = BPMO_OKOK;
  bsacDataBufs_ts * mdBuf_ps = &bsacMDBufs_as[md]; 

  for (buf=0; buf<mdBuf_ps->bufCount;buf++) {
	iss = bsacPrepDataReplyMsg(md, buf);
  }
	
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepAllocateBuffers:allocated %d buffers\n", 
				buf ) );
  /* set flags for the  measurement */
  bsacMDBufs_as[md].nextBuf = 0;
  bsacMD_as[md].msgCounter =  bsacMD_as[md].msgCount; 
  bsacMD_as[md].firstPulse = epicsTrue;
  bsacMD_as[md].timeoutCntr = bsacMD_as[md].timeout;

  return(iss);
}

/*=============================================================================

  Name: bsacPrepAddUnit

  Abs:  Routine adds input primary to primCollect_as and bsacMD_as structures
        in preparation for collection. Calculates, accumulates, and returns total 
		byte count (for entire acq), outgoing message buffer index and offset 
		for each unit.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		const unsigned short  prim       read      primary index
		const unsigned short  unit       read      unit index
        const short           md         read      mdef # 0 to (max-1) (PV's go 1-max)
        const unsigned long  datasize   read      datasize for unit passed inf
		unsigned long *       totbc_p    write     ptr to total byte count
		unsigned long *       offset_p   write     ptr to offset byte count
        short *               bufIndex_p write     buffer index ptr
		
  Rem:

  Side:  from measprep.c,  Measprep1   
        For each primary, if ulist is empty for that primary then each good   
        unit within either range umin:umax or range umin2:umax2 will be used  
        (with 1st calibration).  If ulist is nonempty for that primary then   
        only good units therein (ignoring umin/umax/umin2/umax2) will be used 
        (with their individually indicated calibrations).  "Good" in this     
        context (tested in Bufringbuild rather than here) means either having 
        good HSTA (or if requesting SCP has "use offline BPM's" button lit    
        then having good or ok HSTA), and is quite independent of             
        "maintenance mode" or "checkout mode".                                          

  Ret:  none

  NOTE: for buffered acquisition, an offline unit never gets to this routine, so
        the noffline count is not incremented.  This is OK - in the buffered acquisition case
		the offline unit is not included in the nunits count either.

==============================================================================*/

LOCAL vmsstat_t  bsacPrepAddUnit(const unsigned short prim,
								 const unsigned short unit,
								 const short md,
								 const unsigned long datasize,
								 unsigned long * totbc_p,
								 unsigned long * offset_p,
								 short * bufIndex_p) 				  
{
  unsigned long newOffset;
  vmsstat_t iss = BPMO_OKOK;


  /* HSTAs will be reported/considered by the epics side */
  primCollect_as[prim].unit_as[unit].mdef_as[md].included = epicsTrue;
  /* see if the unit is offline - we will subtract this from nunits later for expected msgcount */

  if (prim==BSACARRY_INDEX){
    bsacMD_as[md].nunits += primCollect_as[prim].unit_as[unit].chanCnt;
	/* start the channel counter */
	primCollect_as[prim].unit_as[unit].chanCntr = 
	  primCollect_as[prim].unit_as[unit].chanCnt;
    /* if the arry is offline, all channels are offline */
    if (primCollect_as[prim].unit_as[unit].hsta_ctrl==CTRL_OFFLINE)
 	  bsacMD_as[md].noffline += primCollect_as[prim].unit_as[unit].chanCnt;	
  } else if (prim==BSACPUID_INDEX) {
	/* don't mult by nrpos here, its done later, and this is never offline */
	bsacMD_as[md].nunits += bsacMD_as[md].navg;
  } else {
	bsacMD_as[md].nunits++;  
    if (primCollect_as[prim].unit_as[unit].hsta_ctrl==CTRL_OFFLINE)
 	  bsacMD_as[md].noffline++;	
  }

  newOffset = *offset_p + datasize  + sizeof(msgheader_ts) + sizeof(fwd_hdr_ts);
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag,
    ("bsacPrepAddUnit: datasize = %lu, newOffset = %lu; lrgmail size = %d\n", 
	 datasize, newOffset, sizeof (lrgmail_ts)));
  
  if ( newOffset  > sizeof(lrgmail_ts) ) {
	/* we'll overrun our buffer, so get a new buffer and new offset */
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag,("bsacPrepAddUnit: getting a new msg buffer\n"));
    bsacPrepNextDataReply(md, offset_p, totbc_p, bufIndex_p);
  }
  primCollect_as[prim].unit_as[unit].mdef_as[md].bufIndex = *bufIndex_p;
  primCollect_as[prim].unit_as[unit].mdef_as[md].offset = *offset_p;
  /* for buf acq, keep track of how many scanblk_prefix's there are */
  bsacMD_as[md].ndatblks++;

  *offset_p += datasize;
  *totbc_p += datasize;
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag,
   ("!!bsacPrepAddUnit: prim=%hd, unit=%hd, ndatblks=%hd, bufIndex=%hd, offset = %ld after block is added.\n",
	prim, unit, bsacMD_as[md].ndatblks,*bufIndex_p, *offset_p));
  DEBUGPRINT(DP_DEBUG, bsacPrepFlag,("bsacPrepAddUnit: totbc = %ld\n", *totbc_p));  
  return iss;
}

/*=============================================================================

  Name: bsacPrepCheckPrim

  Abs:  Routine adds input arg primary to primCollect_as and bsacMD_as structures
        in preparation for collection. Figures out which units of input arg
		primary are desired by consulting scp message ulist or umin-umax, which are 
		stored in bsacMD_as[md] after scp message is parsed.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
	    const unsigned short  measprim   read      SCP measbit primary index
		const unsigned long   datasize   read      datasize for unit passed in
		const short           md         read      mdef # 0 to (max-1) (PV's go 1-max)
		unsigned long *       totbc_p    write     ptr to total byte count
		unsigned long *       offset_p   write     ptr to offset byte count
        short *               bufIndex_p write     buffer index ptr
        unsigned long *       ulistI_p   write     ptr into SCP msg's ulist; advances
	                                                 contents for all units in this prim

  Rem:  from measprep.c,  Measprep1   
        For each primary, if ulist is empty for that primary then each good   
        unit within either range umin:umax or range umin2:umax2 will be used  
        (with 1st calibration).  If ulist is nonempty for that primary then   
        only good units therein (ignoring umin/umax/umin2/umax2) will be used 
        (with their individually indicated calibrations).  "Good" in this     
        context (tested in Bufringbuild rather than here) means either having 
        good HSTA (or if requesting SCP has "use offline BPM's" button lit    
        then having good or ok HSTA), and is quite independent of             
        "maintenance mode" or "checkout mode".                                          

  Ret:  vmsstat_t


==============================================================================*/

LOCAL vmsstat_t bsacPrepCheckPrim(const unsigned short prim,
								  const unsigned short measprim,
								  const unsigned long datasize,
								  const short md,
								  unsigned long * totbc_p,
								  unsigned long * offset_p,
								  short * bufIndex_p,
								  unsigned short * ulistI_p) 
{
  unsigned short cnt, ucnt;
  vmsstat_t iss = BPMO_OKOK;
  int szdatblk = 0;
  int offline = 0;

  /* if called from buf acq request, add to overall basic datasize */
  /* note that PUIDs call AddUnit directly */
  if (!(bsacMD_as[md].measprep)) szdatblk = sizeof(scanblk_prefix_ts);

  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepCheckPrim: Entering for primary index %hd; md # = %hd\n", 
			  prim, md)); 
  if (primCollect_as[prim].unitCnt >0) {

    /* reset mdefUnitCnt for this primary */
    primCollect_as[prim].mdefUnitCnt[md] = 0;
    /* go through prim's unit array looking for each unit in ulist/umin-umax  */
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("bsacPrepCheckPrim: Unit count for primary index %hd is = %hd\n", 
			  prim,primCollect_as[prim].unitCnt )); 
	bsacMD_as[md].nunitsCnt[measprim]=0; /* init matching units counter */

	if (bsacMD_as[md].ulist.nunits[measprim]>0 ){ /* if use ulist */

	  for (ucnt = 0; ucnt <bsacMD_as[md].ulist.nunits[measprim]; ucnt++){		
		DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
				   ("\t: Looking for %hd units for mdef; measprim index = %d; ulistI = %hd\n",
					bsacMD_as[md].ulist.nunits[measprim], measprim, *ulistI_p)); 
	    /* check our unitnum against every unit in list; if found break */

		for (cnt = 0; cnt < primCollect_as[prim].unitCnt; cnt++) {

		  if (primCollect_as[prim].unit_as[cnt].unitnum== 
			  (bsacMD_as[md].ulist.ulist[*ulistI_p] & BPMULIST_UNIT_MASK)){
			  /* a match of one of our units to ulist */
            if ( (bsacMD_as[md].measprep) || (!bsacMD_as[md].measprep && 
                 primCollect_as[prim].unit_as[cnt].hsta_ctrl != CTRL_OFFLINE) ) {
			  /* a slc HSTA of online will add it to the list; 
				 if off-line put in zeros into measprep, or, do not include in wireprep */
			  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
				("\tAdding Primary=%hd, unit_as index cnt=%hd; unit_as[cnt].unitnum=%hd\n", 
						prim,cnt,primCollect_as[prim].unit_as[cnt].unitnum ));
			  iss = bsacPrepAddUnit(prim, cnt, md, 
								  datasize * bsacMD_as[md].nrpos + szdatblk, 
								  totbc_p, offset_p, bufIndex_p);
 
			  bsacMD_as[md].nunitsCnt[measprim]++;		
			  primCollect_as[prim].mdefUnitCnt[md]++;
			} else offline++;
			(*ulistI_p)++; 
		 	break;
		  }
		}
	  }/*for all units in ulist */ 

	} else {
	  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
				 ("!bsacPrepCheckPrim: Using umin-umax\n"));
	  for (cnt = 0; cnt < primCollect_as[prim].unitCnt; cnt++) {
		DEBUGPRINT(DP_DEBUG, bsacPrepFlag,
				   ("\t unit range check unitnum=%hd\n",
					primCollect_as[prim].unit_as[cnt].unitnum));
		if ( (primCollect_as[prim].unit_as[cnt].unitnum >= bsacMD_as[md].umin &&
			  primCollect_as[prim].unit_as[cnt].unitnum <=  bsacMD_as[md].umax) ||
			 (primCollect_as[prim].unit_as[cnt].unitnum >=  bsacMD_as[md].umin2 &&
			  primCollect_as[prim].unit_as[cnt].unitnum <=  bsacMD_as[md].umax2) )
		  {
            if ( (bsacMD_as[md].measprep) ||
				 (!bsacMD_as[md].measprep && 
				   primCollect_as[prim].unit_as[cnt].hsta_ctrl != CTRL_OFFLINE) ) {
			  /* a match of one of our units to ulist */
			  /* a slc HSTA of online will add it to the list; if off-line put in zeros */
			  iss = bsacPrepAddUnit(prim, cnt, md, 
								  datasize * bsacMD_as[md].nrpos + szdatblk, 
								  totbc_p, offset_p, bufIndex_p);	
			}
		  }
	  }  /*for all units in this ioc for this prim */
	}	  	  

	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("\t Total bytecount of  measurement for this MDEF = %lu\n", 
			  *totbc_p)); 
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("\t Total num units involved in measurement for this MDEF = %lu\n", 
			  bsacMD_as[md].nunits)); 
  } /*if unitcnt > 0 */
  else {
	DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
			 ("\t No units involved in measurement!!\n")); 
  }  	  
  if (bsacMD_as[md].ulist.nunits[measprim]>0 ){
	if (bsacMD_as[md].ulist.nunits[measprim] !=bsacMD_as[md].nunitsCnt[measprim]) {
	  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
	   ("\t Mdef %hd Requested # units (%hd) != actual(%hd), due to (%hd) offline units? !!\n", 
		 md,bsacMD_as[md].ulist.nunits[measprim],bsacMD_as[md].nunitsCnt[measprim], offline)); 	  
	} 
  }
  return iss;
}							
/*=============================================================================

  Name: bsacPrepMDEF

  Abs:  Output MDEF parameters to EPICS records.
		
  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        const short            md          read      mdef # 0 to (max-1) (PV's go 1-max)
        const char *         scpName       read      scp name from msgheader

  Rem:  Output data, as specified in original scp message, to EPICs MDEF:IOC:MD:<secondary>
        records.
 

  side: 


  Ret:  vmsstat_t

==============================================================================*/
/* output to epics MDEF records */
/* reads from static global measdef_s to write to EPICS MDEF records */
LOCAL vmsstat_t bsacPrepMDEF (const short md, const char * scpName )
{
  vmsstat_t iss = BPMO_OKOK;
  unsigned short ctrldata = 1;
  unsigned short maskA[5];
  unsigned short maskB[5];
  int cnt;
  unsigned short modedata;

  /* determine proper MDEF MODE value */
  if (bsacMD_as[md].measprep) {
	if (bsacMD_as[md].measbits & MEASPULSID_BIT){
	  modedata = MODE_SLCSINGLE_PUID;
	} else {
	  modedata = MODE_SLCSINGLE;
	}
  } else {
	if (bsacMD_as[md].measbits & MEASPULSID_BIT) {
	  modedata = MODE_SLCBUFFER_PUID;
	}else {
	  modedata = MODE_SLCBUFFER;
	}
  }


  if (mdefAddr_as[md].MODE_s.precord){ 
	if (dbPutField(&mdefAddr_as[md].MODE_s, DBR_USHORT, &modedata, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:MODE failed\n", 
					  slcGetIOCName(),md));
	}else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:MODE = %hd\n", 
					  slcGetIOCName(),md+1, modedata));	
  }

  /* MDEF:IOC:MD:USERNAME - owner name = SCP name */
  if (mdefAddr_as[md].USERNAME_s.precord){   
	if (dbPutField(&mdefAddr_as[md].USERNAME_s, DBR_STRING, scpName, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:USERNAME = %s failed\n", 
					  slcGetIOCName(),md+1, scpName));
	}
	else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:USERNAME = %s\n", 
					  slcGetIOCName(),md+1, scpName));
  }
  /* MDEF:IOC:MD:ID - i_seg measdef id from SCP  */
  if (mdefAddr_as[md].ID_s.precord){
    /* Note - you can output a short to an ao - no need to convert first */
	if (dbPutField(&mdefAddr_as[md].ID_s, DBR_USHORT, &measdef_s.measid, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:ID = %d failed\n", 
					  slcGetIOCName(),md+1, measdef_s.measid ));
	}else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:ID = %d\n", 
					  slcGetIOCName(),md+1, measdef_s.measid ));	
  }
  if (mdefAddr_as[md].AVGCNTMAX_s.precord){
	if (dbPutField(&mdefAddr_as[md].AVGCNTMAX_s, DBR_USHORT, &bsacMD_as[md].navg, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:AVGCNTMAX = %hd failed\n", 
					  slcGetIOCName(),md+1,bsacMD_as[md].navg ));
	}else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:AVGCNTMAX = %d\n", 
					  slcGetIOCName(),md+1, bsacMD_as[md].navg ));
	
  }
  if (mdefAddr_as[md].MEASCNTMAX_s.precord){
	if (dbPutField(&mdefAddr_as[md].MEASCNTMAX_s, DBR_USHORT, &bsacMD_as[md].nrpos, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:MEASCNTMAX = %hd failed\n", 
					  slcGetIOCName(),md,bsacMD_as[md].nrpos ));
	}else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:MEASCNTMAX = %d\n", 
					  slcGetIOCName(),md+1, bsacMD_as[md].nrpos ));	
  }
  if (mdefAddr_as[md].YY_s.precord){
	if (dbPutField(&mdefAddr_as[md].YY_s, DBR_UCHAR, &measdef_s.yy, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:YY = %c failed\n", 
					  slcGetIOCName(),md,measdef_s.yy ));
	}else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:YY = %d\n", 
					  slcGetIOCName(),md+1, measdef_s.yy ));	
  }
  for (cnt=0;cnt<4;cnt++){
    if (mdefAddr_as[md].PP_as[cnt].precord){
      if (dbPutField(&mdefAddr_as[md].PP_as[cnt], DBR_UCHAR, &measdef_s.pp[cnt], 1)) {
		DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
		   ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:PP%hd = %c failed\n", 
		    slcGetIOCName(),md,cnt+1, measdef_s.pp[cnt] ));
	  }	else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:PP = %d\n", 
					  slcGetIOCName(),md+1, measdef_s.pp[cnt] ));
    }
  }
  /* convert the ulong incoming incl masks into u shorts */
  /* debug - set dummy values 0x0000AAAA */
  /* for (cnt=0; cnt<3; cnt++) measdef_s.inclmask[cnt]=43690;*/  
  for (cnt=0; cnt<3; cnt++) {
	cvtULongtoUShorts(measdef_s.inclmask[cnt], &maskB[cnt], &maskA[cnt]);
  }
  /* output to INCLUSION1-3A/B */
  for (cnt=0;cnt<3;cnt++){
    if (mdefAddr_as[md].INCLUSIONA_as[cnt].precord){
      if (dbPutField(&mdefAddr_as[md].INCLUSIONA_as[cnt], DBR_USHORT, &maskA[cnt], 1)) {
		DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
		   ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:INCLUSION%dA = %d failed\n", 
		    slcGetIOCName(),md+1,cnt+1,maskA[cnt] ));
      }	else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:inclusion%dA = %d\n", 
					  slcGetIOCName(),md+1,cnt+1, maskA[cnt] ));
    }
    if (mdefAddr_as[md].INCLUSIONB_as[cnt].precord){
      if (dbPutField(&mdefAddr_as[md].INCLUSIONB_as[cnt], DBR_USHORT, &maskB[cnt], 1)) {
		DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
		   ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:INCLUSION%dB = %d failed\n", 
		    slcGetIOCName(),md+1,cnt+1,maskB[cnt] ));
      }	else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:inclusion%dB = %d\n", 
					  slcGetIOCName(),md+1,cnt+1, maskB[cnt] ));
    }
  }
  /* debug - set dummy values */
  /* for (cnt=0; cnt<3; cnt++) measdef_s.exclmask[cnt]=21845;*/ 

  /* slc threads don't write to inclusion masks 3 and 4; reserved for lcls */  
  /* convert the ulong incoming incl and excl masks into u shorts */
  for (cnt=0; cnt<3; cnt++) {
	cvtULongtoUShorts(measdef_s.exclmask[cnt], &maskB[cnt], &maskA[cnt]);
  }
  /* output to EXCLUSION1-3A/B */
  for (cnt=0;cnt<3;cnt++){
    if (mdefAddr_as[md].EXCLUSIONA_as[cnt].precord){
      if (dbPutField(&mdefAddr_as[md].EXCLUSIONA_as[cnt], DBR_USHORT, &maskA[cnt], 1)) {
		DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
		   ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:EXCLUSION%dA = %d failed\n", 
		    slcGetIOCName(),md+1,cnt+1,maskA[cnt] ));
      }	
    }
    if (mdefAddr_as[md].EXCLUSIONB_as[cnt].precord){
      if (dbPutField(&mdefAddr_as[md].EXCLUSIONB_as[cnt], DBR_USHORT, &maskB[cnt], 1)) {
		DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
		   ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:EXCLUSION%dB = %d failed\n", 
		    slcGetIOCName(),md+1,cnt+1,maskB[cnt] ));
      }	
    }
  }
  /* slc threads don't write to exclusion masks 3 and 4; reserved for lcls */

  /* ALWAYS last - bo MDEF:IOC:MD:CTRL */
  /* output 1 = ON */
  /* output 1 = ON  - will also enable the EVR GO record */
  if (mdefAddr_as[md].CTRL_s.precord){ 
	if (dbPutField(&mdefAddr_as[md].CTRL_s, DBR_USHORT, &ctrldata, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:CTRL failed\n", 
					  slcGetIOCName(),md));
	}else DEBUGPRINT(DP_DEBUG, bsacPrepFlag, 
					 ("bsacPrepMDEF: dbPutField of MDEF:%s:M%d:CTRL = %hd\n", 
					  slcGetIOCName(),md+1, ctrldata));	
  } 

  return iss;
}

/* output to all units involved CTRL records */
/*=============================================================================

  Name: bsacPrepGo

  Abs:  Set PRIM:LOCA:UNIT:CTRL$MD to 1 (ON) for all records
        included in the measurement.
		

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        const short            md          read      mdef # 0 to (max-1) (PV's go 1-max)

  Rem: 

  side: 

  Ret: iss vmsstat_t vms style status: BPMO_FAIL, or BPMO_OKOK

==============================================================================*/
LOCAL vmsstat_t bsacPrepGo (const short md ){
  vmsstat_t iss = BPMO_OKOK;
  int prim, unit;             /* epics records are longouts */
  unsigned short ctrl_on = 1;

  /* for now check devices BPMS, TOROs, ARRYs */
  for (prim=0; prim<=BSACARRY_INDEX; prim++) {
	for (unit=0; unit<primCollect_as[prim].unitCnt; unit++) {
	  if ((primCollect_as[prim].unit_as[unit].mdef_as[md].included) && 
		  (primCollect_as[prim].unit_as[unit].hsta_ctrl!=CTRL_OFFLINE)){
		if (primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord){
		  if (dbPutField(&primAddr_as[prim].unit_as[unit].mdefCTRL_as[md], DBR_USHORT, &ctrl_on, 1)) {
			iss=BPMO_FAIL;
			DEBUGPRINT(DP_ERROR, bsacPrepFlag,
                                   ("bsacPrepGo: dbPutField of %.4s:%s:%s:CTRLM%d failed\n",
                                    (char *)primCollect_as[prim].slcPrim_a,
                                    slcGetLOCAName(),
									(char *)primCollect_as[prim].unit_as[unit].slcUnit_a,md+1));
		  } else DEBUGPRINT(DP_DEBUG, bsacPrepFlag,
                                    ("bsacPrepGo: dbPutField of %.4s:%s:%s:CTRLM%d to 1 successful p=%p\n",
                                     (char *)primCollect_as[prim].slcPrim_a,
                                     slcGetLOCAName(),
									 (char *)primCollect_as[prim].unit_as[unit].slcUnit_a,md+1,
									 &primAddr_as[prim].unit_as[unit].mdefCTRL_as[md]));
		} else DEBUGPRINT(DP_DEBUG, bsacPrepFlag,
                ("bsacPrepGo:unit precord=NULL: %.4s:%s:%s:CTRLM%d \n",
                                     (char *)primCollect_as[prim].slcPrim_a,
                                     slcGetLOCAName(),
									 (char *)primCollect_as[prim].unit_as[unit].slcUnit_a,md+1));
	  } else {
		
		DEBUGPRINT(DP_DEBUG, bsacPrepFlag,
                ("bsacPrepGo:unit not included, or offline:%.4s:%s:%s:CTRLM%d\n",
                                     (char *)primCollect_as[prim].slcPrim_a,
                                     slcGetLOCAName(),
									 (char *)primCollect_as[prim].unit_as[unit].slcUnit_a,md+1));
	  }
	}
  }
 return iss;
}

/*=============================================================================

  Name: bsacPrepImmedReply

  Abs: copy msg header stuff into immed response message

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		msgmail_ts *        rMsg_ps     read/write pointer to the received message header

  Rem:  copies header values into the send message buffer.  Then sets the RESPONSE bit in the 
        msgheader.func. Sets the source and dest fields, puts the BPMO_OKOK reply longword into
		the first two words of reqdata.  

  Side: 

  Ret:  vmsstat_t the status code, BPMO_OKOK or MICR_MALLOC
   
==============================================================================*/
vmsstat_t bsacPrepImmedReply(const msgmail_ts * rMsg_ps,
								   const vmsstat_t iss,
								   const unsigned long totbc,
								   const unsigned short npulstep)
{
  vmsstat_t thisiss = BPMO_OKOK;
  msgmail_ts*     sMsg_ps = NULL;  /* immed reply msg to send */
  bpm_allprep_reply_ts immedReply_s;
  int i;

  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("bsacPrepImmedReply: Entering..\n")); 
  /* get a buffer for the reply data */
  thisiss = msgQGetBuffer( sizeof(msgmail_ts),(void*) &sMsg_ps);
  if ( SUCCESS(thisiss) )
  {

	/* add in the immed reply in the reqdata part of message*/
	/* stuff data part of immed response message & convert to VMS if nec */
	immedReply_s.iss = iss; /* vmsstat_t = int4u */
	immedReply_s.nmuxmin = 1; /* unsigned char, Tony advises = 1 */
	for (i=0; i<3; i++) {
	  immedReply_s.spareb[i]=0;}  /* 3 unsigned chars */
	immedReply_s.datrply_totwc = totbc/2; /* unsigned long */
	immedReply_s.sparel = 0;              /* unsigned long */
	immedReply_s.npulstep = npulstep;     /* unsigned short */
	immedReply_s.sparew = 0;              /* unsigned short */
	cvtLSwap((unsigned long *)&immedReply_s.iss);
	cvtLSwap((unsigned long *)&immedReply_s.datrply_totwc);
	cvtSSwap((unsigned short *)&immedReply_s.npulstep);
    memcpy(&sMsg_ps->reqdata, &immedReply_s, sizeof(immedReply_s));


	/* now fill in the reply message headers */
	thisiss = msgQSlcReply(rMsg_ps->msgheader.func, 
				 sizeof(immedReply_s), 
				 rMsg_ps, sMsg_ps);

    /* debug - print the message */
	/*slcPrintMsgheader(&sMsg_ps->msgheader, bsacPrepFlag );
	for (i=0; i<sMsg_ps->msgheader.datalen; i++) {
	  DEBUGPRINT(DP_DEBUG, bsacPrepFlag, ("\tthe (VMS formatted) data[%d]: 0x%.4x\n", 
										  i, sMsg_ps->reqdata[i]) );
										  }*/

	/* calc some diagnostics ? see msgHdlr sendOKReply */

    /* send the msg to the msgSend thread */
	DEBUGPRINT(DP_INFO, bsacPrepFlag, ("bsacPrepImmedReply: Sending to msgSend..\n")); 
  
    /* calc some diagnostics */
    /*if (!SUCCESS(iss))
	  msgHdlrDiags_s.rplyFails++;
	else
	msgHdlrDiags_s.rplyCnt++;*/

  }
  
 return thisiss;
}

/*-----------------------------------------------------------------------
  Name: bsacPrepForNewCycle

  Abs: setup large msg buffers for this md, prepare it for another yy cycle

  Args: Type		      Name    Access	    Description
	    ----------------- ------- ---------- ----------------------------
        const short       md   read only  mdef whose buffer must be sent
  Rem: 

  Side: 

  Ret:  vmsstat_t vms style status
----------------------------------------------------------------------*/
vmsstat_t bsacPrepForNewCycle(const short md)
{
  unsigned long totbc = 0; /* dummy for this routine */
  vmsstat_t iss = BPMO_OKOK;

  iss = bsacPrepCollectUnits(md, &totbc);
  iss = bsacPrepAllocateBuffers(md);
  return(iss);
}


