/*=============================================================================
 
  Name: bsacPrepInit.c


  Abs:  bsacPrep performs the measurement preparation for BPMO_MEAS_PREP
        and BPMO_WIRE_PREP.
        Reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | BSAC Service Design

  Auth: 20 jul 2005, dROGIND    created 
  Rev:  

NOTE: the device CTRL PVs (hsta equiv, not MDEF ctrl) are RENAMED TO ACCESS, but 
variable names, comments will still refer the CTRL.  Now when you read 
device CTRL, think ACCESS!!
-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  
       13-jul-2007, dfairley renamed device CTRL pv to ACCESS, but beware; vars and 
	      comments will still refer to CTRL
       27-apr-2007, dfairley started LOCA2 fixes
  10 Oct 2006, changed name of DEVICE:LOCA:UNIT:CTRL$MD to 
                     DEVICE:LOCA:UNIT:CTRLM$MD to differentiate from bsac event pvs  
  (newest to oldest) 
        20 jul 2005, DROGIND    created 
=============================================================================*/
/* SCP Operations notes
   BPMSs and TOROs are allowed to be collected together from the SCP
   If, on BSAC panel, Device btn says BPM, the measbits will be set for
     BPMS, TORO, and pulseid.
   For ARRYs, click Device and choose GADC, (then One Shot, etc)
 */
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int bsacPrepInitFlag = DP_ERROR;
#endif


/* c includes */
#include <string.h>        /* memcpy */

/* epics */
#include <dbAccess.h>      /* EPICS dbAddr struct */

/* shared slc micro includes */
#include "bpmfunc.hc"      /* bpmo_xxx func codes */
#include "micrdef.hc"      /* MICR_EXNET_ codes */
#include "slcMsg.h"        /* IOC_STOP, BSAC_* */
#include "msgfunc.hc"      /* RESPONSE for func code */
#include "bpmodef.hc"      /* BPMO_OKOK */

/* slc global vars */
#include "slcGlob.h"       /* slcThreads_as */

/* slcIOC types */
#include "msgTypes.h"      /* msgmail_ts etc */
/* slcIOC utilities */
#include "slcUtil.h"       /* slcPrintMsgheader */
#include "msgQUtil.h"      /* messageQSend, messageQRecv */
#include "cvtUtil.h"       /* slcSSwap, slcLSwap, etc */
#include "slcCmlog.h"      /* slcCmlogLogMsg */
#include "timeUtil.h"      /* timeGetVMS, timeDiffVMS10usec */

/* slcMain function prototypes */
#include "slcMain.h"       /* slcStop(), slcRestart() */

/* bsac includes */
#include    "bsacGlob.h"
#include    "bsacPrep.h"
#include    "bsacPrepTypes.h"
#include    "bpm_reqreply_struc.h"
#include    "bpmstruc.h"
#include    "bsacAbort.h"


/* declare bsacPrep/bsacWire related globals */
mdefAddr_ts mdefAddr_as[MDEF_MAX];
primAddrStruct_ts primAddr_as[BSACPRIM_MAX]; /* contains epics db addrs for */
                  /* records PRIM:LOCA:UNIT:UNITID */
                  /*  and PRIM:LOCA:UNIT:CTRL in bsacStates */


/*-----------------------------------------------------------------------
  Name: bsacInitMDEFAddr

  Abs: initializes the mdefAddr_as structure by calling epics util
       dbNameToAddr() for each MDEF:IOC:MD:SECN used in order to 
       establish epiocs db addresses to be referenced throughout the life 
       of the bsac facilty.

  Args: None

  Rem: 

  Side: MDEFS use slc ioc name in PV name because they are not devices
        in he beam.

  Ret:  void
----------------------------------------------------------------------*/
void bsacPrepInitMDEFAddr (void)
{
  short md;
  char         pvname_ca[80];           /* Epics PV channel name */
  unsigned short mbboValue = 0;
  int cnt;
  
  for (md = 0; md < MDEF_MAX; md++) {
	/* MDEF:IOC:MD:CTRL bo */
	/* inits to 0 by macro on VAL, PINI yes */
	sprintf (pvname_ca, "MDEF:%s:M%d:CTRL.VAL", slcGetIOCName(), md+1);
	if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].CTRL_s))) {
	  mdefAddr_as[md].CTRL_s.precord =NULL;      /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
				 ("bsacPrepInitMDEFAddr: dbNameToAddr of %s failed\n", (char *)pvname_ca));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	} else
	  DEBUGPRINT(DP_DEBUG, bsacPrepInitFlag, 
				 ("bsacPrepInitMDEFAddr: dbNameToAddr of %s successful; CTRL_s = %p\n", 
				  (char *)pvname_ca, mdefAddr_as[md].CTRL_s.pfield));	


	/* MDEF:IOC:MD:USERNAME */
	/* inits to "" by macro on VAL, PINI yes */ 
	/* NAME = MDEF name is inited to "SCPxx"*/
	sprintf (pvname_ca, "MDEF:%s:M%d:USERNAME.VAL", slcGetIOCName(), md+1);
	if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].USERNAME_s))) {
	  mdefAddr_as[md].USERNAME_s.precord =NULL;    /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
				 ("bsacPrepInitMDEFAddr: dbNameToAddr of %s failed\n", (char *)pvname_ca));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	} 

	/* MODE = MDEF meas type is inited to 0, and reset after meas to 0 = USER */
    /* 1=FEEDBACK , 2= SLCSINGLE, 3= SLCBUFFER */
	sprintf (pvname_ca, "MDEF:%s:M%d:MODE.VAL", slcGetIOCName(), md+1);
	if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].MODE_s))) {
	  mdefAddr_as[md].MODE_s.precord =NULL;    /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
				 ("bsacPrepInitMDEFAddr: dbNameToAddr of %s failed\n", (char *)pvname_ca));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	}
	/* MDEF:IOC:MD:ID  SCP Identification*/
    /* inits to 0 by VAL=0; PINI yes */
	sprintf (pvname_ca, "MDEF:%s:M%d:ID.VAL", slcGetIOCName(), md+1);
	if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].ID_s))) {
	  mdefAddr_as[md].ID_s.precord =NULL;      /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
				 ("bsacPrepInitMDEFAddr: dbNameToAddr of %s failed\n", (char *)pvname_ca));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	}
	/* MDEF:IOC:MD:AVGCNTMAX */
	/* inits to $(NAVG) macro on VAL; PINI yes */
	sprintf (pvname_ca, "MDEF:%s:M%d:AVGCNTMAX.VAL", slcGetIOCName(), md+1);
	if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].AVGCNTMAX_s))) {
	  mdefAddr_as[md].AVGCNTMAX_s.precord =NULL;      /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, ("bsacPrepInitMDEFAddr: dbNameToAddr of MDEF:%s:%d:AVGCNTMAX failed\n", slcGetIOCName(), md+1));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	}
	/* MDEF:IOC:MD:MEASCNTMAX */
	/* inits to VAL=1; PINI yes */
	sprintf (pvname_ca, "MDEF:%s:M%d:MEASCNTMAX.VAL", slcGetIOCName(), md+1);
	if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].MEASCNTMAX_s))) {
	  mdefAddr_as[md].MEASCNTMAX_s.precord =NULL;      /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, ("bsacPrepInitMDEFAddr: dbNameToAddr of MDEF:%s:%d:MEASCNTMAX failed\n", slcGetIOCName(), md+1));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	}
	/* MDEF:IOC:MD:YY */
	/* inits to VAL= -1; PINI yes */
	sprintf (pvname_ca, "MDEF:%s:M%d:YY.VAL", slcGetIOCName(), md+1);
	if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].YY_s))) {
	  mdefAddr_as[md].YY_s.precord =NULL;      /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, ("bsacPrepInitMDEFAddr: dbNameToAddr of %s failed\n", (char *)pvname_ca));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	}
	/* MDEF:IOC:MD:PP1-4 */
	/* inits to VAL= -1; PINI yes */
	for (cnt=0; cnt<4; cnt++){
	  sprintf (pvname_ca, "MDEF:%s:M%d:PP%d.VAL", slcGetIOCName(), md+1, cnt+1);
	  if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].PP_as[cnt]))) {
	    mdefAddr_as[md].PP_as[cnt].precord =NULL;      /* keeps us from doing dbGet */
	    DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, ("bsacPrepInitMDEFAddr: dbNameToAddr of %s failed\n", 
						(char *)pvname_ca));
		slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	  }
	}

	/* MDEF:IOC:MD:INCLUSION1-5A and B */
	/* inits to VAL= $(A) or $(B) = 0; PINI yes */
	for (cnt=0; cnt<5; cnt++){
	  sprintf (pvname_ca, "MDEF:%s:M%d:INCLUSION%dA.VAL", slcGetIOCName(), md+1, cnt+1);
	  if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].INCLUSIONA_as[cnt]))) {
	    mdefAddr_as[md].INCLUSIONA_as[cnt].precord =NULL;      /* keeps us from doing dbGet */
	    DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
				   ("bsacPrepInitMDEFAddr: dbNameToAddr of %s failed\n", 
					(char *)pvname_ca));
		slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	  }
	  sprintf (pvname_ca, "MDEF:%s:M%d:INCLUSION%dB.VAL", slcGetIOCName(), md+1, cnt+1);
	  if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].INCLUSIONB_as[cnt]))) {
	    mdefAddr_as[md].INCLUSIONB_as[cnt].precord =NULL;      /* keeps us from doing dbGet */
	    DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
				   ("bsacPrepInitMDEFAddr: dbNameToAddr of %sfailed\n", 
					(char *)pvname_ca));
		slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	  }
	}

	for (cnt=3; cnt<5; cnt++){	
	/* MDEF:IOC:MD:INCLUSION4-5A/B - init to zeroes */
	  /* init to zero; these records are reserved for LCLS, not SLC use */
	  if (mdefAddr_as[md].INCLUSIONA_as[cnt].precord) {
	    if (dbPutField(&mdefAddr_as[md].INCLUSIONA_as[cnt], DBR_USHORT, &mbboValue, 1)) {
		  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
		  ("bsacPrepInitMDEFAddr: dbPutField of MDEF:%s:%d:INCLUSION%dA failed during init\n", 
		   slcGetIOCName(),md+1, cnt+1));
		}
	  }
	}
	/* MDEF:IOC:MD:EXCLUSION1-5A and B */
	/* MDEF:IOC:MD:INCLUSION1-5A and B */
	
	for (cnt=0; cnt<5; cnt++){
	  sprintf (pvname_ca, "MDEF:%s:M%d:EXCLUSION%dA.VAL", slcGetIOCName(), md+1, cnt+1);
	  if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].EXCLUSIONA_as[cnt]))) {
	    mdefAddr_as[md].EXCLUSIONA_as[cnt].precord =NULL;      /* keeps us from doing dbGet */
	    DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
				   ("bsacPrepInitMDEFAddr: dbNameToAddr of %s failed\n", 
					(char *)pvname_ca));
		slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	  }
	  sprintf (pvname_ca, "MDEF:%s:M%d:EXCLUSION%dB.VAL", slcGetIOCName(), md+1, cnt+1);
	  if ((dbNameToAddr(pvname_ca, &mdefAddr_as[md].EXCLUSIONB_as[cnt]))) {
	    mdefAddr_as[md].EXCLUSIONB_as[cnt].precord =NULL;      /* keeps us from doing dbGet */
	    DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
				   ("bsacPrepInitMDEFAddr: dbNameToAddr of %sfailed\n", 
					(char *)pvname_ca));
		slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	  }
	}
  }
}
/*-----------------------------------------------------------------------
  Name: bsacPrepInitDevIDs

  Abs: initializes the primAddr_as structure by calling epics util
       dbNameToAddr() for each PRIM:LOCA:UNIT:UNITID used in order to 
       establish epiocs db addresses to be referenced throughout the life 
       of the bsac facilty.

  Args: None

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/
void bsacPrepInitDevIDs (void)
{
  unsigned long prim, unit;             /* epics records are longouts */
  char         pvname_ca[80];           /* Epics PV channel name */
  char         pvname2_ca[80];          /* Epics PV channle name with second LOCA */

  /* create epics addrs to all the records for this slcIOC's units */
  /* for now do BPMS, TOROs, ARRYs */
  for (prim=0; prim<BSACARRY_INDEX; prim++) {
	for (unit=0; unit<primCollect_as[prim].unitCnt; unit++) {
		/* now do CTRL */
	  sprintf (pvname_ca, "%.4s:%s:%d:ACCESS.VAL", 
			   primCollect_as[prim].slcPrim_a, 
			   slcGetLOCAName(),
			   primCollect_as[prim].unit_as[unit].unitnum );     
	  if ((dbNameToAddr(pvname_ca, &primAddr_as[prim].unit_as[unit].CTRL_s))) {
        if (slcGetLOCA2Name()!=NULL) {
		  sprintf (pvname2_ca, "%.4s:%s:%d:ACCESS.VAL", 
			   primCollect_as[prim].slcPrim_a, 
			   slcGetLOCA2Name(),
			   primCollect_as[prim].unit_as[unit].unitnum );     
		  if ((dbNameToAddr(pvname2_ca, &primAddr_as[prim].unit_as[unit].CTRL_s))) {
			primAddr_as[prim].unit_as[unit].CTRL_s.precord =NULL; /* keeps us from doing dbGet */
			DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
			   ("bsacPrepInitDevIDs: dbNameToAddr of %s failed\n", pvname2_ca));
			slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname2_ca);
		  }	else DEBUGPRINT(DP_DEBUG, bsacPrepInitFlag, 
				 ("bsacPrepInitDevIDs: dbNameToAddr of %s successful; CTRL_s.precord = %p\n", 
				  (char *)pvname2_ca, primAddr_as[prim].unit_as[unit].CTRL_s.precord));
		} else {
			primAddr_as[prim].unit_as[unit].CTRL_s.precord =NULL; /* keeps us from doing dbGet */
			DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
			   ("bsacPrepInitDevIDs: dbNameToAddr of %s failed\n", pvname_ca));
			slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
		}
	  }	else DEBUGPRINT(DP_DEBUG, bsacPrepInitFlag, 
				 ("bsacPrepInitDevIDs: dbNameToAddr of %s successful; CTRL_s.precord = %p\n", 
				  (char *)pvname_ca, primAddr_as[prim].unit_as[unit].CTRL_s.precord));	

                /* now do UNITID */
	  sprintf (pvname_ca, "%.4s:%s:%d:UNITID.VAL", 
			   primCollect_as[prim].slcPrim_a, 
			   slcGetLOCAName(),
			   primCollect_as[prim].unit_as[unit].unitnum );     
	  if ((dbNameToAddr(pvname_ca, &primAddr_as[prim].unit_as[unit].UNITID_s))) {
		if (slcGetLOCA2Name()!=NULL) {
		  sprintf (pvname2_ca, "%.4s:%s:%d:UNITID.VAL", 
			   primCollect_as[prim].slcPrim_a, 
			   slcGetLOCA2Name(),
			   primCollect_as[prim].unit_as[unit].unitnum );     
		  if ((dbNameToAddr(pvname2_ca, &primAddr_as[prim].unit_as[unit].UNITID_s))) {
			primAddr_as[prim].unit_as[unit].UNITID_s.precord =NULL; /* keeps us from doing dbGet */
			DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
			   ("bsacPrepInitDevIDs: dbNameToAddr of %s failed\n", pvname2_ca));
			slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname2_ca);
		  } else {
			/* success - dbPutField the unitindex */
			if (dbPutField( &primAddr_as[prim].unit_as[unit].UNITID_s, 
				   DBR_LONG, &unit, 1)) {
			  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
						 ("bsacPrepInitDevIDs: dbPutField of :%s failed\n",pvname2_ca));
			}
		  }
		}
	    else {
		  primAddr_as[prim].unit_as[unit].UNITID_s.precord =NULL; /* keeps us from doing dbGet */
		  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
			   ("bsacPrepInitDevIDs: dbNameToAddr of %s failed\n", pvname_ca));
		  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
		}
	  } 
	  else {
		/* success - dbPutField the unitindex */
		if (dbPutField( &primAddr_as[prim].unit_as[unit].UNITID_s, 
				   DBR_LONG, &unit, 1)) {
		  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
					 ("bsacPrepInitDevIDs: dbPutField of %s failed\n",pvname_ca));
		}
			   
	  }
	}
	DEBUGPRINT(DP_INFO, bsacPrepInitFlag, 
			 ("bsacPrepInitDevIDs: Unit count for primary index %ld is = %hd\n", 
			  prim, primCollect_as[prim].unitCnt ));
  }

  /* now do the ARRYs, special case, they use the slc IOC name */
  prim=BSACARRY_INDEX;
  for (unit=0; unit<primCollect_as[prim].unitCnt; unit++) {
		/* do CTRL */
	sprintf (pvname_ca, "%.4s:%s:%d:ACCESS.VAL", 
			   primCollect_as[prim].slcPrim_a, 
			   slcGetIOCName(),
			   primCollect_as[prim].unit_as[unit].unitnum );     
	if (dbNameToAddr(pvname_ca, &primAddr_as[prim].unit_as[unit].CTRL_s)) {
	  primAddr_as[prim].unit_as[unit].CTRL_s.precord =NULL; /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
			   ("bsacPrepInitDevIDs: dbNameToAddr of %s failed\n", pvname_ca));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	}
       /* now do UNITID */
    sprintf (pvname_ca, "%.4s:%s:%d:UNITID.VAL", 
			   primCollect_as[prim].slcPrim_a, 
			   slcGetIOCName(),
			   primCollect_as[prim].unit_as[unit].unitnum );     
	if (dbNameToAddr(pvname_ca, &primAddr_as[prim].unit_as[unit].UNITID_s)) {
	  primAddr_as[prim].unit_as[unit].UNITID_s.precord =NULL; /* keeps us from doing dbGet */
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
			   ("bsacPrepInitDevIDs: dbNameToAddr of %s failed\n", pvname_ca));
	  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	}
	else {
		/* success - dbPutField the unitindex */
	  if (dbPutField( &primAddr_as[prim].unit_as[unit].UNITID_s, 
				   DBR_LONG, &unit, 1)) {
	  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
					 ("bsacPrepInitDevIDs: dbPutField of %s failed\n",pvname_ca));
		   
	  }
	}

	DEBUGPRINT(DP_INFO, bsacPrepInitFlag, 
			 ("bsacPrepInitDevIDs: Unit count for primary index %ld is = %hd\n", 
			  prim, primCollect_as[prim].unitCnt ));
  }
}

/*-----------------------------------------------------------------------
  Name: bsacPrepInitGo

  Abs: initializes the mdefAddr_as structure by calling epics util
       dbNameToAddr() for each DEVICE:LOCA:UNIT:CTRL$MD used in
       order to establish epics db addresses to be referenced throughout the
       life of the bsac facilty.
  
  Args: None

  Rem: 

  Side: 

  Ret:  void
----------------------------------------------------------------------*/
void bsacPrepInitGoAddr(void)
{
  unsigned long prim, unit;             /* epics records are longouts */
  char         pvname_ca[80];           /* Epics PV channel name */
  char         pvname2_ca[80];           /* Epics PV channel name with LOCA2 */
  int md;
  /* create epics addrs to all the records for this slcIOC's units */
  /* do BPMS, TOROs, ARRYs */
  for (prim=0; prim<BSACARRY_INDEX; prim++) {
	for (unit=0; unit<primCollect_as[prim].unitCnt; unit++) {
	  for (md=0; md< MDEF_MAX; md++) {
		/* do CTRL record */
		sprintf (pvname_ca, "%.4s:%s:%d:CTRLM%d.VAL", 
				 primCollect_as[prim].slcPrim_a, 
				 slcGetLOCAName(),
				 primCollect_as[prim].unit_as[unit].unitnum,
				 md+1); 

		DEBUGPRINT(DP_DEBUG, bsacPrepInitFlag, 
				 ("bsacPrepInitGo: attempting dbNameToAddr of %s  CTRLM%d .precord = %p\n", 
				  (char *)pvname_ca, md+1,primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord));

		if ((dbNameToAddr(pvname_ca, &primAddr_as[prim].unit_as[unit].mdefCTRL_as[md]))) {
		  if (slcGetLOCA2Name()!=NULL) {
			sprintf (pvname2_ca, "%.4s:%s:%d:CTRLM%d.VAL", 
				 primCollect_as[prim].slcPrim_a, 
				 slcGetLOCA2Name(),
				 primCollect_as[prim].unit_as[unit].unitnum,
				 md+1); 
			if ((dbNameToAddr(pvname2_ca, &primAddr_as[prim].unit_as[unit].mdefCTRL_as[md]))) {
			  primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord =NULL; /* keeps us from doing dbGet */
			  DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
					 ("bsacPrepInitGo: dbNameToAddr of %s failed\n", (char *)pvname2_ca));
			  slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname2_ca);
			} else DEBUGPRINT(DP_DEBUG, bsacPrepInitFlag, 
				 ("bsacPrepInitGo: dbNameToAddr of %s successful; CTRLM%d .precord = %p\n", 
				  (char *)pvname2_ca, md+1,primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord));

		  }
		  else {
			primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord =NULL; /* keeps us from doing dbGet */
			DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
					 ("bsacPrepInitGo: dbNameToAddr of %s failed\n", (char *)pvname_ca));
			slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
		  } 		  
		} else DEBUGPRINT(DP_DEBUG, bsacPrepInitFlag, 
				 ("bsacPrepInitGo: dbNameToAddr of %s successful; CTRLM%d .precord = %p\n", 
				  (char *)pvname_ca, md+1,primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord));
	  }	
	}
  }
  /* special case, ARRY uses slc IOC name instead of LOCA */
  prim=BSACARRY_INDEX;
  for (unit=0; unit<primCollect_as[prim].unitCnt; unit++) {
	for (md=0; md< MDEF_MAX; md++) {
		/* do CTRL record */
		sprintf (pvname_ca, "%.4s:%s:%d:CTRLM%d.VAL", 
				 primCollect_as[prim].slcPrim_a, 
				 slcGetIOCName(),
				 primCollect_as[prim].unit_as[unit].unitnum,
				 md+1); 
		if (dbNameToAddr(pvname_ca, &primAddr_as[prim].unit_as[unit].mdefCTRL_as[md])) {
			primAddr_as[prim].unit_as[unit].mdefCTRL_as[md].precord =NULL; /* keeps us from doing dbGet */
			DEBUGPRINT(DP_ERROR, bsacPrepInitFlag, 
					 ("bsacPrepInitGo: dbNameToAddr of %s failed\n", (char *)pvname_ca));
			slcCmlogLogMsg(BPMO_DBERR, "DBERR", "dbNameToAddr failure for %s.\n", (char *)pvname_ca);
	   } 		  
	}	
  }
}

