/*=============================================================================
 
  Name: cstrAsync.c
         
        cstrAsyncThread       -   main routine for cstrAsync thread.
        cstrAsyncInit
        cstrAsyncSendFuncMsg
        cstrAsyncExit
           

  Abs:  cstrAsync contains the cstrAsyncThread() thread which figures out
        the next cycling function to be executed and sends a message to a 
        handler to execute it.

  Auth: 03 Mar 2005, Ron MacKenzie (ronm)    created 
  Rev:  20 Sep 2005, Stephanie Allison

-----------------------------------------------------------------------------*/


#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  

=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
int cstrFlag = DP_ERROR;           /* Set to DP_INFO for production */ 
#endif

/* c includes */
#include <string.h>                                                /* memcpy */

/* slc ioc prototypes */
#include "epicsTypes.h"                                     /* epicsBoolean  */
#include "slcAsync.h"                                       /* prototypes    */

/* shared slc micro includes */
#include "asyncdef.hc"                              /* ASYNC_OKOK error code */

/* slc global vars */
#include "slcGlob.h"                                        /* slcThreads_as */

/* slcIOC types */
#include    "slcMsg.h"                                 /* IOC_ASYNC msg code */

/* slcIOC utilities */
#include    "slcCmlog.h"                                 /* slcCmlogLogMsg   */
#include    "msgQUtil.h"                                 /* messageQSend     */
#include    "cstrHdlr.h"                                 /* asyncmail_ts     */
#include    "dblist.h"                                   /* dblist proto etc */
#include    "dbl.h"                                      /* dblget proto     */

/* local function prototypes */
LOCAL int           cstrAsyncInit(void);
LOCAL int           cstrAsyncSendFuncMsg(const int cyclIx2Exe,
                                         const char* const funcName_c,  
					 const epicsTimeStamp * const currentTime_ps);
LOCAL void          cstrAsyncExit(void);

static DBLIST(dblistScan_ps, dbhEntry_ts); /* DB list pointers for Supertype II data */
static DBLIST(dbdataScan_ps, slcScan_ts);
int cstrCantSendCount;
/*=============================================================================

  Name: cstrAsyncThread

  Abs:  The CSTR Async thread

  Args: none

  Rem: This thread initializes job global by calling slcAsyncInit.  Then, it
       enters a loop which finds the next cycling function to be run and sends
       a message to a handler to run that function.

  Side: 

  Ret:  none

==============================================================================*/

void  cstrAsyncThread(void)
{

  int ii = 0;                      /* local loop counter                   */
  int iStat = 0;                   /* internal status                      */
  int cyclIx2Exe = 0;              /* index of cycling function to execute */
  unsigned long firstIx = 0;       /* first cycling func index for job     */
  unsigned long numFunc = 0;       /* number of cycling func for job       */
  char     funcName_c[5];          /* null terminated 4-character func name*/
  epicsTimeStamp currentTime_s;    /* current time of day                  */


  /*----------------------------- code ------------------------------------*/

  DEBUGPRINT (DP_DEBUG, cstrFlag, ("Entering cstrAsyncThread !!!\n"));
  
  if (cstrAsyncInit() != 0) {
    DEBUGPRINT (DP_FATAL, cstrFlag, ("cstrAsyncThread:1: cstrAsyncInit failed\n"));
    goto egress;
  }

  /* Tell the exec that we started */
  slcThreads_as[cstrAsync].active = epicsTrue;

  DEBUGPRINT(DP_DEBUG,cstrFlag, ("CSTR_A: cstrAsyncThread set active flag.  Entering main loop.\n"));  

  while (!slcThreads_as[cstrAsync].stop)
  {

    DEBUGPRINT(DP_DEBUG,cstrFlag, ("\n\nCSTR_A: cstrAsyncThread TOP OF LOOP\n"));

    /* Get supertype II now for calls to slcAsyncGetCycleFcn */
    DBCLEAR(dbdataScan_ps);
    if (!SUCCESS(dblget ((const dbheader_ts *const)dblistScan_ps,
                         (dbheader_ts **const)&dbdataScan_ps))) break;

    /* For each job that has cycling functions, find the func most ready to run
     * and send a message to it's handler to run it.
     */

    for (ii=0; ii<MAX_SLC_JOBS; ii++) {

      if (slcJobs_as[ii].numCyclFunc == 0) continue;    /* no cycl funcs for this job*/

      if ((slcThreads_as[slcJobs_as[ii].hdlr_e].tid_ps == NO_THREAD) && /* the job is not running */
          (slcThreads_as[slcJobs_as[ii].async_e].tid_ps == NO_THREAD)) continue;

      firstIx = slcJobs_as[ii].frstCyclFuncIx;
      numFunc = slcJobs_as[ii].numCyclFunc;

      iStat = slcAsyncGetCycleFcn(firstIx, numFunc,
                                  DBPTR_FIRST(dbdataScan_ps),
                                  &cyclIx2Exe, funcName_c, &currentTime_s);
      if (iStat != 0) break;

      if (cyclIx2Exe >= 0)           /* if a ready func was found, send msg to run it*/
        if (cstrAsyncSendFuncMsg(cyclIx2Exe, funcName_c, &currentTime_s) != 0) break;
    }

    epicsThreadSleep(1);      /* Set to 1 sec for production.  Set to 7 for debugging */
  } 

 egress:
  cstrAsyncExit();
  return;

}    /* end cstrAsyncThread */

/*==============================================================================

  Name: cstrAsyncInit()

  Abs:  Initialization for cstrAsync.


  Args: none

  Rem: Call dblistalloc(), dblist(), dblget are called to set up lists for later access to 
       MMSK, CMSK, FMSK, HSTA, and SCAN which has one value per function.  Pointers to lists are in file 
       scope for use by other routines.  Dblget() is called to get initial values 
       which may be overwritten later.

  side: 

  Ret: int  0 = success 
           -1 = Error 

==============================================================================*/
LOCAL int cstrAsyncInit(void)
{

  int             retStat = -1;                  /* return status */  

/*-------------------------------------- code ----------------------------------*/

  DEBUGPRINT(DP_DEBUG, cstrFlag, ("CSTR_A: Entered cstrAsyncInit\n"));

  if (slcAsyncInit() != 0) {
    DEBUGPRINT (DP_FATAL, cstrFlag, ("cstrAsyncInit:1: slcAsyncInit failed\n"));
    goto egress;
  }
  cstrCantSendCount = 0;
  /**********************************************************************/
  /***** Get Supertype II single item slc database database lists. ******/
  /**********************************************************************/

  if (!SUCCESS(DBALLOC(dblistScan_ps, 5))) goto egress;
  if (!SUCCESS(DBALLOC(dbdataScan_ps, 1))) goto egress;
  if (!SUCCESS(dblist ((const char *)"CSTR",
                       (const char *)"1   ",
                       (const char *)"MMSK",
                       (const dbheader_ts **const) &dblistScan_ps))) goto egress;
  if (!SUCCESS(dblist ((const char *)"CSTR",
                       (const char *)"1   ",
                       (const char *)"CMSK",
                       (const dbheader_ts **const) &dblistScan_ps))) goto egress;
  if (!SUCCESS(dblist ((const char *)"CSTR",
                       (const char *)"1   ",
                       (const char *)"FMSK",
                       (const dbheader_ts **const) &dblistScan_ps))) goto egress;
  if (!SUCCESS(dblist ((const char *)"CSTR",
                       (const char *)"1   ",
                       (const char *)"HSTA",
                       (const dbheader_ts **const) &dblistScan_ps))) goto egress;
  if (!SUCCESS(dblist ((const char *)"CSTR",
                       (const char *)"1   ",
                       (const char *)"SCAN",
                       (const dbheader_ts **const) &dblistScan_ps))) goto egress;
  if (!SUCCESS(dblget ((const dbheader_ts *const)dblistScan_ps,
                       (dbheader_ts **const)&dbdataScan_ps))) goto egress;
  retStat = 0;

egress:
  return retStat;
  
}

/*==============================================================================

  Name: cstrAsyncSendFuncMsg()

  Abs:  Find the next cycling function to be run and send a message to a handler 
        to run that function.

  Args: 
       Type:        int
       Name:        cyclIx2Exe
       Access:      R
       Description: Index of cycling function to be run

       Type:        char *
       Name:        funcName_c
       Access:      R
       Description: Name of cycling function to be run

       Type:        epicsTimeStamp
       Name:        currentTime_s
       Access:      R 
       Description: Current time of day


  Rem: Send an IOC_AYNC message by calling messageQSend with the nowait flag set to 
       the appropriate handler to execute the cycling function associated with cycl2Exe.  

  Side:

  Ret: none

  ==============================================================================*/

LOCAL int cstrAsyncSendFuncMsg(const int cyclIx2Exe,
                               const char* const funcName_c,
                               const epicsTimeStamp * currentTime_ps)    
{
  int            retStat = 0;                  /* return status */
  vmsstat_t      iss = ASYNC_OKOK;             /* vms status */
  msgptr_ts      msgptr_s;                     /* message pointer structure */
  msgmail_ts*    msg_ps = NULL;                /* pointer to message buffer */
  asyncmail_ts * asyncmail_ps = NULL;          /* func ix and name in msg to Hdlr */

  /*---------------------------------------------- code -------------------------*/

  DEBUGPRINT(DP_DEBUG, cstrFlag, ("CSTR_A: Entering cstrAsyncSendFuncMsg INDEX = %d  FUNC2EXE = %s \n", 
               cyclIx2Exe, funcName_c));

  /* Sanity Check input parameter */
  if (cyclIx2Exe < 0) {
    DEBUGPRINT(DP_ERROR, cstrFlag, ("CSTR_A: Bad cycle func index passed in = %d \n", cyclIx2Exe));
    retStat = -1;
    goto egress;

  }

  /*  Allocate the standard message structure which is job_id + fwd_dhr + msghdr + async func name and ix. 
   *   The allocated space is enough for the headers and the 4 byte string (funcName) and some padding
   */

  iss = msgQGetBuffer( sizeof(msgmail_ts), (void*)&msg_ps);  /* I added more to length, was wrong?? */

  if (SUCCESS(iss))
  {
    msg_ps->msgheader.func = IOC_ASYNC;
    msgptr_s.msgbytes = MSG_HDRS_BYTES + sizeof(asyncmail_ts); /* standard header and function name */
    msgptr_s.msgptr_p = msg_ps;
    asyncmail_ps = (asyncmail_ts*)msg_ps->reqdata;
    asyncmail_ps->cyclIx = cyclIx2Exe;                         /* transmit cycl func ix and name    */
    strncpy(asyncmail_ps->cyclName, funcName_c, 7);                /* never copy too much... 7 max  */

    iss = messageQSend(slcAsyncGetThreadEnum(cyclIx2Exe), epicsFalse, sizeof(msgptr_s),  &msgptr_s);

    if (!SUCCESS(iss))
    {
	  cstrCantSendCount++;
      msgQFreeBuffer(sizeof(msgmail_ts), (void*)msg_ps);  /* free only if failure sending */ 
	  DEBUGPRINT(DP_ERROR,cstrFlag, ("CSTR_A: cstrAsyncSendFuncMsg sending funcName %s \n", funcName_c));
	  DEBUGPRINT(DP_ERROR,cstrFlag, ("CSTR_A: sending to thread=%d \n", slcAsyncGetThreadEnum(cyclIx2Exe)));
      DEBUGPRINT(DP_ERROR,cstrFlag, ("CSTR_A: messageQSend failed in cstrAsyncSendFuncMsg, count = %d\n", cstrCantSendCount));
    } else cstrCantSendCount = 0;
  }
  else {
    retStat = -1;
    goto egress;
  }

  /* 
   * Update "msg send" statistics in async global.
   * Ignore return status since all it does is set stuff.
   */

  retStat = slcAsyncSetSendStats(cyclIx2Exe, iss, currentTime_ps); 

 egress:

  return retStat;

}

/*==============================================================================

  Name: cstrAsyncExit()

  Abs:  cleanup the globals for this thread before shutdown

  Args: none

  Rem: Clean up job global for this thread and release any other resources.  
       Set executive flags associated with this thread's death.

  side:

  Ret: none

  ==============================================================================*/
LOCAL void cstrAsyncExit(void)    
{

  DEBUGPRINT(DP_DEBUG, cstrFlag, ("CSTR_A: Entering cstrAsyncExit \n"));

  slcAsyncExit();                   /* deallocate job global */
  
  DBFREE(dblistScan_ps);
  DBFREE(dbdataScan_ps);

  DEBUGPRINT(DP_DEBUG, cstrFlag, ("CSTR_A: Setting Task Active flag to false. \n"));

  slcThreads_as[cstrAsync].active = epicsFalse;
  slcThreads_as[cstrAsync].tid_ps = NO_THREAD;
    
  DEBUGPRINT(DP_DEBUG, cstrFlag, ("CSTR_A: cstrAsync Stopped and out of service\n"));

  return;

}


