/*=============================================================================
 
  Name: dbHdlr.c
        dbHdlrThread - main thread entry point
        dbGlobInit   - init DBS globals
        dbCleanup    - cleans up the globals for this thread before shutdown

  Abs:  Waits at message queue to receive inter-thread messages.
        Sends "database service" (DBS) type messages to DBEX port 
        on proxy server for routing to VMS Control System's DBEX process.

        Reference web pages on LCLS Controls | Facilities |SLC-Aware IOC | :
                        Database Service (DBS) Design Document 
                        Functional Requirements | SLC Database Service (DBS)
                        Database Input/Output Utilities
                        C Coding Standard
        Reference RMX_DBS dbmain.c 

  Auth: 4 AUG 2004, drogind created using msgHdlr.c as a template
  Rev: 15 DEC-2004, drogind modified prototype code to production standards
 
-----------------------------------------------------------------------------*/
#include "copyright_SLAC.h"    
/*----------------------------------------------------------------------------- 
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
/* c includes */
#include     <stdio.h>
#include     <stdlib.h>     /* free */
#include     <string.h>     /* memcpy */
#include     <epicsEvent.h> /* epicsEvent */

/* epics libcom osi includes */
#include     <epicsEvent.h>

/* slc global vars */
#include     "slcGlob.h"

/* VMS Control system function/status codes  */
#include     "msgfunc.hc"  /* FUNC_MSG function code */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */

/* slc common includes */
#include     "tcpUtil.h"   /* tcpGetConnection, tcpGeBuffer */
#include     "msgQUtil.h"  /* message queue utilities */
#include     "cvtUtil.h"   /* slc conversion utilities */
#include     "timeUtil.h"  /* slc time conversion utilities */
#include     "slcCmlog.h"  /* slcCmlogLogMsg */
#include     "slcUtil.h"   /* slcGetIOCName */ 
#include     "slcMsg.h"    /* DB_ACK_DBEX... */

/* dbs includes  */  
#include     "db_micro_struc.h"
#include     "dbUtil.h" 
#include     "dbmsgmail.h" 
#include     "dbGlob.h"    /* dbmaincom_p */
#include     "dbdef.hc"
#include     "dbhash.h"
#include     "dblistUtil.h"
#include     "dbST0.h"
#include     "dbdict.h"
/****************************************************************************/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
int dbHdlrFlag = DP_INFO;
#endif
/*****************************************************************************/
/* global vars to db threads only (declared extern in dbGlob.h)              */ 

unsigned long dblongmicro = 0x00000000;   /* 0x584c3031 */
const char v018_c4[4] = "V018";          /* DBEX's message service id.         */

unsigned char  dbsendReceipts = 0;   /* rolling receipt count */
epicsBoolean   dbsendAllUpdates = epicsFalse;
dbsuptype_tu  *dbnodes_a[4];    /* Public structure of pointers to memory allocated */
                                /* to hold supertype blocks 0-3 for this micro.         */

/* watchdog and diagnostic data */
dbdiag_ts    dbdiag_s;  
/* dictionary containing all translated ST0 data */
void * dbdict_p = NULL;
int dbdictSize=DBDICTSIZE;
/* current database version according to  DBEX */
char dbVersion_a[8] = "UNINITED";
/* whether or not DBEX process is known to be up */
epicsBoolean dbex_up = epicsFalse;
/* dbRWMutex */
epicsMutexId dbRWMutex_ps; 
/* free lists */
void * dblistSmFreeList_p= NULL;
void * dblistBigFreeList_p = NULL;
void * dbdictNameTrpFreeList_p = NULL;    /* pointer to dictNameTriplet_ts free list */
void * dbdictNameSglFreeList_p = NULL;    /* pointer to dictNameSgl_ts free list */
void * dbdictPUSFreeList_p = NULL;        /* pointer to dictPUS_ts free list */
void * dbdictPFreeList_p = NULL;          /* pointer to dictP_ts free list */
void * dbdictUFreeList_p = NULL;          /* pointer to dictU_ts free list */
void * dbdictSFreeList_p = NULL;          /* pointer to dictS_ts free list */
/*****************************************************************************/
/* local function prototypes */
LOCAL vmsstat_t     dbGlobInit(void);
LOCAL void          dbCleanup(void);
LOCAL vmsstat_t     dbConvert (void);
/*=============================================================================
 
  Name: dbHdlrThread

  Abs:  This thread accepts messages from dbRecv and slcExec and acts upon them.
  
  Args: none

  Rem:  started with ThreadCreate by the slc Exec thread in slcMain  
        Calls utilities in dbST0 to handle ST0 conversion.

  Side: Manages DBS global variables
 
  Ret:  void
  
=============================================================================*/

epicsShareFunc void epicsShareAPI dbHdlrThread(void)
{
  dbmsgmail_ts         msgRcvDb;            /* message buffer */
  vmsstat_t            iss = DB_OKOK;       /* default ok return */

  /* create the message queue id */
  if ( !SUCCESS(iss = messageQCreate(dbHdlr, 20, MAX_MSG_BYTES)) ) {

	dbCleanup();
	return;
  }
  
  slcThreads_as[dbHdlr].active = epicsTrue;
  DEBUGPRINT(DP_INFO,dbHdlrFlag,("DB HDLR: becoming active\n"));
  
  /* Initialize db thread Globals */
  if (!SUCCESS(iss = dbGlobInit())){
	dbCleanup();
  }

  /* main loop of the db hdlr thread, wait at message queue for incoming 
	 messages & handle them.
  */
  while(slcThreads_as[dbHdlr].active)
  {
    /* read a message at the message queue.    */
    memset ( &msgRcvDb, 0, sizeof(msgRcvDb) );
    if (SUCCESS(iss = messageQRecv(dbHdlr, sizeof(msgRcvDb), NULL, &msgRcvDb ) ))
    {
      /* now handle the message */
      switch(msgRcvDb.func)
		{
		case IOC_STOP:
		  slcThreads_as[dbHdlr].active = epicsFalse;
		  break;
		case DB_CONVERT:
		  dbConvert();
		  break;
		} /* switch */
    } /* if */

    /* check the stop flag just in case */
    if (slcThreads_as[dbHdlr].stop)
      slcThreads_as[dbHdlr].active = epicsFalse;

  } /* while active */

  /* active not set, time to shut down */
  DEBUGPRINT(DP_INFO,dbHdlrFlag,("HDLR: shutdown\n"));
  dbCleanup();
}
/*=============================================================================
 
  Name: dbGlobInit

  Abs:  Initializes DBS - scoped global variables
  
  Args: Type		          Name	      Access	   Description
        ------------------- -----------	---------- ----------------------------
		none

  Rem:  allocates resources 

  Side: Manages DBS global variables
        Initialized globals declared in dbGlob.h
          dbnodes_a
          dblongmicro
          dbdiag_s
          dbdict_p
          dblistSmFreeList_p  (via dblistInitFreeLists())
          dblistBigFreeList_p (via dblistInitFreeLists())

  Ret:  vmsstat_t iss
  
=============================================================================*/

LOCAL vmsstat_t dbGlobInit (void)
{
  vmsstat_t      iss =  iss = MICR_OKOK;
  char *micro = slcGetIOCName();

  memset(&dbnodes_a, 0, sizeof(dbnodes_a));

  /* init the name of this ioc */
  memcpy (&dblongmicro, micro,4);
  DEBUGPRINT(DP_INFO,dbHdlrFlag, ("DB RECV: This is microname =  %s\n", (char *)&dblongmicro));

  memset(&dbdiag_s, 0, sizeof(dbdiag_s));
  dbdiag_s.msgc_snd = JUST_BOOTED_BIT;

  /* init gphash table dictionary pointer */
  dbhInitPvt (&dbdict_p,dbdictSize); 
  if (dbdict_p == NULL) {
    DEBUGPRINT(DP_INFO,dbHdlrFlag, ("DBHDLR: dbdict_p is NULL\n"));
	return (iss=DB_BADCONTEXT);
  }

  /* init free lists for all structures used in dictionary */ 
  dbdictInitFreeLists();

  /* init dict Free Lists Memory Managers  */
  iss = dblistInitFreeLists();  

  /* init read/write mutex around ST data */
  dbRWMutex_ps = epicsMutexCreate();
   
 return iss; 
}
/*=============================================================================
 
  Name: dbCleanup

  Abs:  This thread accepts messages from dbRecv and slcExec and acts upon them.
  
  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
	none

  Rem:  clean up the slcThreads_as struct for this thread
         and release global DBS resources 

  Side: Cleans up DBS global variables:
          dbExists
          dbDownloadEvent_ps
          dbnodes_a
          dbdict_p
          dblistSmFreeList_p  (via dblistDelFreeLists())
          dblistBigFreeList_p (via dblistDelFreeLists())
        Note: the freeListCleanup routine does not reset pointers to NULL!!
        Ensure all wrappers that call this set the pointer to NULL.
          
  Ret:  void
  
=============================================================================*/

LOCAL void dbCleanup(void)
{
  unsigned short supblkx = 0;
  void *   temp_p;

  DEBUGPRINT(DP_INFO,dbHdlrFlag,("DB HDLR: cleanup..\n"));
  DEBUGPRINT(DP_INFO,dbHdlrFlag,("DB HDLR: dbdictPUSFreeList_p = %p\n",dbdictPUSFreeList_p));  
  dbExists = epicsFalse;

  if (dbRWMutex_ps !=NULL){
    epicsMutexMustLock(dbRWMutex_ps);
  }
  /* free the superblock storage in dbnodes_a */
  for (supblkx=0; supblkx <MAX_STX; supblkx++)
    {
      if (dbnodes_a[supblkx]) {
		free(dbnodes_a[supblkx]);
		dbnodes_a[supblkx]=NULL;
	  }
	  else DEBUGPRINT(DP_ERROR,dbHdlrFlag,("DB HDLR: failure to release dbnodes\n"));
    }
  /* free the dictionary free lists */
  DEBUGPRINT(DP_INFO,dbHdlrFlag,("DB HDLR: releasing dictionary free lists \n"));
  dbdictDelFreeLists();
  /* free gphash table dictionary  */
  DEBUGPRINT(DP_INFO,dbHdlrFlag,("DB HDLR: releasing hashtable free lists \n"));
  if (dbdict_p) {
    temp_p = dbdict_p;
    dbdict_p = NULL;
    dbhFreeMem (temp_p);
  }
 
  /* delete list Free Lists  */
  dblistDelFreeLists();
 
  if (dbRWMutex_ps){
    epicsMutexUnlock(dbRWMutex_ps);
  } 
 
  /* unblock other threads, if they are waiting */
  if (!dbExists ) {
    if (dbDownloadEvent_ps != NULL){
        epicsEventSignal(dbDownloadEvent_ps);
    }
  }
  if (dbRWMutex_ps){
    temp_p = dbRWMutex_ps;
    dbRWMutex_ps = NULL;
    epicsMutexDestroy(temp_p);
  }

  if (slcThreads_as[dbHdlr].mqId_ps!=NULL){
	epicsMessageQueueDestroy(slcThreads_as[dbHdlr].mqId_ps);
  }
  slcThreads_as[dbHdlr].mqId_ps = NULL;
  slcThreads_as[dbHdlr].tid_ps = NO_THREAD;

  DEBUGPRINT(DP_INFO,dbHdlrFlag,("DB HDLR: shutdown\n"));
  /* LOG DBS shutdown */
  slcCmlogLogMsg(DB_NOSERVICE, "NOSERVICE", "%.4s Database Service shutdown.\n", 
				 (char *)&dblongmicro);
}

/*=============================================================================
 
  Name: dbConvert

  Abs:  Simply calls the main entry point in dbST0.c - dbST0Convert to do all
        the work of converting ST0 data to the dictionary.
  
  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
	none

  Rem:  If dbST0Convert is successful, then routine:
           deletes ST0
           sets the global dbExists flag true
           signals dbDownloadEvent_ps 

  Side: 
 
  Ret:  vmsstat_t iss
  
=============================================================================*/

LOCAL vmsstat_t dbConvert (void)
{
  vmsstat_t iss = MICR_OKOK;

  DEBUGPRINT(DP_INFO,dbHdlrFlag, ("dbHdlr: dbConvert contacted \n")); 
  if (SUCCESS(iss = dbST0Convert())) {
	DEBUGPRINT(DP_INFO,dbHdlrFlag,("DB Hdlr: Setting dbExists flag! \n"));	
    dbExists = epicsTrue;
	/* LOG DBS successful startup */
	slcCmlogLogMsg(iss, "OKOK", "%.4s Database Service startup.\n", 
				   (char *)&dblongmicro);
    if (dbDownloadEvent_ps) {
      epicsEventSignal(dbDownloadEvent_ps);
    }
  }
  
  return iss; 
}


