/*=============================================================================
 
  Name: dbRecv.c
        dbRecvThread - main thread entry point
        dbRecvInit   - inits file scoped variables
        dbCleanup    - cleans up file scoped vars and resources
        dbDbexSockConnect - establish/re-establish DBEX & CTL socket connections
        dbRead       - Read at the DBEX socket
        dbRecvProcess- Process supertype (ST) data received from DBEX  
        dbMsgCopy    - copy ST data to allocated memory
        dbSTalloc    - allocate memory for entire ST block being downloaded
        dbSendDiagRpy- package up diagnostic message data & send to dbSend
        dbSendSTAck  - send message to dbSend to ack DBEX
        dbRecvDXUPDN  - stores dbVersion_a upon "DXUP" message; sets dbex_up flag
	               accordingly, to True for "DXUP"; false for "DXDN"
        dbSendConvert - request dictionary conversion of dbHdlr

  Abs:  Receives all incoming "database service" (DBS) type messages
        from VMS Control System's DBEX process via the proxy server.
        (dbSend thread sends all DBS messages to DBEX via same proxy
         DBEX socket)

        Reference web pages on LCLS Controls | Facilities |SLC-Aware IOC | :
                        Database Service (DBS) Design Document 
                        Functional Requirements | SLC Database Service (DBS)
                        Database Input/Output Utilities
                        C Coding Standard
        Reference RMX_DBS dbmain.c 

  Auth: 4 AUG 2004, drogind created using msgRecv.c as a template
  Rev:  8 DEC-2004, drogind modified prototype code to production standards
 
-----------------------------------------------------------------------------*/
#include "copyright_SLAC.h"    
/*----------------------------------------------------------------------------- 
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int dbRecvFlag = DP_ERROR;
#endif
/* c includes */
#include     <stdio.h>
#include     <string.h>    /* memcpy */
#include     <stdlib.h>    /* malloc */

/* slc global vars */
#include     "slcGlob.h"

/* VMS Control system function/status codes  */
#include     "msgfunc.hc"  /* FUNC_MSG function code */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */

/* slc common includes */
#include     "msgTypes.h"
#include     "tcpUtil.h"   /* tcpGetConnection, tcpGeBuffer */
#include     "msgQUtil.h"  /* message queue utilities */
#include     "cvtUtil.h"   /* slc conversion utilities */
#include     "timeUtil.h"  /* slc time conversion utilities */
#include     "slcCmlog.h"  /* slcCmlogLogMsg */
#include     "slcUtil.h"   /* slcGetIOCName */ 
#include     "slcMsg.h"    /* DB_ACK_DBEX... */

/* dbs includes  */  
#include     "db_micro_struc.h"
#include     "dbUtil.h" 
#include     "dbmsgmail.h"
#include     "dbGlob.h"
#include     "dbdef.hc"
#include     "dbRecv.h"

/* epics osi includes */
#include     "cantProceed.h"

/* Local structure & defines for dbmicromail's current state. */
                                       /* Definitions of dbmicromail "states":*/
#define dbmmailst_nothingyet 0         /* Bootup: expecting supertype 0.      */
#define dbmmailst_iplzero    1         /* Bootup: receiving supertype 0.      */
#define dbmmailst_iplxone    2         /* Bootup: expecting supertype 1.      */
#define dbmmailst_iplone     3         /* Bootup: receiving supertype 1.      */
#define dbmmailst_iplxtwo    4         /* Bootup: expecting supertype 2 or 3. */
#define dbmmailst_ipltwo     5         /* Bootup: receiving supertype 2.      */
#define dbmmailst_iplxthree  6         /* Bootup: expecting supertype 3.      */
#define dbmmailst_iplthree   7         /* Bootup: receiving supertype 3.      */
#define dbmmailst_ipled      8         /* Database download complete.         */
#define dbmmailst_iplfail    9         /* Supertype sequence error or memory  */
                                       /*  allocation failed or (first)       */
                                       /*  dbget/dbput/micro_dbsend failed.   */
/*****************************************************************************/
/* local file scope data declarations and function prototypes */

typedef struct
{   
  unsigned long  sup_expwc[4],
                 sup_gotwc[4],
                 msgc,
                 errc,
                 ipliss;
  unsigned char  firstf,
                 tmoutf;
} dbmicromail_state_ts;

static dbsndrcvbuf_ts  *dbrcvbuf_ps;       /* receive buffer */
static dbmicromail_state_ts dbmstate_s;
static const unsigned char newfirstf[4] =
  {dbmmailst_nothingyet,dbmmailst_iplone,dbmmailst_ipltwo,dbmmailst_iplthree};
static dbmsgmail_ts    dbRecvmsg_s;       /* message queue buffer for sending out msgs */

/* global diagnostics for dbRecv thread */
dbRecvDiags_ts   dbRecvDiags_s = {0, 0, 0, {0,0}, 0, {'U','N','I','N','I','T','E','D'}, 0, 0, 0 }; 

LOCAL void      dbRecvInit (void);
LOCAL void      dbCleanup(void);
LOCAL vmsstat_t dbDbexSockConnect(void);
LOCAL vmsstat_t dbRead (const unsigned short * const max_bytecount_p,
                        dbsndrcvbuf_ts       * inmsg_ps,
                        int                  * actual_bytecount_p,
                        int                  *sbbytecnt);
LOCAL vmsstat_t dbRecvProcess(const dbsndrcvbuf_ts * const mail2_ps, 
			      const unsigned long actrcvbc, 
			      const int sbbytecnt);
LOCAL void      dbMsgCopy(const dbsndrcvbuf_ts * const mail2_ps, 
			  const unsigned short supblkx, 
			  const unsigned long nwords);
LOCAL vmsstat_t dbSTalloc(const dbsndrcvbuf_ts * const mail2_ps, 
			  const unsigned short supblkx);
LOCAL void      dbSendDiagRpy(const dbsndrcvdiagmsg_ts * const diagmsgbuf_ps, 
			      const unsigned short diagmsgbc);
LOCAL vmsstat_t dbSendSTAck (const unsigned short id);
LOCAL void dbRecvDXUPDN(const dbsndrcvbuf_ts *const dxmsgbuf_ps, 
			const unsigned short msgbc);
LOCAL vmsstat_t dbSendConvert (void); 
/*=============================================================================
 
  Name: dbRecvThread

  Abs:  This thread receives all incoming "database service" (DBS) type messages
        from VMS Control System's DBEX process via the proxy server. 

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
        none
 
  Rem:  This thread does not have a message queue. It waits forever at the DBEX
        socket after init. 
	This thread sets the slcThreads_as[slcDbRecv].active flag when it starts, 
        indicating to slcExec that the thread is alive. 
	The Exec thread (or the slcStop command) sets the sclThreads[slcDbRecv].stop 
        flag to false when it wants to shut down the Recv thread.

  Side: created by the slcExecThread  
 
  Ret:  none

All possible vmsstat_t error messages that originate from dbRecvThread:
 MICR_DBWRNGMICR,
 MICR_DBNOMEM,
 MICR_CANTREIPL,
 MICR_BADSUPTYP,
 MICR_IPLDBDWNSEQ,
 MICR_DBNEVRDWNLD,
 MICR_DBBADMESSG,
 DB_UNRCGMSG,   
=============================================================================*/

epicsShareFunc void epicsShareAPI dbRecvThread(void)
{
  int                  rv;                 /* return variables */
  vmsstat_t            iss = MICR_OKOK;    /* default ok return */
  unsigned short       bufmaxbc;           /* max # bytes we could receive */
  int                  actbc;
  int                  sbbytecnt;

  DEBUGPRINT(DP_DEBUG,dbRecvFlag,("DB RECV:starting\n"));
  
  dbRecvInit();

  slcThreads_as[dbRecv].active = epicsTrue;

  /* GET A db connection now (after setting active ) */
  dbDbexSockConnect();

  /* main loop of the db Recv Task, make sure there's a connection to the server,
	 then listen for messages, send on to correct DBS thread.
  */
  while(slcThreads_as[dbRecv].active)
  {
    /* in case there was an error in the last pass,
     * get a connection with the server*/
    iss = dbDbexSockConnect();

    /* now begin, until connection fails, or we're told to quit */
    while (SUCCESS(iss) && (!slcThreads_as[dbRecv].stop))
    {
      /* read message sent  by server.    */
      DEBUGPRINT(DP_DEBUG,dbRecvFlag,("DB RECV: Waiting at socket - blocking read\n"));
      bufmaxbc = sizeof (dbrcvbuf_ps->supblk); /* NETBUFLEN*2 bytes */
      
      iss = dbRead (&bufmaxbc,
		    dbrcvbuf_ps,
		    &actbc,
		    &sbbytecnt);

      if ( (SUCCESS(iss)) && (actbc>0) )
      { 
		/* if ack req'ed by DBEX, send it now */
		if (dbrcvbuf_ps->nativeSB.id & SUPERTYPE_ID_ACKREQ){
		  DEBUGPRINT(DP_INFO,dbRecvFlag, ("DB RECV: Ack requested on id = 0x%x \n", dbrcvbuf_ps->nativeSB.id ));
		  iss = dbSendSTAck (dbrcvbuf_ps->nativeSB.id);
		  dbRecvDiags_s.dbexUpdCnt++;
		}
		/* dbrcvbuf_ps->supblk is converted to native byte order via _ReadDbs */
        if (!(dbrcvbuf_ps->nativeSB.id & SUPERTYPE_ID_OVERRIDE)) {
		  DEBUGPRINT(DP_INFO,dbRecvFlag,("DB RECV: Rec'ed download at dbex port - calling dbRecvProcess\n"));
          DEBUGPRINT(DP_INFO,dbRecvFlag,("DB RECV: Actual bytecount = %d\n", actbc));
		  dbRecvProcess(dbrcvbuf_ps, actbc, sbbytecnt);
		  ++dbRecvDiags_s.dbexUpdCnt;
		}
		/* -2 on byte count comparison for good measure; takes care of padded architecture: */
        
		else if ((dbrcvbuf_ps->nativeSB.id & SUPERTYPE_ID_WMASK) == SUPERHDR_FUNC_DGINQ &&
                  dbrcvbuf_ps->nativeSB.micr == dblongmicro &&
				 actbc >= sizeof(dbsndrcvdiagmsg_ts) - sizeof(fwd_hdr_ts)-2 ) {
		  dbSendDiagRpy((dbsndrcvdiagmsg_ts *) &dbrcvbuf_ps, actbc);
        }
        else if (dbrcvbuf_ps->nativeSB.id & (SUPERTYPE_ID_OVERRIDE | SUPERHDR_FUNC_NOTFY)){
		  dbRecvDXUPDN(dbrcvbuf_ps, actbc);
        }
        else {
		  iss = DB_UNRCGMSG;
		  slcCmlogLogMsg(iss, "UNRCGMSG", "Unrecognizable message from DBEX with len=%lu; micr=%ld.\n", 
                           dbrcvbuf_ps->nativeSB.len,
                           dbrcvbuf_ps->nativeSB.micr);
		  DEBUGPRINT(DP_ERROR,dbRecvFlag,("DB RECV: db download not validated\n"));
		  DEBUGPRINT(DP_ERROR,dbRecvFlag,("/t**** dbrcvbuf.nativeSB.id = 0x%x\n", dbrcvbuf_ps->nativeSB.id)); 
		  DEBUGPRINT(DP_ERROR,dbRecvFlag,("/t**** dbrcvbuf.nativeSB.micr = 0x%lx\n", dbrcvbuf_ps->nativeSB.micr)); 
		  DEBUGPRINT(DP_ERROR,dbRecvFlag,("/t**** actbc = 0x%x\n", actbc)); 
          rv =  sizeof(dbsndrcvdiagmsg_ts) - sizeof(fwd_hdr_ts);
		  DEBUGPRINT(DP_ERROR,dbRecvFlag,("/t**** message byte cnt = 0x%x\n", rv)); 
        } 
      }
      else /* tcpGetBuffer error */
		{
		  if (slcThreads_as[dbRecv].stop)
			{
			  /* if slcExec has told this thread to stop, then the error is because the 
				 slcExec thread has probably already destroyed the socket */
			  slcThreads_as[dbRecv].active = epicsFalse;
			  iss = MICR_EXNET_RECVRESET;
			  slcCmlogLogMsg(iss,"EXNET_RECVRESET", 
							 "dbRecvThread detected disconnect and stop; time to shutdown.\n");		
			  
			}
		}       
	  
    } /* while (SUCCESS(iss) && slcThreads_as[slcDbRecv].stop);*/
	
    /* if we got here there might have been some error and
	   the socket is invalid, close it */

    if (!SUCCESS(iss) ) 
	  {   
		tcpSockClose(&slcSockets_as[sdDbex]);	
		dbRecvDiags_s.dbexSockConnStatus= slcSockets_as[sdDbex].sd;	
	  }
	
  } /* while slcThreads_as[dbRecv].active);*/
  
  DEBUGPRINT(DP_DEBUG,dbRecvFlag,("DB RECV:shutdown\n"));
  /* make sure all our globals are cleaned up */
  dbCleanup();

}
/*=============================================================================
 
  Name: dbRecvInit

  Abs:  Initializes dbRecv's file scoped variables and DBS global variables
  
  Args: None
 
  Rem:  Init slcThreads_as.active global var
        Initializes file-scoped static global vars

  Side: Malloc's space for dbrcvbuf_ps and clears its memory
        Init's slcThreads_as global var
        clears memory for dbmstate_s, then inits
 
  Ret:  vmsstat_t iss
  
=============================================================================*/
LOCAL void dbRecvInit (void)
{
  /* init slcThreads_as global and file scope vars */
  slcThreads_as[dbRecv].active = epicsTrue;

  /* Allocate receive buffer, */
  dbrcvbuf_ps = callocMustSucceed(1, sizeof(dbsndrcvbuf_ts), "dbRecv cannot calloc dbrcvbuf\n");

  memset(&dbmstate_s, 0, sizeof(dbmstate_s));
  dbmstate_s.firstf = dbmmailst_nothingyet;
  dbmstate_s.ipliss = MICR_CANTIPL;

}
/*===========================================================

  Name: dbDbexSockConnect

  Abs:  get a socket connection for the dbRecvThread. This
        routine will wait forever for a good connection, or until
		the msgRecv thread is told to stop.

  Args: none

  Rem:  returns if this thread is told to stop
        
  Side: 

  Ret:  iss - the status code
===========================================================*/
LOCAL vmsstat_t dbDbexSockConnect(void)
{
  vmsstat_t      iss;
  unsigned short proxy = PROXIES_HOST_MSG_PORT;   /* Proxy port number */
  unsigned short hostdbex =  MICROS_HOST_DBEX_PORT ;     /* our 'porttag'*/
  unsigned long elapsed_count;

  /* for debugging proxy server code- the following will break proxy: */
  /* 
  unsigned short hostmsg =  MICROS_HOST_MSG_PORT ;
  DEBUGPRINT(DP_INFO,dbRecvFlag,("DB RECV: get buggy dbex socket connection..\n")); 
  iss =  tcpSockConnect(dbRecv, proxy, hostmsg, 5.0, &slcSockets_as[sdDbex] ); 
  DEBUGPRINT(DP_INFO,dbRecvFlag,("DB RECV: get buggy  ctl socket connection..\n")); 
  iss =  tcpSockConnect(dbRecv, proxy, hostmsg, 5.0, &slcSockets_as[sdCtl] ); 
  */

  /* the following is the correct set of lines */
  DEBUGPRINT(DP_DEBUG,dbRecvFlag,("DB RECV: get dbex socket connection..\n")); 
  if (SUCCESS(iss=tcpSockConnect(dbRecv, proxy, hostdbex, 5.0, &elapsed_count, &slcSockets_as[sdDbex] ))) {
	++dbRecvDiags_s.dbexSockConnCnt;
	dbRecvDiags_s.dbexSockConnStatus= slcSockets_as[sdDbex].sd;;
  }

  return (iss);
}

/*===========================================================

  Name: dbCleanup()

  Abs:  cleans up the globals for this thread before shutdown

  Args: NONE

  Rem: clean up the slcThreads_as struct for this thread
       and release any resources 

  Side: Modifies slcThreads_as[dbRecv]; 
        frees malloc'ed ST memory pointed to by dbnodes.
        frees malloc'ed memory for dbrcvbuf_ps

  Ret:  NONE

===========================================================*/
LOCAL void dbCleanup(void)
{
  dbsndrcvbuf_ts * temp_ps;

  DEBUGPRINT(DP_DEBUG,dbRecvFlag,("DB RECV: cleanup..\n"));
  /* clean up thread struct */
  slcThreads_as[dbRecv].mqId_ps = NULL;
  slcThreads_as[dbRecv].tid_ps = NO_THREAD;
  slcThreads_as[dbRecv].active = epicsFalse;

  /* free receive buffer */
  if (dbrcvbuf_ps){
	temp_ps = dbrcvbuf_ps;
	dbrcvbuf_ps = NULL;
    free(temp_ps);
  }
  DEBUGPRINT(DP_DEBUG,dbRecvFlag,("DB RECV: shutdown\n"));
}
/*===========================================================

  Name: dbSendSTAck()

  Abs:  send to dbSend a message to Ack DBEX supertype data that was sent

  Args: NONE

  Rem:  Send message to dbSend to actually send out the ack for ST2 update  

  Side: Modifies file-scope variable dbRecvmsg_s
          with .nativSB.id and .func

  Ret:  vmsstat_t: MICR_OKOK or MICR_EXNET_BADACK

===========================================================*/
/**procedure**/
LOCAL vmsstat_t dbSendSTAck (const unsigned short id)
{
  int iss = MICR_EXNET_BADACK;

  /* func code ="DB_ACK_DBEX"; send native dbsuphdr.id to dbSend task to ack */
  memset(&dbRecvmsg_s, 0, sizeof(dbRecvmsg_s));
  dbRecvmsg_s.nativeSB.id = id;

  dbRecvmsg_s.func = DB_ACK_DBEX;
  /* send the message */
  if (!SUCCESS(iss= messageQSend(dbSend, epicsTrue, sizeof(dbRecvmsg_s),(void *)&dbRecvmsg_s)) )
  {
    DEBUGPRINT(DP_ERROR,dbRecvFlag,("SLCMAIN:DB download message failed\n"));
  }
  else 
  { 
    iss = MICR_OKOK;
    DEBUGPRINT(DP_DEBUG,dbRecvFlag,("DB RECV: sent ack to dbSend\n")); 
  }
  return iss;
}
/*=============================================================================
 
  Name: dbRead

  Abs:  Read a message from the DBS_tcp_socket
  
  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        const unsigned short*  max_bytecount_p,   R       max bytes to read not counting tcp hdr
        dbsndrcvbuf_ts*        inmsg_ps,          RW      points to receive buffer alloc'ed
        int*                   actual_bytecount_p W       actual bytes read not counting tcp hdr
        int*                   sbbytecnt          W       superblock hdr (unpadded) byte count

  Rem:  Waits forever at DBEX socket to receive a message from DBEX via proxy.
        Upon receiving data, it converts the supertype header to native in order
        to interpret the id field. If the id field indicates that an Ack was req'ed
        by DBEX, it calls dbSTAck routine to send to dbSend.


  Side: Stores data in global *dbrcvbuf_ps (passed in by inmsg_ps * arg) 
        dbVMStoHSupHdr() fixes endianness and unpacks the dbsuptypehdr portion of the data 
          into a new member of dbsndrcvbuf_ts, nativeSB, once. 
          It cannot be unpacked on top of itself (in place)  because it may occupy 1 more 
          word due to padding on some architectures. Thus, all code should refer to field 
          nativeSB (as opposed to supblk->sbhdr) for the native super block header.

  Ret:  vmsstat_t iss error condition code
  
=============================================================================*/
vmsstat_t dbRead (const unsigned short *max_bytecount_p,
                  dbsndrcvbuf_ts       *inmsg_ps,
                  int                  *actual_bytecount_p,
                  int                  *sbbytecnt)
{
   vmsstat_t      iss = MICR_EXNET_SENDFAIL;

   DEBUGPRINT(DP_DEBUG,dbRecvFlag,("DB RECV: dbRead waiting at socket\n"));

   iss = tcpGetBuffer(&slcSockets_as[sdDbex], 
		      *max_bytecount_p, 			      
		      actual_bytecount_p,
		      &inmsg_ps->pxyhdr,                      
		      &inmsg_ps->supblk);


   if (SUCCESS(iss)) 
   { 
     /*dbprintVMSSupHdr ("This is the VMS dbsuptypehdr_ts: ",
		      (dbsuptype_tu *) 
		      &inmsg_ps->supblk, 0);*/ 

     /* fwd_hdr_ts is the only msg in network order (little endian)  */
     /* Here fix endianness, RISC padding, etc for superblk  - */
     /* transform superblock header from VMS to native here, once */
     *sbbytecnt = dbVMStohSupHdr(&inmsg_ps->supblk, &inmsg_ps->nativeSB);
  
     dbprinthSupHdr ("Here is the native supertype hdr received : ",
       &inmsg_ps->nativeSB, *sbbytecnt);

   }
   return iss;
}   
/*=============================================================================
 
  Name: dbRecvProcess

  Abs:  Validity check contents of received message and store memory locally.
  
  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        dbsndrcvbuf_ts*        inmsg_ps,          R       points to receive buffer alloc'ed
		const unsigned long    actrcvbc,          R       actual bytes received
        const int              sbbytecnt          W       superblock hdr (unpadded) byte count

  Rem: Routine to take network transfers of database information sent by DBEX and 
         write them into micro's local copy of its database.  At micro bootup time 
         (according to flag byte firstf) we receive micro's database download.  
         Download must begin with supertype 0 and end with supertype 3, and supertypes must 
         be sent in increasing order, though we allow supertype 2 to be entirely absent.  
         We do our best with supertype 2 updates (from SCP's other than the SCP doing the IPL
         that arrive before or during the download.  When all of the supertype 3 appears to 
         have been downloaded, we assume IPL is complete, and we send a DB_CONVERT message
         to dbHdlr to convert the ST0 data to the dictionary.  Thenceforth we accept only 
         supertype 2. data.   

         Like each complete superblock (supertype block) for any one micro, 
         any message received here begins with a database superblock header in which:                     
            len    = size in words of entire superblock of which this   
                      message is a portion (size here includes          
                      superblock header).                               
            segbeg = 1-biased subscript into superblock (considered in  
                      its entirety as a word array including superblock 
                      header) telling where this data is to go (copy    
                      operation may or may not overwrite superblock     
                      header).                                          
            blkbeg = 1-biased subscript into this message (considered   
                      as a word array including superblock header)      
                      telling where copy is to begin from (copy         
                      operation may or may not include superblock       
                      header).
         Note: Algorithm fro dbmain.c, and left intact.
               Validity checking changed due to padding in our environment.                                          
  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        dbsndrcvbuf_ts*        mail2_ps,           R      points to receive buffer alloc'ed
        const unsigned long    actrcvbc           R      actual bytes rcv'ed not counting tcp hdr
        int*                   sbbytecnt          R      superblock hdr (unpadded) byte count                  
  Side: Allocates space for superblock and populates coorresponding dbnodes var
        Accesses and updates dbmstate_s - structure containing dbRecvProcess's current state

  Ret:  vmsstat_t iss error condition code
  
=============================================================================*/
LOCAL vmsstat_t dbRecvProcess(const dbsndrcvbuf_ts * const mail2_ps, 
			      const unsigned long actrcvbc, 
			      const int sbbytecnt)

{   
  unsigned         supblkx;     /* supertype number from msg*/
  unsigned long    nwords;
  unsigned long    parmdw;
  /* debug */
  unsigned long    wordcnt=0;
  vmsstat_t iss = MICR_OKOK;

/*----------------------------------------------*/
/* Validate incoming message:  destination micro must be us;  supertype number  */
/* must be 0, 1, 2, or 3 (more conditions checked below);  incoming message     */
/* must contain at least one word of data to be copied besides its superblock   */
/* header;  copy must begin either from exact beginning of or immediately after */
/* superblock header in message (determined by "blkbeg");  size of entire       */
/* superblock (specified in every incoming message) must agree with previously  */
/* seen value for given superblock;  copy destination area (determined by       */
/* "segbeg" and size of message) must not run outside boundaries of superblock. */
/* Copy operation must include superblock header (both in message and in        */
/* micro's superblock) at and only at beginning of download of given superblock.*/

  epicsTimeGetCurrent(&dbRecvDiags_s.dbexLastUpdTime);  /* update time of last dbex udpate */
  DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("dbm state = %x\n",dbmstate_s.firstf));
  ++dbmstate_s.msgc;
  wordcnt = sbbytecnt/2;
  if (mail2_ps->nativeSB.micr != dblongmicro)
  {
	DEBUGPRINT(DP_ERROR,dbRecvFlag, ("Database update for %.4s sent to wrong micro, %.4s.\n", 
				   (char *)mail2_ps->nativeSB.micr,  (char *)&dblongmicro));
    ++dbmstate_s.errc;
	iss = MICR_DBWRNGMICR;
	slcCmlogLogMsg(iss, "DBWRNGMICR", "Database update for %.4s sent to wrong micro, %.4s.\n", 
				   (char *)mail2_ps->nativeSB.micr,  (char *)&dblongmicro);
	++dbRecvDiags_s.dbexUpdDropCnt;
    return iss;
  }
  if ((supblkx = mail2_ps->nativeSB.id & SUPERTYPE_ID_WMASK) > 3 ||
         mail2_ps->nativeSB.id & SUPERTYPE_ID_OVERRIDE)
  {
    ++dbmstate_s.errc;
	iss = MICR_BADSUPTYP;
    parmdw = mail2_ps->nativeSB.id  && SUPERTYPE_ID_WMASK;
	slcCmlogLogMsg(iss, "BADSUPTYP", "Invalid supertype %lu received from DBEX.\n", parmdw);
	++dbRecvDiags_s.dbexUpdDropCnt;
    return iss;
  }
  if ( (int) (nwords = (actrcvbc >> 1) - (mail2_ps->nativeSB.blkbeg - 1)) <= 0 )
    {DEBUGPRINT(DP_ERROR,dbRecvFlag, ("DBRECVPROCESS :bad 1st \n"));
	goto egress;}
  if ( actrcvbc & 1 || actrcvbc <= wordcnt )
    {DEBUGPRINT(DP_ERROR,dbRecvFlag, ("DBRECVPROCESS :bad 2nd \n"));
	goto egress;}            
  if ((mail2_ps->nativeSB.blkbeg != 1 || mail2_ps->nativeSB.segbeg != 1 ||
      (dbmstate_s.firstf != dbmmailst_nothingyet && dbmstate_s.firstf != dbmmailst_iplxone &&
       dbmstate_s.firstf != dbmmailst_iplxtwo && dbmstate_s.firstf != dbmmailst_iplxthree)) &&
       (mail2_ps->nativeSB.blkbeg != wordcnt + 1 || mail2_ps->nativeSB.segbeg <= wordcnt ))
  {
	goto egress;  
  } 
  
  if
    (dbmstate_s.sup_expwc[supblkx] > 0 && dbmstate_s.sup_expwc[supblkx] != mail2_ps->nativeSB.len) {
	DEBUGPRINT(DP_ERROR,dbRecvFlag, ("DBRECVPROCESS :bad 4th; expwc[%u]=%lu, .len = %lu \n",
									 supblkx, dbmstate_s.sup_expwc[supblkx],mail2_ps->nativeSB.len ));
	goto egress;}    
   if (
         mail2_ps->nativeSB.segbeg == 0 ||
         mail2_ps->nativeSB.segbeg > mail2_ps->nativeSB.len ||
         mail2_ps->nativeSB.segbeg + nwords - 1 > mail2_ps->nativeSB.len)
  
/* Explanation of if clause : Here nwords is # of words ("word"="16-bit thing") 
   to actually copy into micro's memory.  Line 1 says that that mustnt look negative.  
   Line 2  says the received message must contain an integral # of words (not an odd
   bytecount), and the received message must be at least as long as the superblock  
   header structure.  By this time we know that flag bit SUPERTYPE_ID_OVERRIDE is  
   clear. Lines 3 through 7 say that if (blkbeg is other than 1 or segbeg is other 
   than 1 or DBMAIN's "state" is none of dbmmailst_nothingyet,  dbmmailst_iplxone, 
   dbmmailst_iplxtwo, or dbmmailst_iplxthree),and (data in  this message begins 
   somewhere other than immediately after the superblock header or copy into the 
   micro's superblock's memory would begin within that  superblock's header in memory) 
   then it's wrong.  Lines 8 & 9 say that if we've  already seen (downloaded) some of 
   this superblock then what this message says should be the length of this whole 
   superblock must agree with what we were told it should be when we first were told 
   anything at all about this superblock.
   Line 10 says destination for copy operation cant be to a strictly lower address 
   than where the beginning of this superblock's superblock header in micro's  memory 
   is or will be.  
   Line 11 says destination for first word of copy  operation cant be to a strictly 
   higher address than where this superblock ends (or will end) in micro's memory.  
   Line 12 says same for last word of copy  operation. 
   (These comparisons are mostly unsigned.)
*/

  /* if  ( 
	 ((int) (nwords = (actrcvbc >> 1) - (mail2_ps->nativeSB.blkbeg - 1)) <= 0) ||
         (actrcvbc & 1) || (actrcvbc <= sizeof(mail2_ps->nativeSB)) ||
	 (
	   (mail2_ps->nativeSB.blkbeg != 1) || (mail2_ps->nativeSB.segbeg != 1) ||
           (
	    (dbmstate_s.firstf != dbmmailst_nothingyet) && 
            (dbmstate_s.firstf != dbmmailst_iplxone) &&
            (dbmstate_s.firstf != dbmmailst_iplxtwo) && 
	    (dbmstate_s.firstf != dbmmailst_iplxthree) &&
            (mail2_ps->nativeSB.blkbeg != sizeof(mail2_ps->nativeSB)/2 + 1 ||
             mail2_ps->nativeSB.segbeg <= sizeof(mail2_ps->nativeSB)/2) 
           ) 
        &supblk,   )
	 ||
	 (
            (dbmstate_s.sup_expwc[supblkx] > 0)
	     && (dbmstate_s.sup_expwc[supblkx] != mail2_ps->nativeSB.len) 
	 )
	 ||
            (mail2_ps->nativeSB.segbeg == 0) ||
            (mail2_ps->nativeSB.segbeg > mail2_ps->nativeSB.len) ||
	    (mail2_ps->nativeSB.segbeg + nwords - 1 > mail2_ps->nativeSB.len)
	 )
  */

  {
    DEBUGPRINT(DP_ERROR,dbRecvFlag, ("DBRECVPROCESS :bad 5th \n"));  
	++dbmstate_s.errc;
    DEBUGPRINT(DP_ERROR,dbRecvFlag, ("DBRECVPROCESS : DB did not validate\n"));
	goto egress;
  }
 
     /* Now check supertype number against our current "state" index.  If bootup  */
     /* time and we havent allocated any memory for the superblock to which the   */
     /* incoming message belongs, then we allocate memory for entire incoming     */
     /* superblock for this micro.  During bootup-time download we count words    */
     /* received for each supertype.  We can check exact completeness of each     */
     /* superblock except supertype 2.  We consider bootup-time download complete */
     /* when supertype 3 looks complete.                                          */
  DEBUGPRINT(DP_INFO,dbRecvFlag, ("DBRECVPROCESS : DB Validated - moving on! \n")); 
  mailswitchagain: switch (dbmstate_s.firstf)
  {
  case dbmmailst_ipled:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _ipled\n"));
    if (supblkx == 2)
	  {
		dbMsgCopy(mail2_ps, supblkx, nwords);
	  }
    else
	  {
		parmdw = mail2_ps->nativeSB.id && SUPERTYPE_ID_WMASK;
		iss = MICR_CANTREIPL;
		slcCmlogLogMsg(iss, "CANTREIPL","Supertype %lu not receivable after micro IPL is complete.\n", parmdw);
	  }
    break;
 
  case dbmmailst_nothingyet:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _nothingyet\n"));
	if (supblkx == 0) {
	  if (SUCCESS(dbSTalloc(mail2_ps, supblkx))) {
		dbmstate_s.firstf = dbmmailst_iplzero;
		DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("\t dbSTalloc success; allocated %ld words \n", mail2_ps->nativeSB.len));
	  }
	  else {
		dbmstate_s.firstf = dbmmailst_iplfail;
		DEBUGPRINT(DP_ERROR,dbRecvFlag, ("\t  supblkx == 0; _iplfail \n"));
		break;
	  }
	}
	else {
	  if (supblkx != 2) {
		parmdw = supblkx;
		iss = MICR_IPLDBDWNSEQ;
		slcCmlogLogMsg(iss, "IPLDBDWNSEQ", "Database download supertype sequence wrong:  expecting 0, got %lu.\n",
					   parmdw);
		dbmstate_s.firstf = dbmmailst_iplfail;
	  }
	  break;
	}
 
  case dbmmailst_iplzero:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _iplzero\n"));
	if (supblkx == 0) {
	  DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: copying %ld words in ST0\n", nwords));
	  dbMsgCopy(mail2_ps, supblkx, nwords);
	  if ((dbmstate_s.sup_gotwc[0] += nwords) == dbmstate_s.sup_expwc[0]) {
		dbmstate_s.firstf = dbmmailst_iplxone;
	  }
	}
	else {
	  if (supblkx != 2) {
		parmdw = supblkx;
		iss = MICR_IPLDBDWNSEQ;			  
		slcCmlogLogMsg(iss, "IPLDBDWNSEQ", "Database download supertype sequence wrong:  expecting 0, got %lu.\n",
					   parmdw);
		dbmstate_s.firstf = dbmmailst_iplfail;
		DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("\t  supblkx !=2; _iplfail \n"));
	  }
	}
	break;
 
  case dbmmailst_iplxone:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _iplxone\n"));
	if (supblkx == 1) {
	  if (SUCCESS(dbSTalloc(mail2_ps, supblkx))) {
		dbmstate_s.firstf = dbmmailst_iplone;
	  }
	  else {
		dbmstate_s.firstf = dbmmailst_iplfail;
		DEBUGPRINT(DP_ERROR,dbRecvFlag, ("\t  supblkx !=1; _iplfail \n"));
		break;
	  }
	}
	else {
	  if (supblkx != 2) {
		parmdw = supblkx;
		iss = MICR_IPLDBDWNSEQ;
		slcCmlogLogMsg(iss, "IPLDBDWNSEQ", "Database download supertype sequence wrong:  expecting 1, got %lu.\n",
					   parmdw);
		dbmstate_s.firstf = dbmmailst_iplfail;
	  }
	  break;
	}
 
  case dbmmailst_iplone:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _iplone\n"));
	if (supblkx == 1) {
	  dbMsgCopy(mail2_ps, supblkx, nwords);
	  if ((dbmstate_s.sup_gotwc[1] += nwords) == dbmstate_s.sup_expwc[1])
        dbmstate_s.firstf = dbmmailst_iplxtwo;
	}
	else {
	  if (supblkx != 2) {
		parmdw = supblkx;
		iss = MICR_IPLDBDWNSEQ;			  
		slcCmlogLogMsg(iss, "IPLDBDWNSEQ", "Database download supertype sequence wrong:  expecting 1, got %lu.\n",
					   parmdw);	  
		dbmstate_s.firstf = dbmmailst_iplfail;
		DEBUGPRINT(DP_ERROR,dbRecvFlag, ("\t  supblkx !=2; _iplfail \n"));
	  }
	}
	break;
 
  case dbmmailst_iplxtwo:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _iplxtwo\n"));
	if (supblkx == 2 || supblkx == 3){
	  if (SUCCESS(dbSTalloc(mail2_ps, supblkx))) {
		dbmstate_s.firstf = newfirstf[supblkx];
		goto mailswitchagain;
	  }
	  else {
		dbmstate_s.firstf = dbmmailst_iplfail;
		break;
	  }
	}
	else {
	  parmdw = supblkx;
	  iss = MICR_IPLDBDWNSEQ;		  
	  slcCmlogLogMsg(iss, "IPLDBDWNSEQ", "Database download supertype sequence wrong:  expecting 2, got %lu.\n",
					 parmdw);	 
	  dbmstate_s.firstf = dbmmailst_iplfail;
	  break;
	}
 
  case dbmmailst_ipltwo:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _ipltwo\n"));
	if (supblkx == 2) {
	  dbMsgCopy(mail2_ps, supblkx, nwords);
	  if ((dbmstate_s.sup_gotwc[2] += nwords) >= dbmstate_s.sup_expwc[2])
		dbmstate_s.firstf = dbmmailst_iplxthree;
	}
	else {
	  parmdw = supblkx;
		iss = MICR_IPLDBDWNSEQ;			  
		slcCmlogLogMsg(iss, "IPLDBDWNSEQ", "Database download supertype sequence wrong:  expecting 2, got %lu.\n",
					   parmdw);	 
		dbmstate_s.firstf = dbmmailst_iplfail;
	}
	break;
 
  case dbmmailst_iplxthree:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _iplxthree\n"));
	if (supblkx == 3) {
	  if (SUCCESS(dbSTalloc(mail2_ps, supblkx))){
		dbmstate_s.firstf = dbmmailst_iplthree;
	  }
	  else {
		dbmstate_s.firstf = dbmmailst_iplfail;
		break;
	  }
	}
	else {
	  if (supblkx == 2) {
		dbMsgCopy(mail2_ps, supblkx, nwords);
	  }
	  else {
		parmdw = supblkx;
		iss = MICR_IPLDBDWNSEQ;			  
		slcCmlogLogMsg(iss, "IPLDBDWNSEQ", "Database download supertype sequence wrong:  expecting 3, got %lu.\n",
					   parmdw);	 
		dbmstate_s.firstf = dbmmailst_iplfail;
	  }
	  break;
	}
 
  case dbmmailst_iplthree:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _iplthree\n"));
	if (supblkx == 3) {
	  dbMsgCopy(mail2_ps, supblkx, nwords);
	  /* CHECK THE FOLLOWING IF; failure would prevent setting state to ipled...... */
	  if ((dbmstate_s.sup_gotwc[3] += nwords) == dbmstate_s.sup_expwc[3]) {
		/* Looks like we just received the last of the bootup download of supertype 3 */
		dbmstate_s.ipliss = MICR_IPLOK;
		dbmstate_s.firstf = dbmmailst_ipled;
		iss = dbSendConvert(); 

                 /* err_send0(&dbmstate_s.ipliss);*/
	  }
	}
	else {
	  if (supblkx == 2) {
		dbMsgCopy(mail2_ps, supblkx, nwords);
	  }
	  else {
		parmdw = supblkx;
		iss = MICR_IPLDBDWNSEQ;		  
		slcCmlogLogMsg(iss, "IPLDBDWNSEQ", "Database download supertype sequence wrong:  expecting 3, got %lu.\n",
					   parmdw);	 
		dbmstate_s.firstf = dbmmailst_iplfail;
	  }
	}
	break;

  case dbmmailst_iplfail:
	DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBRECVPROCESS: case _iplfail - very end\n"));
	;
  }
  return iss;

 egress: {
    iss =  MICR_DBBADMESSG;
    slcCmlogLogMsg(iss, "DBBADMESSG", 
	"Invalid message from DBEX. Bytecount=%lu. Message: network=%hu, id=0x%hx, len=%lu, micr=%.4s, blkbeg=%lu, segbeg=%lu\n", 
	actrcvbc, mail2_ps->nativeSB.network, mail2_ps->nativeSB.id,
	mail2_ps->nativeSB.len, (char *)&mail2_ps->nativeSB.micr, 
    mail2_ps->nativeSB.blkbeg, mail2_ps->nativeSB.segbeg); 
	++dbRecvDiags_s.dbexUpdDropCnt;
	return iss;
  }
 }                                                  /* End local dbRecvProcess. */
/*=============================================================================
 
  Name: dbMsgCopy

  Abs:  Copy (overwrite) piece just received of micro's database into its location in 
        slc ioc's copy of superblock.
  
  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        const dbsndrcvbuf_ts*  mail2_ps           R      ptr to tcp hdr in incoming message
        const unsigned short   supblkx            R      superblock number (0-3)
        const unsigned long    nwords             R      word count of data to be copied

  Rem:  If superblock never got allocated then we complain and lose the update 
        (from the micro's point of view this can happen only for supertype 2, 
        and from any broader point of view it cant happen at all).  
        dbVMStoHSupHdr() fixes endianness and unpacks the dbsuptypehdr portion of the data 
        into a new member of dbsndrcvbuf_ts, nativeSB, once. 
        It cannot be unpacked on top of itself (in place)  because it may occupy 1 more 
        word due to padding on some architectures. Thus, all code should refer to field 
        nativeSB (as opposed to supblk->sbhdr) for the native super block header.


  Side: Called by dbRecvProcess. 
        References dbnodes[], public pointer to superblock data

  Ret:  none
  
=============================================================================*/                        
LOCAL void dbMsgCopy(const dbsndrcvbuf_ts *const mail2_ps, 
                     const unsigned short supblkx, 
                     const unsigned long nwords)
{
  vmsstat_t iss = MICR_OKOK;

  DEBUGPRINT(DP_INFO,dbRecvFlag, 
			 ("DBMSGCOPY: copy incoming buffer for supblkx = 0x%d, word length to copy = %lu...\n", 
			  supblkx, nwords));
  if (dbnodes_a[supblkx] != NULL) {
	memcpy(&dbnodes_a[supblkx]->datw[mail2_ps->nativeSB.segbeg-1],
		   &mail2_ps->supblk.datw[mail2_ps->nativeSB.blkbeg-1], nwords << 1); 
  }
  else {
	iss=MICR_DBNEVRDWNLD;
	slcCmlogLogMsg(iss, "DBNEVRDWNLD", "Database supertype %hu was never downloaded!\n", supblkx);
  }
 }                                                    /* End local dbMsgCopy. */
 /*=============================================================================
 
  Name: dbSTalloc

  Abs:  Allocate memory for whole superblock that just began arriving.      
  
  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        const dbsndrcvbuf_ts*  mail2_ps            R      ptr to tcp hdr in incoming message
        const unsigned short   supblkx            R      superblock number (0-3)

  Rem:  If we need to allocate memory but cant then we complain and kill the download 
        entirely.  If we've already allocated memory for the given supertype then we dont    
        indicate any error (we have already validated superblock size specified in 
        incoming message against size if any already allocated)

  Side: Called by dbRecvProcess. 
        Writes to dbnodes[], public pointer to superblock data
        Writes to dbmstate_s, local structure containing dbRecvProcess's current state.  

  Ret:  vmsstat_t 
  
=============================================================================*/   
LOCAL vmsstat_t dbSTalloc(const dbsndrcvbuf_ts *const mail2_ps, const unsigned short supblkx)
 {   
   vmsstat_t      iss = MICR_OKOK;
   unsigned long  length;
 
   /*----------------------------------------------*/
   DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DBSTALLOC: entering for supblkx = 0x%x...\n", supblkx));
   if (!dbnodes_a[supblkx]) {
	 length = ((mail2_ps->nativeSB.len +1) & -2)<<1;
	 DEBUGPRINT(DP_INFO,dbRecvFlag,("DBSTALLOC: malloc length for supn %d = %ld bytes\n", supblkx, length));
	 dbnodes_a[supblkx] = malloc(length);

	 DEBUGPRINT(DP_INFO,dbRecvFlag,("DBSTALLOC: malloc ptr returned for supblk %d = 0x%p = \n",
									(int) supblkx,
									dbnodes_a[supblkx] ));        

	 if (dbnodes_a[supblkx] == NULL) {
	   iss = MICR_EXNET_RECVFAIL;
	   DEBUGPRINT(DP_INFO,dbRecvFlag, ("DB RECV - dbSTalloc malloc error\n"));
	   iss = MICR_DBNOMEM;
	   slcCmlogLogMsg(iss, "DBNOMEM", "%s cant allocate memory for supertype %hu, size=%lu words.\n", 
					  epicsThreadGetNameSelf(), supblkx, mail2_ps->nativeSB.len);
	 }
	 else {
	   dbmstate_s.sup_expwc[supblkx] = mail2_ps->nativeSB.len; 
	 }
   }
   else {
	 iss = MICR_EXNET_RECVFAIL;
	 DEBUGPRINT(DP_INFO,dbRecvFlag, ("DBSTALLOC - supertype_ptr is NULL\n"));
   }
   return iss;
 }
 /*=============================================================================
 
  Name: dbSendDiagRpy

  Abs:  Fill Q message and send to dbSend to reply to diagnostic inquiry from DBEX      
  
  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        const dbsndrcvdiagmsg_ts*  diagmsgbuf_ps    R      ptr to tcp hdr in incoming message
        const unsigned short   diagmsgbc            R      byte cnt of supertype hdr + data

  Rem:  Diagnostic inquiry/reply uses database main DBEX socket pathway, not ack pathway.  
        Either we or DBEX can be restarted at any time;  hi-order bit of send count 
        indicates restart.  We assume our caller has verified minimum length of received 
        message.
    
  Side: Called by dbRecvThread 
        Writes to dbRecvmsg_s, public queued message structure

  Ret:  none 
  
=============================================================================*/                        
LOCAL void dbSendDiagRpy(const dbsndrcvdiagmsg_ts *const diagmsgbuf_ps, 
			  const unsigned short diagmsgbc)
{   
  vmsstat_t      iss = MICR_OKOK;

  DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DB RECV: DBSENDDIAGRPY: entering \n"));
  /* formulate dbRecvmsg_s to dbSend */
  memset(&dbRecvmsg_s, 0, sizeof(dbRecvmsg_s));
  dbRecvmsg_s.func = DB_DIAG_DBEX;
  /* not used in dbSend dbRecvmsg_s.msgheader.datalen = diagmsgbc;*/

  /* unpack & byte swap the VMS dbex_* and micr_* data to native */
  cvtLSwapMove((const void*)&(diagmsgbuf_ps->diagdata.datw[11]),(void *)&dbRecvmsg_s.dbex_msgc_snd);
  cvtLSwapMove((const void*)&(diagmsgbuf_ps->diagdata.datw[13]),(void *)&dbRecvmsg_s.dbex_msgc_rcv); 
  cvtLSwapMove((const void*)&(diagmsgbuf_ps->diagdata.datw[15]),(void *)&dbRecvmsg_s.micr_msgc_snd); 
  cvtLSwapMove((const void*)&(diagmsgbuf_ps->diagdata.datw[17]),(void *)&dbRecvmsg_s.micr_msgc_rcv);

  /* mail response to dbSend to send to dbex */

  if (SUCCESS(iss = messageQSend(dbSend, epicsTrue, sizeof(dbRecvmsg_s), (void *) &dbRecvmsg_s )))   
  {
    ++dbRecvDiags_s.dbexDiagCnt;
    DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DB RECV: DBSENDDIAGRPY: sent msg to dbSend for reply\n"));
    /*dbprintVMSSupHdr ("\tVMS SupHdr Diag Reply msg: \n", 
		    (dbsuptype_tu *)&dbRecvmsg_s.dbdiag, 
		    sizeof(dbRecvmsg_s.dbdiag)); */
  }
    
 }                                              /* End local dbSendDiagRpy. */

 /*=============================================================================
 
  Name: dbRecvDXUPDN

  Abs:  Decodes either DXUP or DXDN message from DBEX      
  
  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        const dbsndrcvbuf_ts*  dxmsgbuf_ps    R      ptr to tcp hdr in incoming message
        const unsigned short   msgbc          R      byte cnt of supertype hdr + data

  Rem:  
        "DXUP"     Non-msg-svc message from DBEX consisting of:  fwd_hdr_ts, followed by
                   dbsuptypehdr_ts in which member id contains either
                   (SUPERTYPE_ID_OVERRIDE | SUPERHDR_FUNC_NOTFY | SUPERTYPE_ID_IPLDWN)
                   or just (SUPERTYPE_ID_OVERRIDE | SUPERHDR_FUNC_NOTFY),
                   followed by "DXUP", followed by 8-byte ascii version number as updated
                   by most recent run of dbgen.  Ignore members blkbeg and segbeg in this
                   instance of dbsuptypehdr_ts.  This message is represented by
                   dbexmnotifymsg_ts in REF_RMX_INCLUDE:DB_MICRO_STRUC.HM.
                   DBEX will send this message to all SLC-aware IOC's when DBEX is
                   starting up, and to any SLC-aware IOC at database download time for
                   that IOC.  No reply or ack expected for this message.
 
        "DXDN"    Non-msg-svc message from DBEX consisting of:  fwd_hdr_ts, followed by
                  dbsuptypehdr_ts in which member id contains
                  (SUPERTYPE_ID_OVERRRIDE | SUPERHDR_FUNC_NOTFY), followed by
                  "DXDN".  Ignore members blkbeg and segbeg in this instance of
                  dbsuptypehdr_ts.  DBEX will send this message to all SLC-aware IOC's
                  when DBEX is knowingly exiting.  No reply or ack expected for this
                  message.

    
  Side: Called by dbRecvThread 
        Stores dbVersion_a upon "DXUP" message; sets dbex_up flag
	to epicsTrue for "DXUP"; epicsFalse for "DXDN" message

  Ret:  none 
  
=============================================================================*/                        
LOCAL void dbRecvDXUPDN(const dbsndrcvbuf_ts *const dxmsgbuf_ps, 
			  const unsigned short msgbc)
{   
  char           notify_a[4];
  vmsstat_t      iss = MICR_OKOK;

  DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DB RECV: DBSENDDXUPDN: entering \n"));
  /* copy "DXUP" or "DXDN" */
  memcpy(&notify_a, &(dxmsgbuf_ps->supblk.datw[11]), 4);
  if (strncmp(notify_a, "DXUP", 4)==0){
    /* copy latest dbVersion into global */
    memcpy(&dbVersion_a, &(dxmsgbuf_ps->supblk.datw[13]),8);
	/* copy dbVersion into diagnostics global var too */
    memcpy(&dbRecvDiags_s.dbVersion_a, &dbVersion_a, 8); 
	DEBUGPRINT(DP_ERROR,dbRecvFlag, ("DB RECV: dbVersion(): %.8s\n", dbRecvDiags_s.dbVersion_a));
    if (!dbex_up){
      /* send dbupdate ALL* (if db exists) to dbSend; doesn't block on any ackEvent for AlL* case*/
      dbex_up = epicsTrue;
      dbRecvDiags_s.dbexUpState = 1;
	  if (slcGetDbExists()){
		dbRecvmsg_s.func = DB_UPDATE_ALL;
		if (SUCCESS( messageQSend(dbSend, epicsTrue, sizeof(dbRecvmsg_s), (void *) &dbRecvmsg_s )))   
		  {
			DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DB RECV: dbRecvDXUPDN(): sent update all to dbSend \n"));		  
		  }
		/* LOG a DBEX state change */
		iss=DB_DBEX_UP;
		slcCmlogLogMsgFile(iss,"DBEX_UP", __FILE__,__LINE__, "%.4s: DBEX is changing state from down to up.\n",
						(char *)&dblongmicro);	 
	  }
	}
  }
  else {
	if (strncmp(notify_a, "DXDN", 4)==0){
	  if (dbex_up){
		/* LOG a DBEX state change */
		iss=DB_DBEX_DN;
		slcCmlogLogMsgFile(iss,"DBEX_DN", __FILE__,__LINE__, "%.4s: DBEX is changing state from down to up.\n",
						(char *)&dblongmicro);
		++dbRecvDiags_s.dbexDownCnt;
	  }
	  dbex_up = epicsFalse;
      dbRecvDiags_s.dbexUpState = 0;
	}
  }
  DEBUGPRINT(DP_INFO,dbRecvFlag, ("DB RECV: DBSENDDXUPDN: dbVersion = %s \n", dbVersion_a));  
  DEBUGPRINT(DP_INFO,dbRecvFlag, ("DB RECV: DBSENDDXUPDN: dbexUpState = %d \n", dbRecvDiags_s.dbexUpState));   
}
 /*=============================================================================
 
  Name: dbSendConvert

  Abs:  Fill Q message and send to dbHdlr so it will translate downloaded ST
        data into dictionary      
  
  Args: None

  Rem:  
    
  Side: Called by dbRecvProcess after all data has been downloaded 
        Writes to dbRecvmsg_s, public queued message structure

  Ret:  none 
  
=============================================================================*/ 
LOCAL vmsstat_t dbSendConvert(void ) {
  vmsstat_t iss = MICR_OKOK;
  
  dbRecvmsg_s.func = DB_CONVERT;

  /* mail convert request to dbHdlr */

  if (SUCCESS(iss = messageQSend(dbHdlr, epicsTrue, sizeof(dbRecvmsg_s), (void *) &dbRecvmsg_s )))   
  {
    DEBUGPRINT(DP_DEBUG,dbRecvFlag, ("DB RECV: dbSendConvert(): sent request to dbHdlr \n"));

  }
  return iss;
}
