/*=============================================================================
 
  Name: dbST0.c
        dbST0Convert   - main entry point 
        dbST0alloc     - alloc an ST0 list
        dbST0list      - get ST0 data for given prim unit secn
        dbST0units     - get ST0 unit data for given prim
        dbST0GetUpsBlk - called by dbST0list as it ref's ST0 data
        dbST0GetSubBlk - called by dbST0list as it ref's ST0 data
        dbST0ChkClist  - checks input ST0 list
        dbST0CreateDict- creates dictionary from file and ST0 data
        dbST0GetDebugdblist - debug
        dbST0printlist - debug; prints ST0 list

  Abs:  Creates the dictionary (character based hash table) from the downloaded 
        SLC Supertype 0 data and primary.map file.
        Reference web pages on LCLS Controls | Facilities |SLC-Aware IOC | :
                        Database Service (DBS) Design Document 
                        Functional Requirements | SLC Database Service (DBS)
                        Database Input/Output Utilities
                        C Coding Standard

        Referenced SLCRMX:REENTRANT dbgetc.c & dblistu.a38
                     (also Alpha side:) REF_:[DBSSHR]DBLIST_C.C,
                                REF_:[DBSSHR]FNDUPSBLK.C, 
                                REF_:[DBSSHR]FNDSUBBLK.C;
At the end of this module is header documentation describing the ST0 database; 
extracted from dblistu.a38.
 Note:  slc types (int2u, int4u, etc are used in this module)

  Auth: 28-OCT-2004, Debbie Rogind (DROGIND):
  Rev:  dd-mmm-200y,
-----------------------------------------------------------------------------*/
#include "copyright_SLAC.h"    
/*-----------------------------------------------------------------------------  
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
int dbST0Flag = DP_ERROR;
#endif
/* c includes */
#include     <stdio.h>
#include     <stdlib.h>    /* malloc, free */
#include     <string.h>    /* memcpy */
#include     <cantProceed.h>

/* slc global vars */
#include     "slcGlob.h"

/* VMS Control system function/status codes  */
#include     "msgfunc.hc"  /* FUNC_MSG function code */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */

/* slc common includes */
#include     "cvtUtil.h"   /* slc conversion utilities */
#include     "timeUtil.h"  /* slc time conversion utilities */
#include     "slcCmlog.h"  /* slcCmlogLogMsg */
#include     "slcUtil.h"   /* slcGetIOCName */ 

/* dbs includes  */  
#include     "db_micro_struc.h"
#include     "dbUtil.h" 
#include     "dbmsgmail.h" 
#include     "dbGlob.h"    /* dbmaincom_p */
#include     "dbdef.hc"
#include     "dbdict.h"
#include     "freeList.h"
#include     "dbhash.h"
#include     "dblist.h"
#include     "dbget.h"
#include     "dbST0.h"
/****************************************************************************/


/*****************************************************************************/

/* DBST0CLR macro initializes an ST0 list */
#define DBST0CLR(list) if ((list) != NULL) (list)->cur = 0


/* dbdict prototype */
#include     "dbdict.h"
/* freeList prototype */
#include     "freeList.h"
#include     "dblist.h"
#include     "dbST0.h"

LOCAL vmsstat_t  dbST0alloc( const int2u bytes_in_element, 
			     const int2u num_elements,
			     dbST0header_ts **const dbdata_pp);

LOCAL vmsstat_t  dbST0list (const int2u prim,
			    const int2u unit,
			    const int2u secn,
			    dbST0header_ts **const dblist_pp);

LOCAL vmsstat_t  dbST0units (const int2u prim,
			     dbST0header_ts **const dbdata_pp);

LOCAL unsigned long dbST0GetUpsBlk (const unsigned short catn, 
				    const unsigned short unit, 
				    const suptype0_ts *const sb0);

LOCAL unsigned long dbST0GetSubBlk (const unsigned long ups_offset, 
				    const unsigned char secNum, 
				    const suptype0_ts *sb0);
LOCAL vmsstat_t dbST0ChkClist (int4u *orgCnt_p, 
			       int4u *maxCnt_p, 
			       dbST0header_ts * * const dbdata_pp);
 
LOCAL vmsstat_t dbST0CreateDict ( void);

LOCAL vmsstat_t dbST0GetDebugdblist (void);


#define LISTHEAD 0xFFFF           /* Magic value for Primary listhead */ 
#define INIT_LIST 40              /* init # of elements if invalid list */

/*=============================================================================
 
  Name: dbST0Convert

  Abs:  Main routine called by dbHdlr to sequence through the calls to routines
        that handle a) converting ST0 data to native and b) transforming ST0 data
        to char-based hash dictionary used by the slc aware ioc. 
  
  Args: None 
 
  Rem:  This procedure calls other utility procedures in dbST0 to handle
        all conversions and transformations from ST0 to dictionary format.
        1) Converts ST0 to native endian-ness - dbVMStohST0()
        2) Moves ST0 data to dictionary - dbST0CreateDict()

  Side:
 
  Ret:  vmsstat_t VMS style error condition code defined in dbdef.hc
        MICR_OKOK, DB_BADDICT
  
=============================================================================*/
vmsstat_t  dbST0Convert (void)
{
  vmsstat_t iss = MICR_OKOK;
  /* Convert ST0 pointers to native endianness */
  dbVMStohST0();
  /* debug - get a dblist & print */
  /* iss = dbST0GetDebugdblist(); */ 
  /* create dict using primary.map and ST0 data */
  
  iss = dbST0CreateDict(); 
  
  return (iss); 
}
/*=============================================================================
 
  Name: dbST0CreateDict

  Abs:  Creates the dictionary from the downloaded supertype 0 data and file 
        primary.map
  
  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
        None

  Rem:  For every line of PRIMARY.MAP file, it tries to match to ST0 data. If
        there is a match, all ST0 data and accompanying line information is entered
        into the dictionary. 

        For every new primary first encountered when reading the file, a call to
        dbST0units determines whether this ioc is configured for the primary.  
        Reads file primary.map line by line, compares with ST0 data, and:  
	A) For every unique primary name, a dictionary entry is made for "primary". 
	Note : the primary name does not have to match any ST0 data 
	B) Future? For every unique primary secondary name, an entry is made for "primary secondary"
	in the dictionary. Note : the prim/sec name does not have to match any ST0 data 
	C) For every primary, unit, secondary that matches ST0 data, an entry for 
        "primary unit secondary" will be made in the dictionary. Also, the entry 
        will be added to the (ALL* units) linked list associated with that primary entry of A).  
        Question (Ken) - Does this make sense: for ALL* linked list heads, should there be 
        "primary secondary" linked list heads in addition to just "primary" linked lists 
        (for the case of db* calls, other than dbunits, where unit = ALL*)? 
        This would allow faster access when secondary is specified and unit = ALL*. 
        Perhaps this could be added later, if this use case is prevelant in the code.
 
  Side: ASSUMPTIONS: 
        primary.map file location is set by env var. 
        All lines in file have contiguous primary name. 
        prim and secn names are read from file as left justified, and are blank padded
		 by this rte.

  Ret:  VMS style error code
        MICR_OKOK, DB_BADPRIMMAP, DB_BADDICT
  
=============================================================================*/
LOCAL vmsstat_t dbST0CreateDict ( void)

{
#define  SZ_LINE 60
#define PRIMMAP_SIZE 100
  char  filename_c[PRIMMAP_SIZE];
  char *primmap_c;
  vmsstat_t     iss = DB_OKOK;
  vmsstat_t     isss = DB_OKOK;     
  FILE          *file_p = NULL;
  char          *isEnd; 
  char          line[SZ_LINE];  
  char          prim_ca[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char          secn_ca[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  int           curPrimNum;
  unsigned short catn = 0;
  unsigned short secn = 0;
  char          fmt;
  int           width = 0;
  int           supn = 0;
  dbST0header_ts * dbunits_ps = NULL; /* pointer to units list */  
  dbST0header_ts * dblist_ps = NULL;  /* pointer to dblist list */  
  int4u         unitCnt = 0;
  int4u         listCnt = 0; 
  dbhEntry_ts   *dbhdictP_ps = NULL;    /* dict ptr back from inserting prim - only entry */
  epicsBoolean  newPrim;
  int           hasPrimCnt = 0;
  void *        commentChar_p = NULL;
  char          count_a[20];

  DEBUGPRINT(DP_DEBUG,dbST0Flag,("DBST0: Entering routine! \n"));

  /* list init/free processing and file processing */
  if ((primmap_c = getenv("IOC_PRIM_MAP"))) {
    strncpy(filename_c, primmap_c, PRIMMAP_SIZE);
    filename_c[PRIMMAP_SIZE-1] = 0;
  }
  else /* provide default */
  {
    strcpy(filename_c, "/slc/primary.map");
  }
  DEBUGPRINT(DP_INFO,dbST0Flag,("DBST0: primary filename = %s\n",(char *)filename_c));
  DEBUGPRINT(DP_DEBUG,dbST0Flag,("DBST0: IOC_PRIM_MAP = %s\n",primmap_c));

  /* create data list */
  if (! SUCCESS (iss= dbST0alloc( 4, 40, &dblist_ps))) {
    DEBUGPRINT(DP_ERROR,dbST0Flag,("DBST0: dbST0alloc dblist_pps failed \n"));
    return (iss);
  }
  if (! SUCCESS (iss= dbST0alloc( 4, 40, &dbunits_ps))) {
    DEBUGPRINT(DP_ERROR,dbST0Flag,("DBST0: dbST0alloc dbunits_pps failed \n"));
    return (iss);
  }
  DEBUGPRINT(DP_DEBUG,dbST0Flag,("DBST0: dbST0alloc success for both lists \n"));

  file_p = fopen ((char*)filename_c, "r");
  /*file_p = fopen("/TFTP/BOOTP_HOST/g/lcls/ioc/iocBoot/cdvw4/primary.map", "r"); - works - abs path*/
  if (!file_p) {
    DEBUGPRINT(DP_ERROR,dbST0Flag,("DBST0: primary.map file = %s not found \n", (char *)filename_c));
	DEBUGPRINT(DP_ERROR,dbST0Flag,("DBST0: file_p = %p\n", file_p));
	DEBUGPRINT(DP_ERROR,dbST0Flag,("DBST0: IOC_PRIM_MAP=%s\n", (char *)filename_c));
	slcCmlogLogMsg(iss, "NOSERVICE", "primary.map read, format, or can't find error.\n", 
				   (char *)&dblongmicro);
	return (iss=DB_NOSERVICE);
  }
  curPrimNum = -1;
  newPrim = epicsTrue;   
  memset(line,0,SZ_LINE);
  isEnd = fgets(line, sizeof(line)-1, file_p);
  commentChar_p = memchr(line,'#', 5);
  /* a line must be read from primary.map prior to entering this loop */
  while (isEnd != NULL){
	if (!commentChar_p){
	  /* format of ea line: PrimaryName catn SecondaryName secn format width number */
	  sscanf(line, "%4s %hd %4s %hd %1s %d %s %d", 
			 prim_ca, &catn, secn_ca, &secn, &fmt, &width, count_a, &supn);
	  dbgetBlkPad(prim_ca);   /* left justified already; pad out to 4 chars */
	  dbgetBlkPad(secn_ca);
	  DEBUGPRINT(DP_DEBUG,dbST0Flag,
		  ("file line: prim=%.4s, catn=%d, sec=%.4s, secn=%d, fmt=%c, wid=%d, cnt=%s, supn=%d\n",
		  prim_ca, catn, secn_ca, secn, fmt, width, count_a, supn));
	  if (supn!=4){
		if (curPrimNum != catn){ 
		  /* first line of a primary - do some housekeeping */
		  newPrim = epicsTrue;
		  DEBUGPRINT(DP_INFO,dbST0Flag,
					 ("DBST0 1st line of a new primary; inserting prim=%.4s\n", prim_ca));
		  /* clean up lists */
		  DBST0CLR(dblist_ps);
		  DBST0CLR(dbunits_ps);
		  unitCnt = 0;
		  curPrimNum = catn; 
		}
		
		if (newPrim){
		  /* first line of a unique primary - get all units once */
		  DEBUGPRINT(DP_DEBUG,dbST0Flag,
					 ("DBST0 1st line of a new primary: %.4s - get all units via dbST0units \n", prim_ca));
		  isss = dbST0units (catn, &dbunits_ps); 
		  if (isss == DB_NOSUCHUNIT) {
			/* no units for this prim; store prim name in dict */
			DEBUGPRINT(DP_INFO,dbST0Flag,("DBST0 no units; insert primary name = %.4s \n", prim_ca));
			
			dbhdictP_ps = dbdictInsP ((char *)&prim_ca, epicsFalse);
			/* dbhdictP_ps returned will be NULL, as there is no units linked list head
			   entry for this primary */
		  }  
		  else {
			/* units exist  */
			hasPrimCnt++;
			unitCnt = ((dbST0header_ts *) dbunits_ps)->cur;
			DEBUGPRINT(DP_INFO,dbST0Flag,
			 ("DBST0 unit cnt=%ld; insert prim entry for %.4s; insert units linked list\n", 
			  unitCnt, prim_ca));   
			/* insert primary entry */
			dbhdictP_ps= dbdictInsP((char *)&prim_ca, epicsTrue);
			if (dbhdictP_ps == NULL){
			  return (iss = DB_BADPTR);
			}
			/* insert all unit ellist list heads for secns */
			/* for ea unit, add a units linked list head & the secn PUS hash entry */
			dbdictInsAllUnits ((dbdictP_ts *)dbhdictP_ps->usrPvt, dbunits_ps);
		  }
		  newPrim = epicsFalse;
		} /* newPRim */
		/* begin processing each file line for current primary (that has units) */
		/* if the secn in the file doesn't exist in the database, continue on to next */
		/*     or should this be an error? */
		if (unitCnt >0) 
		  { 
		  /* process this secn line for this prim, all units */
		  /* clear list first */
		  DBST0CLR(dblist_ps);
		  if (SUCCESS ( isss = dbST0list(catn, LISTHEAD, secn, &dblist_ps ))){
			DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0list for prim=%.4s, sec=%.4s, secn=%d\n",
										  prim_ca, secn_ca, secn )); 
			/*dbST0printlist (dblist_ps);*/
			listCnt = ((dbST0header_ts *) dblist_ps)->cur;
			if (unitCnt != listCnt) {
			  DEBUGPRINT(DP_ERROR,dbST0Flag,
						 ("\t\t unit count = %ld not equal to dbST0list count = %ld !! Exit \n", 
						  unitCnt, listCnt));
			  return (iss = DB_CORRUPTION);
			}

			/* for each unit of this primary / secn doublet, insert into dictionary */		
			/* note: count will be calculated using width and subblk word length */ 
			DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbdictInstSecnInAllUnits for prim=%.4s, sec=%.4s, secn=%d\n",
										  prim_ca, secn_ca, secn )); 
			if (! SUCCESS (iss=dbdictInsSecnInAllUnits 
						   ((dbdictP_ts *)dbhdictP_ps->usrPvt,dbunits_ps,
							prim_ca, secn_ca,dblist_ps,width,fmt,supn,&dblist_ps) )) {
			  DEBUGPRINT(DP_ERROR,dbST0Flag,("DBST0: dbdictInsSecnInAllUnits failed \n"));
			  return (iss);
			}
			
			/* }  for */
		  }/* if file secn is found in ST0 db */
		} /* unitCnt > 0 ; otherwise scarf all lines of this prim until next prim */
	  } /* supn!=4 */
	  else {
		DEBUGPRINT (DP_DEBUG,dbST0Flag,("dbST0 skipping above prim=%.4s, sec=%.4s  supn=4\n",
									   prim_ca, secn_ca));
	  }
	}/* if (commentChar ==0) */
	else {
		DEBUGPRINT (DP_DEBUG,dbST0Flag,("dbST0 skip comment line %s", line));
	  }
	/* get next line for processing at top of loop */
	isEnd = fgets(line, sizeof(line), file_p);
	commentChar_p = memchr(line,'#', 5);
  } /* while */
  DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0 read in & discovered %d primaries \n", hasPrimCnt));
  DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0 is freeing free lists\n"));  
  if (dblist_ps) free(dblist_ps);
  if (dbunits_ps) free (dbunits_ps);
  fclose (file_p);

  return (iss);
}
/*=============================================================================
 
  Name: dbST0alloc

  Abs:  This routine allocates and initializes a (legacy) data list from
        dynamic memory.
  
  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------  
 	 const int2u       bytes_in_element R      byte size of each element
 	 const int2u       num_elements     R      # elements initially to be in list
         dbST0header_ts **const dbdata_pp  R/W     ptr to ptr of list  
 
  Rem:  Allocates memory from heap for a list that points to legacy ST0 data. 
        It is assumed that the ST0 block has been converted to native format
        prior to this routine.
        

  Side: The list must be passed as a pointer to a pointer so that it
        may be modified. 
        callocMustSucceed's list

 
  Ret:  vmsstat_t VMS style error condition code defined in dbdef.h:
        MICR_OKOK, DB_BADDICT

=============================================================================*/
LOCAL vmsstat_t  dbST0alloc (const int2u bytes_in_element,  
			     const int2u num_elements,
			     dbST0header_ts **const dbdata_pp) 
{ 
   int    new_list;                        /* flag new db list */
   vmsstat_t iss = DB_BADDICT; 
   int bytes_to_allocate; 
  
/*----------------------------------------------------------------------------*/

/*   Check for valid pointer */
   if (dbdata_pp == NULL) {
     DEBUGPRINT(DP_ERROR,dbST0Flag,("dbST0alloc - invalid pointer passed in \n"));
     slcCmlogLogMsgFile(iss, "BADDICT",__FILE__,__LINE__,
			"dbST0alloc - invalid pointer passed in.\n");
     return (iss); 
   } 
/*   calculate size for all elements, plus two-word header  */
   bytes_to_allocate = (num_elements * bytes_in_element) + sizeof(dbST0header_ts); 
   /* If list pointer to pointer is NULL then this is a new list  */
   
   if ( (new_list = (*dbdata_pp == NULL))) {
     DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0alloc: dbST0alloc getting mem from heap \n")); 
     *dbdata_pp = callocMustSucceed(1, bytes_to_allocate, "dbST0alloc error\n"); 
   } 
   /*  Otherwise, re-allocate the requested list and modify list length  */
   
   else {
     DEBUGPRINT(DP_INFO,dbST0Flag,("dbST0alloc: dbST0alloc reallocating mem from heap \n"));  
     *dbdata_pp = realloc(*dbdata_pp, bytes_to_allocate); 
   }
   /*  If returned pointer is NULL then we are out of memory  */
   
   if (*dbdata_pp == NULL) { 
     slcCmlogLogMsgFile(iss, "BADDICT",__FILE__,__LINE__,
			"dbST0alloc - out of memeory.\n");
     DEBUGPRINT(DP_ERROR,dbST0Flag,("dbST0alloc - out of memeory \n"));
     return (iss); 
   } 

   /*  Adjust the header values  */
   /*  Set maximum length to number of words allocated  */
   ((dbST0header_ts*) *dbdata_pp)->max = num_elements; 
   if (new_list) { 
     ((dbST0header_ts*) *dbdata_pp)->cur = 0; 
   } 
   
   return(iss=MICR_OKOK); 
} 
/*=============================================================================
 
  Name: dbST0list

  Abs:  For the given database terms return the corresponding list of
      sptr "pointers", or supertype block 0 offsets.  These pointers may be passed to 
      a subsequent routine (see for example "dblget") to retrieve quickly the database
      information they point to.  

  
  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------  
	const int2u         prim         R           see below
	const int2u         unit         R
	const int2u         secn         R
	dbST0header_ts * *const dbdata_pp R/W

        prim   - A pointer to  a longword containing four ASCII characters.
               These four characters define the database Primary whose
               pointers will be returned.
        unit  -  A pointer to  a longword containing a possible unit number.
               A unit number may range from -32768 to + 32768 (decimal).
               If the high order word of this argument is zero it is
               assumed that a BINARY unit has been passed by the caller.
               If the high order word is non-zero then the routine assumes
               the argument is composed of four ASCII characters which are
               translated as follows:
                      - If the four characters are "ALL*" then the caller
                         has requested that the units be "wildcarded".
                      - Otherwise an ASCII to binary conversion is done
                        and the resultant value is used as the requested
                        unit number.
        secn   - A pointer to  a longword containing four ASCII characters.
               These four characters define the database Secondary whose
               pointers will be returned.
	dbdata_pp  - A pointer to a buffer into which the requested "pointers"
               will be returned. The buffer is formatted as a standard
               VMS database list. The buffer is assumed to be a word
               array, with the first element of the array containing its
               size (in words), and the second element containing the
               count (in words) of elements currently occupied in the
               buffer. It is the callers responsibility to inititalize
               this count to its proper value before calling this routine.
               A value of zero (0) indicates an empty buffer. 

  Rem:  Validates incoming list      
        if unit = ALL*, loops and inserts all prim/unit/secn into list

  Side: The list must be passed as a pointer to a pointer so that it
        may be modified. 
        callocMustSucceed's list

 
  Ret:  vmsstat_t VMS style error condition code defined in dbdef.h:
         DB_NOSUCHUNIT         
         DB_NOSUCHNAME         
         MICR_OKOK
         DB_BADDICT
=============================================================================*/
LOCAL vmsstat_t dbST0list (const int2u prim,
			   const int2u unit,
			   const int2u secn,
			   dbST0header_ts * *const dbdata_pp)

{

  int4u  orgCnt =0;               /* original dbdata list length of elements */
  int4u  maxCnt =0;               /* max elements in dbdata list */
  int4u  *next_p;                 /* next free element location in list */
                                  /* number of bytes to reallocate to list if too short */
  vmsstat_t iss =MICR_OKOK; 
  suptype0_ts    *sb0;            /* pointer to supertype 0 header */
  int4u          ups_offset;      /* word offset to ups blocks within sb0 */
  int4u          next_ups_offset; /* when looping through ups blocks */
  int2u          *ups_ptr;        /* pointer to upsblk (adds ups_offset + sb0) */
  int4u          sec_offset;      /* storage for matching sec offset */
  epicsBoolean   chasing_units;   /* 'ALL*' input for units? */     

  /* validate incoming Clist pointers & list */
  if (!SUCCESS(dbST0ChkClist (&orgCnt, &maxCnt, dbdata_pp) ))
    {
      DEBUGPRINT(DP_ERROR,dbST0Flag,("dbST0list: INVALID dbdata_pp"));
      return (iss);
    }

  /* Set up some list counters and pointers. maxCnt contains max number of
     elements which may be entered into the buffer (less the header),orgCnt contains
     the number of elements used in the buffer before this routine was called, and
     next points to the next available buffer location. */
  next_p = (int4u *)(*dbdata_pp) + 2;/* next element location in buffer */
  maxCnt =  (*dbdata_pp)->max;    /* max elements */
  orgCnt = (*dbdata_pp)->cur;     /* save original # occupied elements (after cur) */
  next_p += orgCnt;               /* next element location in buffer */
  DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0list: List in: dbdata_pp = %p; next_p = %p", dbdata_pp, next_p));

  if(unit==LISTHEAD)                       /* unit = 'ALL*" ? */
    {
      /* Yes, chasing units? */
      chasing_units = epicsTrue;
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0list: ALL* case/n"));
    }
  else
    {
      chasing_units = epicsFalse;  
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0list: unit = %d/n", unit)); 
    }

  /* init pointers */
  sb0 = (suptype0_ts *)dbnodes_a[0];

  /*
   Establish the set of UPS blocks to be searched and point to the
   first block in the set. If the units are chased (ALL*) use the special
   value "LISTHEAD" to find the first block in the set.
  */
  if (!chasing_units)  /* single unit; store one secondary offset */
    {
      /* if unit catn/unit pair match to a ups blk, 
        offset to matching ups blk is returned in ups_offset*/
      if (!(ups_offset = dbST0GetUpsBlk(prim, unit, sb0)))
	{
	    iss = DB_NOSUCHUNIT;
	    return(iss);
        }
      else
	{ 
	  /* establish next_ups_offset to exit loop after first pass */
	  next_ups_offset = ups_offset;
	}
    }
  else /* all units */
    {
      if (!(ups_offset = dbST0GetUpsBlk(prim, unit, sb0)))
	{
	    iss = DB_NOSUCHUNIT; 
            return(iss);
        }
      else /* found listhead for catn; ALL* units;  now use nextc links 
            to find all units in this cat */
	{ 
	  /* */
	  ups_ptr = (int2u *)(ups_offset + (int2u *)sb0 );
	  /* proceed using nextc link for 'ALL*' */
	  memcpy (&ups_offset, (void *)(ups_ptr + OFF_UPS_NEXTC), 4);
	  /* establish next_ups_offset using nextc again  */
	  ups_ptr = (int2u *)(ups_offset + (int2u *)sb0 );
	  memcpy (&next_ups_offset,(void *)(ups_ptr + OFF_UPS_NEXTC), 4);  
	}
    }/* else all units */
  DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0list: ups_offset = %ld; next_ups_offset = %ld", ups_offset, next_ups_offset));
  for (;;)
  /* loop over ups blks following nextc pointer until ups blk points to itself */
    {
      /* validate unit */
      sec_offset = dbST0GetSubBlk(ups_offset,(unsigned char) secn, sb0);
      if (sec_offset ==0 )
	{
	  (*dbdata_pp)->cur = orgCnt;
	  DEBUGPRINT(DP_ERROR,dbST0Flag,("dbST0list - no such subblk in ALL* case- secn = %d", secn));
	  iss = DB_NOSUCHNAME;
	  break;
	}

      /* dbdata list management - store sec_offset returned */
      ((*dbdata_pp)->cur)++;
      DEBUGPRINT(DP_DEBUG,dbST0Flag,
				 ("dbST0list: unit cur count = %ld; max count = %ld \n", 
				  (*dbdata_pp)->cur,  (*dbdata_pp)->max ));
      if ((*dbdata_pp)->cur >  (*dbdata_pp)->max)
	{ 
	  /*(*dbdata_pp)->cur = orgCnt;*/
	  /* if list not large enough, double its size */
	  iss =  dbST0alloc( sizeof(int4u), 20 *(*dbdata_pp)->max, dbdata_pp);
	  if (iss== DB_BADDICT) 
	    return (iss);
	  /* 8/8 point to next element of possible new list returned by realloc */
	  next_p = (int4u *)*dbdata_pp + ( (*dbdata_pp)-> cur - 1) + 2;
      DEBUGPRINT(DP_INFO,dbST0Flag,
				 ("dbST0list: unit cur count after realloc= %ld; max count = %ld \n", 
				  (*dbdata_pp)->cur,  (*dbdata_pp)->max));
	}
      /* store the sptr from subblock in the list */
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0list - successful getting subblk for secn = %d, sec_offset = %ld\n", secn, sec_offset));
      *next_p++ = sec_offset;

      /* does nextc point to current ups offset ? */
      if (ups_offset == next_ups_offset) /* definitely, in case of single unit */
	{

          DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0list - Done - exiting!\n"));
	  break;
	}
      /* no, so advance to next ups in nextc link */
      ups_offset = next_ups_offset;
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t ups_offset = %ld\n", ups_offset));
      ups_ptr = (int2u *)(ups_offset + (int2u *)sb0 ); /*update w nextc*/
      memcpy (&next_ups_offset,(void *)(ups_ptr + OFF_UPS_NEXTC), 4);
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t next_ups_offset = %ld\n", next_ups_offset));
      	      
    } /* for */

  return (iss);
}

/*=============================================================================
  Name: dbST0GetUpsblk

  Abs:
      This routine will scan through the a list of hashed blocks in the      
      database searching for a block which  matches a particular Primary
      and Unit number. The head of the list is implied by specifying an
      entry in the hash table.

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------  
	const int2u         catn         R           see below
	const int2u         unit         R
	const suptype0_ts *const sb0     R

  Args: catn - A value which contains the Category (or Primary) number whose
               corresponding Unit/Primary block is to be found.
        unit - A value which contains the  Unit number whose
               corresponding Unit/Primary block is to be found.              
        sbo  - A pointer to the supertype "0" block in the database.

  Rem: 

  Side: 

  Ret:
       The routine will return a word offset to the matching UPS block. If
       a match is not found the routine will return NULL (0).

==============================================================================*/    
LOCAL unsigned long dbST0GetUpsBlk (const int2u catn, 
				    const int2u unit, 
				    const suptype0_ts *const sb0)
{
/*
  This union is used to represent the key in the UPS block in one of two
  ways:  The first as a two word array with the category number in the low
  order word and the unit in the high order word.  The second is a longword
  which may be used for longword comparisons.
*/
union key
    {
    unsigned short word[2];
    unsigned long  longword;
    };
union key  Tempkey;                             /* Temp. storage for key construction */
unsigned short *upsblk, i;                      /* Returned upsblk, scratch index */
 int4u         offset,                          /* Byte offset into database */
               cnt,
               key,                             /* Node key */
               upskey;                          /* ups key - catn unit combo */

 /* debug */
 int2u         prim, un;
/*
  Compute the hash value. This value is the offset into two tables: The
  first table contains the number of hash collisions and the second contains
  the offsets to the first hashed entry. Note: offset is byte aligned. 
*/
  i = ((catn & 0xFFFF) + (unit & 0xFFFF)) % sb0->hshl;
  cnt = sb0->hcTbl[i];                    /* Get # of blocks on the list */
  offset = sb0->hTbl[i];                  /* Get the initial offset */

/*
  Compose the key which is used to resolve duplicate entrys in the
  Hash table. The key is a longword value with the high order word
  the unit number and the low order word the category number.
*/
  Tempkey.word[0] = catn;
  Tempkey.word[1] = unit;
  key = Tempkey.longword;                 /* hash key */

/*
  Traverse the hash list, looking for a match.
*/
  for (i = 0; i < cnt; i++)
    {
    upsblk = (unsigned short *)(offset + (unsigned short *)sb0);   /* Compute pointer..*/
    memcpy(&upskey, (upsblk + OFF_UPS_CATN), 4); /* match on adjacent catn unit */
    /* debug:*/
    memcpy(&prim,(upsblk + OFF_UPS_CATN), 2);
    DEBUGPRINT(DP_DEBUG,dbST0Flag,("DBIO: getUpsblk catn read = 0x%x\n", prim));
    memcpy(&un,(upsblk + OFF_UPS_UNIT), 2);
    DEBUGPRINT(DP_DEBUG,dbST0Flag,("DBIO: getUpsblk unit read = 0x%x\n", un));
    DEBUGPRINT(DP_DEBUG,dbST0Flag,("DBIO: getUpsblk offset = 0x%lx\n", offset));
    /* end */  
    DEBUGPRINT(DP_DEBUG,dbST0Flag,("DBIO: getUpsblk processing key = 0x%lx; need key %lx\n", upskey, key));
    if(upskey == key)                           /* Found a match? */
      {
		/* debug */
		DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t ups offset = %ld\n", offset));
		DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t catn = %d\n", *(upsblk+OFF_UPS_CATN)));
        DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t unit = %d\n", *(upsblk+OFF_UPS_UNIT)));
        DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t\t nsub = %d\n", *(upsblk+OFF_UPS_NSUB)));
        /* end of debug */
        return(offset);                         /* Yes */
      }
    memcpy (&offset, (upsblk+OFF_UPS_FLINK), 4); /* No, copy in next offset */
    }

return(0);      
}

/*=============================================================================
  Name: dbST0GetSubBlk

  Abs: This routine will scan through the subtype blocks in a UPS block 
       searching for a subtype block which matches a passed "Secondary"
       identifier.

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------  
	const int2u          ups_offset   R        pointer to ups block 
	const unsigned char  secNum       R        secondary number to be found in ups
	const suptype0_ts *const sb0      R        pointer to superblock 0

  Rem: 

  Side: 

  Ret:  The routine will return an offset to the matching sub block. If
       a match is not found the routine will return NULL (zero) pointer.

==============================================================================*/   
LOCAL unsigned long dbST0GetSubBlk (const unsigned long ups_offset, 
				    const unsigned char secNum, 
				    const suptype0_ts *const sb0)
{
struct subblk
    {
    unsigned char  subn;                /* Subtype number */
    unsigned char  supn;                /* Object number */
    unsigned char  cnt;                 /* Data count */
    unsigned char  fmt;                 /* Data format type */
    unsigned long  offset;              /* Word offset to data */
    };

  int cnt, i;
  unsigned long sec_offset = 0;         /* offset to matching secondary */
  short * ups_ptr; 
  unsigned char subn;                   /* subn (secondary #) within subblk */
  /* debug : */
  char * subblk;
  struct subblk subblk_struc;

  DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0GetSubblk passed secn=%d\n",secNum));	      
  ups_ptr =(short *)(ups_offset + (short *)sb0);
  /* get cnt of subblocks in this ups */
  cnt = *(ups_ptr + OFF_UPS_NSUB);
  DEBUGPRINT(DP_DEBUG,dbST0Flag,
			 ("dbST0GetSubblk -  There are %d subblks in upsblk w offset = %ld\n", cnt, ups_offset));
      /*memcpy (&cnt, (void *)(ups_ptr + OFF_UPS_NSUB),2);*/
  /* secondary offset for storing in Clist */
  sec_offset = (ups_offset + SIZE_UPS);
  for (i=0;i<cnt;i++) {
	memcpy (&subn,(void *)(((short *)sb0)+ sec_offset), 1);
	DEBUGPRINT(DP_DEBUG,dbST0Flag,
			   ("dbST0GetSubblk comparing secNum %d with subblk secn=%d\n",secNum, subn));	      
	if (subn == secNum) {
	  subblk = (char *)( ((short *)sb0) + sec_offset );   
	  DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t\t subblk offset = %ld \n", sec_offset));
	  memcpy (&subblk_struc,(((short *)sb0) + sec_offset),sizeof(subblk_struc)); 
	  DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t\t\t subn = %d\n", subblk_struc.subn));
	  DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t\t\t supertype = %d\n", subblk_struc.supn));
	  DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t\t\t slen = %d\n", subblk_struc.cnt));
	  DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t\t\t fmt = %c\n", subblk_struc.fmt));
	  DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t\t\t sptr = %ld\n", subblk_struc.offset));
	  return (sec_offset);
	}
	sec_offset += SIZE_SUB;
  }
return (0);            /* List exhausted, no match */
}
/*=============================================================================
 
  Name: dbST0units

  Abs:  a function to interrogate the database and increment a given
        list with all unit numbers found for a given primary.

  
  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------  
	const int2u         prim         R           see below
	dbST0header_ts * *const dbdata_pp R/W


        prim   - A char pointer - containing four ASCII characters.
               These four characters define the database Primary whose
               pointers will be returned.
	dbdata_pp  - A pointer to a buffer into which the requested "pointers"
               will be returned. The buffer is formatted as a standard
               VMS database list. The buffer is assumed to be a word
               array, with the first element of the array containing its
               size (in words), and the second element containing the
               count (in words) of elements currently occupied in the
               buffer. It is the callers responsibility to inititalize
               this count to its proper value before calling this routine.
               A value of zero (0) indicates an empty buffer.

  Rem:  Validates incoming list      
        uses unit=ALL* to get all prim/unit/secn into list for this prim

  Side: The list must be passed as a pointer to a pointer so that it
        may be modified. 
        callocMustSucceed's list

 
  Ret:  vmsstat_t VMS style error condition code defined in dbdef.h:
        MICR_OKOK, DB_BADDICT

=============================================================================*/
LOCAL vmsstat_t  dbST0units (const int2u prim, 
			     dbST0header_ts **const dbdata_pp)
     /*
vmsstat_t  dbunits(dbST0header_ts **const dbdata_pp, 
const char * const prim)*/   /*aka catn */

{
  vmsstat_t iss = MICR_OKOK;
  int4u  orgCnt =0;               /* original dbdata list length of elements */
  int4u  maxCnt =0;               /* max elements in dbdata list */
  int4u  *next_p;                 /* next free element location in list */
                                 
  int2u  unit = LISTHEAD;
  suptype0_ts    *sb0;            /* pointer to supertype 0 header */
  int4u          ups_offset;      /* word offset to ups blocks within sb0 */
  int4u          next_ups_offset; /* when looping through ups blocks */
  int2u          *ups_ptr;        /* pointer to upsblk (adds ups_offset + sb0) */

  /* validate incoming Clist pointers & list */
  if (!SUCCESS(dbST0ChkClist (&orgCnt, &maxCnt, dbdata_pp) ))
    {
      DEBUGPRINT(DP_ERROR,dbST0Flag,("dbST0units: INVALID dbdata_pp"));
      return (iss);
    }

  /* Set up some list counters and pointers. maxCnt contains max number of
     elements which may be entered into the buffer (less the header),orgCnt contains
     the number of elements used in the buffer before this routine was called, and
     next points to the next available buffer location. */
  next_p = (int4u *)*dbdata_pp + 1;        /* point to count field of list */
  orgCnt = *next_p++;             /* store orgCnt; advance to next element in buffer */
  
  /* init pointers */
  sb0 = (suptype0_ts *)dbnodes_a[0];

  /*
   Establish the set of UPS blocks to be searched and point to the
   first block in the set. Use the special
   value "LISTHEAD" to find the first block in the set.
  */

  if (!(ups_offset = dbST0GetUpsBlk(prim, unit, sb0)))
    {
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0units: no units for primary = %d\n", prim));
	    iss = DB_NOSUCHUNIT;
            return(iss);
    }
  else /* found listhead for catn; ALL* units;  now use nextc links 
            to find all units in this cat */
    { 
      ups_ptr = (int2u *)(ups_offset + (int2u *)sb0 );
      /* proceed using nextc link for 'ALL*' */
      memcpy (&ups_offset, (void *)(ups_ptr + OFF_UPS_NEXTC), 4);
      /* establish next_ups_offset using nextc again  */
      ups_ptr = (int2u *)(ups_offset + (int2u *)sb0 );
      memcpy (&next_ups_offset,(void *)(ups_ptr + OFF_UPS_NEXTC), 4);  
      
	
    }/* found listhead for all units */

  for (;;)
  /* loop over ups blks following nextc pointer until ups blk points to itself */
    {
      /* dbdata list management - update count & store unit */
      ((*dbdata_pp)->cur)++;
      DEBUGPRINT(DP_DEBUG,dbST0Flag,
				 ("dbST0units: unit cur count = %ld; max count = %ld \n", 
				  (*dbdata_pp)->cur,  (*dbdata_pp)->max ));
      if ((*dbdata_pp)->cur >  (*dbdata_pp)->max )
	{ 
	  /*  (*dbdata_pp)->cur = orgCnt; Leave as is 8/8*/
	  /* if list not large enough, double its size */
	  iss =  dbST0alloc(sizeof(int4u),20* (*dbdata_pp)->max, dbdata_pp);
	  if (iss== DB_BADDICT) 
	    return (iss);
	  /* 8/8 point to possible new list returned by realloc */
	  next_p = (int4u *)*dbdata_pp + ( (*dbdata_pp)-> cur - 1) + 2;
      DEBUGPRINT(DP_INFO,dbST0Flag,
				 ("dbST0units: unit cur count after realloc= %ld; max count = %ld \n", 
				  (*dbdata_pp)->cur,  (*dbdata_pp)->max));
	}
      /* store the unit number in the list */
      *next_p++ = *(ups_ptr + OFF_UPS_UNIT);
      /* debug */
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t dbST0units found prim=%d, unit %d \n",prim,*(ups_ptr + OFF_UPS_UNIT)));

      /* does nextc point to current ups offset ? */
      if (ups_offset == next_ups_offset) /* definitely, in case of one unit */
	{
          DEBUGPRINT(DP_DEBUG,dbST0Flag,("dbST0units - dbunits is done - exiting!\n"));
	  break;
	}
      /* no, so advance to next ups in nextc link */
      ups_offset = next_ups_offset;
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t ups_offset = %ld\n", ups_offset));
      ups_ptr = (int2u *)(ups_offset + (int2u *)sb0 ); /*update w nextc*/
      memcpy (&next_ups_offset,(void *)(ups_ptr + OFF_UPS_NEXTC), 4);
      DEBUGPRINT(DP_DEBUG,dbST0Flag,("\t next_ups_offset = %ld\n", next_ups_offset));
      	      
    } /* for */
  /* adjust count */
  /* ((*dbdata_pp)->cur)--;*/
  return (iss);
}
/* ============================================================================= 
   Name: dbST0ChkClist 

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------  
	int4u *             orgCnt         R       original # of elements in list         
        int4u *             maxCnt         R       max # of elements in list
	dbST0header_ts * *const dbdata_pp R/W      ptr to ptr of units data list

   Rem:   

   Side:  

   Ret:  vmsstat_t 
         MICR_OKOK, DB_BADDICT

 ==============================================================================*/
LOCAL vmsstat_t dbST0ChkClist (int4u * orgCnt_p,    /* original # elements */
							   int4u * maxCnt_p,    /* max # elements */
                               dbST0header_ts * * const dbdata_pp)
{ 
#define INIT_LIST 40
   vmsstat_t        iss = DB_BADDICT; 

   /* check for valid pointer to pointer (typed thus in case pointer changes) */
   if (dbdata_pp == NULL) 
   { 
     slcCmlogLogMsgFile(iss, "BADDICT",__FILE__,__LINE__,
			"dbST0ChkClist - invalid pointer to pointer passed in.\n");
     return (iss); 
   } 
   /* remember orig list length in case of list too short; in that case orig */
  /*  list length is returned. */
  *orgCnt_p = (*dbdata_pp)-> cur;
  *maxCnt_p = (*dbdata_pp)->max;    /* max elements less 2 element header */
  /* maxCnt contains max number of
     elements which may be entered into the buffer (less the header),
     orgCnt contains the number of elements currently  in use in the buffer */

  /* check for valid list pointer; if NULL, allocate initial block from heap */
  if (*dbdata_pp == NULL)
  {
    iss = dbST0alloc(sizeof(int4u), INIT_LIST, dbdata_pp );
    if (iss== DB_BADDICT)
      return (iss);
  } 

  /* validate CLIST */
  /* if full, realloc with twice as many elements ? */
  if ( (*maxCnt_p > 0x8001) || (*maxCnt_p==*orgCnt_p))
    {
      /* error */
      slcCmlogLogMsgFile(iss, "BADDICT",__FILE__,__LINE__,
		       "dbST0ChkClist - Invalid list; max or cnt range error.\n");
      DEBUGPRINT(DP_ERROR,dbST0Flag,("DBIO: Invalid list; max or cnt range error\n"));
      return (iss);
    } 

  return (iss);
}

/*=============================================================================

  Abs:  DEBUG routine to get a dblist pointer for testing

  Name: getdblist

  Args: void

  Rem:  

  Side: NOTE: 

  Ret:  vmsstat_t iss 

==============================================================================*/
LOCAL vmsstat_t dbST0GetDebugdblist (void)
{
#define XCOR_XCOR 2
#define MAG_IMMO 24

/* primary/secondary numbers */
/* #include     "magnet_dbs.hm"*/
  vmsstat_t iss = DB_OKOK;
  dbST0header_ts  **dbdata_p = NULL; /* pointer to pointer to dbST0header_ts */

  if (SUCCESS (iss= dbST0alloc ( 4, 40, dbdata_p)))
    {

    dbST0list (XCOR_XCOR, 0xffff, MAG_IMMO, dbdata_p );
    /* proc key 0x200ffff; need key 2ffff*/
    dbST0printlist (*dbdata_p);
    if (*dbdata_p)free (*dbdata_p);
    }
  DEBUGPRINT(DP_DEBUG,dbST0Flag,("DBST0: getDebugdblist complete\n"));
  return (iss);
}

/*=============================================================================

  Abs:  Debug routine to print the db subblk

  Name: dbST0printlist

  Args: super type header block

  Rem:  

  Side: NOTE: 

  Ret:  void

==============================================================================*/
void dbST0printlist (dbST0header_ts *const dbdata_pp)
{
  dbsuptype_tu *sb0;
  char      *subblk;
  int4u      sec_offset;
  int4u      s, cnt;
  int4u      *next_p;    /* next free element location in list */
  int4u      lint;    

  /* init pointers, offset, vars */
  sb0 = dbnodes_a[0];
  cnt =  dbdata_pp->cur;
  next_p = (int4u *)dbdata_pp + 2; /* first element location in buffer */

  DEBUGPRINT(DP_INFO,dbST0Flag,("dbST0:printdblist: Max cnt = %ld; Cur cnt = %ld\n",
	 dbdata_pp->max, cnt));  

  /* Process each subblk  */
  for (s=0; s<cnt; s++)
    {
      sec_offset = *next_p++;
      subblk = (char *)( (int2u *)sb0 + sec_offset );   
      DEBUGPRINT(DP_INFO,dbST0Flag,("\t\t in dblist, ST0 sec_offset =%ld \n", sec_offset));  
      DEBUGPRINT(DP_INFO,dbST0Flag,("\t\t\t subn = %d\n", (unsigned char) *subblk));
      DEBUGPRINT(DP_INFO,dbST0Flag,("\t\t\t supertype = %d\n", (char) *(subblk+OFF_SUB_SUPN_B)));
      DEBUGPRINT(DP_INFO,dbST0Flag,("\t\t\t slen = %d\n", (unsigned char) *(subblk+OFF_SUB_SLEN_B)));
      DEBUGPRINT(DP_INFO,dbST0Flag,("\t\t\t fmt = %c\n", (char) *(subblk+OFF_SUB_FMT_B)));
      memcpy(&lint,subblk+OFF_SUB_SPTR_B,4);
	  DEBUGPRINT(DP_INFO,dbST0Flag,("\t\t\t sptr = %ld\n",lint));	
    }
}

/***************************************************************************/
/*  Header documentation extracted from dblistu.a38: */
/*  The large-scale structure of the database is:
 
   SUPERTYPE_PTR(2,0:4)
           |
           v                               +-------------+
   +----------------+                      | supertype 0 |
   |     (,0)     o-+--------------------->|    /DB/     |
   +----------------+                      +-------------+
   |     (,1)     o-+-------------------+
   +----------------+                   |  +-------------+
   |     (,2)     o-+----------------+  +->| supertype 1 |
   +----------------+                |     |             |
   |     (,3)     o-+-------------+  |     +-------------+
   +----------------+             |  |
   |     (,4)     o-+----------+  |  |     +-------------+
   +----------------+          |  |  +---->| supertype 2 |
                               |  |        |             |
                               |  |        +-------------+
                               |  |
                               |  |        +-------------+
                               |  +------->| supertype 3 |
                               |           |             |
                               |           +-------------+
                               |
                               |           +-------------+
                               +---------->| supertype 4 |
                                           |             |
                                           +-------------+
 
   Supertype 0 contains all information (nodes) describing the
    structure and locations of all data items in the database.
   Supertype 1 contains all constant data (changeable only by DBEDIT).
   Supertype 2 contains all data intended to be transmitted from the
    Vax to the micro.
   Supertype 3 contains all data intended to be transmitted from the
    micro to the Vax.
   Supertype 4 contains Vax-only data, and is not represented in the
    micro.
 
 -----------------------------------------------------------------------
 
   Common block /DBNODES/ contains five supertype base pointers (each
   supertype is one separate segment)   note that we assume that the
   offset portion of each pointer is zero:
 
      INTEGER*4 SUPERTYPE_PTR(2,0:4)
      COMMON /DBNODES/ SUPERTYPE_PTR
 
 -----------------------------------------------------------------------
 
  The format of CLIST is:
 
    0   1   2   3   4   5   6   7   8   9  10   11
  +---+---+--+----+---+---+---+---+---+---+---+---+--- ... ---+---+---+
  |  len1 | len2  |      X1       |      X2       |          Xn       |
  +---+---+--+----+---+---+---+---+---+---+---+---+--- ... ---+---+---+
     (1)     (2)     (3)     (4)     (5)     (6)               (2n+2)
 
  where
 
        len1    is the total max length of CLIST in words.
        len2    is the number of occupied words following len2 in CLIST.
                Len2 must be even.
        X1,X2,...,Xn are unbiased word indices of nodes in superblock 0,
                where n = len2/2.
 
 -----------------------------------------------------------------------
 
   Structure of beginning of supertype 0 block:
 
                +--------+--------+
              1 |     network     |  0
                +--------+--------+
              3 |     netlsn      |  2
                +--------+--------+
              5 |       id        |  4
                +--------+--------+
              7 |                 |  6
                +       len       +
              9 |                 |  8
                +--------+--------+
             11 |                 | 10
                +      micr       +
             13 |                 | 12
                +--------+--------+
             15 |                 | 14
                +     blkbeg      +
             17 |                 | 16
                +--------+--------+
             19 |                 | 18
                +     segbeg      +
             21 |                 | 20
                +--------+--------+
             23 |      hshl       | 22
                +--------+--------+
                |  Array of ptrs  | 24
                |  to listheads.  |
                |                 |
                +--------+--------+
                | Array of counts | 24+hshl*4
                |  of nodes in    |
                |  lists.         |
                +--------+--------+
                | Beginning of    | 24+hshl*6
                | heap storage    |
                | for nodes...    |
 
 
SUPBLK  STRUC
 SP_NET_WORK DW  ?
 SP_NET_LSN  DW  ?
 SP_ID       DW  ?          ; Superblock id #.
 SP_LEN      DD  ?          ; Length in words of entire superblock.
 SP_MICR     DD  ?          ; Micro name.
 SP_BLK_BEG  DD  ?          ; Beginning of actual data (after header)
                            ;  within transmitted block.
 SP_SEG_BEG  DD  ?          ; Beginning source/dest within superblock
                            ;  of transmitted data.
 SP_HSHL     DW  ?          ; Hash modulus = size of array of hash ptrs.
;SP_HSHPT    DD  ? DUP (?)  ; Beginning of two arrays:  hash ptrs and
;SP_HSHCC    DW  ? DUP (?)  ;  collision counts.
SUPBLK  ENDS

 -----------------------------------------------------------------------
 
   Structure of a primary unit node (aka "UPS block"):
 
                +--------+--------+
              1 |                 |  0
                +      nxth       +
              3 |                 |  2
                +--------+--------+
              5 |                 |  4
                +      nxtp       +
              7 |                 |  6
                +--------+--------+
              9 |                 |  8
                +      nxtc       +
             11 |                 | 10
                +--------+--------+
             13 |                 | 12
                +      nlen       +
             15 |                 | 14
                +--------+--------+
             17 |      catn       | 16
                +--------+--------+
             19 |      unit       | 18
                +--------+--------+
             21 |      nsub       | 20
                +--------+--------+
 
UPSBLK  STRUC
 SP_NXTH     DD  ?            Next CATN,UNIT in same hash residue.
 SP_NXTP     DD  ?            Next CATN's first unit.
 SP_NXTC     DD  ?            Next unit of same CATN.
 SP_NLEN     DD  ?            Size of UPS block + its sub-blocks.
 SP_CATN     DW  ?            CATN id number.
 SP_UNIT     DW  ?            UNIT number (adjacent to CATN).
 SP_NSUB     DW  ?            Number of secondaries (subn's).
UPSBLK  ENDS
 
 -----------------------------------------------------------------------
 
   Structure of a secondary node (aka "sub-block" of some UPS block):
 
                +--------+--------+
              1 |  supn  |  subn  | 0
                +--------+--------+
              3 |  fmt   |  slen  | 2
                +--------+--------+
              5 |                 | 4
                +      sptr       +
              7 |                 | 6
                +--------+--------+

SUBBLK  STRUC
 SP_SUBN     DB  ?          ; Subtype (secondary) number.
 SP_SUPN     DB  ?          ; Supertype number.
 SP_SLEN     DB  ?          ; Wordcount of subtype data in supertype.
 SP_FMT      DB  ?          ; Format indicator (I,A,R,Z,S,T).
 SP_SPTR     DD  ?          ; Subscript (unbiased word offset) into
SUBBLK  ENDS                ;  supertype block.
--------------------------------------------------------------------------*/
