/*=============================================================================
 
  Name: dbSend.c
        dbSendThread - main thread entry point
        dbDownladMe
        dbWriteDbs
        dbAck 
        dbDiagReply
        dbSendCleanup    - cleans up the globals for this thread before shutdown

  Abs:  Waits at message queue to receive inter-thread messages.
        Sends "database service" (DBS) type messages to DBEX port 
        on proxy server for routing to VMS Control System's DBEX process.

        Reference web pages on LCLS Controls | Facilities |SLC-Aware IOC | :
                        Database Service (DBS) Design Document 
                        Functional Requirements | SLC Database Service (DBS)
                        Database Input/Output Utilities
                        C Coding Standard

        Reference RMX_DBS dbmain.c 

  Auth: 4 AUG 2004, drogind created using msgSend.c as a template
  Rev: 14 DEC-2004, drogind modified prototype code to production standards
 
-----------------------------------------------------------------------------*/
#include "copyright_SLAC.h"    
/*----------------------------------------------------------------------------- 
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
int dbSendFlag = DP_ERROR;
#endif
/* c includes */
#include     <stdio.h>
#include     <stdlib.h>    /* malloc, free */
#include     <string.h>    /* memcpy */

/* slc global vars */
#include     "slcGlob.h"

/* VMS Control system function/status codes  */
#include     "msgfunc.hc"  /* FUNC_MSG function code */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */

/* slc common includes */
#include     "tcpUtil.h"   /* tcpGetConnection, tcpGeBuffer */
#include     "msgQUtil.h"  /* message queue utilities */
#include     "cvtUtil.h"   /* slc conversion utilities */
#include     "timeUtil.h"  /* slc time conversion utilities */
#include     "slcCmlog.h"  /* slcCmlogLogMsg */
#include     "slcUtil.h"   /* slcGetIOCName */ 
#include     "slcMsg.h"    /* DB_ACK_DBEX... */

/* dbs includes  */  
#include     "db_micro_struc.h"
#include     "dbUtil.h" 
#include     "dbmsgmail.h" 
#include     "dbGlob.h"  
#include     "dbdef.hc" 
#include     "dbSendCtl.h"

/* epics osi includes */
#include     "cantProceed.h"
#include     "dbSend.h"

/*****************************************************************************/
#define  MAX_RECEIPT_LOOP 15
/*****************************************************************************/
/* local file scope data declarations and function prototypes */

static dbsndrcvbuf_ts  *dbsendbuf_ps;       /* send buffer */
static dbmsgmail_ts    dbSendmsg_s;          /* slc message buffer */
static unsigned char DBEXACK_count=0;  /* Tag to validate return receipt      */ 

/* diagnostics - global extern in dbGlob.h */
dbSendDiags_ts dbSendDiags_s = {0,0,0,0,0,{0,0}, {0,0}, 0.0, 0.0, 0, 0, 0.0, 0.0};

/* internal function prototypes */

LOCAL vmsstat_t dbDownloadMe(void);
LOCAL vmsstat_t dbAck (dbsuptypehdr_ts *suphdr_ps);
LOCAL vmsstat_t dbDiagReply (void);
LOCAL vmsstat_t dbSendUpdates(const slcJob_te slcJobId_e);
LOCAL epicsBoolean dbexSockCheck(void);
LOCAL vmsstat_t dbSendUpdateAll(void);
/*=============================================================================
 
  Name: dbSendThread

  Abs:  This thread sends out all "database service" (DBS) type tcp/ip messages
        to the VMS Control System's DBEX process via the proxy server. 

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
        none
 
  Rem:  This thread waits at its message queue and acts upon messages. 
        All messages sent from dbRecv thread act upon incoming data by:
        converting native format to VMS, and then sending to Alpha DBEX process
        via dbex socket on proxy.
		For ST2 or ST3 update messages to VMS, dbSend waits at the CTL port to
		receive an ack. dbSend manages the CTL port.( dbRecv manages DBEX port).
		This thread sets the slcThreads_as[slcDbSend].active flag when it starts, 
        indicating to slcExec that the thread is alive. 
		The Exec thread (or the slcStop command) sets the sclThreads[slcDbSend].stop 
        flag to false when it wants to shut down the Send thread.

  Side: created by the slcExecThread  
 
  Ret:  none

All possible vmsstat_t error messages that originate from dbSendThread:
 DB_CANTSEND,  
=============================================================================*/

/**procedure**/
epicsShareFunc void epicsShareAPI dbSendThread(void)
{
  vmsstat_t            iss; 
  vmsstat_t            isss; /* for tcp send error codes */

  /* create the message queue id */
  if ( !SUCCESS(iss = messageQCreate(dbSend, 20, MAX_MSG_BYTES)) ) {
    dbSendCleanup();
	return;
  }

  /* Allocate send buffer, */
  dbsendbuf_ps = callocMustSucceed(1, sizeof(dbsndrcvbuf_ts), "dbSend cannot calloc dbsendbuf\n");
  /* main loop of the dbSendmsg_s Send Task, wait at message queue for outgoing 
	 messages, then send message on to proxy server
  */

  DEBUGPRINT(DP_INFO,dbSendFlag,("DB SEND: becoming active\n"));
  slcThreads_as[dbSend].active = epicsTrue;

  /* make a CTL connection - do this after setting active */
  dbCtlSockConnect();

  while(slcThreads_as[dbSend].active)
  { 
    /* now begin receiving messages, until connection fails, or we're told to stop */
	
	/* read a message at the message queue.    */
	memset ( &dbSendmsg_s, 0, sizeof(dbSendmsg_s) );
	DEBUGPRINT(DP_DEBUG,dbSendFlag, ("DB SEND: Waiting at mbx\n"));
	if (dbsendAllUpdates){
	  dbsendAllUpdates = epicsFalse;
	  dbSendUpdateAll();
	}
	isss = MICR_OKOK;
        if (SUCCESS(iss = messageQRecv(dbSend, sizeof(dbmsgmail_ts),NULL, (void *)&dbSendmsg_s)) )
      {

		switch (dbSendmsg_s.func) 
		  {
		  case IOC_STOP: /* if this is a STOP message, stop */
			DEBUGPRINT(DP_DEBUG,dbSendFlag,("DB SEND:got a IOC_STOP message\n"));
			slcThreads_as[dbSend].active = epicsFalse;
			break;
			
		  case DB_DOWNLOAD_REQ: /* IPL- request download */
			DEBUGPRINT(DP_DEBUG,dbSendFlag,("DB SEND: Rec'ed download message - request download!\n"));
			/* if proxy is down when we started, then comes up, we've missed */
			/* the opp. to get the db and must be restarted */
			isss = dbDownloadMe();
			break;
			
		  case DB_ACK_DBEX: /* dbex req'ed ack for download chunk, or latest ST2; so send it now */
			DEBUGPRINT(DP_DEBUG,dbSendFlag,("DB SEND: Rec'ed Ack Request from dbex \n"));
			isss = dbAck(&dbSendmsg_s.nativeSB);
			break;

		  case DB_DIAG_DBEX : /* diagnostic reply back to dbex */
			DEBUGPRINT(DP_DEBUG,dbSendFlag,("DB SEND: Rec'ed diag request from dbex \n"));
			isss=dbDiagReply();
			break;

		  case DB_UPDATE:  /* ST data */
			DEBUGPRINT(DP_DEBUG,dbSendFlag,
				   ("DB SEND: Rec'ed dbupdate req from job enum %d\n",dbSendmsg_s.jobId ));
		    isss=dbSendUpdates((const slcJob_te)dbSendmsg_s.jobId );
			break;
		  case DB_UPDATE_ALL:  /* all jobs ST data */
			DEBUGPRINT(DP_DEBUG,dbSendFlag,
				   ("DB SEND: Rec'ed dbupdate ALL req from\n" ));
		    isss=dbSendUpdateAll();
			break;
		  default:  /* if not inter-process msg, this is a mail message to be sent to Proxy ??? */ 
			DEBUGPRINT(DP_ERROR,dbSendFlag,("DB SEND: unknown msg at queue!\n"));
			break;
			
		  } /*switch*/
      } /* if success */
	while( (!SUCCESS(isss)) && (!slcThreads_as[dbSend].stop) ) {
	  if ( dbexSockCheck() ){
		isss = MICR_OKOK;
	  }
	}	
	/* check the stop flag */
	if (slcThreads_as[dbSend].stop)
	  slcThreads_as[dbSend].active = epicsFalse;		
  } /* while active */

  /* active not set, time to shut down */
  DEBUGPRINT(DP_INFO,dbSendFlag,("DB SEND: shutdown\n"));
  dbSendCleanup();
}


/*===========================================================

  Name: dbexSockCheck

  Abs:  Checks integrity of dbex socket connection and
        waits forever, in increments of wait_time, for
        dbRecv to connect or to stop.

  Args: none

  Rem:  returns if this thread is told to stop
        
  Side: 

  Ret:  iss - the status code
===========================================================*/
LOCAL epicsBoolean dbexSockCheck(void)
{
  float          wait_time = 1.0;
  unsigned long  elapsedCount=0;
  epicsBoolean   check;
    
  check=tcpSockCheck(&slcSockets_as[sdDbex], dbSend, dbRecv, wait_time, &elapsedCount);
  if (elapsedCount>0) {
	if ( (slcGetDbExists()) && (slcThreads_as[dbSend].active)){
	  DEBUGPRINT(DP_INFO,dbSendFlag,("DB SEND: dbexSockCheck setting dbsendAllUpdates flag!\n"));
	  dbsendAllUpdates = epicsTrue;
	  dbSendDiags_s.dbexLastDisconnTime = (elapsedCount * wait_time);
	}
  }
  return (check);
}
/*===========================================================

  Name: dbSendCleanup()

  Abs:  cleans up the globals for this thread before shutdown

  Args: NONE

  Rem: clean up the slcThreads_as struct for this thread
       and release any resources 

  Side: Modifies slcThreads_as[dbSEnd]; 
        frees malloc'ed memory for dbsendbuf_ps

  Ret:  NONE

===========================================================*/
void dbSendCleanup(void)
{
  DEBUGPRINT(DP_INFO,dbSendFlag,("DB SEND: cleanup..\n"));
  if (slcThreads_as[dbSend].mqId_ps != NULL)
	 epicsMessageQueueDestroy(slcThreads_as[dbSend].mqId_ps);
  slcThreads_as[dbSend].mqId_ps = NULL;
  slcThreads_as[dbSend].tid_ps = NO_THREAD;

  /* free send buffer */
  if (dbsendbuf_ps) {
    free (dbsendbuf_ps);
	dbsendbuf_ps = NULL;
  }

  tcpSockClose(&slcSockets_as[sdCtl]);

  DEBUGPRINT(DP_INFO,dbSendFlag,("DB SEND: shutdown\n"));
}

/*=============================================================================

  Name: dbAck

  Abs:  dbAck sends an ack to Alpha's dbex process

  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        dbsuptypehdr_ts *      nativeSB_ps       R W     pointer to native superheader
                                                           sent by dbRecv

  Rem:  This procedure handles all acks requested by dbex, sent from dbSend, upon:
        1) ST0-3 IPL downloaded database pieces
        2) ST2 updates sent to slc ioc from dbex.
        An ST2 update was previously received via the dbRecv task. It then
	formulates an ack msg and sends to dbSend's mbx.

  Side: NOTE: 

  Ret:  vmstat_t

==============================================================================*/
/**procedure**/
LOCAL vmsstat_t dbAck (dbsuptypehdr_ts *nativeSB_ps)
{
  unsigned long  bytes_to_send; 
  dbrcvackbuf_ts read_ack_s; 
  vmsstat_t      iss = MICR_EXNET_SENDFAIL;
  unsigned long  bytecnt = 0;

  read_ack_s.pxyhdr.ip_port_u.word  = *(long *) v018_c4;
  read_ack_s.pxyhdr.cmd = PX_FORWARD_ALIAS_FUNC;
  read_ack_s.pxyhdr.user.byte0 = ++DBEXACK_count;
  read_ack_s.pxyhdr.len = htonl(sizeof(dbsuptypehdr_ts)-2);
  read_ack_s.pxyhdr.crc = 0x55;

  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("DBSend: Incoming id to mask for Ack = %x\n", nativeSB_ps->id));
  /* end of moved block */

  nativeSB_ps->id = SUPERHDR_FUNC_ACK | SUPERTYPE_ID_OVERRIDE |(nativeSB_ps->id & SUPERTYPE_ID_IPLDWN);
  nativeSB_ps->micr = dblongmicro;
 
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("DBSend: Sending this Ack id = %x\n", nativeSB_ps->id));
  bytecnt = dbhtoVMSSupHdr(nativeSB_ps, (dbsuptype_tu *)&read_ack_s.suphdr); 

  
  bytes_to_send = sizeof(read_ack_s)-2;
  dbexSockCheck();
  if (SUCCESS(iss= tcpSendBuffer(&slcSockets_as[sdDbex], bytes_to_send, (void *) &read_ack_s)))
  {
    iss = MICR_OKOK;
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("DB SEND: Sending Ack to dbex\n"));
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("DB SEND: PXYHDR: \n"));
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t read_ack_s.pxyhdr.ip_port_u.s.ip = 0x%x\n", read_ack_s.pxyhdr.ip_port_u.s.ip));
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t read_ack_s.pxyhdr.ip_port_u.s.port = 0x%x\n", read_ack_s.pxyhdr.ip_port_u.s.port));
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t read_ack_s.len = 0x%lx\n", read_ack_s.pxyhdr.len));
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t read_ack_s.pxyhdr.user.byte0 = 0x%x\n", read_ack_s.pxyhdr.user.byte0));
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t read_ack_s.pxyhdr.user.byte1 = 0x%x\n", read_ack_s.pxyhdr.user.byte1));
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t read_ack_s.pxyhdr.cmd = 0x%x\n", read_ack_s.pxyhdr.cmd));
    DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t read_ack_s.pxyhdr.crc = 0x%x\n", read_ack_s.pxyhdr.crc));

    dbprintVMSSupHdr ("DB SEND: VMS SupHdr Ack msg: \n", (dbsuptype_tu *)&read_ack_s.suphdr, sizeof(read_ack_s));
  }
  return iss; 
}
/*===========================================================

  Name: dbDiagReply

  Abs:  dbDiagReply sends a diagnostic reply to Alpha's DBEX process
          after dbRecv had received the request

  Args: NONE

  Rem:  This procedure handles all diagnostic messages requested by dbex

  Side: Writes to file-scoped variables 
          dbdiag_s
          dbSendmsg_s

  Ret:  vmsstat_t MICRO_OKOK or  MICR_EXNET_SENDFAIL

===========================================================*/
/**procedure**/
LOCAL vmsstat_t dbDiagReply (void)
{ 
  unsigned short      bytes_to_send;
  unsigned short      id;
  dbsndrcvdiagmsg_ts  diag_rply_s;  
  vmsstat_t           iss =  MICR_OKOK;


  /* If we were just booted, then our msgc_snd has 0x80000000 bit set, and */
  /* we tell DBEX;  if DBEX was just restarted, then his dbex_msgc_snd has */
  /* 0x80000000 bit set, and we believe him.  DBEX does comparisons and    */
  /* err_send's;  we dont.
                                                  */
  /* adjust incoming dbex buffer copy & local var */   
  dbSendmsg_s.micr_msgc_snd = dbdiag_s.msgc_snd; /*stuff micr_msgc_snd*/
  if (dbdiag_s.msgc_snd & JUST_BOOTED_BIT || dbSendmsg_s.dbex_msgc_snd & JUST_BOOTED_BIT)
  {
    dbdiag_s.msgc_snd &= ~JUST_BOOTED_BIT;
    dbdiag_s.msgc_rcv = dbSendmsg_s.dbex_msgc_snd & ~JUST_BOOTED_BIT;
  }
  dbSendmsg_s.micr_msgc_rcv = dbdiag_s.msgc_rcv;
 
  /* start filling diag_rply */
  memset(&diag_rply_s, 0, sizeof(diag_rply_s));

  id  = SUPERTYPE_ID_OVERRIDE |  SUPERHDR_FUNC_DGRPY;  
  /* swap id to VMS from native; fill in micr field  */    
  cvtSSwapMove ((const void *)&id, (void *)&(diag_rply_s.diagdata.data.suphdr.id));
  memcpy(&diag_rply_s.diagdata.datw[5], &dblongmicro, 4); 

  /* pack & byte swap data part (temp buffer) back to VMS */
  cvtLSwapMove((const void *)&dbSendmsg_s.dbex_msgc_snd, (void *)&(diag_rply_s.diagdata.datw[11]));
  cvtLSwapMove((const void *)&dbSendmsg_s.dbex_msgc_rcv, (void *)&(diag_rply_s.diagdata.datw[13]));
  cvtLSwapMove((const void *)&dbSendmsg_s.micr_msgc_snd, (void *)&(diag_rply_s.diagdata.datw[15]));
  cvtLSwapMove((const void *)&dbSendmsg_s.micr_msgc_rcv, (void *)&(diag_rply_s.diagdata.datw[17]));

  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("DB SEND: Sending diag reply\n"));

  bytes_to_send = sizeof(dbsndrcvdiagmsg_ts);

  if (!SUCCESS(iss = dbWriteDbs(slcSockets_as[sdDbex].sd,
								id,
								&bytes_to_send, 
								(dbsndrcvbuf_ts *)&diag_rply_s )))
  { 
    iss = MICR_EXNET_SENDFAIL;
    /*         err_send0(&iss);*/
    DEBUGPRINT(DP_ERROR,dbSendFlag, ("DB SEND: error sending diag reply\n"));
  }
  
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("DB SEND: Sending DIAG REPLY to dbex\n"));
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("DB SEND: PXYHDR: \n"));
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t diag_rply_s.pxyhdr.ip_port_u.s.ip = 0x%x\n", diag_rply_s.pxyhdr.ip_port_u.s.ip));
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t diag_rply_s.pxyhdr.ip_port_u.s.port = 0x%x\n", diag_rply_s.pxyhdr.ip_port_u.s.port));
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t diag_rply_s.len = 0x%lx\n", diag_rply_s.pxyhdr.len));
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t diag_rply_s.pxyhdr.user.byte0 = 0x%x\n", diag_rply_s.pxyhdr.user.byte0));
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t diag_rply_s.pxyhdr.user.byte1 = 0x%x\n", diag_rply_s.pxyhdr.user.byte1));
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t diag_rply_s.pxyhdr.cmd = 0x%x\n", diag_rply_s.pxyhdr.cmd));
  DEBUGPRINT(DP_DEBUG,dbSendFlag, ("\t diag_rply_s.pxyhdr.crc = 0x%x\n", diag_rply_s.pxyhdr.crc));
  
  /*dbprintVMSSupHdr ("DB SEND: VMS Diag Reply msg: \n", (dbsuptype_tu *)&diag_rply_s.diagdata.datw[0], sizeof(diag_rply_s));*/
  return iss; 
}


/*=============================================================================

  Name: dbDownloadMe
  
  Abs:  send a "download me" request to DBEX process in Alpha

  Args: None

  Rem:  requests a download from DBEX by formatting the correct supertype
        header id field and sending via tcp/ip

  Scop: writes to file-scoped global variable
        dbsendbuf_ps                                                   

  Ret:  vmsstat_t returned from dbWriteDbs()
==============================================================================*/
/**procedure**/
LOCAL vmsstat_t dbDownloadMe(void)
{   
  unsigned short dbmsgbc, temp;
  vmsstat_t      iss;
  unsigned short id;
  dbsuptype_tu  supblk;
  

  dbmsgbc = sizeof(dbsendbuf_ps->supblk.sbhdr);
  memset(&dbsendbuf_ps->supblk.sbhdr, 0, sizeof(dbsendbuf_ps->supblk.sbhdr));
  id = SUPERTYPE_ID_OVERRIDE | SUPERHDR_FUNC_BTDN;

  dbsendbuf_ps->supblk.sbhdr.id =  id;
  /*memcpy(dbsendbuf_ps->supblk.nativeSB.micr, micr, 4);*/
  dbsendbuf_ps->supblk.sbhdr.micr = dblongmicro;

  dbprinthSupHdr ("dbDownloadMe host formatted super type header - ",
		   &dbsendbuf_ps->supblk.sbhdr, 
		   dbmsgbc);
  /* format download req in VMS super type format */
  memset(&supblk,0,dbmsgbc);
 
  temp = dbhtoVMSSupHdr(&dbsendbuf_ps->supblk.sbhdr, &supblk); 
  
  memcpy(&dbsendbuf_ps->supblk.datw[0], &supblk, dbmsgbc);

  dbprintVMSSupHdr ("dbDownloadMe VMS-formatted super type header - ",
		   &(dbsendbuf_ps->supblk), 
		   dbmsgbc); 
  /*     
  if (dbmstate_s.tmoutf == 0)
  {
    dbmstate_s.tmoutf = 1;
    rqcreatetask(190, DB_timeout_task, (selector) &dbmstate_s, NULL, 0x200, 0, &rqexc);
    if (!SUCCESS(errchk("timeout task", 12, &rqexc)))
             dbmstate_s.tmoutf = 0;
  }
  */

  if (!SUCCESS(iss = dbWriteDbs(slcSockets_as[sdDbex].sd,
				    id,
				    &dbmsgbc, 
				    dbsendbuf_ps )))
  { ;  /*         err_send0(&iss);*/ }
  return iss;
}                                                 
 /* End local dbDownloadMe. */
/*=============================================================================

  Name: dbSendUpdates

  Abs:  dbSendUpdates sends a job's ST2 or ST3 updates that have accumulated in a job's
        dbupdateHiLo buffer.

  Args: Type		       Name	       Access	 Description
	-------------------  -----------  -------   ----------------------------
	    const slcJob_te   slcJobId_e   R         job enum for which to send updates
	    const short       supn                   supertype (2 or 3) data to send


  Rem:  Calls dbSendWriteUpdates, which sends all items on calling job's 
        (slcJobId_e) list of pending supertype 2
        or 3 updates. This procedure is the result of another job thread issuing a dbupate().
		dbupdate sends a message to dbSend and waits at its mutex:
		slcJobs_as[slcJobId_e].dbupdate_ps->dbupdateHiLoMutex_ps
        until it is signalled back by this routine.

  Side: NOTE: Calling job thread is blocked waiting for an ack back from DBEX
        on the CTL port. 

  Ret:  vmstat_t
Replace DB_CANTSEND with other error code message to indicate connection went down, then up
The socket down/up could be DBEX/CTL or both.
==============================================================================*/

LOCAL vmsstat_t dbSendUpdates(const slcJob_te slcJobId_e)
{
  vmsstat_t      iss = DB_OKOK;
  /*----------------------------------------------*/
	
  iss = dbSendWriteUpdates(slcJobId_e, 0, dbsendbuf_ps);  /* ST2 index=0 */
  iss = dbSendWriteUpdates(slcJobId_e, 1, dbsendbuf_ps);  /* ST3 index=1 */
  /* Store status and signal event for calling job thread */ 

  /* report back first error status */
  slcJobs_as[slcJobId_e].dbupdate_ps->dbudpateHiLoAckStatus_s = iss;
	
  /* release event */
  /* job thread who called dbupdate() is waiting at its ..AckEvent */
  if (slcJobs_as[slcJobId_e].dbupdate_ps->dbupdateHiLoAckEvent_ps != NULL)
	{
	  DEBUGPRINT(DP_INFO,dbSendFlag, ("dbSendUpdates: signalling ack event to slcJobId=%d\n",
									  (int)slcJobId_e));
	  epicsEventSignal(slcJobs_as[slcJobId_e].dbupdate_ps->dbupdateHiLoAckEvent_ps);
	}
  if (!(SUCCESS(iss))) 
	{
	  if (iss==DB_CANTSEND){
		dbSendUpdateAll(); }
	}
  return iss;  
} 
/*=============================================================================

  Name: dbSendUpdateAll

  Abs:  Sends all updates accumulated for all jobs

  Args: None

  Rem:  For all jobs configured on this ioc, 
        calls dbSendWriteUpdates, which sends all items on job's 
        list of pending supertype 2 or 3 updates. 
		This procedure is called as the result of: 
		1) A DXUP message from DBEX
		2) Loosing a socket connection (either CTL or DBEX) and regaining it
		dbSend does not wait for any of the ack events to occur before
		proceeding to next job.

  Side: 

  Ret:  vmstat_t
==============================================================================*/

LOCAL vmsstat_t dbSendUpdateAll(void) {
  vmsstat_t iss = MICR_OKOK;
  int jobInd;
  slcJob_te  jobId_e;

  for (jobInd=0; jobInd<MAX_SLC_JOBS; jobInd++)
	{
	  if (slcJobs_as[jobInd].dbupdate_ps)
		{
		  jobId_e = slcGetJobEnumByName (slcJobs_as[jobInd].jobName_a); 
		  /* send to VMS ST2 updates for this job, if any (dbSendWriteUpdates checks count) */		 
		  if (SUCCESS(iss = dbSendWriteUpdates (jobId_e, 0, dbsendbuf_ps))) { /* ST2 index=0 */
			/* send to VMS ST3 updates for this job, if any */
			iss = dbSendWriteUpdates (jobId_e, 1, dbsendbuf_ps); /* ST3 index=1 */
		  }
		} 
	}
  return iss;
}   
/*=============================================================================

  Name: dbWriteDbs
  
  Abs:  Write a message to the dbex tcp/ip socket 


  Args: Type		       Name	       Access	 Description
	-------------------    -----------     -------   ----------------------------
        const int              sd                 R       socket descriptor
        const unsigned short   id                 R       native id field of supertype header
        const unsigned short*  max_bytecount_p,   R       max bytes to write not counting tcp hdr
        dbsndrcvbuf_ts*        outmsg_ps,         RW      outgoing VMS formatted msg buff, 
	                                                      pointer points to top of tcp hdr
                                                           this rte writes the tcp hder
        int*                   actual_bytecount_p W       actual bytes written not counting tcp hdr




  Rem:  Called from dbDownloadMe(), dbDiagRpy(), and dbSendUpdate
        formats tcp header by writing to outmsg_ps->pxyhdr
        sends max_bytecount_p bytes of data already stored in outmsg_ps + tcp header
          to DBEX socket
        If it is determined that we just sent an ST data update to Alpha, then
          we wait at the CTL socket for an ack back from DBEX via dbSendWaitCtlAck().

  Side: Incoming send buffer argument must already contain VMS formatted data 

  Ret:  vmsstat_t MICR_OKOK or MICR_EXNET_SENDFAIL

==============================================================================*/
/**procedure**/
vmsstat_t dbWriteDbs(const int sd,
						   const unsigned short id,
						   const unsigned short *max_bytecount_p,
						   dbsndrcvbuf_ts       *outmsg_ps )

{
   unsigned long         bytes_to_send;    
   vmsstat_t             iss = MICR_EXNET_SENDFAIL;
   vmsstat_t             isss = MICR_OKOK;

   /*----------------------------------------------------------------*/
 
   DEBUGPRINT(DP_DEBUG,dbSendFlag, ("dbWriteDbs: id=0x%hx \n", id));
   /* Set up the forward header information. */
   outmsg_ps->pxyhdr.ip_port_u.word = *(long *) v018_c4; /* DBEX's message service id */
   outmsg_ps->pxyhdr.cmd = PX_FORWARD_ALIAS_FUNC;   /* forward using alias */
 
   /* Add code to set the proper bit in the .cmd element to request a */
   /* receipt from the other end, which is DBEX in this case.         */
   if (!(id & SUPERTYPE_ID_OVERRIDE))
   {
     /* take care of byte swapping .cmd ? */
     outmsg_ps->pxyhdr.cmd |= PX_REQUEST_APP_RECEIPT;
     outmsg_ps->pxyhdr.user.byte0 = ++dbsendReceipts;
	 DEBUGPRINT(DP_INFO,dbSendFlag, 
				("dbWriteDbs: sending dbsendReceipts = %hd\n", outmsg_ps->pxyhdr.user.byte0));
   }
   else
   {
     outmsg_ps->pxyhdr.user.byte0 = 0;
   }
   outmsg_ps->pxyhdr.user.byte1 = 0;
   outmsg_ps->pxyhdr.len = htonl(*max_bytecount_p);
   outmsg_ps->pxyhdr.crc = 0x55;       /* default crc value, future=routine call */
   bytes_to_send = *max_bytecount_p + sizeof(fwd_hdr_ts);

   /* In the micros, the send routines will create a new connection if
      the connection fails, I will code this thread to simply wait for
      the receive task to detect the lost connection and re-create.
   */
   iss = MICR_EXNET_SENDFAIL;
   /* this while loop may be overkill; attempt to reesend, but */
   while( (!SUCCESS(iss)) && (!slcThreads_as[dbSend].stop)) {
	 if (SUCCESS(dbexSockCheck())) {
	   
	   if (SUCCESS (iss= tcpSendBuffer( &slcSockets_as[sdDbex], bytes_to_send, (void *) outmsg_ps)))
		 /*if ((cnt = send(sd, (char *) outmsg_ps, bytes_to_send, 0)) > 0)*/
		 {  
		   /*  if (!(outmsg_ps->supblk.nativeSB.id & SUPERTYPE_ID_OVERRIDE))*/
		   if (!(id & SUPERTYPE_ID_OVERRIDE))
			 {
			   /* This ack pertains to supertype 2 & 3 updates going to DBEX      */
			   /* Now wait here for ack back from DBEX, since we have a separate  */
			   /* connection for ack in this direction.  This could take a few    */
			   /* milliseconds so provide a timeout just in case this doesnt work */
			   /* as planned (such as re-connect, proxy dead, etc).  If we see a  */
			   /* failure, then we do return bad status.                          */
			   
			   DEBUGPRINT(DP_INFO,dbSendFlag, ("dbWriteDbs: Calling dbSendWaitCtlAck()\n"));
			   isss = dbSendWaitCtlAck();			   
			 }
		 } /* if send succeeds */
	   else 
		 { 
		   ++dbSendDiags_s.slcUpdDropCnt;
		   /* avoid hogging CPU time */
		   DEBUGPRINT(DP_ERROR,dbSendFlag, ("dbWriteDbs: Unsuccessful tcpSendBuffer; sleep & try again\n"));
		   epicsThreadSleep(1.0);
		 }
	 }
   }/*while*/
   /* now check return status of dbSendWaitCtlAck call (isss) */
   /* MICR_OKOK ok; MICR_EXNET_BADACK for bad ack;
      MICR_EXNET_RECVTMOUT if the call timed out without getting a ack after 15x1sec loop
	  MICR_EXNET RECVFAIL if there was socket error 
   */
   
   return isss;
}

