/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/* gpHashLib.c,v 1.24.2.1 2004/03/29 19:35:58 evans Exp */
/* Author:  Marty Kraimer Date:    04-07-94 */
/*=============================================================================
 
  Name: dbhash.c

  Abs:  Utilities for dictionary database (db) hash table useage
        Reference web pages on LCLS Controls | Facilities |SLC-Aware IOC | :
                        Database Service (DBS) Design Document 
                        Functional Requirements | SLC Database Service (DBS)
                        Database Input/Output Utilities
                        C Coding Standard

  Auth: 29-DEC-2004, Debbie Rogind (DROGIND) created from gphash.c- 
           See detailed mods below
  Rev:  dd-mmm-200y,
-----------------------------------------------------------------------------*/

/****************   Modifications for dbHash ******************************/
/*----------------------------------------------------------------------------- 

  Mod:  (newest to oldest)  
        29-OCT-2004, DROGIND:
		Copied $EPICS_BASE/src/libCom/gphash/gpHash* to dbHash*
		substitued 'db' for 'gp' in the files
		added dbInitFreeLists(), dbDelFreeLists(), and associated pointers for
	      free list list heads and free list entries
	    dbhInitPvt calls dbInitFreeLists()
		dbhFreeMem calls dbDelFreeLists()
		dbhAdd gets list heads and entries from ass'ed free list
		dbhDelete frees entries (note: not heads) back to entry free list
		removed epicsMutexId lock from dbhPvt struct and commented out usage everywhere
		removed pvtid from being passed to procedures and commented out usage everywhere
		changed strcmp to strncmp everywhere; added namelen arg to 
		  dbhFind, dbhAdd, dbhDelete
		Changed GPHENTRY structure name to dbhEntry_ts (in dbhash.h)
		removed void * pvtid from dbhEntry_ts struct
		Added:   
	      #include "dbdict.h" for dbdictPUS_ts for dbhDumpFP()
	      NENTRIES defines how many entries can be placed into hash table
	      DBDICTSIZE defines the hash table size and how many list heads there are
		  Changed some variable names to be consistent with our C Coding standards
        
		Reference LCLS Controls | Facilities |SLC-Aware IOC |  
	      C Coding Standards
        14-FEB-2005, Debbie Rogind:
		  Added DEBUGPRINT 
		  Modified hash() prototype & code to include length of string
		1-MAR-2005 Modified dbhDump fprintf statements to contain precisions
		  solaris may init memory w '\0'
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int dbhFlag = DP_ERROR;
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdlib.h>
#include <stddef.h>
/* VMS Control system function/status codes  */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */
#include     "slcTypes.h"  /* vmsstat_t */

#define epicsExportSharedSymbols
#include "cantProceed.h"
#include "epicsMutex.h"
#include "dbDefs.h"
#include "ellLib.h"
#include "epicsPrint.h"
#include "freeList.h"
#include "dbGlob.h"
#include "dbdict.h"
#include "dbhash.h"
#include "slcCmlog.h"      /* slcCmlogLogMsg */

typedef struct dbhPvt {
    int		tableSize;
    int		nShift;
    ELLLIST	**paplist; /*pointer to array of pointers to ELLLIST */
  /*epicsMutexId lock;*/
}dbhPvt;


LOCAL void *dbhEntryFreeList_ps = NULL;
LOCAL void *dbhHeadFreeList_ps = NULL;

/*The hash algorithm is the algorithm described in			*/
/* Fast Hashing of Variable Length Text Strings, Peter K. Pearson,	*/
/* Communications of the ACM, June 1990					*/

static unsigned char T[256] = {
 39,159,180,252, 71,  6, 13,164,232, 35,226,155, 98,120,154, 69,
157, 24,137, 29,147, 78,121, 85,112,  8,248,130, 55,117,190,160,
176,131,228, 64,211,106, 38, 27,140, 30, 88,210,227,104, 84, 77,
 75,107,169,138,195,184, 70, 90, 61,166,  7,244,165,108,219, 51,
  9,139,209, 40, 31,202, 58,179,116, 33,207,146, 76, 60,242,124,
254,197, 80,167,153,145,129,233,132, 48,246, 86,156,177, 36,187,
 45,  1, 96, 18, 19, 62,185,234, 99, 16,218, 95,128,224,123,253,
 42,109,  4,247, 72,  5,151,136,  0,152,148,127,204,133, 17, 14,
182,217, 54,199,119,174, 82, 57,215, 41,114,208,206,110,239, 23,
189, 15,  3, 22,188, 79,113,172, 28,  2,222, 21,251,225,237,105,
102, 32, 56,181,126, 83,230, 53,158, 52, 59,213,118,100, 67,142,
220,170,144,115,205, 26,125,168,249, 66,175, 97,255, 92,229, 91,
214,236,178,243, 46, 44,201,250,135,186,150,221,163,216,162, 43,
 11,101, 34, 37,194, 25, 50, 12, 87,198,173,240,193,171,143,231,
111,141,191,103, 74,245,223, 20,161,235,122, 63, 89,149, 73,238,
134, 68, 93,183,241, 81,196, 49,192, 65,212, 94,203, 10,200, 47 
};

#define NSIZES 9
static int allowSize[NSIZES] = {256,512,1024,2048,4096,8192,16384,32768,65636};


static void *myCalloc(size_t nobj,size_t size)
{
    void *p;

    p = callocMustSucceed(nobj,size,"freeListLib");
    return(p);
}

static int hash(const char *name_p,int nShift,  const size_t namelen)
{
    unsigned char	h0=0;
    unsigned char	h1=0;
    unsigned short	ind0,ind1;
    int			even = TRUE;
    unsigned char	c;
	int loopi = 0;

    while(loopi<namelen) {
	  c = *name_p;
	  if(even) {h0 = T[h0^c]; even = FALSE;}
	  else {h1 = T[h1^c]; even = TRUE;}
	  name_p++;
      loopi++;
    }
    ind0 = (unsigned short)h0;
    ind1 = (unsigned short)h1;
    return((ind1<<nShift) ^ ind0);
}

static void dbhInitFreeLists (void)
{
  if (!dbhEntryFreeList_ps){
    freeListInitPvt(&dbhEntryFreeList_ps, sizeof(dbhEntry_ts),NENTRIES);
  } 
  else DEBUGPRINT(DP_ERROR,dbhFlag,("dbhEntryFreeList_ps ptr is not NULL; trying to init; error\n"));
  if (!dbhHeadFreeList_ps){
    freeListInitPvt(&dbhHeadFreeList_ps, sizeof(ELLLIST), dbdictSize);
  } 
  else DEBUGPRINT(DP_ERROR,dbhFlag,("dbhHeadFreeList_ps ptr is not NULL; trying to init; error\n"));    
}
static void dbhDelFreeLists (void)
{
  if (dbhEntryFreeList_ps) {
    freeListCleanup(dbhEntryFreeList_ps);
	dbhEntryFreeList_ps = NULL;
  }
  if (dbhHeadFreeList_ps) {
   freeListCleanup(dbhHeadFreeList_ps);
   dbhHeadFreeList_ps = NULL;
  }
}

void dbhInitPvt(void **pvt_pps,int size)
{
    dbhPvt *dbhPvt_ps;
    int	   i;
    int	   tableSize=0;
    int	   nShift=0;

    for(i=0; i<NSIZES; i++) {
	  if(size==allowSize[i]) {
	    tableSize = size;
	    nShift = i;
	  }
    }
    if(tableSize==0) {
	  epicsPrintf("dbhInitPvt: Illegal size\n");
	  return;
    }
    dbhPvt_ps = myCalloc(1,sizeof(dbhPvt));
    dbhPvt_ps->tableSize = tableSize;
    dbhPvt_ps->nShift = nShift;
    dbhPvt_ps->paplist = myCalloc(tableSize, sizeof(ELLLIST *));
    /*dbhPvt_ps->lock = epicsMutexMustCreate();*/
    *pvt_pps = (void *)dbhPvt_ps;
    DEBUGPRINT(DP_INFO,dbhFlag, ("dbhInitPvt: ptr = %p\n", dbhPvt_ps));
    /* init freeLists */
    dbhInitFreeLists();
    return;
}
	
dbhEntry_ts * dbhFind(void *pvt_ps,const char *name_p, const size_t namelen)
{
    int		hashInd;
    dbhPvt	*dbhPvt_ps = (dbhPvt *)pvt_ps;
    ELLLIST	**paplist;
    ELLLIST	*dbhlist;
    dbhEntry_ts	*dbhNode_ps;
    
    if(dbhPvt_ps==NULL) return(NULL);
    paplist = dbhPvt_ps->paplist;
    hashInd = hash(name_p,dbhPvt_ps->nShift, namelen);
    if (namelen == SZ_NAME_SGL) {
	  DEBUGPRINT(DP_DEBUG, dbhFlag, ("\ndbhFind: hashInd= %d for %.4s\n", hashInd, name_p));}
	else {
	  DEBUGPRINT(DP_DEBUG, dbhFlag, ("\ndbhFind: hashInd= %d for %.12s\n", hashInd, name_p));}
    /* epicsMutexMustLock(dbhPvt_ps->lock);*/
    if ((dbhlist=paplist[hashInd]) == NULL) {
	  dbhNode_ps = NULL;
	  DEBUGPRINT(DP_DEBUG, dbhFlag, ("dbhFind: dbhNode_ps=NULL\n"));
    } else {
	  dbhNode_ps = (dbhEntry_ts *) ellFirst(dbhlist);
    }
    while(dbhNode_ps) {
	  if(strncmp(name_p,(char *)dbhNode_ps->name, namelen) == 0) {
		/*if(pvtid==dbhNode_ps->pvtid)*/ break;
	  }
	  dbhNode_ps = (dbhEntry_ts *) ellNext((ELLNODE*)dbhNode_ps);
    }
    /* epicsMutexUnlock(dbhPvt_ps->lock);*/
	DEBUGPRINT(DP_DEBUG, dbhFlag, ("dbhFind: return dbhNode_ps=%p\n", dbhNode_ps));
    return(dbhNode_ps);
}

dbhEntry_ts * dbhAdd(void *pvt_ps,const char *name_p, const size_t namelen)
{
    dbhPvt	*dbhPvt_ps = (dbhPvt *)pvt_ps;
    int		hashInd;
    ELLLIST	**paplist;
    ELLLIST	*plist;
    dbhEntry_ts	*dbhNode_ps;


    if(dbhPvt_ps==NULL){
      DEBUGPRINT(DP_ERROR,dbhFlag, ("dbhAdd: dbhPvt_ps = NULL\n"));
      return(NULL);
    }
    paplist = dbhPvt_ps->paplist;
    hashInd = hash(name_p,dbhPvt_ps->nShift, namelen);
    /* epicsMutexMustLock(dbhPvt_ps->lock);*/
    if(paplist[hashInd] == NULL) {
      /*paplist[hashInd] = myCalloc(1, sizeof(ELLLIST));*/
      plist = freeListCalloc (dbhHeadFreeList_ps);
	  if (!plist){
		slcCmlogLogMsg(MICR_MALLOC, "MALLOC",
					   "%s unable to malloc dbhHead buffer from freelist.\n", 
					   epicsThreadGetNameSelf());
		return (NULL);
	  }
      paplist[hashInd] = plist;
      ellInit(paplist[hashInd]);
    }
    plist=paplist[hashInd];
    dbhNode_ps = (dbhEntry_ts *) ellFirst(plist);
    while(dbhNode_ps) {
	if(strncmp(name_p,(char *)dbhNode_ps->name, namelen) == 0)
	   /*&&(pvtid == dbhNode_ps->pvtid))*/ {
	  /* epicsMutexUnlock(dbhPvt_ps->lock);*/
	    return(NULL);
	}
	dbhNode_ps = (dbhEntry_ts *) ellNext((ELLNODE*)dbhNode_ps);
    }
    /*dbhNode_ps = myCalloc(1, (unsigned) sizeof(dbhEntry_ts));*/
    dbhNode_ps = freeListCalloc (dbhEntryFreeList_ps);
	if (!dbhNode_ps){
	  slcCmlogLogMsg(MICR_MALLOC, "MALLOC",
					 "%s unable to malloc dbhEntry buffer from freelist.\n", 
					 epicsThreadGetNameSelf());
	  return (NULL);
	}
    dbhNode_ps->name = name_p;
    /*dbhNode_ps->pvtid = pvtid;*/
    ellAdd(plist, (ELLNODE*)dbhNode_ps);
    /* epicsMutexUnlock(dbhPvt_ps->lock);*/
    return (dbhNode_ps);
}

void dbhDelete(void *pvt_ps,const char *name_p, const size_t namelen)
{
    int		hashInd;
    dbhPvt	*dbhPvt_ps = (dbhPvt *)pvt_ps;
    ELLLIST	**paplist;
    ELLLIST	*plist = NULL;
    dbhEntry_ts	*dbhNode_ps;
    
    if(dbhPvt_ps==NULL) return;
    paplist = dbhPvt_ps->paplist;
    hashInd = hash(name_p,dbhPvt_ps->nShift, namelen);
    /* epicsMutexMustLock(dbhPvt_ps->lock); */
    if(paplist[hashInd] == NULL) {
	dbhNode_ps = NULL;
    } else {
	plist=paplist[hashInd];
	dbhNode_ps = (dbhEntry_ts *) ellFirst(plist);
    }
    while(dbhNode_ps) {
	if(strncmp(name_p,(char *)dbhNode_ps->name, namelen) == 0)
	  /*&&(pvtid == dbhNode_ps->pvtid))*/ {
	    ellDelete(plist, (ELLNODE*)dbhNode_ps);
	    freeListFree (dbhEntryFreeList_ps, (void *)dbhNode_ps);
	    /*free((void *)dbhNode_ps);*/
	    break;
	}
	dbhNode_ps = (dbhEntry_ts *) ellNext((ELLNODE*)dbhNode_ps);
    }
    /* epicsMutexUnlock(dbhPvt_ps->lock);*/
    return;
}

void dbhFreeMem(void *pvt_ps)
{
    dbhPvt	*dbhPvt_ps = (dbhPvt *)pvt_ps;

    /*caller must ensure that no other thread is using *pvt */
    if(dbhPvt_ps==NULL) return;
    
    /* epicsMutexDestroy(dbhPvt_ps->lock); */
    free((void *) dbhPvt_ps->paplist);
    free((void *)dbhPvt_ps);
    /* free the free lists */
    dbhDelFreeLists();
}
void dbhDump(void *pvt_ps)
{
    dbhDumpFP(stdout,pvt_ps);
}

void dbhDumpFP(FILE *fp,void *pvt_ps)
{
    int		hashInd;
    dbhPvt	*dbhPvt_ps = (dbhPvt *)pvt_ps;
    ELLLIST	**paplist;
    ELLLIST	*plist;
    dbhEntry_ts	*dbhNode_ps;
    int		number;
    ELLLIST      unitList;
    char        name[4];

    memcpy (&name, &dblongmicro, 4);
    if(dbhPvt_ps==NULL) return;
    paplist = dbhPvt_ps->paplist;
    fprintf(fp, "This is the dbhDump file for slc IOC: %s", name );
    for (hashInd=0; hashInd<dbhPvt_ps->tableSize; hashInd++) {
	  if(paplist[hashInd] == NULL) continue;
	  plist=paplist[hashInd];
	  dbhNode_ps = (dbhEntry_ts *) ellFirst(plist);
	  fprintf(fp,"\n %3.3hd=%3.3d",hashInd,ellCount(plist));
	  number=0;
	  while(dbhNode_ps) {
	    /* fprintf(fp," %s %p",dbhNode_ps->name,dbhNode_ps->pvtid);*/
		if (dbhNode_ps->kind_e == primUnitSecn){
	    fprintf(fp," %.12s",dbhNode_ps->name);
	      fprintf(fp, "; supn=%d; ", (*(dbdictPUS_ts *)(dbhNode_ps->usrPvt)).supn);
	    }
	    else if (dbhNode_ps->kind_e == prim){
		  fprintf(fp," %.4s",dbhNode_ps->name);
	      if (dbhNode_ps->usrPvt){

			unitList = (*(dbdictP_ts *)(dbhNode_ps->usrPvt)).listUnits; 
			fprintf(fp,"; unit cnt=%d; ",unitList.count);
	      }
	      else {
			fprintf(fp,"; no units; ");
		  }
	    }
	    if(number++ ==2) {number=0;fprintf(fp,"\n        ");}
	    dbhNode_ps = (dbhEntry_ts *) ellNext((ELLNODE*)dbhNode_ps);
	  }
    }
    fprintf(fp,"\n End of DB Hash\n");
}
