/*=============================================================================
 
  Name: dbl.c
        dblunits - list all units for a given primary
        dblget   - read data for a given list
        dblput   - write data for a given list

  Abs:  Utilities which read and write database lists 
        Reference web pages on LCLS Controls | Facilities |SLC-Aware IOC | :
                        Database Service (DBS) Design Document 
                        Functional Requirements | SLC Database Service (DBS)
                        Database Input/Output Utilities
                        C Coding Standard


  Auth: 4-JAN-2005, Debbie Rogind (DROGIND) created
  Rev:  dd-mmm-200y,
-----------------------------------------------------------------------------*/
#include "copyright_SLAC.h"    
/*-----------------------------------------------------------------------------  
  Mod:  (newest to oldest)  
        09-NOV-2006 DROGIND
		   Removed debug statements

=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int dblFlag = DP_ERROR;
#endif
/* c includes */
#include     <stdio.h>
#include     <stdlib.h>    /* malloc, free */
#include     <string.h>    /* memcpy */
#include     <cantProceed.h>
#include     "freeList.h"

/* slc global vars */
#include     "slcGlob.h"

/* VMS Control system function/status codes  */
#include     "msgfunc.hc"  /* FUNC_MSG function code */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */

/* slc common includes */
#include     "cvtUtil.h"   /* slc conversion utilities */
#include     "timeUtil.h"  /* slc time conversion utilities */
#include     "slcCmlog.h"  /* slcCmlogLogMsg */
#include     "slcUtil.h"   /* slcGetIOCName */ 

/* dbs includes  */  
#include     "db_micro_struc.h"
#include     "dbUtil.h" 
#include     "dbGlob.h"   
#include     "dbdef.hc"    /* DB_OKOK,,*/
#include     "dbdict.h"

/* dbdict */
#include      "dbhash.h"
#include      "dbdict.h"
/* dbget */
#include      "dbget.h"
/* dblist  prototypes */
#include    "dblistUtil.h"
#include    "dblist.h"
#include    "dbcheck.h"
#include    "dbl.h"
#include    "dbupdate.h"
LOCAL void dbLogST1Change (const dbdictPUS_ts * dbdictPUS_ps,
						   const dbheader_ts * oldValue_ps,
						   const dbheader_ts * newValue_ps,
						   const char * name);
/*=============================================================================
  Name: dblunits

  Abs: For the given primary, return the corresponding list of
       units.  

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
	const char * const   prim_a       R          see below
        dbheader_ts * *const dbdata_pps   R/W        see below

        prim_a   - A pointer to four ASCII characters.
               These four characters define the database Primary whose
               units will be returned.
        dbdata_pps - A pointer to a buffer into which the requested units
               will be returned.

  Rem:  
  Side: Memory allocation.

  Ret:  vmsstat_t:

==============================================================================*/
vmsstat_t dblunits (const char  * const prim_a,  
					dbheader_ts ** const dbdata_pps)
{
  vmsstat_t iss =DB_OKOK;  
  dbhEntry_ts *  dbhEntry_ps;
  dbdictP_ts  *  dbdictP_ps;      /* pointer to dbdictP primary only entry in dict */     
  dbdictU_ts  *  dbdictU_ps;      /* pointer to dbdictU unit entry */
  ELLLIST        listUnits;       /* unit list head for input arg prim_a when units=ALL* */          
  
  DEBUGPRINT(DP_INFO,dblFlag,("dblunits: prim = %.4s \n", prim_a ));
  if (!SUCCESS (iss=dbcheckWithWait((const dbheader_ts ** const)dbdata_pps))) {
    return (iss); }
  DEBUGPRINT(DP_INFO,dblFlag,("dblunits: dbdata_pps=%p; *dbdata_pps=%p \n",dbdata_pps, *dbdata_pps ));  
  if (!SUCCESS (iss=dbcheckPrim(prim_a,
			   (const dbheader_ts ** const) dbdata_pps, &dbhEntry_ps))) {
    return (iss); 
  }

  if (!dbhEntry_ps->usrPvt){
	return (iss=DB_BADPTR); 
  }
  
  /* progress linearly through units linked list */
  dbdictP_ps = (dbdictP_ts *)dbhEntry_ps->usrPvt;
  if (!dbdictP_ps){
	return (iss=DB_BADPTR);}
  listUnits = dbdictP_ps->listUnits;
  dbdictU_ps = (dbdictU_ts *)ellFirst(&(listUnits));
  if (!dbdictU_ps){
	DEBUGPRINT(DP_INFO,dblFlag,("dblunits: no units for prim=%4s\n",prim_a));	
	return (iss=DB_NOTALLUNIT);
  }
  while (dbdictU_ps)
	{
	  /* store units into list */
	  DEBUGPRINT(DP_INFO,dblFlag,("dblunits: dblistAdd for unit = %.4s\n", (char *)dbdictU_ps->name ));
	  dblistAdd( 4, 1, (void *)dbdictU_ps->name,(const dbheader_ts ** const) dbdata_pps);
	  DEBUGPRINT(DP_INFO,dblFlag,("dblunits: prim %.4s has unit = %.4s\n", 
								  prim_a, (char *)dbdictU_ps->name));
	  dbdictU_ps = (dbdictU_ts *)ellNext((ELLNODE *)dbdictU_ps);
	}
    
  return (iss);
}
/*=============================================================================
  Name: dblsecns

  Abs: For the given primary, return the corresponding list of
       secondaries  

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
	    const char * const   prim_a       R          see below
        dbheader_ts * *const dbdata_pps   R/W        see below

        prim_a   - A pointer to four ASCII characters.
               These four characters define the database Primary whose
               units will be returned.
        dbdata_pps - A pointer to a buffer into which the requested secns
               will be returned.

  Rem:  
  Side: Memory allocation.

  Ret:  vmsstat_t:

==============================================================================*/
vmsstat_t dblsecns (const char  * const prim_a,  
					dbheader_ts ** const dbdata_pps)
{
  vmsstat_t iss =DB_OKOK;  
  dbhEntry_ts *  dbhEntry_ps;
  dbdictP_ts  *  dbdictP_ps;      /* pointer to dbdictP primary only entry in dict */     
  dbdictU_ts  *  dbdictU_ps;      /* pointer to dbdictU unit entry */
  dbdictS_ts  *  dbdictS_ps;      /* pointer to dbdictS secn entry */
  ELLLIST        listUnits;       /* unit list head for input arg prim_a when units=ALL* */  
  ELLLIST        listSecns;        /* secn list head for input arg prim_a  */         

  DEBUGPRINT(DP_INFO,dblFlag,("dblsencs: prim = %.4s \n", prim_a ));
  if (!SUCCESS (iss=dbcheckWithWait((const dbheader_ts ** const)dbdata_pps))) {
    return (iss); }
  DEBUGPRINT(DP_INFO,dblFlag,("dblsecns: dbdata_pps=%p; *dbdata_pps=%p \n",dbdata_pps, *dbdata_pps ));  
  if (!SUCCESS (iss=dbcheckPrim(prim_a,
			   (const dbheader_ts ** const) dbdata_pps, &dbhEntry_ps))) {
    return (iss); }
  else
    {
      if (!dbhEntry_ps->usrPvt){
		return (iss=DB_BADPTR); }

      /* get first unit linked list */
      dbdictP_ps = (dbdictP_ts *)dbhEntry_ps->usrPvt;
      if (!dbdictP_ps){
		return (iss=DB_BADPTR);}
      listUnits = dbdictP_ps->listUnits;
      dbdictU_ps = (dbdictU_ts *)ellFirst(&(listUnits));
      listSecns = dbdictU_ps->listSecns;
      dbdictS_ps = (dbdictS_ts *)ellFirst(&(listSecns));

      while (dbdictS_ps)
		{
		  /* store secn into list */
		  DEBUGPRINT(DP_INFO,dblFlag,("dblsecns: prim %.4s has secn = %.4s\n", 
				      prim_a, (char *)dbdictS_ps->name));
		  dblistAdd( 4, 1, (void *)dbdictS_ps->name,(const dbheader_ts ** const) dbdata_pps);
		  dbdictS_ps = (dbdictS_ts *)ellNext((ELLNODE *)dbdictS_ps);
		}
    }  
  
  return (iss);
}
/*=============================================================================
  Name: dblget

  Abs: For the given data list of pointers, return the corresponding list of
       data.

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
        const dbheader_ts * const dblist_ps   R    pointer to list of dict data ptrs
        dbheader_ts **const dbdata_pps        W    pointer to pointer to list of data

  Rem:  

  Side: Memory allocation.

  Ret:  vmsstat_t:
Note: dblist_ps contents are checked to see if list is zero (cur=0); if so error
      if dbdata_ps contents are not zeroed, list cur is set to zero. 
==============================================================================*/
vmsstat_t  dblget  (const dbheader_ts * const dblist_ps, 
					dbheader_ts **const dbdata_pps)

{
  vmsstat_t iss = MICR_OKOK;  
  unsigned long   *next_p;            /* next dict ptr location in dblist */
  char            *nextData_p;        /* next data element in dbdata list */
  unsigned long   byteCnt;            /* # bytes to convert */
  unsigned short  curElems=0;         /* current # of elements in dblist (ptrs to hash entries) */
  unsigned short  loop;               /* current # bytes in  data list; loop cnter */
  unsigned short  dataElems =0;       /* number of data elements in data list */
  dbhEntry_ts     *dbhEntry_ps;       /* ptr elements in dblist_ps */
  dbdictPUS_ts    *dbdictPUS_ps;      /* dbhEntry_ps->usrPvt points to this element */
  char            *vmsData_p;         /* ptr to ST data */

  if (!SUCCESS(iss=dbcheck((const dbheader_ts ** const)dbdata_pps))) {
	/*DEBUGPRINT(DP_DEBUG,dblFlag,
	  ("\t Bad data list for 1) PUS=%.12s \n",(char *)dbhEntry_ps->name));*/	
    return (iss);}
  if (!SUCCESS(iss=dbcheckHeader((dbheader_ts * const)dblist_ps, epicsTrue))) {
	/*DEBUGPRINT(DP_DEBUG,dblFlag,
	  ("\t Bad list list for 1) PUS=%.12s \n",(char *)dbhEntry_ps->name));*/
    return (iss);}

  /* init read only dblist stuff */
  curElems = dblist_ps->cur/ (sizeof(dbhEntry_ts *)/2);
  next_p = (unsigned long *)dblist_ps;
  next_p += sizeof(dbheader_ts)/sizeof(long); 
  dbhEntry_ps = (dbhEntry_ts *)*next_p;        /* get hash entry for this P-U-S */
  DEBUGPRINT(DP_DEBUG,dblFlag,("dblget: dblist num elements=%u for 1) PUS=%.12s\n", 
							   curElems,(char *)dbhEntry_ps->name ));

  /*if (dblist_ps->cur == 0){
    DEBUGPRINT(DP_ERROR,dblFlag,("dblget: Error: dblist list cur is zero!\n"));
	return (iss=DB_BADPRIM);
	}*/

  for (loop=0;loop<curElems;loop++){
    dbhEntry_ps = (dbhEntry_ts *)*next_p;        /* get hash entry for this P-U-S */
    DEBUGPRINT(DP_DEBUG,dblFlag,("dblget: processing dbhEntry_ps %u =%p, dblist ptr=%p \n", 
								 loop, dbhEntry_ps, next_p)); 
	if (!dbhEntry_ps){
      DEBUGPRINT(DP_ERROR,dblFlag,("dblget: bad dbhEntry_ps dblist ptr \n"));
      DEBUGPRINT(DP_ERROR,dblFlag,("\t dblist ptr=%p; next_p=%p \n", dblist_ps, next_p));
      return (iss= DB_BADPTR);
    }
    else {
      dbdictPUS_ps = dbhEntry_ps->usrPvt;
      if (!dbdictPUS_ps){
		DEBUGPRINT(DP_ERROR,dblFlag,("dblget: bad dict ptr \n"));
		return (iss= DB_BADPTR);
      }
      byteCnt = dbdictPUS_ps->count * dbdictPUS_ps->width;
      vmsData_p = (char *)dbnodes_a[dbdictPUS_ps->supn]+(dbdictPUS_ps->sptr*2);
	  if (!vmsData_p){
		DEBUGPRINT (DP_ERROR,dblFlag,("dblget: bad vmsData_p \n"));
		return (iss=DB_BADPTR);
	  } 
	  DEBUGPRINT(DP_DEBUG,dblFlag,
	    ("dblget: vmsData_p=%p using dbnodes_a[%hd]=%p,sptr=%ld\n", 
	    vmsData_p,  dbdictPUS_ps->supn, dbnodes_a[dbdictPUS_ps->supn], dbdictPUS_ps->sptr));
	  DEBUGPRINT(DP_DEBUG,dblFlag,
	    ("\t PUS=%.12s with format=%c, width=%hd, count=%hd, supn=%hd, list ptr=%p \n", 
		(char *)dbhEntry_ps->name,dbdictPUS_ps->format, dbdictPUS_ps->width, dbdictPUS_ps->count,
				  dbdictPUS_ps->supn, next_p));

	  /* since we're storing to the dbdata list, check for overflow */
	  /* can do the logic prior to calling, or call dblistalloc every time */
	  /* could consider changing dblistAdd to have an arg of mutex Y/N and calling */
      if ( (sizeof(dbheader_ts)+ ( (* dbdata_pps)->cur)*2 + byteCnt) > ((* dbdata_pps)->max)*2 ){
		DEBUGPRINT(DP_INFO,dblFlag,("dblget: dbheader size = %d, calc'ed length = %ld, max byte length=%d\n",
				  sizeof(dbheader_ts), ((* dbdata_pps)->cur)*2 + byteCnt, ((* dbdata_pps)->max)*2) );
		if (!SUCCESS(iss=dblistalloc (byteCnt,1,  
									  (const dbheader_ts ** const) dbdata_pps))){
		  return iss;
		}
		else DEBUGPRINT(DP_INFO,dblFlag,("dblget: Enlarging dbdata list for PUS=%.12s!\n",
										 (char *)dbhEntry_ps->name ));
	  }
	  /* always recalc next ptr into data list from base in case is was realloc'ed above */
	  nextData_p = (char *)(*dbdata_pps) + sizeof(dbheader_ts) + ((* dbdata_pps)->cur)*2; 
 
	  /* read data from correct supertype area; lock first */
	  if (dbRWMutex_ps)
		epicsMutexMustLock(dbRWMutex_ps);
	  else {
		DEBUGPRINT(DP_ERROR,dblFlag,("dblget: bad dbRWMutex_ps \n"));
		return (iss=DB_BADPTR);
		}
      /* copy ST data into input data buffer */

	  memcpy (nextData_p,vmsData_p, byteCnt);

	  DEBUGPRINT(DP_DEBUG,dblFlag,("dblget: memcopied %ld bytes \n", byteCnt));

	  if (dbRWMutex_ps)
		epicsMutexUnlock(dbRWMutex_ps);

	  DEBUGPRINT(DP_DEBUG,dblFlag,
				 ("dblget: cvtFromVMS with format=%c, width=%d, count=%d \n", 
				 dbdictPUS_ps->format, dbdictPUS_ps->width, dbdictPUS_ps->count));
	
      cvtFromVMS((const char)dbdictPUS_ps->format, 
		 (const int)dbdictPUS_ps->width, 
		 (const int)dbdictPUS_ps->count, 
		 (const void *)nextData_p,
		 (void * const)nextData_p);
		  
	  next_p++;               /* next dict addr */

	  (*dbdata_pps)->cur += byteCnt/2;
      dataElems++;      
    }/*else*/     

  }/*for*/ 
  
  /*dblistPrint(dblist_ps); */

  if (dataElems != curElems ){
    DEBUGPRINT(DP_ERROR,dblFlag,
			   ("dblget: data list cnts out of synch; data elems=%u; dblist ptrs=%u \n", 
				dataElems, curElems));    
  }

  return (iss);
}

/*=============================================================================
  Name: dblput

  Abs: For the given data list of dictionary pointers, write the corresponding 
       list of data.

  Args: Type		    Name 	   Access	   Description
	------------------- -----------	   ---------- ----------------------------
        const dbheader_ts * const dblist_ps  R        pointer to ST data offsets
        const dbheader_ts * const dbdata_ps  R        pointer to list of data

  Rem:  

  Side: Memory allocation.

  Ret:  vmsstat_t:
Note: dblist_ps contents are checked to see if list is zero (cur=0); if so error
      if dbdata_ps contents are not zeroed, list cur is set to zero. 
==============================================================================*/
vmsstat_t  dblput  (const dbheader_ts * const dblist_ps,  
					const dbheader_ts * const dbdata_ps,
					slcJob_te slcJobId_e)
{
  vmsstat_t iss =  DB_OKOK;  
  unsigned long    *next_p;            /* next dict ptr location in dblist */
  char             *nextData_p;        /* next data element beginning location in dbdata */
  unsigned short   cur, loop;  /* current # of elements in list; loop cnter */
  dbhEntry_ts      *dbhEntry_ps;       /* ptr elements in dblist_ps */
  dbdictPUS_ts     *dbdictPUS_ps;      /* dbhEntry_ps->usrPvt points to this element */
  char             *vmsData_p;         /* ptr to ST area where native is conv'ed to VMS */

  if (!SUCCESS(iss=dbcheckHeader((dbheader_ts * const)dblist_ps, epicsTrue))){
    return (iss);}
  if (!SUCCESS(iss=dbcheckHeader((dbheader_ts * const)dbdata_ps, epicsFalse))){
    return (iss);}  
  if ((slcJobId_e < 0)||(slcJobId_e >= invalidJob)) {
	return (iss=DB_BADPTR);}

  cur = dblist_ps->cur/(sizeof(dbhEntry_ts *)/2);
  next_p = (unsigned long *)dblist_ps;
  next_p += sizeof(dbheader_ts)/sizeof(unsigned long); /* point to first dict ptr loc in list */

  nextData_p = (char *)dbdata_ps;
  nextData_p += sizeof(dbheader_ts);

  if (dblist_ps->cur == 0){
    DEBUGPRINT(DP_ERROR,dblFlag,("dblget: Error: dblist list cur is zero!\n"));
	return (iss=DB_BADPRIM);
  }
  /* Log an error and return if going beyond cur in dbdata_ps */
  
  DEBUGPRINT(DP_INFO,dblFlag,("dblput: cur number of elements in list=%u\n",cur ));   
  DEBUGPRINT(DP_INFO,dblFlag,("dblput: dbhEntry in dblist =%lx\n",(unsigned long)*next_p)); 
  DEBUGPRINT(DP_INFO,dblFlag,("dblput: ulong entry in data list =%lu\n",(unsigned long)*nextData_p));   
  for (loop=0;loop<cur;loop++){
    dbhEntry_ps = (dbhEntry_ts *)(*next_p); /* get hash entry for this P-U-S */
    if (!dbhEntry_ps){
      DEBUGPRINT(DP_ERROR,dblFlag,("dblput: bad list ptr \n"));        
      return (iss= DB_BADPTR);
    }
    else {
      dbdictPUS_ps = dbhEntry_ps->usrPvt;
      if (!dbdictPUS_ps){
		DEBUGPRINT(DP_ERROR,dblFlag,("dblput: bad dict ptr \n"));
		return (iss= DB_BADPTR);
      }

	  /* vmsData_p is calc'ed from dictionary entry */
      vmsData_p = (char *)dbnodes_a[dbdictPUS_ps->supn] + (dbdictPUS_ps->sptr*2);  

	  /* write data to correct supertype area; lock first */
	  if (dbRWMutex_ps)
		epicsMutexMustLock(dbRWMutex_ps);
	  else {
		DEBUGPRINT(DP_ERROR,dblFlag,("dblput: bad dbRWMutex_ps \n"));
		return (iss=DB_BADPTR);
		}  

      /* lock db; send ptr to ST space */      

      cvtToVMS((const char)dbdictPUS_ps->format, 
		 (const int)dbdictPUS_ps->width, 
		 (const int)dbdictPUS_ps->count, 
		 (const void *)nextData_p,
		 (void * const)vmsData_p);

	  if (dbRWMutex_ps)
		epicsMutexUnlock(dbRWMutex_ps);	
	  /* advance to next dict entry in list */
	  next_p++;
	  /* advance to next piece of data in dbdata list */
	  nextData_p += dbdictPUS_ps->count * dbdictPUS_ps->width;
	  /* update this piece of data in approp ST2 or ST3 job hilo update list */
	  iss =dbupdateHiLo(slcJobId_e, 
						dbdictPUS_ps->supn,
						dbdictPUS_ps->sptr,  /* lo sptr */
						dbdictPUS_ps->sptr + (dbdictPUS_ps->count * dbdictPUS_ps->width)-1 ); /* hi sptr */
     
    }/*else*/
  }/*for*/
  return (iss);
}
/*=============================================================================
  Name: dblputST1

  Abs: For the given data list of dictionary pointers, write the corresponding 
       list of data. This routine is specifically called ONLY BY the ioc shell for the
       purpose of editing ST1 data.

  Args: Type		    Name 	   Access	   Description
	------------------- -----------	   ---------- ----------------------------
        const dbheader_ts * const dblist_ps  R        pointer to ST data offsets
        const dbheader_ts * const dbdata_ps  R        pointer to list of data

  Rem:  

  Side: Memory allocation.

  Ret:  vmsstat_t:

==============================================================================*/
vmsstat_t  dblputST1  (const dbheader_ts * const dblist_ps,  
					   const dbheader_ts * const dbdata_ps,
					   slcJob_te slcJobId_e)
{
  vmsstat_t iss =  DB_OKOK;  
  unsigned long    *next_p;            /* next dict ptr location in dblist */
  char             *nextData_p;        /* next data element beginning location in dbdata */
  unsigned short   cur, loop;  /* current # of elements in list; loop cnter */
  dbhEntry_ts      *dbhEntry_ps;       /* ptr elements in dblist_ps */
  dbdictPUS_ts     *dbdictPUS_ps;      /* dbhEntry_ps->usrPvt points to this element */
  char             *vmsData_p;         /* ptr to ST area where native is conv'ed to VMS */
                                       
  dbheader_ts      *oldValue_ps = NULL;/* stores old native value */
  dbheader_ts      *newValue_ps = NULL;/* stores new native value */
  char             *newVal_p    = NULL;    
  char             *oldVal_p    = NULL;


  if (!SUCCESS(iss=dbcheckHeader((dbheader_ts * const)dblist_ps, epicsTrue))){
    return (iss);}
  if (!SUCCESS(iss=dbcheckHeader((dbheader_ts * const)dbdata_ps, epicsFalse))){
    return (iss);}  

  cur = dblist_ps->cur/(sizeof(dbhEntry_ts *)/2);
  next_p = (unsigned long *)dblist_ps;
  next_p += sizeof(dbheader_ts)/sizeof(unsigned long); /* point to first dict ptr loc in list */

  nextData_p = (char *)dbdata_ps;
  nextData_p += sizeof(dbheader_ts);
  /* Log an error and return if going beyond cur in dbdata_ps */
  
  DEBUGPRINT(DP_INFO,dblFlag,("dblputST1: cur number of elements in list=%u\n",cur ));   
  DEBUGPRINT(DP_INFO,dblFlag,("dblputST1: dbhEntry in dblist =%lx\n",(unsigned long)*next_p)); 
  DEBUGPRINT(DP_INFO,dblFlag,("dblputST1: ulong entry in data list =%lu\n",(unsigned long)*nextData_p));   
  for (loop=0;loop<cur;loop++){
    dbhEntry_ps = (dbhEntry_ts *)(*next_p); /* get hash entry for this P-U-S */
    if (!dbhEntry_ps){
      DEBUGPRINT(DP_ERROR,dblFlag,("dblputST1: bad list ptr \n"));        
      return (iss= DB_BADPTR);
    }
    else {
      dbdictPUS_ps = dbhEntry_ps->usrPvt;
      if (!dbdictPUS_ps){
		DEBUGPRINT(DP_ERROR,dblFlag,("dblputST1: bad dict ptr \n"));
		return (iss= DB_BADPTR);
      }

	  /* vmsData_p is calc'ed from dictionary entry */
      vmsData_p = (char *)dbnodes_a[dbdictPUS_ps->supn] + (dbdictPUS_ps->sptr*2);  

	  /* copy old value first, into a psuedo dblist, before it is overwritten */
	  /* old value will be logged */

	  if (SUCCESS(iss =dblistalloc(sizeof(unsigned long),1,(const dbheader_ts ** const)&oldValue_ps))){
		oldVal_p = (char *)oldValue_ps;
		oldVal_p += sizeof(dbheader_ts);
		memcpy(oldVal_p, vmsData_p, dbdictPUS_ps->width);
	
		cvtFromVMS((const char)dbdictPUS_ps->format, 
				   (const int)dbdictPUS_ps->width, 
				   (const int)dbdictPUS_ps->count, 
				   (const void *)oldValue_ps,
				   (void * const)oldValue_ps);
	  }
	  /* store new value before it is overwritten w VMS conversion */
	  if (SUCCESS(iss =dblistalloc(sizeof(unsigned long),1,(const dbheader_ts ** const)&newValue_ps))){
		newVal_p = (char *)newValue_ps;
		newVal_p += sizeof(dbheader_ts);
		memcpy(newVal_p, nextData_p, dbdictPUS_ps->width);
		DEBUGPRINT(DP_INFO,dblFlag,("dblputST1: ulong entry in newValue list =%lu\n",
									(unsigned long)*newVal_p));
	  }

	  /* write data to correct supertype area; lock first */
	  if (dbRWMutex_ps)
		epicsMutexMustLock(dbRWMutex_ps);
	  else {
		DEBUGPRINT(DP_ERROR,dblFlag,("dblputST1: bad dbRWMutex_ps \n"));
		return (iss=DB_BADPTR);
		}  

      cvtToVMS((const char)dbdictPUS_ps->format, 
		 (const int)dbdictPUS_ps->width, 
		 (const int)dbdictPUS_ps->count, 
		 (const void *)nextData_p,
		 (void * const)vmsData_p);

	  if (dbRWMutex_ps)
		epicsMutexUnlock(dbRWMutex_ps);	
	  /* advance to next dict entry in list */
	  next_p++;
	  /* advance to next piece of data in dbdata list */
	  nextData_p += dbdictPUS_ps->count * dbdictPUS_ps->width;
	  /*
		for array data: 
		if (dbdictPUS_ps->count > 1){
		iss = DB_DBEDITST1;
		slcCmlogLogMsg(iss, "DB_DBEDITST1","%s ST1 data changed.\n", dbhEntry_ps->name); 
		} else */
	  dbLogST1Change ((const dbdictPUS_ts *) dbdictPUS_ps,
					  (const dbheader_ts *) oldValue_ps,
					  (const dbheader_ts *) newValue_ps,
					  dbhEntry_ps-> name);     
    }/*else*/
  }/*for*/

 
  return (iss);
}
/*=============================================================================
  Name: dblputST1

  Abs:  logs using slcCmlogLogMsg
        "<prim> <unit> <secn> changed from <old_value> to <new_value>"


  Args: Type		    Name 	   Access	   Description
  ----- -------------- ----------- ---------- ----------------------------
   const dbdictPUS_ts * dbdictPUS_ps   R      pointer to PUS dict struture
   const dbheader_ts *  oldValue_ps    R      pointer to list containing old value
   const dbheader_ts *  newValue_ps    R      pointer to list containing new value
   const char *         name           R      pointer to 12 char name of PUS

  Rem:  

  Side: switches on format to process incoming old and new values contained in 
        their respective dblists.

  Ret:  logs vmsstat_t of DB_DBEDITST1

==============================================================================*/
LOCAL void dbLogST1Change (const dbdictPUS_ts * dbdictPUS_ps,
						   const dbheader_ts * oldValue_ps,
						   const dbheader_ts * newValue_ps,
						   const char * name) 
{
  long oldlint, newlint;           /* for accessing 8 byte 'T' as 2 long words */
  unsigned long oldulint, newulint;
  short oldsint, newsint; 
  float oldtfloat, newtfloat;
  char             *oldVal_p;
  char             *newVal_p;  
  vmsstat_t        iss=DB_DBEDITST1;

  if (oldValue_ps && newValue_ps){
	oldVal_p = (char *)oldValue_ps;
	oldVal_p += sizeof(dbheader_ts);
	newVal_p = (char *)newValue_ps;
	newVal_p += sizeof(dbheader_ts);
  }
  else return;

  switch (dbdictPUS_ps->format){
  case 'I':
	if (dbdictPUS_ps->width==4){
	  memcpy (&oldlint, oldVal_p, dbdictPUS_ps->width);
	  memcpy (&newlint, newVal_p, dbdictPUS_ps->width);
	  slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 data changed from %ld to %ld.\n", 
					 name, oldlint, newlint); 
	}
	else {
	  memcpy (&oldsint, oldVal_p, dbdictPUS_ps->width);
	  memcpy (&newsint, oldVal_p, dbdictPUS_ps->width);
	  slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 data changed from %d to %d.\n", 
					 name, oldsint, newsint); 
	}
	break;
  case 'A':
	/* count is total # of chars; space between every 4 chars */
	  slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 data changed from %.4s to %.4s\n", 
					 name, oldVal_p, newVal_p); 
	break;
  case 'R':	
	memcpy (&oldtfloat, oldVal_p, dbdictPUS_ps->width);
	memcpy (&newtfloat, newVal_p, dbdictPUS_ps->width);
	slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 data changed from %12.6g to %12.6g\n", 
					 name, oldtfloat, newtfloat); 
	break;
  case 'Z':
	if (dbdictPUS_ps->width==4){
	  memcpy (&oldlint, oldVal_p, dbdictPUS_ps->width);
	  memcpy (&newlint, newVal_p, dbdictPUS_ps->width);
	  slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 data changed from %.8lx to %.8lx\n", 
					 name, oldlint, newlint); 
	}
	else {
	  memcpy (&oldsint, oldVal_p, dbdictPUS_ps->width);
	  memcpy (&newsint, newVal_p, dbdictPUS_ps->width);
	  slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 data changed from %.4hx to %.4hx\n", 
					 name, oldlint, newlint);
	}
	break;
  case 'S':
	slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 data changed from %.*s to %.*s\n", 
					 name, oldVal_p, newVal_p); 		
	break;
  case 'T':
	memcpy (&oldulint, oldVal_p, dbdictPUS_ps->width);
	memcpy (&newulint, newVal_p, dbdictPUS_ps->width);
	slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 timestamp[0] changed from 0x%.8lx to 0x%.8lx\n", 
					 name, oldulint, newulint); 
	oldVal_p += 4;
	newVal_p += 4;
	memcpy (&oldulint, oldVal_p, dbdictPUS_ps->width);
	memcpy (&newulint, newVal_p, dbdictPUS_ps->width);
	slcCmlogLogMsg(iss, "DB_DBEDITST1","%.12s ST1 timestamp[1] changed from 0x%.8lx to 0x%.8lx\n", 
					 name, oldulint, newulint);
	break;
  }
  return;
}
