/*=============================================================================
 
  Name: dblist.c
        dblist  
        dblistalloc  
        dblistfree 

  Abs:  Utilities which manage database lists 
        Reference web pages on LCLS Controls | Facilities |SLC-Aware IOC | :
                        Database Service (DBS) Design Document 
                        Functional Requirements | SLC Database Service (DBS)
                        Database Input/Output Utilities
                        C Coding Standard


  Auth: 1-DEC-2004, Debbie Rogind (DROGIND) created
  Rev:  dd-mmm-200y,
-----------------------------------------------------------------------------*/
#include "copyright_SLAC.h"    
/*-----------------------------------------------------------------------------  
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int dblistFlag = DP_ERROR;
#endif
/* c includes */
#include     <stdio.h>
#include     <stdlib.h>    /* malloc, free */
#include     <string.h>    /* memcpy */
#include     <cantProceed.h>

/* slc global vars */
#include     "slcGlob.h"

/* VMS Control system function/status codes  */
#include     "msgfunc.hc"  /* FUNC_MSG function code */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */

/* slc common includes */
#include     "cvtUtil.h"   /* slc conversion utilities */
#include     "timeUtil.h"  /* slc time conversion utilities */
#include     "slcCmlog.h"  /* slcCmlogLogMsg */
#include     "slcUtil.h"   /* slcGetIOCName */ 

/* dbs includes  */  
#include     "db_micro_struc.h"
#include     "dbUtil.h" 
#include     "dbGlob.h"   
#include     "dbdef.hc"    /* DB_OKOK,,*/
#include     "dbdict.h"
#include     "freeList.h"
#include     "cantProceed.h"
/* dbdict */
#include      "dbdict.h"
/* dbget */
#include      "dbget.h"
/* dbcheck */
#include      "dbcheck.h"
/* dblist  prototypes */
#include    "dblistUtil.h"
#include    "dblist.h"

/*=============================================================================
  Name: dblistalloc

  Abs:  This routine allocates and initializes a data list from
         either of two free list memory pools, small and big.
         If the big list size is too small, it allocates from the heap.

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
	dbheader_ts **const dbdata_pp, 	    R/W    pointer to pointer to structure
 	const unsigned long bytes_in_element R     Size in bytes of individual list elements  
 	const unsigned long num_elements     R     Number of elements initially to be in list

   Rem:  The list must be passed as a pointer to a pointer so that it
         may be modified.
         Caller must determine how many elements it needs

   Side: Memory allocation.

   Ret:  vmsstat_t:
         DB_BADPTR          - bad pointer supplied
         DB_NOHEAP          - insufficient HEAP memory available
         DB_BADVERS         - DEFINE: list version and current version don't match 

==============================================================================*/ 
vmsstat_t dblistalloc (const unsigned short bytes_in_element,  
					   const unsigned short num_elements, 
					   const dbheader_ts ** const dbdata_pps)
{ 
  vmsstat_t iss = MICR_OKOK; 
  unsigned long bytes_to_alloc, temp;
  unsigned short max, cur;

  
/*----------------------------------------------------------------------------*/

  /*   Check for valid pointer */
  if (dbdata_pps == NULL) { 
    iss = DB_BADPTR; 
    return (iss); 
  } 
  else DEBUGPRINT(DP_DEBUG,dblistFlag, ("dblistalloc: dbdata_pps=%p; *dbdata_pps=%p \n", 
									   dbdata_pps, *dbdata_pps));
  /* if free lists have been destroyed, must be initiating shutdown; 
     don't recreate freeList...just exit with bad status  */
  if (!(SUCCESS(iss=dbcheckFreeLists()))) return iss;
  /*   calculate size for all elements, plus header size  */
  bytes_to_alloc = (num_elements * bytes_in_element); 

  DEBUGPRINT(DP_DEBUG,dblistFlag, ("dblistalloc: bytes to alloc = %ld\n", bytes_to_alloc)); 
  /* If list pointer to pointer is NULL then this is a new list  */
  if (*dbdata_pps == NULL) { 
    iss = dblistNew(bytes_to_alloc + sizeof(dbheader_ts), dbdata_pps);
	DEBUGPRINT(DP_INFO,dblistFlag, ("dblistalloc: dblistNew returned *dbdata_pps=%p \n", *dbdata_pps));  
  } 
  /*  Otherwise, possibly re-allocate the requested list and modify list length  */
  else { 
	max = ((dbheader_ts *)(*dbdata_pps))->max;
	cur = ((dbheader_ts *)(*dbdata_pps))->cur;
    temp =  max - (cur + (sizeof(dbheader_ts)/2) );
	DEBUGPRINT(DP_INFO,dblistFlag, ("dblistalloc: max = %u; cur = %u ; bytes_to_alloc = %ld; list # words = %ld\n", max, cur, bytes_to_alloc, temp)); 
    if ( (bytes_to_alloc/2) >  ( max - (cur + (sizeof(dbheader_ts)/2) )) )
    {
	  DEBUGPRINT(DP_INFO,dblistFlag, ("dblistalloc: dblistEnlarge \n")); 
      iss = dblistEnlarge(bytes_to_alloc, (dbheader_ts ** const)dbdata_pps);
    }
    /*  else there is enough room in existing list */
  }
  return(iss); 
}

/*=============================================================================
  Name: dblistfree

  Abs:  This routine frees a list to its appropriate free list manager,
         if any. If the list was created on the heap, it returns it to the heap.

  Args: Type		    Name	        Access	   Description
	------------------- -----------	---------- ----------------------------
    dbheader_ts * const dbdata_or_list_ps   W    pointer to list to be freed

   Rem:  Resource is freed back to free list space. 

   Side: Uses epics freeList to manage big and small list memory pools
         Note: dbdata_or_list_ps pointer is not NULLED!

   Ret:  vmsstat_t: DB_OKOK


==============================================================================*/
vmsstat_t dblistfree (const dbheader_ts *const dbdata_or_list_ps)    
{
  vmsstat_t iss = MICR_OKOK;

  /*
  ** Return OK if NULL pointer.
  */
  if (dbdata_or_list_ps == NULL) return iss;

  if (!(SUCCESS(iss=dbcheckFreeLists()))) return iss;

  if (((dbheader_ts*) dbdata_or_list_ps)->max == DBL_WSZSM  ) {
	freeListFree (dblistSmFreeList_p, (void *) dbdata_or_list_ps);
	DEBUGPRINT(DP_DEBUG,dblistFlag, ("dblistfree: freeing small list\n")); 
	
  }
  else if (((dbheader_ts*)  dbdata_or_list_ps)->max == DBL_WSZBIG  ) {
	freeListFree (dblistBigFreeList_p, (void *) dbdata_or_list_ps); 
	DEBUGPRINT(DP_DEBUG,dblistFlag, ("dblistfree: freeing big list\n"));
  } 
  else {    
    free( (void *)dbdata_or_list_ps);
  } 
  /*dbdata_or_list_ps = NULL;  requires ** ptr to ptr input arg */
  return (iss);
}
/*=============================================================================
  Name: dblist

  Abs: For the given database terms return the corresponding list of
      "pointers" These pointers may be passed to a subsequent routine
      (see for example "dblget") to retrieve quickly the database
      information they point to.

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
	const char * const   prim_a        R         see below
	const char * const   unit_a        R         see below
	const char * const   secn_a        R         see below
        dbheader_ts * *const dblist_pps   R/W       see below

  Args: prim_a - A pointer to four ASCII characters.
               These four characters define the database Primary whose
               pointers will be returned.
        unit_a -  A pointer to char containing 4 char ASCII unit number.
               If the four characters are "ALL*" then the caller
               has requested that the units be "wildcarded".
        secn_a - A pointer to  four ASCII characters.
               These four characters define the database Secondary whose
               pointers will be returned.

	dblist_pps - ptr to ptr to data list which contains dict hash entries (ptrs)

  Rem:  

  Side: Memory allocation.

  Ret:  vmsstat_t:
         DB_BADPTR          - bad pointer supplied
         DB_NOHEAP          - insufficient HEAP memory available
         DB_BADVERS         - DEFINE: list version and current version don't match 
DB_OKOK       
DB_BADPRIM    
DB_BADCHASE   
DB_LISTSHORT  
DB_UNITORDER  
DB_NOSUCHUNIT  
DB_BADSECN    
==============================================================================*/
vmsstat_t dblist (const char * const prim_a,  
				  const char * const unit_a,  
				  const char * const secn_a,  
				  const dbheader_ts **const dblist_pps)
{
  vmsstat_t iss =DB_OKOK; 
  dbhEntry_ts *  dbhEntry_ps;     /* pointer to dict entry in hash tbl */
  dbdictP_ts  *  dbdictP_ps;      /* pointer to dbdictP primary only entry in dict */     
  dbdictU_ts  *  dbdictU_ps;      /* pointer to dbdictU unit entry */
  dbdictS_ts  *  dbdictS_ps;      /* pointer to dbdictS secn entry */
  ELLLIST        listUnits;       /* unit list head for input arg prim_a when units=ALL* */
  ELLLIST        listSecn;        /* secn list head for each unitinput arg prim_a when units=ALL* */
  char           name_a[SZ_NAME_PUS] = {' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' '};
  epicsBoolean   found = epicsFalse;

  /* validate a) that db exists and b) valid list */
  if (!(SUCCESS(iss=dbcheckWithWait(dblist_pps))))
    {
      return (iss);
    }
  DEBUGPRINT(DP_DEBUG,dblistFlag, ("dblist for %.4s %.4s %.4s\n",prim_a, unit_a, secn_a));  
  /* unit can be input as ASCII unit or 'ALL*' */
  if(strncmp(unit_a,"ALL*",SZ_NAME_SGL)==0)     /* unit = 'ALL*" ? */
    {
      /* Yes, after all units */
	  if (!SUCCESS(iss=dbcheckPrim(prim_a,dblist_pps,&dbhEntry_ps))) 
		{
		  return (iss);
		}
      else
		{
		  /* progress linearly through units linked list */
		  dbdictP_ps = (dbdictP_ts *)dbhEntry_ps->usrPvt;
		  if (!dbdictP_ps)
			{
			  return (iss=DB_BADPTR);
			}
          listUnits = dbdictP_ps->listUnits;
          dbdictU_ps = (dbdictU_ts *)ellFirst(&(listUnits));
		  while (dbdictU_ps)
			{
			  /* get secn list head for this unit */
			  listSecn = dbdictU_ps->listSecns;
              dbdictS_ps = (dbdictS_ts *)ellFirst(&(listSecn));
			  while (dbdictS_ps)
				{
				  if(strncmp(secn_a, (char *)(dbdictS_ps->name),
							 SZ_NAME_SGL)==0)     /* secn names match?*/
					{
                      /* add PUS ptr to list */
					  DEBUGPRINT(DP_INFO,dblistFlag, ("dblist: secn match = %s \n", (char *)dbdictS_ps->name));
					  if (!dbdictS_ps->dbdictEntry){
						DEBUGPRINT(DP_ERROR,dblistFlag, 
								   ("dblist: secn %s has not dict entry; return\n",(char *) dbdictS_ps->name)); 
					  }
					  dblistAdd(sizeof(dbhEntry_ts *),1,(void *)&(dbdictS_ps->dbdictEntry), dblist_pps);	   
					  found = epicsTrue;
					  break;  /* done w this unit, go to next unit */
					}
				  dbdictS_ps = (dbdictS_ts *)ellNext((ELLNODE *)dbdictS_ps);
				}
			  if (!found) return (iss = DB_BADSECN);
			  found = epicsFalse;
			  /* go to next unit list head*/
			  dbdictU_ps = (dbdictU_ts *)ellNext((ELLNODE *)dbdictU_ps);
			}
		}
    }
  else
    {
      /* single unit #, in ASCII, is specified */
      dbgetCatPUSName(prim_a, unit_a, secn_a, (char *) &name_a);
	  /*DEBUGPRINT(DP_DEBUG,dblistFlag, ("dblist: dbhFind for %.12s\n",(char *)&name_a));*/		
      dbhEntry_ps = dbhFind(dbdict_p,(char *)&name_a,SZ_NAME_PUS);
      if (!dbhEntry_ps)
		{
		  DEBUGPRINT(DP_ERROR,dblistFlag, 
					 ("dblist: couldn't find %.12s; dbhEntry_ps=%p (DB_NOSUCHNODE)\n" , 
					  (char *)&name_a, dbhEntry_ps));	
		  return (iss=DB_NOSUCHNODE);
		}
      else {
		/*dblistAdd(sizeof(dbhEntry_ts *), 1,(void *)&dbhEntry_ps, dblist_pps);*/ 
		dblistAdd(sizeof(dbhEntry_ps), 1,(void *)&dbhEntry_ps, dblist_pps); 
		DEBUGPRINT(DP_DEBUG,dblistFlag,
				   ("dblist: dblistAdding %p entry \n",dbhEntry_ps));
	  }
	}
  return (iss);
}

