/*=============================================================================
 
  Name: dblist.h

  Abs:  Public prototypes for dblist.c   
        Reference LCLS Controls | Facilities |SLC-Aware IOC |  
	   Database Service (DBS) Design Document 
        Reference RMX_DBS dbmain.c 

  Auth: 1 DEC 2004, drogind created
  Rev: 17 DEC-2004, drogind modified prototype code to production standards
 
-----------------------------------------------------------------------------*/
#include "copyright_SLAC.h"    
/*----------------------------------------------------------------------------- 
  Mod:  (newest to oldest)
        28-Jul-2005, K. Luchini (LUCHINI):
           Port db macros from dbgetc.h on the alpha   
 
=============================================================================*/
#ifndef DBLIST_HM  /* guard macro */
#define DBLIST_HM  /* guard macro */
#ifdef __cplusplus
extern "C" {
#endif

#include "stdio.h"
#include "dbhash.h"
  
/* 
 * Declaration and initialization macros:(general)
 * DBLIST_DECLARE macro used to declare a pointer to a dynamic database list. 
 */
#define DBLIST_DECLARE(type) \
    struct                   \
    {                        \
        dbheader_ts hdr_s;   \
        type        dat[1];  \
    } *

/* DBLIST_INIT macro to initialize a pointer to a dynamic database list. */
#define     DBLIST_INIT(list) list = NULL

/* DBLIST macro used to declare a NULL initialized pointer to a list */
#define     DBLIST(list, type)  DBLIST_INIT( DBLIST_DECLARE(type) list )

/* DBFREE macro used to release a dynamic list pointer */
#define     DBALLOC(list, num)  (dblistalloc(DBSIZ(list), num, \
                                 (const dbheader_ts ** const)&(list)))
  
/* DBFREE macro used to release a dynamic list pointer */
#define     DBFREE(list)  (list != NULL)? \
                           dblistfree((const dbheader_ts *const)list):1; \
                           list = NULL

/* DBCLEAR macro initializes a list */
#define     DBCLEAR(list) (((list) != NULL)?(list)->hdr_s.cur = 0:0)

/* DBMAX macro calculates the maximum number of "type" items in list */
#define     DBMAX(list) \
    (((list) != NULL) ?  (2*((list)->hdr_s.max - (sizeof(dbheader_ts)/2)))/sizeof((list)->dat[0]) : 0)

/* DBCNT macro calculates the current number of "type" items in list */
#define     DBCNT(list) \
    (((list) != NULL) ? (2*((list)->hdr_s.cur))/sizeof((list)->dat[0]) : 0)

/* DBLISTSIZE macro calculates the number of words in the data list */
#define     DBLISTSIZE(list) \
    (((list) != NULL) ? (list)->hdr_s.cur : 0)

/* 
 * DBPTR macro calculates the address of the nth "type" item in a list 
 *  (where the first "type" item is n=zero, etc.)                       
 *  Note that this does not check to see if
 *  list is a NULL pointer.  That is because there is no well-defined
 *  value that the expressionPress RETURN to continue= NULL, since
 *
 *  (list)->dat[num] could be of almost any type. Only use this
 *  macro if you are sure that list != NULL.
 *
 * (Also note that we say list->dat+num, rather than &list->dat[num].
 *  If 'dat' is a typedef'ed array, then applying the '&' operator to it
 *  will not work;  the '+' operator will). 
 */
#define     DBPTR(list, num)  ( ( (list)->dat ) + ( num ) )

/* 
 * Macro to calculate pointer to the first item in a database list.
 *  Only use this macro if you are sure that list is not NULL
 *  (see comments for DBPTR).
 *
 *  If the list is non-NULL, but DBCNT(list) is zero, this will
 *  (of course) not point to a valid item.
 *
 *  However, DBPTR_FIRST(list) will always evaluate to a greater value
 *  than DBPTR_LAST(list) if DBCNT(list) is zero.  This is useful
 *  if you are looping from the first item in a list to the last item;
 *  if the list is empty, you never enter the loop. 
 */
#define     DBPTR_FIRST(list)  ( DBPTR( (list), 0 ) )

/*
 * Macro to calculate pointer to the last item in a database list.
 *  Only use this macro if you are sure that list is not NULL
 *  (see comments for DBPTR) and that there is at least one item
 *  in the list.
 *
 *  If the list is non-NULL, but DBCNT(list) is zero, this will
 *  (of course) not point to a valid item.
 *
 * However, DBPTR_FIRST(list) will always evaluate to a greater value
 *  than DBPTR_LAST(list) if DBCNT(list) is zero.  This is useful
 *  if you are looping from the first item in a list to the last item;
 *  if the list is empty, you never enter the loop. 
 */
#define     DBPTR_LAST(list) ( DBPTR( (list), DBCNT( (list) ) - 1 ) )

/* 
 * These macros are all analagous to the DBPTR* macros, except that
 *  they evaluate to the VALUE of an item in the database list, rather
 *  than the ADDRESS of an item in the list.
 *
 *  Note that these macros do not check to see if
 *  list is a NULL pointer.  That is because there is no well-defined
 *  value that the expression could take on if list == NULL, since
 *  (list)->dat[num] could be of almost any type. Only use these
 *  macro if you are sure that list != NULL. Only use
 *  DBITM_FIRST and DBITM_LAST if DBCNT(list) is not zero. 
 */
#define     DBITM(list, num)  ( (list)->dat[(num)] )
#define     DBITM_FIRST(list) ( DBITM( (list), 0 ) )
#define     DBITM_LAST(list)  ( DBITM( (list), DBCNT( (list) ) - 1 ) )

/*
 * DBSIZ macro computes the size of a single item in a list.
 * This will work even if list == NULL.
 */
#define     DBSIZ(list)  (sizeof((list)->dat[0]))


/* structure definitions */
typedef char dbVersion_ta[8];

typedef struct 
{
  /* dbVersion_ta dbVersion_a;*/    /* dbVersion stamp when list was formed */
  unsigned short max;           /* max length of list (words, includes this hdr) */
  unsigned short cur;           /* current occupied number of words, not including this hdr */
                               /*  (when max = 0, list is empty) */
                               /*  empty list cur = 0; caller knows type(size)of element */
} dbheader_ts;



/* prototype definitions */

  vmsstat_t dblist (const char * const prim_a,
                    const char * const unit_a,
                    const char * const secn_a,
                    const dbheader_ts **const dblist_pps);

  vmsstat_t dblistalloc (const unsigned short bytes_in_element,
                         const unsigned short num_elements,
                         const dbheader_ts ** const dbdata_pps);

  vmsstat_t dblistfree (const dbheader_ts *const dbdata_or_list_ps);

#ifdef __cplusplus
}
#endif

#endif  /* DBLIST_HM guard macro */
