/*=============================================================================
 
  Name: dbshell.c
        dbshellDumpHash 
        dbshellDumpHashFile
        dbshellGetMeta 
        dbshellGet
		dbshellGetUnits
        dbshellEdit
        dbshellDump
        dbshellGetExists 
        dbshellGetVersion
        dbshellUpdate 
        dbshellSetHashSize

  Abs:  DBS Utilities commanded from ioc shell (see ../exec/slcMain.c) 
        All Database Service commands start with 'slcdb'

        Reference web pages on LCLS Controls | Facilities |SLC-Aware IOC | :
                        Database Service (DBS) Design Document 
                        Functional Requirements | SLC Database Service (DBS)
                        Database Input/Output Utilities
                        C Coding Standard


  Auth: 24-JAN-2005, Debbie Rogind (DROGIND) created
  Rev:  dd-mmm-200y,
-----------------------------------------------------------------------------*/
#include "copyright_SLAC.h"    
/*-----------------------------------------------------------------------------  
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int dbshellFlag = DP_ERROR;
#endif

/* c includes */
#include     <stdio.h>
#include     <stdlib.h>    /* malloc, free */
#include     <string.h>    /* memcpy */
#include     <cantProceed.h>

/* slc global vars */
#include     "slcGlob.h"

/* VMS Control system function/status codes  */
#include     "msgfunc.hc"  /* FUNC_MSG function code */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */

/* slc common includes */
#include     "cvtUtil.h"   /* slc conversion utilities */
#include     "timeUtil.h"  /* slc time conversion utilities */
#include     "slcCmlog.h"  /* slcCmlogLogMsg */
#include     "slcUtil.h"   /* slcGetIOCName */ 

/* dbs includes  */  
#include     "db_micro_struc.h"
#include     "dbUtil.h" 
#include     "dbGlob.h"   
#include     "dbdef.hc"    /* DB_OKOK,,*/
#include     "dbdict.h"
#include     "dbget.h"     /* dbgetVersion */
#include     "freeList.h"
#include     "cantProceed.h"
/* dbdict */
#include      "dbhash.h"
#include      "dbdict.h"
/* dbget */
#include      "dbget.h"
/* dblist  prototypes */
#include    "dblistUtil.h"
#include    "dblist.h"
#include    "dbcheck.h"
#include    "dbl.h"
#include    "dbdump.h"
#include    "dbshell.h"
#include    "dbedit.h"
#include    "dbupdate.h"


LOCAL void dbshellBlkPad (char *name_p);
/*=============================================================================
  Name: dbshellDumpHash 

  IOC SHELL COMMAND: 'slcdbDumpHash'
  Note: if wanting to dump hash table to file, start ioc and pipe to file, then type
  command(s), then exit and read file.

  Abs:  Dump the hash table to the console  

  Args: None

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		Calls dbhDumpFP with stdout, dictionary pointer

  Side: 

  Ret:  void

==============================================================================*/
void dbshellDumpHash (void)
{

  if (epicsThreadGetId(EXEC_NAME)){

	if (slcGetDbExists()){
	  dbhDumpFP(stdout,dbdict_p);
	}
	else DEBUGPRINT(DP_INFO,dbshellFlag, ("dbshell: error: db doesn't exist \n")); 

  } 
  else printf("Error - issue 'slcStart' first.\n");
}

/*=============================================================================
  Name: dbshellDumpHashFile 

  IOC SHELL COMMAND: 'slcdbDumpHashFile filename'
  Temporary for testing.

  Abs: For the given primary, return the corresponding list of
       units.  

  Args: Type		    Name	Access	   Description
        ----------- -----------	---------- ----------------------------
		char *      filename     W         filename to dump to

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		Calls dbhDumpFP with stdout, file pointer
  
  Side: File I/O

  Ret:  void

==============================================================================*/
void dbshellDumpHashFile (const char * const filename)
{
  FILE *   file_p = NULL;
  if (epicsThreadGetId(EXEC_NAME)){

	if (filename){
	  DEBUGPRINT(DP_INFO,dbshellFlag, ("dbshellHashFile: filename = %s\n", filename)); 
	  if (slcGetDbExists()){
		file_p = fopen (filename, "w");
		if (!file_p) {
		  DEBUGPRINT(DP_ERROR,dbshellFlag,("dbshellHashFile: input filename not writeable \n"));
		}
		else {
		  DEBUGPRINT(DP_INFO,dbshellFlag, ("dbshellHashFile: Calling dbhDumpFP()! \n")); 
		  dbhDumpFP(file_p,dbdict_p);
		  fclose (file_p);
		}	  
	  }
	  else {
		DEBUGPRINT(DP_ERROR,dbshellFlag, ("dbshellHashFile: error: database doesn't exist \n")); 
	  }
	}else {
	  DEBUGPRINT(DP_ERROR,dbshellFlag, ("dbshellHashFile: error - improper filename \n")); 
	}
  } else printf("Error - issue 'slcStart' first.\n");
}
/*-----------------------------------------------------------------------
  Name: dbshellGetMeta

  IOC SHELL COMMAND: 'slcdbGetMeta prim unit secn'

  Abs:  Gets meta data for given prim unit secn input on command line at 
        prompt.
        

  Args: Type		        Name	  Access	   Description
	    ---------------   --------  ---------- ----------------------------
        const char * const  prim_p     R     pointer to prim input on keyboard
        const char * const  unit_p     R     pointer to unit input on keyboard
        const char * const  secn_p     R     pointer to secn input on keyboard

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		All input args should be checked to be non-NULL pointers.

        Uses local dbshellBlkPad() versus public dbgetBlkPad() to right pad 
		each input arg, as each input argument will not necessarily be null terminated.
        Calls dbgetMeta()

  Ret:  none
----------------------------------------------------------------------*/
void dbshellGetMeta  (const char * const prim_p, 
					  const char * const unit_p, 
					  const char * const secn_p)
{
  char format = ' '; 
  int  width = 0;
  int  count = 0;
  int  supn = 0;
  char prim_pa[SZ_NAME_SGL] = "    ";
  char unit_pa[SZ_NAME_SGL] = "    ";
  char secn_pa[SZ_NAME_SGL] = "    ";
  char primUnitSecn_a[SZ_NAME_PUS];
  vmsstat_t iss = MICR_OKOK;

  if (epicsThreadGetId(EXEC_NAME)!=NO_THREAD){
   
	if (slcGetDbExists()){

	  if (! (prim_p && unit_p && secn_p) ){
		printf("Missing arguments -  'slcdbGetMeta prim unit secn' \n");
		return;
	  }
	  DEBUGPRINT(DP_DEBUG,dbshellFlag,
				 ("Before pad: prim=%.4s; unit=%.4s; secn=%.4s \n", prim_p, unit_p, secn_p));           
	  strncpy (prim_pa, prim_p, SZ_NAME_SGL);
      strncpy (unit_pa, unit_p, SZ_NAME_SGL);
      strncpy (secn_pa, secn_p, SZ_NAME_SGL);
	  DEBUGPRINT(DP_DEBUG,dbshellFlag,
				 ("After copy: prim=%.4s; unit=%.4s; secn=%.4s \n", prim_pa, unit_pa, secn_pa)); 
	  dbshellBlkPad((char *)prim_pa);
      dbshellBlkPad((char *)unit_pa);
      dbshellBlkPad((char *)secn_pa);
	  DEBUGPRINT(DP_INFO,dbshellFlag,
				 ("After pad: prim=%.4s; unit=%.4s; secn=%.4s \n", prim_pa, unit_pa, secn_pa));      
	  if (! SUCCESS(iss = dbgetMeta(prim_pa, unit_pa,secn_pa, &format, &width, &count, &supn))) {

		if (iss==DB_BADPRIM) {
		  printf ("No such primary found: %.4s\n",prim_pa);
		}
		else { 
		  if (iss==DB_NOSUCHNODE){ 
			dbgetCatPUSName(prim_pa, unit_pa, secn_pa, (char *)&primUnitSecn_a);
			printf ("No dictionary entry for '%.12s'\n",(char *)primUnitSecn_a);
		  }
		}
		return;
	  }
	  printf ("format: %c\n", format);
	  printf ("width: %d\n", width);
	  printf ("count: %d\n", count);
      printf ("supertype number:  %d\n", supn); 
	}
	else {	  
	  printf ("Command invalid; database does not exist. \n"); 
	}
  } else printf("Error - issue 'slcStart' first.\n");
}
/*-----------------------------------------------------------------------
  Name: dbshellDump

  IOC SHELL COMMAND: 'slcdbDump prim <unit> <secn>'

  Abs:  Gets data for input primary, optional unit and optional secondary.
        

  Args: Type		        Name	  Access	   Description
	    ---------------   --------  ---------- ----------------------------
        const char * const  prim_p     R     pointer to prim input on keyboard
        const char * const  unit_p     R     pointer to unit input on keyboard
        const char * const  secn_p     R     pointer to secn input on keyboard

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		All input args should be checked to be non-NULL pointers.

        Uses local dbshellBlkPad() versus public dbgetBlkPad() to right pad 
		each input arg, as each input argument will not necessarily be null terminated.
        Calls dbgetMeta()

  Ret:  none
----------------------------------------------------------------------*/
void dbshellDump (const char * const prim_p, 
				 const char * const unit_p, 
				 const char * const secn_p)
{ 
  char prim_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char unit_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char secn_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char primUnitSecn_a[SZ_NAME_PUS]; /* contains concatenated prim  unit secn name */
  vmsstat_t iss = MICR_OKOK;

  if (epicsThreadGetId(EXEC_NAME)){

	if (slcGetDbExists()){

	  if (! prim_p ){
		printf("Missing arguments\n");
		printf("Usage: slcdbDump prim <unit> <secn>, where unit and secn are optional\n");
		return;
	  } 
	  
	  strncpy (prim_pa, prim_p, SZ_NAME_SGL);	
	  if (unit_p) 
		strncpy (unit_pa, unit_p, SZ_NAME_SGL);
	  if (secn_p) 
		strncpy (secn_pa, secn_p, SZ_NAME_SGL);

	  dbshellBlkPad((char *)prim_pa);
      dbshellBlkPad((char *)unit_pa);
      dbshellBlkPad((char *)secn_pa);
	  DEBUGPRINT(DP_INFO,dbshellFlag,
				 ("After pad: prim=%s; unit=%s; secn=%s \n", prim_pa, unit_pa, secn_pa));   
	  if (!SUCCESS(iss=dbdumpDump(prim_pa, unit_pa, secn_pa))){
		if (iss==DB_NOTALLUNIT) {
		  printf ("No units found for primary: %.4s\n",prim_pa);
		}
		else {
		  if (iss==DB_BADPRIM) {
			printf ("No such primary found: %.4s\n",prim_pa);
		  }
		  else { 
			if (iss==DB_NOSUCHNODE){
			  dbgetCatPUSName(prim_pa, unit_pa, secn_pa, (char *)&primUnitSecn_a);
			  printf ("No dictionary entry for '%.12s'\n",primUnitSecn_a);
			}
		  }
		}
	  }
	} else printf ("Command invalid; database does not exist. \n"); 
  } else printf("Error - issue 'slcStart' first.\n");
}
/*-----------------------------------------------------------------------
  Name: dbshellGetUnits

  IOC SHELL COMMAND: 'slcdbGetUnits'

  Abs:  Gets units for given primary input on command line at 
        prompt.
        

  Args: Type		        Name	  Access	   Description
	    ---------------   --------  ---------- ----------------------------
        const char * const  prim_p     R     pointer to prim input on keyboard

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		All input args should be checked to be non-NULL pointers.


  Ret:  none
----------------------------------------------------------------------*/
void dbshellGetUnits (const char * const prim_p)
{
  dbheader_ts * dbunits_ps = NULL;
  vmsstat_t    iss = MICR_OKOK;
  unsigned long cnt;
  unsigned long * next_p;

  if (epicsThreadGetId(EXEC_NAME)!= NO_THREAD )
  {
	if (slcGetDbExists()){

	  if (!prim_p ){
		printf("Missing argument -  'slcdbGetUnits prim' \n");
		return;
	  }
	  if (!SUCCESS(iss = dblistalloc(4,13,(const dbheader_ts ** const)&dbunits_ps)))
		{
		  DEBUGPRINT(DP_ERROR,dbshellFlag,("list error \n"));
		  printf ("Error with list\n");
		  return;		  
		}
	  if (!SUCCESS(iss = dblunits(prim_p, &dbunits_ps)))
		{
		  if (iss==DB_BADPRIM)
			printf ("No such primary found: %.4s\n",prim_p);
		  else printf ("Error with dblunits/ dblistalloc\n");
		  return;
		}	  
	  if (!dbunits_ps) {
		DEBUGPRINT(DP_ERROR,dbshellFlag,("list error \n"));
		printf ("Error with list\n");
		return;
	  }
	  /* print out all the units */
	  next_p = (unsigned long *)(dbunits_ps) + 4;
      for (cnt=0; (cnt < dbunits_ps->cur/WSZ_NAME_SGL); cnt++){
		printf("%.4s\n", (char *)next_p);
		next_p++;
	  }	  
	  dblistfree(dbunits_ps);
      
	}
	else {
	  printf ("Command invalid; database does not exist. \n"); 
	}
  } else printf("Error - issue 'slcStart' first.\n");
}
/*-----------------------------------------------------------------------
  Name: dbshellGetExists

  Abs:  Gets dbExists global flag to tell user whether the database has
        been downloaded.
        
  Args: None

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Issues sclGetDbExists and outputs True or False based on its value.

  Ret:  none
----------------------------------------------------------------------*/
void dbshellGetExists (void)
{

  if (epicsThreadGetId(EXEC_NAME)){

	if (slcGetDbExists())
	  printf ("True\n");
	else printf  ("False\n");

  } else printf("Error - issue 'slcStart' first.\n");
}
/*-----------------------------------------------------------------------
  Name: dbshellGetVersion

  IOC SHELL COMMAND: 'slcdbGetVersion'

  Abs:  Gets dbVersion_a global to output to user 

  Args: None

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		Issues dbgetVersion and outputs its value.

  Ret:  none
----------------------------------------------------------------------*/
void dbshellGetVersion (void)
{

  if (epicsThreadGetId(EXEC_NAME)){
	
	if (slcGetDbExists()){
	  printf ("Downloaded Version from DBEX = %s\n", dbgetVersion());
	}
	else {
	  printf ("Command invalid; database does not exist. \n"); 
	}
  } else printf("Error - issue 'slcStart' first.\n");
}
/*-----------------------------------------------------------------------
  Name: dbshellUpdate

  IOC SHELL COMMAND: 'slcdbUpdate jobName'
        Note: After 'slcdbEdit'(s), issue 'slcdbUpdate IOCS' to send to Alpha,
		where IOCS = IOC shell jobName, since the previous slcdbEdit's placed data
		 into ioc shell's hilo update buffer

  Abs:  Causes updates for jobName to be sent to DBEX

  Args: None

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		Checks for valid jobName
		If jobName = ALL*, calls dbupdateAll()
		otherwise calls dbupdateNoWait(), since the ioc shell shouldn't have
		  to ever wait

  Ret:  none
----------------------------------------------------------------------*/
void dbshellUpdate (const char * const jobName)
{
  slcJob_te  jobId_e;

  if (epicsThreadGetId(EXEC_NAME)){

	if (slcGetDbExists()){
	  if (!jobName ){
		printf("Missing argument -  'slcdbUpdate jobName' \n");
		printf("Valid jobNames: %s, %s, %s, %s, or %s  \n", 
			   IOCS_JOB_NAME, CSTR_JOB_NAME, TIME_JOB_NAME,BPMO_JOB_NAME, MGNT_JOB_NAME  );
		return;
	  }
	  if (strncmp(jobName,"ALL*", 4)==0){
		printf ("Update command being sent for ST2/ ST3 data, all jobs...\n"); 
		dbupdateAll();
	  }
	  else 
	  {
		jobId_e = slcGetJobEnumByName (jobName);  
		if (jobId_e == invalidJob) {
		  printf ("Error: :Invalid job name\n");
		  return;
		}
		printf ("Update command being sent for ST2/3, job %s; job enum = %d...\n", 
				jobName, jobId_e);   
		dbupdateNoWait(jobId_e);
	  }
  }
	else {
	  printf ("Command invalid; database does not exist. \n"); 
	}
} else printf("Error - issue 'slcStart' first.\n");
}
/*-----------------------------------------------------------------------
  Name: dbshellSetHashSize

  IOC SHELL COMMAND: 'slcdbSetHashSize int'

  Abs:  Adjust this ioc's hash table size via command line
        

  Args: Type		   Name	     Access	   Description
        ---------   --------  ---------- ----------------------------
		int           hashSize     R     hash tbl size input via epics shell

  Rem:  

  Ret:  none
----------------------------------------------------------------------*/
void dbshellSetHashTableSize(int hashSize)
{ 
#define NSIZES 9
  static int allowSize[NSIZES]={256, 512, 1024, 2048, 4096, 8192, 16386, 32768, 65636};
  int loopi;
  epicsBoolean found = epicsFalse;

  for (loopi=0; loopi<NSIZES; loopi++){
	if (hashSize==allowSize[loopi]){
	  found = epicsTrue;
      dbdictSize = hashSize;
	}
  }  
  if (found) printf ("hash table size = %d \n", dbdictSize);
  else 
	printf ("hash table size must be a power of 2, in the set: \n 256, 512, 1024, 2048, 4096, 8192, 16386, 32768, 65636\n"); 

}
/*-----------------------------------------------------------------------
  Name: dbshellBlkPad

  Abs:  Pads given 4 char input string with trailing blanks,
        rendering a left justified, blank padded, non-null terminated char[4]
        

  Args: Type		   Name	     Access	   Description
	---------------   --------  ---------- ----------------------------
        char *            name_pa     W     pointer to 4 char primary, secn, unit name


  Rem:  Input string may be null-terminated (unit will be if <4 chars, as it
        is converted from int to ASCII via "sprintf()". Input string should be
        4 chars long.

  Ret:  none
----------------------------------------------------------------------*/
LOCAL void dbshellBlkPad (char *name_p)
{ 
  char name[SZ_NAME_SGL + 1] = {' ',' ',' ',' ', '\0'};
  size_t cnt = 0;

  /* diff pattern than dbgetBlkPad so as to echo lower case letters back to user */
  cnt = strspn(name_p,"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789*abcdefghijklmnopqrstuvwxyz");
  if (cnt<=SZ_NAME_SGL)
	memcpy (&name, name_p, cnt);
  else 
	memcpy (&name, name_p, SZ_NAME_SGL);
  memcpy (name_p, &name, SZ_NAME_SGL); 
}
/*-----------------------------------------------------------------------
  Name: dbshellEdit

  IOC SHELL COMMAND: 'slcdbEdit prim unit secn value'
  Note: Use for supertype 2 or 3 data. For ST1, use 'slcdbEditST1'.

  Abs:  Edit a primary unit secondary that has count =1 scalar value.
                

  Args: Type		        Name	  Access	   Description
	    ---------------   --------  ---------- ----------------------------
        const char * const  prim_p     R     pointer to prim input on keyboard
        const char * const  unit_p     R     pointer to unit input on keyboard
        const char * const  secn_p     R     pointer to secn input on keyboard
		const char * const  val_p      R     pointer to new edit value

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		All input args should be checked to be non-NULL pointers.

        Uses local dbshellBlkPad() versus public dbgetBlkPad() to right pad 
		each input arg, as each input argument will not necessarily be null terminated.
        After processing input args, calls dbeditEdit with input args.
		If the PUS is an array, only the first element will be modified.
		Editing arrays is not supported.

  Ret:  none
----------------------------------------------------------------------*/
  void dbshellEdit  (const char * const prim_p, const char * const unit_p, 
					 const char * const secn_p, const char * const val_p)
{ 
  char prim_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char unit_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char secn_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char primUnitSecn_a[SZ_NAME_PUS]; /* contains concatenated prim  unit secn name */
  vmsstat_t iss = MICR_OKOK;
  char format = ' '; 
  int  width = 0;
  int  count = 0;
  int  supn = 0;

  if (epicsThreadGetId(EXEC_NAME)){

	if (slcGetDbExists()){

	  if (! (prim_p && unit_p && secn_p && val_p )) {
		printf("Missing arguments\n");
		printf("Usage: slcdbEdit prim unit secn value\n");
		return;
	  } 
	  DEBUGPRINT(DP_INFO,dbshellFlag,
		     ("dbshellEdit: value entered =%s \n", val_p));   
	  
	  strncpy (prim_pa, prim_p, SZ_NAME_SGL);   
	  strncpy (unit_pa, unit_p, SZ_NAME_SGL);
	  strncpy (secn_pa, secn_p, SZ_NAME_SGL);

	  dbshellBlkPad((char *)prim_pa);
	  dbshellBlkPad((char *)unit_pa);
      dbshellBlkPad((char *)secn_pa);

	  DEBUGPRINT(DP_INFO,dbshellFlag,
		     ("dbshellEdit: After pad: prim=%s; unit=%s; secn=%s \n", prim_pa, unit_pa, secn_pa));  

	  if (! SUCCESS(iss = dbgetMeta(prim_pa, unit_pa, secn_pa, &format, &width, &count, &supn))) {
		  printf("Error: %.4s %.4s %.4s not found\n", prim_pa, unit_pa, secn_pa);
		  return;
	  }
	  if (supn==1){
		  printf("Error: Supertype 1 data; use the 'slcdbEditST1' command.\n");
		  return;
	  }

	  if (!SUCCESS(iss=dbeditEdit(prim_pa, unit_pa, secn_pa, val_p))){
		if (iss==DB_NOTALLUNIT) {
		  printf ("No units found for primary: %.4s\n",prim_pa);
		}
		else {
		  if (iss==DB_BADPRIM) {
			printf ("No such primary found: %.4s\n",prim_pa);
		  }
		  else { 
			if (iss==DB_NOSUCHNODE){
			  dbgetCatPUSName(prim_pa, unit_pa, secn_pa, (char *)&primUnitSecn_a);
			  printf ("No dictionary entry for '%.12s'\n",primUnitSecn_a);
			}
		  }
		}
	  }
	} else printf ("Command invalid; database does not exist. \n"); 
  } else printf("Error - issue 'slcStart' first.\n");
}
/*-----------------------------------------------------------------------
  Name: dbshellEditST1

  IOC SHELL COMMAND: 'slcdbEditST1 prim unit secn value'
  Note: used for supertype 1 data only. For testing only.

  Abs:  Edit a Supertype 1 primary unit secondary that has count =1 scalar value
                
  Args: Type		        Name	  Access	   Description
	    ---------------   --------  ---------- ----------------------------
        const char * const  prim_p     R     pointer to prim input on keyboard
        const char * const  unit_p     R     pointer to unit input on keyboard
        const char * const  secn_p     R     pointer to secn input on keyboard
		const char * const  val_p      R     pointer to new edit value

  Rem:  Checks to make sure Exec thread is up and running in order for global
        variables (memory)to be valid.
		Checks to see if the database has been downloaded to see if it should
        bother accessing.
		All input args should be checked to be non-NULL pointers.

        Uses local dbshellBlkPad() versus public dbgetBlkPad() to right pad 
		each input arg, as each input argument will not necessarily be null terminated.
        After processing input args, calls dbeditEdit with input args.
		If the PUS is an array, only the first element will be modified.
		Editing arrays is not supported.

		For TESTING ONLY!

  Ret:  none
----------------------------------------------------------------------*/
  void dbshellEditST1  (const char * const prim_p, const char * const unit_p, 
						const char * const secn_p, const char * const val_p)
{ 
  char prim_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char unit_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char secn_pa[SZ_NAME_SGL+1]= {' ',' ',' ',' ','\0'};
  char primUnitSecn_a[SZ_NAME_PUS]; /* contains concatenated prim  unit secn name */
  vmsstat_t iss = MICR_OKOK;
  char format = ' '; 
  int  width = 0;
  int  count = 0;
  int  supn = 0;

  if (epicsThreadGetId(EXEC_NAME)){

	if (slcGetDbExists()){

	  if (! (prim_p && unit_p && secn_p && val_p )) {
		printf("Missing arguments\n");
		printf("Usage: slcdbEdit prim unit secn value\n");
		return;
	  } 
	  DEBUGPRINT(DP_INFO,dbshellFlag,
		     ("dbshellEdit: value entered =%s \n", val_p));   
	  
	  strncpy (prim_pa, prim_p, SZ_NAME_SGL);   
	  strncpy (unit_pa, unit_p, SZ_NAME_SGL);
	  strncpy (secn_pa, secn_p, SZ_NAME_SGL);

	  dbshellBlkPad((char *)prim_pa);
	  dbshellBlkPad((char *)unit_pa);
      dbshellBlkPad((char *)secn_pa);

	  DEBUGPRINT(DP_INFO,dbshellFlag,
		     ("dbshellEdit: After pad: prim=%s; unit=%s; secn=%s \n", prim_pa, unit_pa, secn_pa));  

	  if (! SUCCESS(iss = dbgetMeta(prim_pa, unit_pa, secn_pa, &format, &width, &count, &supn))) {
		  printf("Error: %.4s %.4s %.4s not found\n", prim_pa, unit_pa, secn_pa);
		  return;
	  }
	  if (!(supn==1)){
		  printf("Error: NOT Supertype 1 data; use the 'slcdbEdit' command.\n");
		  return;
	  }

	  if (!SUCCESS(iss=dbeditEdit(prim_pa, unit_pa, secn_pa, val_p))){
		if (iss==DB_NOTALLUNIT) {
		  printf ("No units found for primary: %.4s\n",prim_pa);
		}
		else {
		  if (iss==DB_BADPRIM) {
			printf ("No such primary found: %.4s\n",prim_pa);
		  }
		  else { 
			if (iss==DB_NOSUCHNODE){
			  dbgetCatPUSName(prim_pa, unit_pa, secn_pa, (char *)&primUnitSecn_a);
			  printf ("No dictionary entry for '%.12s'\n",primUnitSecn_a);
			}
		  }
		}
	  }
	} else printf ("Command invalid; database does not exist. \n"); 
  } else printf("Error - issue 'slcStart' first.\n");
}
