/*=============================================================================
 
  Name: devSLC.c

  Abs: This file contains all device support for the slc-aware IOC base
  Rem: When adding support for a new record, rec, (such as rec=stringin, 
       as was done 18-APR-2005) do the following:
	   1) add init_rec and read_rec into this file
       2) add PV name into the recData[] array, as well as pointer to diagnostic element
	       (diag element is a var setup in slc code - include approp. .h file)
       3) make new saIocApp/Db/slcrec.db file
       4) edit saIocApp/Db/slc.template to add in record file db/slcrec.db and
	       substitution pattern
	   5) in exec/devSLC.dbd, add any new record: 
	       device(rec,INST_IO,devrecSLC, "SLC")
       6) gmake clean, gmake

	   When adding a diagnostic that already has device support in this file, 
	     simply 1) add the diagnostic to the appropriate recData[] array below and 
         2) add into saIocApp/Db/slc.template
		 Then do a gmake clean, gmake

	  To test, boot ioc, look for debug/error messages (devSLCFlag = DP_DEBUG). If all is
	     well, do a "dbl" at the epics prompt to see your new PV. 
		 On another terminal, do a probe of the PV to see if probe connects to it. 
         Make sure to include the ioc host IP into your probe session's EPICS_CA_ADDR_LIST.
		 (for XL03, on noric06, do a nslookup for noric06)

  Auth: 13-apr-2005, dfairley 
  Rev:  18-APR-2005, drogind - added above remarks  

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        13-apr-2005, dfairley:
		   created
        18-APR-2005, drogind:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "dbDefs.h"
#include "dbAccess.h"
#include "recSup.h"
#include "recGbl.h"
#include "alarm.h"
#include "devSup.h"
#include "devLib.h"
#include "aiRecord.h"
#include "biRecord.h"
#include "aoRecord.h"
#include "eventRecord.h"
#include "longinRecord.h"
#include "stringinRecord.h"
#include "epicsExport.h"
#include "slcGlob.h"     /* execDiagnostics_s */
#include "slcMain.h"     /* slcRestart, slcStop */
#include "dbRecv.h"      /* dbRecvDiags_ts */
#include "dbGlob.h"      /* dbRecvDiags_s */

#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int devSLCFlag = DP_ERROR;
#endif

/*=============================================================================
 
  Name: ai record device support for the slc-aware IOCs

  Abs:  This is the device support for ai records used by the slc-aware IOCs.
        the SLC ai record can be used to report the value of any slc-aware 
		global variable of the type "double".  It is mainly used to report 
		diagnostic values.
  
  Args:
 
  Rem:  The device support routines needed for the ai record

  Side:

  Ret:  
  
=============================================================================*/
/*======================== AI RECORD ======================================*/

/* create the dset for the slc ai record */
LOCAL long init_ai(struct aiRecord *prec);
LOCAL long read_ai(struct aiRecord *prec);

struct {
  long		number;
  DEVSUPFUN	report;
  DEVSUPFUN	init;
  DEVSUPFUN	init_record;
  DEVSUPFUN	get_ioint_info;
  DEVSUPFUN	read_ai;
  DEVSUPFUN special_linconv;
} devAiSLC = {
	6,
	NULL,
	NULL,
	init_ai,
	NULL,
	read_ai,
    NULL
};
epicsExportAddress(dset, devAiSLC);

LOCAL long init_ai( struct aiRecord *prec) 
{
  static struct {
	char	*string;
	double	*data_p;
  } aiData[] = {
	{ "EXRESTARTCNT",	&execDiagnostics_s.numStarts },
	{ "EXSTOPCNT",	    &execDiagnostics_s.numStops  },
    { "DBEXUPDCNT",     &dbRecvDiags_s.dbexUpdCnt }, 
    { "DBEXUPDDROP",    &dbRecvDiags_s.dbexUpdDropCnt }, 
    { "DBEXDOWNCNT",    &dbRecvDiags_s.dbexDownCnt },
	{ NULL,		0 }
  };

  char *inString;
  int i;

  /* ai.inp must be INST_IO */
  if (prec->inp.type != INST_IO) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: Bad INP field type in ai record; !=INST_IO");
	goto error;
  }

  inString = ((struct instio *)&(prec->inp.value))->string;
  DEBUGPRINT(DP_DEBUG,devSLCFlag,("AI:the inp.string is %s\n",inString));

  /* get the pointer to the data we want*/
  prec->dpvt=NULL;
  for (i=0; aiData[i].string != NULL; i++)
	if (strcmp(inString, aiData[i].string) == 0)
	    prec->dpvt = (void *)aiData[i].data_p;

  if (prec->dpvt == NULL) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: data for ai record does not exist.");

	goto error;
  }
  return 0;

error:
  return S_db_badField;
}

LOCAL long read_ai(struct aiRecord *prec)
{
	if (prec->dpvt == NULL) {
	    prec->pact = TRUE;
	    return S_dev_noDevice;
	}

	prec->val = *((double*)prec->dpvt);
	prec->udf = FALSE;
	return 2; /* return 2 to indicate value put directly into val */
}

/*=============================================================================
 
  Name: longin record device support for the slc-aware IOCs

  Abs:  This is the device support for longin records used by the slc-aware IOCs.
        the SLC ai record can be used to report the value of any slc-aware 
		global variable of the "unsigned long" data type.  It is mainly used to report 
		diagnostic values.
  
  Args:
 
  Rem:  The device support routines needed for the longin record

  Side:

  Ret:  
  
=============================================================================*/
/*======================== LONGIN RECORD ======================================*/

/* create the dset for the slc longin record */
LOCAL long init_longin(struct longinRecord *prec);
LOCAL long read_longin(struct longinRecord *prec);

struct {
	long		number;
	DEVSUPFUN	report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_longin;
} devLonginSLC = {
	5,
	NULL,
	NULL,
	init_longin,
	NULL,
	read_longin
};
epicsExportAddress(dset, devLonginSLC);

LOCAL long init_longin( struct longinRecord *prec) 
{
  static struct {
	char		    *string;
	unsigned long	*data_p;
  } longinData[] = {
	/*	{ "numStarts",	&execDiagnostics_s.numStarts },*/
		/*{ "numStops",	&execDiagnostics_s.numStops },*/
	{ NULL,		0 }
  };

  char *inString;
  int i;

  /* longin.inp must be a INST_IO */
  if (prec->inp.type != INST_IO) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: Bad INP field type for longin record; !=INST_IO");
	goto error;
  }

  inString = ((struct instio *)&(prec->inp.value))->string;
  DEBUGPRINT(DP_DEBUG,devSLCFlag,("LONGIN:the inp.string is %s\n",inString));


  /* get the pointer to the data we want*/
  prec->dpvt=NULL;
  for (i=0; longinData[i].string != NULL; i++)
	if (strcmp(inString, longinData[i].string) == 0)
	    prec->dpvt = (void *)longinData[i].data_p;

  if (prec->dpvt == NULL) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: data for longin record does not exist");

	goto error;
  }
  return 0;

error:
  return S_db_badField;
}

LOCAL long read_longin(struct longinRecord *prec)
{
	if (prec->dpvt == NULL) {
	    prec->pact = TRUE;
	    return S_dev_noDevice;
	}

	prec->val = *((unsigned long*)prec->dpvt);
	return 0;
}

/*=============================================================================
 
  Name: event record device support for the slc-aware IOCs

  Abs:  This is the device support for event records used by the slc-aware IOCs.
        the SLC event record can be used to execute command functions on
		any slc-aware IOC. 
  
  Args:
 
  Rem:  The device support routines needed for the event record

  Side:

  Ret:  
  
=============================================================================*/
/*======================== EVENT RECORD ======================================*/
/* Create the dset for event records  */
LOCAL long init_event(struct eventRecord *prec);
LOCAL long read_event(struct eventRecord *prec);

struct {
	long		number;
	DEVSUPFUN	report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_event;
} devEventSLC = {
	5,
	NULL,
	NULL,
	init_event,
	NULL,
	read_event
};
epicsExportAddress(dset, devEventSLC);

LOCAL long init_event( struct eventRecord *prec )
{

  static struct {
	char	*string;
	void    (*func_p)();
  } eventData[] = {
	{ "EXRESTART",  slcRestart }, 
	{ "EXSTOP" ,    slcStop    }, 
	{ NULL,		0      }
  };

  char *inString;
  int i;

  /* longin.inp must be a INST_IO */
  if (prec->inp.type != INST_IO) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: Bad INP field type for longin record; !=INST_IO");
	goto error;
  }

  inString = ((struct instio *)&(prec->inp.value))->string;
  DEBUGPRINT(DP_DEBUG,devSLCFlag,("EVENT:the inp.string is %s\n",inString));


  /* get the pointer to the data we want*/
  prec->dpvt=NULL;
  for (i=0; eventData[i].string != NULL; i++)
	if (strcmp(inString, eventData[i].string) == 0)
	    prec->dpvt = (void *)eventData[i].func_p;

  if (prec->dpvt == NULL) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: data for event record does not exist");

	goto error;
  }
  return 0;

error:
  return S_db_badField;
}

LOCAL long read_event(struct eventRecord *prec)
{
  void (*func)() = NULL;

  if (prec->dpvt == NULL) {
	prec->pact = TRUE;
	return S_dev_noDevice;
  }
  /* execute the function */
  func = prec->dpvt;
  (*func)();
  prec->udf = FALSE;
  return 0;
}
/*=============================================================================
 
  Name: stringin record device support for the slc-aware IOCs

  Abs:  This is the device support for stringin records used by the slc-aware IOCs.
        the SLC ai record can be used to report the value of any slc-aware 
		global variable of the "char *" string data type.  It is mainly used to report 
		diagnostic values.
  
  Args:
 
  Rem:  The device support routines needed for the stringin record

  Side:

  Ret:  
  
=============================================================================*/
/*======================== STRINGIN RECORD ======================================*/

/* create the dset for the slc stringin record */
LOCAL long init_stringin(struct stringinRecord *prec);
LOCAL long read_stringin(struct stringinRecord *prec);

struct {
	long		number;
	DEVSUPFUN	report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_stringin;
} devStringinSLC = {
	5,
	NULL,
	NULL,
	init_stringin,
	NULL,
	read_stringin
};
epicsExportAddress(dset, devStringinSLC);

LOCAL long init_stringin( struct stringinRecord *prec) 
{
  static struct {
	char		    *string;
	char	        *data_p;
  } stringinData[] = {
	{ "DBVERSION",	  dbRecvDiags_s.dbVersion_a },
	{ NULL,		0 }
  };

  char *inString;
  int i;

  /* stringin.inp must be a INST_IO */
  if (prec->inp.type != INST_IO) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: Bad INP field type for stringin record; !=INST_IO");
	goto error;
  }

  inString = ((struct instio *)&(prec->inp.value))->string;
  DEBUGPRINT(DP_DEBUG,devSLCFlag,("STRINGIN:the inp.string is %s\n",inString));


  /* get the pointer to the data we want*/
  prec->dpvt=NULL;
  for (i=0; stringinData[i].string != NULL; i++)
	if (strcmp(inString, stringinData[i].string) == 0)
	    prec->dpvt = (void *)stringinData[i].data_p;

  if (prec->dpvt == NULL) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: data for stringin record does not exist");

	goto error;
  }
  return 0;

error:
  return S_db_badField;
}

LOCAL long read_stringin(struct stringinRecord *prec)
{
	if (prec->dpvt == NULL) {
	    prec->pact = TRUE;
	    return S_dev_noDevice;
	}

	strcpy((char *)prec->val,(char *)prec->dpvt);

	return 0;
}

/*======================== BI RECORD ======================================*/

/* create the dset for the slc bi record */
LOCAL long init_bi(struct biRecord *prec);
LOCAL long read_bi(struct biRecord *prec);

struct {
  long		number;
  DEVSUPFUN	report;
  DEVSUPFUN	init;
  DEVSUPFUN	init_record;
  DEVSUPFUN	get_ioint_info;
  DEVSUPFUN	read_bi;
  DEVSUPFUN special_linconv;
} devBiSLC = {
	6,
	NULL,
	NULL,
	init_bi,
	NULL,
	read_bi,
    NULL
};
epicsExportAddress(dset, devBiSLC);

LOCAL long init_bi( struct biRecord *prec) 
{
  static struct {
	char	*string;
	unsigned short	*data_p;
  } biData[] = {
    { "DBEXSTATE",      &dbRecvDiags_s.dbexUpState }, 
    { NULL,		0 }
  };

  char *inString;
  int i;

  /* bi.inp must be INST_IO */
  if (prec->inp.type != INST_IO) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: Bad INP field type in bi record; !=INST_IO");
	goto error;
  }

  inString = ((struct instio *)&(prec->inp.value))->string;
  DEBUGPRINT(DP_DEBUG,devSLCFlag,("BI:the inp.string is %s\n",inString));

  /* get the pointer to the data we want*/
  prec->dpvt=NULL;
  for (i=0; biData[i].string != NULL; i++)
	if (strcmp(inString, biData[i].string) == 0)
	    prec->dpvt = (void *)biData[i].data_p;

  if (prec->dpvt == NULL) {
	recGblRecordError(S_db_badField,(void *)prec,
					  "devSLC: data for bi record does not exist.");

	goto error;
  }
  return 0;

error:
  return S_db_badField;
}

LOCAL long read_bi(struct biRecord *prec)
{

	if (prec->dpvt == NULL) {
	    prec->pact = TRUE;
	    return S_dev_noDevice;
	}
	/*DEBUGPRINT(DP_DEBUG,devSLCFlag,("AI:the value of memory = %hd\n", 
									dbRecvDiags_s.dbexUpState) );
	DEBUGPRINT(DP_DEBUG,devSLCFlag,("AI:the value of dpvt = %hd\n", 
	                                 *((unsigned short *)prec->dpvt)) ); */

	prec->val = *((unsigned short *)prec->dpvt);
	prec->udf = FALSE;
	return 2; /* return 2 to indicate value put directly into val */
}
