/*=============================================================================
 
  Name: slcExec.c
        slcExecThread - the main routine of the SLC Executive Thread 
		anyJobActive
		reqdThreadsActive
		startDatabaseService
		waitForDbThreadsActive
		dbThreadsActive
		waitForDbExists
		waitForAsyncReady
		waitForMgntAsyncReady
		startHdlrThreads
		waitForHdlrThreadsActive
		hdlrThreadsActive
		waitForRecvThreadActive
		waitForThreadsTerminated
		allThreadsTerminated
		setStopFlagsTrue
		setStopFlagsFalse
		startThreads
		stopThreads
		closeSocketConnections
		cleanup

  Abs:  slcExec contains the slcExec thread.  
    See the SLC Executive and slc-IOC Initialization Design document

  Auth: 13-dec-2004, diane fairley (DFAIRLEY): created from slcMain.c
  Rev:  02-feb-2005 Debbie Rogind


-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)
        10-Apr-2004, Ron MacKenzie (ronm):
                   slcAsyncGetJMSK argument changed to reference
        07-Apr-2005, Ron MacKenzie (ronm):
                   Call slcAsync utility to access asyncExists. 
        25-Jan-2005, Ron MacKenzie (ronm):
                   Add cstrAsync startup   
        13-dec-2004, diane fairley (DFAIRLEY):
		   split slcExec thread code out of slcMain into this file
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int slcExecFlag = DP_DEBUG;
#include <stdio.h>
#endif


/* c library includes */
#include	<string.h>        /* memcpy */

/* epics includes */
#include    <freeList.h>      /* freeListFree */


/* slc common includes */
#include    "slcGlob.h"       /* globals, vmsstat_t, SUCCESS, NULL, etc MUST BE FIRST INCLUDE*/
#include    "msgTypes.h"      /* msgmail_ts */
#include    "tcpTypes.h"      /* MAX_SOCKETS */
#include    "slcMsg.h"        /* slcIOC internal msg function codes */
#include    "cstrAsync.h"     /* slcIOC CSTR async job */
#include    "msgfunc.hc"      /* FUNC_TEST job code */
#include    "micrdef.hc"      /* MICR_EXNET error codes */
#include    "db_micro_struc.h"/* stuctures for db tasks */
#include    "dbmsgmail.h"     /* dbex universal msg struc */

/* slc utility prototypes */
#include    "msgQUtil.h"      /* message queue utilities */
#include    "slcCmlog.h"      /* slcCmlogLogMsg */
#include    "slcAsync.h"      /* slcAsyncGetJMSK */
#include    "tcpUtil.h"       /* tcpCloseSocket */

/* thread prototypes  */
#include    "msgRecv.h"    /* msg receive thread */
#include    "msgSend.h"    /* msg send thread */
#include    "msgHdlr.h"    /* msg handler thread */
#include    "cstrHdlr.h"   /* cstr handler thread */
#include    "cstrAsync.h"  /* cstr async thread */
#include    "dbRecv.h"     /* db receive thread */
#include    "dbSend.h"     /* db send thread */
#include    "dbHdlr.h"     /* db handler thread */
#include    "bsacHdlr.h"   /* bsac hdlr thread */
#include    "unsupportedHdlr.h"   /* unsupported threads */
#include    "mgntHdlr.h"   /* mgnt hdlr thread */
#include    "mgntAsync.h"  /* mgnt async thread */

/* internal function prototypes */
LOCAL   epicsBoolean     anyJobActive(void);
LOCAL   epicsBoolean     startDatabaseService(void);
LOCAL   epicsBoolean     waitForDbThreadsActive(void);
LOCAL   epicsBoolean     dbThreadsActive(void);
LOCAL   epicsBoolean     waitForDbExists(void);
LOCAL   epicsBoolean     waitForAsyncReady(void);
LOCAL   epicsBoolean     waitForMgntAsyncReady(void);
LOCAL   epicsBoolean     startHdlrThreads(void);
LOCAL   epicsBoolean     waitForBsacHdlrActive(void);
LOCAL   epicsBoolean     bsacHdlrActive(void);
LOCAL   epicsBoolean     waitForMgntHdlrActive(void);
LOCAL   epicsBoolean     mgntHdlrActive(void);
LOCAL   epicsBoolean     waitForMicroHdlrActive(void);
LOCAL   epicsBoolean     microHdlrActive(void);
LOCAL   epicsBoolean     waitForHdlrThreadsActive(void);
LOCAL   epicsBoolean     hdlrThreadsActive(void);
LOCAL   epicsBoolean     waitForRecvThreadActive(void);
LOCAL   epicsBoolean     allThreadsTerminated(void);
LOCAL   epicsBoolean     nonDBThreadsTerminated(void);
LOCAL   epicsBoolean     waitForNonDBThreadsTerminated(void);
LOCAL   epicsBoolean     waitForThreadsTerminated(void);
LOCAL   void             setStopFlagsTrue( void);
LOCAL   void             setStopFlagsFalse( void);
LOCAL   epicsBoolean     startThreads(void);
LOCAL   epicsBoolean     stopThreads(void);
LOCAL   void             closeSocketConnections(void);
LOCAL   void             cleanup(void);
/*=============================================================================
 
  Name: slcExecThread

  Abs:  Start all the SLC aware IOC threads.

  
  Args: None

  Rem: The slcExecThread handles the IOC_RESTART and IOC_STOP messages
       IOC_RESTART: first attempts to stop all threads except slcExec, 
	                 then starts all slcIOC threads
       IOC_STOP: stops all running slcIOC threads, except slcExec

  Side: The slcExec threads attaches to the socket library, creates a message queue
 
  Ret:  none
        generates EXNET_INITFAIL if cannot attach to socket lib
        generates errors from message queue utilities upon message queue failures
=============================================================================*/
void slcExecThread(void)
{
  int                 rv;
  vmsstat_t           iss;
  msgmail_ts*         msg_ps = NULL;
  epicsBoolean        go = epicsTrue;
  slcThread_ts*       slcExec_ps = &slcThreads_as[slcExec];
  unsigned short      vnum = 9;

  /* attach to the OSI sockets */
  if ((rv=osiSockAttach()) < 0) {
    iss = MICR_EXNET_INITFAIL;
    slcCmlogLogMsg(iss, "EXNET_INITFAIL",
				   "Socket library initialization failed. Shutting down.\n");
    slcExec_ps->tid_ps = NO_THREAD;
	return;
  }

  /* create this thread's message queue */
  if ( !SUCCESS(messageQCreate(slcExec, MAX_QUEUED_MSGS, MAX_MSGPTR_BYTES)) )
  {
    osiSockRelease();
    slcExec_ps->tid_ps = NO_THREAD;
	return;
  }

  /* set this thread to active */
  slcExec_ps->active = epicsTrue;
  DEBUGPRINT(DP_INFO,slcExecFlag, ("slcExec is active now\n"));

       
  /* now we wait at the message queue for thread status messages */
  go = epicsTrue; /* go is set to false only when slcExec must exit */
  while (go)
  {
	/* wait for a message at the queue */
	if ( SUCCESS(msgQMsgRecv(slcExec, NULL, (void*)&msg_ps)) )
	{

	  /* if a message is received, it will be a thread stop or start msg */
	  /* set the thread.active flag accordingly, */
	  /* no need to swap bytes for internal messages */
	  switch ( msg_ps->msgheader.func)
	  {
	    case IOC_RESTART:
		  DEBUGPRINT(DP_INFO,slcExecFlag, 
					 ("EXEC:executing RESTART command.\n"));
		  if ((go = stopThreads()) )
		  {
			if ( (go = startThreads()) )
			{
			  iss = MICR_IPLOK;
			  slcCmlogLogMsg(iss, "IPLOK","%s v.%u started and at your service\n", 
							 slcGetIOCName(), vnum);
			  /* all necessary threads are going, start and continue our main loop */
			}
		  }
                  /* Reset flag that tells cstrHdlr that this was a SCP IPL */
                  slcSetNotify(epicsFalse);
		  break;
	    case IOC_STOP:
		  DEBUGPRINT(DP_INFO,slcExecFlag, 
					 ("EXEC:executing STOP command.\n"));
	
		  go = stopThreads();
		  break;
	  } 

	  /* release the message buffer */
	  msgQFreeBuffer(sizeof(msgmail_ts), (void *)msg_ps);

	} /* if message in queue */

    /* if no jobs are active we have shut down, 
	   close the sockets and wait for a "start" msg
	*/
    if (!anyJobActive())
	{
	  /* if we got here, all the job threads are done */
	  DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC: all threads have stopped\n"));
	  DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC: closing sockets\n"));
	  closeSocketConnections();
	}
	  
  } /*while go */

  /* if we got here somehow there's been a fatal error, exit */
  DEBUGPRINT(DP_FATAL,slcExecFlag, ("EXEC: FATAL ERROR during restart or stop\n"));

  /*close the socket connections just to be sure */
  closeSocketConnections();
  cleanup();
}


/*=============================================================================
 
  Name: cleanup

  Abs:  clean up before exiting thread.  release all resources.

  
  Args: None
 
  Rem:  called before all returns from this thread.  

  Side: releases the socket library, 
        destroys the message queue
		destroys the dbDownloadEvent_ps
        sets the thread id = NO_THREAD
 
  Ret:  none
    may generate message queue errors from message queue destroy utility
  
=============================================================================*/
LOCAL void cleanup(void)
{
  /* release all resources */
  osiSockRelease();

  messageQDestroy(slcExec);

  slcDbDownloadEventDestroy();

  slcThreads_as[slcExec].tid_ps = NO_THREAD;
}



/*=============================================================================

  Name: startThreads

  Abs:  start the slc threads

  Args: None


  Rem:  orchestrates the startup of all slc-aware threads. Thread startup order
        is important.
         
       
  Side: all threads are started with the epics medium priority, and large stack size.

        We may what to include a timeout - which would then cause an epicsFalse
        to be returned if the timeout is reached.  would remove a possible infinite loop.
        Right now the slcIOC must be rebooted if this startup fails.


  Ret: epicsTrue if threads started successfully, otherwise epicsFalse

==============================================================================*/
LOCAL epicsBoolean  startThreads(void)
{
  unsigned int  stackSize;
  unsigned long jmsk = 0;
  unsigned long bitmsk = 1;
  /*
   * Start all of the SLC msg threads except the receive threads.
   * The receive threads are started last, after we know all other threads
   * have been started successfully.
   */
  /* set up thread defaults */
  stackSize = epicsThreadGetStackSize(epicsThreadStackBig);

  /************ PHASE I: start the Database Service *********************/

  DEBUGPRINT(DP_DEBUG,slcExecFlag, ("EXEC: starting Database Service\n"));

  if (startDatabaseService() == epicsFalse) {
	return(epicsFalse);
  }

  /* now that we have a database, determine if this is a FECC ioc */
  if (!slcInitFeccIOC() ) {
	return (epicsFalse);
  }

  /**************** PHASE II: Now create the Async thread *******************/
  /* this must be done first in an slcIOC so that the async table is created
	 before the other hdlrs .
     Start mgntAsync prior, as, for debug, cstrAsync sends a message soon                                       
     NOTE: FECC does not use cstrAsync.
  */

  if (feccIOC) {
	DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC:this is a FECC!\n"));
	/* both FECC and slcIOC use the message service threads, so start
	   msgSend and msgHdlr now for FECC.  MsgRecv is started later, after all 
	   the other threads are up and running.
	*/

	DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC: FECC msgSend, msgHdlr starting\n"));
	/* create the send thread now  */
	slcThreads_as[msgSend].tid_ps = epicsThreadCreate(MSG_SEND_NAME,
						   epicsThreadPriorityHigh, stackSize,
						   (EPICSTHREADFUNC)msgSendThread, 0);

	/* create the msg handler thread */
	slcThreads_as[msgHdlr].tid_ps  = epicsThreadCreate(MSG_HDLR_NAME, 
						  epicsThreadPriorityMedium, stackSize,
						  (EPICSTHREADFUNC)msgHdlrThread, 0);

    /* now start the FECC thread micromain, which is will take over and
	   start and manage all the FECC threads. */
	DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC:FECC Micro Job starting\n"));
	/* create the microMain job and let the FECC start itself up */
	slcThreads_as[microHdlr].tid_ps  = epicsThreadCreate(MICRO_HDLR_NAME, 
						    epicsThreadPriorityMedium, stackSize,
							(EPICSTHREADFUNC)micromain, 0);

	/* wait for the micromain thread to be active before continuing to msgRecv */
	if (waitForMicroHdlrActive() == epicsFalse) {
	  return(epicsFalse);
	}
  }
  else /* this is not a FECC */
  {
	/* get CSTR JMSK to determine what Hdlrs to start */
	if (!(slcAsyncGetJMSK (&jmsk))){ 
	  DEBUGPRINT(DP_ERROR,slcExecFlag, ("EXEC: Error reading CSTR JSMK\n"));
	  /* other error processing goes here ... */
	}
	/* create the mgnt Async  thread */
	if ((jmsk&(bitmsk<<mgntJob))>0) { 
	  DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC: starting mgntAsync\n"));
	  slcThreads_as[mgntAsync].tid_ps  = epicsThreadCreate(MGNT_ASYN_NAME, 
														   epicsThreadPriorityMedium, stackSize,
														   (EPICSTHREADFUNC)mgntAsyncThread, 0);
	  
	  /* wait for the thread to be active   */
	  if (waitForMgntAsyncReady() == epicsFalse) {
		return(epicsFalse);
	  }
	}

	if ((jmsk&(bitmsk<<mgntJob))>0) { 
	  DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC:JMSK-> MGNT Job starting\n"));
	  /* start mgnt job */
	  slcThreads_as[mgntHdlr].tid_ps  = epicsThreadCreate(MGNT_HDLR_NAME, 
														  epicsThreadPriorityMedium, stackSize,
														  (EPICSTHREADFUNC)mgntHdlrThread, 0);
	  /* wait for the mgnt Hdlr thread to set active flags   DMF timeout here? 
		 We don't want to go on until we know the mgntHdlr is working */
	  if (waitForMgntHdlrActive() == epicsFalse) {
		return(epicsFalse);
	  }
	}

	DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC: starting cstrAsync\n"));

	/* be sure asycExists is false */
	slcAsyncSetAsyncExists(epicsFalse);
	/* create the cstr Async  thread */

	slcThreads_as[cstrAsync].tid_ps  = epicsThreadCreate(CSTR_ASYN_NAME, 
						    epicsThreadPriorityMedium, stackSize,
							(EPICSTHREADFUNC)cstrAsyncThread, 0);

	/* wait for the thread to be active, and the asyncExists flag true */
	if (waitForAsyncReady() == epicsFalse) {
	  return(epicsFalse);
	}


	/**************** PHASE III: Start all the Hdlr threads  ******************/

	DEBUGPRINT(DP_DEBUG,slcExecFlag, ("EXEC: starting Hdlr Threads\n"));

	if (startHdlrThreads() == epicsFalse) {
	  return(epicsFalse);
	}
 
  } /* if !feccIOC */

  /*****************PHASE IV: start the msgRecv thread ********************/
  /* msgRecv is started last to hold off any incoming SLC messages until 
	 we know all threads are up and running.  Both slcIOC and FECC use the
	 msgRecv thread.
  */

  DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC: starting msgRecv\n"));

  /* now we know the threads are ready, create the msg receive thread */
  slcThreads_as[msgRecv].tid_ps = epicsThreadCreate(MSG_RECV_NAME, 
						  epicsThreadPriorityMedium, stackSize, 
						  (EPICSTHREADFUNC)msgRecvThread, 0);

  /* increment the starts count */
  epicsTimeGetCurrent(&execDiagnostics_s.lastStartTime);
  execDiagnostics_s.numStarts++;

  /* wait for the msgRecv thread active flag */
  return (waitForRecvThreadActive());  /* DMF timeout here? */
}

/*=============================================================================

  Name: startDatabaseService

  Abs:  start the db threads

  Args: None


  Rem:  orchestrates the startup of all Database threads, send the dbdownload msg, 
        then waits for the dbExists flag to be set before returning
         
       
  Side: all db threads are started with the epics medium priority, and large stack size.

  Ret: epicsTrue if threads started successfully, otherwise epicsFalse

==============================================================================*/
LOCAL epicsBoolean  startDatabaseService(void)
{
  unsigned int  stackSize;
  dbmsgmail_ts  dbMsg_s;
  vmsstat_t     iss=MICR_OKOK;

  /*
   * Start all of the database service threads.
   * The receive thread is started last, after we know all other threads
   * have been started successfully.
   */
  /* set up thread defaults */
  stackSize = epicsThreadGetStackSize(epicsThreadStackBig);

  /* be sure dbExists is false */
  dbExists = epicsFalse;

  /* create the dbDownloadEvent_ps for database download */
  dbDownloadEvent_ps = epicsEventMustCreate(epicsEventEmpty); /* DMF take out later?*/

  /************ PHASE I: start the Database Service *********************/

  /* create the db send thread */
  slcThreads_as[dbSend].tid_ps = epicsThreadCreate(DB_SEND_NAME, 
						   epicsThreadPriorityHigh, stackSize,
						   (EPICSTHREADFUNC)dbSendThread, 0);

  /* create the db handler thread */
  slcThreads_as[dbHdlr].tid_ps  = epicsThreadCreate(DB_HDLR_NAME, 
						   epicsThreadPriorityMedium, stackSize,
						   (EPICSTHREADFUNC)dbHdlrThread, 0);
  
  /* and...create the db receive thread */
  slcThreads_as[dbRecv].tid_ps = epicsThreadCreate(DB_RECV_NAME, 
						 epicsThreadPriorityMedium, stackSize, 
						 (EPICSTHREADFUNC)dbRecvThread, 0);

  /* wait for the db service  threads to set active flags   DMF timeout here? */
  if (waitForDbThreadsActive() == epicsFalse) {
	return(epicsFalse);
  }

  /* send a request to dbSend thread to send db download request to Alpha */
  /* dbSend does not us a mem pool, just send the dbmsgmail_ts struct */
  memset(&dbMsg_s,0, sizeof(dbMsg_s));
  DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC: Sending download message\n"));
  dbMsg_s.func = DB_DOWNLOAD_REQ;
  /* send the message, wait at queue if it is full */
  if ((iss = messageQSend(dbSend, epicsTrue, sizeof(dbMsg_s), &dbMsg_s))<0) {
	DEBUGPRINT(DP_DEBUG,slcExecFlag, ("EXEC: Error sending download message\n"));
	return(epicsFalse);
  }

  if (waitForDbExists() == epicsFalse) {
	DEBUGPRINT(DP_DEBUG,slcExecFlag, ("EXEC: Error waiting for DbExists\n"));
	return(epicsFalse);
  }

  return (epicsTrue);
}

  
/*=============================================================================

  Name: waitForDbThreadsActive

  Abs:  wait for the Database Service to be active

  Args: none

  Rem:  Wait for all database service threads to be active
       
  Side: 

   NOTE:We may what to include a timeout - which would then cause an epicsFalse
        to be returned if the timeout is reached. This can be an  infinite loop
        if the threads fail to start.


  Ret: epicsTrue if all conditions successful, otherwise no return!

==============================================================================*/
LOCAL epicsBoolean  waitForDbThreadsActive(void)
{
  /* wait for threads active, dbExists */
  
  while (!dbThreadsActive())
  {
	epicsThreadSleep(0.5);
  } /*while  not all threads active */
	
  return epicsTrue;
}

/*=============================================================================

  Name: dbThreadsActive

  Abs:  returns true if all the first dbs threads are active 

  Args: none

  Rem:  this utility checks the threads array to see if the dbs threads have started
       
  Side: 

  Ret: epicsTrue if database service is active, else epicsFalse

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean dbThreadsActive(void)
{

  if (slcThreads_as[slcExec].active)
  {
	if ( (!slcThreads_as[dbSend].active)||
		 (!slcThreads_as[dbHdlr].active)||
		 (!slcThreads_as[dbRecv].active) ) {
	  return(epicsFalse);
	}
	return(epicsTrue);
  }
  else
	return(epicsFalse);
}

/*=============================================================================

  Name: waitForDbExists

  Abs:  wait for the database to be downloaded successfully

  Args: none

  Rem:  Wait for the dbExists flag to be set true
       
  Side: 

   NOTE:We may what to include a timeout - which would then cause an epicsFalse
        to be returned if the timeout is reached. This can be an  infinite loop
        if the threads fail to start.


  Ret: epicsTrue if successful, otherwise no return!

==============================================================================*/
LOCAL epicsBoolean  waitForDbExists(void)
{
  /* wait for threads active, dbExists */
  
  while (slcGetDbExists()==epicsFalse)
  {
	epicsThreadSleep(0.5);
  } /*while  not all threads active */
	
  return epicsTrue;
}

/*=============================================================================

  Name: waitForAsyncReady

  Abs:  wait for the cstrAsync thread to become active, and asyncExists true

  Args: none

  Rem:  lets the slcExec thread know when the async functions are ready to go.
       
  Side:  
  
  NOTE: We may want to add a timeout here, which would allow an epicsFalse return.
        This would eliminate a possible infinite loop. Of course, if there's a failure
		we would probably want to re-boot anyway....


  Ret: epicsTrue if thread started successfully, otherwise no return!

==============================================================================*/
LOCAL epicsBoolean  waitForAsyncReady(void)
{
  
  while ((!slcThreads_as[cstrAsync].active) || 
		 (!slcAsyncGetAsyncExists()) ) {
	epicsThreadSleep(0.5);
  } /*while*/
  return epicsTrue;
}

/*=============================================================================

  Name: waitForMgntAsyncReady

  Abs:  wait for the mgntAsync thread to become active

  Args: none

  Rem:  lets the slcExec thread know when the async functions are ready to go.
       
  Side:  
  
  NOTE: We may want to add a timeout here, which would allow an epicsFalse return.
        This would eliminate a possible infinite loop. Of course, if there's a failure
		we would probably want to re-boot anyway....


  Ret: epicsTrue if thread started successfully, otherwise no return!

==============================================================================*/
LOCAL epicsBoolean  waitForMgntAsyncReady(void)
{
  
  while (!slcThreads_as[mgntAsync].active) {
	epicsThreadSleep(0.5);
  } /*while*/
  return epicsTrue;
}

/*=============================================================================

  Name: startHdlrThreads

  Abs:  start the required hdlr threads

  Args: None


  Rem:  orchestrates the startup of all handler threads and the msgSend thread
        The msgSend thread is included here to be sure it is ready when msgRecv
		is started later.
         
       
  Side: all these threads are started with the epics medium priority,
        and large stack size.

  Ret: epicsTrue if threads started successfully, otherwise epicsFalse

==============================================================================*/
LOCAL epicsBoolean  startHdlrThreads(void)
{
  unsigned int  stackSize;
  unsigned long jmsk = 0;
  unsigned long bitmsk = 1;


  /* set up thread defaults */
  stackSize = epicsThreadGetStackSize(epicsThreadStackBig);


  /* get CSTR JMSK to determine what Hdlrs to start */
  if (!(slcAsyncGetJMSK (&jmsk))){ 
	DEBUGPRINT(DP_ERROR,slcExecFlag, ("EXEC: Error reading CSTR JSMK\n"));
	/* other error processing goes here ... */
  }
  DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC: CSTR JMSK: %lx\n", jmsk ));

  /* get this just for the info message */
  if ((jmsk&(bitmsk<<dbsJob))>0) {
	DEBUGPRINT(DP_DEBUG,slcExecFlag, 
	 ("EXEC:JMSK-> says DBS Job should start, but it already has...by design...\n"));
  }

  /************ PHASE III: start the Handlers *********************/

  if ((jmsk&(bitmsk<<msgJob))>0) {
	DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC:JMSK-> MSG Job starting\n"));
	/* create the send thread now  */
	slcThreads_as[msgSend].tid_ps = epicsThreadCreate(MSG_SEND_NAME,
						   epicsThreadPriorityHigh, stackSize,
						   (EPICSTHREADFUNC)msgSendThread, 0);

	/* create the msg handler thread */
	slcThreads_as[msgHdlr].tid_ps  = epicsThreadCreate(MSG_HDLR_NAME, 
						  epicsThreadPriorityMedium, stackSize,
						  (EPICSTHREADFUNC)msgHdlrThread, 0);
  }

  /*
   * Start all of the handler threads for the slcIOC
   */

  /* bsacHdlr*/
  if ((jmsk&(bitmsk<<bpmoJob))>0) {
	DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC:JMSK-> BSAC Job starting\n"));
	/* start bsac (bpm) job */
	slcThreads_as[bsacHdlr].tid_ps  = epicsThreadCreate(BSAC_HDLR_NAME, 
										epicsThreadPriorityMedium, stackSize,
						  				(EPICSTHREADFUNC)bsacHdlrThread, 0);
	/* wait for the bsac Hdlr thread to set active flags   DMF timeout here? 
	 We don't want to go on until we know the bsacHdlr is working */
	if (waitForBsacHdlrActive() == epicsFalse) {
	  return(epicsFalse);
	}
  }

/* MOVED mgntHdlr initialization in front of mgntAsync in startThreads 
   if ((jmsk&(bitmsk<<mgntJob))>0) { 
	DEBUGPRINT(DP_INFO,slcExecFlag, ("EXEC:JMSK-> MGNT Job starting\n"));
	 start mgnt job 
  	slcThreads_as[mgntHdlr].tid_ps  = epicsThreadCreate(MGNT_HDLR_NAME, 
														  epicsThreadPriorityMedium, stackSize,
														  (EPICSTHREADFUNC)mgntHdlrThread, 0);
	 wait for the mgnt Hdlr thread to set active flags   DMF timeout here? 
	 We don't want to go on until we know the mgntHdlr is working 
	if (waitForMgntHdlrActive() == epicsFalse) {
	  return(epicsFalse);
	}
  }
*/

  /* 
	 create the cstr handler thread - cstrHdlr MUST ALWAYS BE LAST! 
	 cstrHdlr checks that all the other threads required for THIS slcIOC
	 have been started - so it must be the last hdlr created.
  */

  if ((jmsk&(bitmsk<<cstrJob))>0) {
	DEBUGPRINT(DP_DEBUG,slcExecFlag, ("EXEC:JMSK-> CSTR Hdlr starting\n"));

	slcThreads_as[cstrHdlr].tid_ps  = epicsThreadCreate(CSTR_HDLR_NAME, 
						   epicsThreadPriorityMedium, stackSize,
						   (EPICSTHREADFUNC)cstrHdlrThread, 0);
  }

  /* wait for the key Hdlr threads to set active flags   DMF timeout here? 
	 We don't want to go on until we know the threads are humming along */
  if (waitForHdlrThreadsActive() == epicsFalse) {
	  return(epicsFalse);
  }

  return(epicsTrue);
}

/*=============================================================================

  Name: waitForBsacHdlrActive

  Abs:  returns true when all the required bsacHdlr is active

  Args: none

  Rem: 

  Side: 

  NOTE: We may want to add a timeout here, which would allow an epicsFalse return.
        This would eliminate a possible infinite loop. Of course, if there's a failure
		we would probably want to re-boot anyway....

  Ret: epicsTrue when bsacHdlr is active, else does not return!

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean  waitForBsacHdlrActive(void)
{
  /* wait for threads active,  */
  
  while (!bsacHdlrActive())
  {
	epicsThreadSleep(0.5);
  }
	
  return epicsTrue;
}

/*=============================================================================

  Name: bsacHdlrActive

  Abs:  returns true if the required hdlr thread is active

  Args: none

  Rem:  this utility checks the threads array to see if the mgntHdlr thread
        is active
       
  Side: 

  Ret: epicsTrue if bsacHdlr thread active, else epicsFalse

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean bsacHdlrActive(void)
{

  /* replace this with whatever jobs you want to check on...*/
  if (slcThreads_as[slcExec].active)
  {
	if ((!slcThreads_as[bsacHdlr].active) ) {
	  return(epicsFalse);
	}

	return(epicsTrue);
  }
  else
	return(epicsFalse);
}


/*=============================================================================

  Name: waitForMgntHdlrActive

  Abs:  returns true when all the required mgntHdlr is active

  Args: none

  Rem: 

  Side: 

  NOTE: We may want to add a timeout here, which would allow an epicsFalse return.
        This would eliminate a possible infinite loop. Of course, if there's a failure
		we would probably want to re-boot anyway....

  Ret: epicsTrue when mgntHdlr is active, else does not return!

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean  waitForMgntHdlrActive(void)
{
  /* wait for threads active,  */
  
  while (!mgntHdlrActive())
  {
	epicsThreadSleep(0.5);
  }
	
  return epicsTrue;
}

/*=============================================================================

  Name: mgntHdlrActive

  Abs:  returns true if the required hdlr thread is active

  Args: none

  Rem:  this utility checks the threads array to see if the mgntHdlr thread
        is active
       
  Side: 

  Ret: epicsTrue if mgntHdlr thread active, else epicsFalse

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean mgntHdlrActive(void)
{

  /* replace this with whatever jobs you want to check on...*/
  if (slcThreads_as[slcExec].active)
  {
	if ((!slcThreads_as[mgntHdlr].active) ) {
	  return(epicsFalse);
	}

	return(epicsTrue);
  }
  else
	return(epicsFalse);
}



/*=============================================================================

  Name: waitForMicroHdlrActive

  Abs:  returns true when all the required FECC threads are active

  Args: none

  Rem: this utility waits for the fecc threads to become active by waiting for
       the microHdlr.active flag to be set to epicsTrue.  The micromain thread 
	   must set it's active flag AFTER all of its threads are up and running
	   properly.

  Side: 

  NOTE: We may want to add a timeout here, which would allow an epicsFalse return.
        This would eliminate a possible infinite loop. Of course, if there's a failure
		we would probably want to re-boot anyway....

  Ret: epicsTrue when microHdlr is active, else does not return!

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean  waitForMicroHdlrActive(void)
{
  /* wait for threads active,  */
  
  while (!microHdlrActive())
  {
	epicsThreadSleep(0.5);
  }
	
  return epicsTrue;
}

/*=============================================================================

  Name: microHdlrActive

  Abs:  returns true if the required hdlr thread is active

  Args: none

  Rem:  this utility checks the threads array to see if the micrHdlr thread
        is active
       
  Side: 

  Ret: epicsTrue if microHdlr thread active, else epicsFalse

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean microHdlrActive(void)
{

  /* BOB replace this with whatever jobs you want to check on...*/
  if (slcThreads_as[slcExec].active)
  {
	if ((!slcThreads_as[microHdlr].active) ) {
	  return(epicsFalse);
	}

	return(epicsTrue);
  }
  else
	return(epicsFalse);
}


/*=============================================================================

  Name: waitForHdlrThreadsActive

  Abs:  returns true when all the required handler threads are active

  Args: none

  Rem: this utility waits for the hdlr threads to become active. 
  Side: 

  NOTE: We may want to add a timeout here, which would allow an epicsFalse return.
        This would eliminate a possible infinite loop. Of course, if there's a failure
		we would probably want to re-boot anyway....

  Ret: epicsTrue when handlers are active, else does not return!

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean  waitForHdlrThreadsActive(void)
{
  /* wait for threads active, dbExists */
  
  while (!hdlrThreadsActive())
  {
	epicsThreadSleep(0.5);
  } /*while  not all threads active */
	
  return epicsTrue;
}

/*=============================================================================

  Name: hdlrThreadsActive

  Abs:  returns true if all the required hdlr threads; msgHdlr, msgSend, cstrHdlr ,
  are active.  these are the minimum threads needed for the slcIOC to function

  Args: none

  Rem:  this utility checks the threads array to see if the required hdlr 
        threads have started
       
  Side: 

  Ret: epicsTrue if msgHdlr, msgSend, cstrHdlr are active, else epicsFalse

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean hdlrThreadsActive(void)
{

  if (slcThreads_as[slcExec].active)
  {
	if ( (!slcThreads_as[msgSend].active) ||
		 (!slcThreads_as[msgHdlr].active) ||
		 (!slcThreads_as[cstrHdlr].active) ) {
	  return(epicsFalse);
	}

	return(epicsTrue);
  }
  else
	return(epicsFalse);
}

/*=============================================================================

  Name: waitForRecvThreadActive

  Abs:  wait for the msgRecv thread to become active

  Args: none

  Rem:  lets the slcExec thread know when the recv thread is ready to go.
       
  Side: 

  NOTE: We may want to add a timeout here, which would allow an epicsFalse return.
        This would eliminate a possible infinite loop. Of course, if there's a failure
		we would probably want to re-boot anyway....


  Ret: epicsTrue if thread started successfully, otherwise no return!

==============================================================================*/
LOCAL epicsBoolean  waitForRecvThreadActive(void)
{
  
  while (!slcThreads_as[msgRecv].active)
  {
	epicsThreadSleep(0.5);
  } /*while*/
  return epicsTrue;
}


/*=============================================================================

  Name: closeSocketConnections

  Abs:  close all of the slc socket connections 

  Args: none

  Rem:  closes all sockets. This is called when all threads should stop, or if
        the slcExec thread must exit (not a normal condition)
         
       
  Side: The socket Recv threads must have their sockets closed for them, as they are
		most likely waiting on a recv call at the socket and cannot check their stop flags
        to see that they should stop.  Closing the socket will interrupt that recv call.

  Ret: none

==============================================================================*/
LOCAL void closeSocketConnections(void)
{
  int i;
  /*  SOCKET sd;*/

  for (i=0; i<MAX_SOCKETS; i++) {
	if (slcSockets_as[i].sd != INVALID_SOCKET) {
	  DEBUGPRINT(DP_DEBUG,slcExecFlag, 
				 ("EXEC:destroying socket %d!\n", 
				  slcSockets_as[i].sd));
	  tcpSockClose(&slcSockets_as[i]);
	}
  }
}

/*=============================================================================

  Name: anyJobActive

  Abs:  return false if all "job" threads are done

  Args: none

  Rem:  this utility checks the threads array to see if any "job" thread is active
        the "job" threads are all threads other than slcExec, msgRecv, dbRecv, slcEnd.
        
  Side: this function indicates when all 'job' threads have stopped, so then slcExec
        knows it's ok to close the sockets and stop the recv threads.

  Ret: epicsTrue if any of the "job" threads is active, else epicsFalse

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean anyJobActive(void)
{
  int i;

  for (i=msgSend; i<slcEnd; i++)
  {
    if (slcThreads_as[i].active) 
    {
	  DEBUGPRINT(DP_DEBUG,slcExecFlag, 
				 ("EXEC: a thread enum #%d still active\n",i));
	  return(epicsTrue);

    }
  }
  return(epicsFalse);
}
	

/*=============================================================================

  Name: stopThreads

  Abs:  stop all the slc threads, except exec

  Args: none

  Rem:  this function stops all the slc threads.
        it is called by the slcExec thread when it receives the IOC_STOP msg;

  Side: stops all the slc threads by sending a IOC_STOP message to each thread 
        with a message queue.  Threads without message queues will check their 
        slcThread[].stop flag. Once all threads are told to stop, the sockets
		are closed to allow the recv threads to check their .stop flag.
        This routine does not return until all threads are stopped. If there is
		a problem and the threads don't stop, a reboot is required.

  Ret: epicsTrue, or it does not return
     can generates message queue errors from message queue utilities -see utilities
==============================================================================*/
/**procedure**/
LOCAL epicsBoolean stopThreads(void)
{
  int            cnt;
  dbmsgmail_ts   dbmsg_s;
  epicsBoolean   ok;
  msgptr_ts      msgptr_s;
  msgmail_ts*    msg_ps = NULL;

  /* set all thread stop flags to true */
  setStopFlagsTrue(); /* dmf except dont stop dbSend or dbHdlr */


  /* send a STOP message to all of the non-db threads except slcExec */
  for (cnt=slcExec+1; cnt < dbSend; cnt++)
  {
	/*send stop message to all threads that have a valid mQ (skips the recv threads) */
	if (slcThreads_as[cnt].mqId_ps!=NULL)
	{
	  /* create a STOP message for each of the  non-db threads */
	  if ( SUCCESS(msgQGetBuffer( sizeof(msgmail_ts), (void*)&msg_ps)) )
	  {
		msg_ps->msgheader.func = IOC_STOP;
        msgptr_s.msgbytes = MSG_HDRS_BYTES;
        msgptr_s.msgptr_p = msg_ps;
		/* send the message, wait at queue if full */
		if (!SUCCESS(messageQSend(cnt, epicsTrue, sizeof(msgptr_s), &msgptr_s)))
		{
		  DEBUGPRINT(DP_ERROR, slcExecFlag, 
					 ("EXEC: stop failed on thread enum #%d\n",cnt));
		} else
		  DEBUGPRINT(DP_DEBUG, slcExecFlag, 
					 ("EXEC: created and sent a msg service message\n"));
		  
	  }
	}
  }

  /* now close the recv sockets */
  /* DMF there is a timing issue here, the stop command gets to this place before the
	 recv threads may detect the stop flag is true.  Here the socket is destroyed, which
	 causes the Recv threads to err out of the "recv()" call they are typically
	 blocked on.
	 that error gets reported, and then the recv threasd note it's stop flag is true, and
	 shuts down.  There is no real "error" state here, but I'm trying to avoid using 
	 non-blocking sockets ...
  */
  closeSocketConnections();
  DEBUGPRINT(DP_INFO, slcExecFlag, ("EXEC: stopped threads\n"));

  /* wait for these threads to stop */
  ok = waitForNonDBThreadsTerminated();

  /* create the stop message for db threads */
  memset(&dbmsg_s,0, sizeof(dbmsg_s));
  dbmsg_s.func = IOC_STOP;
  /* send a STOP message to the db threads */
  for (cnt=dbSend; cnt < slcEnd; cnt++)
  {
	/*send stop message to all threads that have a valid mQ */
	if (slcThreads_as[cnt].mqId_ps!=NULL)
	{
	  if ((!SUCCESS(epicsMessageQueueSend(slcThreads_as[cnt].mqId_ps, 
									(void *)&dbmsg_s, 
									sizeof(dbmsg_s)))) <0)
	  {
		DEBUGPRINT(DP_ERROR, slcExecFlag, (
			   "EXEC: stop failed on thread enum #%d\n",cnt));
	  }
	}
  }

  /* now wait for threads to stop */
  ok = waitForThreadsTerminated();

  DEBUGPRINT(DP_INFO, slcExecFlag, ("EXEC: all threads are gone\n"));
  setStopFlagsFalse();

  if (dbDownloadEvent_ps != NULL)
  {
	epicsEventDestroy(dbDownloadEvent_ps);
    dbDownloadEvent_ps = NULL;
  }

  /* incr. the Stops count */
  execDiagnostics_s.numStops++;

  /*slcCmlogLogMsg(MICR_STOPPED, "STOPPED", 
	"%s Stopped and out of service\n", slcGetIOCName); */

  return (ok);
}

/*=============================================================================

  Name: waitForNonDBThreadsTerminated

  Abs:  wait for all threads to terminate, except the slcExec thread

  Args: none

  Rem:  Waits for all the thread tid_ps's to be NO_THREAD
         
  Side: 

  NOTE: We may want to include a timeout - which would then cause an epicsFalse
        to be returned if the timeout is reached.  would remove a possible infinite loop.


  Ret: epicsTrue if threads terminated successfully, otherwise does not return

==============================================================================*/
LOCAL epicsBoolean  waitForNonDBThreadsTerminated(void)
{
  /*
   * wait to receive startup complete msgs from the threads?
  */
  while (!nonDBThreadsTerminated())
  {
	epicsThreadSleep(0.5);
  } /*while  not all threads terminated */
	
  return epicsTrue;
}

/*=============================================================================

  Name: nonDBThreadsTerminated

  Abs:  return true if all threads, except slcExec, are terminated

  Args: none

  Rem:  this utility checks the threads array to see if all threads have terminated
       
  Side:   

  Ret: epicsTrue if no threads exist (except slcExec), else epicsFalse

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean nonDBThreadsTerminated(void)
{
  int cnt;

  for (cnt=slcExec+1; cnt< dbSend; cnt++)
  {
	if ( slcThreads_as[cnt].tid_ps != NO_THREAD) {
	  if (!(epicsThreadIsSuspended(slcThreads_as[cnt].tid_ps))) 
		return(epicsFalse);
	}
  }
  return(epicsTrue);
}

/*=============================================================================

  Name: waitForThreadsTerminated

  Abs:  wait for all threads to terminate, except the slcExec thread

  Args: none

  Rem:  Waits for all the thread tid_ps's to be NO_THREAD
         
  Side: 

  NOTE: We may want to include a timeout - which would then cause an epicsFalse
        to be returned if the timeout is reached.  would remove a possible infinite loop.


  Ret: epicsTrue if threads terminated successfully, otherwise does not return

==============================================================================*/
LOCAL epicsBoolean  waitForThreadsTerminated(void)
{
  /*
   * wait to receive startup complete msgs from the threads?
  */
  while (!allThreadsTerminated())
  {
	epicsThreadSleep(0.5);
  } /*while  not all threads terminated */
	
  return epicsTrue;
}

/*=============================================================================

  Name: allThreadsTerminated

  Abs:  return true if all threads, except slcExec, are terminated

  Args: none

  Rem:  this utility checks the threads array to see if all threads have terminated
       
  Side:   

  Ret: epicsTrue if no threads exist (except slcExec), else epicsFalse

==============================================================================*/
/**procedure**/
LOCAL epicsBoolean allThreadsTerminated(void)
{
  int cnt;

  for (cnt=slcExec+1; cnt<slcEnd; cnt++)
  {
	if ( slcThreads_as[cnt].tid_ps != NO_THREAD) {
	  if (!(epicsThreadIsSuspended(slcThreads_as[cnt].tid_ps))) 	  
		return(epicsFalse);
	}
  }
  return(epicsTrue);
}

/*=============================================================================

  Name: setStopFlagsTrue

  Abs:  set all the slcThreads_as[].stop = epicsTrue;

  Args: none

  Rem:  this function sets all the slcThreads_as[].stop = epicsTrue;
        slcExec thread is not included, since it never stops
       
  Side: dmf 12/13/07 made this stop at dbSend so that dbSend does not stop here

  Ret: none

==============================================================================*/
/**procedure**/
LOCAL void setStopFlagsTrue(void)
{
  int cnt;

  for (cnt=slcExec+1; cnt<dbSend; cnt++)
  {
	slcThreads_as[cnt].stop = epicsTrue;
  }
}

/*=============================================================================

  Name: setStopFlagsFalse

  Abs:  set all the slcThreads_as[].stop = epicsFalse;

  Args: none

  Rem:  this function sets all the slcThreads_as[].stop = epicsFalse;
       
  Side: 

  Ret: none

==============================================================================*/
/**procedure**/
LOCAL void setStopFlagsFalse(void)
{
  int cnt;

  for (cnt=slcExec; cnt<slcEnd; cnt++)
  {
	slcThreads_as[cnt].stop = epicsFalse;
  }
}

