/*=============================================================================
 
  Name: slcMain.c
        slcStart      - the start command called at initialization
        slcRestart    - the restart command available via IPL, iocsh, db routine
        slcStop       - the stop command available via IPL, iocsh, db routine
        slcFreeLists  - Diagnostic free list reporting
        slcfreelistUtil - Local support routine for slcFreeLists

  Abs:  slcMain contains the slcExec thread and the slcStart, slcRestart, and
        slcStop iocsh commands for the SLC-IOC
    See the SLC Executive and slc-IOC Initialization Design document

  Auth: 22-jul-2004, diane fairley (DFAIRLEY): created
  Rev:  02 feb 2005 drogind


-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        13-dec-2004, diane fairley (DFAIRLEY):
		   moved all slcExec thread code into new slcExec.c and .h files
		06-dec-2004, diane fairley (DFAIRLEY): 
		   added this header, coding standard changes
        
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int slcMainFlag = DP_DEBUG;
#endif


/* c library includes */
#include    <stdio.h>
#include	<string.h>

/* epics includes */
#include    <dbDefs.h>        /* NELEMENTS */
#include    <iocsh.h>         /* ioc shell stuff */
#include    <registryFunction.h>
#include    <epicsExport.h>   /* for registering slcRestart, slcStop */
#include    <epicsEvent.h>    /* epicsEventId */
#include    <freeList.h>      /* freeListItemsAvail */

/* slc common includes */
#include    "slcGlob.h"       /* globals, types, */
#include    "msgTypes.h"      /* msgmail_ts */
#include    "slcMsg.h"        /* slcIOC internal msg function codes */
#include    "msgfunc.hc"      /* FUNC_TEST job code */

/* slc utility prototypes */
#include    "msgQUtil.h"      /* message queue utilities */
#include    "slcCmlog.h"      /* slcCmlogLogMsg */
#include    "cvtMacr.h"       /* macro used for endian-ness */

/* thread prototypes  */
#include    "slcMain.h"    /* slcStart, slcRestart, slcStop */
#include    "slcExec.h"    /* slcExec thread */

/* dbs utility prototypes (for db free list commands) */
#include    "dbGlob.h"
#include    "dblistUtil.h"
#include    "dbdict.h"

/* for dbNameToAddr test */
#include    "mgntParam.h"
#include <dbAccess.h>      /* Epics Runtime Database        */
#include    "mgntHdlr.h"     /*        */
/*=============================================================================
 
  Name: slcStart

  Abs:  The function slcStart initializes the slc-aware thread globals and
        starts the slcExec thread. Once the slcExec thread is active, it calls
		the slcRestart function to tell slcExec to start all other slc-aware 
        threads.
  
  Args: none
 
  Rem:  This function is called once to start the slc-aware portion of the slcIOC.
        It is called from a startup file, after all other IOC initialization is done.

  Side: the slc-aware thread globals are initialized to known values. 
        When this function completes successfully, all the slc-aware threads
		will have been started.
. 
  Ret:  none
  
=============================================================================*/
void slcStart(void)
{
  unsigned int    stackSize, threadPriority;

  if ( !epicsThreadGetId(EXEC_NAME) )
  {
	threadPriority = epicsThreadPriorityMedium;
	stackSize = epicsThreadGetStackSize(epicsThreadStackBig);

	/* initialize the global structures - ONLY DONE ONCE RIGHT HERE */
	slcInitGlobals();

	/* create the slcExec thread */
	DEBUGPRINT(DP_INFO,slcMainFlag, ("starting exec thread\n"));
	slcThreads_as[slcExec].tid_ps  = epicsThreadCreate(EXEC_NAME, 
	      threadPriority, stackSize,
	      (EPICSTHREADFUNC)slcExecThread, 0);

	/* check for good thread start */
	if (slcThreads_as[slcExec].tid_ps==NO_THREAD) {
	  DEBUGPRINT(DP_FATAL,slcMainFlag, 
				 ("SLCSTART: slcStart command failed to start exec thread!\n"));
	  return;
	}

	while (slcThreads_as[slcExec].active==epicsFalse) epicsThreadSleep(0.5);
  
	/* restart the threads */
	slcRestart();
  }
  else printf("Invalid Command. The slc-aware threads have already started\n");

}


/*=============================================================================

  Name: slcStop

  Abs:  Stop slc threads

  Args: none

  Rem:  stopThreads issues the message to stop all threads, close all sockets. 
        This is the main shutdown command for the slcIoc.  It can be called as an iocsh
		command, or from the epics subroutine record.
         
  Side: a IOC_STOP message is sent to the slcExec thread message queue

  Ret: None

==============================================================================*/
/**procedure**/
void slcStop(void)
{
  msgptr_ts         msg_s;
  msgmail_ts*       sMsg_ps = NULL;

 
  if (epicsThreadGetId(EXEC_NAME))
  {
	if (!feccIOC)
	{
	  if ( SUCCESS(msgQGetBuffer(sizeof(msgmail_ts), (void*)&sMsg_ps)) )
	  {
		sMsg_ps->msgheader.func = IOC_STOP;

		/* create the STOP message for the slcExec thread  */
		memset(&msg_s,0, sizeof(msg_s));
		msg_s.msgptr_p = sMsg_ps;
		msg_s.msgbytes = MSG_HDRS_BYTES;

		/*send stop message to slcExe thread  */
		if (!SUCCESS(messageQSend(slcExec, epicsTrue, sizeof(msg_s),  &msg_s)))
		{
		  DEBUGPRINT(DP_ERROR,slcMainFlag, 
					 ("SLCSTOP:stop failed on send to Exec thread mq.\n"));
		}
	  }
	} else printf(" Invalid Command for a FECC IOC.\n");
  } else printf(" Invalid Command. The slc-aware threads have not started.\n");
}

/*=============================================================================

  Name: slcRestart

  Abs:  Restart slc threads

  Args: none

  Rem:  slcRestart issues the message to restart all threads. This is the main
        Restart command. Can be called as an iocsh command, or from an epics 
        subroutine record.

  Side: sends the IOC_RESTART msg to the slcExec thread

  Ret: None

==============================================================================*/
/**procedure**/
void slcRestart(void)
{
  msgptr_ts         msg_s;
  msgmail_ts*       sMsg_ps = NULL;

  if (epicsThreadGetId(EXEC_NAME))
  {
	if (!feccIOC)
	{
	  if ( SUCCESS(msgQGetBuffer(sizeof(msgmail_ts), (void*)&sMsg_ps)) )
	  {
		sMsg_ps->msgheader.func = IOC_RESTART;
	

		/* create the Restart message for the slcExec thread  */
		memset(&msg_s,0, sizeof(msg_s));
		msg_s.msgptr_p = sMsg_ps;
		msg_s.msgbytes = MSG_HDRS_BYTES;

		/*send restart message to slcExe thread - wait at queue if full */
		if (!SUCCESS(messageQSend(slcExec, epicsTrue, sizeof(msg_s),  &msg_s)))
		{
		  DEBUGPRINT(DP_ERROR,slcMainFlag, 
					 ("SLCRESTART:restart failed on send to Exec thread mq.\n"));
		}
	  }
	} else printf(" Invalid Command for a FECC IOC.\n");
  } else printf(" Invalid Command. The slc-aware threads have not started.\n");
}
/*=============================================================================

  Name: slcfreelistUtil

  Abs:  Diagnostic to dump information about various ioc free list items available
        called by slcFreeLists, below

  Args: Type		          Name	      Access	   Description
        ------------------- -----------	---------- ----------------------------
		const void *        freeList_ps     R      pointer to free list to diagnose
        const int           numLists        R      number of lists in free list
        const int           sizeEaList      R      size of each list in free list
        const char *        message         R      string describing name of free list

  Rem:  Called as an iocsh command

  Side: Uses epics libCom freeListLib utilities freeListItemsAvail to view the 
        number of lists available to the ioc.
        freeListLib memory allocates the free list after the first list request,
        and not upon initialization. Thus, if the  'slcFreeLists' shell routine is 
		called prior to any programmatic requests, (as may be the case with dblist
        or message lists), this routine will go ahead and request a free list, then 
        free it right away in order for the entire free list structure to get allocated.
        It then reports the items avail, total  # lists in free list, and size of each
        list.

  Ret: None

==============================================================================*/
/**procedure**/
void slcfreelistUtil(const void * freeList_ps, 
						   const int numLists, 
						   const int sizeEaList,
						   const char * message)
{
  size_t  itemsAvail = 0;
  void * dblist_ps = NULL;

  if (freeList_ps)
	{
	  itemsAvail=freeListItemsAvail((void *)freeList_ps);
	  if (itemsAvail==0)
		{
		  /* freelistLib allocates entire space after first list request */
		  /*  so, make request, then free */
		  dblist_ps= freeListCalloc((void *)freeList_ps);
		  freeListFree((void *)freeList_ps, dblist_ps);
		  itemsAvail=freeListItemsAvail((void *)freeList_ps);		   
		}
	  printf("Free lists for %s avail: %d (out of %d; list size=%d bytes)\n", 
			 message, itemsAvail, numLists, sizeEaList);
	} 
  else
	{
	  printf ("Free lists for %s do not exist; error \n", message);
	} 
}
/*=============================================================================

  Name: slcFreeLists

  Abs:  Diagnostic to dump information about various ioc free list items available

  Args: none

  Rem:  Called as an iocsh command
        Free list pointers must be global in order to report upon them
        Calls local utility slcfreelistUtil with free list args in order to output
        to console
        Feel free to insert your free lists here

  Side: 

  Ret: None

==============================================================================*/
/**procedure**/
void slcFreeLists(void)
{
  if (epicsThreadGetId(EXEC_NAME))
  {
	if (slcGetDbExists())
	{
	  /* Database Service free lists diagnostics */
	  /* dblist free lists */
	  slcfreelistUtil(dblistSmFreeList_p,  DBL_NSM, DBL_SZSM, "dblist small");
	  slcfreelistUtil(dblistBigFreeList_p, DBL_NBIG, DBL_SZBIG, "dblist big");
	  /* dbdict free lists */
      slcfreelistUtil(dbdictNameTrpFreeList_p, NENTRIES, 
						sizeof(dbdictNameTriplet_ts), 
						"dbdictNameTrpFreeList" );
      slcfreelistUtil(dbdictNameSglFreeList_p, NENTRIES, 
						sizeof(dbdictNameSgl_ts), 
						"dbdictNameSglFreeList" );
      slcfreelistUtil(dbdictPUSFreeList_p, NENTRIES, 
						sizeof(dbdictPUS_ts), 
						"dbdictPUSFreeList" );
      slcfreelistUtil(dbdictPFreeList_p, NENTRIES, 
						sizeof(dbdictP_ts), 
						"dbdictPFreeList" );
      slcfreelistUtil(dbdictUFreeList_p, NENTRIES, 
						sizeof(dbdictU_ts), 
						"dbdictUFreeList" );
      slcfreelistUtil(dbdictSFreeList_p, NENTRIES, 
						sizeof(dbdictS_ts), 
						"dbdictSFreeList" );
	  /* message service free lists */
	} else printf (" Invalid Command. The database does not exist.\n");

	/* the msg service lists do not need the database */
	slcfreelistUtil(slcSmMsgPvt_p, SMALL_MSGPOOL_ELEMENTS, 
					sizeof(msgmail_ts),
					"smallmsgFreeList" );
	slcfreelistUtil(slcLrgMsgPvt_p, LARGE_MSGPOOL_ELEMENTS, 
					sizeof(lrgmail_ts),
					"largemsgFreeList" );

  } else printf(" Invalid Command. The slc-aware threads have not started.\n");
}
/*=============================================================================

  Name: slcdbAddr

  Abs:  Diagnostic to dump information about various ioc free list items available

  Args: none

  Rem:  Called as an iocsh command
        Tests out dbNametoAddr calls that are failing on linux in the
		slc aware code.

  Side: 

  Ret: None

==============================================================================*/
/**procedure**/
void slcdbAddr(void)
{
  char         pvname_ca[80];           /* Epics PV channel name */
  DBADDR       pvAddr_s;

  mgntHdlrInit();

  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_CTRL);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_EPICSHSTA);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_STAT);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_TTIM);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_ITRY);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_BDES);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_BACT);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_BMON);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_IDES);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_IACT);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_RIPL);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_IPRV);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_IMON);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_IMAX);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_CTIM);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_KTIM);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);
  sprintf (pvname_ca, "LGPS:XX02:1:%s", PV_ZTIM);
  if ((dbNameToAddr(pvname_ca, &pvAddr_s))) 
	printf ("dbNameToAddr of %s failed\n", (char *)pvname_ca);


/* #define PV_TTIM "TRIMTOD.TIME" */ /* core dump - linux */
/*#define PV_CTIM "CALBTOD.TIME"*/ /* core dump - linux */
/*#define PV_KTIM "STDZTOD.TIME"*/ /* core dump - linux */
/*#define PV_ZTIM "STDZLOSTTOD.TIME"*/ /* core dump - linux */

}
