/*=============================================================================
 
  Name: slcRegister.c
        slcRegistrar - registers all "slcxxx" commands with the iocsh

  Abs:  shell commands must be registered before they can be recognized by
        IOC shell. Registration is achieved by calling a registration function
		with the appropiate parameters for each command.

  Auth: 02-Feb-2005, Debbie Rogind (drogind): created with some initial code 
               split from slcMain.c
  Rev:  02-Feb-2005 dfairley


-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        01-Apr-2005, Ron MacKenzie (ronm):
	   Add slcAsyncDumpGlob
        
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int slcRegFlag = DP_DEBUG;
#endif


/* c library includes */
#include	<string.h>

/* epics includes */
#include    <dbDefs.h>        /* NELEMENTS */
#include    <iocsh.h>         /* ioc shell stuff */
#include    <registryFunction.h>
#include    <epicsExport.h>   /* for registering slcRestart, slcStop */
#include    <epicsEvent.h>    /* epicsEventId */

/* slc common includes */
#include    "slcGlob.h"       /* globals, types, */
#include    "msgTypes.h"      /* msgmail_ts */
#include    "slcMsg.h"        /* slcIOC internal msg function codes */
#include    "msgfunc.hc"      /* FUNC_TEST job code */

/* slc utility prototypes */
#include    "msgQUtil.h"      /* message queue utilities */
#include    "slcCmlog.h"      /* slcCmlogLogMsg */
#include    "cvtMacr.h"       /* macro used for endian-ness */

/* thread prototypes  */
#include    "slcMain.h"    /* slcStart, slcRestart, slcStop */
#include    "dbshell.h"    /* slcdbXxxYyy shell commands */
#include    "slcExec.h"    /* slcExec thread */
#include    "slcAsync.h"   /* slcAsync thread */
#include    "slcRegister.h"

/* slcStart */
LOCAL iocshFuncDef slcStartFuncDef = 
  {"slcStart", 0, NULL };
LOCAL void slcStartCallFunc(const iocshArgBuf *args)
{
  slcStart();
}

/* slcStop */
LOCAL iocshFuncDef slcStopFuncDef = 
  {"slcStop", 0, NULL };
LOCAL void slcStopCallFunc(const iocshArgBuf *args)
{
  slcStop();
}

/* slcRestart */
LOCAL iocshFuncDef slcRestartFuncDef = 
  {"slcRestart", 0, NULL };
LOCAL void slcRestartCallFunc(const iocshArgBuf *args)
{
  slcRestart();
}
/* slcFreeLists */
LOCAL iocshFuncDef slcFreeListsFuncDef = 
  {"slcFreeLists", 0, NULL };
LOCAL void slcFreeListsCallFunc(const iocshArgBuf *args)
{
  slcFreeLists();
}
/* slcdbAddr */
LOCAL iocshFuncDef slcdbAddrFuncDef = 
  {"slcdbAddr", 0, NULL };
LOCAL void slcdbAddrCallFunc(const iocshArgBuf *args)
{
  slcdbAddr();
}

/*  dr add db service shell commands here */
/* slcdbSetHashTableSize */
static const iocshArg dbSetHashTableSizeArg0 = { "dbdictSize", iocshArgInt};
static const iocshArg * const dbSetHashTableSizeArgs[1] = {&dbSetHashTableSizeArg0};
static const iocshFuncDef dbSetHashTableSizeFuncDef = {"slcdbSetHashTableSize",1,dbSetHashTableSizeArgs};
LOCAL void dbSetHashTableSizeCallFunc(const iocshArgBuf *args)
{
  dbshellSetHashTableSize(args[0].ival);
}
/* slcdbDumpHash */
static const iocshFuncDef dbDumpHashFuncDef = {"slcdbDumpHash",0,NULL};
LOCAL void dbDumpHashCallFunc(const iocshArgBuf *args)
{
  dbshellDumpHash();
}
/* slcdbDumpHashFile */
static const iocshArg dbDumpHashFileArg0 = { "filename", iocshArgString};
static const iocshArg * const dbDumpHashFileArgs[1] = {&dbDumpHashFileArg0};
static const iocshFuncDef dbDumpHashFileFuncDef = {"slcdbDumpHashFile",1,dbDumpHashFileArgs};
LOCAL void dbDumpHashFileCallFunc(const iocshArgBuf *args)
{
  dbshellDumpHashFile(args[0].sval);
}
/* slcdbGetMeta */
static const iocshArg dbGetMetaArg0 = { "primary", iocshArgString};
static const iocshArg dbGetMetaArg1 = { "unit", iocshArgString};
static const iocshArg dbGetMetaArg2 = { "secondary", iocshArgString};
static const iocshArg * const dbGetMetaArgs[3] = 
  {&dbGetMetaArg0, &dbGetMetaArg1, &dbGetMetaArg2};
static const iocshFuncDef dbGetMetaFuncDef = {"slcdbGetMeta",3,dbGetMetaArgs};
LOCAL void dbGetMetaCallFunc(const iocshArgBuf *args)
{
  dbshellGetMeta(args[0].sval, args[1].sval, args[2].sval);
}
/* slcdbDump */
static const iocshArg dbDumpArg0 = { "primary", iocshArgString};
static const iocshArg dbDumpArg1 = { "unit", iocshArgString};
static const iocshArg dbDumpArg2 = { "secondary", iocshArgString};
static const iocshArg * const dbDumpArgs[3] = 
  {&dbDumpArg0, &dbDumpArg1, &dbDumpArg2};
static const iocshFuncDef dbDumpFuncDef = {"slcdbDump",3,dbDumpArgs};
LOCAL void dbDumpCallFunc(const iocshArgBuf *args)
{
  dbshellDump(args[0].sval, args[1].sval, args[2].sval);
}
/* slcdbGetUnits */
static const iocshArg dbGetUnitsArg0 = { "primary", iocshArgString};
static const iocshArg * const dbGetUnitsArgs[1] = {&dbGetUnitsArg0};
static const iocshFuncDef dbGetUnitsFuncDef = {"slcdbGetUnits",1,dbGetUnitsArgs};
LOCAL void dbGetUnitsCallFunc(const iocshArgBuf *args)
{
  dbshellGetUnits(args[0].sval);
}
/* slcdbGetExists */
static const iocshFuncDef dbGetExistsFuncDef = {"slcdbGetExists",0,NULL};
LOCAL void dbGetExistsCallFunc(const iocshArgBuf *args)
{
  dbshellGetExists();
}
/* slcdbGetVersion */
static const iocshFuncDef dbGetVersionFuncDef = {"slcdbGetVersion",0,NULL};
LOCAL void dbGetVersionCallFunc(const iocshArgBuf *args)
{
  dbshellGetVersion();
}
/* slcAsyncDumpGlob */
static const iocshFuncDef slcAsyncDumpGlobFuncDef = {"slcAsyncDumpGlob",0,NULL};
LOCAL void slcAsyncDumpGlobCallFunc(const iocshArgBuf *args)
{
  slcAsyncDumpGlob();
}
/* slcdbUpdate */
static const iocshArg dbUpdateArg0 = { "primary", iocshArgString};
static const iocshArg * const dbUpdateArgs[1] = {&dbGetUnitsArg0};
static const iocshFuncDef dbUpdateFuncDef = {"slcdbUpdate",1,dbGetUnitsArgs};
LOCAL void dbUpdateCallFunc(const iocshArgBuf *args)
{
  dbshellUpdate(args[0].sval);
}
/* slcdbEdit */
static const iocshArg dbEditArg0 = { "primary", iocshArgString};
static const iocshArg dbEditArg1 = { "unit", iocshArgString};
static const iocshArg dbEditArg2 = { "secondary", iocshArgString};
static const iocshArg dbEditArg3 = { "value", iocshArgString};
static const iocshArg * const dbEditArgs[4] = 
  {&dbEditArg0, &dbEditArg1, &dbEditArg2, &dbEditArg3};
static const iocshFuncDef dbEditFuncDef = {"slcdbEdit",4,dbEditArgs};
LOCAL void dbEditCallFunc(const iocshArgBuf *args)
{
  dbshellEdit(args[0].sval, args[1].sval, args[2].sval, args[3].sval);
}
/* slcdbEditST1 */
static const iocshArg dbEditST1Arg0 = { "primary", iocshArgString};
static const iocshArg dbEditST1Arg1 = { "unit", iocshArgString};
static const iocshArg dbEditST1Arg2 = { "secondary", iocshArgString};
static const iocshArg dbEditST1Arg3 = { "value", iocshArgString};
static const iocshArg * const dbEditST1Args[4] = 
  {&dbEditST1Arg0, &dbEditST1Arg1, &dbEditST1Arg2, &dbEditST1Arg3};
static const iocshFuncDef dbEditST1FuncDef = {"slcdbEditST1",4,dbEditST1Args};
LOCAL void dbEditST1CallFunc(const iocshArgBuf *args)
{
  dbshellEditST1(args[0].sval, args[1].sval, args[2].sval, args[3].sval);
}
LOCAL registryFunctionRef slcRef[] = {
  {"slcStart", (REGISTRYFUNCTION)slcStart},
  {"slcStop", (REGISTRYFUNCTION)slcStop},
  {"slcRestart",(REGISTRYFUNCTION)slcRestart},
  {"slcFreeLists",(REGISTRYFUNCTION)slcFreeLists},
  {"slcdbSetHashTableSize",(REGISTRYFUNCTION)dbshellSetHashTableSize},
  {"slcdbDumpHash",(REGISTRYFUNCTION)dbshellDumpHash},
  {"slcdbDumpHashFile",(REGISTRYFUNCTION)dbshellDumpHashFile},
  {"slcdbGetMeta",(REGISTRYFUNCTION)dbshellGetMeta},
  {"slcdbDump",(REGISTRYFUNCTION)dbshellDump},
  {"slcdbGetUnits",(REGISTRYFUNCTION)dbshellGetUnits},
  {"slcdbGetExists",(REGISTRYFUNCTION)dbshellGetExists},
  {"slcdbGetVersion",(REGISTRYFUNCTION)dbshellGetVersion},
  {"slcAsyncDumpGlob",(REGISTRYFUNCTION)slcAsyncDumpGlob},
  {"slcdbUpdate",(REGISTRYFUNCTION)dbshellUpdate},
  {"slcdbEdit",(REGISTRYFUNCTION)dbshellEdit},
  {"slcdbEditST1",(REGISTRYFUNCTION)dbshellEditST1},
  {"slcdbAddr", (REGISTRYFUNCTION)slcdbAddr}

};

/*=============================================================================

  Name: slcRegistrar

  Abs: add slcStart and slcStop to the function registry

  Args: none

  Rem: epics requires global functions to be registered for proper linking

  Side: 

  Ret:  none

==============================================================================*/
void slcRegistrar(void)
{
    registryFunctionRefAdd(slcRef,NELEMENTS(slcRef));
    iocshRegister(&slcStartFuncDef, slcStartCallFunc);
    iocshRegister(&slcStopFuncDef, slcStopCallFunc);
    iocshRegister(&slcRestartFuncDef, slcRestartCallFunc);
    iocshRegister(&slcFreeListsFuncDef, slcFreeListsCallFunc);
    iocshRegister(&dbSetHashTableSizeFuncDef, dbSetHashTableSizeCallFunc);
    iocshRegister(&dbDumpHashFuncDef, dbDumpHashCallFunc);
    iocshRegister(&dbDumpHashFileFuncDef, dbDumpHashFileCallFunc);
    iocshRegister(&dbGetMetaFuncDef, dbGetMetaCallFunc);
    iocshRegister(&dbGetUnitsFuncDef, dbGetUnitsCallFunc);
    iocshRegister(&dbDumpFuncDef, dbDumpCallFunc);
    iocshRegister(&dbGetExistsFuncDef, dbGetExistsCallFunc);
    iocshRegister(&dbGetVersionFuncDef, dbGetVersionCallFunc);
    iocshRegister(&slcAsyncDumpGlobFuncDef, slcAsyncDumpGlobCallFunc);
    iocshRegister(&dbUpdateFuncDef, dbUpdateCallFunc);
    iocshRegister(&dbEditFuncDef, dbEditCallFunc);
    iocshRegister(&dbEditST1FuncDef, dbEditST1CallFunc);
    iocshRegister(&slcdbAddrFuncDef, slcdbAddrCallFunc);
}
epicsExportRegistrar(slcRegistrar);

