/*=============================================================================
 
  Name: iocCmlog.c

       iocCmlogLogMsg      - entry point      (log message)
       iocCmlogVLogMsg     - entry point      (log message va_args)
       iocCmlogRegistrar   - external routine (register iocsh commands)
       iocCmlogSetCode     - internal routine 

  Abs:  This file contains the entry functions 
	used in the implementation of the IOC LOG Service.

  Auth: 13-Dec-2004, James Silva (JSILVA)
  Rev:  20-Jan-2005, S A Allison (SAA)

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        20-Sep-2006, Ron MacKenzie:
           Removed several functions like iocCmlogStart as part of project
            to log messages from RTEMS into cmlog.
           Change iocCmlogVLogMsg to add cmlog tags to the message and 
            send the message via errlogPrintf instead of sending it to 
            iocCmlogHdlr thread.
 
=============================================================================*/

#include        <stdio.h>              /* for printf               */
#include        <string.h>             /* for strcpy, strcmp, etc. */
#include        <stdarg.h>             /* for va_args              */
#include        "iocCmlog.h"           /* iocCmlog protos          */
#include        "debugPrint.h"         /* DEBUGPRINT               */

#include        "errlog.h"             /* for errlogPrintf         */
#include        "alarm.h"              /* NO_ALARM, etc            */
#include        "alarmString.h"        /* alarmStatusString        */
#include        "epicsThread.h"        /* for epicsThreadGetNameSelf */
#include        "epicsExport.h"        /* for epicsExportRegistrar */
#include        "epicsStdio.h"         /* for epicsVsnprintf       */
#include        "iocsh.h"              /* for iocshRegister        */

/***************************************************************************/

#define IOC_CMLOG_FACILITY_STR_LEN    8
#define IOC_CMLOG_TEXT_STR_LEN        256
#define IOC_CMLOG_STATUS_STR_LEN      32
#define IOC_CMLOG_SEVERITY_STR_LEN    16
#define IOC_CMLOG_PROCESS_STR_LEN     16
#define IOC_CMLOG_DEST_STR_LEN        8  
#define IOC_CMLOG_MSG_STR_LEN         400
#define IOC_CMLOG_WARN                "WARN"
#define IOC_CMLOG_SUCC                "SUCC"
#define IOC_CMLOG_EROR                "EROR"
#define IOC_CMLOG_INFO                "INFO"
#define IOC_CMLOG_FATL                "FATL"

typedef enum {
        cmlogVerbosityFatal     = 0, 
        cmlogVerbosityInvalid   = 1, 
        cmlogVerbosityError     = 2, 
        cmlogVerbosityWarning	= 3, 
        cmlogVerbosityInfo	= 4, 
        cmlogVerbositySuccess	= 5
}cmlogVerbosity_te;

/*
 * Global and File-Scope Data
 * --------------------------
 */
#ifdef  DEBUG_PRINT
  int iocCmlogFlag = DP_NONE;
#endif

/* For EPICS severity to verbosity translation */
#define IOC_CMLOG_SEVERITY_EPICS_MAX   4
static const int epicsVerbosity_a[]  = {cmlogVerbositySuccess,
                                        cmlogVerbosityWarning,
                                        cmlogVerbosityError,
                                        cmlogVerbosityInvalid};
/* For VMS status code translation */
#define IOC_CMLOG_SEVERITY_VMS_MAX     5
#define IOC_CMLOG_SEVERITY_VMS_DEFAULT 3
static const int  cmlogVerbosity_a[] = {cmlogVerbosityWarning,
                                        cmlogVerbositySuccess,
                                        cmlogVerbosityError,
                                        cmlogVerbosityInfo,
                                        cmlogVerbosityFatal};
static const char * vmsSeverity_ac[] = {IOC_CMLOG_WARN,
                                        IOC_CMLOG_SUCC,
                                        IOC_CMLOG_EROR,
                                        IOC_CMLOG_INFO,
                                        IOC_CMLOG_FATL};

/*
 * Internal function declarations
 * ------------------------------
 */
       void iocCmlogRegistrar(void);
static void iocCmlogSetCode  (const unsigned long codeI,
                              const char * const  code_c,
                              const char * const  facility_c,
                              const char * const  severity_c,
                                    char * const  result_c);



/*=============================================================================
 
  Name: iocCmlogLogMsg

  Abs:  Log message to IOC LOG Service. (variable arguments)

  
  Args: Type		    Name	     Access	   Description
	------------------- -------------    ---------- -----------------------
	const char * const  facility_c	     read-only  Facility string (if 
                                                        NULL, "EPICS" is used)
	const unsigned long status	     read-only  Status code
                                                        (EPICS or VMS)
	const int           severity	     read-only  EPICS severity code
                                                        (-1: status code is VMS
                                                          4: informational)
	const int           doPrintf	     read-only  0=LOG only
                                                        1=printf and LOG
	const char * const  status_c	     read-only  Status string
                                                        (will default if NULL)
	const char * const  destination_c    read-only  Destination string
                                                        (can be NULL)
	const char * const  filename_c	     read-only  File name (can be NULL)
	const unsigned long lineno	     read-only  File line number
                                                        (default 0)
	const char * const  format_c 	     read-only  Format string
	...                 (none)           read-only  Variable arguments

  Rem:  Function call to log message to LOG Service with variable
        arguments. Acts as a wrapper for iocCmlogVLogMsg.

  Side: Puts variable arguments into va_arg type. Passes arguments to
        iocCmlogVLogMsg
 
  Ret:  return status of iocCmlogVLogMsg (int)
  
=============================================================================*/

int iocCmlogLogMsg (const char * const  facility_c,
		    const unsigned long status,
		    const int           severity,
                    const int           doPrintf,
		    const char * const  status_c,
		    const char * const  destination_c,
		    const char * const  filename_c,
		    const unsigned long lineno,
		    const char * const  format_c, ...)
{
  va_list format_args_p;

  va_start(format_args_p, format_c);
    
  return iocCmlogVLogMsg(facility_c, status, severity, doPrintf, status_c,
                         destination_c, filename_c, lineno, format_c,
                         format_args_p);
}

/*=============================================================================
 
  Name: iocCmlogVLogMsg

  Abs:  Log message to IOC Log Service (with va_args argument)

  
  Args: Type		    Name	     Access	   Description
	------------------- -------------    ---------- -----------------------
	const char * const  facility_c	     read-only  Facility string (if 
                                                        NULL, "EPICS" is used)
	const unsigned long status	     read-only  Status code
                                                        (EPICS or VMS)
	      int           severity	     read-only  EPICS severity code
                                                        (-1: status code is VMS
                                                          4: informational)
	const int           doPrintf	     read-only  0=LOG only
                                                        1=printf and LOG
	const char * const  status_c	     read-only  Status string
                                                        (will default if NULL)
	const char * const  destination_c    read-only  Destination string
                                                        (can be NULL)
	const char * const  filename_c	     read-only  File name (can be NULL)
	const unsigned long lineno	     read-only  File line number
                                                        (default 0)
	const char * const  format_c 	     read-only  Format string
	va_list             format_args_p    read-only  Variable arguments

 
  Rem: Generates message to be sent to LOG server. Performs an errlogPrintf to
       send the message to the iocLogClient.

  Side: None
 
  Ret: Return from errlogPrintf.
  
=============================================================================*/

int iocCmlogVLogMsg(const char * const  facility_c, 
		    const unsigned long status, 
		          int           severity, 
                    const int           doPrintf,
		    const char * const  status_c, 
		    const char * const  destination_c,             
		    const char * const  filename_c,  
		    const unsigned long lineno,           
		    const char * const  format_c,  
		    va_list             format_args_p)
{
  int           len = 0;
  unsigned long verbosity;
  char        * process_c;
  char          msg_c[IOC_CMLOG_MSG_STR_LEN];
  char          stat_c[IOC_CMLOG_STATUS_STR_LEN];
  const int maxSeverity= (IOC_CMLOG_SEVERITY_EPICS_MAX < ALARM_NSEV) ?
                          IOC_CMLOG_SEVERITY_EPICS_MAX : ALARM_NSEV;
  
  /* Copy from input strings - check for maximum string sizes */
  msg_c[0] = 0;
  if (facility_c)
    len += sprintf(msg_c+len, "fac=%.*s ",  IOC_CMLOG_FACILITY_STR_LEN,
                   facility_c);
  if (destination_c)
    len += sprintf(msg_c+len, "dest=%.*s ", IOC_CMLOG_DEST_STR_LEN,
                   destination_c);
  /* Now copy our thread name */
  if ((process_c = (char *)epicsThreadGetNameSelf()))
    len += sprintf(msg_c+len, "proc=%.*s ", IOC_CMLOG_PROCESS_STR_LEN,
                   process_c);
  /* If an EPICS status is passed in, set verbosity,    */
  /*   severity string, and status string appropriately */
  if (severity >= 0) {
    if (severity < maxSeverity) {
      verbosity = epicsVerbosity_a[severity];
      len += sprintf(msg_c+len, "sevr=%.*s ", IOC_CMLOG_SEVERITY_STR_LEN,
                     alarmSeverityString[severity]);
    }
    else {
      verbosity = cmlogVerbosityInfo;
    }
    if (status_c)
      len += sprintf(msg_c+len, "stat=%.*s ", IOC_CMLOG_STATUS_STR_LEN,
                     status_c);
    else if (status < ALARM_NSTATUS)
      len += sprintf(msg_c+len, "stat=%.*s ", IOC_CMLOG_STATUS_STR_LEN,
                     alarmStatusString[status]);
  }  
  /* Get severity and verbosity based on the condition code. */
  /* Condition code is assumed to be VMS so lower 3 bits give severity */
  else {
    severity = status & 0x7;
    if (severity < IOC_CMLOG_SEVERITY_VMS_MAX) {
      verbosity = cmlogVerbosity_a[severity];
      len += sprintf(msg_c+len, "sevr=%.*s ", IOC_CMLOG_SEVERITY_STR_LEN,
                     vmsSeverity_ac[severity]);
    }
    else {
      severity  = IOC_CMLOG_SEVERITY_VMS_DEFAULT;
      verbosity = cmlogVerbosityInfo;
    }
    /* Now construct the condition code string which consists of */
    /*   the subsystem, severity character, and input code with  */
    /*   -'s as delimiters */
    iocCmlogSetCode(status, status_c, facility_c,
                    vmsSeverity_ac[severity], stat_c);
    len += sprintf(msg_c+len, "stat=%.*s ", IOC_CMLOG_STATUS_STR_LEN,
                   stat_c);
    len += sprintf(msg_c+len, "code=%lu ", status);
  }
  len += sprintf(msg_c+len, "verb=%lu ", verbosity);
  /* Now format the message text */
  if (format_c)
    len += epicsVsnprintf(msg_c+len, IOC_CMLOG_MSG_STR_LEN - len - 1,
                          format_c, format_args_p);
  else
    len += sprintf(msg_c+len, "\n");

  DEBUGPRINT(DP_ERROR, iocCmlogFlag,
             ("Message = %s\n", msg_c));

  if (doPrintf) return (errlogPrintf(msg_c));
  return (errlogPrintfNoConsole(msg_c));
}

/*=============================================================================
 
  Name: iocCmlogSetCode

  Abs:  Fill in condition code string that matches input message code and
        input condition.
  
  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
	const unsigned long codeI       read-only  integer-based message code
	const char * const  code_c	read-only  String version of code
                                                   Can be NULL.
	const char * const  facility_c	read-only  Facility string
                                                   Can be NULL.
        const char * const  severity_c  read-only  Severity string
                                                   Must be non-NULL.
        char * const        result_c    read/write Condition code string
                                                   Must be non-NULL.
 
  Rem: First, find facility code from input message code and fill in matching
       facility string from existing list of facility code
       translations. If none exists, use the input facility.
       The char string were taken from PREFIX in ref_slcmsg:*def.msg.
       The values were taken from *$FACILITY in ref_c_inc:*def.h*.

       Then add the severity character and the input code string with -'s
       as delimiters.

  Side: None.
 
  Ret:  None
  
=============================================================================*/

static void iocCmlogSetCode(const unsigned long codeI,
                            const char * const  code_c,
                            const char * const  facility_c,
                            const char * const  severity_c,
                                  char * const  result_c)
{
  /*
   * facilityCode - from top 2 bytes of the message code -  
   * mask and shift down.
   */
  int facilityCode = (codeI >> 16) & 0x00000FFF;
  int resultLen;

  /* blank char first */
  memset(result_c, 0, IOC_CMLOG_STATUS_STR_LEN);
  switch (facilityCode) {

      case 0x00000871: strcat(result_c, "ASYNC"   ); break;
      case 0x00000844: strcat(result_c, "BPMO"    ); break;
      case 0x00000847: strcat(result_c, "CRAT"    ); break;
      case 0x00000801: strcat(result_c, "DATABASE"); break;
      case 0x00000841: strcat(result_c, "LGPS"    ); break;
      case 0x00000840: strcat(result_c, "MGNT"    ); break;
      case 0x00000814: strcat(result_c, "MICR"    ); break;
      case 0x00000820: strcat(result_c, "MPG"     ); break;
      case 0x0000081C: strcat(result_c, "MSG"     ); break;
      case 0x00000805: strcat(result_c, "OPS"     ); break;
      case 0x00000848: strcat(result_c, "TIME"    ); break;
      case 0x00000000: strcat(result_c, "SYSTEM"  ); break;
      default:         if (facility_c && strlen(facility_c))
                          strcat(result_c, facility_c);
                       else
                          strcat(result_c, "EPICS");
                       break;
  }
  resultLen = strlen(result_c);
  result_c[resultLen++] = '-';
  result_c[resultLen++] = severity_c[0];
  result_c[resultLen++] = '-';
  if (code_c && code_c[0])
    strncat(result_c, code_c,     IOC_CMLOG_STATUS_STR_LEN - resultLen - 1);
  else
    strncat(result_c, severity_c, IOC_CMLOG_STATUS_STR_LEN - resultLen - 1);
}

/*=============================================================================
 
  Name: iocCmlogRegistrar

  Abs:  Register external functions with ioc shell.

  Args: None.
 
  Rem:  Register external function
        iocCmlogMsg  using iocshRegister. 

  Side: Registers functions for use in IOC command-line shell.
 
  Ret:  None.
 
=============================================================================*/

static const iocshArg iocCmlogMsgArg0_s    = {"facility",    iocshArgString}; 
static const iocshArg iocCmlogMsgArg1_s    = {"status",      iocshArgInt   }; 
static const iocshArg iocCmlogMsgArg2_s    = {"severity",    iocshArgInt   }; 
static const iocshArg iocCmlogMsgArg3_s    = {"text",        iocshArgString};
static const iocshArg iocCmlogMsgArg4_s    = {"statusString",iocshArgString}; 
static const iocshArg iocCmlogMsgArg5_s    = {"destination", iocshArgString}; 

static const iocshArg *const iocCmlogMsgArgs_as[6] =
        {&iocCmlogMsgArg0_s, &iocCmlogMsgArg1_s, &iocCmlogMsgArg2_s,
         &iocCmlogMsgArg3_s, &iocCmlogMsgArg4_s, &iocCmlogMsgArg5_s};

static const iocshFuncDef iocCmlogMsgDef_s       = {"iocCmlogMsg",       6,
                                                    iocCmlogMsgArgs_as      }; 

static void iocCmlogMsgCall       (const iocshArgBuf * args_as) {
  iocCmlogLogMsg(args_as[0].sval,
                 (unsigned long)args_as[1].ival,
                 args_as[2].ival,
                 IOC_CMLOG_PRINTF,
                 args_as[4].sval, args_as[5].sval, NULL, 0, args_as[3].sval);
  errlogPrintf("\n");
}

void epicsShareAPI iocCmlogRegistrar(void) {
	iocshRegister(&iocCmlogMsgDef_s       , iocCmlogMsgCall      );
}
epicsExportRegistrar(iocCmlogRegistrar);
