/*=============================================================================
 
  Name: mgntAsync.c
        mgntAysncThread   - Main routine for magnet async thread
		mgntAsyncCheckDoProceed - see if 10sec MGNT CHCK was performed recently
		mgntAsyncInit     - Init, beginning of thread
        mgntAsyncIoc      - IOC_ASYNC message from cstrHdlr 
		mgntAsyncCleanup  - cleanup, end of thread
		mgntAsyncCTRLHdlr - Called when receive message from mgntMonCTRLCallback
		mgntAsyncFuncDone - Called when mgntAsyncCTRLHdlr determines scp req completed
		mgntAsyncHSTAHdlr - Called when receive message from mgntMonHSTACallback
		mgntAsyncBDESHdlr - Called when receive message from mgntMonBDESCallback
		mgntAsyncCheck    - writes changes for many devices to HSTA, BDES, BACT, BMON, STAT, 
		                    IDES, IACT, IMON, RIPL, IPRV, TTIM, ITRY, KTIM, ZTIM, IMAX, CTIM.
							Performed when scp request is not active.
							Called from IOC_ASYNC 10 sec processing, and as final check when a 
							scp request is done/timed out. 
		mgntAsyncCheckFeedback - writes changes to secns above for devices under feedback ctrl
		mgntAsyncCheckUnit- utility routine used by mgntAsyncCheck and mgntAsyncCheckFeedback,
		                    also called from CTRL=IDLE/unit processing, 
		mgntAsyncCheckUnitSigChange - determines if unit's STAT changes, or if unit's
		                    BACT or BMON has significant change
		mgntAsyncBactSigChange - determines if BACT has sig change, using TOLS ST1 data
		mgntAsyncBmonSigChange - determines if BMON has sig change, using TOLF ST1 data
		mgntAsyncPVToSlcFloat - writes float secn from epics to slc 
		mgntAsyncPVToSlcUShort - writes unsigned short secn from epiocs to slc
		mgntAsyncPVToSlcTime - writes converted epics time -> vms time to slc
		mgntAsyncSlcToPVBdes - writes slc BDES value to epics
		mgntAsyncSlcToPVHstaCtrl - writes slc HSTA value to epics, HSTACTRL record
		fmax              - returns floating point maximum of two input arguments
		mgntAsyncTime     - Called from 1 sec IOC_ASYNC function

  Abs:  Perform magnet a local async message request

  Args: none

  Rem: This thread receives periodic IOC_ASYNC messages which it receives from the
       the CSTR handler. The mgntCheck() routine is eventuallyl called and it 
       reads all magnet adc pv's from the EPCIS db and then determine if any value
       need to be update in the SLC database.

       Please note that as the message is local a reply message is not requried.

  Side: This thread is started by SLC Exec after the magnet handler active
        flag has been set, which MUST be done after the database initialization 
        for the magnet handler is complete.

        reference document:
          http://www.slac.stanford.edu/grp/lcls/controls/global/facilities/slc_ioc/requirments/mgnt_designSpec.html

  Ret:  None

  Auth: 10-Jul-2005, K.  Luchini      (LUCHINI)
  Rev:  dd-mmm-yyyy, Reviewer's Name  (USERNAME)

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  
        dd-mmm-yyyy, First Lastname (USERNAME):
           comments

=============================================================================*/

/* C system header files */
#include <string.h>        /* memcpy                             */
#include <stdio.h>
#include <math.h>          /* fabs                               */
/* Debug header files */
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int mgntAsyncFlag = DP_ERROR;
#endif

/* epics */
#include <dbAccess.h>      /* EPICS dbAddr struct */
#include "epicsTime.h"     /* epicsTimeGetCurrent()              */
#include <epicsEvent.h>    /* epicsEventId                       */

/* SLC header files */
#include "slcTypes.h"      /* vmsstat_t,mgntAsync,                */
#include "slcGlob.h"       /* slcThreads_as                       */
#include "slcCmlog.h"      /* slcCmlogLogMsg()                    */
#include "slcUtil.h"       /* slcGetJobEnumByName()               */
#include "slcAsync.h"      /* asyncmail_ts                        */
#include "slcMsg.h"        /* MGNT_NOTIFY                         */
#include "timeUtil.h"      /* time utils                          */
#include "opdef.hc"        /* error codes                         */

/* Database header files */
#include "dbupdate.h"      /* dbupdate()                          */
#include "dbl.h"           /* dblget, dblput                      */
#include "dblist.h"        /* DB macros                           */

/* Message and miscellaneous header files */
#include "msgTypes.h"      /* msgmail_ts etc                      */
#include "msgQUtil.h"      /* msgQMsgProcess(),messageQDestroy()  */

/* Magnet header files */
#include "mgntdef.hc"      /* Magnet message codes                 */
#include "mgntdefMsg.h"    /* Magnet message code strings          */
#include "mgntstat.h"      /* HSTA_* bits                          */
#include "mgntMacros.h"    /* MGNT_LOCAL_FUNC,MGNT_NUM_LOCAL_FUNCS */
#include "mgntTypes.h"     /* mgntEHdlrInfo_ts                     */
#include "mgntAsync.h"     /* mgntAsyncPVToSlc*()                   */

/*
** Local prototypes  
*/
LOCAL vmsstat_t    mgntAsyncInit(void);
LOCAL vmsstat_t    mgntAsyncIoc (const msgmail_ts* rMsg_ps);
LOCAL int          mgntAsyncCleanup(void);
LOCAL vmsstat_t    mgntAsyncCTRLHdlr (msgmail_ts * rMsg_ps); 
LOCAL vmsstat_t    mgntAsyncFuncDone(void);
LOCAL vmsstat_t    mgntAsyncHSTAHdlr (msgmail_ts * rMsg_ps);
LOCAL vmsstat_t    mgntAsyncBDESHdlr (msgmail_ts * rMsg_ps); 
LOCAL vmsstat_t    mgntAsyncCheck(const epicsBoolean useTolerances,
							   const epicsBoolean idleUnitsOnly,
							   const epicsBoolean doDbUpdate,
							   const mgntMaster_te mastertype_e,
							   const slcReqType_te requestType);
LOCAL vmsstat_t    mgntAsyncCheckFeedback( const slcReqType_te requestType);
LOCAL epicsBoolean mgntAsyncCheckUnitSigChange (mgntUnitStruct_ts * const unit_ps);
LOCAL vmsstat_t    mgntAsyncCheckUnit (mgntUnitStruct_ts * const unit_ps,
									   const mgntMaster_te masterType_e);  
LOCAL float        fmax (float x, float y);
LOCAL epicsBoolean mgntAsyncBactSigChange(mgntUnitStruct_ts * unit_ps,
										  float * bactSigChgTol); 
LOCAL epicsBoolean mgntAsyncBmonSigChange(mgntUnitStruct_ts * unit_ps,
										  float bactSigChgTol);  
LOCAL vmsstat_t    mgntAsyncPVToSlcTime ( DBADDR *dbaddr_ps,
										  const dbheader_ts * const dblist_ps,
										  dbheader_ts **const dbdata_pps); 
LOCAL vmsstat_t    mgntAsyncSlcToPVBdes( mgntUnitStruct_ts * const unit_ps);
LOCAL vmsstat_t    mgntAsyncSlcToPVHstaCtrl  (mgntUnitStruct_ts * const unit_ps);
LOCAL vmsstat_t    mgntAsyncTime (void);
/*
** Global variables
*/
extern mgntHdlrInfo_ts         *mgnt_ps;
extern mgntPrimStruct_ts       mgntPrim_s[MGNT_NUM_PRIMS];
/*=============================================================================

  Name: mgntAsyncInit

  Abs:  Init globals shared between mgntHdlr and mgntAsync

  Args: none

  Rem: 

  Side: 

  Ret:  None

==============================================================================*/
LOCAL vmsstat_t  mgntAsyncInit( void )
{
  vmsstat_t        status      = MGNT_OKOK;
  static const int maxMsgs     = MAX_QUEUED_MSGS;
  static const int maxMsgBytes = MAX_MSGPTR_BYTES;

  /* Create the message queue id */
  status = messageQCreate(mgntAsync,maxMsgs,maxMsgBytes);

  return( status );
}

/*=============================================================================

  Name: mgntAsyncThread

  Abs:  The magnet asyncthread

  Args: none

  Rem: This thread accepts and acts on FUNC_TEST messages 
       from the SLC control system

  Side:
  mgntAsync can recieve the following messages at its queue:
  A)"MGNT TIME" from cstrAsync (1 sec intervals) 
    counts down active scp function timer if > 0; 
    checks for function timeout
    meters dbupdates 1/sec if update flag set from MGNT_NOTIFY
    TBD anything else
  B)"MGNT CHCK" from cstrAsync (10 sec intervals)
    If active SCP function - do nothing.
    Performs check function - for all prim/units updates sig changes (within TOLS) to 
    slc database from EPICS for non-fbck devices.  Does all the dblputs, then dbupdate 
    if anything changed.
  C)"MGNT CHKF" from cstrAsync (1 min intervals)
    for all fbck prim/unit devices updates changes (BDES,STAT,ACT) to slc from EPICS 
    (TOLS ignored) do all the dblputs, then dbupdate at the end if anything changed 
  D) MNGT_CHCK  from mgntHdlr (user inititated Check function) 
    Perform MGNT CHCK, igore tolerances 
    Perform MGNT CHKF (tolerances already ignored)
  E) MGNT_*_NOTIFY from CA Monitor Callback, where *=CTRL, BDES, HSTA
    Data: 1) ptr to unit structure, 2) value
    CTRL != IDLE sets unit's busy flag (scp func units have add'l flag, scp=T, set in mgntHdlr)
    CTRL == IDLE && active scp function
      reads STATUS, updates unit and summary status,
      read associated PVs, update slc, set update flag if changes decs monitor counter, 
      clears busy flag, checks monitor counter=0 to see if function done
      if done, does a MGNT CHCK to update everything, then sets event flag to release mgntHdlr
    CTRL == IDLE && fbck unit
      clear busy flag; nothing else ("MGNT CHKF" takes care of updates); ?
    CTRL == IDLE && CA user
      clear busy flag, update BACT, STAT (not BDES), set update flag if changed
      TBD HSTA/DES - dbput if changed, dbupdate if changed		
      TBD on DES
  F) IOC_STOP from slcExec
    stop and cleanup

  Ret:  None

==============================================================================*/
void  mgntAsyncThread(void)
{
  vmsstat_t       status=MGNT_OKOK;         /* status return                  */
  unsigned short  func;                     /* incoming message function      */
  unsigned long   update=0;                 /* update SLC database            */

  slcJob_te       jobName_e=invalidJob;     /* job name                       */
  epicsTimeStamp  msgReqTime_s;             /* request message read time      */
  msgmail_ts     *recvMsg_ps   = NULL;      /* incoming mail message          */
  slcThread_ts   *mgntAsync_ps = NULL;      /* pointer to my thread struct    */
  epicsBoolean    relMsgBuf = epicsTrue;

  /* 
  ** First verify that the magnet handler initialization 
  ** has completed successfully. If not, then simply
  ** exit.
  */
  /*if ( !mgnt_ps->diag_s.init ) return;*/

  /* Get the thread information */
  mgntAsync_ps = &slcThreads_as[mgntAsync];

  /* Get job name */
  jobName_e = slcGetJobEnumByName(MGNT_JOB_NAME);

  /* Create the message queue id */
  status = mgntAsyncInit();

  if (SUCCESS (status)) mgntAsync_ps->active = epicsTrue;

  DEBUGPRINT(DP_INFO, mgntAsyncFlag, ("MGNT_A: active\n"));
  
  /*
  ** Wait for an incoming messages, handle it, then send a reply to the
  ** msgSend thread but only after the function has been completed.
  */
  while(mgntAsync_ps->active) {
	/* 
	** Read an incoming message from the queue. 
	** Note: this is a local message and we only expect
	** to receive an IOC_ASYNC or an IOC_STOP message.
	** The IOC_ASYNC messages are sent to this thread 
	** periodically by the CSTR handler. The period that
	** these messages are sent is defined in a CSTR secondary.
	** this time can only be changed upon an reboot of the SLC
	** Aware ioc.
	*/
	status = msgQMsgRecv(mgntAsync,&msgReqTime_s,(void*)&recvMsg_ps);
	if (SUCCESS(status)) {
	  relMsgBuf = epicsTrue;
	  /*
	  ** Should we drop the IOC_ASYNC on the floor?
	  ** This is done if the adc's have all been recently
	  ** updated. Remember that the adc's are all updated
	  ** not just upon receiving a MGNT_CHCK function but
	  ** prior to performing any of the magnet functions.
	  ** So here we look at the last time that the magnet
	  ** handler processed an incoming magnet message from an
	  ** Alpha process.
	  ** Note: lastChckTime_s will be updated in the mgntAsyncCheck() 
	  ** function. 
	  */

	  func = recvMsg_ps->msgheader.func;

	  switch(recvMsg_ps->msgheader.func){
	  case IOC_STOP:
		slcThreads_as[mgntHdlr].active = epicsFalse; 
		break;
	  case IOC_ASYNC:
		/* Note: make sure CMSK bits 0x00E0 are set for mgnt cycle functions */
		/* Note:make sure CSTR:MICRO:HSTA "magnets online" bit is set to get async messages 
		   ! BIT    CDSC                       HSTA                           BIT
		   !
		   ! 0001   VIRTUAL MICRO              MAGNETS ONLINE                0001  <snip>
		*/
		status = mgntAsyncIoc(recvMsg_ps);
		break;
	  case MGNT_CHCK:
		relMsgBuf = epicsFalse;
		DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("MGNT_A: Received MGNT_CHCK from MGNT_H\n"));
		/* perform scp requested MGNT CHCK forwarded from mgntHdlr (no tols) */
		status = mgntAsyncCheck( epicsFalse,  /* dblput any difference, regardless of tolerances */
						epicsTrue,   /* all devices with CTRL = IDLE only */
						epicsTrue,   /* do the update now */
						slcBdesHstaMaster_e,    /* slcAsyncMeterDbSend, not dbupdate */
						slcMsg);      /* request type (slcMsg, slcAsync) */
		/* perform scp requested MGNT CHKF forwarded from mgntHdlr (no tols) */
		status = mgntAsyncCheckFeedback(slcMsg);		
		/* now release mgntHdlr from waiting for completion */
		if (mgnt_ps->func_s.funcDoneEvent_ps != NULL){
		  DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("MGNT_A: signaling funcDoneEvent_ps \n"));
		  epicsEventSignal(mgnt_ps->func_s.funcDoneEvent_ps);
		}
		break;
	  case MGNT_NOTIFY_CTRL:
		/*message from CTRL callback, with val of CTRL  */
		/* and ptr to proper unit in mgntPrim_s.unit_ps */
		DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("MGNT_A: Received CTRL callback message \n"));
		status = mgntAsyncCTRLHdlr(recvMsg_ps);
		break;
	  case MGNT_NOTIFY_HSTA:
		/*message from HSTA callback, with val of HSTA  */
		/* and ptr to proper unit in mgntPrim_s.unit_ps */
		DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("MGNT_A: Received HSTA callback message \n"));
		status = mgntAsyncHSTAHdlr (recvMsg_ps);   
		break;
	  case MGNT_NOTIFY_BDES:
		/*message from BDEs callback, with val of BDES */
		/* and ptr to proper unit in mgntPrim_s.unit_ps */
		DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("MGNT_A: Received BDES callback message \n"));
		status = mgntAsyncBDESHdlr(recvMsg_ps);		  
		break;
	  default:
		/* unexpected message */
		DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
				   ("MGNT_A: Received unexpected msg func code = %x\n", 
					recvMsg_ps->msgheader.func));
		break;
	  } /* switch */		  
	  
	  /*
	  ** Determine if the Alpha db should be updated.
	  */
	  if ( update )
		status = dbupdate( jobName_e );   
	}  /* if successful read of message from the queue */

	  /* 
	  ** Release the message receive buffer before
	  ** checking for another message in the queue.
      ** This buffer is released by mgntHdlr after it sends a reply 
	  */
	if (relMsgBuf)
	msgQFreeBuffer(sizeof(msgmail_ts),(void *)recvMsg_ps);

	/* check the stop flag just in case */
	if (mgntAsync_ps->stop) 
	  mgntAsync_ps->active = epicsFalse;
  } /* while active */
  
  /* active not set, time to shut down */
  DEBUGPRINT(DP_INFO, mgntAsyncFlag, ("MGNT_A: shutdown\n"));
  
  /* set event flag if empty, delete queue,.. */
  mgntAsyncCleanup();
  return;
}
/*=============================================================================

  Name: mgntAsyncIoc ()

  Abs:  Decodes data from IOC_ASYNC message from cstrHdlr. For magnet, will be


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  

  side: 
  Data:
  "MGNT CHCK" is for 10 second update, 
  "MGNT CHKF" is for 1 minute update, and 
  "MGNT TIME" is for the 1 second actions.  

  For the 10 second function MGNT CHCK, read all the non-FBCK devices, 
  check TOLS/TOLF and do the proper dblputs and dbupdate at the end.  
  For the 1 minute function MGNT CHKF, you'll read all the FBCK devices, 
  and dblput anything that is different (no TOLS/TOLF check) and dbupdate at the end. 
  Devices are under feedback control when the feedback-control bit in their SLC HSTA is set,
   (dblget HSTA).

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL vmsstat_t mgntAsyncIoc (const msgmail_ts* rMsg_ps)
{
  asyncmail_ts asyncmail_s;
  vmsstat_t status = MGNT_OKOK;

  memcpy(&asyncmail_s, rMsg_ps->reqdata, sizeof(asyncmail_s)); 
  if ( strncmp((char *)&asyncmail_s.cyclName, "TIME", 4)==0 ) {
	DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
	  ("MGNT_A: IOC_ASYNC - MGNT TIME\n"));
	status = mgntAsyncTime();
  } else if ( strncmp((char *)&asyncmail_s.cyclName, "CHCK", 4)==0 ) {
	DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
	  ("MGNT_H: IOC_ASYNC - MGNT CHCK\n"));
	/* see if enough time has elapsed to perform check again first */
	if (mgntAsyncCheckDoProceed())
		status = mgntAsyncCheck( epicsTrue,    /* use tolerances */
						epicsTrue,    /* all devices with CTRL = IDLE only */
						epicsTrue,    /* do the update now */
						slcBdesHstaMaster_e,     /* dbupdate, don't slcAsyncMeterDbSend */ 
						slcAsync);      /* request type (slcMsg, slcAsync) */  
  } else if ( strncmp((char *)&asyncmail_s.cyclName, "CHKF", 4)==0 ) {
     DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
	  ("MGNT_H: IOC_ASYNC - MGNT CHKF\n"));
	status = mgntAsyncCheckFeedback(slcAsync);    /*MGNT CHKF (defaults no tols) */
  } else
	DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
				 ("MGNT_A: can't decode IOC_ASYNC\n"));
  return(status);
}
/*=============================================================================

  Name: mgntAsyncBDESHdlr ()

  Abs:  Decodes data from MGNT_BDES_NOTIFY message. Handles a change for BDES
        PV in epics db. Update slc db locally from epics if slc db value differs 
		from epics, and only if device is not under feedback control, as determined
		by reading HSTA. Update to Alpha is done by 1 sec processing.


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  

  Side: When BDES changes, we know that either
        1) a CA user could have changed the BDES PV
		2) as a result of a scp function request, mgntHdlr output to epics BDES PV
		3) SLC is master and output to epics BDES PV (?)

  Data: 1st 4 bytes is pointer to unit in mgntPrim_s.units_ps
        2nd 4 bytes is float value of BDES
 


  Ret: iss vmsstat_t vms style status

===============================================================================*/

LOCAL vmsstat_t mgntAsyncBDESHdlr (msgmail_ts * rMsg_ps) 
{
  vmsstat_t status = MGNT_OKOK;
  mgntUnitStruct_ts * unit_ps;
  long  nreqs=1;      /* these are for fetching Epics PV's */
  long  options=0;
  /* debug */
  float bdesslc, bdesepics;

  memcpy(&unit_ps, &rMsg_ps->reqdata[0], 4);  
  if (!unit_ps) {
     DEBUGPRINT(DP_ERROR, mgntAsyncFlag, ("MGNT_A: Error accessing BDES Callback msg data \n"));
	 return (status = MGNT_DBERR);
  }
  /* update unit's bdesVal */
  memcpy(&(unit_ps->bdesVal), &(rMsg_ps->reqdata[2]), sizeof(float));
  if (unit_ps->hstaAddr_s.precord) {
	/* see if unit is under feedback control, as determined by HSTA bit */
	if ( dbGetField( &(unit_ps->hstaAddr_s), DBR_USHORT, (void *)&(unit_ps->hstaVal),
					 &options, &nreqs, NULL) ) {
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, ("MGNT_A: dbGetField error\n"));
	  return (MGNT_DBERR);	
	}
	/* if under feedback control, don't update */
	if (unit_ps->hstaVal & HSTA_FBKCONTROL) {
	  DEBUGPRINT(DP_INFO, mgntAsyncFlag, ("MGNT_A: Under feedback control; hsta val=0x%x\n", unit_ps->hstaVal));
	  return status;
	}
  }
	
  /* update local db if changed */
  DBCLEAR(unit_ps->bdesDlist_ps);
  dblget( (const dbheader_ts * const)unit_ps->bdesPlist_ps, 
		  (dbheader_ts **const)&(unit_ps->bdesDlist_ps) );
  /* debug */
  bdesslc = *(float *)DBPTR_FIRST(unit_ps->bdesDlist_ps);
  bdesepics = unit_ps->bdesVal;
  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("MGNT_A: BDES slc: %f; BDES epics: %f \n",bdesslc, bdesepics ));
  /* */

  if (*(float *)DBPTR_FIRST(unit_ps->bdesDlist_ps) != unit_ps->bdesVal ) {
	memcpy (DBPTR_FIRST(unit_ps->bdesDlist_ps), &(unit_ps->bdesVal), sizeof(float));
	DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("mgntAsyncBDESHdlr: dblput; BDES slc: %f; BDES epics: %f \n",bdesslc, bdesepics )); 
	dblput((const dbheader_ts * const)unit_ps->bdesPlist_ps, 
		   (const dbheader_ts * const)unit_ps->bdesDlist_ps, mgntJob);
	/* wait for 1 sec async to update Alpha */
	mgnt_ps->updateTimeout = 1;

  }

  return status;
}
/*=============================================================================

  Name: mgntAsyncHSTA Hdlr ()

  Abs:  Decodes data from MGNT_HSTA_NOTIFY message. Handles a change for HSTA
        PV in epics db.  Update slc db locally if epics pv value differs from
		the slc value.  Update to Alpha occurs on the 1 sec async processing.
		Update slc regardless of whether or not the unit is under feedback control.
        Update HSTACTRL PV (slc hsta version)


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  

  Side: There is a HSTACTRL pv and a HSTA readback PV.
        This is the HSTA readback PV.

  Data: 1st 4 bytes is pointer to unit in mgntPrim_s.units_ps
        2nd 2 bytes is unsigned short value of the CTRL
 


  Ret: iss vmsstat_t vms style status

===============================================================================*/

LOCAL vmsstat_t mgntAsyncHSTAHdlr (msgmail_ts * rMsg_ps) 
{
  vmsstat_t status = MGNT_OKOK;
  mgntUnitStruct_ts * unit_ps;
  long nreqs = 1;

 
  memcpy(&unit_ps, &rMsg_ps->reqdata[0], 4);  
  if (!unit_ps) {
     DEBUGPRINT(DP_ERROR, mgntAsyncFlag, ("MGNT_A: Error accessing HSTA Callback msg data \n"));
	 return (status = MGNT_DBERR);
  }
  /* update unit's hstaVal */
  memcpy(&(unit_ps->hstaVal), &(rMsg_ps->reqdata[2]), 2);
  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("mgntAsyncHSTAHdlr: HSTA changed for %s = %x\n", 
										(&(unit_ps->hstaCtrlAddr_s))->precord->name, unit_ps->hstaVal));
  /* update local & Alpha db if changed */
  DBCLEAR(unit_ps->hstaDlist_ps);
  dblget( (const dbheader_ts * const)unit_ps->hstaPlist_ps, 
		  (dbheader_ts **const)&(unit_ps->hstaDlist_ps) );
  if (*(unsigned short *)DBPTR_FIRST(unit_ps->hstaDlist_ps) != unit_ps->hstaVal ) {
	memcpy (DBPTR_FIRST(unit_ps->hstaDlist_ps), &(unit_ps->hstaVal), sizeof(unsigned short));
	dblput((const dbheader_ts * const)unit_ps->hstaPlist_ps, 
		 (const dbheader_ts * const)unit_ps->hstaDlist_ps, mgntJob);
	/* wait for 1 sec async to update Alpha */
	mgnt_ps->updateTimeout = 1;
	/* output to HSTACTRL, which is the slc version of hsta */
	DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("mgntAsyncHSTAHdlr: dbPutField HSTA for %s of %x\n", 
										(&(unit_ps->hstaCtrlAddr_s))->precord->name, unit_ps->hstaVal));
	if (dbPutField(&(unit_ps->hstaCtrlAddr_s), DBR_USHORT, (void *) &(unit_ps->hstaVal),nreqs) ){
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, ("mgntAsyncHSTAHdlr : dbPutField error\n"));	  
	  return (MGNT_DBERR);
	}
  } 

  return status;
}
/*=============================================================================

  Name: mgntAsyncCTRLHdlr ()

  Abs:  Decodes data from MGNT_CTRL_NOTIFY message. Handles transition of CTRL
        to IDLE and to non-IDLE


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  CTRL != IDLE sets unit's busy flag (scp func units have add'l flag, scp=T, set in mgntHdlr)
        CTRL == IDLE && active scp func
          reads STATUS, updates unit and summary status,
          read associated PVs, update slc, set update flag if changes decs monitor counter, 
          clears busy flag, checks monitor counter=0 to see if function done
          if done, does a MGNT CHCK to update everything, then sets event flag to release mgntHdlr
		  Note: dblput any change to HSTA, BDES, BACT, BMON, STAT, IDES, IACT, IMON, IRIPL,
		  IACT_PREVOK, TRIM_TOD, TRIM_ITRYCNT,STDZ_TOD, STDZ_LOSTTOD, STDZ_IMAX, CALB_TOD
		  (any ST3 secn) from EPICS to SLC, whether or not a scp user request is being processed.		  
        CTRL == IDLE && fbck unit 
		  ABOVE, except for reading status, decs monitor counter, check monitor counter
        CTRL == IDLE && CA user
		  same as IDLE && fbck unit

  Side: for device under scp func control, check status for timeout.
        if timed out by epics records, call approp routine.

  Data: 1st 4 bytes is pointer to unit in mgntPrim_s.units_ps
        2nd 2 bytes is unsigned short value of the CTRL
 


  Ret: iss vmsstat_t vms style status

===============================================================================*/

LOCAL vmsstat_t mgntAsyncCTRLHdlr (msgmail_ts * rMsg_ps) 
{
  vmsstat_t status = MGNT_OKOK;
  mgntUnitStruct_ts * unit_ps;
  unsigned short ctrlVal;
  long  nRequest=1;      /* these are for fetching Epics PV's */
  long  options=0;
  unsigned short pvshort = 0;
  epicsBoolean scp = epicsFalse;
  
  memcpy(&unit_ps, &rMsg_ps->reqdata[0], 4);
  memcpy(&ctrlVal, &(rMsg_ps->reqdata[2]), 2);  
  if (!unit_ps) {
     DEBUGPRINT(DP_ERROR, mgntAsyncFlag, ("mgntAsyncCTRLHdlr: Error accessing CTRL Callback data \n"));
	 return (status = MGNT_DBERR);
  }
	
  if (ctrlVal == CTRL_IDLE ) {
	unit_ps->busy = epicsFalse;
	DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
			   ("mgntAsyncCTRLHdlr: activeFunc=%u; scp=%u for %s \n",
				mgnt_ps->func_s.activeFunc,unit_ps->scp, (&(unit_ps->ctrlAddr_s))->precord->name));

	/*if ( mgnt_ps->func_s.activeFunc && unit_ps->scp ) {
	if this device is under scp control...........      */
	if (mgnt_ps->func_s.busyCounterMutex_ps)		 
		epicsMutexMustLock(mgnt_ps->func_s.busyCounterMutex_ps);
	scp = unit_ps->scp;
	if (mgnt_ps->func_s.busyCounterMutex_ps)
	  epicsMutexUnlock(mgnt_ps->func_s.busyCounterMutex_ps);
	
	if ( scp ) {
   	  /* read SNL STATUS for this unit */
	  if (unit_ps->ctrlStateAddr_s.precord) {
		if ( (dbGetField(&(unit_ps->ctrlStateAddr_s), DBR_USHORT, 
						 &(pvshort), &options, &nRequest, NULL))) {
		  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
					 ("mgntAsyncCTRLHdlr: Error for dbGetField  \n"));
		  unit_ps->iss = MGNT_DBERR;
		}else { /* successful STATUS PV read */
		  /* translate CTRL STATE to unit_ps->iss */
		  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
					 ("mgntAsyncCTRLHdlr: Active Function going to IDLE for %s; ctrl = %u\n", 
					  (&(unit_ps->ctrlStateAddr_s))->precord->name, pvshort ));
		  switch (pvshort) {
		  case STATUS_OK: case STATUS_DONE:
			unit_ps->iss = MGNT_OKOK;
			break;
		  case STATUS_UNAVAILABLE: case STATUS_DISABLED:
			/*unit_ps->iss = MGNT_OPERR;
			DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
					 ("mgntAsyncCTRLHdlr: ctrlstate=STATUS_UNAVAILABLE; iss=MGNT_OPERR, for %s\n",
					 (&(unit_ps->ctrlStateAddr_s))->precord->name ));*/
			unit_ps->iss = MGNT_OKOK;
			break;
			/*  if status indicates timeout or abort */
		  case STATUS_RAMP_TIMEOUT: case STATUS_TIMEOUT:
			unit_ps->iss = MGNT_TIMEOUT;
			DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
					 ("mgntAsyncCTRLHdlr: ctrlstate=STATUS<_RAMP>_TIMEOUT; iss=MGNT_TIMEOUT, for %s\n",
					  (&(unit_ps->ctrlStateAddr_s))->precord->name ));
			break;
		  case STATUS_ABORT:
			unit_ps->iss = MGNT_ABORTED;
			DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
					 ("mgntAsyncCTRLHdlr: ctrlstate=STATUS_ABORT; iss=MGNT_ABORTED, for %s\n",
					  (&(unit_ps->ctrlStateAddr_s))->precord->name ));
			break;
		  case STATUS_ERROR:
			unit_ps->iss = MGNT_OPERR;
			DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
					 ("mgntAsyncCTRLHdlr: ctrlstate=STATUS_ERROR; iss=MGNT_OPERR, for %s\n",
					  (&(unit_ps->ctrlStateAddr_s))->precord->name ));
			break;
		  default:
			unit_ps->iss = MGNT_OPERR;
			DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
					 ("mgntAsyncCTRLHdlr: ctrlstate=other; iss=MGNT_OPERR, for %s\n",
					  (&(unit_ps->ctrlStateAddr_s))->precord->name ));
			break;
		  } /* switch */	 
		  /* else */
		  /* update mgnt_ps->func_s.summaryStatus */

		  /* then do equiv of MAGTRCK = MAX(MAGTRCK, iss); */
			 /* debug */
		  /*unit_ps->iss = MGNT_OKOK;*/
		  mgnt_ps->func_s.summaryStatus =  max(unit_ps->iss, mgnt_ps->func_s.summaryStatus);;   
		  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
					 ("mgntAsyncCTRLHdlr: summaryStatus = 0x%lx for %s\n", 
					  mgnt_ps->func_s.summaryStatus,(&(unit_ps->ctrlAddr_s))->precord->name ));		  
		  /*if (SUCCESS(unit_ps->iss)) {
			DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
					   ("mgntAsyncCTRLHdlr: Calling mgntAsyncCheckUnit  \n"));
			status = mgntAsyncCheckUnit (unit_ps, epicsBdesHstaMaster_e);
			} */
		  status = mgntAsyncCheckUnit (unit_ps, epicsBdesHstaMaster_e);
			  
		}
	  } /* if a scp function in progress & this unit involved */
	  /* see if scp function is complete */
	  /* update counters */
	  /* put mutex around the increment */
	  if (mgnt_ps->func_s.busyCounterMutex_ps)		 
		epicsMutexMustLock(mgnt_ps->func_s.busyCounterMutex_ps);
	  mgnt_ps->func_s.busyCounter--;

	  /* check if scp requested function is done */
	  if (mgnt_ps->func_s.busyCounter<= 0) { 
		if (mgnt_ps->func_s.busyCounterMutex_ps != NULL) 
		  epicsMutexUnlock(mgnt_ps->func_s.busyCounterMutex_ps);
		status = mgntAsyncFuncDone();
	  } else if (mgnt_ps->func_s.busyCounterMutex_ps != NULL)
		epicsMutexUnlock(mgnt_ps->func_s.busyCounterMutex_ps);

	} /* unit part of scp function */
	else { /* feedback or CA (/EDM) unit  */ 
	  /* update ST 3 secns for this unit */
	  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
					   ("mgntAsyncCTRLHdlr: No active scp function; ctrlVal=%u\n", ctrlVal));
	  status = mgntAsyncCheckUnit (unit_ps, epicsBdesHstaMaster_e);
	}
  }else /* transition to non-idle state */{ 
	/* is now busy */
	unit_ps->busy = epicsTrue;
	DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
			   ("mgntAsyncCTRLHdlr: CTRL transition for %s to non-idle; = %u!\n", 
				 (&(unit_ps->ctrlAddr_s))->precord->name,ctrlVal));
  }
	   
  return status;
}
/*=============================================================================

  Name: mgntAsyncCheckUnitSigChange ()

  Abs:  For given unit, compares epics BDES,BMON values to local slc db values 
        and looks for signficant changes, using tolerances. Also compares STAT
		for any change. Outputs a flag as to whether or not there is a significant
		change.


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		unsigned short      func        read        active function
		mgntUnitStruct_ts  *unit_ps     read/write  ptr to unit completing function

  Rem:  

  side: 
  Data: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/

LOCAL epicsBoolean mgntAsyncCheckUnitSigChange (mgntUnitStruct_ts * const unit_ps)
{
  /* get computed BACT sig change delta for use in calculating BMON's */
  float   bactSigChgTol = 0;
  epicsBoolean sigChange = epicsFalse;
  long  nRequest=1;      /* these are for fetching Epics PV's */
  long  options=0;
  unsigned short pvshort = 0;
  unsigned char * db_ps;

  /* STAT - see if any change */

  /* dbGetField STAT PV value */
  if (unit_ps->statAddr_s.precord) {
	if ( (dbGetField(&(unit_ps->statAddr_s), DBR_USHORT, 
			   &(pvshort), &options, &nRequest, NULL))) {
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
		 ("mgntAsyncCheckUnitSigChange: Error for STAT dbGetField  \n")); 
	}else { /* successful PV read */
	  /* read slc value */
	  DBCLEAR(unit_ps->statDlist_ps);
	  if (SUCCESS (dblget((const dbheader_ts * const)unit_ps->statPlist_ps, 
						  (dbheader_ts **const) &(unit_ps->statDlist_ps)) )) {
		/* compare slc value with PV */
		db_ps = (unsigned char *) (unit_ps->statDlist_ps);
		db_ps += sizeof(dbheader_ts);		
		if (*(unsigned short *)db_ps != pvshort ) {
		  sigChange= epicsTrue;
		  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
					 ("mgntAsyncCheckUnitSigChange:  updating STAT  \n")); 
		}
	  }
	}
  }
  if (sigChange) return sigChange;

  /* BACT - Actual B field - Calculated from IACT */
  sigChange = mgntAsyncBactSigChange(unit_ps, &bactSigChgTol);
  if (sigChange) return sigChange;

  /* BMON - B field calculated from IMON */
  sigChange = mgntAsyncBmonSigChange (unit_ps, bactSigChgTol );
  return sigChange;
}
/*=============================================================================

  Name: mgntBactSigChange ()

  Abs:  Reads float PV value and corresponding SLC float value.
        Compares the two values and dblputs to loca SLC db if different, 
		with the option of using a tolerance window around the difference.
		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		DBADDR  *           dbaddr_ps    read       ptr to unit's epics address 
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit


  Rem:  

  side: Sets the mgnt_ps->doUpdate flag to true if dblputting to slc db.
        This way the 1 second async message will cause a call to dbupdate to
		update the Alpha after waiting for more units to "report in".

  Data: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL epicsBoolean mgntAsyncBactSigChange(mgntUnitStruct_ts * unit_ps,
										  float * bactSigChgTol) 
									  
{
  long           nRequest=1;      /* these are for fetching Epics PV's */
  long           options=0;
  float          bactVal = 0;
  unsigned char  *db_ps;
  float          sigtol;
  float          sigchangetol = 1.0;
  float          actualTolAbs;
  float          *tol_p;
  epicsBoolean   sigChange = epicsFalse;

  /* dbGetField PV value */
  if (unit_ps->bactAddr_s.precord) {
	if ( (dbGetField(&(unit_ps->bactAddr_s), DBR_FLOAT, 
			   &(bactVal), &options, &nRequest, NULL))) {
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
		 ("mgntAsyncBactSigChange : Error for dbGetField  \n")); 
	}else { /* successful PV read */
	  /* read slc value */
	  DBCLEAR (unit_ps->bactDlist_ps);
	  if (SUCCESS(dblget( (const dbheader_ts * const)unit_ps->bactPlist_ps,
                   (dbheader_ts **const)& (unit_ps->bactDlist_ps) ) )) {
		/* compare slc value with PV */
		db_ps = (unsigned char *) (unit_ps->bactDlist_ps);
		db_ps += sizeof(dbheader_ts);		

		tol_p = DBPTR(mgnt_ps->dlist_s.tols_ps,(unit_ps->slcUnitsI*4));
		DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("mgntAsyncBactSigChange: BACT 1st tol=%f; 2nd tol=%f\n", *tol_p, *tol_p+1));
		sigtol = fmax ((*tol_p), (*(tol_p+1)) * fabs(unit_ps->bdesVal) ) * sigchangetol;
		*bactSigChgTol = sigtol;
		actualTolAbs =  fabs (*(float *)db_ps - bactVal);
		DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("mgntAsyncBactSigChange: BACT sigtol=%f; actual abs tol = %f\n", sigtol, actualTolAbs)); 
		if ( fabs (*(float *)db_ps - bactVal)> sigtol)  {
		  /* report */
		  sigChange = epicsTrue;
		}		  
		
	  }
	}
  }
  return sigChange;
}
/*=============================================================================

  Name: mgntAsyncBmonSigChange ()

  Abs:  Reads float PV value and corresponding SLC float value.
        Compares the two values and dblputs to loca SLC db if different, 
		with the option of using a tolerance window around the difference.
		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		DBADDR  *           dbaddr_ps    read       ptr to unit's epics address 
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit


  Rem:  

  side: Sets the mgnt_ps->doUpdate flag to true if dblputting to slc db.
        This way the 1 second async message will cause a call to dbupdate to
		update the Alpha after waiting for more units to "report in".

  Data: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL epicsBoolean mgntAsyncBmonSigChange(mgntUnitStruct_ts * unit_ps,
										  float bactSigChgTol)  
									  
	                                 
{
  epicsBoolean    sigChange= epicsFalse;
  long            nRequest=1;      /* these are for fetching Epics PV's */
  long            options=0;
  float           bmonVal = 0;
  unsigned char   *db_ps;
  float           sigtol;
  float           *tol_p;

  /* dbGetField PV value */
  if (unit_ps->bmonAddr_s.precord) {
	if ( (dbGetField(&(unit_ps->bmonAddr_s), DBR_FLOAT, 
			   &(bmonVal), &options, &nRequest, NULL))) {
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
		 ("mgntAsyncBmonPVToSlc : Error for dbGetField  \n")); 
	}else { /* successful PV read */
	  /* read slc value */
	  DBCLEAR (unit_ps->bmonDlist_ps);
	  if (SUCCESS(dblget( (const dbheader_ts * const)unit_ps->bmonPlist_ps,
                   (dbheader_ts **const)& (unit_ps->bmonDlist_ps) ) )) {
		/* compare slc value with PV */
		db_ps = (unsigned char *) (unit_ps->bmonDlist_ps);
		db_ps += sizeof(dbheader_ts);		
 
		tol_p = DBPTR(mgnt_ps->dlist_s.tolf_ps,(unit_ps->slcUnitsI*4));
		DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("mgntAsyncBmonPVToSlc: BMON TOLF = %f\n", *tol_p ));
		sigtol = fabs( *tol_p *  bactSigChgTol);
		
		DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
				   ("mgntAsyncBmonPVToSlc: BMON sigtol=%f \n", sigtol)); 
		if ( fabs (*(float *)db_ps - bmonVal)> sigtol)  {
		  /* report */
		  sigChange = epicsTrue;
		}		  
	  }
	}
  }
  return sigChange;
}
/*=============================================================================

  Name: mgntAsyncCheckUnit ()

  Abs:  Compares local slc db and epics values for all possible magnet secondaries 
        and writes latest value (If the secondary doesn't exist - no error)  
        For given unit, compares ST3 epics values to local slc db values and dblputs 
        any changes from EPICS to SLC.
		If masterType_e is slcBdesHstaMaster_e, then slc db values are dbPutFieldd
		to epics for BDES and HSTA if they differ.



  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		unsigned short      func        read        active function
		mgntUnitStruct_ts  *unit_ps     read/write  ptr to unit completing function

  Rem:  

  side: Each mgntAsyncPVToSlc*() utility called will set mgnt_ps->doUpdate upon
        detecting change and issue a dblput().

  Data: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/

LOCAL vmsstat_t mgntAsyncCheckUnit (mgntUnitStruct_ts * const unit_ps,
									const mgntMaster_te masterType_e) 
{
  vmsstat_t status = MGNT_OKOK;

  /* All units: */

  if (masterType_e == epicsBdesHstaMaster_e) {
	/* update SLC BDES and HSTA from EPICS */
	/*  occurs for CTRL=IDLE callbacks */
	/* BDES - Desired B field */
	mgntAsyncPVToSlcFloat( &(unit_ps->bdesAddr_s),
						   (const dbheader_ts *) unit_ps->bdesPlist_ps, 
						   (dbheader_ts **const) &(unit_ps->bdesDlist_ps));
	
	/* HSTA */
	DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("mgntAsyncCheckUnit: Check HSTA \n"));
	mgntAsyncPVToSlcUShort( &(unit_ps->hstaAddr_s),
							(const dbheader_ts *) unit_ps->hstaPlist_ps, 
							(dbheader_ts **const) &(unit_ps->hstaDlist_ps),
							unit_ps);
  } else { /* masterType_e == slcBdesHstaMaster_e */
	/* update ST 2 EPICS BDES and HSTA CTRL from SLC */
	/* occurs on 10 sec async checks */
	mgntAsyncSlcToPVBdes(unit_ps);
	mgntAsyncSlcToPVHstaCtrl(unit_ps);

  }
  /* STAT */
  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("mgntAsyncCheckUnit: Check STAT  \n"));
  mgntAsyncPVToSlcUShort( &(unit_ps->statAddr_s),
						  (const dbheader_ts *) unit_ps->statPlist_ps, 
						  (dbheader_ts **const) &(unit_ps->statDlist_ps),
												  NULL);  
  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("mgntAsyncCheckUnit: Check ITRY for %s  \n",
									  (&(unit_ps->itryAddr_s))->precord->name));
  /* ITRY - # of attempts required for trim */
  mgntAsyncPVToSlcUShort( &(unit_ps->itryAddr_s),
						  (const dbheader_ts *) unit_ps->itryPlist_ps, 
						  (dbheader_ts **const) &(unit_ps->itryDlist_ps),
												  NULL);  

  /* BACT - Actual B field - Calculated from IACT - at last check */
  mgntAsyncPVToSlcFloat( &(unit_ps->bactAddr_s),
						 (const dbheader_ts *) unit_ps->bactPlist_ps, 
						 ( dbheader_ts **const) &(unit_ps->bactDlist_ps));
  /* BMON - B field calculated from IMON */
  mgntAsyncPVToSlcFloat( &(unit_ps->bmonAddr_s),
						 (const dbheader_ts *) unit_ps->bmonPlist_ps, 
						 ( dbheader_ts **const) &(unit_ps->bmonDlist_ps));
  /* IDES - Desired Current calculated from BDES */
  mgntAsyncPVToSlcFloat( &(unit_ps->idesAddr_s),
						 (const dbheader_ts *) unit_ps->idesPlist_ps, 
						 ( dbheader_ts **const) &(unit_ps->idesDlist_ps));
  /* IACT - Current Readback - at last check */
  mgntAsyncPVToSlcFloat( &(unit_ps->iactAddr_s),
						(const dbheader_ts *) unit_ps->iactPlist_ps, 
						( dbheader_ts **const) &(unit_ps->iactDlist_ps));
  /* RIPL - ripple current readback */
  mgntAsyncPVToSlcFloat( &(unit_ps->riplAddr_s),
						(const dbheader_ts *) unit_ps->riplPlist_ps, 
						( dbheader_ts **const) &(unit_ps->riplDlist_ps));
  /* IPRV - last good current readback in tolerance */
  mgntAsyncPVToSlcFloat( &(unit_ps->iprvAddr_s),
						(const dbheader_ts *) unit_ps->iprvPlist_ps, 
						( dbheader_ts **const) &(unit_ps->iprvDlist_ps));
  /* IMON - second (redundant) current readback */
  mgntAsyncPVToSlcFloat( &(unit_ps->imonAddr_s),
						(const dbheader_ts *) unit_ps->imonPlist_ps, 
						( dbheader_ts **const) &(unit_ps->imonDlist_ps));
  /* IMAX - second (redundant) current readback */
  mgntAsyncPVToSlcFloat( &(unit_ps->imaxAddr_s),
						(const dbheader_ts *) unit_ps->imaxPlist_ps, 
						( dbheader_ts **const) &(unit_ps->imaxDlist_ps));
  
  /* TTIM - Last trim time */
  mgntAsyncPVToSlcTime ( &(unit_ps->ttimAddr_s),
						 (const dbheader_ts *) unit_ps->ttimPlist_ps, 
						 ( dbheader_ts **const) &(unit_ps->ttimDlist_ps));
  /* CTIM - Last successful calibration time */
  mgntAsyncPVToSlcTime( &(unit_ps->ctimAddr_s),
						 (const dbheader_ts *) unit_ps->ctimPlist_ps, 
						 ( dbheader_ts **const) &(unit_ps->ctimDlist_ps));
  /* KTIM - Last successful standardization time */
  mgntAsyncPVToSlcTime( &(unit_ps->ktimAddr_s),
						 (const dbheader_ts *) unit_ps->ktimPlist_ps, 
						 ( dbheader_ts **const) &(unit_ps->ktimDlist_ps));
  /* ZTIM- Last time standardization lost */
  mgntAsyncPVToSlcTime( &(unit_ps->ztimAddr_s),
						 (const dbheader_ts *) unit_ps->ztimPlist_ps, 
						 ( dbheader_ts **const) &(unit_ps->ztimDlist_ps)); 
  return status;

}

LOCAL float fmax (float x, float y) {
  return (x > y) ? x : y;
}
/*=============================================================================

  Name: mgntAsyncPVToSlcFloat ()

  Abs:  Reads float PV value and corresponding SLC float value.
        Compares the two values and dblputs to loca SLC db if different, 
		with the option of using a tolerance window around the difference.
		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		DBADDR  *           dbaddr_ps    read       ptr to unit's epics address 
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit


  Rem:  

  side: Sets the mgnt_ps->doUpdate flag to true if dblputting to slc db.
        This way the 1 second async message will cause a call to dbupdate to
		update the Alpha after waiting for more units to "report in".

  Data: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
vmsstat_t mgntAsyncPVToSlcFloat ( DBADDR *dbaddr_ps,
										const dbheader_ts * const dblist_ps,
										dbheader_ts **const dbdata_pps)

									  
{
  vmsstat_t     status = MGNT_OKOK;
  long          nRequest=1;      /* these are for fetching Epics PV's */
  long          options=0;
  float         pvfloat = 0;
  unsigned char * db_ps;


  /* dbGetField PV value */
  if (dbaddr_ps->precord) {
	if ( (dbGetField(dbaddr_ps, DBR_FLOAT, 
			   &(pvfloat), &options, &nRequest, NULL))) {
	  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
		 ("mgntAsyncPVToSlcFloat: Error for dbGetField  \n")); 
	}else { /* successful PV read */
	  /* read slc value */
	  ((dbheader_ts *)(*dbdata_pps))->cur = 0;
	  if (SUCCESS (dblget( dblist_ps, dbdata_pps) )) {
		/* compare slc value with PV */
		db_ps = (unsigned char *) (*dbdata_pps);
		db_ps += sizeof(dbheader_ts);		
		if (*(float *)db_ps != pvfloat ) {
		  memcpy (db_ps, &(pvfloat), sizeof(float));
		  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
					 ("mgntAsyncPVToSlcFloat: dblput %f\n", pvfloat));
		  dblput (dblist_ps, *dbdata_pps, mgntJob); 
		  mgnt_ps->doUpdate = epicsTrue;
		}
	  }
	}
  }
  return status;
}

/*=============================================================================

  Name: mgntAsyncPVToSlcShort ()

  Abs:  Reads float PV value and corresponding SLC float value.
        Compares the two values and dblputs to loca SLC db if different, 
		with the option of using a tolerance window around the difference.
		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		DBADDR  *           dbaddr_ps    read       ptr to unit's epics address 
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit


  Rem:  

  side: Sets the mgnt_ps->doUpdate flag to true if dblputting to slc db.
        This way the 1 second async message will cause a call to dbupdate to
		update the Alpha after waiting for more units to "report in".

  Data: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
vmsstat_t mgntAsyncPVToSlcUShort ( DBADDR *dbaddr_ps,
								   const dbheader_ts * const dblist_ps,
								   dbheader_ts **const dbdata_pps,
								   mgntUnitStruct_ts * const unit_ps)		
						   
	 
	                                   /* prob. want a tol here; if =0, don't use */
{
  vmsstat_t status = MGNT_OKOK;
  long  nRequest=1;      /* these are for fetching Epics PV's */
  long  options=0;
  unsigned short pvshort = 0;
  unsigned char * db_ps;

  /* dbGetField PV value */
  if (dbaddr_ps->precord) {
	if ( (dbGetField(dbaddr_ps, DBR_USHORT, 
			   &(pvshort), &options, &nRequest, NULL))) {
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
		 ("mgntAsyncPVToShort: Error for dbGetField  \n")); 
	}else { /* successful PV read */
	  /* read slc value */
	  ((dbheader_ts *)(*dbdata_pps))->cur = 0;
	  if (SUCCESS (dblget( dblist_ps, dbdata_pps) )) {
		/* compare slc value with PV */
		db_ps = (unsigned char *) (*dbdata_pps);
		db_ps += sizeof(dbheader_ts);
		DEBUGPRINT(DP_DEBUG, mgntAsyncFlag,("mgntAsyncPVToShort: epics =%u; slc=%u\n", 
										   pvshort, *(unsigned short *)db_ps )); 		
		if (*(unsigned short *)db_ps != pvshort ) {
		  memcpy (db_ps, &(pvshort), sizeof(unsigned short));
		  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
					 ("mgntAsyncPVToShort: dblput %u \n", pvshort)); 
		  dblput (dblist_ps, *dbdata_pps, mgntJob); 
		  mgnt_ps->doUpdate = epicsTrue;
		  /* in addition, update the slc hsta pv */
		  if (unit_ps != NULL) {
			DEBUGPRINT(DP_DEBUG, mgntAsyncFlag,("mgntAsyncPVToShort: updating HSTACTRL, epics =%x; slc=%x\n", 
										   pvshort, *(unsigned short *)db_ps )); 
			if (dbPutField(&(unit_ps->hstaCtrlAddr_s), DBR_USHORT, (void *)db_ps,nRequest) )
			  return (MGNT_DBERR);			
		  }
		}
	  }else DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
		 ("mgntAsyncPVToShort: Error for dblget\n")); 
	}
  }
  return status;
}
/*=============================================================================

  Name: mgntAsyncPVToSlcTime ()

  Abs:  Reads float PV value and corresponding SLC float value.
        Compares the two values and dblputs to loca SLC db if different, 
		with the option of using a tolerance window around the difference.
		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		DBADDR  *           dbaddr_ps    read       ptr to unit's epics address 
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit


  Rem:  

  side: Sets the mgnt_ps->doUpdate flag to true if dblputting to slc db.
        This way the 1 second async message will cause a call to dbupdate to
		update the Alpha after waiting for more units to "report in".

  Data: 

  Ret:  iss vmsstat_t vms style status

  Side: To read timestamp record, create (from $EPICS_BASE/include/dbAccessDefs.h)   
 
 * struct {
 *	DBRstatus
 *	DBRtime
 *	epicsFloat32 value[10]
 * } buffer;
 *
 * IMPORTANT!! The DBRoptions must be given in the order that they
 *             appear in the Data Base Request Options #defines
 *
 * The associated dbGetField call is:
 *
 * long options,number_elements;
 * ...
 * options = DBR_STATUS|DBR_TIME;
 * number_elements = 10;
 * rtnval=dbGetField(paddr,DBR_FLOAT,&buffer,&options,&number_elements, NULL);
 *	
 * When dbGetField returns:
 *	rtnval is error status (0 means success)
 *	options has a bit set for each option that was accepted
 *	number_elements is actual number of elements obtained
 *
 * The individual items can be refered to by the expressions::
 *
 * buffer.status
 * buffer.severity ..
 * buffer.time
 * buffer.value[i]
 *

===============================================================================*/
LOCAL vmsstat_t mgntAsyncPVToSlcTime ( DBADDR *dbaddr_ps,
									   const dbheader_ts * const dblist_ps,
									   dbheader_ts **const dbdata_pps) 
									  
	                                   /* prob. want a tol here; if =0, don't use */
{
 typedef struct {
   DBRstatus;
   DBRtime;
   epicsFloat32 value[10];
 } buffer_ts;

  vmsstat_t        status = MGNT_OKOK;
  long             nRequest=10;    
  long             options;  
  unsigned char    *db_ps;
  vmsTimeStamp_ta  vmsTS ;
  long int         diff10ms;
  epicsTimeStamp   time_s;  
  buffer_ts        buffer;

  options = DBR_STATUS|DBR_TIME;
  /* dbGetTimeStamp of PV link */
  if (dbaddr_ps->precord) {
	if ( (dbGetField(dbaddr_ps, DBR_FLOAT, 
			   &buffer, &options, &nRequest, NULL))) {
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
		 ("mgntAsyncPVToSlcTime: Error for dbGetField  \n")); 

	}else { /* successful PV read */
	  /* read slc value */
	  ((dbheader_ts *)(*dbdata_pps))->cur = 0;
	  if (SUCCESS (dblget(dblist_ps, dbdata_pps) )) {
		time_s = buffer.time;
		timeEpicsToVMS(&time_s, &vmsTS[0]);
		db_ps = (unsigned char *) (*dbdata_pps);
		db_ps += sizeof(dbheader_ts);		
		
		/* use tolerances ? */
		diff10ms = timeDiffVMS10msec(*(vmsTimeStamp_ta *)db_ps,vmsTS);
		DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
				 ("mgntAsyncPVToSlcTime: diff= %ld \n", diff10ms)); 
		if (diff10ms >10 || diff10ms < -10) {
		  memcpy (db_ps, &(vmsTS),sizeof(vmsTimeStamp_ta));
		  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("mgntAsyncPVToTime: dblput\n"));
		  dblput (dblist_ps, *dbdata_pps, mgntJob); 
		  mgnt_ps->doUpdate = epicsTrue;
		}
	  }else DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
					   ("mgntAsyncPVToTime: dblget error for - %s \n", 
						dbaddr_ps->precord->name));
	}
  } /*else DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
					("mgntAsyncPVToTime: precord error for - %s \n", 
					dbaddr_ps->precord->name));*/ 
  return status;
}


/*=============================================================================

  Name: mgntAsyncFuncDone ()

  Abs:  Called when the requested scp function has been determined to be 
		complete(/timed out/aborted). Completeness occurs when the busyCounter counts 
		down to 0, when the timeoutCounter counts down to zero, or upon an abort.
		Updates and checks counter; if done, does a MGNT CHCK
		to update everything, then sets event flag to reelase mgntHdlr


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rMsg_ps     read only   incoming CLR_SCP message    

  Rem:  

  side: 
  Data:

  Ret: None

===============================================================================*/

LOCAL vmsstat_t mgntAsyncFuncDone(void) 
{
  vmsstat_t status = MGNT_OKOK;

  /* scp function is complete - heard from all units  */
  if (mgnt_ps->func_s.busyCounterMutex_ps)		 
	epicsMutexMustLock(mgnt_ps->func_s.busyCounterMutex_ps);
  mgnt_ps->func_s.busyCounter = 0;
  mgnt_ps->func_s.timeoutCounter = 0;
  if (mgnt_ps->func_s.busyCounterMutex_ps)		 
	epicsMutexUnlock(mgnt_ps->func_s.busyCounterMutex_ps);
  /* do a mgntAsyncCheck to update slc db */
  /* for this final check, check all units, no tols, all devices regardless of CTRL status*/
  status = mgntAsyncCheck( epicsFalse,   /* don't use tolerances */
						   epicsFalse,   /* all devices regarless of CTRL status */
						   epicsTrue,    /* do the update now */
						   slcBdesHstaMaster_e,   /* async slcAsyncMeterDbSend, not dbupdate */
						   slcMsg);      /* request type (slcMsg, slcAsync) */
  /* now release mgntHdlr from waiting for completion */
  if (mgnt_ps->func_s.funcDoneEvent_ps != NULL){
	DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
			   ("MGNT_A: signaling funcDoneEvent_ps from mgntAsyncFuncDone() \n"));
	epicsEventSignal(mgnt_ps->func_s.funcDoneEvent_ps);
	
  }
  return status;
}

/*=============================================================================

  Name: mgntAsyncCheckDoProceed

  Abs:  Determine if the magnet function should be dropped on the floor

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        unsigned short    func          read-only    function code from req msg
        epicsTimeStamp *  curTime_ps    read-only    ptr to current time 
        epicsTimeStamp *  prevTime_ps   read-only    ptr last time of a magnet request

  Rem: The purpose of this function is to determine if the specified
       magnet request function should be dropped on the floor. The reason that
       this might be done is that the function may have just been processed and
       and there is no need to process the same function at this time. 

       The type of functions that this occurs is a local request for a adc update (IOC_ASYNC).
       which is received by the mgntAsync thread and a magnet check function, which is the same
       except for this  message is received by the magnet handler from a SCP process.
       Basically, a check is  done to determine if enough time has past since the last adc
       update was performed. Remember that and adc update is done every time a magent 
       function is requested and not just when a MGNT_CHCK or IOC_ASYNC is found in the message
       queueu of the magnet handler or the magnet async threads.

  Side: None

  Ret:  int
            0  - Flag check (adc read) should NOT be aboted
            1  - Flag check (adc read) should be aborted

==============================================================================*/
epicsBoolean mgntAsyncCheckDoProceed (void)
{
  double         secs;
  epicsBoolean   goAhead= epicsTrue;
  epicsTimeStamp  curTime_s;                /* current time stamp             */  
  /* 
  ** Is this a local ASYN function? If yes and an adc
  ** update was just done, then abort this request.
  ** Otherwise, continue.
  */
/* get the curren time */
  epicsTimeGetCurrent(&curTime_s);
  secs = epicsTimeDiffInSeconds(&curTime_s,&(mgnt_ps->lastCheckTime_s));
  if ( secs < MGNT_TIME_BETWEEN_CHECK)
	goAhead = epicsFalse;

  return(goAhead);
}
/*=============================================================================

  Name: mgntAsyncTime ()

  Abs:  For the 10 second function MGNT CHCK, read all the non-FBCK devices, 
        check TOLS/TOLF and do the proper dblputs and dbupdate at the end.


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const epicsBoolean  useTolerances read only whether to use PRIM:UNIT:SECN:TOLS    

  Rem:  

  side: 
  Data:
  "MGNT CHCK" is for 10 second update, 
  "MGNT CHKF" is for 1 minute update, and 
  "MGNT TIME" is for the 1 second actions.  

  For the 10 second function MGNT CHCK, read all the non-FBCK device readbacks.
  When called from user via scp, useTolerances is False; otherwise it is True.
  When using tols, check agains TOLS/TOLF. Do the proper dblputs and dbupdate at the end.  
  For the 1 minute function MGNT CHKF, you'll read all the FBCK devices, 
  and dblput anything that is different (no TOLS/TOLF check) and dbupdate at the end. 
  Devices are under feedback control when the feedback-control bit in their SLC HSTA is set,
   (dblget HSTA).

   **  Note: when called upon scp function completion - in this case, 
  For all prim/units, update SLC STAT and ACT values from records in EPICS db

  Ret: iss vmsstat_t vms style status

===============================================================================*/
 
LOCAL vmsstat_t mgntAsyncTime (void) {
  int                asyncCycFuncIndex;    /* cycling function index */
  vmsstat_t status = MGNT_OKOK;

  /* if no scp request being processed, dbupdate if dblput was done in any 
	  BDES, HSTA, or CTRL callback */
  if (! mgnt_ps->func_s.activeFunc){

	if (mgnt_ps->updateTimeout || mgnt_ps->doUpdate) {
	  DEBUGPRINT(DP_INFO, mgntAsyncFlag, ("mgntAsyncTime: dbupdating !!!!\n"));
	  status = dbupdate(mgntJob);
	  mgnt_ps->updateTimeout = 0;
	  mgnt_ps->doUpdate = epicsFalse;
	}
  } else {
	mgnt_ps->func_s.timeoutCounter--;
	if (mgnt_ps->func_s.timeoutCounter <=0){
	  /* scp request has timed out */
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, ("mngtAsyncTime timed out active function. \n"));
	  /* do a final check, like the request finished successfully; which is like the
		 mgntAsyncCheck() called from 10 sec processing */
	  /* check all units, no tols, all devices regardless of CTRL status*/
	  status = mgntAsyncFuncDone();
	  /* perhaps send ABORT to the device so that the SNL exits tidily */

	}
  }
   /* !!! Now call slcAsyncMeterDbSend and slcasyncSetFcStats */

  /* HERES WHERE I LEFT OFF!!!! is this the right place?*/
  epicsTimeGetCurrent(&(mgnt_ps->lastTime_s)); 

  if ((asyncCycFuncIndex = slcAsyncGetFuncIxByName("TIME")) == -1) {	
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "mgntAsyncTime failed to find its function index\n");
    return(MGNT_DBERR);
  }

  if (slcAsyncMeterDbSend(asyncCycFuncIndex, slcAsync, epicsTrue)!=0) 
	return (MGNT_DBERR);

  if (slcAsyncSetFcnStats(asyncCycFuncIndex, &(mgnt_ps->lastTime_s), slcAsync)!=0)
	return (MGNT_DBERR);


  return (status);
}
/*=============================================================================

  Name: mgntAsyncCheck ()

  Abs:  For the 10 second function MGNT CHCK, read all the non-FBCK devices, 
        check TOLS/TOLF and do the proper dblputs and dbupdate at the end.


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const epicsBoolean  useTolerances read only whether to use PRIM:UNIT:SECN:TOLS    

  Rem:  

  side: Go through all devices first and look for a significant change and then if 
        there is a change, go through all devices again and dblput anything that is 
        different, regardless of significance.
  Data:
  "MGNT CHCK" is for 10 second update, 
  "MGNT CHKF" is for 1 minute update, and 
  "MGNT TIME" is for the 1 second actions.  

  For the 10 second function MGNT CHCK, read all the non-FBCK device readbacks.
  When called from user via scp, useTolerances is False; otherwise it is True.
  When using tols, check agains TOLS/TOLF. Do the proper dblputs and dbupdate at the end.  
  For the 1 minute function MGNT CHKF, you'll read all the FBCK devices, 
  and dblput anything that is different (no TOLS/TOLF check) and dbupdate at the end. 
  Devices are under feedback control when the feedback-control bit in their SLC HSTA is set,
   (dblget HSTA).

   **  Note: when called upon scp function completion - in this case, 
  For all prim/units, update SLC STAT and ACT values from records in EPICS db

  Ret: iss vmsstat_t vms style status

===============================================================================*/
 
LOCAL vmsstat_t mgntAsyncCheck(const epicsBoolean useTolerances,
							   const epicsBoolean idleUnitsOnly,
							   const epicsBoolean doDbUpdate,
							   const mgntMaster_te mastertype_e,
							   const slcReqType_te requestType)
	 
{
  vmsstat_t          status = MGNT_OKOK;
  unsigned short     i_prim;               /* primary index        */ 
  unsigned short     i_unit;               /* unit index           */
  mgntPrimStruct_ts  *prim_ps=NULL;        /* pointer to primary unit info */
  mgntUnitStruct_ts  *unit_ps=NULL;
  long               nreqs=1;      /* these are for fetching Epics PV's */
  long               options=0;
  int 		         asyncCycFuncIndex = 0;
  epicsBoolean       sigChange = epicsFalse;

  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
("mgntAsyncCheck: useTol=%d, idleUnitsOnly=%d, doDbUpdate=%d, mastertype=%d (0=slc; 1=epics), requestType=%d ( 0=slcAsync, 1=slcMsg); active func?=%d \n", useTolerances, idleUnitsOnly, doDbUpdate, mastertype_e, requestType, mgnt_ps->func_s.activeFunc)); 
  /* only do the check function if active scp function not ongoing */
  if (mgnt_ps->func_s.activeFunc) return status;
  DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, ("mngtAsyncCheck() reached; no active scp function \n"));
  /* check to see if sufficient time has elapsed */
  
  epicsTimeGetCurrent(&(mgnt_ps->lastCheckTime_s)); 
  /*
  ** Get the current adc readings from EPICS for non-feedback devices 
  ** Only update the SLC db secondaries where there was
  ** significant change. This change or tolerance is setup
  */
  if (requestType == slcAsync) {
	/* first pass - see if ANY unit has sig change for BACT, BMON, or any change for STAT  */
	/*  if so - break from first pass and go onto second pass */
	for (i_prim =0; (i_prim<MGNT_NUM_PRIMS) && (!sigChange); i_prim++ ) {
	  prim_ps = &(mgntPrim_s[i_prim]);
	  unit_ps = DBPTR_FIRST(prim_ps->units_ps);
	  for (i_unit=0; (i_unit<prim_ps->nunits) && (!sigChange) && unit_ps; i_unit++, unit_ps++) {

		/* read CTRL to see if = IDLE; if not, skip this unit */
		if (unit_ps->ctrlAddr_s.precord) {
		  if (dbGetField(&(unit_ps->ctrlAddr_s),  DBR_USHORT, 
						 &(unit_ps->ctrlVal), &options, &nreqs, NULL))
			continue;
		  /* successful PV read */
		  /* note that feedback devices have CTRL=PRTB and won't be included */
		  if (unit_ps->ctrlVal==CTRL_IDLE ) 
			sigChange = mgntAsyncCheckUnitSigChange(unit_ps );
		} /* if unit has a CTRL record */
	  } /* for all units */
	}/* for all prims */
	/* if no sig change for ANY prim/unit - return now */
	if (!sigChange) return status;
  } /* 10 sec periodic async check */

  /* second pass - there was sig change for ANY unit */
  /* so now go through ALL units and upate AnY change, regardless of sig tol */
  for (i_prim =0; i_prim<MGNT_NUM_PRIMS; i_prim++ ) {
	prim_ps = &(mgntPrim_s[i_prim]);
	unit_ps = DBPTR_FIRST(prim_ps->units_ps);
	for (i_unit=0; (i_unit<prim_ps->nunits) && SUCCESS(status) && unit_ps; i_unit++, unit_ps++) { 
	  if (idleUnitsOnly) { /* only scp request completion processing wants ALL units regardless */
		/* read CTRL to see if = IDLE; if not, skip this unit */
		if (unit_ps->ctrlAddr_s.precord) {
		  if (!dbGetField(&(unit_ps->ctrlAddr_s),  DBR_USHORT, 
						 &(unit_ps->ctrlVal), &options, &nreqs, NULL)){
		   
			/* successful PV read */
			/* note that feedback devices have CTRL=PRTB and won't be included */
			DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
			 ("mgntAsyncCheck: Calling mgntAsyncCheckUnit ctrlval=%d (if 0) for prim=%d, unit=%d,\n",
			   unit_ps->ctrlVal, i_prim, i_unit));
			if (unit_ps->ctrlVal==CTRL_IDLE ){
			  status = mgntAsyncCheckUnit(unit_ps, mastertype_e );
			}
		  }	
		} /* if unit has a CTRL record */
	  } else  {/* if idle units only */ 
		status = mgntAsyncCheckUnit(unit_ps, mastertype_e );
		DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
				   ("mgntAsyncCheck: Calling mgntAsyncCheckUnit for prim=%d, unit=%d,\n",i_prim, i_unit));
	  }
	} /* for all units */
  }/* for all prims */
  

  if ((asyncCycFuncIndex = slcAsyncGetFuncIxByName("CHCK")) == -1) {	
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "mgntAsyncCheck failed to find its function index\n");
    return(MGNT_DBERR);
  }

  if (slcAsyncMeterDbSend(asyncCycFuncIndex, requestType, epicsTrue)!=0)
	return (MGNT_DBERR);

  if (requestType==slcMsg) {
	if (slcAsyncSetFcnStats(asyncCycFuncIndex, &(mgnt_ps->func_s.reqStartTime_s), requestType)!=0)
	  return (MGNT_DBERR);
  }
  else{
	if (slcAsyncSetFcnStats(asyncCycFuncIndex, &(mgnt_ps->lastCheckTime_s), requestType)!=0)
	  return (MGNT_DBERR);
  }
  /*  }*/
  
  return(status);
}
/*=============================================================================

  Name: mgntAsyncCheckFeedback ()

  Abs:  For the 1 minute function MGNT CHKF, read all theFBCK devices, 
        and do the proper dblputs and dbupdate at the end. Only execute if
		no scp request is being processed


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const epicsBoolean  useTolerances read only whether to use PRIM:UNIT:SECN:TOLS    

  Rem:  

  side: 
  Data:
  "MGNT CHCK" is for 10 second update, 
  "MGNT CHKF" is for 1 minute update, and 
  "MGNT TIME" is for the 1 second actions.  

  For the 10 second function MGNT CHCK, read all the non-FBCK device readbacks.
  When called from user via scp, useTolerances is False; otherwise it is True.
  When using tols, check agains TOLS/TOLF. Do the proper dblputs and dbupdate at the end.  
  For the 1 minute function MGNT CHKF, read all the FBCK devices, 
  and dblput anything that is different (no TOLS/TOLF check) and dbupdate at the end. 
  Devices are under feedback control when the feedback-control bit in their SLC HSTA is set,
   (dblget HSTA).

   **  Note: when called upon scp function completion - in this case, 
  For all prim/units, update SLC STAT and ACT values from records in EPICS db

  Ret: iss vmsstat_t vms style status

===============================================================================*/
 
LOCAL vmsstat_t mgntAsyncCheckFeedback( const slcReqType_te requestType)
	 
{
  vmsstat_t          status = MGNT_OKOK;
  unsigned short     i_prim;               /* primary index        */ 
  unsigned short     i_unit;               /* unit index           */
  mgntPrimStruct_ts  *prim_ps=NULL;        /* pointer to primary unit info */
  mgntUnitStruct_ts  *unit_ps=NULL;
  long               nreqs=1;      /* these are for fetching Epics PV's */
  long               options=0;
  int                asyncCycFuncIndex;    /* cycling function index */


   /* only do the check function if active scp function not ongoing */
  DEBUGPRINT(DP_INFO, mgntAsyncFlag, ("mngtAsyncCheckFeedback reached; active scp function?=%d \n",
										   mgnt_ps->func_s.activeFunc ));
   if ( ! mgnt_ps->func_s.activeFunc){
	 /* check to see if sufficient time has elapsed */
	 
	 epicsTimeGetCurrent(&(mgnt_ps->lastCheckFeedbackTime_s)); 
	 /*
	 ** Get the current adc readings from EPICS for all feedback devices 
	 ** Update values that change, ignoring significant tolerances
	 */
	 for (i_prim =0; i_prim<MGNT_NUM_PRIMS; i_prim++ ) {
	   prim_ps = &(mgntPrim_s[i_prim]);
	   unit_ps = DBPTR_FIRST(prim_ps->units_ps);

	   for (i_unit=0; (i_unit<prim_ps->nunits) && unit_ps; i_unit++, unit_ps++) {
		 /* read HSTA to see if under feedback control  */
		 if (unit_ps->hstaAddr_s.precord) {

		   if (!(dbGetField(&(unit_ps->hstaAddr_s),  DBR_USHORT, 
						  &(unit_ps->hstaVal), &options, &nreqs, NULL))){
			 /* successful PV read */
		   
			 if (unit_ps->hstaVal & HSTA_FBKCONTROL ){
			   DEBUGPRINT(DP_INFO, mgntAsyncFlag, ("mngtAsyncCheckFeedback: %s under FBCK control\n",
												 (&(unit_ps->hstaAddr_s))->precord->name));			 
			   mgntAsyncCheckUnit(unit_ps, epicsBdesHstaMaster_e );		 
			 } /* if unit under feedback control */
		   }
		 }
	   } /* for all units */
	 }/* for all prims */
   } /* no active scp request being processed */

   /* !!! Now call slcAsyncMeterDbSend and slcasyncSetFcStats */
   /*  regardless of whether a scp function is being processed */


   if ((asyncCycFuncIndex = slcAsyncGetFuncIxByName("CHKF")) == -1) {	
     slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "mgntAsyncCheckFeedback failed to find its function index\n");
     return(MGNT_DBERR);
   }

  if (slcAsyncMeterDbSend(asyncCycFuncIndex, requestType, epicsTrue)!=0)
	return (MGNT_DBERR);

  if (requestType==slcMsg) {
	if (slcAsyncSetFcnStats(asyncCycFuncIndex, &(mgnt_ps->func_s.reqStartTime_s), requestType)!=0)
	  return (MGNT_DBERR);
  }
  else{
	if (slcAsyncSetFcnStats(asyncCycFuncIndex, &(mgnt_ps->lastCheckFeedbackTime_s), requestType)!=0)
	  return (MGNT_DBERR);
  }

   return(status);

}
/*=============================================================================

  Name: mgntAsyncSlcToPVBdes ()

  Abs:  Reads float PV value and corresponding SLC float value.
        Compares the two values and updates EPICS from SLC if different, 
		with the option of using a tolerance window around the difference.
		Used for BDES.		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		DBADDR  *           dbaddr_ps    read       ptr to unit's epics address 
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit


  Rem:  

  side: Sets the mgnt_ps->func_s.doUpdate flag to true if dblputting to slc db.
        This way the 1 second async message will cause a call to dbupdate to
		update the Alpha after waiting for more units to "report in".

  Data: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL vmsstat_t mgntAsyncSlcToPVBdes( mgntUnitStruct_ts * const unit_ps)
							   
{
  vmsstat_t status = MGNT_OKOK;
  long  nreqs=1;      /* these are for fetching Epics PV's */
  long  options=0;
  float pvfloat = 0;
  float * db_ps;

  /* dbGetField PV value */
  if (unit_ps->bdesAddr_s.precord){
	if ( (dbGetField(&(unit_ps->bdesAddr_s), DBR_FLOAT, 
			   &(pvfloat), &options, &nreqs, NULL))) {
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
		 ("mgntAsyncSlcToPVFloat: Error for dbGetField  \n")); 
	}else { /* successful PV read */
	  /* read slc value */
	  ((dbheader_ts *)(unit_ps->bdesDlist_ps))->cur = 0;
	  if (SUCCESS(dblget( (const dbheader_ts * const)unit_ps->bdesPlist_ps,
                   (dbheader_ts **const)&(unit_ps->bdesDlist_ps) ) )) {	
		/* compare slc value with PV */
		db_ps = DBPTR_FIRST(unit_ps->bdesDlist_ps);
		DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
		 ("mgntAsyncSlcToPVBdes for %s: slc bdes=%f; epics bdes=%f\n",
		              (&(unit_ps->bdesAddr_s))->precord->name, *(float *)db_ps, pvfloat)); 		
		if (*(float *)db_ps != pvfloat ) {
		  if (dbPutField(&(unit_ps->bdesAddr_s), DBR_FLOAT, (void *)db_ps,nreqs) )
			return (MGNT_DBERR);
		}
	  }else DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
					   ("mgntAsyncSlcToPVBdes: Error with dblget\n")); 
	}
  }
  return status;
}
/*=============================================================================

  Name: mgntAsyncSLCToPVHstaCtrl ()

  Abs:  Reads float PV value and corresponding SLC float value.
        Compares the two values and dblputs to loca SLC db if different, 
		with the option of using a tolerance window around the difference.
		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		DBADDR  *           dbaddr_ps    read       ptr to unit's epics address 
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit


  Rem:  


  Data: 

  Ret: iss vmsstat_t vms style status

===============================================================================*/
LOCAL vmsstat_t mgntAsyncSlcToPVHstaCtrl (mgntUnitStruct_ts * const unit_ps)
									  
	                                   /* prob. want a tol here; if =0, don't use */
{
  vmsstat_t status = MGNT_OKOK;
  long  nreqs=1;                       /* these are for fetching Epics PV's */
  long  options=0;
  unsigned short pvshort = 0;
  unsigned short * db_ps;

  /* dbGetField PV value; read and compare with epics "HSTA"; set slc "HSTACTRL" */
  if (unit_ps->hstaAddr_s.precord){
	if ( (dbGetField(&(unit_ps->hstaAddr_s), DBR_USHORT, 
			   &(pvshort), &options, &nreqs, NULL))) {
	  DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
		 ("mgntAsyncSlcToPVHsta: Error for dbGetField  \n")); 
	}else { /* successful PV read */
	  /* read slc value */
	  ((dbheader_ts *)(unit_ps->hstaDlist_ps))->cur = 0;
	  if (SUCCESS(dblget( (const dbheader_ts * const)unit_ps->hstaPlist_ps,
                   (dbheader_ts **const)&(unit_ps->hstaDlist_ps) ) )) {	
		/* compare slc value with PV */
		db_ps = DBPTR_FIRST(unit_ps->hstaDlist_ps);
		DEBUGPRINT(DP_DEBUG, mgntAsyncFlag, 
		 ("mgntAsyncSlcToPVHsta for %s: slc hsta=%x; epics hsta=%x\n",
		              (&(unit_ps->hstaAddr_s))->precord->name, *(unsigned short *)db_ps, pvshort)); 		
		if (*(unsigned short *)db_ps != pvshort ) {
		  if (dbPutField(&(unit_ps->hstaCtrlAddr_s), DBR_USHORT, (void *)db_ps,nreqs) )
			return (MGNT_DBERR);
		}
	  }else DEBUGPRINT(DP_ERROR, mgntAsyncFlag, 
					   ("mgntAsyncSlcToPVHsta: Error with dblget\n")); 
	}
  }
  return status;
}
/*=============================================================================

  Name: mgntAsyncCleanup

  Abs:  Cleanup magnet handler before exiting

  Args: none

  Rem: This funtion is called to clean up any memory allocation and
       message queues before exiting the magnet handler.

  Side: 

  Ret:  int 
              0 = Successful
             -1 = failure

==============================================================================*/
LOCAL int mgntAsyncCleanup(void) 
{
  int status = MGNT_OKOK;

  DEBUGPRINT(DP_INFO, mgntAsyncFlag, ("MGNT_A: Cleaning up and exiting!!! \n"));

  /* if event flag is reset, set it to release mgntHdlr */
  if (mgnt_ps->func_s.funcDoneEvent_ps !=NULL) 
  {
	/* false if empty */
	/*if (!( epicsEventTryWait(mgnt_ps->func_s.funcDoneEvent_ps))) {
	  DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
	  ("MGNT_A: signaling funcDoneEvent_ps from mgntAsyncCleanup() \n"));*/
	  epicsEventSignal(mgnt_ps->func_s.funcDoneEvent_ps);
	  /*} else
	  DEBUGPRINT(DP_INFO, mgntAsyncFlag, 
	  ("MGNT_A: funcDoneEvent_ps returned true in mgntAsyncCleanup()\n")); */
  }
  /* Delete CA monitors - let mgntHdlr do this */

  /* Delete message queue */
  messageQDestroy(mgntAsync);
  slcThreads_as[mgntAsync].mqId_ps = NULL;
  slcThreads_as[mgntAsync].tid_ps = NO_THREAD;

  return( SUCCESS(status) ? 0 : -1 );
}

