/*=============================================================================
 
  Name: mgntGlob.c

         Initialization 
         --------------
         mgntGlobInit        - Initialize local database information
     *   mgntGlobInitPrims   - Initialize prim units in global structure
     *   mgntGlobInitSecn    - Initialize secn EPICS db addr lists for prim/unit
	 *   slcDbToEpicsDbAddr-  Get the EPICS db address for a pv mapped to
                             an SLC db of the form PRIM:MICR:UNIT:SECN
     *   mgntGloblist        - Creates the dblist for given prim, unit, secn
     *   mgntGlobInitLongLists- Creates "long" dblists, for all prim/units, db order

         Cleanup
         -------
         mgntGlobExit    - Free dblist memory used in magnet utilities


	 Note:  (*) - indicates static function

  Abs:  Magnet database utilities used for initializing epics db addresses, slc dblists
        contained in the global variables shared between mgntHdlr and mgntAsync.

       reference document:
          http://www.slac.stanford.edu/grp/lcls/controls/global/facilities/slc_ioc/mgnt_designSpec.html

  Auth: 10-Jul-2005, K.  Luchini      (LUCHINI)
  Rev:  20-Oct-2005, Debbie Rogind    (DROGIND)

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  
        24-jul-2007, dfairley
		   removed all references to mgntMonTypes.h, calls mgntMonInitUnits once
		   per primary, instead of mgntMonInitUnit per unit of each primary
        dd-mmm-yyyy, First Lastname (USERNAME):
           comment
   
=============================================================================*/

/* C system header files */
#include <stdio.h>
#include <string.h>            /* strncmp() */
#include <stdlib.h>            /* atoi      */

/* Debug header file */
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
   int mgntGlobFlag = DP_ERROR;
#endif

 
/* EPICS includes */
#include <dbDefs.h>        /* NELEMENTS                     */
#include <epicsEvent.h>    /* epicsEventId                  */
#include <epicsTypes.h>    /* epicsBoolean                  */
#include <freeList.h>      /* freeListItemsAvail            */
#include <dbAccess.h>      /* Epics Runtime Database        */
#include <dbAccessDefs.h>  /* dbPutField()                  */               
#include <dbFldTypes.h>    /* DBR_FLOAT,DBR_USHORT          */

/* SLC header files */
#include "micrdef.hc"      /* MICR_MALLOC                   */
#include "msgfunc.hc"      /* for TERSE                     */
#include "slcTypes.h"      /* MGNT_JOB_NAME,vmsstat_t       */
#include "slcUtil.h"       /* slcGetJobByEnum proto         */
#include "slcGlob.h"       /* slcGetIOCName()               */
#include "slcCmlog.h"      /* slcCmlogLogMsg                */
#include "sysUtil.h"       /* Find_index_c()                */

/* Database utility prototypes */
#include "dbdef.hc"        /* DB_NOTALLUNIT                 */
#include "dbGlob.h"
#include "dblist.h"        /* DBLIST,DBCLEAR,DBCNT,dblget() */
#include "dbl.h"           /* dblunits()                    */
#include "dbupdate.h"      /* dbupdateNoWait()              */
#include "dbdict.h"
#include "dblistUtil.h"

/* Magnet header files */
#include "mgntdef.hc"      /* magnet error codes            */
#include "mgntTypes.h"     /* mgntHdlrInfo_ts               */
#include "mgntMacros.h"    /* MGNT_PRIM,MGNT_PRIM_NNUM      */

#include "mgntMon.h"       /* mgntMon functions             */
/*#include "mgntMonTypes.h"   monPrimStruct_ts              */

/*
** Global variables for mgntHdlr, mgntAsync threads
*/

mgntHdlrInfo_ts  mgntInfo_s;
mgntHdlrInfo_ts  *mgnt_ps = &mgntInfo_s;
/* array of primaries; contains unit info */
mgntPrimStruct_ts mgntPrim_s[MGNT_NUM_PRIMS];
/* array of primaries; contins unit epics ca monitor info */
/*monPrimStruct_ts monPrim_s[MGNT_NUM_PRIMS];*/
/* timeouts, adjustable in rtems cexp */
double     funcTrimTimeout;
double     funcCalbTimeout;
double     funcDacZeroTimeout;
double     funcPerturbTimeout;
double     funcCheckTimeout;
double     funcStdzTimeout;

/* 
** Local prototypes 
*/
static vmsstat_t mgntGlobInitPrims( void );
static vmsstat_t mgntGlobInitSecn( mgntPrim_te prim_e );
static vmsstat_t mgntGlobInitSecnDbLists(const mgntPrim_te  prim_e, 
									   mgntUnitStruct_ts * unit_ps );
static vmsstat_t slcDbToEpicsDbAddr(const char * const prim_p,
                                    const char * const ioc_p,
                                    const char * const unit_p,
                                    const char * const secn_p,
                                    DBADDR     * const dbAddr_ps,
                                    char       * const pvName_p );
static vmsstat_t mgntGloblist (const char * const prim_a,  
							   const char * const unit_a,  
							   const char * const secn_a,  
							   dbheader_ts **const dblist_pps);
LOCAL void mgntGlobZeroSlcFloat (const dbheader_ts * const dblist_ps,
								 dbheader_ts **const dbdata_pps,
								 epicsBoolean * doUpdate );
LOCAL void mgntGlobZeroSlcUShort (const dbheader_ts * const dblist_ps,
								  dbheader_ts **const dbdata_pps,
								  epicsBoolean * doUpdate );
static vmsstat_t mgntGlobInitLongLists (void);

/*=============================================================================

  Name:  slcDbToEpicsDbAddr

  Abs:  

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        const char *     prim_p        read-only    ptr to primary name
        const char *     ioc_p         read-only    ptr to ioc/loca name
        const char *     unit_p        read-only    ptr to unit name
        const char *     secn_p        read-only    ptr to secn name
        const char *     DBADDR        read-only    ptr to dbAddr 

  Rem: 

  Side: secn_p must be null terminated

  Ret:  vmsstat_t  
             MGNT_OKOK     - Successful operation
             MGNT_DBERR    - Failed, due to database error
             MICR_MALLOC   - Memory allocation error.
             DB_NOTALLUNIT - Failure, due to no db units
             Otherwise, a failure code from the 
                mgntGlobInitPrim()

==============================================================================*/
static vmsstat_t slcDbToEpicsDbAddr(const char * const prim_p,
                                    const char * const ioc_p,
                                    const char * const unit_p,
                                    const char * const secn_p,
                                    DBADDR     * const dbAddr_ps,
                                    char       * const pvName_p )
{
    vmsstat_t            status=MGNT_OKOK;     /* return status            */
    int                  unitNo;               /* unit number              */
    char                 pvName_c[41];         /* EPICS pv name            */
											   /*   solaris doesn't like 40 */
	                                           /*   linux doesn't like 60 */
    unsigned long        bcnt;                 /* byte count               */    
    static unsigned long maxBcnt=sizeof(pvName_c)-1;    

    /*
    ** Verify that the pv name isn't longer than maximum size of our
    ** pv name string
    */
    unitNo = atoi(unit_p);
	memset (pvName_c, 0, 60);
	/*DEBUGPRINT(DP_INFO, mgntGlobFlag,("prim=%.4s, unitNum=%d, secn=%.4s\n",prim_p, unitNo, secn_p));*/ 
       
    bcnt   = sprintf(pvName_c,"%.4s:%.4s:%d:%s",prim_p,ioc_p,unitNo,secn_p);
    bcnt += sizeof(secn_p); 
    if (bcnt<maxBcnt) {
      bcnt = sizeof(pvName_c)-1;
      sprintf(&pvName_c[bcnt],":%s",secn_p);
    }
    if ( dbNameToAddr(pvName_c,dbAddr_ps) ) {
       status = MICR_DBBADMESSG;
       dbAddr_ps->precord=NULL;     /* keeps us from doing dbGet */
       DEBUGPRINT(DP_ERROR, mgntGlobFlag,
                 ("slcDbToEpicsDbAddr: dbNameToAddr of %s failed\n",pvName_c));
	   slcCmlogLogMsg(status, "MICR_DBBADMESSG", "%s SLC/EPICS db mismatch: No EPICS PV for %s \n", 
						  epicsThreadGetNameSelf(), pvName_c);
    }
    /* Return the pv name if requested to do so */
    else if ( pvName_p )
      strcpy(pvName_p,pvName_c);
    return(status);
}

/*=============================================================================

  Name: mgntGlobInit

  Abs:  Initialize magnet local database 

  Args: None

  Rem: The purpose of this function is to update the CSTR.1.MAGF 
       with the supplied magnet function and then to notify the Alpha.

  Side: A database update is done to reflect the current magnet function
        that is being processed.

  Ret:  vmsstat_t  
             MGNT_OKOK     - Successful operation
             MGNT_DBERR    - Failed, due to database error
             MICR_MALLOC   - Memory allocation error.
             DB_NOTALLUNIT - Failure, due to no db units
             Otherwise, a failure code from the 
                mgntGlobInitPrim()

==============================================================================*/
vmsstat_t mgntGlobInit( void ) 
{
    MGNT_PRIM_ENUM;                       /* magnet primary enum list */
    vmsstat_t       status = MGNT_OKOK;   /* status return            */
    unsigned long   i_prim;               /* primary index            */       


   /* init the ca context prior to initing the indiv channels */
   status = mgntMonInitContext();  

    /* Get all primary units */
    status = mgntGlobInitPrims();
	DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
                 ("mgntGlobInit: Initing EPICS db addrs... \n"));
    /* Get EPICS db addr id's for pv's associated with magnet primary secondaries */
    for (i_prim = 0; (i_prim<MGNT_NUM_PRIMS && SUCCESS(status)); i_prim++) {
	  DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
                 ("mgntGlobInit: Number of units=%d for primary index = %lu\n", 
				  mgntPrim_s[i_prim].nunits, i_prim ));
      if ( mgntPrim_s[i_prim].nunits ) 
        if (!(SUCCESS(status = mgntGlobInitSecn(mgntPrim_ae[i_prim]) ))) {
		  DEBUGPRINT(DP_ERROR, mgntGlobFlag,
                 ("mgntGlobInit: Failure during mgntGlobInitSecn()!!!\n"));
		  return status;
		}
		  
    } /* End of i_prim FOR loop */

	/* now get the master dblists for dblgetting values all at once */
	if (!(SUCCESS(status=mgntGlobInitLongLists() ))) {
	  DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
                 ("mgntGlobInit: Failure during mgntGlobInitLongLists()!!!\n"));
	}

	DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
                 ("mgntGlobInit: Return from mgntGlobInit\n"));

	funcTrimTimeout    = MGNT_TRIM_TIMEOUT;
    funcCalbTimeout    = MGNT_CALB_TIMEOUT;
    funcDacZeroTimeout = MGNT_ZERO_TIMEOUT;
    funcPerturbTimeout = MGNT_PTRB_TIMEOUT;
	funcCheckTimeout   = MGNT_CHCK_TIMEOUT;
    funcStdzTimeout    = MGNT_STDZ_TIMEOUT;

    return(status);
}

/*=============================================================================

  Name: mgntGlobInitPrims

  Abs:  Get all magnet units

  Args: None

  Rem: The purpose of this function is to get all of the magnet units
       and initialize the information in the global structure for the units
       based on primary.

  Side: No wildcard primary names.
        Iterates through (using i_prim) all possible primaries for an ioc and inits the
		mgntPrim_s[i_prim] structure.
		Also populates  mgnt_ps->dlist_s.slcAllUnits_ps as it iterates thru
		all possible primaries.
        Reads the tolerances for each primary once and populates
        prim_ps->tolsPlist_ps, prim_ps->tolsDlist_ps 
		Allocates space for mgntPrim_s[i_prim].units_ps structure and 
		monPrim_s[i_prim].units_ps structure based on number of units found
		for each primary, i_prim.


  Ret:  vmsstat_t  
        MGNT_OKOK    - Successful operation
        Otherwise, a failure code from dblunits()

==============================================================================*/
static vmsstat_t  mgntGlobInitPrims( void )
{ 
    vmsstat_t           status = MGNT_OKOK;   /* status return            */
    unsigned short      prevCnt=0;            /* previous count           */
    unsigned long       i_prim=0;             /* primary index            */
    unsigned long       i_unit=0;             /* unit index               */
    unsigned long       nbytes;               /* number of bytes          */
    unsigned long       nelm=20;              /* # of elements            */
    mgntUnitStruct_ts   *unit_ps=NULL;         /* unit information         */
    mgntPrimStruct_ts  *prim_ps=NULL;         /* primary info             */
	unsigned long      *db_p;                 /* iterative ptr for db unit access */
    MGNT_PRIM;                                /* list of magnet primaries */
    MGNT_PRIM_ENUM;                           /* list of magnet prim enums*/
                                              /* same order as MGNT_PRIM  */
   /* 
   ** Allocate some memory for the db lists and then
   ** set the number of items in the db lsit to zero.
   */
   if (! SUCCESS(status=DBALLOC( mgnt_ps->dlist_s.slcAllUnits_ps, nelm ))){
	 return (status);
   }
   memset (DBPTR( mgnt_ps->dlist_s.slcAllUnits_ps,0), 0, DBMAX( mgnt_ps->dlist_s.slcAllUnits_ps));
   DBCLEAR( mgnt_ps->dlist_s.slcAllUnits_ps );

   /* 
   ** Get the units for each magnet primary and init the
   ** global structure mgnt_ps
   */
   for (i_prim=0; (i_prim<MGNT_NUM_PRIMS /*&& SUCCESS(status)*/); i_prim++) {
     prim_ps  = &mgntPrim_s[i_prim];
	 /*     monPrim_ps = &monPrim_s[i_prim];                          set up monitor prim var too */

     prim_ps->prim_e  = mgntPrim_ae[i_prim];                  /* primary enum */
	 /*     monPrim_ps->prim_e =  mgntPrim_ae[i_prim]; */ 

     strcpy((char *)prim_ps->prim_ac,mgntPrim_ac[i_prim]);    /* primary name */
     /* 
     ** Get unit numbers for this primary.
     ** Add them to the complete list of units. 
     ** Note: the unit numbers are 4-chars strings
     **  not NULL terminated.
     */
     prevCnt = mgnt_ps->nunits;
     if (SUCCESS(dblunits(mgntPrim_ac[i_prim],
						  (dbheader_ts ** const)&mgnt_ps->dlist_s.slcAllUnits_ps ))) {
       mgnt_ps->nunits = DBCNT(mgnt_ps->dlist_s.slcAllUnits_ps);   /* total # of units so far     */
       prim_ps->nunits = mgnt_ps->nunits - prevCnt;;      /* # of units for this primary */
	   /*       monPrim_ps->nunits = prim_ps->nunits; */

	   DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
       ("mgntGlobInitPrim: dblunits for mgntPrim_ac[%lu]=%.4s yielded %d units; %d total units so far; prevCnt=%u\n",
				  i_prim, mgntPrim_ac[i_prim], prim_ps->nunits, mgnt_ps->nunits , prevCnt));
       /* does this primary have units? */
       if ( prim_ps->nunits ) {
         prim_ps->slcUnitsI   = prevCnt; /* set the first unit index into the master unit list. */
		 /*	 prim_ps->slcUnits_ps = DBPTR(mgnt_ps->dlist_s.slcAllUnits_ps, prim_ps->slcUnitsI);*/
		 DEBUGPRINT(DP_DEBUG, mgntGlobFlag,("\tfirst unit index = %u; slcAllUnits_ps=%p\n", 
										  prim_ps->slcUnitsI,mgnt_ps->dlist_s.slcAllUnits_ps));

		 /* allocate memory for all units of this primary  */
		 nbytes = prim_ps->nunits * sizeof(mgntUnitStruct_ts);
		 if (! SUCCESS (status=DBALLOC (prim_ps->units_ps, prim_ps->nunits))) {
		   return (status);
		 }
		 /* zero out (all ptrs) in this unit structure */
		 /*memset (DBPTR( prim_ps->units_ps, 0), 0, nbytes);*/
		 db_p = DBPTR(mgnt_ps->dlist_s.slcAllUnits_ps, prim_ps->slcUnitsI);
		 unit_ps = DBPTR_FIRST(prim_ps->units_ps);
         for ( i_unit=0; (i_unit<prim_ps->nunits && unit_ps); i_unit++, db_p++, unit_ps++ ) {
		   /*unit_ps = &(mgntPrim_s[i_prim].units_s[i_unit]);*/
		   unit_ps->prim_pc = &(prim_ps->prim_ac[0]);   /* primary ascii information           */
           unit_ps->slcUnitsI  = prim_ps->slcUnitsI + i_unit; /* index into master slcAllUnits list */
		   unit_ps->prim_e = prim_ps->prim_e;
		   memset(&(unit_ps->unit_ac[0]), 0, 5);
		   memcpy(&(unit_ps->unit_ac[0]), db_p, 4); /* 4-char unit number null-terminated */
		   DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
                 ("mgntGlobInitPrim: %.4s %.4s is setup with slcAllUnits index %u\n",
				 unit_ps->prim_pc, unit_ps->unit_ac, unit_ps->slcUnitsI ));   
		 }

       }/* Have units for this primary */
     }/* dblist was successful */
	 else {
	   DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
                 ("mgntGlobInitPrim: No units for primary index %lu\n",i_prim ));
	   /* pointer to units_ps is NULL */
	   prim_ps->units_ps = NULL;
	   /*monPrim_ps->units_ps = NULL;*/
	 }
   }/* End of i_prim FOR loop */
   
   /* 
   ** Do we have any magnet units in this ioc? 
   ** If not, then set the return status appropriately.
   ** Note: an ioc without units is NOT considered an error.
   ** We should simply exit the magnet handler gracefully
   ** if no units are found in this ioc and a message should
   ** be issued to inform the operator as to why the magnet
   ** handler won't be running.
   */
   if ((mgnt_ps->nunits==0) && SUCCESS(status)){ 
     status = DB_NOTALLUNIT;
	 DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
                 ("mgntGlobInitPrim: No magnet units in this ioc\n"));
   } else {
	 /* debug */
	 db_p = DBPTR(mgnt_ps->dlist_s.slcAllUnits_ps, 0);
	 for (i_unit=0; i_unit<DBCNT(mgnt_ps->dlist_s.slcAllUnits_ps); i_unit++, db_p++) {
	   DEBUGPRINT(DP_DEBUG, mgntGlobFlag, 
				  ("mgntGlobUnits: Unit %ld = %.4s; ptr=%p\n",i_unit,(char *)db_p, db_p)); 	  
	 }
   }

   return(status);
}
/*=============================================================================

  Name: mgntGlobInitSecnDbLists

  Abs:  For all units of input primary, initialize its structures to EPICS 
        pv names and SLC dblists for all primaries

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        mgntPrim_te      prim_e        read-only    primary enum

  Rem: Note that not all magnet primaries will have the same set of EPICS pv's
       associated with a SLC db unit.

  Side: None

  Ret:  vmsstat_t  
        MGNT_OKOK     - Successful operation
        MGNT_NOTMAGPRIM - Invalid magnet primary enum

==============================================================================*/
static vmsstat_t mgntGlobInitSecnDbLists(const mgntPrim_te  prim_e, 
									   mgntUnitStruct_ts * unit_ps ) 
{
  MGNT_PRIM;
  vmsstat_t            status = MGNT_OKOK;   /* status return                  */

  char                 slcHSTA_ca[5] = "HSTA";
  char                 slcSTAT_ca[5] = "STAT";
  char                 slcTTIM_ca[5] = "TTIM";
  char                 slcBDES_ca[5] = "BDES";
  char                 slcITRY_ca[5] = "ITRY";
  char                 slcBACT_ca[5] = "BACT";
  char                 slcBMON_ca[5] = "BMON";
  char                 slcIDES_ca[5] = "IDES";
  char                 slcIACT_ca[5] = "IACT";
  char                 slcRIPL_ca[5] = "RIPL";
  char                 slcIPRV_ca[5] = "IPRV";
  char                 slcIMON_ca[5] = "IMON";
  char                 slcIMAX_ca[5] = "IMAX";
  char                 slcCTIM_ca[5] = "CTIM";
  char                 slcKTIM_ca[5] = "KTIM";
  char                 slcZTIM_ca[5] = "ZTIM";
  char                 unit_ac[5];

  memset (unit_ac, 0, 5);
  memcpy (unit_ac, unit_ps->unit_ac, 4);
  
  /* now init the db ptr lists Plist, and data lists, Dlist */
  if (!SUCCESS(status=DBALLOC(unit_ps->hstaPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)unit_ac, &slcHSTA_ca[0], 
						  (dbheader_ts **const) &(unit_ps->hstaPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->hstaDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->bdesPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcBDES_ca[0], 
						  (dbheader_ts **const) &(unit_ps->bdesPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->bdesDlist_ps,1))) return (status);
	DEBUGPRINT(DP_INFO, mgntGlobFlag,
		   ("mgntGlob: bdesDlist_ps = %p for %.4s\n",unit_ps->bdesDlist_ps, unit_ps->unit_ac));			
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->statPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcSTAT_ca[0], 
						  (dbheader_ts **const) &(unit_ps->statPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->statDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->ttimPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcTTIM_ca[0], 
						  (dbheader_ts **const) &(unit_ps->ttimPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->ttimDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->itryPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcITRY_ca[0], 
						  (dbheader_ts **const) &(unit_ps->itryPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->itryDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->bactPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcBACT_ca[0], 
						  (dbheader_ts **const) &(unit_ps->bactPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->bactDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->bmonPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcBMON_ca[0], 
						  (dbheader_ts **const) &(unit_ps->bmonPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->bmonDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->idesPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcIDES_ca[0], 
						  (dbheader_ts **const) &(unit_ps->idesPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->idesDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->iactPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcIACT_ca[0], 
						  (dbheader_ts **const) &(unit_ps->iactPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->iactDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->riplPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]),  &slcRIPL_ca[0], 
						  (dbheader_ts **const) &(unit_ps->riplPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->riplDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->iprvPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcIPRV_ca[0], 
						  (dbheader_ts **const) &(unit_ps->iprvPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->iprvDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->imonPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcIMON_ca[0], 
						  (dbheader_ts **const) &(unit_ps->imonPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->imonDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->imaxPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcIMAX_ca[0], 
						  (dbheader_ts **const) &(unit_ps->imaxPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->imaxDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->ctimPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcCTIM_ca[0], 
							 (dbheader_ts **const) &(unit_ps->ctimPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->ctimDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->ktimPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcKTIM_ca[0], 
						  (dbheader_ts **const) &(unit_ps->ktimPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->ktimDlist_ps,1))) return (status);		
  }
  if (!SUCCESS(status=DBALLOC(unit_ps->ztimPlist_ps,1))) return (status);
  if (SUCCESS (mgntGloblist(mgntPrim_ac[prim_e], (char *)&(unit_ac[0]), &slcZTIM_ca[0], 
						  (dbheader_ts **const) &(unit_ps->ztimPlist_ps)))) {
	/* we have this triplet, so alloc space for the Data list */
	if (!SUCCESS(status=DBALLOC(unit_ps->ztimDlist_ps,1))) return (status);		
  }
  return status;
}

/*=============================================================================

  Name: mgntGlobInitSecn

  Abs:  For all units of input primary, initialize its structures to EPICS 
        pv names and SLC dblists for all primaries

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        mgntPrim_te      prim_e        read-only    primary enum

  Rem: Note that not all magnet primaries will have the same set of EPICS pv's
       associated with a SLC db unit.

  Side: None

  Ret:  vmsstat_t  
        MGNT_OKOK     - Successful operation
        MGNT_NOTMAGPRIM - Invalid magnet primary enum

==============================================================================*/
static vmsstat_t mgntGlobInitSecn( mgntPrim_te prim_e ) 
{
   vmsstat_t           status = MGNT_OKOK;   /* status return                         */
   mgntPrimStruct_ts   *prim_ps = NULL;      /* primary unit information              */
   /*   monPrimStruct_ts    *monPrim_ps = NULL;    primary monitor unit info             */
   mgntUnitStruct_ts   *unit_ps = NULL;      /* unit structure w/epics dbaddr ptrs    */
   char                *prim_pc  = NULL;     /* ptr to primary name (NULL terminated) */
   char                *micr_pc  = NULL;     /* ptr to 4-char micro name              */
   unsigned short       i_unit;              /* unit index                            */
   epicsBoolean         doUpdate = epicsFalse;  /* perform dbupdate if true              */
   char                 secnCTRL_ca[] = PV_CTRL;
   char                 secnCTRLSET_ca[] = PV_CTRL_SETPT;
   char                 secnCTRLSTATE_ca[] = PV_CTRLSTATE;
   char                 secnHSTA_ca[] = PV_EPICSHSTA;
   char                 secnHSTACTRL_ca[] = PV_SLCHSTA;
   char                 secnSTAT_ca[] = PV_STAT;
   char                 secnTTIM_ca[] = PV_TTIM;   
   char                 secnBDES_ca[] = PV_BDES;
   char                 secnITRY_ca[] = PV_ITRY;
   char                 secnBACT_ca[] = PV_BACT;
   char                 secnBMON_ca[] = PV_BMON;
   char                 secnIDES_ca[] = PV_IDES;
   char                 secnIACT_ca[] = PV_IACT;
   char                 secnRIPL_ca[] = PV_RIPL;
   char                 secnIPRV_ca[] = PV_IPRV;
   char                 secnIMON_ca[] = PV_IMON;
   char                 secnIMAX_ca[] = PV_IMAX;
   char                 secnCTIM_ca[] = PV_CTIM;
   char                 secnKTIM_ca[] = PV_KTIM;
   char                 secnZTIM_ca[] = PV_ZTIM;
   char                 unit_ac[5];


   if ((prim_e<lgps_e) || (prim_e>sext_e)) 
	 return(MGNT_NOTMAGPRIM);
   
   prim_ps = &mgntPrim_s[prim_e];


   if (!prim_ps->nunits) {
	 DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
				("mgntGlobInitSecn: Exiting - No units for %.4s\n", 
				 prim_ps->prim_ac ));	
	 return(status);
   }

   prim_pc   = prim_ps->prim_ac;   /* get ptr to primary name */
   micr_pc   = slcGetLOCAName();  /* get location name            */
		  
   /* 
   ** Get EPICS db address id for each unit for pv's associated
   ** with SLC an db magnet unit 
   ** Note: don't abort if you don't find a corresponding EPICS pv.
   **/

   unit_ps = DBPTR_FIRST(prim_ps->units_ps);

   for (i_unit=0; i_unit<prim_ps->nunits && unit_ps; unit_ps++,i_unit++) {

	 memset (unit_ac, 0, 5);
	 memcpy (unit_ac, unit_ps->unit_ac, 4);

	 status = mgntGlobInitSecnDbLists(prim_e,unit_ps);
     /*DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
				("mgntGlobInitSecn: Calling slcDBToEpicsDbAddr for  %.4s:%.4s:%.4s\n", 
				prim_pc, micr_pc, (char*)unit_ps->unit_ac));*/

	 /* For ea unit, if the pv doesn't exist and it has a corresponding slc db counterpart,  
		set the slc db value to zero */

	 /*DEBUGPRINT(DP_DEBUG, mgntGlobFlag, ("mgntGlobInitSecn: unit = 0x%lx\n", *(unsigned long *)unit_ac));*/
	 
	 if (SUCCESS (slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnCTRL_ca[0],&(unit_ps->ctrlAddr_s),NULL)))
	 {
	   
	   status = slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnCTRLSET_ca[0],&(unit_ps->ctrlSetptAddr_s),NULL);
	   status = slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnCTRLSTATE_ca[0],&(unit_ps->ctrlStateAddr_s),NULL);   
	   status = slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnBDES_ca[0],&(unit_ps->bdesAddr_s),NULL);
	   
	   status = slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnHSTA_ca[0],&(unit_ps->hstaAddr_s),NULL);
	   
	   status = slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnHSTACTRL_ca[0],
								   &(unit_ps->hstaCtrlAddr_s),NULL);
	   
	   if (!SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnSTAT_ca[0],&(unit_ps->statAddr_s),NULL))) 
		 mgntGlobZeroSlcUShort ( (const dbheader_ts * const)unit_ps->statPlist_ps,
								 (dbheader_ts **const)&(unit_ps->statDlist_ps), &doUpdate );
	   
	   
	   status= slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnTTIM_ca[0],&(unit_ps->ttimAddr_s),NULL);
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnITRY_ca[0],&(unit_ps->itryAddr_s),NULL)))
		 mgntGlobZeroSlcUShort ( (const dbheader_ts * const)unit_ps->itryPlist_ps,
								 (dbheader_ts **const)&(unit_ps->itryDlist_ps), &doUpdate );
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnBACT_ca[0],&(unit_ps->bactAddr_s),NULL)))
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->bactPlist_ps,
								(dbheader_ts **const)&(unit_ps->bactDlist_ps), &doUpdate );
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnBMON_ca[0],&(unit_ps->bmonAddr_s),NULL)))
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->bmonPlist_ps,
								(dbheader_ts **const)&(unit_ps->bmonDlist_ps), &doUpdate );
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnIDES_ca[0],&(unit_ps->idesAddr_s),NULL)))
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->idesPlist_ps,
								(dbheader_ts **const)&(unit_ps->idesDlist_ps), &doUpdate );
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnIACT_ca[0],&(unit_ps->iactAddr_s),NULL)))
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->iactPlist_ps,
								(dbheader_ts **const)&(unit_ps->iactDlist_ps), &doUpdate );
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnRIPL_ca[0],&(unit_ps->riplAddr_s),NULL)))
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->riplPlist_ps,
								(dbheader_ts **const)&(unit_ps->riplDlist_ps), &doUpdate );
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnIPRV_ca[0],&(unit_ps->iprvAddr_s),NULL)))
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->iprvPlist_ps,
								(dbheader_ts **const)&(unit_ps->iprvDlist_ps), &doUpdate );
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnIMON_ca[0],&(unit_ps->imonAddr_s),NULL)))
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->imonPlist_ps,
								(dbheader_ts **const)&(unit_ps->imonDlist_ps), &doUpdate );
	   
	   if (! SUCCESS(slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnIMAX_ca[0],&(unit_ps->imaxAddr_s),NULL)))
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->imaxPlist_ps,
								(dbheader_ts **const)&(unit_ps->imaxDlist_ps), &doUpdate );
	   
	   status = slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnCTIM_ca[0],&(unit_ps->ctimAddr_s),NULL);
	   
	   status = slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnKTIM_ca[0],&(unit_ps->ktimAddr_s),NULL);

	   /* by commenting this line out, unit strings are restored and it behaves properly */
	   status = slcDbToEpicsDbAddr(prim_pc,micr_pc,(char *)&(unit_ac),&secnZTIM_ca[0],&(unit_ps->ztimAddr_s),NULL);
	 } else
	   { /* magnet is not in the epics datbase, so don't try to connect to the other pv fields, to save time*/
		 unit_ps->ctrlSetptAddr_s.precord = NULL;	   
	     unit_ps->bdesAddr_s.precord = NULL;	   
	     unit_ps->hstaAddr_s.precord = NULL;	   
	     unit_ps->hstaCtrlAddr_s.precord = NULL;	   
	     unit_ps->statAddr_s.precord = NULL; 
		 mgntGlobZeroSlcUShort ( (const dbheader_ts * const)unit_ps->statPlist_ps,
								 (dbheader_ts **const)&(unit_ps->statDlist_ps), &doUpdate );
         unit_ps->ttimAddr_s.precord = NULL;	   
	     unit_ps->itryAddr_s.precord = NULL;
		 mgntGlobZeroSlcUShort ( (const dbheader_ts * const)unit_ps->itryPlist_ps,
								 (dbheader_ts **const)&(unit_ps->itryDlist_ps), &doUpdate );	   
	     unit_ps->bactAddr_s.precord = NULL;
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->bactPlist_ps,
								(dbheader_ts **const)&(unit_ps->bactDlist_ps), &doUpdate );	   
	     unit_ps->bmonAddr_s.precord = NULL;
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->bmonPlist_ps,
								(dbheader_ts **const)&(unit_ps->bmonDlist_ps), &doUpdate );		 
		 unit_ps->idesAddr_s.precord = NULL;
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->idesPlist_ps,
								(dbheader_ts **const)&(unit_ps->idesDlist_ps), &doUpdate );	   
		 unit_ps->iactAddr_s.precord = NULL;
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->iactPlist_ps,
								(dbheader_ts **const)&(unit_ps->iactDlist_ps), &doUpdate );	   
		 unit_ps->riplAddr_s.precord = NULL;
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->riplPlist_ps,
								(dbheader_ts **const)&(unit_ps->riplDlist_ps), &doUpdate );	   
		 unit_ps->iprvAddr_s.precord = NULL;
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->iprvPlist_ps,
								(dbheader_ts **const)&(unit_ps->iprvDlist_ps), &doUpdate );	   
		 unit_ps->imonAddr_s.precord = NULL;
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->imonPlist_ps,
								(dbheader_ts **const)&(unit_ps->imonDlist_ps), &doUpdate );		 
		 unit_ps->imaxAddr_s.precord = NULL;
		 mgntGlobZeroSlcFloat ( (const dbheader_ts * const)unit_ps->imaxPlist_ps,
								(dbheader_ts **const)&(unit_ps->imaxDlist_ps), &doUpdate );	   
		 unit_ps->ctimAddr_s.precord = NULL;	   
		 unit_ps->ktimAddr_s.precord = NULL;
		 unit_ps->ztimAddr_s.precord = NULL;
	   }
	 if (doUpdate) status = dbupdate(mgntJob);
	 
   } /* End of i_unit FOR loop */ 
   
   /* now do the monitors for this primary */
   status = mgntMonInitUnits(prim_e);

   return(MGNT_OKOK);
}

/*=============================================================================

  Name: mgntGlobZeroSlcFloat ()

  Abs:  Called if a float PV doesn't exist and it has a slc db counterpart.
        The slc db value is set to zero.
		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit
		epicsBoolean      * doUpdate,    write      perform a dbupate after all done

  Rem:  

  side: Allow for ANY PV to be nonexistent on EPICS side. If its a PV that the
        code does a dbget from (like during a check function), set the slc db
		value to 0 once at initialization.

  Data: 

  Ret: void

===============================================================================*/
LOCAL void mgntGlobZeroSlcFloat (const dbheader_ts * const dblist_ps,
							   dbheader_ts **const dbdata_pps,
							   epicsBoolean * doUpdate )
{

  /*  float         initVal = 0.0;          
  unsigned char * db_ps;
  
  if (dblist_ps && *dbdata_pps){
	db_ps = (unsigned char *) (*dbdata_pps);
	db_ps += sizeof(dbheader_ts);	
	memcpy (db_ps, &initVal, sizeof(float));
	dblput (dblist_ps, *dbdata_pps, mgntJob);  
	*doUpdate = epicsTrue;
	}*/
}
/*=============================================================================

  Name: mgntGlobZeroSlcUShort ()

  Abs:  Called if an unsigned short PV doesn't exist and it has a slc db counterpart.
        The slc db value is set to zero.
		


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const dbheader_ts * const dblist_ps,  read  slc dictionary ptr dblist for unit
	    dbheader_ts **const dbdata_pps,  read/write slc data dblist for unit
		epicsBoolean      * doUpdate,    write      perform a dbupate after all done

  Rem:  

  side: Allow for ANY PV to be nonexistent on EPICS side. If its a PV that the
        code does a dbget from (like during a check function), set the slc db
		value to 0 once at initialization.

  Data: 

  Ret: void

===============================================================================*/
LOCAL void mgntGlobZeroSlcUShort (const dbheader_ts * const dblist_ps,
								dbheader_ts **const dbdata_pps,
								epicsBoolean * doUpdate )
{
  /* unsigned short      initVal;             
  unsigned char * db_ps;

  if (dblist_ps && *dbdata_pps){
	db_ps = (unsigned char *) (*dbdata_pps);
	db_ps += sizeof(dbheader_ts);	
	memcpy (db_ps, &initVal, sizeof(unsigned short));
	dblput (dblist_ps, *dbdata_pps, mgntJob); 
	*doUpdate = epicsTrue;
	}*/
}
/*=============================================================================

  Name: mgntGlobInitLongLists

  Abs:  Initialize pointers to slc db lists that are part of mgntInfo_s structure

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------

  Rem:  Initialize mgntInfo_s.dlist and .plist

  Side: allocate space and populate
        mgnt_ps->plist_s and  mgnt_ps->dlist_s (ptr and dblists)
		for hsta, stat, bdes, tols, tolf

		just allocates space for  mgnt_ps->dlist_s.cstrMagf_s and 
		mgnt_ps->plist_s.cstrMagf_s  (no need to read)

  Ret:  vmsstat_t  
             MGNT_OKOK     - Successful operation
             MGNT_NOTMAGPRIM - Invalid magnet primary enum

==============================================================================*/
LOCAL vmsstat_t mgntGlobInitLongLists(void)
{
  vmsstat_t status = MGNT_OKOK;
  unsigned short     i_prim;
  mgntPrimStruct_ts  *prim_ps;
  mgntUnitStruct_ts  *unit_ps = NULL;     
  unsigned short     i_unit;
   /* 
   ** Allocate some memory for the db lists and then
   ** set the number of items in the db lsit to zero.
   */

   if (! SUCCESS(status=DBALLOC( mgnt_ps->plist_s.hsta_ps, mgnt_ps->nunits ))) {
	 return (status);
   }

   if (! SUCCESS(status=DBALLOC( mgnt_ps->plist_s.stat_ps, mgnt_ps->nunits ))){
	 return (status);
   }

   if (! SUCCESS(status=DBALLOC( mgnt_ps->plist_s.bdes_ps, mgnt_ps->nunits ))){
	 return (status);
   }

   if (! SUCCESS(status=DBALLOC( mgnt_ps->dlist_s.hsta_ps, mgnt_ps->nunits ))){
	 return (status);
   }

   if (! SUCCESS(status=DBALLOC( mgnt_ps->dlist_s.stat_ps, mgnt_ps->nunits ))){
	 return (status);
   }

   if (! SUCCESS(status=DBALLOC( mgnt_ps->dlist_s.bdes_ps, mgnt_ps->nunits ))){
	 return (status);
   }

   /* allocate memeory for the ST 1 BDES Tolerances, TOLS, for this primary */
   if (! SUCCESS(status=DBALLOC( mgnt_ps->plist_s.tols_ps, (mgnt_ps->nunits * 4) ))){
	 return (status);
   }
   DBCLEAR( mgnt_ps->plist_s.tols_ps);
   if (! SUCCESS(status=DBALLOC( mgnt_ps->dlist_s.tols_ps, (mgnt_ps->nunits *4) ))){
	 return (status);
   }
   DBCLEAR( mgnt_ps->dlist_s.tols_ps );

   /* allocate memeory for the ST 1 BMON Tolerances, TOLF, for this primary */
   if (! SUCCESS(status=DBALLOC( mgnt_ps->plist_s.tolf_ps, mgnt_ps->nunits ))){
	 return (status);
   }
   DBCLEAR( mgnt_ps->plist_s.tolf_ps);
   if (! SUCCESS(status=DBALLOC( mgnt_ps->dlist_s.tolf_ps, mgnt_ps->nunits ))){
	 return (status);
   }
   DBCLEAR( mgnt_ps->dlist_s.tolf_ps );
      
   /* now get the db order ptrs for all primaries and their associated units */
   for (i_prim = 0; i_prim < MGNT_NUM_PRIMS; i_prim++ ) {
	 prim_ps = &mgntPrim_s[i_prim];
	 unit_ps = DBPTR_FIRST(prim_ps->units_ps);
	 for (i_unit = 0; i_unit<prim_ps->nunits; i_unit ++, unit_ps++) {

	   if (!(SUCCESS (mgntGloblist(&(prim_ps->prim_ac[0]), 
								 (char *)&(unit_ps->unit_ac[0]), "HSTA", 
								 (dbheader_ts **const) &((mgnt_ps->plist_s.hsta_ps)))) )) {
		 DBFREE (mgnt_ps->dlist_s.hsta_ps);
		 mgnt_ps->dlist_s.hsta_ps = NULL;
		 return (status);
	   }
	   if (!(SUCCESS (mgntGloblist(&(prim_ps->prim_ac[0]), 
								 (char *)&(unit_ps->unit_ac[0]), "BDES", 
								 (dbheader_ts **const) &((mgnt_ps->plist_s.bdes_ps)))) )) {
		 DBFREE (mgnt_ps->dlist_s.bdes_ps);
		 mgnt_ps->dlist_s.bdes_ps = NULL;
		 return (status);
	   }
	   if (!(SUCCESS (mgntGloblist(&(prim_ps->prim_ac[0]), 
								 (char *)&(unit_ps->unit_ac[0]), "STAT", 
								 (dbheader_ts **const) &((mgnt_ps->plist_s.stat_ps)))) )) {
		 DBFREE (mgnt_ps->dlist_s.stat_ps);
		 mgnt_ps->dlist_s.stat_ps = NULL;
		 return (status);
	   }

	   if (!(SUCCESS (mgntGloblist(&(prim_ps->prim_ac[0]), 
								 (char *)&(unit_ps->unit_ac[0]), "TOLS", 
								 (dbheader_ts **const) &((mgnt_ps->plist_s.tols_ps)))) )) {
		 DBFREE (mgnt_ps->dlist_s.tols_ps);
		 mgnt_ps->dlist_s.tols_ps = NULL;
		 return (status);
	   }

	   if (!(SUCCESS (mgntGloblist(&(prim_ps->prim_ac[0]), 
								 (char *)&(unit_ps->unit_ac[0]), "TOLF", 
								 (dbheader_ts **const) &((mgnt_ps->plist_s.tolf_ps)))) )) {
		 DBFREE (mgnt_ps->dlist_s.tolf_ps);
		 mgnt_ps->dlist_s.tolf_ps = NULL;
		 return (status);
	   }

	   /*status = dblget((const dbheader_ts * const) mgnt_ps->plist_s.hsta_ps,  
					 (dbheader_ts **const) mgnt_ps->dlist_s.hsta_ps);
	   status = dblget((const dbheader_ts * const) mgnt_ps->plist_s.stat_ps,  
					 (dbheader_ts **const) mgnt_ps->dlist_s.stat_ps);
	   status = dblget((const dbheader_ts * const) mgnt_ps->plist_s.bdes_ps,  
	   (dbheader_ts **const)mgnt_ps->dlist_s.bdes_ps);*/
	 }
   }

   /* read supertype 1 tols and tolf for all prim/units once, during init */
   dblget( (const dbheader_ts * const)mgnt_ps->plist_s.tols_ps, 
		   (dbheader_ts **const)&(mgnt_ps->dlist_s.tols_ps) ); 
   dblget( (const dbheader_ts * const)mgnt_ps->plist_s.tolf_ps, 
		   (dbheader_ts **const)&(mgnt_ps->dlist_s.tolf_ps) ); 
   
   /* get 1 unit of CSTR MAGF */
   /* Allocate memory for CSTR MAGF dblists  */
   if (!SUCCESS(status=DBALLOC (mgnt_ps->plist_s.cstrMagf_ps, 1))) return (status);
   if (SUCCESS (mgntGloblist("CSTR", "1   ", "MAGF", 
						   (dbheader_ts **const) &(mgnt_ps->plist_s.cstrMagf_ps)))) {
	 /* we have this triplet, so alloc space for the Data list */
	 if (!SUCCESS(status=DBALLOC(mgnt_ps->dlist_s.cstrMagf_ps,1))) return (status);		
   }  
 
   DEBUGPRINT(DP_DEBUG, mgntGlobFlag,
			  ("mgntGlobInitLongLists: Done initializing hsta, stat, bdes, cstr magf db lists\n")); 
   return (status);
}

/*=============================================================================
  Name: mgntGlobList

  Abs: For the given database terms return the corresponding list of
      "pointers" These pointers may be passed to a subsequent routine
      (see for example "dblget") to retrieve quickly the database
      information they point to.

  Args: Type		    Name	Access	   Description
	------------------- -----------	---------- ----------------------------
	const char * const   prim_a        R         see below
	const char * const   unit_a        R         see below
	const char * const   secn_a        R         see below
        dbheader_ts * *const dblist_pps   R/W       see below

  Args: prim_a - Null term pointer to four ASCII characters.
               These four characters define the database Primary whose
               pointers will be returned.
        unit_a -  Null term pointer to char containing 4 char ASCII unit number.
               If the four characters are "ALL*" then the caller
               has requested that the units be "wildcarded".
        secn_a - Pointer to  four ASCII characters.
               These four characters define the database Secondary whose
               pointers will be returned.

	dblist_pps - ptr to ptr to data list which contains dict hash entries (ptrs)

  Rem:  

  Side: Memory allocation.

  Ret:  vmsstat_t:
  
==============================================================================*/
static vmsstat_t mgntGloblist (const char * const prim_a,  
							 const char * const unit_a,  
							 const char * const secn_a,  
							 dbheader_ts **const dblist_pps)
{
  vmsstat_t iss =DB_OKOK; 
  /*char prim_ac[5];
	char unit_ac[5];*/

  /*memset (&prim_ac[0],0,5);
  memset (&unit_ac[0],0,5);
  strncpy (&prim_ac[0], prim_a, 4);
  strncpy (&unit_ac[0], unit_a, 4);*/
  /*  if (!SUCCESS (iss = dblist(&prim_ac[0],
							 &unit_ac[0],
							 secn_a,
							 dblist_pps))) {*/
  if (!SUCCESS (iss = dblist(prim_a,
							 unit_a,
							 secn_a,
							 (const dbheader_ts **const) dblist_pps))) {
	DBFREE (*dblist_pps);
	*dblist_pps= NULL;
    DEBUGPRINT(DP_INFO, mgntGlobFlag,("mgntGloblist: %s %s %.4s Not Found.\n", 
									prim_a, unit_a, secn_a));	
  } /* else (* dblist_pps)->cur = 0; */
  return iss;
}
/*=============================================================================

  Name: mgntGlobExit

  Abs:  Free dblist memory used in magnet utilities

  Args: None

  Rem: The purpose of this function is deallocate memory for any 
       database lists used in magnet utilities..

  Side: None

  Ret:  vmsstat_t
             MGNT_OKOK - Successful operation
             Otherwise, return from dblistfree()
          
==============================================================================*/
vmsstat_t mgntGlobExit(void)
{
   vmsstat_t       status = MGNT_OKOK;  /* status return */
   mgntPrimStruct_ts * prim_ps;
   unsigned short i_prim;

   /* Free db data lists */
   if (mgnt_ps->dlist_s.cstrMagf_ps)         DBFREE(mgnt_ps->dlist_s.cstrMagf_ps);
   if (mgnt_ps->dlist_s.slcAllUnits_ps) DBFREE(mgnt_ps->dlist_s.slcAllUnits_ps);
   if (mgnt_ps->dlist_s.bdes_ps)  DBFREE(mgnt_ps->dlist_s.bdes_ps);
   if (mgnt_ps->dlist_s.hsta_ps)  DBFREE(mgnt_ps->dlist_s.hsta_ps);
   if (mgnt_ps->dlist_s.stat_ps)  DBFREE(mgnt_ps->dlist_s.stat_ps);

   /* Free db ptr lists */
   if (mgnt_ps->plist_s.cstrMagf_ps)         DBFREE(mgnt_ps->plist_s.cstrMagf_ps);
   if (mgnt_ps->plist_s.bdes_ps)  DBFREE(mgnt_ps->plist_s.bdes_ps);
   if (mgnt_ps->plist_s.hsta_ps)  DBFREE(mgnt_ps->plist_s.hsta_ps);
   if (mgnt_ps->plist_s.stat_ps)  DBFREE(mgnt_ps->plist_s.stat_ps);
 

   /* Free db lists used for unit array */
   for (i_prim=0; (i_prim<MGNT_NUM_PRIMS) && SUCCESS(status); i_prim++) {
     prim_ps  = &mgntPrim_s[i_prim];
	 if ((prim_ps->nunits ) && (prim_ps->units_ps)) {
	   status = DBFREE(prim_ps->units_ps);
	 }
   }
   /* free the ca monitors and free lists associated with monPrim_s */
   mgntMonCleanup();  

   return( status );
}
/* debug */
void printStatus(vmsstat_t status) 
{
  if (SUCCESS(status)){
    DEBUGPRINT(DP_DEBUG, mgntGlobFlag,("printStatus: status is OK; = 0x%lx\n", status ));
  }
  else {
	DEBUGPRINT(DP_DEBUG, mgntGlobFlag,("printStatus: status is BAD; = 0x%lx\n", status ));
  }
}
