/*=============================================================================
 
  Name: mgntHdlr.c
        mgntHdlrThread     - Main routine for magnet handler thread
		mgntHdlrInit       - Initializes all globals common to Hdlr and mgntAsync
		mgntHdlrGetUnits   - Gets units specified in slc message
		mgntHdlrPrimToEnum - Gets prim enum from ascii prim contained in slc message
		mgntHdlrSetMagf    - Update CSTR.1.MAGF w/ function being processed by
                             the magnet handler thread.
		mgntHdlrProcess    - Process the prim/units requested in scp message
		mgntHdlrUpdateBDES - update local BDES from slc db and check for changes
		mgntHdlrUpdateHSTA - update local HSTA from slc db and check for changes
		mgntHdlrWaitFuncDone - wait on event flag for mgntAsync to signal for
		                       any scp request
		mgntHdlrFuncDone   - "wake up" from event and finish up request
		mgntHdlrCleanup    - cleanup globals upon thread exit

       reference document:
http://www.slac.stanford.edu/grp/lcls/controls/global/facilities/slc_ioc/requirments/mgnt_designSpec.html

  Auth: 10-Jul-2005, K.  Luchini      (LUCHINI)
  Rev:  dd-mmm-yyyy, Debbie Rogind    (DROGIND)

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  
        30-Jan-2008, K. Luchini (LUCHINI):
	   In mgntHdlrGetUnits return error if unit not found or
           first unit or last unit not found in range. Later add
           code to find next valid first unit in range or first valid
           last unit in range.
           In mgntHdlrProcess, given a primary list, if no units are found
           for a primary, skip to next primary in list.
        23-Jan-2008, K. Luchini (LUCHINI):
           change ALL* unit check 
        24-jul-2007, dfairley
		   removed all references to mgntMonTypes.h
        04-Oct-2005, K. Luchini (LUCHINI):
           call messageQDestroy() in mgntHdlrCleanup()
        27-Sep-2005, K. Luchini (LUCHINI):
           added call to mgntGlobExit() to mgntHdlrCleanup()

=============================================================================*/

/* C system header files */
#include <string.h>        /* memcpy                           */
#include <stdio.h>
#include <stdlib.h>            /* malloc()  */

/* epics libCom header files */
#include     "freeList.h"

/* Debug header files */
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int mgntHdlrFlag = DP_ERROR;
#endif

#include "epicsTime.h"     /* epicsTimeGetCurrent()            */
#include "epicsTypes.h"    /* epicsBoolean                     */

#include <dbAccess.h>      /* Epics Runtime Database               */
#include <dbAccessDefs.h>  /* dbPutField()                         */
#include <dbFldTypes.h>    /* DBR_STRING                           */

/* SLC header files */
#include "slcTypes.h"      /* vmsstat_t                         */
#include "slcGlob.h"       /* slcThreads_as                     */
#include "slcCmlog.h"      /* slcCmlogLogMsg()                  */
#include "slcUtil.h"       /* slcCleanupThread(),               */
			               /* slcGetJobEnumByName()             */
#include "cvtUtil.h"       /* cvtLSwap(),                       */
#include "micrdef.hc"      /* MICR_MALLOC                       */
#include "msgfunc.hc"      /* for TERSE                         */

/* Dataabse header files */
#include "dbupdate.h"      /* dbupdate()                        */
#include "dbl.h"           /* dblget()                          */
#include "dbget.h"         /* dbgetUnitAsString()               */

/* Message and miscellaneous header files */
#include "msgTypes.h"      /* msgmail_ts etc                    */
#include "msgQUtil.h"      /* msgQMsgProcess(),messageQDestroy()*/
#include "sysUtil.h"       /* Find_index_c()                    */

/* Magnet header files */
#include "mgntdef.hc"      /* Magnet message codes               */
#include "mgntstat.h"      /*  HSTA_FBKCONTROL                   */
#include "mgntdefMsg.h"    /* Magnet message code strings        */
#include "mgntTypes.h"     /* mgntHdlrInfo_ts,etc                */
#include "mgntMacros.h"    /* MGNT_SCP_FUNC                      */
#include "mgntGlob.h"      /* mgntGlobExit(),mgntGlobSetMagf()   */
#include "mgntAsync.h"     /* mgntAsyncPVToSlcFLoat, UShort      */
#include "mgntHdlr.h"      /* mgntHdlrCleanup()                  */

/* local prototypes */
/* LOCAL vmsstat_t  mgntHdlrInit( void );*/

/* Get the unit list for the specified primary. */
LOCAL vmsstat_t mgntHdlrGetUnits( unsigned long         iprim,
                       unsigned long * const firstUnit_p,
                       unsigned long         lastUnit,
                       unsigned long * const iunit_p,
                       unsigned long * const nunits_p );
/* 
** Get the primary enum  for the specified primary name.
** A wildcard is permitted for the primary name, in which 
** case a list of primary enums will be returned and the
** argument nprims_p will be set to the number of items
** in the list.
*/
LOCAL vmsstat_t mgntHdlrPrimToEnum( unsigned long * const prim_p,
                            unsigned long * const iprim_p,
                            unsigned long * const nprims_p );


/* Set the CSTR.1.MAGF db secondary with the current
** magnet function being processed. 
** Note: Only those magnet that take a long time to
** complete are set in the MAGF secondary. The reason that
** this is done is that only one magnet requrest can take 
** place at a time. So if the request is time consuming the 
** micro hangs up and this is a way that the operator can tell
** if the micro is busy performing a task. This is a 
*/
LOCAL vmsstat_t mgntHdlrSetMagf(unsigned short func);
LOCAL vmsstat_t mgntHdlrClearMagf( unsigned short func ); 
LOCAL vmsstat_t mgntHdlrProcess (const unsigned short func,
								 const unsigned long  timeout,
								 const msgmail_ts * rMsg_ps);
LOCAL vmsstat_t mgntHdlrProcessUnit (const unsigned long nunits,
									 mgntUnitStruct_ts * unit_ps,
									 const unsigned short func );
LOCAL vmsstat_t mgntHdlrUpdateBDES (mgntUnitStruct_ts * unit_ps);
LOCAL vmsstat_t mgntHdlrUpdateHSTA (mgntUnitStruct_ts * unit_ps);
LOCAL vmsstat_t mgntHdlrWaitFuncDone (const unsigned long  timeout, const unsigned short func);
LOCAL vmsstat_t mgntHdlrFuncDone (void);
LOCAL int       mgntHdlrCleanup(void);
/*
** Local datbase lists and variables
*/ 

static float          *prev_bdes_a = NULL;
static unsigned short *prev_hsta_a = NULL;
static unsigned short *prev_stat_a = NULL;

/* 
** Global variables (declared in mgntGlob)
*/
extern mgntHdlrInfo_ts         mgntInfo_s;
extern mgntHdlrInfo_ts        *mgnt_ps;
extern mgntPrimStruct_ts       mgntPrim_s[MGNT_NUM_PRIMS];
/*extern monPrimStruct_ts        monPrim_s[MGNT_NUM_PRIMS];*/
extern double     funcTrimTimeout;
extern double     funcCalbTimeout;
extern double     funcDacZeroTimeout;
extern double     funcPerturbTimeout;
extern double     funcCheckTimeout;
extern double     funcStdzTimeout;

/*=============================================================================

  Name: mgntHdlrInit

  Abs:  Initialization for this thread; also calls the mgntDbInit() proc to
        Init globals shared between mgntHdlr and mgntAsync

  Args: none

  Rem: 

  Side: 

  Ret:  None

==============================================================================*/
vmsstat_t  mgntHdlrInit( void )
{
  vmsstat_t        status      = MGNT_OKOK;
  static const int maxMsgs     = MAX_QUEUED_MSGS;
  static const int maxMsgBytes = MAX_MSGPTR_BYTES;


  /* Initialization magnet information structure */
  memset((void *)mgnt_ps,0,sizeof(mgntHdlrInfo_ts));
  memset(&mgntPrim_s,0,sizeof(mgntPrim_s));
  mgnt_ps->diag_s.init = epicsFalse;
  /* create event flag that mgntHdlr waits on */
  /*  mgntAsync sets the event after receiving monitor callback msgs */
  /*  for all units requested */ 
  mgnt_ps->func_s.funcDoneEvent_ps = epicsEventMustCreate(epicsEventEmpty);
  mgnt_ps->func_s.busyCounterMutex_ps = epicsMutexMustCreate();
  /* Create the message queue id */
  status = messageQCreate(mgntHdlr,maxMsgs,maxMsgBytes);

  if (SUCCESS(status)) {
    status = mgntGlobInit(); /* get db lists */
    if (SUCCESS(status))
       mgnt_ps->diag_s.init = epicsTrue;
  }
  return( SUCCESS(status) ? MGNT_OKOK : status );
}


/*=============================================================================

  Name: mgntHdlrThread

  Abs:  The magnet handler thread

  Args: none

  Rem: This thread accepts and acts on FUNC_TEST messages 
       from the SLC control system

  Side: 
    * Waits on its queue for a message from Alpha or slcExec.
	* If slcExec says stop, cleanup and stop.
	* Check if any HSTA or DES value has changed in the SLC database and
	  write new values to EPICS HSTACTRL and DES records.
	* Parse message from Alpha to get list of devices (prim/units).
	* Update CSTR MAGF if applicable.
	* If not Alpha check function:
	  - update shared memory with magAsync if necessary.  Reset event flag.
	  - for each prim/unit, check EPICS CTRL field.  If not IDLE, then
	    set bad status for this unit.  if EPICS CTRL is IDLE, set CTRL 
	    to desired action (TRIM, STDZ, etc) based on function code.
	  - if at least one CTRL was set, wait for magAsync to set 
	    event flag.
	* If Alpha check function - forward message to mgntAsync
	* For all prim/units, update SLC STAT and ACT values from records in 
	  the EPICS database.
	* For any function, update the Alpha database.  What to do for
	  check function from Paranoia?
	* Based on status information in memory shared with mgntAsync, reply 
	  with status (summary?) back to the Alpha if reply expected.
	* Go back to top.



  Ret:  None

==============================================================================*/
void  mgntHdlrThread(void)
{
  vmsstat_t       status;                   /* status return                  */ 
  unsigned long   update=0;                 /* update SLC database            */ 
  unsigned long   nbytes;                   /* number of bytes                */
  slcJob_te       jobName_e=invalidJob;     /* job name                       */
  epicsTimeStamp  curTime_s;                /* current time stamp             */
  msgmail_ts     *recvMsg_ps  = NULL;       /* incoming mail message          */ 
  slcThread_ts   *mgntHdlr_ps = NULL;       /* pointer to my thread struct    */
  double          secs=0;
  msgptr_ts       msgptr_s;                 /* structure to forward msgs to async */

  /* Initialization */
  status = mgntHdlrInit();

  if (!SUCCESS(status) || !mgntInfo_s.nunits) {
     slcCleanupThread(mgntHdlr,NULL);
     return;
  }

  /* Get job name */
  jobName_e = slcGetJobEnumByName(MGNT_JOB_NAME);

  /* 
  ** Get the current time this handler was initialized
  ** and set to an active state.
  */
  epicsTimeGetCurrent(&mgntInfo_s.diag_s.startTime_s);

  mgntHdlr_ps         = &slcThreads_as[mgntHdlr];
  mgntHdlr_ps->active = epicsTrue;
  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: active\n"));
  
  /*
  ** Wait for an incoming messages, handle it, then send a reply to the
  ** msgSend thread but only after the function has been completed.
  */
  while(mgntHdlr_ps->active) {
	/* Wait for an incoming message from the message queue.*/
	status = msgQMsgRecv(mgntHdlr,&mgntInfo_s.diag_s.lastReqTime_s,(void*)&recvMsg_ps);
	if (SUCCESS(status)) {

	  /* time messae request was started */
	  mgntInfo_s.diag_s.reqStartCnt++;  
	  nbytes = sizeof(epicsTimeStamp);
	  mgntInfo_s.diag_s.processingReq = epicsTrue;
	  memset((char *)&mgntInfo_s.diag_s.lastReqFinishedTime_s,0,nbytes);
	  memcpy((char *)&mgntInfo_s.diag_s.lastReqStartTime_s,(char *)&curTime_s,nbytes);
	  mgntInfo_s.diag_s.curFunc = recvMsg_ps->msgheader.func;
	       
	  /* 
	  ** Process the magnet function request
	  */
	  /*mgnt_ps->func_s.rMsg_ps = recvMsg_ps;*/  /* so mgntHdlrProcess can access */
	  mgnt_ps->func_s.activeFunc = 0;        /* assume no active func */
	  switch(recvMsg_ps->msgheader.func){
	  case IOC_STOP:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received STOP request \n"));
		slcThreads_as[mgntHdlr].active = epicsFalse;  
		break;
	  case MGNT_CALB:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received CALB request \n"));
		status = mgntHdlrProcess(recvMsg_ps->msgheader.func, funcCalbTimeout, recvMsg_ps);
		break;
	  case MGNT_STDZ:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received STDZ request \n"));
		status = mgntHdlrProcess(recvMsg_ps->msgheader.func, funcStdzTimeout, recvMsg_ps);
		break;
	  case MGNT_CHCK:
		/* forward message to mgntAsync */
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
					 ("MGNT_H: Forward MGNT_CHCK msg to mgntAsync\n"));
        msgptr_s.msgbytes = MSG_HDRS_BYTES; 
        msgptr_s.msgptr_p = recvMsg_ps; 
		if (!SUCCESS(messageQSend(mgntAsync, epicsTrue, sizeof(msgptr_s), &msgptr_s)))
		{
		  DEBUGPRINT(DP_ERROR, mgntHdlrFlag, 
					 ("MGNT_H: Failure to forward MGNT_CHCK msg to mgntAsync\n"));
		}
		/*wait on event flag */
		if (mgnt_ps->func_s.funcDoneEvent_ps){
		  DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				  ("MGNT_H: locking funcDoneEvent_ps for MGNT_CHCK\n"));
		  epicsEventMustWait(mgnt_ps->func_s.funcDoneEvent_ps);
		  DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
					 ("MGNT_H: signal recevied from funcDoneEvent_ps\n"));
		  status = msgQSlcIssReply( (const void *) recvMsg_ps, mgnt_ps->func_s.summaryStatus);
		}		
		break;
	  case MGNT_TRIM:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received TRIM request \n"));
		if (SUCCESS( status = mgntHdlrProcess(recvMsg_ps->msgheader.func, funcTrimTimeout, recvMsg_ps))) {
		  status = msgQSlcIssReply( (const void *)recvMsg_ps, mgnt_ps->func_s.summaryStatus);

		  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Sent TRIM iss reply = 0x%lx \n",
											 mgnt_ps->func_s.summaryStatus ));
		  /*printStatus( mgnt_ps->func_s.summaryStatus);*/
		}
		break;
	  case MGNT_PTRB:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received PTRB request \n"));
		if (SUCCESS(status = mgntHdlrProcess(recvMsg_ps->msgheader.func, funcPerturbTimeout, recvMsg_ps))){
		  status = msgQSlcIssReply( (const void *) recvMsg_ps, mgnt_ps->func_s.summaryStatus);
		}																									  
		break;
	  case MGNT_ZERO:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received ZERO request \n"));
		if (SUCCESS(status = mgntHdlrProcess(recvMsg_ps->msgheader.func, funcDacZeroTimeout, recvMsg_ps))) {
		  status = msgQSlcIssReply( (const void *) recvMsg_ps, mgnt_ps->func_s.summaryStatus);
		}
		break;
		/*
	  case MGNT_TOUC:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received TOUC request \n"));
		if (SUCCESS(status = mgntHdlrProcess(recvMsg_ps->msgheader.func, MGNT_TOUC_TIMEOUT, recvMsg_ps))){
		  status = msgQSlcIssReply( (const void *) recvMsg_ps, mgnt_ps->func_s.summaryStatus);
		}
		break;
		case MGNT_DCAL:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received DCAL request \n"));
		status = mgntHdlrProcess(recvMsg_ps->msgheader.func, MGNT_DCAL_TIMEOUT, recvMsg_ps);
		break;
	  case MGNT_DAOC:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received DAOC request \n"));
		status = mgntHdlrProcess(recvMsg_ps->msgheader.func, MGNT_DAOC_TIMEOUT, recvMsg_ps);
		break;
	  case MGNT__ACT:
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Received _ACT request \n"));
		if (SUCCESS(status = mgntHdlrProcess(recvMsg_ps->msgheader.func, MGNT__ACT_TIMEOUT, recvMsg_ps))) {
		  status = msgQSlcIssReply( (const void *) recvMsg_ps, mgnt_ps->func_s.summaryStatus);
		}
		break;
		*/
	  default:
		/* unexpected message */
		DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				   ("MGNT_H: Received unexpected msg func code = %x\n", 
					recvMsg_ps->msgheader.func));
        status = MGNT_UNKFUNC;
		slcCmlogLogMsg(status, "UNKFUNC", "Slc aware magnet IOC does not handle magnet function %d\n", 
                       recvMsg_ps->msgheader.func);
		break;
	  } /* switch */
	  
		/*
		** Determine if the Alpha db should be updated.
		** This is dor for
		** 1) all Alpha requests that do not have the TERSE bit set.
		** 2) periodiacally scheduled async tasks IOC_ASYNC
		** 3) after a BDES or HSTA change
		*/
	  if ( update ) {
		status = dbupdate( jobName_e ); 
		epicsTimeGetCurrent(&mgntInfo_s.diag_s.lastDbupdateTime_s);
		/* }*/ 
		
		/*
		** We have finished processing this function
		** so now update some statisticts.
		*/
		mgntInfo_s.diag_s.processingReq = epicsFalse;
		mgntInfo_s.diag_s.lastFunc = mgntInfo_s.diag_s.curFunc;
		mgntInfo_s.diag_s.curFunc  = 0;
		mgntInfo_s.diag_s.reqFinishedCnt++;
		epicsTimeGetCurrent(&mgntInfo_s.diag_s.lastReqFinishedTime_s);
		secs = epicsTimeDiffInSeconds(&mgntInfo_s.diag_s.lastReqFinishedTime_s,
									  &mgntInfo_s.diag_s.lastReqStartTime_s);
		mgntInfo_s.diag_s.lastDeltaRespTime = secs;
		
		
	  } /* Processing request message found in the queue */
	  
	    /* 
	    ** Release the message receive buffer before
	    ** checking for another message in the queue.
	    */
	  msgQFreeBuffer(sizeof(msgmail_ts),(void *)recvMsg_ps);
	  
	} /* if successful read of message from the queue */
	
      /* check the stop flag just in case */
	if (mgntHdlr_ps->stop) 
	  mgntHdlr_ps->active = epicsFalse;
	
  } /* while active */
  
  /* active not set, time to shut down */
  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: shutdown\n"));
  /* Note -move all threads to this slc utility at one time: */
  /*slcCleanupThread(mgntHdlr,mgntHdlrCleanup);              */
  mgntHdlrCleanup();
  return;
}

/*=============================================================================

  Name: mgntHdlrSetMagf

  Abs:  Update CSTR.1.MAGF in the database and notify the Alpha

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        unsigned short   func          read-only    function code from req msg

  Rem: The purpose of this function is to update the CSTR.1.MAGF 
       with the supplied magnet function and then to notify the Alpha.

	   Question - do we need to reset CSTR.1.MAGF when func is complete?

  Side: A database update is done to reflect the current magnet function
        that is being processed.

  Ret:  vmsstat_t  
             MGNT_OKOK    - Successful operation
             Otherwise, a failure code from the 
                dbsend,dblistalloc,dblist,dblput,dblget

==============================================================================*/
LOCAL vmsstat_t mgntHdlrSetMagf( unsigned short func ) 
{
     vmsstat_t              status=MGNT_OKOK;        /* return status        */
     unsigned short         i_func;                  /* function index       */
     unsigned long          found=0;                 /* found flag           */
     slcJob_te              jobName_e = invalidJob;  /* job enum             */
     static unsigned short *cstrMagf_p=NULL;         /* CSTR.1.MAGF data ptr */
     MGNT_FUNC_TO_ALPHA;

     /*
     ** Is this one of the magnet function that requires 
     ** that the Alpha be notified  which function is currently
     ** being executed on the SLC-Aware IOC? If yes, then
     ** set the function code in by CSTA.1.MAGF
     */
     for (i_func=0; (i_func<MGNT_NUM_FUNCS) && !found ; i_func++ ) { 
      if ((func & !TERSE) == mgntFuncToAlpha_a[i_func]) 
         found=1;
     }
     if (!found) return(status);

	 DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, ("mngtHdlrSetMagf: TERSE bit set..\n"));
     /* Get job name if necessary */
     if ( jobName_e == invalidJob ) 
	   jobName_e = slcGetJobEnumByName(MGNT_JOB_NAME);


	 if (SUCCESS(status)) {
	   DBCLEAR (mgnt_ps->dlist_s.cstrMagf_ps);
	   status = dblget((const dbheader_ts * const)mgnt_ps->plist_s.cstrMagf_ps,
					   (dbheader_ts **const)&(mgnt_ps->dlist_s.cstrMagf_ps));
	   if (SUCCESS(status)) 
		 cstrMagf_p = DBPTR_FIRST(mgnt_ps->dlist_s.cstrMagf_ps);
	 }
      
	 DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, ("mngtHdlrSetMagf: Updating cstr magf\n"));
     /* Update the MAGF secondary of CSTR */
     if ( cstrMagf_p && (*cstrMagf_p!= func)) {
       *cstrMagf_p = func;
        status     = dblput((const dbheader_ts *const) mgnt_ps->plist_s.cstrMagf_ps,
                            (const dbheader_ts * const)mgnt_ps->dlist_s.cstrMagf_ps,jobName_e);

        /* ship this information to the Alpha NOW */
        if (SUCCESS(status)) 
          status = dbupdateNoWait( jobName_e );

		DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, ("mngtHdlrSetMagf: Updating alpha to set MAGF..\n"));
     }
     return( SUCCESS(status) ? MGNT_OKOK : status );
}
/*=============================================================================

  Name: mgntHdlrClearMagf

  Abs:  Update CSTR.1.MAGF in the database and notify the Alpha

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        unsigned short   func          read-only    function code from req msg

  Rem: The purpose of this function is to update the CSTR.1.MAGF
       that the requested magnet function is complete and notify the Alpha.
 	   We need to reset CSTR.1.MAGF when func is complete.

  Side: A database update is done to reflect the current magnet function
        that is being processed.

  Ret:  vmsstat_t  
             MGNT_OKOK    - Successful operation
             Otherwise, a failure code from the 
                dbsend,dblistalloc,dblist,dblput,dblget

==============================================================================*/
LOCAL vmsstat_t mgntHdlrClearMagf( unsigned short func ) 
{
     vmsstat_t              status=MGNT_OKOK;        /* return status        */
     unsigned short         i_func;                  /* function index       */
     unsigned long          found=0;                 /* found flag           */
     slcJob_te              jobName_e = invalidJob;  /* job enum             */
     static unsigned short *cstrMagf_p=NULL;         /* CSTR.1.MAGF data ptr */
     MGNT_FUNC_TO_ALPHA;

     /*
     ** Is this one of the magnet function that requires 
     ** that the Alpha be notified  which function is currently
     ** being executed on the SLC-Aware IOC? If yes, then
     ** set the function code in by CSTA.1.MAGF
     */
     for (i_func=0; (i_func<MGNT_NUM_FUNCS) && !found ; i_func++ ) { 
      if ((func & !TERSE) == mgntFuncToAlpha_a[i_func]) 
         found=1;
     }
     if (!found) return(status);

	 DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, ("mngtHdlrClearMagf: TERSE bit set..\n"));
     /* Get job name if necessary */
     if ( jobName_e == invalidJob ) 
	   jobName_e = slcGetJobEnumByName(MGNT_JOB_NAME);

	 DBCLEAR (mgnt_ps->dlist_s.cstrMagf_ps);
	 cstrMagf_p = DBPTR_FIRST(mgnt_ps->dlist_s.cstrMagf_ps);
	 
	 *cstrMagf_p = 0;
	 status  = dblput((const dbheader_ts *const) mgnt_ps->plist_s.cstrMagf_ps,
					  (const dbheader_ts * const)mgnt_ps->dlist_s.cstrMagf_ps,jobName_e);

	 /* ship this information to the Alpha NOW */
	 if (SUCCESS(status)) 
	   status = dbupdateNoWait( jobName_e );

	 DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, ("mngtHdlrClearMagf: Updating alpha to clear MAGF..\n"));	 
	 return( SUCCESS(status) ? MGNT_OKOK : status );
}

/*=============================================================================

  Name: mgntHdlrProcess

  Abs:  Process a magnet function from a scp message

  Args: Type                Name        Access     Description
        ------------------- ----------- ---------- ----------------------------
        char  *               data_c    read-only  Ascii CTRL function string (10 chars null terminated).
        unsigned long *       timeout   read-only  timeout in milliseconds


  Rem: The purpose of this function is pass determine which magnet
       devices (units/prim) the magnet request is for and then
       to issue the proper EPICS CTRL command to do just that. 
	   Called as a result of receiving MGNT_*** commands in mgntHdlr Q (except MGNT_CHCK).
	   For an ALL* request, if there is a missing CTRL, BDES or HSTA, skip - don't flag bad status
	     also if unit is not idle (CA, or feedback user), skip, no bad status
	   For a specific device, if there is a missing CTRL, BDES or HSTA, skip - is bad status
	     if unit is under feedback control, bad status
	   When finished processing units list, call mgntHdlrWaitFuncDone (func, timeout)
	     to determine if a wait is necessary.
	   
  Side: None

  Ret:  vmsstat_t 
             MGNT_OKOK     - Successful operation 
             MGNT_ABORTED  - Failure, due to aborted function
             MGNT_TIMEDOUT - Faulure, due to timeout

==============================================================================*/
LOCAL vmsstat_t mgntHdlrProcess (const unsigned short func, /* says epics output to CTRL */
								 const unsigned long  timeout,
								 const msgmail_ts * rMsg_ps)

{
  vmsstat_t                status = MGNT_OKOK;
  vmsstat_t                status2 = MGNT_OKOK;
  unsigned long            unitCnt;
  unsigned short           nitems=0;
  unsigned long            nunits=0;
  unsigned long            nprims=0;
  unsigned short           i_item=0;
  unsigned long            i_unit=0;
  unsigned long            i_prim=0;    
  mgntPrimStruct_ts        *prim_ps=NULL;          /* prims struct containing units   */
  mgntUnitStruct_ts        *unit_ps=NULL;
  mgntHdlrReqData_ts       *msgData_as;

 
  /*msgData_as = (mgntHdlrReqData_ts *)(mgnt_ps->func_s.rMsg_ps->reqdata);*/
  msgData_as = (mgntHdlrReqData_ts *)(rMsg_ps->reqdata);
  /* determine how many items (diff primaries) there are in this message */
  /*nitems = 2*(mgnt_ps->func_s.rMsg_ps->msgheader.datalen)/sizeof(mgntHdlrReqData_ts);*/
  nitems = 2*(rMsg_ps->msgheader.datalen)/sizeof(mgntHdlrReqData_ts);
  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("mngtHdlrProcess; datalen = %u, nitems=%u\n",
									 rMsg_ps->msgheader.datalen, nitems)); 
  /* initialize summaryStatus field for reply message */
  mgnt_ps->func_s.summaryStatus = MGNT_OKOK;
  
  /* if this is "ALL*" magnets ..*/
  if ( strncmp((char *)&(msgData_as[0].prim),_WILDCARD, 4)==0 ) {
	DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, ("mngtHdlrProcess: ALL* magnets!!\n"));
	for (i_item=0;SUCCESS(status) && (i_item < MGNT_NUM_PRIMS); i_item++) {
	  prim_ps = &mgntPrim_s[i_item];
	  if (prim_ps->nunits > 0) {
		unit_ps = DBPTR_FIRST(prim_ps->units_ps);
		for ( unitCnt=0; (unitCnt<prim_ps->nunits) && unit_ps ; unitCnt++, unit_ps++ ) {
		  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("mngtHdlrProcess: ALL* magnets; prim %d, ptr=%p\n",
											 i_item, unit_ps));
		  mgntHdlrProcessUnit(9999, unit_ps, func);
	   } /* for all units */	   
	  } /* if this primary is present */
	} /* for all primarys */
  } /* if ALL* magnets */
  
  else {
	/* discrete list of prims/units from scp */
	for (i_item=0;i_item<nitems; i_item++) {
	  DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				 ("mngtHdlrProcess: Got %.4s \n",(char *) &(msgData_as[i_item].prim )));
	  
	  /* convert firstUnit, lastUnit inplace; it is number format */
	  /* Note for "ALL*" units, a range from 1 - 9984 is sent */
	  cvtLSwap((const unsigned long *)&(msgData_as[i_item].firstUnit));
	  cvtLSwap((const unsigned long *)&(msgData_as[i_item].lastUnit));
	  /* first and last units are integers */
	  DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
				 ("mngtHdlrProcess: i_prim=%ld, [%d]firstUnit=0x%.8lx, [%d]lastUnit=0x%.8lx; msgData fu ptr=%p \n",
				  i_prim, i_item, msgData_as[i_item].firstUnit,
				  i_item, msgData_as[i_item].lastUnit, &(msgData_as[i_item].firstUnit)));
	  
	  if (SUCCESS(status2 = mgntHdlrPrimToEnum( &(msgData_as[i_item].prim),&i_prim,&nprims ))) {
		if (SUCCESS (status2 = mgntHdlrGetUnits( i_prim,
							&(msgData_as[i_item].firstUnit),
							 msgData_as[i_item].lastUnit,
						        &i_unit,
						        &nunits ))) 
		  {
			prim_ps = &mgntPrim_s[i_prim];
			unit_ps = DBPTR(prim_ps->units_ps, i_unit);
			for ( unitCnt=0; (unitCnt<nunits)&& unit_ps ; unitCnt++, unit_ps++ ) {
			  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("mngtHdlrProcess calling ProcessUnits, nunits=%ld, busyCounter=%ld\n", nunits, mgnt_ps->func_s.busyCounter ));
			  mgntHdlrProcessUnit(nunits, unit_ps, func);
			} /* for all units */		 
			
		  } /* if success */
	  } /*if success */
          /* return maximum severity */
          if (status2>status) status=status2;
	} /* End of i_item FOR loop */
  }/* discrete list of prim/units */
  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("mngtHdlrProcess: busyCounter = %ld\n",mgnt_ps->func_s.busyCounter));   
  status = mgntHdlrWaitFuncDone(timeout, func);
  
  return( status );
}
/*=============================================================================

  Name: mgntHdlrProcessUnit

  Abs:  Process a magnet function from scp request message for a particular unit

  Args: Type                Name        Access     Description
        ------------------- ----------- ---------- ----------------------------
		unsigned long       nunits      read       # units in this request
		mgntUnitStruct_ts   * unit_ps   read/write unit


  Rem:
	   For an ALL* request, if there is a missing CTRL, BDES or HSTA, skip - don't flag bad status
	     also if unit is not idle (CA, or feedback user), skip, no bad status
	   For a specific device, if there is a missing CTRL, BDES or HSTA, skip - is bad status
	     if unit is under feedback control, bad status

	   
  Side: None

  Ret:  vmsstat_t 
             MGNT_OKOK     - Successful operation 
             MGNT_ABORTED  - Failure, due to aborted function
             MGNT_TIMEDOUT - Faulure, due to timeout

==============================================================================*/
LOCAL vmsstat_t mgntHdlrProcessUnit (const unsigned long nunits,
									 mgntUnitStruct_ts * unit_ps,
									 const unsigned short func )
									 
{
  vmsstat_t      status = MGNT_OKOK;
  long           nreqs=1;      /* these are for fetching Epics PV's */
  long           options=0;
  unsigned short ctrlFunc=0;

/* DEBUG : with no db, test timeout */
  /*unit_ps->scp = epicsTrue;*/
	/*unit_ps->ireq = i_item;*/
	/*mgnt_ps->func_s.busyCounter++;
	status = mgntHdlrUpdateBDES(unit_ps);
	status = mgntHdlrUpdateHSTA(unit_ps);
	*/
/* END OF DEBUG */  

  /* read CTRL and check for ! IDLE instead of checking busy flag */
  /* if CTRL doesn't exist, can't control it - so skip it */
  if (!unit_ps->ctrlAddr_s.precord){
	if (nunits<=1) {  /* if request wants specific unit, then this is an error */
	  mgnt_ps->func_s.summaryStatus = max(MGNT_NOUNITS, mgnt_ps->func_s.summaryStatus); /* error */
	  DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				 ("mngtHdlrProcessUnit: No CTRL record for req'ed unit %.4s %.4s\n", 
				  unit_ps->prim_pc,(char *) &(unit_ps->unit_ac[0])));
	  slcCmlogLogMsg(MGNT_NOUNITS, "MGNT_NOUNITS", "%s function requested for %.4s %.4s and PV doesn't exist \n", 
				   epicsThreadGetNameSelf(),unit_ps->prim_pc ,(char *) &(unit_ps->unit_ac[0]) );
	}
	return status;
  }
  /* we have CTRL record */
  if (dbGetField(&(unit_ps->ctrlAddr_s),  DBR_USHORT, 
				 &(unit_ps->ctrlVal), &options, &nreqs, NULL)) {
	mgnt_ps->func_s.summaryStatus = max(MGNT_NOUNITS, mgnt_ps->func_s.summaryStatus); /* error reading CTRL */
	return status;
  } 
  /* successful CTRL PV read */
  /* note that devices under fbck have CTRL=PTRB */
  if (unit_ps->ctrlVal != CTRL_IDLE ){
	if (nunits <= 1) 
	  mgnt_ps->func_s.summaryStatus = max(MGNT_NOUNITS, mgnt_ps->func_s.summaryStatus); /* new "busy" status */
	/*device is already busy - if SLC and EPICS have different BDES or HSTA values, */
	/* update SLC BDES and HSTA from EPICS */
	/*  occurs for CTRL=IDLE callbacks */
	/* BDES - Desired B field */
	mgntAsyncPVToSlcFloat( &(unit_ps->bdesAddr_s),
							 (const dbheader_ts *) unit_ps->bdesPlist_ps, 
							 (dbheader_ts **const) &(unit_ps->bdesDlist_ps));
	
	/* HSTA */
	mgntAsyncPVToSlcUShort( &(unit_ps->hstaAddr_s),
							(const dbheader_ts *) unit_ps->hstaPlist_ps, 
							(dbheader_ts **const) &(unit_ps->hstaDlist_ps), unit_ps);
	return status;
  }
  /* has CTRL, & unit is IDLE */  
  /* see if it has HSTA */
  if (unit_ps->hstaAddr_s.precord) {
	/* see if unit is under feedback control, as determined by HSTA bit */
	if ( dbGetField( &(unit_ps->hstaAddr_s), DBR_USHORT, (void *)&(unit_ps->hstaVal),
					 &options, &nreqs, NULL) ){
	  mgnt_ps->func_s.summaryStatus =  max(MGNT_NOUNITS, mgnt_ps->func_s.summaryStatus); /* error reading HSTA */
	  return status;
	} 
	
	/* if under feedback control, don't include */
	if (unit_ps->hstaVal & HSTA_FBKCONTROL) {
	  if (nunits <=1)
		mgnt_ps->func_s.summaryStatus =  max(MGNT_NOUNITS, mgnt_ps->func_s.summaryStatus); /* busy - feedback ctrl status */
	  return status;
	}
	
	/* check to see if it has BDES record */
	if (!unit_ps->bdesAddr_s.precord) {
	  if (nunits<=1)
		mgnt_ps->func_s.summaryStatus = max(MGNT_NOUNITS, mgnt_ps->func_s.summaryStatus); /* error */
	  DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				 ("mngtHdlrProcess: No BDES record for req'ed unit %.4s %.4s\n", 
				  unit_ps->prim_pc,(char *) &(unit_ps->unit_ac[0])));
	  return status;
	}
	
	/* unit is IDLE, not under feedback control, and has CTRL, BDES, HSTA records  */
	/* set up unit level and function global flags/counters */
	/* unit_ps->busy = epicsTrue; */ /* callback Hdlr sets */
	if (unit_ps->ctrlAddr_s.precord) {

	  /*unit_ps->ireq = i_item;*/
	  
	  /*
	  ** Before starting ANY user request 
	  ** secondaries with the lastet values
	  ** from the SLC db and check to see if
	  ** there have been any chagnes since we
	  ** were last here. If so, then udpate the
	  ** appropriate EPICS pv's.
	  */

	  status = mgntHdlrUpdateBDES(unit_ps);
	  status = mgntHdlrUpdateHSTA(unit_ps);
	
	  /* output  mgnt_ps->func_s.activeFunc function to epics */
	  switch  (func) {
	  case  MGNT_CALB : 
		ctrlFunc = FUNC_CALB;
		break;
	  case  MGNT_STDZ:  
		ctrlFunc = FUNC_STDZ;
		break;
	  case  MGNT_TRIM:  
		ctrlFunc = FUNC_TRIM;
		break;
	  case  MGNT_PTRB:  
		ctrlFunc = FUNC_PERTURB;
		break;
	  case MGNT_ZERO:   
		ctrlFunc = FUNC_DAC_ZERO;
		break;
	  case MGNT_RSET:   
		ctrlFunc = FUNC_RESET;	
		break;
	  default: break;
	  } /* switch */
	  /* incr busyCounter before output to CTRL in case callback->mgntAsync decrements before
         it is even incremented here */
	  /* put mutex around the increment */
	  if (mgnt_ps->func_s.busyCounterMutex_ps)		 
		epicsMutexMustLock(mgnt_ps->func_s.busyCounterMutex_ps);
	  unit_ps->scp = epicsTrue;
  	  mgnt_ps->func_s.busyCounter++;
	  if (mgnt_ps->func_s.busyCounterMutex_ps)	
		epicsMutexUnlock(mgnt_ps->func_s.busyCounterMutex_ps);
	
	  if (dbPutField(&(unit_ps->ctrlSetptAddr_s),DBR_USHORT,&ctrlFunc,nreqs )){
		if (mgnt_ps->func_s.busyCounterMutex_ps)		 
		  epicsMutexMustLock (mgnt_ps->func_s.busyCounterMutex_ps);
		unit_ps->scp = epicsFalse;
		mgnt_ps->func_s.busyCounter--;
		if (mgnt_ps->func_s.busyCounterMutex_ps)	
		  epicsMutexUnlock(mgnt_ps->func_s.busyCounterMutex_ps);		
		DEBUGPRINT(DP_ERROR, mgntHdlrFlag,("mngtHdlrProcessUnit: dbPutField error \n"));
		return (MGNT_DBERR);
	  }
	  DEBUGPRINT(DP_INFO, mgntHdlrFlag,("mngtHdlrProcessUnit: dbPutField to %s ctrlFunc=%d busyCounter =%ld\n",
						 (&(unit_ps->ctrlSetptAddr_s))->precord->name,ctrlFunc,mgnt_ps->func_s.busyCounter));


	}
	else  DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				   ("mngtHdlrProcess: No CTRL record for req'ed unit %.4s %.4s\n", 
					unit_ps->prim_pc,(char *) &(unit_ps->unit_ac[0]))); 

	
  } /* unit has a hsta record */
  else  DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				   ("mngtHdlrProcess: No HSTA record for req'ed unit %.4s %.4s\n", 
					unit_ps->prim_pc,(char *) &(unit_ps->unit_ac[0])));
  return status;
}
/*=============================================================================

  Name: mgntHdlrUpdateBDES

  Abs:  Set EPICS BDES value from slc value, if different, when CTRL=IDLE.
        When CTRL!=IDLE, set slc BDES value from EPICS.
        

  Args: Type               Name          Access        Description
        -----------        -----------   ----------    ----------------------------
        mgntUnitStruct_ts  *unit_ps     read-only ptr  ptr to unit to which
		                                               its BDES val needs to be synced
		
		Rem: 

  
  Side: Called on behalf of a unit either involved in the current scp function being
        setup, or by the Check function. 
		Before starting ANY SCP request, for every device included in the request
        that has CTRL set to IDLE, all BDES (and HSTA) values from SLC different
		from EPICS are written to EPICS. For every device included in the request
		that has CTRL != IDLE, all BDES (and HSTA) values from EPICS different 
		than SLC are written to SLC.
		
  Ret:  vmsstat_t  
        MGNT_OKOK    - Successful operation
		
==============================================================================*/
LOCAL vmsstat_t mgntHdlrUpdateBDES (mgntUnitStruct_ts * unit_ps) {
  vmsstat_t          status = MGNT_OKOK;
  unsigned char      *db_ps;
  long               nreqs=1;              /* number of requests           */
  long               options=0;
  

  /* Get the current BDES from SLC (current EPICS bdes set via monitor */
  DBCLEAR(unit_ps->bdesDlist_ps);
  status = dblget( (const dbheader_ts * const)unit_ps->bdesPlist_ps,
                   (dbheader_ts **const)& (unit_ps->bdesDlist_ps) );
  db_ps = (unsigned char *) (unit_ps->bdesDlist_ps);
  db_ps += sizeof(dbheader_ts);

  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: mgntHdlrUpdateBDES slc bdes=%f\n",*(float *)(db_ps)));					   
  /* see if epics and slc differ */
  if ( (*(float *)(db_ps)) != unit_ps->bdesVal ) {
	/* epics and slc differ, so.... */
	if (unit_ps->ctrlAddr_s.precord) {
	  /* we have CTRL record */
	  if (dbGetField(&(unit_ps->ctrlAddr_s),  DBR_USHORT, 
				 &(unit_ps->ctrlVal), &options, &nreqs, NULL)) {
		/* error reading CTRL */
		return status = MGNT_DBERR;
	  } 
	  if (unit_ps->ctrlVal == CTRL_IDLE){
		/* if CTRL=IDLE, update EPICS from SLC */
		if (unit_ps->hstaAddr_s.precord && (!(unit_ps->hstaVal & HSTA_FBKCONTROL)) ) {
		  /* update only if unit is not under feedback control */		  
		  memcpy (&(unit_ps->bdesVal), db_ps, sizeof(float));
		  if ( dbPutField( &(unit_ps->bdesAddr_s), DBR_FLOAT, (void *)&(unit_ps->bdesVal),nreqs) )
			return (MGNT_DBERR);	
		}
	  }
	  else { /* not IDLE */
		/* update SLC from EPICS */
		memcpy (&db_ps, &(unit_ps->bdesVal), sizeof(float));
		/* let mgntAsync 1 sec processing do the update */
		mgnt_ps->doUpdate=epicsTrue;		
	  }
	}
  } /* epics and slc differ */
  return status;
}
/*=============================================================================

  Name: mgntHdlrUpdateHSTA

  Abs:  Set EPICS HSTA value from slc value, if different, when CTRL=IDLE.
        When CTRL!=IDLE, set slc HSTA value from EPICS.
        

  Args: Type               Name          Access        Description
        -----------        -----------   ----------    ----------------------------
        mgntUnitStruct_ts  *unit_ps     read-only ptr  ptr to unit to which
		                                               its HSTA val needs to be synced
		
		Rem: 

  
  Side: Called on behalf of a unit either involved in the current scp function being
        setup, or by the Check function. 
		Before starting ANY SCP request, for every device included in the request
        that has CTRL set to IDLE, all HSTA (and HSTA) values from SLC different
		from EPICS are written to EPICS. For every device included in the request
		that has CTRL != IDLE, all HSTA (and HSTA) values from EPICS different 
		than SLC are written to SLC.
		
  Ret:  vmsstat_t  
        MGNT_OKOK    - Successful operation
		
==============================================================================*/
LOCAL vmsstat_t mgntHdlrUpdateHSTA (mgntUnitStruct_ts * unit_ps) {
  vmsstat_t          status = MGNT_OKOK;
  unsigned char      *db_ps;
  long               nreqs=1;              /* number of requests           */
  long               options=0;

  /* Get the current HSTA from SLC (current EPICS hsta set via monitor */
  DBCLEAR (unit_ps->hstaDlist_ps);
  status = dblget( (const dbheader_ts * const)unit_ps->hstaPlist_ps,
                   (dbheader_ts **const)& (unit_ps->hstaDlist_ps) );
  db_ps = (unsigned char *) (unit_ps->hstaDlist_ps);
  db_ps += sizeof(dbheader_ts);

  /* CTRL just read for this unit */
  /* note that devices under fbck have CTRL=PTRB */
	
  /* see if epics and slc differ */
  if ( (*(unsigned short *)(db_ps)) != unit_ps->hstaVal ) {
	/* epics and slc differ, so.... */
	if (unit_ps->ctrlAddr_s.precord) {
	  /* we have CTRL record */
	  if (dbGetField(&(unit_ps->ctrlAddr_s),  DBR_USHORT, 
				 &(unit_ps->ctrlVal), &options, &nreqs, NULL)) {
		/* error reading CTRL */
		return status = MGNT_DBERR;
	  } 
	  if (unit_ps->ctrlVal == CTRL_IDLE ){
		/* if CTRL=IDLE, update EPICS HSTA from SLC */
		if (unit_ps->hstaAddr_s.precord && (!(unit_ps->hstaVal & HSTA_FBKCONTROL)) ) {
		  /* update only if unit is not under feedback control */
		  memcpy (&(unit_ps->hstaVal), db_ps, sizeof(unsigned short));
		  if (unit_ps->hstaCtrlAddr_s.precord){
			if ( dbPutField( &(unit_ps->hstaCtrlAddr_s), DBR_USHORT, (void *)&(unit_ps->hstaVal),nreqs) )
			  return (MGNT_DBERR);	
		  }
		}
	  }
	  else { /* not IDLE */
		/* update SLC from EPICS */
		memcpy (&db_ps, &(unit_ps->hstaVal), sizeof(unsigned short));
		/* let mgntAsync 1 sec processing do the update */
		mgnt_ps->doUpdate=epicsTrue;
		
	  }
	}
  } /* epics and slc differ */
  return status;
}
/*=============================================================================

  Name: mgntHdlrWaitFuncDone

  Abs:  Wait on the mgnt_ps->func_s.funcDoneEvent_ps event flag shared between
        mgntAsync and mgntHdlr for mgntAsync to signal. mgntAsync signals when
		it receives all callbacks for all devices involved in current scp
		function, or upon timeout or abort.

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        unsigned short   func          read-only    function code from req msg

  Rem: 


  Side: A database update is done to reflect the current magnet function
        that is being processed.

  Ret:  vmsstat_t  
        MGNT_OKOK    - Successful operation

==============================================================================*/
LOCAL vmsstat_t mgntHdlrWaitFuncDone (const unsigned long  timeout, const unsigned short func ) {
  vmsstat_t status = MGNT_OKOK;

   /* check to see if we should wait at the event flag */
   if (mgnt_ps->func_s.busyCounter > 0) {
	 /*
	 ** Determine if the Alpha needs to be notifed
	 ** that a time consuming function is about to
	 ** begin processing. If yes, then go ahead
	 ** and notify the Alpha via CSTR.1.MAGF
	 **  see MGNT_FUNC_TO_ALPHA macro
	 */
	 mgnt_ps->func_s.activeFunc = func;
	 status = mgntHdlrSetMagf(mgnt_ps->func_s.activeFunc);
	 
	 /* first set function timeout Coutner */
	 mgnt_ps->func_s.timeoutCounter = timeout;
	 DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("mngtHdlrProcess: Waiting on eventflag; busyCounter = %ld\n", 
										mgnt_ps->func_s.busyCounter));					   
	 /*wait on event flag */
	 if (mgnt_ps->func_s.funcDoneEvent_ps){
	   /* debug */
	   
	   DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				  ("MGNT_H: waiting on funcDoneEvent_ps .... \n"));
	   epicsEventMustWait(mgnt_ps->func_s.funcDoneEvent_ps);
	   DEBUGPRINT(DP_INFO, mgntHdlrFlag, 
				  ("MGNT_H: signal from funcDoneEvent_ps\n"));
	   
	   /* Done with func - go cleanup */
	   status = mgntHdlrFuncDone();
	   
	 }	 
   }
   return status;
  
}
/*=============================================================================

  Name: mgntHdlrFuncDone

  Abs:  Cleanup after the scp function has completed.

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        unsigned short   func          read-only    function code from req msg

  Rem: 


  Side: A database update is done to reflect the current magnet function
        that is being processed.

  Ret:  vmsstat_t  
        MGNT_OKOK    - Successful operation

==============================================================================*/
LOCAL vmsstat_t mgntHdlrFuncDone (void) {
  vmsstat_t          status = MGNT_OKOK;
  unsigned short     i_prim;               /* primary index        */ 
  unsigned short     i_unit;               /* unit index           */
  mgntPrimStruct_ts  *prim_ps=NULL;        /* pointer to primary unit info */
  mgntUnitStruct_ts  *unit_ps=NULL;

  /* clear MAGF */
  mgntHdlrClearMagf (mgnt_ps->func_s.activeFunc);

  if (mgnt_ps->func_s.busyCounterMutex_ps)		 
	epicsMutexMustLock(mgnt_ps->func_s.busyCounterMutex_ps);
  /* Done with scp request - now clear all scp func related flags (in case of timeout) */
  for (i_prim =0; i_prim<MGNT_NUM_PRIMS; i_prim++ ) {
	prim_ps = &(mgntPrim_s[i_prim]);
	unit_ps = DBPTR_FIRST(prim_ps->units_ps);
	for (i_unit=0; (i_unit<prim_ps->nunits) && SUCCESS(status) && unit_ps; i_unit++, unit_ps++) {
	  if (unit_ps->scp) unit_ps->busy = epicsFalse;
	  unit_ps->scp = epicsFalse;
	  unit_ps->ireq = 0;
	}
  }
  mgnt_ps->func_s.activeFunc = 0;
  mgnt_ps->func_s.timeoutCounter = 0;
  mgnt_ps->func_s.busyCounter = 0;

  if (mgnt_ps->func_s.busyCounterMutex_ps)	
	epicsMutexUnlock(mgnt_ps->func_s.busyCounterMutex_ps);
  
  return status;
}
/*=============================================================================

  Name: mgntHdlrGetUnits

  Abs:  Get unit list requested by incoming magnet message  

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        unsigned long    iprim         read-only    primary index into primary list 
        unsigned long    firstUnit     read-only    first unit number in range
        unsigned long    lastUnit      read-only    last unit number in range
        unsigned long *  iunit         read-write   index into the primary unit list 
        unsigned long *  nunits        read-write   number of units in range 
		                                               of firstUnit to lastUnit

  Rem: The purpose of this function is to decode a single item (triplet) from an
       incoming message into the unit range for the specified primary.A  wildcard
       primary (ie "ALL*") is not permitted in this function. However, a wildcard
       is permitted for the firstUnit argument. If the firstUnit argument is set to
       zero then a wildcard is assumed.

  Side: No wildcard primary names.

  Ret:  vmsstat_t  
             MGNT_OKOK    - Successful operation
             Otherwise, a failure code from

==============================================================================*/
LOCAL vmsstat_t mgntHdlrGetUnits( unsigned long         iprim,
                       unsigned long * const firstUnit_p,
                       unsigned long         lastUnit,
                       unsigned long * const iunit_p,
                       unsigned long * const nunits_p )
{
  vmsstat_t           status = MGNT_NOTMAGPRIM; /* return status                     */
  long                i_first;        /* index into unit list for first unit # */
  long                i_last;         /* index to unit list for last unit #    */
  unsigned long       nelm;           /* number of elements                    */
  unsigned long       keysiz;         /* size of key                           */
  unsigned long       keyoff = 0;     /* byte offset to item of interest       */

  unsigned char       firstUnit_ca[4]; /* first unit number as a 4-char string  */
  unsigned char       lastUnit_ca[4];  /* last unit number as a 4-char string   */
  mgntPrimStruct_ts   *prim_ps = NULL;
  mgntUnitStruct_ts   *unit_ps = NULL;
  unsigned long       *db_p;
  int i;
  unsigned long       *temp_p;        /* for debugprint message */


  /* debug */
  /*
	db_p = DBPTR(mgnt_ps->dlist_s.slcAllUnits_ps, 0);
	DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
			   ("These are all units in this ioc:\n")); 
  for (i=0; i<DBCNT(mgnt_ps->dlist_s.slcAllUnits_ps); i++, db_p++) {
	DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
			   ("mgntHdlrGetUnits: Unit %d = %.4s; ptr=%p\n",i,(char *)db_p, db_p)); 	  
  } 
  */
  *nunits_p = 0;   /* set number of units in range to zero */ 

  /* check for failure conditions */
  if ( iprim > sext_e ) return(status);
  if ( mgntPrim_s[iprim].nunits==0 ) {
	  DEBUGPRINT(DP_ERROR, mgntHdlrFlag, 
				 ("mgntHdlrGetUnits: no units associated with requ'ed iprim %lu\n", iprim ));
	return(status);}

  prim_ps = &mgntPrim_s[iprim];

  /* check for request for "ALL*" units for this primary */
  if ( (strncmp((char *)firstUnit_p,_WILDCARD, 4)==0) || 
       ((*firstUnit_p==0) && (lastUnit>900))          ||
       ((*firstUnit_p==1) && (lastUnit>900)) ) {
      DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
			    ("mgntHdlrGetUnits: request for ALL* units, iprim %lu\n", iprim ));
     *iunit_p  = 0;
     *nunits_p = prim_ps->nunits;
      if (prim_ps->nunits > 0) {
 	 unit_ps = DBPTR_FIRST(prim_ps->units_ps);
        *iunit_p = unit_ps->unitIndex;
          DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
	   	               ("mgntHdlrGetUnits: request for a ALL* units, iprim %lu (firstUnit=%ld nunits=%ld)\n", 
                                 iprim,*firstUnit_p,*nunits_p ));
      }
  }
  /* check for a range of units or a single unit for this primary */
  else {
        DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
	   	                 ("mgntHdlrGetUnits: request for a range of units, iprim %lu\n", iprim ));
       /* 
        ** Convert the first and last unit numbers from numbers
        ** to 4-char unit numbers. We want the unit number to
        ** be left justified.
        */
        dbgetUnitAsString((const unsigned short) (*firstUnit_p), (char * const)&firstUnit_ca);
        dbgetUnitAsString((const unsigned short) lastUnit, (char * const)&lastUnit_ca);
        temp_p = (unsigned long *) &lastUnit_ca[0];
        DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
			     ("mgntHdlrGetUnits: locate first unit %.4s, iprim %lu (lastUnit_ca=0x%.8lx, %.4s)\n", 
			      &firstUnit_ca[0],iprim, *temp_p, &lastUnit_ca[0]  ));

        /* search for the first unit number */
        nelm     = prim_ps->nunits;
        keysiz   = sizeof(unsigned long);

        /*
        ** search the primary unit list for a match. This 
        ** function returns an index into this list.
        */
	db_p = DBPTR(mgnt_ps->dlist_s.slcAllUnits_ps, prim_ps->slcUnitsI);
	DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
				 ("mgntHdlrGetUnits: slcUnits_ps = %p, nelem=%lu, firstUnit=%.4s, keysiz=%lu, keyoff=%lu\n",
				  db_p, nelm, (char *)&firstUnit_ca[0], keysiz, keyoff));
        for (i=0; i<nelm; i++, db_p++) {
	  DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
				 ("mgntHdlrGetUnits: Unit %d = %.4s, 0x%.8lx \n",i,(char *)db_p, *db_p)); 	  
	}
	db_p = DBPTR(mgnt_ps->dlist_s.slcAllUnits_ps, prim_ps->slcUnitsI);
        i_first = Find_index_c( (void *)(db_p),
				 keysiz,
				 nelm,
				 (void *)firstUnit_ca,
                                 keysiz,
                                 keyoff );
        if (MATCH_FOUND(i_first)) {
	  DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
		                   ("mgntHdlrGetUnits: Found first unit %.4s in slcAllUnts list\n", 
                                       (char *)&firstUnit_ca[0]));
          *iunit_p = i_first; 
      /* 
      ** We found the first unit in the unit list 
      ** Now look for the last unit. However, first
      ** make sure that there is a last unit. Please
      ** note that the last unit does not have to be specified.
      ** If it is not,or the unit last unit number is zero then
      ** assume that only one unit is in our range.
      */
	  if ( (*firstUnit_p==lastUnit) || (lastUnit==0)) {
		*nunits_p = 1;		
	  }
	  else {
		db_p = DBPTR(mgnt_ps->dlist_s.slcAllUnits_ps, prim_ps->slcUnitsI);
		/* Now search for the last unit number in the primary unit list. */
		/*nelm = (nelm - i_first) - 1;*/
		i_last = Find_index_c( (void *)db_p,
					keysiz,
					nelm,
					(void *)lastUnit_ca,	
					keysiz,
					keyoff );
		if (MATCH_FOUND(i_last)) {
		  DEBUGPRINT(DP_DEBUG, mgntHdlrFlag, 
				       ("mgntHdlrGetUnits: Found last unit %.4s in slcAllUnts list\n", 
					   lastUnit_ca)); 
		  *nunits_p = (i_last - i_first) + 1;
                }
	        else {
		 *nunits_p = 0;
                 *iunit_p  = 0;
	         status = MGNT_ABORTED;
                 DEBUGPRINT(DP_ERROR, mgntHdlrFlag, 
			      	       ("mgntHdlrGetUnits: Last Unit %.4s not found in slcAllUnits list\n", 
					 lastUnit_ca));
	        }
	   }
        /* Is this a single unit then we have an invalid unit.*/
        } else if ( *firstUnit_p == lastUnit ) {
  	     *nunits_p = 0;
             *iunit_p  = 0;
	      status = MGNT_ABORTED;
              DEBUGPRINT(DP_ERROR, mgntHdlrFlag, 
			      	       ("mgntHdlrGetUnits: Unit %.4s not found in slcAllUnits list\n", 
					 firstUnit_ca));
	/* Otherwise, we have a range of units...so find the first valid unit. */
        } else {
  	  *nunits_p = 0;
          *iunit_p  = 0;
	   status = MGNT_ABORTED;
           DEBUGPRINT(DP_ERROR, mgntHdlrFlag, 
			      	    ("mgntHdlrGetUnits: First Unit %.4s not found in slcAllUnits list\n", 
        			     firstUnit_ca));
	}	 
  }
  return(status=MGNT_OKOK);
}


/*=============================================================================

  Name: mgntHdlrPrimToEnum

  Abs:  Get primary enum or list of enums given a 4-char primary name or names

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        unsigned long    prim          read-only    primary name (4-char) not NULL terminated
        unsigned long *  iprim_p       read-write   primary enum 
        unsigned long *  nprims_p      read-write   number of primaries in list
                                                    This is 1 unless the agrument
                                                    prim was a wildcard

  Rem: The purpose of this function is get primary enum associated with the primary
       name. If a wildcard primary (ie "ALL*") is specified then the first primary 
       enum in the list of primary enums is specified and the number of primaries is
       set to the total number of magnet primaries.

  Side: Wildcard primary name is permitted.

  Ret:  vmsstat_t  
             MGNT_OKOK       - Successful operation
             MGNT_NOTMAGPRIM - Failure, due to unknown primary name.

==============================================================================*/
LOCAL vmsstat_t mgntHdlrPrimToEnum( unsigned long * const prim_p,
                            unsigned long * const iprim_p,
                            unsigned long * const nprims_p )
{
  MGNT_PRIM;  /*mgntPrim_ac */
  vmsstat_t status = MGNT_OKOK;

  /* 
  ** Has a wildcard been speicified for the primary name.
  ** If yes, then just return the entire list of magnet
  ** primaries.
  */
  DEBUGPRINT(DP_DEBUG, mgntHdlrFlag,("mgntHdlrPrimToEnum: prim in msg=0x%.8lx\n", *prim_p ));
  DEBUGPRINT(DP_DEBUG, mgntHdlrFlag,("mgntHdlrPrimToEnum: prim in msg=%.4s\n", (char *)prim_p ));
  if ( strncmp((char *)prim_p,_WILDCARD,sizeof(*prim_p))==0) { 
    *nprims_p = MGNT_NUM_PRIMS; 
    *iprim_p = 0;   
  } 
  /* 
  ** We have a single primary name specified.
  ** Determine if this is a valid primary
  */
  else {
    *nprims_p = MGNT_NUM_PRIMS;
    *iprim_p  = Find_index_c( (void *)mgntPrim_ac[0],
                              sizeof(prim_ta), /* sizeof(prim_ta)+1,*/
                              *nprims_p,
                             (void *)prim_p,
                             sizeof(*prim_p),
                             sizeof(prim_ta) );
    if ( MATCH_FOUND(*iprim_p) )
     *nprims_p = 1;
    else {
     *nprims_p = 0;
     status = MGNT_NOTMAGPRIM;
    }
  }
  return(status);
}
/*=============================================================================

  Name: mgntHdlrCleanup

  Abs:  Cleanup magnet handler before exiting

  Args: none

  Rem: This funtion is called to clean up any memory allocation and
       message queues before exiting the magnet handler.

  Side: 

  Ret:  int 
              0 = Successful
             -1 = failure

==============================================================================*/
LOCAL int mgntHdlrCleanup(void) 
{
  epicsMutexId temp_ps; 
  int status = MGNT_OKOK;

  DEBUGPRINT(DP_INFO, mgntHdlrFlag, ("MGNT_H: Cleaning up and exiting!!! \n"));
  /* Free up any allocated memory */
  status = mgntGlobExit();

  /* Free allocated memory */
  if (prev_bdes_a) { free(prev_bdes_a); prev_bdes_a=NULL; }
  if (prev_hsta_a) { free(prev_hsta_a); prev_hsta_a=NULL; }
  if (prev_stat_a) { free(prev_stat_a); prev_stat_a=NULL; }
  
  if (mgnt_ps->func_s.funcDoneEvent_ps !=NULL) 
  {
	epicsEventDestroy(mgnt_ps->func_s.funcDoneEvent_ps);
	mgnt_ps->func_s.funcDoneEvent_ps = NULL;
  }

  if (mgnt_ps->func_s.busyCounterMutex_ps != NULL)
  {
	temp_ps = mgnt_ps->func_s.busyCounterMutex_ps;
	mgnt_ps->func_s.busyCounterMutex_ps = NULL;
	epicsMutexDestroy(temp_ps);
  }
  

  /* Delete message queues */
  messageQDestroy(mgntHdlr);
  /* clean up thread struct */
  slcThreads_as[mgntHdlr].mqId_ps = NULL;
  slcThreads_as[mgntHdlr].tid_ps = NO_THREAD;

  return( SUCCESS(status) ? 0 : -1 );
}

