/*=============================================================================
 
  Name: mgntMon.c
		mgntMonInitUnit    - establishes ca monitors for unit's BDES, HSTA,CTRL
		                     pvs, it they exist, for a valid unit in slc db
        mgntMonInitContext - init the ca monitor context
        getChannel         - get ca montitor channel
		setCaMonitor       - establish the ca monitor with channel
		mgntMonCTRLCallback- Callback routine for a change in a unit's CTRL value
		mgntMonHSTACallback- Callback routine for a change in a unit's HSTA value
		MonBDESCallback    - Callback routine for a change in a unit's BDES value
		mgntMonCleanup     - Clears all units' ca channels and subscriptions
         
  Abs:  The mgntMon module contains utilities to keep 
        SLC and Epics database HSTA, CTRL, BDES information sychronized by
		using EPICS channel access to monitor the EPICS device HSTA, CTRL
		and BDES records for all units found to be in the slc db. 
		*******************************************************
		bsacHsta:
		Upon initialization, the bsacHdlr thread compares the CTRL values 
		to the SLC database HSTA values.
		The EPICS values are used as 'master' upon initialization.  From then
		on the HSTA and CTRL values reflect the last control command from either
		the SCP or from an EPICS client. 

		The epics CTRL value is monitored.  When the CTRL value is changed by an
		epics user, a callback routine is executed that will send an CTRL updated
		message	to the bsacHdlr thread. When the message is received the bsacHdlr
		will immediately update the hsta_ctrl global, and update the SLC database
		with a dbupdate request.

		The SLC HSTA value is polled.  The cstrAsync thread sends a message to the
		bsacHdlr periodically to check all the unit HSTAs.  If any one has changed
		from the hsta_ctrl global value, then that hsta_ctrl value, and the epics


		CTRL record are both updated immediately.

		The bsacHdlrThread is essentially a channel access client, but we
		rely on the assumption that when a channel access client and server are 
		in the same 'address space and host' the client calls bypass the normal 
		server and work directly with the server tool component.

  Auth: 20 oct 2005, debbie rogind (DROGIND)    created 
  Rev:  

-----------------------------------------------------------------------------*/


#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
  24-jul-2007 dfairley
    cnanged mgntMonInitUnit to mgntMonInitUnits and now it loops through all units
	of a primary, creating the channel access monitors for the PVs HSTA, CTRL, BDES
    This allows us to isolate all channel access includes to just this module
    This fixes a problem in the memory allocation for these monUnitStruct_ts records
	needed for each unit.
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
int mgntMonFlag = DP_ERROR; /* Set to DP_INFO for production */ 
#endif

/* c includes */
#include <string.h>         /* memcpy */
#include <stdlib.h>         /* atoi      */

#include    "cadef.h"      /* channel access */
#include    "dbDefs.h"
#include    "epicsMutex.h" /* mutex utilities */

/* slcIOC utilities */
#include    "slcCmlog.h"   /* slcCmlogLogMsg */
#include    "msgQUtil.h"   /* messageQSend  */
#include    "slcMsg.h"     /* MGNT_NOTIFY */
#include    "slcGlob.h"    /* slcGetIOCName()               */
#include    "micrdef.hc"   /* MICR_OKOK */
#include    "cvtUtil.h"    /* conversion utils */

/* SLC database utilities */
#include    "dbdef.hc"     /* db error codes */
#include    "dblist.h"     /* dblist macros and utils */
#include    "dbl.h"
#include    "dbupdate.h"   

/* mgnt includes */
/* Magnet header files */
#include "mgntdef.hc"      /* magnet error codes            */
#include "mgntTypes.h"     /* mgntHdlrInfo_ts               */
#include "mgntParam.h"     /* CB_CTRL, HSTA, BDES           */
#include "mgntMacros.h"    /* MGNT_PRIM,MGNT_PRIM_NNUM      */

/* channel access channel information for caMonitors */

typedef struct {
  chid chId;
  evid evId;
  int  chStatus;
} caMonStruct_ts;

typedef struct {
  caMonStruct_ts   ctrl_s;
  caMonStruct_ts   hsta_s;
  caMonStruct_ts   bdes_s;
} monUnitStruct_ts;

/* Magnet Monitor information */
/*  the units_ps will be type-cast to monUnitStruct_ts within mgntMon.c */
typedef struct {
  mgntPrim_te         prim_e;      /* primary emum */
  unsigned short      nunits;      /* number of units for this primary            */
  DBLIST_DECLARE    (monUnitStruct_ts) units_ps;
                                   /* ptr to dblist whose sole purpose is to contain */
                                   /*  heap memory for the units array using db free list */
				   /* type it generically so that ca/dbaddr can include */
} monPrimStruct_ts;


/* global vars */
monPrimStruct_ts   monPrim_s[MGNT_NUM_PRIMS];

extern mgntPrimStruct_ts  mgntPrim_s[MGNT_NUM_PRIMS];

/* local function prototypes */


LOCAL vmsstat_t getChannel(const char * name_c,
						   caMonStruct_ts * monUnit_ps) ;

LOCAL vmsstat_t setCaMonitor(const void * unit_ps, 
							 caMonStruct_ts * monUnit_ps,
							 const unsigned short callback);

LOCAL void mgntMonCTRLCallback(struct event_handler_args eha);
LOCAL void mgntMonHSTACallback(struct event_handler_args eha);
LOCAL void mgntMonBDESCallback(struct event_handler_args eha);

/*=============================================================================

  Name:  mgntMonInitUnit

  Abs:  

  Args: Type             Name          Access       Description
        -----------      -----------   ----------   ----------------------------
        const char * const prim_p,     read
		const char * const ioc_p,      read
		const char * const unit_p,     read
	    const void * const unit_ps,    read         unit in mgntPrim_s[prim].unit_ps 
		const unsigned short i_prim,   read         which prim in monPrim_s[_iprim]  
		const unsigned short i_unit

  Rem: 

  Side: For prim/loca/unit/ passed in , try to establish ca monitors
        for secondaries HSTA, CTRL, BDES. If they don't exist, it is 
	not an error. 
	This routine is called when a unit exists in the slc db.

  Ret:  vmsstat_t  
             MGNT_OKOK     - Successful operation
             MGNT_DBERR    - Failed, due to database error
             MICR_MALLOC   - Memory allocation error.
             DB_NOTALLUNIT - Failure, due to no db units
             Otherwise, a failure code from the 
                mgntDbInitPrim()

==============================================================================*/
vmsstat_t mgntMonInitUnits (const unsigned short i_prim)  /*  */
{
  vmsstat_t           iss = MGNT_OKOK;
  mgntPrimStruct_ts   *prim_ps = NULL;      /* primary unit information              */
  monPrimStruct_ts    *monPrim_ps = NULL;   /* primary monitor unit info             */
  mgntUnitStruct_ts   *unit_ps = NULL;      /* unit structure w/epics dbaddr ptrs    */
  char                *prim_pc  = NULL;     /* ptr to primary name (NULL terminated) */
  char                *ioc_pc  = NULL;     /* ptr to 4-char ioc name              */
  unsigned short      i_unit;              /* unit index                            */
  monUnitStruct_ts    *monUnit_ps;
  char                pvname_ca[80];
  unsigned short      callback;
  int                 unitNo;
  char                unit_ac[5];
  mgntUnitStruct_ts   *u_ps; /* to check addr, see if PV exists */
  unsigned long       nbytes;               /* number of bytes          */
 

  prim_ps = &mgntPrim_s[i_prim];
  monPrim_ps = &monPrim_s[i_prim];
  monPrim_ps->nunits = prim_ps->nunits;
  monPrim_ps->prim_e = prim_ps->prim_e;

  if (!prim_ps->nunits) {
	 DEBUGPRINT(DP_DEBUG, mgntMonFlag,
				("mgntMonInitUnits: Exiting - No units for %.4s\n", 
				 prim_ps->prim_ac ));	
	 return(iss);
  }

  prim_pc   = prim_ps->prim_ac;   /* get ptr to primary name */
  ioc_pc   = slcGetLOCAName();  /* get location name            */
		  

  /* allocate memory for the units in the monitor var */
  nbytes = monPrim_ps->nunits * sizeof(monUnitStruct_ts);
  DEBUGPRINT(DP_INFO, mgntMonFlag,
				("mgntMonInitUnits:  No units for %.4s is %u, tot.bytes=%lu\n", 
				 prim_ps->prim_ac, monPrim_ps->nunits ,nbytes));	
  if (! SUCCESS (iss=DBALLOC (monPrim_ps->units_ps, monPrim_ps->nunits))) {
    return (iss);
  }
  /* zero out (all ptrs) in this unit structure */
  memset (DBPTR( monPrim_ps->units_ps, 0), 0, nbytes);

  unit_ps = DBPTR_FIRST(prim_ps->units_ps);
  if (monPrim_s[i_prim].units_ps) {

    for (i_unit=0; i_unit<prim_ps->nunits && unit_ps; unit_ps++,i_unit++) {

      memset (unit_ac, 0, 5);
      memcpy (unit_ac, unit_ps->unit_ac, 4);
      unitNo = atoi((char *)&unit_ac);
	  /* cast the unit_ps to the correct type to check addr*/
      u_ps = (mgntUnitStruct_ts *)unit_ps;
	  /* convert i_prim, i_unit into correct ptr */
	  monUnit_ps = (monUnitStruct_ts *)DBPTR(monPrim_s[i_prim].units_ps, i_unit);

	  callback = CB_CTRL;
      if (u_ps->ctrlAddr_s.precord != NULL) {
	    sprintf (pvname_ca, "%s:%s:%d:%s", prim_pc, ioc_pc, unitNo, PV_CTRL);
	    iss = getChannel(pvname_ca, &(monUnit_ps->ctrl_s) );
	    if (SUCCESS(iss)) setCaMonitor(unit_ps, &(monUnit_ps->ctrl_s), callback );
	  }

	  callback = CB_HSTA;	
      if (u_ps->hstaAddr_s.precord != NULL) {
	    sprintf (pvname_ca, "%s:%s:%d:%s.VAL", prim_pc, ioc_pc, unitNo, PV_EPICSHSTA);
	    iss = getChannel(pvname_ca,  &(monUnit_ps->hsta_s));
	    if (SUCCESS(iss)) setCaMonitor(unit_ps,  &(monUnit_ps->hsta_s), callback );
	  }

	  callback = CB_BDES;	
      if (u_ps->bdesAddr_s.precord != NULL) {
	    sprintf (pvname_ca, "%s:%s:%d:%s.VAL", prim_pc, ioc_pc, unitNo, PV_BDES);
	    iss = getChannel(pvname_ca,  &(monUnit_ps->bdes_s));
	    if (SUCCESS(iss)) setCaMonitor(unit_ps, &(monUnit_ps->bdes_s), callback );
	  }
    }
  } else
    DEBUGPRINT (DP_ERROR, mgntMonFlag, 
	("mgntMonInitUnits: called to create a monitor on i_prim=%hd that doesn't exist\n",
				 i_prim));

  return iss;
}
/*=============================================================================

  Name: getChannel

  Abs:  create a CA channel to the EPICS PV named

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		char *              name_c      read only  pointer to name of EPICS pv
        unsigned short      pIndex      read only  primary device index
        unsigned short      uIndex      read only  unit index

  Rem:   

  Side: 

  Ret:  int channel access error code
==============================================================================*/
LOCAL vmsstat_t getChannel(const char * name_c,
						   caMonStruct_ts * monUnit_ps)
						 
{ 
  vmsstat_t iss = MICR_OKOK;
  int status = ECA_NORMAL;
  chid chId;

  if ((status = (ca_create_channel(name_c,NULL,NULL,10, &chId)) )!=ECA_NORMAL ) {
    DEBUGPRINT (DP_INFO, mgntMonFlag, 
				("mgntMon:ca_create_channel failure, status = %d\n", status));
	 monUnit_ps->chStatus = status; 
	 /*iss = BPMO_DBERR;
	 slcCmlogLogMsg(BPMO_DBERR, "DBERR", 
	 "CA Create Channel failure for %s.\n", (char *)name_c);	*/
	 return(iss); 
  }
  else {
	monUnit_ps->chId = chId;
	monUnit_ps->chStatus = status;
    DEBUGPRINT (DP_INFO, mgntMonFlag, 
				("mgntMon:created ca channel for %s\n", (char *)name_c ));
  }

  if ((status = (ca_pend_io(1.0)))!=ECA_NORMAL ) {
    DEBUGPRINT (DP_ERROR, mgntMonFlag, ("mgntMon:ca_pend_io failure on connect to %s\n",name_c));
     /* failure, release the channel */
  }

  return(iss); 
}
/*=============================================================================

  Name: setCaMonitor

  Abs:  create a CA subscriiption to the EPICS PV named.  Pass in a pointer
        to the global unit structure needed by the subscription callback.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		const unsigned short pIndex       read only  primary index
		const unsigned short uIndex       read only  uIndex index

  Rem:  this routine will assign the subscription callback and pass in the
        pointer to the corresponding primCollect_as[p].unit_as[u] structure.

  Side: 

  Ret:  vmsstat_t vms style return code
   
==============================================================================*/
LOCAL vmsstat_t setCaMonitor(const void * unit_ps, 
							 caMonStruct_ts * monUnit_ps,
							 unsigned short callback) 
{
  vmsstat_t iss = MICR_OKOK;
  int status = ECA_NORMAL;
  mgntUnitStruct_ts * u_ps = NULL;
 
  DEBUGPRINT (DP_DEBUG, mgntMonFlag, 
			  ("mgntMon:setCaMonitor on unit_ps= %p\n",(unsigned long *)unit_ps));

  switch (callback) {
  case CB_CTRL:
	status = (ca_create_subscription(DBR_SHORT, 1, monUnit_ps->chId, DBE_VALUE,  
									 mgntMonCTRLCallback, (void *)unit_ps, &(monUnit_ps->evId))); 
	break;
  case CB_HSTA:
	status = (ca_create_subscription(DBR_SHORT, 1, monUnit_ps->chId, DBE_VALUE,  
									 mgntMonHSTACallback, (void *)unit_ps, &(monUnit_ps->evId))); 
	break;
  case CB_BDES:
	status = (ca_create_subscription(DBR_FLOAT, 1, monUnit_ps->chId, DBE_VALUE,  
									 mgntMonBDESCallback, (void *)unit_ps, &(monUnit_ps->evId)));
	break;
  }/* switch */
 	
  if (status==ECA_NORMAL) DEBUGPRINT (DP_INFO, mgntMonFlag, 
				("mgntMon:Success on create_subscr, status=%d\n", status));
	
  if ((status = (ca_pend_io(1.0)))!=ECA_NORMAL ) {
	u_ps = (mgntUnitStruct_ts *)unit_ps;
    DEBUGPRINT (DP_ERROR, mgntMonFlag, 
	("mgntMon:ca_pend_io failure on create_subscr on %s %s, status=%d\n",u_ps->prim_pc,(char *)&u_ps->unit_ac,status ));
  }  
  return (iss);
}


/*==============================================================================

  Name: mgntMonInitContext()

  Abs:  create a channel access context,
  Args: none

  Rem:  this routine uses the previously initialized primCollect_as global
        to search through all units and create a channel to each unit's 
		epics CTRL record.  It then assigns a monitor to each channel.

  side: the bsacInitGlobals must return successful before this routine is called.

  Ret: vmsstat_t iss the vms style return code

  ==============================================================================*/
vmsstat_t mgntMonInitContext (void)    
{
  vmsstat_t iss = MICR_OKOK;
  int status = ECA_NORMAL;

  /* create a channel access context */
  if ((status = (ca_context_create (ca_enable_preemptive_callback )))!=ECA_NORMAL )
    DEBUGPRINT (DP_FATAL, mgntMonFlag, 
				("mgntMon:ca_create_context failure status = %d\n", status));
  
  return (iss);
}


/*=============================================================================

  Name: mgntMonCTRLCallback

  Abs:  callback for the caMonitor channel access subscription request.
        Send a MGNT_NOTIFY_CTRL message to the mgntAsync.  
		The first two words of reqdata are the pointer to the unit_as[u] 
		struct needed to identify the unit whose CTRL val changed.
  
		The next word is the new CTRL value.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        struct event_handler_args eha   read only channel access event callback arguments

  Rem:  

  Side: a message is sent to the mgntAsync message queue 

  Ret: 
   
==============================================================================*/
LOCAL void mgntMonCTRLCallback(struct event_handler_args eha)
{
  msgptr_ts  mSndptr_s;
  msgmail_ts * mSnd_ps = NULL;
  unsigned short *ctrlVal_p = NULL;
  unsigned long * usr_ps = NULL;

  if(eha.status==ECA_NORMAL) {
	usr_ps = (unsigned long *)eha.usr;
	ctrlVal_p = (unsigned short *)eha.dbr;
	DEBUGPRINT(DP_INFO, mgntMonFlag,
			   ("mgntMonCTRLCallback: got event %s = %hd, unit_ps=%p\n",
				ca_name(eha.chid), *ctrlVal_p, usr_ps) );

	mSndptr_s.msgbytes = sizeof(msgmail_ts);
	if (SUCCESS( msgQGetBuffer(mSndptr_s.msgbytes, (void*)&mSndptr_s.msgptr_p ) )) {
	  mSnd_ps = (msgmail_ts *)mSndptr_s.msgptr_p;

	  mSnd_ps->msgheader.func = MGNT_NOTIFY_CTRL;
	  memcpy( &mSnd_ps->reqdata[0], &usr_ps, 4);
	  mSnd_ps->reqdata[2] = *ctrlVal_p;
	  if (!(SUCCESS(messageQSend(mgntAsync, epicsFalse, sizeof(mSndptr_s), &mSndptr_s))))
	  {
		DEBUGPRINT(DP_ERROR, mgntMonFlag,
			   ("mgntMonCTRLCallback: msg failed sending event %s = %d\n",
				ca_name(eha.chid),*ctrlVal_p) );
		msgQFreeBuffer(sizeof(msgmail_ts), (void *)mSnd_ps);
	  }
	}
  }
}
/*=============================================================================

  Name: mgntMonHSTACallback

  Abs:  callback for the caMonitor channel access subscription request.
        Send a MGNT_NOTIFY_HSTA message to the mgntAsync.  
		The first two words of reqdata are the pointer to the unit_as[u] 
		struct needed to identify the unit whose CTRL val changed.
  
		The next word is the new CTRL value.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        struct event_handler_args eha   read only channel access event callback arguments

  Rem:  

  Side: a message is sent to the mgntAsync message queue 

  Ret: 
   
==============================================================================*/
LOCAL void mgntMonHSTACallback(struct event_handler_args eha)
{
  msgptr_ts  mSndptr_s;
  msgmail_ts * mSnd_ps = NULL;
  unsigned short *ctrlVal_p = NULL;
  unsigned long * usr_ps = NULL;

  if(eha.status==ECA_NORMAL) {
	usr_ps = (unsigned long *)eha.usr;
	ctrlVal_p = (unsigned short *)eha.dbr;
	DEBUGPRINT(DP_INFO, mgntMonFlag,
			   ("mgntMonHSTACallback: got event %s = %d, unit_ps=%p\n",
				ca_name(eha.chid), *ctrlVal_p, usr_ps) );

	mSndptr_s.msgbytes = sizeof(msgmail_ts);
	if (SUCCESS( msgQGetBuffer(mSndptr_s.msgbytes, (void*)&mSndptr_s.msgptr_p ) )) {
	  mSnd_ps = (msgmail_ts *)mSndptr_s.msgptr_p;

	  mSnd_ps->msgheader.func = MGNT_NOTIFY_HSTA;
	  memcpy( &mSnd_ps->reqdata[0], &usr_ps, 4);
	  mSnd_ps->reqdata[2] = *ctrlVal_p;
	  if (!(SUCCESS(messageQSend(mgntAsync, epicsFalse, sizeof(mSndptr_s), &mSndptr_s))))
	  {
		DEBUGPRINT(DP_ERROR, mgntMonFlag,
			   ("mgntMonHSTACallback: msg failed sending event %s = %d\n",
				ca_name(eha.chid),*ctrlVal_p) );
		msgQFreeBuffer(sizeof(msgmail_ts), (void *)mSnd_ps);
	  }
	}
  }
}
/*=============================================================================

  Name: mgntMonBDESCallback

  Abs:  callback for the caMonitor channel access subscription request.
        Send a MGNT_NOTIFY_CTRL message to the mgntAsync.  
		The first two words of reqdata are the pointer to the unit_as[u] 
		struct needed to identify the unit whose CTRL val changed.
  
		The next word is the new CTRL value.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        struct event_handler_args eha   read only channel access event callback arguments

  Rem:  

  Side: a message is sent to the mgntAsync message queue 

  Ret: 
   
==============================================================================*/
LOCAL void mgntMonBDESCallback(struct event_handler_args eha)
{
  msgptr_ts  mSndptr_s;
  msgmail_ts * mSnd_ps = NULL;
  float *ctrlVal_p = NULL;
  unsigned long * usr_ps = NULL;

  if(eha.status==ECA_NORMAL) {
	usr_ps = (unsigned long *)eha.usr;
	ctrlVal_p = (float *)eha.dbr;
	DEBUGPRINT(DP_INFO, mgntMonFlag,
			   ("mgntMonBDESCallback: got event %s = %f, unit_ps=%p\n",
				ca_name(eha.chid), *ctrlVal_p, usr_ps));

	mSndptr_s.msgbytes = sizeof(msgmail_ts);
	if (SUCCESS( msgQGetBuffer(mSndptr_s.msgbytes, (void*)&mSndptr_s.msgptr_p ) )) {
	  mSnd_ps = (msgmail_ts *)mSndptr_s.msgptr_p;

	  mSnd_ps->msgheader.func = MGNT_NOTIFY_BDES;
	  memcpy( &mSnd_ps->reqdata[0], &usr_ps, 4);
	  memcpy( &mSnd_ps->reqdata[2], ctrlVal_p, sizeof(float));
	  if (!(SUCCESS(messageQSend(mgntAsync, epicsFalse, sizeof(mSndptr_s), &mSndptr_s))))
	  {
		DEBUGPRINT(DP_ERROR, mgntMonFlag,
			   ("mgntMonBDESCallback: msg failed sending event %s = %f\n",
				ca_name(eha.chid),*ctrlVal_p) );
		msgQFreeBuffer(sizeof(msgmail_ts), (void *)mSnd_ps);
	  }
	}
  }
}


/*==============================================================================

  Name: cleanup()

  Abs:  cleanup the globals for this thread before shutdown

  Args: none

  Rem: Clean up the dblists used for synchronizing hstas / ctrl
  side:

  Ret: none

  ==============================================================================*/
void mgntMonCleanup(void)    
{
  int status = ECA_NORMAL;
  unsigned short pIndex;
  unsigned short uIndex;
  monPrimStruct_ts * monPrim_ps;
  monUnitStruct_ts * monUnit_ps;

  /*ca_client_status(5);*/

  monPrim_ps = &monPrim_s[0];
  for (pIndex=0; pIndex<MGNT_NUM_PRIMS; pIndex++, monPrim_ps++) {

	if (monPrim_s[pIndex].units_ps) {
	  monUnit_ps = (monUnitStruct_ts *) DBPTR_FIRST(monPrim_ps->units_ps);
	  for (uIndex=0; uIndex< monPrim_ps->nunits; uIndex++, monUnit_ps++) {
		if ( (monUnit_ps->ctrl_s.chId) && (monUnit_ps->ctrl_s.chStatus==ECA_NORMAL) ) {
		  status = ca_clear_subscription(monUnit_ps->ctrl_s.evId);
		  status = ca_clear_channel(monUnit_ps->ctrl_s.chId);
		  status = ca_pend_io(0.1);
		}
		if ( (monUnit_ps->hsta_s.chId) && (monUnit_ps->hsta_s.chStatus==ECA_NORMAL) ) {
		  status = ca_clear_subscription(monUnit_ps->hsta_s.evId);
		  status = ca_clear_channel(monUnit_ps->hsta_s.chId);
		  status = ca_pend_io(0.1);
		}		
		if ( (monUnit_ps->bdes_s.chId) && (monUnit_ps->bdes_s.chStatus==ECA_NORMAL) ) {
		  status = ca_clear_subscription(monUnit_ps->bdes_s.evId);
		  status = ca_clear_channel(monUnit_ps->bdes_s.chId);
		  status = ca_pend_io(0.1);
		}
	  }/* for all units in prim */
		/* now free the primary's units list */
	  DBFREE(monPrim_ps->units_ps);
	  monPrim_s[pIndex].units_ps = NULL;
	} /* for all prims */
  }
  /* destroy the channel access context */
  ca_context_destroy();

  return;
}


