/*=============================================================================
 
  Name: msgHdlr.c
        msgHdlrThread   - main routine for msgHdlr thread
        cleanup         - reset globals, release resources before terminating
		createReply     - create a reply msg
		printMsgheader  - debug routine

  Abs:  msgHdlr contains the msgHdlrThread() thread that receives TEST
        function code messages from the Recv thread, acts on them,
		then sends a reply to the msgSend thread.
        reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | Message Service Design

  Auth: 19 jul 2004, dfairley    created 
  Rev:  04 feb 2005, Debbie Rogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest) 
        08 Mar 2005, dfairley    removed handling TEST_ msgs, gave them to cstrHdlr.c
        30 Nov 2004, dfairley    re-write to production level coding standards, var names etc
        19 jul 2004, dfairley    created 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int msgHdlrFlag = DP_DEBUG;
#endif


/* c includes */
#include <string.h>        /* memcpy */

/* shared slc micro includes */
#include "msgfunc.hc"      /* msg_slc_iocnotify */
#include "micrdef.hc"      /* MICR_EXNET_ codes */
#include "slcMsg.h"        /* IOC_STOP */

/* slc global vars */
#include "slcGlob.h"       /* slcThreads_as */

/* slcIOC types */
#include "msgTypes.h"      /* msgmail_ts etc */

/* slcIOC utilities */
#include "slcUtil.h"       /* slcPrintMsgheader */
#include "msgQUtil.h"      /* messageQSend, messageQRecv */
#include "cvtUtil.h"       /* slcSSwap, slcLSwap, etc */
#include "slcCmlog.h"      /* slcCmlogLogMsg */
#include "timeUtil.h"      /* timeGetVMS, timeDiffVMS10usec */

/* slcMain function prototypes */
#include "slcMain.h"       /* slcStop(), slcRestart() */

/* msgHdlr include */
#include     "msgHdlr.h" /* msgHdlrDiags_ts */

/* local function prototypes */
LOCAL void       cleanup(void);
LOCAL vmsstat_t  sendOKReply(const msgmail_ts *rcv_ps);
LOCAL void       doStop(void);
LOCAL vmsstat_t  replyIOCNotify(const msgmail_ts *rcv_ps);
LOCAL vmsstat_t  doIOCNotify(const msgmail_ts *rcv_ps);

/* msgHdlr diagnostics global variable */
msgHdlrDiags_ts msgHdlrDiags_s = { {0, 0}, 0, {0, 0}, 0, 0, 0, 0, 0, 0};


/* IOCNotify commands to compare */
LOCAL unsigned long boot;
LOCAL unsigned long rset;
LOCAL unsigned long iocNotifyCmd;

/*=============================================================================

  Name: msgHdlrThread

  Abs:  The msg handler thread

  Args: none

  Rem: This thread accepts and acts on FUNC_TEST messages from the SLC control system

  side: 

  Ret:  none

==============================================================================*/
/**procedure**/
void  msgHdlrThread(void)
{
  vmsstat_t     iss;
  msgmail_ts*   rMsg_ps = NULL;   /* read msg buffer pointer */
  slcThread_ts* msgHdlr_ps = &slcThreads_as[msgHdlr]; /* pointer to my thread struct */ 

  /* initialize the boot and reset cmd comparators */
  memcpy(&boot, "BOOT", 4);
  memcpy(&rset, "RSET", 4);
  iocNotifyCmd = 0;
  memset(&msgHdlrDiags_s, 0, sizeof(msgHdlrDiags_s) );
  /* create the message queue id */
  if ( !SUCCESS(messageQCreate(msgHdlr, MAX_QUEUED_MSGS, MAX_MSGPTR_BYTES)) ) {
	cleanup();
	return;
  }

  msgHdlr_ps->active = epicsTrue;
  DEBUGPRINT(DP_INFO, msgHdlrFlag, ("MSG_H: becoming active\n"));
  epicsTimeGetCurrent(&msgHdlrDiags_s.lastStartTime);

  /* main loop of the msg hdlr thread, wait at message queue for incoming 
	 messages, handle it, then send a reply to the msgSend thread
  */
  while(msgHdlr_ps->active)
  {
	/* read a message at the message queue.*/
	if (SUCCESS(iss = msgQMsgRecv(msgHdlr, &msgHdlrDiags_s.lastReqTime, (void*)&rMsg_ps) ))
	{
	  msgHdlrDiags_s.reqsSinceStart++;

	  DEBUGPRINT(DP_INFO, msgHdlrFlag, 
				 ("MSG_H:incoming msg func code = %x\n", rMsg_ps->msgheader.func));
	  switch(rMsg_ps->msgheader.func){
	  case IOC_STOP:
		doStop();
		break;
	  case MSG_IOC_SLCNOTIFY:
		iss = replyIOCNotify(rMsg_ps);
		msgHdlrDiags_s.iocSlcNotifyCnt++;
		break;
	  case IOC_IOC_SLCNOTIFY:
		iss = doIOCNotify(rMsg_ps);
		break;
	  default:
		/* unexpected message */
		break;
	  } /* switch */

	  /* release the receive buffer */
	  msgQFreeBuffer(sizeof(msgmail_ts), rMsg_ps);
	} /* if successful read */

    /* check the stop flag just in case */
	if (msgHdlr_ps->stop)
	  msgHdlr_ps->active = epicsFalse;

  } /* while active */

  /* active not set, time to shut down */
  DEBUGPRINT(DP_INFO, msgHdlrFlag, ("MSG_H: shutdown\n"));
  cleanup();
}

/*=============================================================================

  Name: doStop()

  Abs:  terminate this thread by setting the active flag to false

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		none

  Rem:  setting active to false will break the thread out of it's loop and cause 
        it to begin the termination process.

  side: the slcThreads_as[msgHdlr].active flag is set to false.

  Ret: none

===============================================================================*/
LOCAL void doStop(void)
{
  slcThreads_as[msgHdlr].active = epicsFalse;
}


/*==============================================================================

  Name:replyIOCNotify()

  Abs:  reply to the MSG_IOCSLCNOTIFY with a MICR_OKOK status. Save the 4 char cmd
        in the iocNotifyCmd static var, to be used in part 2 of this function.

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rcv_ps      read only   pointer to the incoming message 

  Rem: The SLC Control System expects a reply to this request BEFORE the slc-aware
       threads are stopped.  The msgSend thread will send back an IOC_IOC_SLCNOTIFY
	   msg when this reply is actually sent off to the SLC Control System. The
	   doIOCNotify function will actually execute the request.

  side: The command requested is stored in the static var iocNotifyCmd, to be used
        by the doIOCNotify function later.

  Ret: vmsstat_t the return code of the sendOKReply function

==============================================================================*/
LOCAL vmsstat_t replyIOCNotify(const msgmail_ts *rcv_ps)
{
  DEBUGPRINT(DP_INFO, msgHdlrFlag, ("MSG_H:got IOC_SLCNotify message - sending reply\n"));
  memcpy(&iocNotifyCmd, &rcv_ps->reqdata, 4);
  return ( sendOKReply(rcv_ps) );
}
/*==============================================================================

  Name:doIOCNotify()

  Abs: Execute the IOCNotify requested (BOOT or RSET), slcRestart or slcStop

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
		const msgmail_ts*   rcv_ps      read only   pointer to the incoming message 

  Rem: This function executes the requested boot command after the reply has
       gone out on TCP

  side: once the command is executed, the iocNotify static var is reset to 0

  Ret: MICR_OKOK for now

==============================================================================*/
LOCAL vmsstat_t doIOCNotify(const msgmail_ts *rcv_ps)
{

  /* now execute according to value */
  if (iocNotifyCmd==boot) {
	DEBUGPRINT(DP_INFO, msgHdlrFlag,("MSG_H:got a restart msg - excuting!\n"));
        /* Set flag that tells cstrHdlr that this was a SCP IPL */
        slcSetNotify(epicsTrue);
	slcRestart();
  } 
  else if (iocNotifyCmd==rset) {
	DEBUGPRINT(DP_INFO, msgHdlrFlag,("MSG_H:got a stop msg - excuting!\n"));
	slcStop();
  }
  /* reset the notify cmd */
  iocNotifyCmd = 0;
  return(MICR_OKOK);
}

/*==============================================================================

  Name: cleanup()

  Abs:  cleanup the globals for this thread before shutdown

  Args: none

  Rem: clean up the slcThreads_as struct for this thread
       and release any resources 

  side: 

  Ret: none

==============================================================================*/
LOCAL void cleanup(void)
{
  messageQRelease(msgHdlr);
  slcThreads_as[msgHdlr].active = epicsFalse;
  slcThreads_as[msgHdlr].tid_ps = NO_THREAD;
}

/*=============================================================================

  Name: sendOKReply

  Abs: copy msg header stuff into response message

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		msgmail_ts          rMsg_ps     read/write pointer to the received message header

  Rem:  copies header values into the send message buffer.  Then sets the RESPONSE bit in the 
        msgheader.func. Sets the source and dest fields, puts the MICR_OKOK reply longword into
		the first two words of reqdata.  

  Side: 

  Ret:  vmsstat_t the status code, MICR_OKOK or MICR_MALLOC
   
==============================================================================*/
LOCAL vmsstat_t sendOKReply(const msgmail_ts *rMsg_ps)
{
  vmsstat_t       iss = MICR_OKOK;
  vmsTimeStamp_ta now;
  long int        timeDiff;
  
  /* send an iss reply  */
  iss = msgQSlcIssReply((void*)rMsg_ps, MICR_OKOK);

  /* calc some diagnostics */
  timeGetVMS(now);
  timeDiff = timeDiffVMS10usec(rMsg_ps->msgheader.timestamp, now);
  msgHdlrDiags_s.totalDeltaRespTime += timeDiff;
  if (timeDiff > msgHdlrDiags_s.maxDeltaRespTime) 
	  msgHdlrDiags_s.maxDeltaRespTime = timeDiff;
  msgHdlrDiags_s.avgDeltaRespTime = 
	  msgHdlrDiags_s.totalDeltaRespTime / msgHdlrDiags_s.reqsSinceStart;
   
  /* calc some more diagnostics */
  if (!SUCCESS(iss))
	  msgHdlrDiags_s.rplyFails++;
  else
	  msgHdlrDiags_s.rplyCnt++;
  return (iss);
}

