/*=============================================================================
 
  Name: msgRecv.c
        msgRecvThread   - main routine of the msgRecv thread
        cleanup         - reset globals, release resources before terminating
		msgSockConnect  - connect to proxy
	    prepMessage     - handle endian issues on incoming fwd and msg header 
		printMsgheader  - debug routine


  Abs:  msgRecv contains the msgRecvThread() thread, which
        receives all incoming "message service" type messages from the proxy 
		server. The message is relayed to the correct 'job' thread based
		on the function code within the msg header.
        reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | Message Service Design

  Auth: 1  jul 2004, dfairley    created
  Rev:  04 feb 2005, Debbie Rogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        08 Mar 2005, dfairley    removed call to osiSockAttach, osiSockRelease
        30 Nov 2004, dfairley   re-write for production level coding standards
        1  jul 2004, dfairley    created 
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int msgRecvFlag = DP_ERROR;
#endif

/* c includes */
#include     <stdio.h>
#include     <string.h>    /* memcpy */

/* slc global vars */
#include     "slcGlob.h"

/* VMS Control system function/status codes  */
#include     "msgfunc.hc"  /* FUNC_MSG function code */
#include     "micrdef.hc"  /* MICR_EXNET_ codes */

/* slc common includes */
#include     "tcpUtil.h" /* tcpGetConnection, tcpGeBuffer */
#include     "msgQUtil.h" /* message queue utilities */
#include     "cvtUtil.h" /* slc conversion utilities */
#include     "timeUtil.h"/* slc time conversion utilities */
#include     "slcCmlog.h" /* slcCmlogLogMsg */

/* msgRecv include */
#include     "msgRecv.h" /* msgRecvDiags_ts */

LOCAL void           cleanup(void);
LOCAL vmsstat_t      msgSockConnect(void);
LOCAL vmsstat_t      getMessage(msgmail_ts **const msg_ps, int *cnt_p);
LOCAL unsigned short prepMessage(msgmail_ts* msg_ps);
LOCAL void   printHeaders(const msgmail_ts* msg_ps);
LOCAL vmsstat_t slcSendImAlive(void);

/* msgRecv diagnostics global variable */  
msgRecvDiags_ts msgRecvDiags_s = { {0,0}, 0, {0,0}, 0, 0, 0, 0, 0, 0, 0 };
/*=============================================================================


  Name: msgRecvThread

  Abs:  The msgReceive thread

  Args: none

  Rem:  This thread will handle all incoming message service msgs from the alpha, via 
        the proxy server.  This thread will decode the forward header and pass
		the message on to the correct 'job' thread.
 
  Side: The recv thread does not have a message queue.  The slcThreads_as[slcRecv].stop 
        flag is used to tell the the Recv thread to stop.  The Recv task sets itself 
		to active when it starts, the Exec thread sets the sclThreads_as[slcRecv].stop 
		flag to true when it wants to shut down the Recv thread.

  Ret:  none

==============================================================================*/
void  msgRecvThread(void)
{
  int             cnt;                  /* number of bytes received from tcp */
  vmsstat_t       iss;                  /* return status for msgq functions */
  vmsstat_t       issr = MICR_OKOK;      /* default ok return for 'socket recv' functions */
  msgmail_ts*     rMsg_ps = NULL;        /* receive buffer */
  msgptr_ts       mqMsg_s;               /* local buffer of msgQ pointer */
  slcThread_ts*   msgRecv_ps = &slcThreads_as[msgRecv]; /* pointer to my thread struct */
  msgRecvDiags_ts* diags_ps = &msgRecvDiags_s;


  /* I'm ready to go */
  msgRecv_ps->active = epicsTrue;
  DEBUGPRINT(DP_INFO, msgRecvFlag, ("RECV:becoming active\n"));

  /* initialize diagnostics */
  memset(diags_ps, 0, sizeof(msgRecvDiags_ts));
  epicsTimeGetCurrent(&diags_ps->lastStartTime);  

  /* make a connection - do this after setting active */
  msgSockConnect();

  /* all set - now tell slc to start Paranoia */
  /* this should be moved to cstrHdlr when ready */
  slcSendImAlive();


  /* main loop of the msg Task, make sure there's a connection to the server,
	 then listen for messages, send on to correct "job" thread.
  */
  while(msgRecv_ps->active)
  {

    /* in case there was an error in the last pass,
     * get a connection with the server*/
    msgSockConnect(); /* dont bother checking for errors,
                         the next read will fail anyway if there was one */

	/* initialize issr tcp receive status code */
	issr = MICR_OKOK;

	/* now begin receiving messages, until connection fails, or we're told to quit */
	while (SUCCESS(issr) && (!msgRecv_ps->stop))
	{
	  /* get the message sent by server.  This will always be the small size  */
      issr = getMessage(&rMsg_ps, &cnt);

      /* if msg cnt>0 then we must handle it, BTW we can
		 get a count of 0 with no error - keepalive */
      if ( (SUCCESS(issr)) && (cnt>0) ) {

        /* prepare the incoming msg  - conversions */
		mqMsg_s.msgbytes = prepMessage(rMsg_ps);
        mqMsg_s.msgptr_p = rMsg_ps;

        /* save the source SLC process name for use by slcCmlogLogMsg */
        /* the source of this msg is the destination for the reply */
        memcpy(&msgRecv_ps->destination, rMsg_ps->msgheader.source, 4);

		/* send the message, wait at queue if full */
		if (SUCCESS(iss = messageQSend(rMsg_ps->job_id, epicsTrue, sizeof(mqMsg_s),  &mqMsg_s)))
		{ 
		  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("RECV: sent msg to %lu (thread enum)\n",rMsg_ps->job_id));
		}
		else 
		{
		  /* note: messageQSend waits forever for room in the message Q */
		  /* if there is a failure there's something really wrong, 
			 don't try again, just dump the message */
		  msgQFreeBuffer(sizeof(msgmail_ts), (void*)rMsg_ps);
		  DEBUGPRINT(DP_ERROR, msgRecvFlag, 
					 ("RECV: trying to send a message  0x%.4x to an unknown thread\n",
					  rMsg_ps->msgheader.func) );
		}
		/* reset the destination name */
		msgRecv_ps->destination = V017_UL;
	  }
	  else /* error result from tcpGetBuffer */
	  {
		msgQFreeBuffer(sizeof(msgmail_ts), (void*)rMsg_ps);
		rMsg_ps = NULL;

		if (msgRecv_ps->stop)
		{
		  msgRecv_ps->active = epicsFalse;
		   /* if as stop is requested, then the thread is shutting down. the 
              slcExec thread has probably already destroyed the socket */
		  issr = MICR_EXNET_RECVRESET;
          slcCmlogLogMsg(issr,"EXNET_RECVRESET", 
						 "msgRecvThread detected disconnect and stop; time to shutdown.\n");
		}  
	  }

	} /* while (SUCCESS(issr) && !slcThreads_as[slcRecv].stop);*/

    /* if we got here and there was some error we should try to close the socket */
	if ( !SUCCESS(issr) ) tcpSockClose(&slcSockets_as[sdMsgSrvc]);
  } /* while active */

  DEBUGPRINT(DP_INFO, msgRecvFlag, ("RECV:shutdown\n"));  
  cleanup();
}

/*===========================================================

  Name: cleanup()

  Abs:  cleanup the globals for this thread before shutdown

  Args: none

  Rem: clean up the slcThreads_as struct for this thread
       and release any resources 

  Side: 

  Ret: none

===========================================================*/
LOCAL void cleanup(void)
{
  /*  clean up thread struct */
  slcThreads_as[msgRecv].active = epicsFalse;
  slcThreads_as[msgRecv].mqId_ps = NULL; /* has no mq */
  slcThreads_as[msgRecv].tid_ps = NO_THREAD;
}

/*===========================================================

  Name: msgSockConnect

  Abs:  get a socket connection for the msgRecvThread. This
        routine will wait forever for a good connection, or until
		the msgRecv thread is told to stop.

  Args: none

  Rem:  returns if this thread is told to stop
        
  Side: increments the diagnostic totalProxyConnects when necessary

  Ret:  iss - the status code
===========================================================*/
LOCAL vmsstat_t msgSockConnect(void)
{
  vmsstat_t      iss = MICR_OKOK;
  unsigned short proxy = PROXIES_HOST_MSG_PORT;   /* Proxy port number */
  unsigned short host = MICROS_HOST_MSG_PORT;     /* our 'porttag'*/
  unsigned long sleepCount = 0;
 
  if (slcSockets_as[sdMsgSrvc].sd == INVALID_SOCKET) {
	iss =  tcpSockConnect(msgRecv, proxy, host, 5.0, &sleepCount, &slcSockets_as[sdMsgSrvc] ); 
	if (SUCCESS(iss)) msgRecvDiags_s.totalProxyConnects++;
  }
  msgRecvDiags_s.msgSocketStatus = slcSockets_as[sdMsgSrvc].sd;
  return (iss);
}

/*===========================================================

  Name: getMessage

  Abs:  get the message  

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        msgmail_ts*         msg_ps      read/write  pointer to the msg struct

  Rem:  get a message buffer, and read in message from tcp

  Side: sets several msgRecv thread diagnostics

  Ret: iss - the status code

===========================================================*/
LOCAL vmsstat_t  getMessage(msgmail_ts **const msg_ps, int *cnt_p)
{
  /* get the maximum number of bytes we could receive from TCP (after the fwd hdr)*/
  int          maxBytes = sizeof((*msg_ps)->msgheader) + sizeof((*msg_ps)->reqdata);          
  vmsstat_t    iss  = MICR_OKOK;

  /* get a buffer for the  message sent by server.  This will always be the small size  */
  iss = msgQGetBuffer( sizeof(msgmail_ts), (void**)msg_ps );
  if ( SUCCESS(iss) )
  {
  
	DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("RECV:waiting at socket for msg....\n"));
	iss = tcpGetBuffer(&slcSockets_as[sdMsgSrvc], maxBytes,
						cnt_p, &(*msg_ps)->tcpfwdhdr,  &(*msg_ps)->msgheader);
    /* diagnostics */
    if (!SUCCESS(iss) ) msgRecvDiags_s.tcpReadFails++;

	DEBUGPRINT(DP_INFO, msgRecvFlag, ("RECV:Got a msg!\n"));

	/* more diagnostics */
    msgRecvDiags_s.msgSocketStatus =slcSockets_as[sdMsgSrvc].sd; 
	epicsTimeGetCurrent(&msgRecvDiags_s.lastReqTime); 
	msgRecvDiags_s.reqsSinceStart++;
	msgRecvDiags_s.reqsPerSec = msgRecvDiags_s.reqsSinceStart / 
	                            ( epicsTimeDiffInSeconds(
							      &msgRecvDiags_s.lastReqTime, 
							      &msgRecvDiags_s.lastStartTime) );
  }
  /* return status */
  return (iss);
}

/*===========================================================

  Name: prepMessage

  Abs:  prepare the msgheader struct for this host architecture

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        msgmail_ts*         msg_ps      read/write  pointer to the msg struct

  Rem:  swap in place the members of the msgheader structure
        replace the timestamp with the current time in vms format

  Side: conversion utilities only convert if reqd by this processor
        several msgRecv thread diagnostics are calculated

  Ret: none

===========================================================*/
LOCAL unsigned short  prepMessage( msgmail_ts *msg_ps)
{
  vmsTimeStamp_ta vmsTime;
  long int        dest_slcJobId; /* the slc job id of msg destination */
  unsigned short  msgBytes = 0;

  /* handle the endian-ness of the incoming msgheader */
  /* first swap the function code */
  cvtSSwap(&msg_ps->msgheader.func);
  /* now swap the datalen word count */
  cvtSSwap(&msg_ps->msgheader.datalen);

  /* get the slc job id and convert to saIoc thread enumeration */
  dest_slcJobId = (msg_ps->msgheader.func >> 8);
  msg_ps->job_id = slcJobs_as[dest_slcJobId].hdlr_e;

  /* DMF below for debugging */
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("RECV: slc micro JobId = %lu\n", dest_slcJobId) );
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("RECV: after swap!\n") );
  printHeaders(msg_ps);
  /* DMF end for debugging */

  /* replace the timestamp */
  timeGetVMS(vmsTime);
  memcpy(&msg_ps->msgheader.timestamp, vmsTime, 8);
  /* the char arrays should be fine */

  msgBytes = sizeof(msg_ps->job_id)+sizeof(fwd_hdr_ts)+ntohl(msg_ps->tcpfwdhdr.len);

  /* calc some diagnostics */
  if (msgBytes > msgRecvDiags_s.maxReqBytes) 
	msgRecvDiags_s.maxReqBytes = msgBytes;
  msgRecvDiags_s.totalReqBytes+=msgBytes;
  msgRecvDiags_s.avgReqBytes = msgRecvDiags_s.totalReqBytes / msgRecvDiags_s.reqsSinceStart;

  /* return the number of bytes recvd */
  return (msgBytes);
}


/*=============================================================================

  Name printHeaders

  Abs: print the msgheader structure

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        msgmail_ts        msgheader   read-only  message structure

  Rem:  a debug routine to print interesting things in the msgheader

  Side: 

  Ret: 
   
==============================================================================*/
LOCAL void printHeaders(const msgmail_ts *msg_ps)
{
   fwd_hdr_ts*    fwd_ps = NULL;
   msgheader_ts*  msgheader = NULL;
  vmsTimeStamp_ta  curTime;
  long int         usec_diff, msec_diff;


  fwd_ps = &((msgmail_ts*)msg_ps)->tcpfwdhdr;
  msgheader = &((msgmail_ts*)msg_ps)->msgheader;

  /* print the job code */
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("RECV:job code is: %lu\n",msg_ps->job_id) );
 
  /* print the forward header */ 
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("RECV:fwd header is filled with: \n") );
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe ip_port: %lu\n",fwd_ps->ip_port_u.word) );
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe cmd: %u\n",fwd_ps->cmd) );
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe crc: 0x%x\n", fwd_ps->crc) );
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe user.byte0:  %u\n", fwd_ps->user.byte0) );
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe user.byte1:  %u\n", fwd_ps->user.byte1) );
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe len:  %lu\n", fwd_ps->len) );

  /* print the message header */
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("RECV:msg header is filled with:\n"));
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe source: %c%c%c%c\n",msgheader->source[0],
		msgheader->source[1],msgheader->source[2],
		msgheader->source[3]));
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe dest: %c%c%c%c\n",msgheader->dest[0],
		msgheader->dest[1],msgheader->dest[2],
		msgheader->dest[3]));
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe timestamp[0]:  0x%.8lx\n",msgheader->timestamp[0]));
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe timestamp[1]:  0x%.8lx\n",msgheader->timestamp[1]));
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe func:  0x%.4x\n",msgheader->func));
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\tthe datalen: %u\n",msgheader->datalen));

  /* more debugging, show the difference in time */
  /* get the current time in VMS format*/
  timeGetVMS(curTime);
  usec_diff = timeDiffVMS10usec(msgheader->timestamp, curTime);
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\ttime diff in 10usec: %li\n", usec_diff));
  msec_diff = timeDiffVMS10msec(msgheader->timestamp, curTime);
  DEBUGPRINT(DP_DEBUG, msgRecvFlag, ("\ttime diff in 10msec: %li\n", msec_diff));
}

/*=============================================================================

  Name slcSendImAlive

  Abs: 

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        msgheader_ts        msgheader   read-only  message header structure

  Rem: 

  Side: 

  Ret: 
   
==============================================================================*/
LOCAL vmsstat_t slcSendImAlive(void)
{
  vmsstat_t iss;
  char *microname = slcGetIOCName();
  msgmail_ts * msg_ps=NULL;

  if (SUCCESS( iss = msgQGetBuffer( sizeof(msgmail_ts),(void*) &msg_ps ) ) ) {
	msg_ps->msgheader.func = MICRREQ_IMALIVE;
	memcpy(&msg_ps->msgheader.dest,"V016",4);
	memcpy(&msg_ps->msgheader.source, microname, 4);
	DEBUGPRINT(DP_INFO, msgRecvFlag, ("RECV:sending ImAlive to Paranoia\n"));
	iss = msgQMsgReply(msgSend, msg_ps);
  }
  return (iss);
}
