/*=============================================================================
 
  Name: msgSend.c
        msgSendThread      - the main routine for the msgSend thread
        cleanup            - release all resources before termination
	    swapMsgheader      - swap msg header 
		msgSendBuffer      - send a reply to the slc Control System

  Abs:  msgSend contains the msgSendThread() thread, which
        receives all outgoing "message" type messages
	    from the asIoc threads. The message is sent to the 
	    proxy server. 
        reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | Message Service Design

  Auth  19 jul 2004, dfairley    created 
  Rev:  04 feb 2005, Debbie Rogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        08 Mar 2005, dfairley    removed call to osiSockAttach, osiSockRelease
        29 Nov 2004, dfairley    re-write for production, coding standards, var names etc
        19 jul 2004, dfairley    created 
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>   /* printf */
  int msgSendFlag = DP_ERROR;
#endif

/* c includes */
#include <string.h>       /* memcpy */

/* shared slc micro includes */
#include  "micrdef.hc"    /* MICR_OKOK, MICR_EXNET_x */
#include  "msgfunc.hc"      /* MSG_IOC_SLCNOTIFY */
#include  "slcMsg.h"      /*IOC_IOC_SLCNOTIFY */

/* slc includes */
#include  "slcGlob.h"   /* slcThreads_as, slcSockets_as */

/* slc types */
#include "msgTypes.h"     /* fwd_hdr_ts, msgmail_ts */

/* slc function prototypes */
#include "msgQUtil.h"     /* msg q utilities */
#include "cvtUtil.h"      /* slc conversion utilities */
#include "timeUtil.h"     /* timeGetVMS */
#include "tcpUtil.h"      /* tcpSockCheck, tcpSendBuffer */

/* msgSend include */
#include     "msgSend.h" /* msgSendDiags_ts */

/* internal function prototypes */
LOCAL void cleanup(void);
LOCAL vmsstat_t msgSendBuffer( const unsigned long dataBytes,
							   const lrgmail_ts*   msg_ps);
LOCAL void swapMsgheader(msgheader_ts *msghdr_ps);
LOCAL void printHeaders(const msgmail_ts *msg_ps);

/*diagnostics global variable */
msgSendDiags_ts msgSendDiags_s = { 
  {0,0}, 0, {0,0}, 0, 0, 0, 0, 0 
};
/*=============================================================================

  Name: msgSendThread

  Abs:  The msgSend thread main routine

  Args: none

  Rem:  This thread that will handle all outgoing messages from the ioc, via 
        the proxy server.  This thread gets a message from its queue and send the
		message out over TCP/IP.

  Side: NOTE: in this design, if the sendBuffer call fails we keep trying forever or
              until a STOP is requested.
              When the STOP flag is set, the thread is immediately
			  stopped and all other messages in the message queue are lost.

  Ret:  none
==============================================================================*/
void msgSendThread(void)
{
  unsigned long dataBytes=0;
  unsigned long totalBytes = 0;
  lrgmail_ts*   msg_ps = NULL;  /* a large message buffer - 
								   all messages will fit here */
  vmsstat_t     iss = MICR_OKOK;  /* for error codes */
  slcThread_ts* msgSend_ps = &slcThreads_as[msgSend];
  unsigned long sleepCount= 0;

  /* create the message queue id */
  if ( !SUCCESS(messageQCreate(msgSend, MAX_QUEUED_MSGS, MAX_MSGPTR_BYTES) ) ) {
	cleanup();
	return;
  }
  /* initialize the diagnostics */
  memset(&msgSendDiags_s, 0, sizeof(msgSendDiags_s));

  msgSend_ps->active = epicsTrue;
  DEBUGPRINT(DP_INFO,msgSendFlag, ("SEND: becoming active\n"));
  epicsTimeGetCurrent(&msgSendDiags_s.lastStartTime);

  /* main loop of the msg Send Task, wait at message queue for outgoing 
	 messages, then send message on to proxy server
  */
  while( msgSend_ps->active)
  {
    
	if (SUCCESS(msgQMsgRecv( msgSend, NULL, (void*)&msg_ps) ) )
	{
	  /* get the data length in bytes before swapping header for endian-ness */
	  dataBytes= (msg_ps->msgheader.datalen) * 2;
	  /* get the total message buffer size*/
	  totalBytes = dataBytes + sizeof(fwd_hdr_ts) + sizeof(msgheader_ts);

	  /* if this is a stop message from slcExec we ignore it. instead wait
		 for the stop flag to be set and act on that */
	  if (msg_ps->msgheader.func != IOC_STOP) {

		DEBUGPRINT(DP_INFO, msgSendFlag, ("SEND:sending msg to my VMS master\n"));
 
		/* now handle endian issues for msgheader */
		swapMsgheader(&msg_ps->msgheader);
	  
		/* We must check that we have a good connection before attempting
		   to send the message.  If this fails, keep trying, do not 
		   dump the message.
		*/
	 	iss = MICR_EXNET_SENDFAIL;
		while( (!SUCCESS(iss)) && (!msgSend_ps->stop) )
		{
		  if ( tcpSockCheck(&slcSockets_as[sdMsgSrvc], 
							msgSend, msgRecv, 5.0, &sleepCount) )
		  {
			if (!SUCCESS(iss = (msgSendBuffer(dataBytes, msg_ps))) )
			{
			  DEBUGPRINT(DP_ERROR, msgSendFlag, 
						 ("SEND: error sending; close socket and  wait for new\n"));
			  tcpSockClose(&slcSockets_as[sdMsgSrvc]);
 		 	  epicsThreadSleep(5.0);
			}
		  }
		}      
	  }else {
		DEBUGPRINT(DP_INFO, msgSendFlag, ("SEND: got a STOP msg!\n"));
	  }
	  /* free the message buffer */
	  msgQFreeBuffer(totalBytes, msg_ps);
	  /* reset the destination name */
	  slcThreads_as[msgSend].destination = V017_UL;

      /* check the stop flag */
	  if (msgSend_ps->stop)
		msgSend_ps->active = epicsFalse;
	} 

  } /* while active */

  /* active not set, time to shut down */
  DEBUGPRINT(DP_INFO,msgSendFlag, ("SEND: shutdown\n"));
  cleanup();
}

/*===========================================================

  Name: cleanup()

  Abs:  cleanup the globals for this thread before shutdown

  Args: none

  Rem: clean up the slcThreads_as struct for this thread
       and release any resources 

  Side:

  Ret: none

===========================================================*/
LOCAL void cleanup(void)
{
  messageQRelease(msgSend);
  slcThreads_as[msgSend].active = epicsFalse;
  slcThreads_as[msgSend].tid_ps = NO_THREAD;
}

/*=============================================================================

  Name: putBuffer

  Abs: send a message to the proxy server

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		int                 sd         read/write  the socket used for sending messages
        fwd_hdr_ts          fwd_ps     read/write  pointer to the forward header
        char*               dest       read/write  pointer to destination name
        insigned long       dataBytes  read/write  the number of bytes in the reqdata field

  Scop: called by msgSendThread

  Rem:  fills the forward message header and sends the message to the
        proxy server.

  Side: 

  Ret:  vmsstat_t a return code / error indicator
        Possible return codes are: MICR_EXNET_SENDFAIL, MICR_OKOK

==============================================================================*/
/**procedure**/
LOCAL vmsstat_t msgSendBuffer( const unsigned long dataBytes,
					           const lrgmail_ts*   msg_ps)
{
   unsigned long  bytes_to_send;
   vmsstat_t      iss = MICR_EXNET_SENDFAIL;
   fwd_hdr_ts*    fwd_ps = NULL;
   msgheader_ts*  mhd_ps = NULL;
   lrgmail_ts*    tmpMsg_ps = NULL;
   msgmail_ts*    mNotify_ps = NULL;
   unsigned short tmpfunc = 0;
   unsigned long  this_bc = 0;
   unsigned long  done_bc = 0;  /* multipacket bytes counters */
   unsigned long  largest_bc = (NETMICMSGLEN*2) - sizeof(msgheader_ts) - sizeof(fwd_hdr_ts);
   unsigned short u = 0; /* user byte indicator for multipacket sends */
   unsigned long  packet_bc = 0; /* byte count of a single packet */
   

   fwd_ps = &((lrgmail_ts *)msg_ps)->tcpfwdhdr;
   mhd_ps = &((lrgmail_ts *)msg_ps)->msgheader;

   DEBUGPRINT(DP_INFO,msgSendFlag,("SEND:data bytes = %ld\n",dataBytes) );

   /* calculate the full number of bytes to send */
   bytes_to_send = dataBytes + sizeof(fwd_hdr_ts) + sizeof(msgheader_ts);

   /* fill in the forward header with info - no swapping reqd */
   memcpy(&fwd_ps->ip_port_u.word, &mhd_ps->dest, 4); 
   fwd_ps->cmd = PX_FORWARD_ALIAS_FUNC;
   fwd_ps->crc = 0x55;
   fwd_ps->user.byte0 = 1;
   fwd_ps->user.byte0 |= DONE_MASK;  /* had been used for slcnet */
   fwd_ps->user.byte1 = 0;  /* had been used for slcnet */
   fwd_ps->len = htonl( (dataBytes+sizeof(msgheader_ts)) );

   printHeaders( (msgmail_ts*)msg_ps );

   DEBUGPRINT(DP_INFO, msgSendFlag, ("SEND:total bytes to send = %ld\n", bytes_to_send) );

   if (bytes_to_send <= (NETMICMSGLEN*2) ) 
   {
	 /* this is a small message - send it as is */
	 iss = tcpSendBuffer(&slcSockets_as[sdMsgSrvc], bytes_to_send, fwd_ps);
     if (!SUCCESS(iss)) msgSendDiags_s.writeFails++;
   }
   else
   {   
	 /* its a big one, do multiple sends, breaking up the buffer */
	 DEBUGPRINT(DP_INFO, msgSendFlag, ("SEND: breaking up buffer for send!\n"));
 
	 /* get a temp buffer and copy in headers*/
	 msgQGetBuffer(bytes_to_send, (void*)&tmpMsg_ps );
     memcpy(&tmpMsg_ps->tcpfwdhdr, &msg_ps->tcpfwdhdr, sizeof(fwd_hdr_ts) );
     memcpy(&tmpMsg_ps->msgheader, &msg_ps->msgheader, sizeof(msgheader_ts) );
	 
	 for (this_bc = 0, u=1; ; ++u, done_bc +=this_bc) {

	   /* figure out the data byte count for this packet */
	   if ( (this_bc = dataBytes - done_bc) > largest_bc)
		 this_bc = largest_bc;
	   else 
		 u |= DONE_MASK;
		 
	   /* fill in the forward header for this packet */
	   tmpMsg_ps->tcpfwdhdr.user.byte0 = u; 
	   tmpMsg_ps->tcpfwdhdr.len = htonl( (this_bc+sizeof(msgheader_ts)) );

	   /* fill in the msgheader for this packet */
	   tmpMsg_ps->msgheader.datalen = this_bc/2;
	   DEBUGPRINT(DP_INFO, msgSendFlag, ("SEND:sending buf size %lu; datalen = %u\n", 
										 this_bc, tmpMsg_ps->msgheader.datalen  ));
       cvtSSwap(&tmpMsg_ps->msgheader.datalen);

	   /* copy this portion of the data bytes */
       memcpy(&tmpMsg_ps->reqdata, &msg_ps->reqdata[(done_bc/2)], this_bc);
       
	   /* now send the message */
       packet_bc = this_bc + sizeof(msgheader_ts) + sizeof(fwd_hdr_ts);
	   iss = tcpSendBuffer(&slcSockets_as[sdMsgSrvc], packet_bc, &tmpMsg_ps->tcpfwdhdr);
       if (!SUCCESS(iss)) msgSendDiags_s.writeFails++;

	   if ( (u & DONE_MASK) || (!SUCCESS(iss)) )
		 break;

	 } /* for */
	 
	 /* release the temporary buffer */
	 msgQFreeBuffer(bytes_to_send, tmpMsg_ps);
   } /* if large buffer */


   /* check for the MSG_IOC_SLCNOTIFY message - we need to send
		an IOC_IOC_SLCNOTIFY back to msgHdlr if it is. */
   cvtSSwapMove(&mhd_ps->func, &tmpfunc);
   if (tmpfunc == (MSG_IOC_SLCNOTIFY | RESPONSE ) ) {

	 if (SUCCESS( msgQGetBuffer( sizeof(msgmail_ts), (void*)&mNotify_ps ) )) {
	   mNotify_ps->msgheader.func = IOC_IOC_SLCNOTIFY;
	   DEBUGPRINT(DP_INFO, msgSendFlag, ("SEND: sending SLCNOTIFY to msgHdlr\n") );
	   msgQMsgReply(msgHdlr, mNotify_ps);
	 }
   }


   if (SUCCESS(iss) )
   {
	 /* calc diagnostics */
     epicsTimeGetCurrent(&msgSendDiags_s.lastRplyTime);
     msgSendDiags_s.rplysSinceStart++;
     msgSendDiags_s.rplysPerSec = msgSendDiags_s.rplysSinceStart / 
	                            ( epicsTimeDiffInSeconds(
							      &msgSendDiags_s.lastRplyTime, 
							      &msgSendDiags_s.lastStartTime) );
     msgSendDiags_s.totalRplyBytes+=bytes_to_send;
     msgSendDiags_s.avgRplyBytes = 
	   msgSendDiags_s.totalRplyBytes/msgSendDiags_s.rplysSinceStart;
     if (bytes_to_send>msgSendDiags_s.maxRplyBytes)
	   msgSendDiags_s.maxRplyBytes = bytes_to_send;
     
   }
   DEBUGPRINT(DP_INFO, msgSendFlag, ("SEND: exiting msgSendBuffer \n") );
   return iss;
}                                             

/*===========================================================

  Name: swapMsgheader

  Abs:  swap the msgheader struct from vms format to host format

  Args: msgheader_ts* msghdr_ps read/write pointer to the msgheader

  Rem:  swap in place the members of the msgheader structure
        !!this assumes the byte alignment is alway 4 or less!!

  Ret: none

===========================================================*/
LOCAL void swapMsgheader(msgheader_ts *msghdr_ps)
{
  vmsTimeStamp_ta curTime;
  
  /* first swap the function code */
  cvtSSwap(&msghdr_ps->func);
  /* now swap the datalen word count */
  cvtSSwap(&msghdr_ps->datalen);
  /* now swap the timestamp words  - no dont - they stay in VMS format throughout! 
     here we insert the current time before sending it out */
  if (timeGetVMS(curTime) == epicsTimeOK) 
	  memcpy(&msghdr_ps->timestamp[0], curTime, 4);
  /* the char arrays should be fine */
}

/*=============================================================================

  Name printHeaders

  Abs: print the msgheader structure

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        msgmail_ts        msgheader   read-only  message structure

  Rem:  a debug routine to print interesting things in the msgheader

  Side: 

  Ret: 
   
==============================================================================*/
LOCAL void printHeaders(const msgmail_ts *msg_ps)
{
   fwd_hdr_ts*    fwd_ps = NULL;
   msgheader_ts*  msgheader = NULL;
   
  fwd_ps = &((msgmail_ts*)msg_ps)->tcpfwdhdr;
  msgheader = &((msgmail_ts*)msg_ps)->msgheader;

  /* print the job code */
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("SEND:job code is: %lu\n",msg_ps->job_id) );
 
  /* print the forward header */ 
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("SEND:fwd header is filled with: \n") );
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe ip_port: %lu\n",fwd_ps->ip_port_u.word) );
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe cmd: %u\n",fwd_ps->cmd) );
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe crc: 0x%x\n", fwd_ps->crc) );
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe user.byte0:  %u\n", fwd_ps->user.byte0) );
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe user.byte1:  %u\n", fwd_ps->user.byte1) );
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe len:  %lu\n", fwd_ps->len) );

  /* print the message header */
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("SEND:msg header is filled with:\n"));
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe source: %c%c%c%c\n",msgheader->source[0],
		msgheader->source[1],msgheader->source[2],
		msgheader->source[3]));
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe dest: %c%c%c%c\n",msgheader->dest[0],
		msgheader->dest[1],msgheader->dest[2],
		msgheader->dest[3]));
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe timestamp[0]:  0x%.8lx\n",msgheader->timestamp[0]));
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe timestamp[1]:  0x%.8lx\n",msgheader->timestamp[1]));
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe func:  %x\n",msgheader->func));
  DEBUGPRINT(DP_DEBUG, msgSendFlag, ("\tthe datalen: %u\n",msgheader->datalen));


}
