/*=============================================================================
 
  Name: cvtUtil.c
        cvtSSwap - swap the bytes of a short int
		cvtLSwap - swap the bytes of a long int
		cvtSSwapMove - swap the bytes of a short int, moving the result to a new location
		cvtLSwapMove - swap the bytes of a short int, moving the result to a new location
		cvtFloatVMSToIEEE - convert a VMS float to an IEEE float
		cvtFloatIEEEToVMS - convert an IEEE float to a VMS float
		cvtFromVMS - convert an array of data from VMS to native format
		cvtToVMS - convert an array of data from native to VMS format

        

  Abs : datatype conversion utility routines.  These routines convert word and float
        datatypes from vms format to host format and vice versa.  There is a built-in
        test for endian-ness on the host for all conversions.  
        refer to web page LCLS Controls | Facilities | SLC-Aware IOC | General Purpose Utilities
                     
  Auth: 18 AUG 2004, Diane Fairley (dfairley): created datatype conversion routines host->vms->host
  Rev:  02 feb 2005 drogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        12 DEC 2004, diane fairley:
		   added this header, updated prototypes to coding standards
        29 NOV 2004, Diane Fairley:
		   moved time functions/defines to timeUtil.c
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int cvtUtilFlag = DP_FATAL;
#endif

#define INDEF 1.6e38		/* or whatever you want it to be */

/* c includes */
#include <string.h>     /* memcpy */

/* slc includes */
#include "slcGlob.h"    /* slcIocIsLittleEndian */

/* function prototypes */
#include "cvtUtil.h"

/* local typedefs */
typedef struct {  /* this is the ieee floating point bit structure big-endian */
      unsigned s   : 1;
      unsigned exp : 8;
      unsigned man : 23;
} be_float_bits;

typedef struct {  /* this is the ieee floating point bit structure little-endian */
      unsigned man : 23;
      unsigned exp : 8;
      unsigned s   : 1;
} le_float_bits;

typedef union {   /* union used for accessing 4 bytes in different formats little-endian*/
    float          f;
    unsigned long  l;
    unsigned short s[2];
    le_float_bits  b;
} le_float_tu;

typedef union {   /* union used for accessing 4 bytes in different formats big-endian*/
    float          f;
    unsigned long  l;
    unsigned short s[2];
    be_float_bits  b;
} be_float_tu;
  
typedef union { /* represent the vms timestamp in different formats */
  double vmstime;
  unsigned long int liTime[2];
  unsigned short int siTime[4];
} vmstime_tu;


/*=============================================================================

  Name: cvtSSwap

  Abs:  swap 2 bytes in place  1-2 to 2-1

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
        const unsigned short* s_p           read/write pointer to the variable being converted

  Rem:  swap in place the hi / lo bytes of the unsigned short s
        Can be used for vms -> host,  and host -> vms. 

  Side: If the host is	little-endian the swap is no-op'd.

  Ret:  none

==============================================================================*/
/* procedure */
void cvtSSwap(const unsigned short* s_p) 
{
  register unsigned char tmp, *p;
  
  if (!slcIocIsLittleEndian)
  {
	/* flip bytes */
	p = (unsigned char*)s_p;
	tmp = *p;
	*p = *(p+1);
	*(p+1) = tmp;
  }
}

/*=============================================================================

  Name: cvtLSwap

  Abs:  swap 4 bytes in place  1-2-3-4 -> 4-3-2-1

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
        const unsigned long*  l_p           read/write pointer to the variable being converted

  Rem:  Swap (rotate really) in place the 4 bytes of the input var l.
        Can be used for vms -> host,  and host -> vms. 

  Side: if the host is little endian the swap is no-op'd

  Ret: none - bytes are swapped in place

=======================================================================================*/
void cvtLSwap(const unsigned long* l_p)
{
  register unsigned char tmp, *p;

  if (!slcIocIsLittleEndian)
  {
	/* flip outside bytes */
	p = (unsigned char *)l_p;
	tmp = *p;
	*p = *(p+3);
	*(p+3) = tmp;

	/* flip inside bytes */
	tmp = *(p+1);
	*(p+1) = *(p+2);
	*(p+2) = tmp;
  }
}

/*=============================================================================

  Name:  cvtSSwapMove

  Abs:  swap 2 bytes; returning val in a new var, the original is unchanged

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
        const void*           u2In_p        read only  pointer to the variable being converted
        void*                 u2Out_p       read/write pointer to the destination 

  Rem:  copies the input value into the pOut2u location, then flips the bytes 
        The original input variable is never altered.
        Can be used for vms -> host,  and host -> vms. 

  Side: Uses cvtSSwap, which will noop the swap if host is little-endian. 

  Ret:  none

==============================================================================*/
/* procedure */
void  cvtSSwapMove( const void* u2In_p, void * u2Out_p) 
{
 
  /* flip bytes */
  memcpy(u2Out_p,u2In_p,2);
  cvtSSwap((unsigned short *)u2Out_p);
}

/*=============================================================================

  Name: cvtLSwapMove

  Abs:  swap 4 bytes returning val in new var, original is unchanged 

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
        const void*           u4In_p         read only  pointer to the variable being converted
        void *                u4Out_p        read/write pointer to the destination variable

  Rem:  copies the input value into the urOut_p location, then flips the bytes 
        The original input variable is never altered.
        Can be used for vms -> host,  and host -> vms. 

  Side: Uses cvtLSwap, which will noop the swap if host is little-endian. 

  Ret:   none

=======================================================================================*/
void cvtLSwapMove(const void* u4In_p, void* u4Out_p)
{
  /* rotate bytes */
  memcpy(u4Out_p, u4In_p,4);
  cvtLSwap((unsigned long *)u4Out_p);
}

/*=============================================================================

  Name: cvtFloatVMSToIEEE

  Abs:  convert a vms float ( 4-byte) to a IEEE float (4-byte)

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
        const void*           r4In_p        read only  pointer to the real*4 variable being converted
        void *                r4Out_p       read/write pointer to the output IEEE float variable

  Rem:  copies the input value into a temp location, then flips adjacent bytes and
        subtracts 2 from the exponent. the final value is in the r4Out_p location;  the original 
        is not altered if r4In_p != r4Out_p

  Side: uses cvtLSwap which will noop the swap if the host is little-endian

  Ret: none

=======================================================================================*/
void cvtFloatVMSToIEEE( const void* r4In_p, void* r4Out_p)
{      
  register unsigned short tmpS, *pS;
  be_float_tu temp;
  le_float_tu* ltemp;
 
  memcpy(&temp, r4In_p, 4);

  /* swap bytes little endian to big if reqd */
  cvtLSwap(&temp.l);

  /* swap words for vms to ieee */
  pS = &temp.s[0];
  tmpS = *pS;                       
  *pS = *(pS+1);
  *(pS+1) = tmpS; 

  if (!slcIocIsLittleEndian) {
	/* if it is big endian just manipulate using the struct*/
	/* test exponent for 0,1 or 2 (underflow) */
	if (temp.b.exp > 2) 
	  temp.b.exp-=2;
	else /* if exp < 3 then return a 0 */
	  temp.f = 0;
  }else{
	/* if it is little endian use a pointer cast to the correct bit-mapping*/
	/* test exponent for 0,1 or 2 (underflow) */
	ltemp = (le_float_tu*)&temp;
	if (ltemp->b.exp > 2) 
	  ltemp->b.exp-=2;
	else /* if exp < 3 then return a 0 */
	  ltemp->f = 0;
	}

  /* now put the final result into r4Out_p */
  memcpy(r4Out_p, &temp, 4);
}

/*=============================================================================

  Name: cvtFloatIEEEToVMS

  Abs:  convert a ieee float ( 4-byte) to a vms (4-byte) float

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
        const void *          r4In_p        read only  pointer to the native float variable being converted
        void *                r4Out_p       read/write pointer to the output VMS float variable location

  Rem:  tests the exponent for NaN, Inf, neq zero, then swaps bytes and words
		the original is not altered if the r4in_p and r4Out_p point to different memory
		locations.  This routine should work if the r4In_p  = r4Out_p;
        Can be used for host -> vms.

  Side: uses cvtLSwap to noop the swap if the host is little-endian

  Ret:  none

=======================================================================================*/
void cvtFloatIEEEToVMS(const void * r4In_p, void* r4Out_p)
{      
  unsigned short tmpS, *pS;
  be_float_tu temp;
  le_float_tu* ltemp;
  
  memcpy(&temp, r4In_p, 4);

  if (!slcIocIsLittleEndian) {
	/* test exponent for NaN, Infinity, neg zero... */
	if (temp.b.exp == 255)
	  temp.f = INDEF;
	else if (temp.b.exp == 0){
	  if (temp.b.man == 0)
		temp.f = 0.0;
	  else
		temp.b.exp+= 1;
	}
	else {
	  if (temp.b.exp > 254)
		temp.f = INDEF;
	  else {
		temp.b.exp+=2;
	  }
	}
  }
  else {
	/* if it is little endian use a pointer cast to the correct bit-mapping*/
	ltemp = (le_float_tu*)&temp;
	/* test exponent for NaN, Infinity, neg zero... */
	if (ltemp->b.exp == 255)
	  ltemp->f = INDEF;
	else if (ltemp->b.exp == 0){
	  if (ltemp->b.man == 0)
		ltemp->f = 0.0;
	  else
		ltemp->b.exp+= 1;
	}
	else {
	  if (ltemp->b.exp > 254)
		ltemp->f = INDEF;
	  else {
		ltemp->b.exp+=2;
	  }
	}
  }


  /* swap for big endian to little endian if reqd */
  cvtLSwap(&temp.l);

  /* swap words for ieee to vms */
  pS = &temp.s[0];
  tmpS = *pS;                       
  *pS = *(pS+1);
  *(pS+1) = tmpS; 

  /* now copy the result into r4Out_p */
  memcpy(r4Out_p, &temp, 4);
}
  
/*=============================================================================

  Name: cvtFromVMS

  Abs:  convert an array of VMS data to native format

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
		const char            format        read only  data format of input array 
		const int             width         read only  data width in bytes of an element of the array
        const int             count         read only  number of elements in the array
		const void *          vmsData_p     read only  pointer to the vms data array
        void * const          nativeData_p             pointer to the native data array

  Rem: 
      The format argument is a single character:

	  'R' means call cvtFloatVMSToIEEE or cvtFloatIEEEToVMS for each element in the array. 
	  'I' or 'Z' and width=2 means call cvtSSwapMove for each element in the array 
	  'I', 'Z', and width=4 means call cvtLSwapMove for each element in the array 
	  'A' means 4-byte chars array, do nothing
	  'T' means VMS time value, do nothing - we assume VMS Time
	  'S' means char array, do nothing

  Side: 

  Ret: none

=======================================================================================*/
void cvtFromVMS(const char format, const int width, 
				const int count, const void * vmsData_p,
				void * const nativeData_p)
{
  int cnt;
  char *in_p;
  char *out_p;

  in_p = (char *)vmsData_p;
  out_p = (char *)nativeData_p;

  switch (format) {
  case 'R':
	/* convert and move real values to native array */
	DEBUGPRINT(DP_DEBUG, cvtUtilFlag,("CVT:converting reals from VMS:\n") );
	for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) {
	  cvtFloatVMSToIEEE(in_p, out_p);
	}
	break;
  case 'Z':
  case 'I':
	/* convert and move integer values to native array */
	DEBUGPRINT(DP_DEBUG, cvtUtilFlag,("CVT:converting ints from VMS, size = %d:\n", width) );
	if (width==2) {
	  for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) 
		cvtSSwapMove(in_p, out_p);
	}
	else if (width==4) {
	  for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) 
		cvtLSwapMove(in_p, out_p);
	}
	else {
	  for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) 
		memcpy(out_p, in_p, width);
	}
	break;
  case 'S':
  case 'A':
  case 'T':
	DEBUGPRINT(DP_DEBUG, cvtUtilFlag,("CVT:converting time or chars from VMS - simple copy:\n") );
	/* these data must simply be copied to native array */
	for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) {
	  memcpy(out_p, in_p, width);
	}
	break;
  }
}
 
/*=============================================================================

  Name: cvtToVMS

  Abs:  convert an array of native data to VMS format

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
		const char            format        read only  data format of input array 
		const int             width         read only  data width in bytes of an element of the array
        const int             count         read only  number of elements in the array
		const void *          nativeData_p  read only  pointer to the native data array
        void * const          VMSData_p             pointer to the vms data array

  Rem: 
      The format argument is a single character:

	  'R' means call cvtFloatVMSToIEEE or cvtFloatIEEEToVMS for each element in the array. 
	  'I' or 'Z' and width=2 means call cvtSSwapMove for each element in the array 
	  'I', 'Z', and width=4 means call cvtLSwapMove for each element in the array 
	  'A' means 4-byte chars array, do nothing
	  'T' means VMS time value, do nothing - we assume VMS Time
	  'S' means char array, do nothing

  Side: 

  Ret: none

=======================================================================================*/
void cvtToVMS(const char format, const int width, 
			  const int count, const void * nativeData_p, 
			  void * const vmsData_p)
{
  int cnt;
  char *in_p;
  char *out_p;
  
  in_p = (char *)nativeData_p;
  out_p = (char *)vmsData_p;

  switch (format) {
  case 'R':
	/* convert and move real values to native array */
	DEBUGPRINT(DP_DEBUG, cvtUtilFlag,("CVT:converting reals to VMS:\n") );
	for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) {
	  cvtFloatIEEEToVMS(in_p, out_p);
	}
	break;
  case 'Z':
  case 'I':
	/* convert and move integer values to native array */
	DEBUGPRINT(DP_DEBUG, cvtUtilFlag,("CVT:converting ints to VMS, size = %d:\n", width) );
	if (width==2) {
	  for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) 
		cvtSSwapMove(in_p, out_p);
	}
	else if (width==4) {
	  for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) 
		cvtLSwapMove(in_p, out_p);
	}
	else { /* if width is not 2 or 4, this is an error - just copy bytes */
	  for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) 
		memcpy(out_p, in_p, width);
	}
	break;
  case 'S':
  case 'A':
  case 'T':
	/* these data must simply be copied to native array */
	DEBUGPRINT(DP_DEBUG, cvtUtilFlag,("CVT:converting time or chars to VMS - simple copy:\n") );
	for (cnt=0; cnt<count; cnt++, in_p+=width, out_p+=width) {
	  memcpy(out_p, in_p, width);
	}
	break;
  default: DEBUGPRINT(DP_ERROR, cvtUtilFlag,("cvtToVMS:format invalid: %c\n",format));
  }

}
/*=============================================================================

  Name: cvtULongtoUShorts

  Abs:  split an unsigned long into two unsigned short components, one high order, 
         one low order

  Args: Type	              Name          Access	   Description
        -------------------   -----------	---------- ----------------------------
        const unsigned long   thelong       read only  long to be split
        unsigned short *      hishort_p     read/write pointer to the output high order u short
        unsigned short *      lowshort_p    read/write pointer to the output low order u short

  Rem: 

  Side: 

  Ret:  none

=======================================================================================*/
void cvtULongtoUShorts(const unsigned long thelong,
					   unsigned short * hishort_p,
					   unsigned short * lowshort_p)
{
  *lowshort_p = (unsigned short) (thelong & 0x0000ffff);
  *hishort_p = (unsigned short) ((thelong & 0xffff0000)>>16);
}
