/*=============================================================================
 
  Name: msgQUtil.c
        messageQSend - low-level: send a message to a message queue
		messageQRecv - low-level: receive a message from a message queue
		messageQCreate - create a message queue
        messageQDestroy - destroy a mesage queue
        messageQRelease  - release all buffers then destroy the queue
        msgQMsgRecv - hi level: receive from a queue
        msgQMsgReply - hi level: send a reply to a queue
        msgQGetBuffer - get a message buffer from the proper pool
		msgQFreeBuffer - return a message buffer to the proper pool
        msgQMsgProcess - Process mail message based on function code

  Abs:  slc-aware thread utilities for using epics message queues
        refer to web page LCLS Controls | Facilities | SLC-Aware IOC 
		| General Purpose Utilities
 

  Auth: 09-dec-2004, diane fairley (dfairley):
  Rev:  02 feb 2005, Debbie Rogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int msgQUtilFlag = DP_INFO;
#endif

/* c includes */
#include <string.h>        /* memset, memcpy */

/* epics includes */
#include <freeList.h>      /* freeListCalloc, freeListFree */
#include <epicsTime.h>     /* epicsTimeShow */

/* shared slc micro includes */
#include "micrdef.hc"     /* MICR_EXNET_ error codes */
#include "msgdef.h"       /* MSG_NOMBX error code    */
#include "msgfunc.hc"     /* RESPONSE bit            */
#include "slcMsg.h"       /* IOC_STOP                */

/* slc global vars */
#include    "slcCmlog.h"  /* slcCmlogLogMsg          */
#include    "slcGlob.h"   /* slcThreads              */
#include    "slcUtil.h"   /* slcGetThreadEnumByTid   */
#include    "cvtUtil.h"   /* for swapping Iss in msgQSLCReply */
#include    "sysUtil.h"   /* for Find_index_c() prototype     */


/* function prototypes */
#include "msgQUtil.h"

/*===========================================================

  Name: messageQSend

  Abs:  send a message queue message

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        slcThread_te        thrd_e      read only  thread enumeration of the
		                                           message queue to send to
        void *              msg_p       read/write message structure being sent 
        const int           msgBytes    byte count of the message structure

  Rem: Utility for tasks to send message to another threads message queue.
       Handles any size msg structure up to the maxsize the 
       message queue was created with.  

       This routine will return an error if the thread sending to does not 
	   exist, if the mq does not exist, or if the message is too big for the 
	   queue.  All are fatal problems that indicate a programmer error or an 
	   entire slcIOC failure.  no reason to try to re-send if this function 
	   call returns an error.

  Side: issues all relevent error messages with slcCmloglogMsg

  Ret: iss vmsstat_t error codes MICR_OKOK, or MSG_NOMBX, MICR_SENDMBX

===========================================================*/
vmsstat_t messageQSend(const slcThread_te thrd_e, 
					   const epicsBoolean wait_at_queue,
					   const int msgBytes, 
					   const void *msg_p )
{
  int               rv; /* return variable */
  vmsstat_t         iss=MICR_OKOK;
  char              name_a[MAX_THREADNAME_SIZE];
  
  /* see that the input thread enum is valid */
  if (thrd_e >= slcInvalid)
  {
	iss = MSG_NOJOBTOKEN;
	slcCmlogLogMsg(iss,"NOJOBTOKEN",
				   "trying to send msg to an unknown thread.\n");
	return iss;
  }

  /* see that the message queue is valid */
  if (slcThreads_as[thrd_e].mqId_ps == NULL) 
  {
	iss = MSG_NOMBX;
	slcCmlogLogMsg(iss,"NOMBX",
				   "trying to send a message to an unknown message queue.\n");
	return iss;
  }

  /* now try to send the message */
  if (wait_at_queue == epicsTrue) {
	if ((rv=epicsMessageQueueSend(slcThreads_as[thrd_e].mqId_ps, 
								  (void *)msg_p, msgBytes))<0) {
	  iss = MICR_SENDMBX;
	  /* get the name of the thread we're sending to */
	  epicsThreadGetName(slcThreads_as[thrd_e].tid_ps, 
						 (char *)&name_a, MAX_THREADNAME_SIZE);
	  slcCmlogLogMsg(iss, "SENDMBX", 
					 "send message to %s message queue failed.\n",
					 &name_a);
	}
  }
  else {
	if ((rv=epicsMessageQueueTrySend(slcThreads_as[thrd_e].mqId_ps, 
									   (void *)msg_p, msgBytes))<0) {
	  iss = MICR_SENDMBX;
	  /* get the name of the thread we're sending to */
	  epicsThreadGetName(slcThreads_as[thrd_e].tid_ps, 
						 (char *)&name_a, MAX_THREADNAME_SIZE);
	  slcCmlogLogMsg(iss, "SENDMBX", 
					 "send message to %s message queue failed, or the queue may be full.\n",
					 &name_a);
	  /* debug code */
	  
	  /*	  if (epicsTimeGetCurrent(&currentTS)==epicsTimeERROR) {
		iss = MSG_BADTIME;
		slcCmlogLogMsg(iss, "BADTIME", 
					   "messageQSend cannot get the current time.\n");
	  } else {
		DEBUGPRINT(DP_ERROR,msgQUtilFlag,("send message to %s message queue failed, or the queue may be full; TIME:\n",
					 (char *)&name_a));
		epicsTimeShow (&currentTS, 1); 
		}*/
	}
  }
  return iss;
}

/*===========================================================

  Abs:  receive a message queue message

  Name: messageQRecv

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		slcThread_te        thrd_e      read only  thread enumeration for queue
        void*               msg_p       read/write message structure being sent 
        const int           msgBytes    byte count of the message structure
		epicsTimestamp*     currentTime_ps read/write pointer to current time value

  Rem: Utility for tasks to recv message from their own message queue.
        Handles any size msg structure

		if the currentTime_ps is NULL then this routine will not get a current time.

		This routine will return an error if the thread receiving from does not 
		exist, if the mq does not exist, or if the message is too big for the 
		receiving buffer, or if the epicsTimeGet current call returns an error.
		All are fatal problems that indicate a programmer error or an entire 
		slcIOC failure.  no reason to try to receive again
		if this function call returns an error.

  Side: handles all error checking. waits forever for a message

  Ret:  iss vmsstat_t error code MICR_OKOK, or MSG_NOMBX, MICR_READ_MBX,
        MICR_BADTIME

===========================================================*/
vmsstat_t messageQRecv(const slcThread_te thrd_e, 
					   const int msgBytes, 
					   epicsTimeStamp* currentTime_ps,
					   void* msg_p )
{
  int               rv; /* return variable */
  vmsstat_t         iss=MICR_OKOK;
  
  /* see that the input thread enum is valid */
  if (thrd_e >= slcInvalid)
  {
	iss = MSG_NOJOBTOKEN;
	slcCmlogLogMsg(iss,"NOJOBTOKEN",
				   "trying to send msg to an unknown thread.\n");
	return iss;
  }

  /* see that the message queue is valid */
  if (slcThreads_as[thrd_e].mqId_ps == NULL) 
  {
	iss = MSG_NOMBX;
	slcCmlogLogMsg(iss,"NOMBX",
				   "cannot recv message; its message queue does not exist.\n");
	return iss;
  }

  /* now try to receive the message */
  if ((rv=epicsMessageQueueReceive(slcThreads_as[thrd_e].mqId_ps, 
								   msg_p, msgBytes))<0) {
	iss = MICR_READ_MBX;
	slcCmlogLogMsg(iss, "READ_MBX", 
				   "recv message from message queue failed.\n");
  } 
  else if (currentTime_ps != NULL) {
	if (epicsTimeGetCurrent(currentTime_ps)==epicsTimeERROR) {
	  iss = MSG_BADTIME;
	  slcCmlogLogMsg(iss, "BADTIME", 
					 "messageQRecv cannot get the current time.\n");
	}
  }

  return iss;
}

/*===========================================================

  Abs:  create a message queue

  Name: messageQCreate()

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        const int           maxMsgs     read only  max number of messages this queue may hold
        const int           maxMsgBytes read only  max byte count of each message this queue may hold

  Rem: Utility for threads to create their own  message queue.
      
  Side:  handles all error checking and places the mqId into the 
	     the slcThreads_as global for the calling thread.

  Ret:  iss       vmsstat_t error codes MICR_OKOK, or MICR_CREMBX

===========================================================*/
vmsstat_t messageQCreate(const slcThread_te thrd_e, const int maxMsgs, const int maxMsgBytes)
{
  vmsstat_t     iss = MICR_OKOK;

  /* see that the input thread enum is valid */
  if (thrd_e >= slcInvalid)
  {
	iss = MICR_EEXIST;
	slcCmlogLogMsg(iss,"EEXIST",
				   "trying to receive msg but is not a known slcIOC thread.\n");
	return iss;
  }

  /* create a message queue for this thread */
  slcThreads_as[thrd_e].mqId_ps = epicsMessageQueueCreate( maxMsgs, maxMsgBytes);
  if (slcThreads_as[thrd_e].mqId_ps == NULL )
  {
	iss =  MICR_CREMBX;
	slcCmlogLogMsg(iss, "CREMBX","message queue create failed.\n" );
  }
  return iss;
}
/*===========================================================

  Abs:  release all message in the queue, then destroy  message queue 

  Name: messageQRelease

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		slcThreads_te       thrd_e      read only  enumeration of the thread 
      
  Rem: release all message in the queue, then destroy the message queue
       ONLY FOR THREADS WORKING WITH MESSAGE SERVICE Messages that are allocated from the
	   message service free lists.

  Side: sets the slcThreads_as[].mqId_ps to NULL after storing the mqid in a local var.
        this is done to protect other jobs from accessing a message queue that is in the
		process of being destroyed.

  Ret: none

===========================================================*/
void messageQRelease(const slcThread_te thrd_e )
{
  msgptr_ts msgptr_s;
  int rv = 1;
  epicsMessageQueueId mqId_ps = NULL;

  /* see that the input thread enum is valid */
  if (thrd_e < slcInvalid)
  {
	/* see that the message queue is valid */
	if (slcThreads_as[thrd_e].mqId_ps != NULL) 
	{
	  mqId_ps = slcThreads_as[thrd_e].mqId_ps;
      slcThreads_as[thrd_e].mqId_ps = NULL;
	  /* read thru the queue and release message buffers! */
      while (rv > 0)
	  {
		if ((rv=epicsMessageQueueTryReceive(mqId_ps, &msgptr_s, MAX_MSGPTR_BYTES))>0)
		  msgQFreeBuffer(msgptr_s.msgbytes, msgptr_s.msgptr_p);
      }
	  epicsMessageQueueDestroy(mqId_ps);
	}
  }
}

/*===========================================================

  Abs:  destroy a message queue message

  Name: messageQDestroy

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
		slcThreads_te       thrd_e      read only  enumeration of the thread 
      
  Rem: destroy the message queue - Used by Database Service threads that do not
       use a memory pool for messages.

  Side: sets the slcThreads_as[].mqId_ps to NULL after storing the mqid in a local var.
        this is done to protect other jobs from accessing a message queue that is in the
		process of being destroyed.

  Ret: none

===========================================================*/
void messageQDestroy(const slcThread_te thrd_e )
{
  epicsMessageQueueId mqId = NULL;

  /* see that the input thread enum is valid */
  if (thrd_e < slcInvalid)
  {
	mqId = slcThreads_as[thrd_e].mqId_ps;
 
    slcThreads_as[thrd_e].mqId_ps = NULL;
 	/* see that the message queue is valid */
	if (mqId != NULL) 
	{
	  epicsMessageQueueDestroy(mqId);
	}

  }
}

/*===========================================================

  Name: msgQMsgRecv

  Abs:   receive a msg from queue, return the msg buffer pointer, set destination, 

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcThread_te        thrd_e      read only  thread enumeration
		void *              msg_pp      read/write pointer to the pointer to the 
		                                           incoming message

  Rem:  this routine recieves a message from a queue, and passes back a pointer 
        to the message buffer. that message buffer may be a small or large buffer, 
		hence the void * pointer type.  the calling function must figure the 
		actual buffer size from the msgheader.datalen field

  Side: sets the thread destination field according to the incoming message

  Ret:  vmsstat_t messageQRecv return code see messageQRecv header

===========================================================*/
vmsstat_t msgQMsgRecv(const slcThread_te thrd_e,  
					  epicsTimeStamp* currentTime_ps,
					  void **const msg_pp)
{
  msgptr_ts msgptr_s;
  vmsstat_t iss = MICR_OKOK;
  
  memset ( &msgptr_s, 0, sizeof(msgptr_s) );
  iss = messageQRecv( thrd_e, sizeof(msgptr_s), currentTime_ps, &msgptr_s);
  if ( SUCCESS(iss) )
  {
    *msg_pp =(void *)msgptr_s.msgptr_p;
	/* save the source SLC process name for use by slcCmlogLogMsg */
    /* though this may be a lrgmail_ts, it still has the msgheader.source in the same place */
	memcpy(&slcThreads_as[thrd_e].destination, ((msgmail_ts*)(*msg_pp))->msgheader.source, 4);

  }
  return (iss);
}

/*===========================================================

  Name: msgQMsgReply

  Abs:   send the message buffer to the queue, reset the thread destination field
         wait at the queue for room if there is none

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcThread_te        thrd_e      read only  thread enumeration
		msgmail_ts*         msg_ps      read only  pointer to the outgoing message

  Rem:  must cast the buffer pointer to a msgmail to get the data byte counts from
        the msgmail_ts header.

  Side: sets the thread destination field to the default after sending the message

  Ret:  vmsstat_t messageQSend return code see messageQSend header

===========================================================*/
vmsstat_t msgQMsgReply(const slcThread_te thrd_e, const void* msg_p)
{
  msgptr_ts msgptr_s;
  vmsstat_t iss = MICR_OKOK;
  msgmail_ts* msg_ps;

  msg_ps = (msgmail_ts*)msg_p;
  msgptr_s.msgptr_p = msg_ps;
  msgptr_s.msgbytes = sizeof(msg_ps->job_id)+sizeof(fwd_hdr_ts)+
	                  sizeof(msgheader_ts)+(msg_ps->msgheader.datalen*2);
  /* send the message, wait at queue if there is no room */
  iss = messageQSend( thrd_e, epicsTrue,  sizeof(msgptr_s), &msgptr_s );
  slcThreads_as[thrd_e].destination =  V017_UL;

  return (iss);
}

/*===========================================================

  Name: msgQGetBuffer

  Abs:  get a message buffer 

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        unsigned long       bytes       read only   minimum number of bytes required in buffer
		void **             msg_pp      read/write  pointer to the pointer to the message buffer

  Rem:  gets a  message buffer from the correct memory pool based on the number of bytes requested

  Side:

  Ret: vmsstat_t return code 
         MICR_OKOK  - alls well
		 MICR_MALLOC- could not get memory

===========================================================*/
vmsstat_t  msgQGetBuffer( const unsigned long bytes, void **const msg_pp) 
{
  vmsstat_t iss = MICR_OKOK;
  void *buf_p;

  if (bytes <= MAX_SMALL_MSG_BYTES) {
	buf_p = freeListCalloc(slcSmMsgPvt_p);

	if (!buf_p) {
	  iss = MICR_MALLOC;
	  slcCmlogLogMsg(iss, "MALLOC",
					 "unable to malloc small msg buffer from freelist.\n"); 
	}
  }
  else {
	buf_p = freeListCalloc(slcLrgMsgPvt_p);

	if (!buf_p) {
	  iss = MICR_MALLOC;
	  slcCmlogLogMsg(iss, "MALLOC",
					 "unable to malloc large msg buffer from freelist.\n"); 
	}
  }

  *msg_pp = buf_p;
  DEBUGPRINT(DP_DEBUG, msgQUtilFlag, ("%s got a msg buffer size %lu\n", 
									  epicsThreadGetNameSelf(), bytes));
  return iss;
}

/*===========================================================

  Name: msgQFreeBuffer

  Abs:  free a message buffer 

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        unsigned long       bytes       read only   minimum number of bytes required in buffer
		void *              msg_p       read/write pointer to the message buffer

  Rem:  releases a message buffer back to the correct memory pool 
        based on the size of the buffer 

  Side: 

  Ret: none

===========================================================*/
void  msgQFreeBuffer(const unsigned long bytes, const void* msg_p) 
{
  if (msg_p != NULL) {
	if (bytes <= MAX_SMALL_MSG_BYTES)
	  freeListFree(slcSmMsgPvt_p, (void *)msg_p);
	else
	  freeListFree(slcLrgMsgPvt_p, (void *)msg_p);
	DEBUGPRINT(DP_DEBUG, msgQUtilFlag, ("%s free'd a msg buffer size %lu\n", 
										epicsThreadGetNameSelf(), bytes));
  }
}

/*===========================================================

  Name: msgQSlcReply

  Abs:  hi-level utility for sending an slc reply message to msgSend thread 
        To be used ONLY for sending reply messages to SLC via msgSend thread, called after
        the sMsg_ps->reqdata member is filled with converted data.
 
        This utility waits at the msgSend messageQ until there is room for the message

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        const int           func        read only  function code of the reply message
		const int           bytes       read only  data length in BYTES of the message data
  const	msgmail_ts          rMsg_ps     read/write ptr to the received message header
		msgmail_ts          sMsg_ps     read/write ptr to the reply (send) message header

  Rem:  this routine prepares the SLC Reply message headers and sends the reply message to
        the msgSend thread. The RESPONSE bit is set in the return func code. 

		see msgHdlr.c for example usage

  Side: The message data must already be converted and placed in the 
        sMsg_ps->reqdata member BEFORE
        calling this routine to send the data.

  Ret: vmsstat_t iss the vms style return status

===========================================================*/
vmsstat_t  msgQSlcReply(const int func, const int bytes, const void* rMsg_p, void* sMsg_p) 
{
  msgmail_ts* sMsg_ps; 
  msgmail_ts* rMsg_ps;
  char  *microname = slcGetIOCName();

  /* cast the pointers to the msgmail_ts struct so we can access the members */
  rMsg_ps = (msgmail_ts*)rMsg_p;
  sMsg_ps = (msgmail_ts*)sMsg_p;

  /* fill in the reply message headers */
  sMsg_ps->job_id = rMsg_ps->job_id;
  /* put incoming time here, msgSend will replace with outgoing time*/
  sMsg_ps->msgheader.timestamp[0] = rMsg_ps->msgheader.timestamp[0];
  sMsg_ps->msgheader.timestamp[1] = rMsg_ps->msgheader.timestamp[1];
  memcpy(&sMsg_ps->msgheader.dest, &rMsg_ps->msgheader.source, 4);
  memcpy(&sMsg_ps->msgheader.source, microname, 4);
  sMsg_ps->msgheader.func         = func | RESPONSE;

  /* Copy tcpfwdhdr from the incoming msg into the outgoing msg */
  memcpy(&sMsg_ps->tcpfwdhdr, &rMsg_ps->tcpfwdhdr, sizeof(fwd_hdr_ts));
  
  sMsg_ps->msgheader.datalen = bytes/2;

   /* send the msg to the msgSend thread */
  return ( msgQMsgReply(msgSend, sMsg_ps) );

}

/*===========================================================

  Name: msgQSlcIssReply

  Abs:  hi-level utility for sending an slc iss reply message to msgSend thread 
        This routine assumes the message contains ONLY the iss value, and that the
        message is to be sent to the SLC via the msgSend thread.

        This utility waits at the msgSend messageQ until there is room for the message


  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
  const	msgmail_ts          rMsg_ps     read/write ptr to the received message header
		vmsstat_t           iss         read only  the iss to be returned in the reply message

  Rem:  this routine gets a small message buffer, prepares the SLC Reply message headers,
        and send the iss reply message to the msgSend thread. Calls the msgQSlcReply utility
        to fill in message headers and set RESPONSE bit.
   
		see bsacHdlr.c for example usage

  Side: only useful for ISS only reply messages for the SLC control system

  Ret:  vmsstat_t iss the vms style return status

===========================================================*/
vmsstat_t  msgQSlcIssReply( const void * rMsg_p, const vmsstat_t iss) 
{
  vmsstat_t   thisiss = MICR_OKOK;
  msgmail_ts* rMsg_ps;
  msgmail_ts* sMsg_ps = NULL;  /* immed reply msg to send */

  /* cast the pointer to the msgmail_ts struct so we can access the members */
  rMsg_ps = (msgmail_ts*)rMsg_p;

  /* get a buffer for the reply data */
  thisiss = msgQGetBuffer( sizeof(msgmail_ts),(void*) &sMsg_ps);
  if ( SUCCESS(thisiss) )
  {	
	/* convert iss & copy to req'ed data area of sMsg */
	cvtLSwap((unsigned long *)&iss);
    memcpy(&sMsg_ps->reqdata, &iss, sizeof(iss));

	/* fill in the reply message headers */
    thisiss = msgQSlcReply(rMsg_ps->msgheader.func, sizeof(iss), rMsg_p, (void *)sMsg_ps);
  } 
  return (thisiss);
}


/*=============================================================================
 
  Name: msgQMsgProcess.c

  Abs:  Process mail message based on function code

  Args: Type                Name        Access     Description
        ------------------- ----------- ---------- ----------------------------
        unsigned long       nr_func     read_only   No of items in array 
        slcScpFunc_ts       slc_func_as read-only   Array of SLC scp functions
        slcThread_te        thread_e    read-only   Thread enum
        msgmail_ts          rMsg_ps     read-write  Mail message
        unsigned long       update_p    read-write  Flag SLC db update needed

  Rem: The purpose of this function is to process a message recieved from the 
       incoming message queue. The message is processed based on the function
       code which is in the header of the message. The format of the function
       is as follows:

         bit 0-7 : function code
         bit 8-14: job id 
         bit 15  : response flag

  Side: None
  
  Ret:    
          MICR_OKOK       - Successful operation
          MSG_NOJOBTOKEN  - Failure, due to unknow thread

=============================================================================*/
vmsstat_t msgQMsgProcess( slcThread_te                thread_e, 
                          unsigned long               nr_func,
                          const slcScpFunc_ts * const scp_func_as,
                          const void          * const rMsg_p,
                          void                * const sMsg_p,
                          unsigned long       * const update_p )
{
    vmsstat_t            status = MICR_OKOK;  /* status return                    */
    unsigned short       reply;               /* relpy flag                       */
    unsigned short       func;                /* function code from msg           */
    slcJob_te            job_e;               /* job enum                         */
    unsigned long        i_func;              /* function index                   */
    int                  nbytes=0;            /* number of bytes                  */
    unsigned long        timeout;             /* timeout in milliseconds          */
    char                *cmd_c;               /* ascii function (NULL terminated) */
    slcThread_ts        *thread_ps=NULL;
    msgmail_ts          *recvMsg_ps=(msgmail_ts *)rMsg_p;
    vmsstat_t          (*routine_pf)(int                 * const nbytes_p,
                                     const void          * const rMsg_ps,
                                     void                * const sMsg_ps,
                                     const char          * const cmd_c,
                                     const unsigned long * const timeout_p,
                                     unsigned long       * const update_p);

    *update_p = 0;
    thread_ps = slcGetThread(thread_e);
    if ( !thread_ps ) {
       status = MSG_NOJOBTOKEN;
       slcCmlogLogMsg(status,
                      "NOJOBTOKEN",
                      "msgQMsgprocess trying to process a msg from an unknown thread.\n");
       return(status);
    }

    job_e = slcGetJobEnumByThreadEnum(thread_e);
    DEBUGPRINT(DP_INFO, msgQUtilFlag,
               ("msgQMsgProcess: incoming %s msg func code = %x\n", 
                 slcJobs_as[job_e].jobName_a,
                 recvMsg_ps->msgheader.func) );

    /*
    ** First get the function code from the func
    ** found in the message received from the scp or
    ** locally. Then search the input table of supported magnet scp functions
    ** for this slc-aware hjob and return an index into the table.
    ** NOTE: if a match is not found then a -1 is returned
    */
    func   = recvMsg_ps->msgheader.func; /* & FUNC_CODE;*/
    reply  = recvMsg_ps->msgheader.func & RESPONSE;
    i_func = Find_index_c( (void *)&scp_func_as[0].func,
                           sizeof(slcScpFunc_ts),
                           nr_func,
                           (void *)&func, /* key */ 
                           sizeof(func),
                           sizeof(scp_func_as[0].ascii_func) ); /* key offset */
	DEBUGPRINT(DP_INFO, msgQUtilFlag,
			   ("msgQMsgProcess: Find_index_c found i_func = %lu\n",
				i_func) );

    /* If match found then process the message function */
    if ( func == IOC_STOP ) 
      slcStopThread(thread_e);
    else if ( MATCH_FOUND(i_func) && scp_func_as[i_func].routine_pf ) {
       routine_pf = scp_func_as[i_func].routine_pf;
       timeout    = scp_func_as[i_func].timeout;
       cmd_c      = (char *)scp_func_as[i_func].ascii_func;
       if ( reply ) {
          status = (*routine_pf)(&nbytes,rMsg_p,sMsg_p,cmd_c,&timeout,update_p);
		  DEBUGPRINT(DP_INFO, msgQUtilFlag,
					 ("msgQMsgProcess: sending reply = %u for func %.10s\n",
					  reply, cmd_c) );
          status = msgQSlcReply((int)func,nbytes,rMsg_p,sMsg_p);
       }
       else
		  DEBUGPRINT(DP_INFO, msgQUtilFlag,
					 ("msgQMsgProcess: Calling routine for func = %.10s\n",
					  cmd_c) );
          status = (*routine_pf)(&nbytes,rMsg_p,NULL,cmd_c,&timeout,update_p);

    }else {
      DEBUGPRINT(DP_INFO, msgQUtilFlag,
                ("msgQMsgProcess: Received unknown message from %s\n",
                 slcJobs_as[job_e].jobName_a) );
    }
    return(status);
}
