/*=============================================================================
 
  Name: slcAsync.c  
        Public Utility functions:
        -------------------------
        slcAsyncInit
        slcAsyncGetJMSK
        slcAsyncGetFuncIxByName
        slcAsyncSetSendStats
        slcAsyncDumpGlob
        slcAsyncSetAsyncExists
        slcAsyncGetAsyncExists
        slcAsyncSetCYCL
        slcAsyncSetMTRL
        slcAsyncSetMTRC
        slcAsyncSetMAXT
        slcAsyncGetCycleFcn
        slcAsyncProcChk1
        slcAsyncProcChk2
        slcAsyncProcChk3
        slcAsyncReady
        slcAsyncMeterDbSend
        slcAsyncLockStats
        slcAsyncUnlockStats
        slcAsyncSetFcnStats
        slcAsyncGetThreadEnum
        slcAsyncExit

        Local functions:
        ----------------
        slcAsyncInitVerify
        slcAsyncSetDbStats


  Abs:  There are two types of utility funcitions in this file.
        1) General purpose utilities, many of which access job global.
        2) Functions called by cstrAsync to access job global which is 
           primarily slcJobFunc_as.  

        Job Global is never accessed directly by any functions other than
        those in this file.  So, in that sense, these funcitions encapsulate
        Job Global.

  Auth: 29-Mar-2005, Ron MacKenzie 

  Rev:  20-Sep-2005, Stephanie Allison

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------

  Mod:  (newest to oldest)  
  07-Apr-2005, Ron MacKenzie (ronm)
     asyncExists flag is now in this file, not in slcGlob.
  01-Apr-2005, Ron MacKenzie (ronm)
     Add slccstrDumpGlob() for calling from ioc consonle.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
int slcFlag = DP_FATAL;     /* Set to DP_INFO for production */  
#endif

/* Turns on debug printf's for asyncReady and Lock/unlock if defined */
/*
#define TESTREADYLOCKS 0
*/
/* c includes */
#include    <stdlib.h>                                              /* getenv */
#include    <string.h>                                              /* memcpy */

/* epics includes */ 
#include    <epicsTime.h>                                   /* epicsTimeStamp */
#include    <cantProceed.h>                              /* callocMustSucceed */

/* slc IOC  includes */
#include    "slcGlob.h"                        /* slcThreads_as, slcIocName_a */
#include    "slcTypes.h"                                         /* int4u etc */
#include    "dblist.h"                                    /* dblget proto etc */
#include    "dbl.h"                                       /* dblget dblput    */
#include    "dbupdate.h"                                  /* dbupdate proto   */
#include    "slcUtil.h"                                /* slcGetJobEnumByName */
#include    "slcAsync.h"                            /* public funcs from here */
#include    "slcCmlog.h"                                    /* slcCmlogLogMsg */
#include    "asyncdef.hc"                                   /* error codes    */
#include    "opdef.hc"                                      /* error codes    */
#include    "msgdef.h"                  /* NOTE: .h extension..error codes    */
#include    "timeUtil.h"                                    /* timeEpicsToVms */

/* local function prototypes */

LOCAL int slcAsyncInitVerify(void);

LOCAL int slcAsyncSetDbStats(const int asyncFuncIx,
                             const vmsstat_t dbUpdateStatus,
                             const epicsTimeStamp * const currentTime_ps);

/* defines */


#define  ASYN_MGNT_HSTA_MASK 0x00000001        /* CSTR:HSTA bit for magnet job    */
#define BIT0_MASK            0x00000001        /* shift a bit over then mask against this*/

/* Async cycling function information */
typedef struct {  
  slcJob_te            slcJobId_e;    /* enum into slcJob_as for this job             */     
  char                 funcName_c[5]; /* null terminated 4-character func name        */ 
  short                cyclTimeSec;   /* cycle time from CSTR:CYCL                    */ 
  short                mtrlSec;       /* dpupdate meter length in seconds             */
  short                mtrcSec;       /* dbupdate meter max # dbupdates in mrtl       */
  short                maxtSec;       /* dbupdate meter max time between dbupdates    */
  epicsTimeStamp       meterStartTime_s;     /*   time metering started                      */  
  unsigned long        meterCount;           /*   # of calls to MeterDbSend in mrtl          */
  vmsstat_t            lastDbupdateStatus;   /*   last database update status                */  
  epicsTimeStamp       updateTimeStamp_s;    /* timestamp of last dbupdate                   */
  epicsTimeStamp       cycleTimeStamp_s;     /* last time this cycl function was executed    */
  float                betweenCycle;         /* time between last 2 cycles for ELPS          */ 
  unsigned long        numFuncExe;           /* total # function executions in calc interval */ 
  unsigned long        numVmsMsg;            /* # of function executions from vms msg interv */ 
  unsigned long        numDbUpdates;         /* # of dbupdates in calc interval              */ 
  unsigned long        numDbUpdateFail;      /* # of dbupdates that failed  in calc interval */ 
  epicsTimeStamp       sendTimeStamp_s;      /* time when last msg sent attempt by cstr to hdlr */  
  unsigned long        sendErrCount;         /* number of failures in msg send               */
} slcJobFunc_ts;     

/**********************File scoped data (static makes it private to file) ** *****************/
static unsigned long   asyncFmsk = 0;            /* CSTR:FMSK */
static epicsBoolean    asyncExists = epicsFalse; /* Flag that exec/others use to tell if we're setup*/
static slcJobFunc_ts * slcJobFunc_as = NULL;     /* Job Global: Pointer to array of structures*/
static int             numCyclFunc = 0;          /* number of cycling functions for ioc */
static epicsMutexId    asyncRWMutex;             /* Protects async job global */

/*===========================================================

  Name: slcAsyncInit

  Abs:  Called by slcAsync to initialize "Job Global".
  
  Args: None

  Rem:  Reads CSTR database and initializes slcJobFunc_as
  
  Side: Allocates slcJobFunc_as
 
  Ret:  0 = success
       -1 = failure



===========================================================*/
int slcAsyncInit(void)
{
  int retStat = 0;

  /****************** CMLOG EXAMPLES ************

 Remember to make the message you generate similar to that associated with the slc message
 code you choose.  In this first example, the message code is ASYNC_INIT. 

 Don't forget to put a newline at the end of the text (vXworks needs it or you'll get ugly
 messages on the conole.


 * This is a simple way it works:
   ------------------------------
    slcCmlogLogMsg(ASYNC_INIT, "INIT",  "Initialization Failed.\n");

 * The parameters are like printf (variable number of them)
   --------------------------------------------------------
      slcCmlogLogMsg(iss, "INIT",  
        "Init Failed. Cycling Job %4.4s from CSTR SLC DB didn't match any in slcJob_as\n ", dbdata_p );

 * This one is cool.  It logs file and line number where we are... 1234 is the line number 
      You must use the correct cmlog browser config file to display them, though.
   ------------------------------------------------------------------------------------

  slcCmlogLogMsgFile(OP_STARTUP, "STARTUP",  __FILE__,  1234,  "Startup TEST MSG: Enter Init. INFO Severity\n");

   ****************** END CMLOG EXAMPLES */



 /******************************* CODE ********************************/

  numCyclFunc = 0;       /* filescoped.  On restart, init here is needed*/

  /* Indicate that global has not been set up ok */
  slcAsyncSetAsyncExists(epicsFalse);

  retStat = slcAsyncInitVerify(); 
  if (retStat != 0) goto egress;         /* get out if init. failed */

  /* Create mutex for use in protecting Async Job Global */
  asyncRWMutex = epicsMutexCreate();


  /* Indicate that global has been set up ok */
  slcAsyncSetAsyncExists(epicsTrue);

egress:
  if (retStat != 0) {
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Init FAILED!!!!!!!!!!!!\n"));
    slcCmlogLogMsg(ASYNC_INIT, "INIT",  "Initialization Failed.\n");
  }
  return retStat;  
}

/*===========================================================

  Name: slcAsyncInitVerify

  Abs:  Called by slcAsync to Initialize "Job Global" and 
        Verify what is being put in there from the slc DB.
  
  Args: None

  Rem:  Reads CSTR database and initializes slcJobFunc_as
  
  Side: Allocates slcJobFunc_as
 
  Ret:  epicsFalse (error)
        epicsTrue (success)



===========================================================*/
LOCAL int slcAsyncInitVerify(void)
{
  vmsstat_t       iss       = ASYNC_OKOK;
  int             ii;                       /* loop counter */
  int             funcIx;                   /* loop counter */
  epicsBoolean listFinished = epicsFalse;
  char tmpChar_a[8];
  char *         currJob  = NULL;                   /* current Job */
  epicsTimeStamp tempTimeStamp_s;
  slcJob_te      job_e;                   /* job table enum (index) */
  unsigned long  asynJmsk;

  /* General purpose lists.  Might be reused */

  DBLIST(dblist_ps, dbhEntry_ts);     
  DBLIST(dbdata_ps, long);  

  /* Specific one time usage lists */

  DBLIST(dblistCYCL_ps, dbhEntry_ts);     
  DBLIST(dbdataCYCL_ps, short);  

  DBLIST(dblistMTRC_ps, dbhEntry_ts);     
  DBLIST(dbdataMTRC_ps, short);  

  DBLIST(dblistMTRL_ps, dbhEntry_ts);     
  DBLIST(dbdataMTRL_ps, short);  

  DBLIST(dblistMAXT_ps, dbhEntry_ts);     
  DBLIST(dbdataMAXT_ps, short);  

  /*--------------------------------  code --------------------------*/


  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncInit!\n"));

  if (!(slcAsyncGetJMSK(&asynJmsk) )) {
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: getJMSK FAILED\n"));
    iss = ASYNC_INIT;
    goto egress;
  }

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: JMSK is: %x \n", (unsigned int)asynJmsk) );

  /* Get CNAM from the database.  
   * Use it to determine how many cycl functions there are. Store in global*/
  
  DBCLEAR(dblist_ps);
  DBCLEAR(dbdata_ps);
  if (!SUCCESS(iss=dblist ((const char *)"CSTR", 
			   (const char *)"1   ", 
			   (const char *)"CNAM",
			   (const dbheader_ts **const) &dblist_ps))) goto egress;
  if (!SUCCESS(iss=dblget ((const dbheader_ts *const)dblist_ps,
                           (dbheader_ts **const)&dbdata_ps)))  goto egress;

  /****************************************************************************************
   * Go through list of CNAM:CYCL and compute how many cycl functions there are and calloc 
   * slcJobFunc based on that number.
   * Remember,
   * Cycl functions must be grouped together for a given job.  
   * NO gaps allowed and must start at first entry in database
   ****************************************************************************************/

  for (ii = 0; ii < SLC_MAX_N_ASYN*2; ii=ii+2) {

    if (memcmp(DBPTR(dbdata_ps, ii), "NONE", 4) != 0) {       /* If not end of real function names */
      numCyclFunc++;                              /* Found a new function */
      if (listFinished) {                         /* If end already readhed, */
        iss = ASYNC_INIT;                         /*  must be a gap in list of names. Error */
        slcCmlogLogMsg(iss, "INIT",  "Initialization Failed because of gap in CSTR:CNAMs in SLC DB\n");
	goto egress;
      }
    }
    else                                         /* else, we're past end of functions */
      listFinished = epicsTrue;
  } 

  DEBUGPRINT(DP_INFO, slcFlag, ("CSTR_A: %d Cycling Functions read from SLC DB\n",numCyclFunc));

  if (numCyclFunc == 0) {
    iss = ASYNC_INIT;
    slcCmlogLogMsg(iss, "INIT",  "Initialization Failed because no Cycling Functions found in SLC DB\n");
    goto egress;
  }

  /* Allocate Job-Func global table.  Point to it with a pointer to array of structures.  We access it like
   * an array of structs, so we name the pointer (xxx_as) */

  slcJobFunc_as = (slcJobFunc_ts *)callocMustSucceed
                   ( numCyclFunc, sizeof(slcJobFunc_ts),"no memory for slcJobFunc_ts in slcAsyncInit\n");

  /* 
   * Get db lists of CYCL, MTRC, MTRL, MAXT.  
   * We'll verify and save these into slcJobFunc_as 
   */

  if (!SUCCESS(iss=dblist ((const char *)"CSTR", 
			   (const char *)"1   ", 
			   (const char *)"CYCL",
			   (const dbheader_ts **const) &dblistCYCL_ps))) goto egress;
  if (!SUCCESS(iss=dblget ((const dbheader_ts *const)dblistCYCL_ps,
                           (dbheader_ts **const)&dbdataCYCL_ps)))  goto egress;
  /* MTRC */

  if (!SUCCESS(iss=dblist ((const char *)"CSTR", 
			   (const char *)"1   ", 
			   (const char *)"MTRC",
			   (const dbheader_ts **const) &dblistMTRC_ps))) goto egress;
  if (!SUCCESS(iss=dblget ((const dbheader_ts *const)dblistMTRC_ps,
                           (dbheader_ts **const)&dbdataMTRC_ps)))  goto egress;

  /* MTRL */

  if (!SUCCESS(iss=dblist ((const char *)"CSTR", 
			   (const char *)"1   ", 
			   (const char *)"MTRL",
			   (const dbheader_ts **const) &dblistMTRL_ps))) goto egress;
  if (!SUCCESS(iss=dblget ((const dbheader_ts *const)dblistMTRL_ps,
                           (dbheader_ts **const)&dbdataMTRL_ps)))  goto egress;

  /* MAXT */

  if (!SUCCESS(iss=dblist ((const char *)"CSTR", 
			   (const char *)"1   ", 
			   (const char *)"MAXT",
			   (const dbheader_ts **const) &dblistMAXT_ps))) goto egress;
  if (!SUCCESS(iss=dblget ((const dbheader_ts *const)dblistMAXT_ps,
                           (dbheader_ts **const)&dbdataMAXT_ps)))  goto egress;


  /*****************************************************************************************
   * Go through CSTR:CNAM list again.
   * Copy cycling func. names from slc db list into global.  Remember,
   * cycl functions must be grouped together for a given job.  NO
   * gaps allowed and must start at first entry in database
   ****************************************************************************************/ 

  for (ii = 0, funcIx = 0; ii < numCyclFunc*2; ii=ii+2, funcIx++) {

    /* Compute index into slcJobs_as[] for this job. 
     * Store it into job global.
     */
    memcpy(tmpChar_a,DBPTR(dbdata_ps, ii),4);  /* saved job name */  
    tmpChar_a[4] = '\0';                       /* next func wants null terminated string */
    job_e = slcGetJobEnumByName(tmpChar_a);    
    slcJobFunc_as[funcIx].slcJobId_e = job_e;

    /* Store number of cycling functions for the job in job table.
     * Also store first cycl func index for the job into job table.
     */
    if (currJob == NULL || memcmp(currJob, tmpChar_a, 4) != 0) { /* new job found in list */
      slcJobs_as[job_e]. numCyclFunc   = 1;                            /* job's # cycl func = 1 */
      slcJobs_as[job_e].frstCyclFuncIx = funcIx;
    }
    else
      slcJobs_as[job_e]. numCyclFunc++;     

    currJob = (char*)DBPTR(dbdata_ps, ii);

    /* Store CNAM into job global */
    memcpy(slcJobFunc_as[funcIx].funcName_c, (char *)DBPTR(dbdata_ps, (ii+1)),
           sizeof(unsigned long)); /* warning: might be 3 chars */
    
    DEBUGPRINT(DP_INFO, slcFlag, ("CSTR_A: CNAM FUNC copied to job global = : %4.4s \n", (char *)DBPTR(dbdata_ps, (ii+1))));

    if (slcJobFunc_as[funcIx].slcJobId_e == invalidJob) {
      iss = ASYNC_INIT;
      slcCmlogLogMsg(iss, "INIT",  
        "Init Failed. Cycling Job %4.4s from CSTR SLC DB didn't match any in slcJob_as\n ", (char *)DBPTR(dbdata_ps, ii) );
      goto egress;
    }

    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: JOB ENUM set in global = %d \n", slcJobFunc_as[funcIx].slcJobId_e));

    /*
     * Initialize epicsTime items in async job global.
     */

    if (epicsTimeGetCurrent(&tempTimeStamp_s)==epicsTimeERROR) {
      iss = ASYNC_INIT;
      slcCmlogLogMsg(iss, "INIT", "Init Failed because epicTimeGetCurrent Failed at 1\n");
      goto egress;
    }

    memcpy(&(slcJobFunc_as[funcIx].meterStartTime_s),  &tempTimeStamp_s, sizeof(epicsTimeStamp));
    memcpy(&(slcJobFunc_as[funcIx].updateTimeStamp_s), &tempTimeStamp_s, sizeof(epicsTimeStamp));
    memcpy(&(slcJobFunc_as[funcIx].cycleTimeStamp_s),  &tempTimeStamp_s, sizeof(epicsTimeStamp));
    memcpy(&(slcJobFunc_as[funcIx].sendTimeStamp_s),  &tempTimeStamp_s, sizeof(epicsTimeStamp));

    /* 
     * Initialize a couple of misc. things in async job global.
     */
    slcJobFunc_as[funcIx].betweenCycle = 0.0;
    slcJobFunc_as[funcIx].lastDbupdateStatus = ASYNC_OKOK;

    /*
     * Store CYCL and Metering parameters for this function from db list into Job Global.
     */ 
    slcAsyncSetCYCL(funcIx, *DBPTR(dbdataCYCL_ps, funcIx));
    slcAsyncSetMTRL(funcIx, *DBPTR(dbdataMTRL_ps, funcIx));
    slcAsyncSetMTRC(funcIx, *DBPTR(dbdataMTRC_ps, funcIx));
    slcAsyncSetMAXT(funcIx, *DBPTR(dbdataMAXT_ps, funcIx));
  }

  egress:

  /* Free these lists since we're not going to use them again */

  DBFREE(dblist_ps);
  DBFREE(dbdata_ps);
  DBFREE(dblistCYCL_ps);
  DBFREE(dbdataCYCL_ps);
  DBFREE(dblistMTRL_ps);
  DBFREE(dbdataMTRL_ps);
  DBFREE(dblistMTRC_ps);
  DBFREE(dbdataMTRC_ps);
  DBFREE(dblistMAXT_ps);
  DBFREE(dbdataMAXT_ps);

  if (!SUCCESS(iss)) {
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Init FAILED in slcAsyncInitVerify !!!!!!!!!!!!\n"));
    return -1;
  }
  return 0;

} /* end slcAsyncInitVerify */



/*=============================================================================

  Name: slcAsyncGetJMSK

  Abs:  read CSTR JMSK.  Update value in file scope and return it too.

  Args: Type		        Name	    Access	    Description
         -------------------    ---------   ----------      ------------------
        unsigned long           jmsk_p      R/W             JMSK value returned

  Rem:  Reads the CSTR 1 JMSK field from database and returns this 'Z' value in
        input arg. jmsk.
		To read from the slc db, first call dblistalloc() to allocate a pointer list 
          (dblist_ps) that will contain dictionary entries (pointers of type 
          dbhEntry_ts, see dbdict.h) 
        Next calls dblistalloc() to allocate a data list (dbdata_ps)that will contain
          the data read from the slc db (pointed to by related dblist_ps entry). 
		Next calls dblist("CSTR", "1   ", "JMSK", &dblist_ps) to populate the pointer
          list with the dictionary entry ponter.NOte that unit numbers must be left justified
		  and blank padded out to 4 chars.
        Next calls dblget() with both lists to populate the data list with data pointed
          to by pointer list.
		Next establishes a pointer to the location that contains the data within
          the data list.
		Next copies the data into R/W arg jmsk from data list using memcpy().
        Frees both lists (if their pointers are valid) by calling dblistfree()
        Returns jmsk  

  Side: Include dblist.h (dbheader_ts, dblist, dblistalloc, dblistfree), 
                dbl.h    (dblget, dblput)

  Ret: epicsTrue if database read successfully, otherwise epicsFalse indicates
       that jmsk is returning as invalid. In either case, the lists are attempted
       to be freed.
       
==============================================================================*/
epicsBoolean slcAsyncGetJMSK (unsigned long * const jmsk_p)
{
  DBLIST(dblist_ps, dbhEntry_ts);      
  DBLIST(dbdata_ps, long);
  vmsstat_t     iss;
  epicsBoolean  retStat = epicsFalse;

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncGetJMSK!\n"));

  if (SUCCESS(iss=dblist ((const char *)"CSTR",
                          (const char *)"1   ",
                          (const char *)"JMSK",
                          (const dbheader_ts **const) &dblist_ps))){
    if (SUCCESS(iss=dblget ((const dbheader_ts *const) dblist_ps,
                            (dbheader_ts **const) &dbdata_ps))){
      *jmsk_p = (unsigned long)*DBPTR_FIRST(dbdata_ps);
      retStat = epicsTrue;
    }
  }  
  
  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: JMSK just fetched in GetJMSK =  %x!\n", (unsigned int)*jmsk_p));

  DBFREE(dblist_ps);      
  DBFREE(dbdata_ps);

  return retStat;
}


/*=============================================================================

  Name: slcAsyncGetCDSC

  Abs:  read CSTR CDSC.  Update value in file scope and return it too.

  Args: Type		        Name	    Access	    Description
         -------------------    ---------   ----------      ------------------
        unsigned long           cdsc_p      R/W             CDSC value returned

  Rem:  Reads the CSTR 1 CDSC field from database and returns this 'Z' value in
        input arg. cdsc.
		To read from the slc db, first call dblistalloc() to allocate a pointer list 
          (dblist_ps) that will contain dictionary entries (pointers of type 
          dbhEntry_ts, see dbdict.h) 
        Next calls dblistalloc() to allocate a data list (dbdata_ps)that will contain
          the data read from the slc db (pointed to by related dblist_ps entry). 
		Next calls dblist("CSTR", "1   ", "CDSC", &dblist_ps) to populate the pointer
          list with the dictionary entry ponter.NOte that unit numbers must be left justified
		  and blank padded out to 4 chars.
        Next calls dblget() with both lists to populate the data list with data pointed
          to by pointer list.
		Next establishes a pointer to the location that contains the data within
          the data list.
		Next copies the data into R/W arg cdsc from data list using memcpy().
        Frees both lists (if their pointers are valid) by calling dblistfree()
        Returns cdsc  

  Side: Include dblist.h (dbheader_ts, dblist, dblistalloc, dblistfree), 
                dbl.h    (dblget, dblput)

  Ret: epicsTrue if database read successfully, otherwise epicsFalse indicates
       that cdsc is returning as invalid. In either case, the lists are attempted
       to be freed.
       
==============================================================================*/
epicsBoolean slcAsyncGetCDSC (unsigned short * const cdsc_p)
{
  DBLIST(dblist_ps, dbhEntry_ts);      
  DBLIST(dbdata_ps, short);
  vmsstat_t     iss;
  epicsBoolean  retStat = epicsFalse;

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncGetCDSC!\n"));

  if (SUCCESS(iss=dblist ((const char *)"CSTR",
                          (const char *)"1   ",
                          (const char *)"CDSC",
                          (const dbheader_ts **const) &dblist_ps))){
    if (SUCCESS(iss=dblget ((const dbheader_ts *const) dblist_ps,
                            (dbheader_ts **const) &dbdata_ps))){
      *cdsc_p = (unsigned short)*DBPTR_FIRST(dbdata_ps);
      retStat = epicsTrue;
    }
  }  
  
  DEBUGPRINT(DP_DEBUG, slcFlag, 
			 ("CSTR_A: CDSC just fetched in GetCDSC =  %ux!\n", (unsigned short)*cdsc_p));

  DBFREE(dblist_ps);      
  DBFREE(dbdata_ps);

  return retStat;
}
/*==============================================================================

  Name: slcAsyncGetFuncIxByName

  Abs:  Public utility that gets the index to async job global based on cycling
        function name.

  Args:        
       Type:        pointer to string
       Name:        name_p
       Access:      R
       Description: Name of cycling function


  Rem: This utility is useful if a cycling handler needs to call one of the 
       utilities in this file that require JobFuncIx and they don't already
       know what that index is.

       Normally, a handler wouldn't need to call this utility because cstrAsync
       sends the async function index and name in the request packet to the handler.

  side:none

  Ret: slcJobFuncIx. Index to async job global.
       -1 if index not found.

  ==============================================================================*/
int slcAsyncGetFuncIxByName(const char * const name_p)    
{
  int ii;

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncGetFuncIxByName!\n"));

  for (ii = 0; ii < numCyclFunc; ii++) {
    if (strcmp((char *)slcJobFunc_as[ii].funcName_c , name_p)==0)
    {
	  return ii;
    }
  }

  DEBUGPRINT(DP_ERROR, slcFlag, ("CSTR_A: ERROR: didn't find async func name for %6.6s \n", name_p));

  return -1;     /* error condition */

}


/*==============================================================================

  Name: slcAsyncSetSendStats()

  Abs:  Public utility that sets message send statistics in job global for
        a given cycling function.

  Args: 
       Type:        int
       Name:        asyncFuncIx
       Access:      R
       Description: index into slcJobFunc_as for the cycling function.

       Type:        vmstat_t
       Name:        status
       Access:      R
       Description: indicates the message send vms status.

       Type:        epicsTimeStamp *
       Name:        currentTime_ps
       Access:      R
       Description: current time of day.

  Rem: 

  side:none

  Ret: int
        0 = success
       -1 = failure

  ==============================================================================*/
int slcAsyncSetSendStats(const int asyncFuncIx, const vmsstat_t status,
                         const epicsTimeStamp * const currentTime_ps) 
{


 /*----------------------------------------- code ------------------------------*/

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncSetSendStats. vmsstat(hex) = %x \n",
				 (unsigned int)status));

  if (slcAsyncLockStats() != 0)  return -1;

       
  /* If status indicates failure
   * increment send error counter (sendErrCount ) in job global.  This is not zeroed on success.
   */

  if (!SUCCESS(status)) {
    slcJobFunc_as[asyncFuncIx].sendErrCount++;
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: BAD MSG SEND STATUS in slcAsyncSetSendStats. Expected? \n"));
  }
  /* 
   * Update send timestamp (sendErrCount) in job global.  This is set for both success and failure.
   */
  slcJobFunc_as[asyncFuncIx].sendTimeStamp_s = *currentTime_ps;

  slcAsyncUnlockStats();       /* Unlock Job Global. */

  return 0;
}


/*==============================================================================

  Name: slcAsyncDumpGlob()

  Abs:  dump global when called by iocshell

  Args: none

  Rem: 

  side:

  Ret: none

  ==============================================================================*/
void slcAsyncDumpGlob(void)
{
  int ii;
  time_t today = 0;
  struct tm *DayToday;

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: entering slcAsyncDumpGlob !!!!\n"));

  printf ("***Async Global (slcJobFunc_as) ***\n\n");

  for (ii = 0; ii < numCyclFunc; ii++) {
    printf ("\n\nCycling function number %d\n", ii);
    printf ("ENUM JOB  FUNC CYCL MTRL MTRC MAXT \n");
    printf ("----------------------------------\n");

    printf ("%d  %s   %4.4s %d    %d  %d    %d\n", 
            slcJobFunc_as[ii].slcJobId_e,
            slcJobs_as[(slcJobFunc_as[ii].slcJobId_e)].jobName_a  ,
            (char*)&(slcJobFunc_as[ii].funcName_c),
            (int)slcJobFunc_as[ii].cyclTimeSec,
            (int)slcJobFunc_as[ii].mtrlSec,
            (int)slcJobFunc_as[ii].mtrcSec,
            (int)slcJobFunc_as[ii].maxtSec
            );

     epicsTimeToTime_t (&today, &(slcJobFunc_as[ii].meterStartTime_s));
     DayToday = localtime(&today);
     printf ( "meterStartTime is: %s ", asctime (DayToday) );  
     epicsTimeToTime_t (&today, &(slcJobFunc_as[ii].updateTimeStamp_s));
     DayToday = localtime(&today);
     printf ( "updateTimeStamp is: %s ", asctime (DayToday) );  
     epicsTimeToTime_t (&today, &(slcJobFunc_as[ii].cycleTimeStamp_s));
     DayToday = localtime(&today);
     printf ( "cycleTimeStamp is: %s ", asctime (DayToday) );  

     printf("---\n");
     printf("Associated Job Table entries for this cycling function\n");
     printf("Number of cycling functions = %d\n", 
       slcJobs_as[(slcJobFunc_as[ii].slcJobId_e)].numCyclFunc);
     printf("First cycling function index in slcJobFunc_as = %d\n",
       slcJobs_as[(slcJobFunc_as[ii].slcJobId_e)].frstCyclFuncIx);
  }

  return;

  /**********
    memcpy(&(slcJobFunc_as[ii].meterStartTime_s),  &tempTimeStamp_s, sizeof(epicsTimeStamp));
    memcpy(&(slcJobFunc_as[ii].updateTimeStamp_s), &tempTimeStamp_s, sizeof(epicsTimeStamp));
    memcpy(&(slcJobFunc_as[ii].cycleTimeStamp_s),  &tempTimeStamp_s, sizeof(epicsTimeStamp));

    epicsTimeToTime_t (time_t *dest, epicsTimeStamp *eTime)


    time_t today = 1101995558L;

    struct tm *DayToday;

    DayToday = localtime(&today);
    printf ( "date and time are: %s", asctime (DayToday) );
  ********/


}

/*==============================================================================

  Name: slcAsyncSetAsyncExists()

  Abs:  Public utility that sets asyncExists which is in file scope.

  Args: 
       Type:        epicsBoolean
       Name:        newVal
       Access:      R
       Description: new value to set it to.

  Rem: none

  side:none

  Ret: none

  ==============================================================================*/
void slcAsyncSetAsyncExists(const epicsBoolean newVal)    
{
  asyncExists = newVal;

 DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncSetAsyncExists!\n"));

  return;
}

/*==============================================================================

  Name: slcAsyncGetAsyncExists()

  Abs:  Public utility that gets val of asyncExists which is in file scope.

  Args: none

  Rem: 

  side:none

  Ret: Returns the value of asyncExists flag (which is file scoped).
       epicsFalse if global is not set up.
       epicsTrue  if global is set up.

  ==============================================================================*/
 epicsBoolean slcAsyncGetAsyncExists(void)    
{
#ifdef TESTREADYLOCKS
 DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncGetAsyncExists!\n"));
#endif

  return asyncExists;
}


/*==============================================================================

  Name: slcAsyncSetCYCL()

  Abs:  Public utility that sets CYCL into job global.

  Args:
       Type:        int
       Name:        jobFuncIx
       Access:      R
       Description: index into job func array

       Type:        short
       Name:        newVal
       Access:      R
       Description: Value to set it to.

  Rem: Validates value passed in.  Sets to default if out of range.

  side:none

  Ret: void


  ==============================================================================*/
void slcAsyncSetCYCL(const int jobFuncIx, short newVal)    
{
  short defaultVal = 0;

 DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncCYCL!\n"));


  /* test input values within range.  */

  if (jobFuncIx < 0 || jobFuncIx > numCyclFunc-1) {   
      slcCmlogLogMsg(ASYNC_BAD_ARG, "BAD_ARG",  
        "Bad Argument. jobFuncIx = %d when seeting CYCL value = %d \n ",
         jobFuncIx, newVal  );
      goto egress;
  }

  if (newVal < 0 ) {    
      slcCmlogLogMsg(ASYNC_BAD_ARG, "BAD_ARG",  
        "Bad Argument. jobFuncIx = %d, CYCL value = %d Set to %d \n ",
         jobFuncIx, newVal, defaultVal  );
      newVal = defaultVal;
  }

  slcJobFunc_as[jobFuncIx].cyclTimeSec = newVal;  /* Store CYCL for this function */

 egress:
  return;
}


/*==============================================================================

  Name: slcAsyncSetMTRL()

  Abs:  Public utility that sets MTRL into job global.

  Args: 
       Type:        int
       Name:        jobFuncIx
       Access:      R
       Description: index into job func array

       Type:        short
       Name:        newVal
       Access:      R
       Description: Value to set it to.

  Rem: Validates value passed in.  Sets to default if out of range.

  side:none

  Ret: void


  ==============================================================================*/
void slcAsyncSetMTRL(const int jobFuncIx, short newVal)    
{
  short defaultVal = 60;

 DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncSetMTRL!\n"));

  /* test input values within range.  */

  if (jobFuncIx < 0 || jobFuncIx > numCyclFunc-1) {   
      slcCmlogLogMsg(ASYNC_BAD_ARG, "BAD_ARG",  
        "Bad Argument. jobFuncIx = %d when seeting MTRL value = %d \n ",
         jobFuncIx, newVal  );
      goto egress;
  }

  if (newVal < 1 ) {    
      slcCmlogLogMsg(ASYNC_BAD_ARG, "BAD_ARG",  
        "Bad Argument. jobFuncIx = %d, MTRL value = %d Set to %d \n ",
         jobFuncIx, newVal, defaultVal  );
      newVal = defaultVal;
  }

  slcJobFunc_as[jobFuncIx].mtrlSec = newVal;

 egress:
  return;
}


/*==============================================================================

  Name: slcAsyncSetMTRC()

  Abs:  Public utility that sets MTRC into job global.

  Args: 
       Type:        int
       Name:        jobFuncIx
       Access:      R
       Description: index into job func array

       Type:        short
       Name:        newVal
       Access:      R
       Description: Value to set it to.

  Rem: Validates value passed in.  Sets to default if out of range.

  side:none

  Ret: void


  ==============================================================================*/
void slcAsyncSetMTRC(const int jobFuncIx, short newVal)    
{

  short defaultVal = 1;

 DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncMTRC!\n"));


  /* test input values within range.  */

  if (jobFuncIx < 0 || jobFuncIx > numCyclFunc-1) {   
      slcCmlogLogMsg(ASYNC_BAD_ARG, "BAD_ARG",  
        "Bad Argument. jobFuncIx = %d when seeting MTRC value = %d \n ",
         jobFuncIx, newVal  );
      goto egress;
  }

  if (newVal < 1 ) {    
      slcCmlogLogMsg(ASYNC_BAD_ARG, "BAD_ARG",  
        "Bad Argument. jobFuncIx = %d, MTRC value = %d Set to %d \n ",
         jobFuncIx, newVal, defaultVal  );
      newVal = defaultVal;
  }

  slcJobFunc_as[jobFuncIx].mtrcSec = newVal;  

 egress:
  return;
}

/*==============================================================================

  Name: slcAsyncSetMAXT()

  Abs:  Public utility that sets MAXT into job global.

  Args: 
       Type:        int
       Name:        jobFuncIx
       Access:      R
       Description: index into job func array

       Type:        short
       Name:        newVal
       Access:      R
       Description: Value to set it to.

  Rem: Validates value passed in.  Sets to default if out of range.

  side:none

  Ret: void


  ==============================================================================*/
void slcAsyncSetMAXT(const int jobFuncIx, short newVal)    
{
  short defaultVal = 0;

 DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncSetMAXT!\n"));


  /* test input values within range.  */

  if (jobFuncIx < 0 || jobFuncIx > numCyclFunc-1) {   
      slcCmlogLogMsg(ASYNC_BAD_ARG, "BAD_ARG",  
        "Bad Argument. jobFuncIx = %d when seeting MAXT value = %d \n ",
         jobFuncIx, newVal  );
      goto egress;
  }

  if (newVal < 0 ) {    
      slcCmlogLogMsg(ASYNC_BAD_ARG, "BAD_ARG",  
        "Bad Argument. jobFuncIx = %d, MAXT value = %d Set to %d \n ",
         jobFuncIx, newVal, defaultVal  );
      newVal = defaultVal;
  }

  slcJobFunc_as[jobFuncIx].maxtSec = newVal;  

 egress:
  return;
}

/*==============================================================================

  Name: slcAsyncGetCycleFcn()

  Abs:  For a given job, get the cycling function most ready to run.

  Args: 

       Type:        int
       Name:        firstIx
       Access:      R
       Description: tells us first cycling func index in async job global for the job.

       Type:        int
       Name:        numFunc
       Access:      R
       Description: tells us number of cycling func for the job.

       Type:        int *
       Name:        cyclIx2Exe
       Access:      W
       Description: job global index of function most ready to run.

       Type:        char *
       Name:        cyclName2Exe_c
       Access:      W
       Description: Points to 5 char string storage (null terminated)
                    name of function most ready to run.

       Type:        epicsTimeStamp *
       Name:        currentTime_ps
       Access:      W
       Description: Current time passed back to caller.

  Rem:  This is a public function called by cstrAsync.  
        Goes through all entries in Job Global for the current job as indicated 
        by 1stCycleFuncIx and numCycleFun.  It returns the index of the "most ready
        to run function in cycl2Exe (index) and  cyclIx2Exe output arguments.
        If one is not found, then -1 is returned in cycl2Exe and cyclIx2Exe is 
        unaffected.

  side: 

  Ret: -1 (error)
        0 (success)

==============================================================================*/
int slcAsyncGetCycleFcn(const int firstIx , const int numFunc,
                        const slcScan_ts * const scan_ps,
                        int * const cyclIx2Exe, 
			char* const cyclName2Exe_c,
                        epicsTimeStamp * const currentTime_ps)    
{
  int ii = 0;                    /* loop counter */
  unsigned long tmpInt = 0;      /* temporary    */
  unsigned long cyclPeriod = 0;  /* Cycle period */
  int timeSinceLastExeSec = 0;   /* time since last execution in seconds */
  int timeSinceLastMsendSec = 0; /* time since last message send in seconds */
  int timeSinceSec = 0;          /* greater of previous two                 */
  int mostDueSec = 0;            /* most number of past due seconds of any func on job */

  /*----------------------------------------- code -----------------------------*/

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A:************** Entering slcAsyncGetCycleFcn ix = %d numFunc = %d\n",
				 firstIx, numFunc));

  /* Check input argumements */
  if (firstIx < 0 || firstIx > SLC_MAX_N_ASYN-1 || numFunc <=0) {
    slcCmlogLogMsg(OP_BADARGS, "BADARGS", "Illegal call to slcAsyncGetCycleFcn\n");
    return -1;
  }

  *cyclIx2Exe = -1;         /* this says none found.  Pass back to caller */

  if (!slcAsyncReady()) return -1;

  /* Get current time for use later */
  if (epicsTimeGetCurrent(currentTime_ps)==epicsTimeERROR) {
      slcCmlogLogMsg(MSG_BADTIME, "BADTIME", "Unable to get Epics TimeOfDay loca 12\n");
      return -1;
  }

  if (slcAsyncLockStats() != 0)  return -1;

  asyncFmsk = scan_ps->fmsk;  /* store away for use by other utilities */

  for (ii=firstIx; ii<firstIx+numFunc; ii++) {

    /*****
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: CHECKING JOB %4.4s \n",
				   slcJobs_as[(slcJobFunc_as[ii].slcJobId_e)].jobName_a));
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: FIRST = %d \n", 
                                   slcJobs_as[(slcJobFunc_as[ii].slcJobId_e)].frstCyclFuncIx));
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: NUM = %d \n", 
                                   slcJobs_as[(slcJobFunc_as[ii].slcJobId_e)].numCyclFunc));
    ***/

    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: ******************************* CHECKING FUNCTION IX = %d \n",ii  ));

    if (slcJobFunc_as[ii].cyclTimeSec <= 0) continue;
    
    /* If this is the magnet job and it's hsta bit not set, 
     * loop around and check next function. Note: Only the magnet job supports being
     * HSTA's on and off, not the test job.
     */

    if ((memcmp(slcJobs_as[(slcJobFunc_as[ii].slcJobId_e)].jobName_a, "MGNT", 4) == 0) && 
        ((scan_ps->hsta & ASYN_MGNT_HSTA_MASK) == 0)) continue;

    /* if ii+1 bit is set in cmsk, then function is enabled.  
     * NOTE: CMSK is funny in the database.  
     * The first function is bit 1 (not zero as you might expect.
     */

    tmpInt = (scan_ps->cmsk)>>(ii+1);         /* shift bit in question into zeroTH */
    if ((tmpInt & BIT0_MASK) == 0) continue;  /* if bit not set, loop for next func*/

    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: %4.4s FUNCTION IS ENABLED \n",slcJobFunc_as[ii].funcName_c  ));

    /*
     * Get the cycling_period.
     * It is either CYCL or SCAN
     * Use SCAN if the bit for the function is set in MMSK and SCAN is less than CYCL.
     */

    cyclPeriod = slcJobFunc_as[ii].cyclTimeSec;

    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: MMSK IS %x \n",
                                   (unsigned int)scan_ps->mmsk));

    tmpInt = ((unsigned long)scan_ps->mmsk)>>(ii+1);  /* shift this func's bit into zeroTH bit */
    if ((tmpInt & BIT0_MASK) != 0) {     /* if bit set in MMSK for this function  */
      if (scan_ps->scan_a[ii] < cyclPeriod) {   /* if scan < cycl                        */
        DEBUGPRINT(DP_DEBUG, slcFlag, 
		   ("CSTR_A: SAW MMSK SET. Using SCAN = %d instead of CYCL = %d !!!!!!!!!!!!!!!!!!!!!\n",
                    scan_ps->scan_a[ii], (int)cyclPeriod));
	cyclPeriod = scan_ps->scan_a[ii];
      }
    }

    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: CYCLE PERIOD COMPUTED = %d\n", (int)cyclPeriod ));

    /* 
     * If Current_time -  last_execution_time > cycling_period, this function is ready to run
     * -or- 
     * If Current_time -  time_of_last_msg_send > cycling_period (don't keep trying if q full),
     *    this function is ready to run.
     *  Save greater of those last 2 items in timeSinceSec 
     */

    /* compute time since last execution. Warning: converting double return value to int 
     */
    timeSinceLastExeSec = epicsTimeDiffInSeconds( currentTime_ps, &slcJobFunc_as[ii].cycleTimeStamp_s); 

    /* compute time since last msg send so we don't keep trying if Q is full or error. 
     */
    timeSinceLastMsendSec =  epicsTimeDiffInSeconds(currentTime_ps, &slcJobFunc_as[ii].sendTimeStamp_s); 

    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: SINCELASTEXE = %d  SINCELAST_MSEND = %d MOSTDUE = %d\n", 
				   timeSinceLastExeSec, timeSinceLastMsendSec, mostDueSec )); 

    /* if time since last cycle execution AND time since last successful dbupdate are both  
     * greater than cycle period, use the greater of those two as our time since cycle executed. 
     */

    if  ((timeSinceLastExeSec >= cyclPeriod) &&  (timeSinceLastMsendSec >= cyclPeriod)) {
      if (timeSinceLastExeSec >  timeSinceLastMsendSec) 
        timeSinceSec = timeSinceLastExeSec;
      else
	timeSinceSec = timeSinceLastMsendSec;

      DEBUGPRINT(DP_DEBUG, slcFlag, 
		 ("CSTR_A: getCyclFcn checking if IX = %d ready to run. timeSince = %d  mostDueSec = %d \n", 
		  ii, timeSinceSec, mostDueSec));

      /* Pass back to caller, the index and name of the cycling function */
      if (timeSinceSec > mostDueSec) {           /* found one more due than last one found */
	mostDueSec = timeSinceSec;
        *cyclIx2Exe  = ii;
        strncpy(cyclName2Exe_c, slcJobFunc_as[ii].funcName_c, 5);
	DEBUGPRINT(DP_DEBUG, slcFlag, 
		   ("CSTR_A: getCyclFcn FOUND A CYCLE FUNC READY TO RUN.index = %d !!!!!!! \n", ii));
     }
    } 

  } /* end for */

  slcAsyncUnlockStats();

  if ( *cyclIx2Exe > 0) DEBUGPRINT(DP_DEBUG, slcFlag, 
	  ("CSTR_A: ***************GETCYCLFCN exiting!!!!!!! LETS RUN THIS FUNC IX = %d NAME = %s \n",
	  *cyclIx2Exe,  cyclName2Exe_c ));

  return 0;

} /* end slcAsyncGetCyclFcn() */


/*==============================================================================

  Name: slcAsyncProcChk1()

  Abs:  Update dblget data for UTIM, CTIM, and ELPS.

  Args: TODO *** FILL IN ARGS

  Rem: These secondaries are always sent (dbupdated) by the caller after we do
       this function without checking if they changed because  
       UTIM and CTIM change when this function runs.

  side:

  Ret: -1 (error)
        0 (success)

==============================================================================*/

int slcAsyncProcChk1(vmsTimeStamp_ta * const dataCTIM_aa,
                     vmsTimeStamp_ta * const dataUTIM_aa,
                     float           * const dataELPS_a)
{
  int retStat = 0;                         /* return status */
  int ii;                                  /* loop counter  */  

  /*----------------------------------- code ----------------------------------*/

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncProcChk1!\n"));

  if (!slcAsyncReady()) return -1;
  if (slcAsyncLockStats() != 0)  return -1;

  
  /* For each cycling function in the Global Area do the following:
   *   Put the last database update time from Global into the UTIM dblist.
   *   Put the last cycling function update time from Global into the CTIM dblist.
   *   Put elapsed time from Global in the ELPS dblist.
   * Note: Timestamps are converted to VMS format using  timeEpicsToVms()  before putting 
   *   them into the slc database.
   */

  for (ii = 0; ii < numCyclFunc; ii++) {
		 
   /* UTIM */

   if (timeEpicsToVMS(&slcJobFunc_as[ii].updateTimeStamp_s, dataUTIM_aa[ii]) == epicsTimeERROR) {
     /* TODO LOG CMLOG ERROR HERE */
     DEBUGPRINT(DP_ERROR, slcFlag, ("CSTR_A: timeEpicsToVMS failed for UTIM!!!!!!\n"));
     retStat = -1;
     goto egress;
   }

   /* CTIM */

   if (timeEpicsToVMS(&slcJobFunc_as[ii].cycleTimeStamp_s, dataCTIM_aa[ii]) == epicsTimeERROR) {
     DEBUGPRINT(DP_ERROR, slcFlag, ("CSTR_A: timeEpicsToVMS failed for CTIM!!!!!!\n"));
     retStat = -1;
     goto egress;
   }

    /* ELPS */

    dataELPS_a[ii] = slcJobFunc_as[ii].betweenCycle;
  }
  
 egress:

  slcAsyncUnlockStats();

  return retStat;

} /* end slcAsyncProcChk1 */


/*==============================================================================

  Name: slcAsyncProcChk2()

  Abs:  Update dblget data for NRUN, FAIL, PUPD, and PVAX

  Args:  TODO *** FILL IN ARGS

  Rem: 

  side:

  Ret: -1 (error)
        0 (success)

  ==============================================================================*/
int slcAsyncProcChk2(long  * const dataNRUN_a,
                     long  * const dataFAIL_a,
                     float * const dataPVAX_a,
                     float * const dataPUPD_a,
                     epicsBoolean * const NRUN_changed_p,
                     epicsBoolean * const FAIL_changed_p,
                     epicsBoolean * const PVAX_changed_p,
                     epicsBoolean * const PUPD_changed_p)
{
  int retStat = 0;                         /* return status */
  int         ii;                          /* loop counter  */
  float       tmpFloat;                    /* temporary     */
  
  /*----------------------------------- code ----------------------------------*/

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsycProcChk2!\n"));
  *NRUN_changed_p = epicsFalse;
  *FAIL_changed_p = epicsFalse;
  *PVAX_changed_p = epicsFalse;
  *PUPD_changed_p = epicsFalse;

  if (!slcAsyncReady()) return -1;
  if (slcAsyncLockStats() != 0)  return -1;
  
  /******************
  For each cycling function in the Global Area:

    If Number of Executions (NRUN) has changed in Global vs what's in the previous used data,
      put the new value from Global into the NRUN data and set dbUpdateNeeded flag 
      indicating that something changed 

    If number of failures (FAIL) has changed in Global vs what's in the previous used date, 
      put the new value from Global into the FAIL data and set dbUpdateNeeded flag that 
      something changed.

    Calculate Percent executions triggered by Vax Messages (PVAX). 
      If it has changed, store it in the PVAX data and set flag.
 
    Calculate Percent executions triggering successful database updates (PUPD). 
      If it has changed, store it in the PUPD data and set flag.

     Zero the counters for NRUN, FAIL, DBUPDATES, and VAXFUNCS in global.
  ****************/
  for (ii = 0; ii < numCyclFunc; ii++) {
    
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk2 checking cycl func %d \n", ii));

    if (dataNRUN_a[ii] != slcJobFunc_as[ii].numFuncExe) {
      dataNRUN_a[ii] = slcJobFunc_as[ii].numFuncExe;
      *NRUN_changed_p = epicsTrue;
      DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk2 sees that NRUN changed \n"));
    }

    /* FAIL */

    if (dataFAIL_a[ii] != slcJobFunc_as[ii].numDbUpdateFail) {
      dataFAIL_a[ii] = slcJobFunc_as[ii].numDbUpdateFail;
      *FAIL_changed_p = epicsTrue;
      DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk2 sees that FAIL changed \n"));
    }

    /* PVAX */

    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk2  slcJobFunc_as[ii].numVmsMsg = %d\n", 
				  (int)slcJobFunc_as[ii].numVmsMsg ));
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk2  slcJobFunc_as[ii].numFuncExe = %d\n", 
				  (int)slcJobFunc_as[ii].numFuncExe ));

    /* Be careful with use of Floats here.  Is there a problem with the casts here?
     * The items on the right are cast to int's. It does not get an arithmetic exception and
     * the resulting tmpFloat is a very large number if numFuncExe is zero. Why?      
     * My solution is to just check if numFuncExe is 0.
     */

    if (slcJobFunc_as[ii].numFuncExe == 0) {   /* prevent divide by zero */
      tmpFloat = 0.0;
      DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk numFuncExe was zero*******************XXXXXXX****\n"));
    }
    else {
      tmpFloat = 100.0 * (float)slcJobFunc_as[ii].numVmsMsg / (float)slcJobFunc_as[ii].numFuncExe;
      /*** This does get an arithmetic exception (note: no cast ) 
        tmpFloat = slcJobFunc_as[ii].numVmsMsg / slcJobFunc_as[ii].numFuncExe;
      ****/
    }

    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk2 tmpFloat = %f\n", 
				  tmpFloat ));

    if (tmpFloat != dataPVAX_a[ii]) {            /* Warning, testing equality of float */
      dataPVAX_a[ii] = tmpFloat;
      *PVAX_changed_p = epicsTrue;
      DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk2 sees that PVAX changed \n"));
    }

    /* PUPD */
    if (slcJobFunc_as[ii].numDbUpdates == 0) {
      tmpFloat = 0.0;
    }
    else {
      tmpFloat = 100.0 * (float)slcJobFunc_as[ii].numDbUpdates / (float)slcJobFunc_as[ii].numDbUpdates; 
    }
    if (tmpFloat != dataPUPD_a[ii]) {         /* Warning, testing equality of float */
      dataPUPD_a[ii] = tmpFloat;
      *PUPD_changed_p = epicsTrue;
      DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: slcAsyncProcChk2 sees that PUPD changed \n"));
    }

    /* Zero counters for  NRUN, FAIL, DBUPDATES, and VAXFUNCS.
     * The reason we do it here is that the "calculation interval" for these items
     * is the cycl length (CYCL) of CHK2.
     */

    slcJobFunc_as[ii].numDbUpdateFail = 0;
    slcJobFunc_as[ii].numFuncExe      = 0; 
    slcJobFunc_as[ii].numDbUpdates    = 0;
    slcJobFunc_as[ii].numVmsMsg       = 0;

  } /* end loop over cycling functions */
  
  slcAsyncUnlockStats();

  return retStat;
}


/*==============================================================================

  Name: slcAsyncProcChk3()

  Abs:  

  Args: 

  Rem: 

  side:

  Ret: -1 (error)
        0 (success)

  ==============================================================================*/
int slcAsyncProcChk3(void)

{
  int retStat = 0;                         /* return status */
  int         ii;                          /* loop counter  */
  float       tmpFloat;                    /* temporary     */
  
  /*----------------------------------- code ----------------------------------*/

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsycProcChk3!\n"));

  if (!slcAsyncReady()) return -1;

  return retStat;
}



/*==============================================================================

  Name: slcAsyncReady()

  Abs:  Check if the job global area has been set up.

  Args: none

  Rem: Simply returns the value of asyncExists flag by using the public accessor
       function.

  side:

  Ret: Returns the value of asyncExists flag (which is file scoped).
       epicsFalse if global is not set up.
       epicsTrue  if global is set up.


  ==============================================================================*/
epicsBoolean slcAsyncReady(void)    
{

  epicsBoolean retStat;

#ifdef TESTREADYLOCKS
  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncReady!\n"));
#endif
  
  retStat = slcAsyncGetAsyncExists();

  if (!retStat) {
    slcCmlogLogMsg(ASYNC_INIT, "INIT",  "Initialization Failed.\n");
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Global doesn't exist in AsyncReady. OK????\n"));
  } 

  return retStat;
}


/*==============================================================================

  Name: slcAsyncMeterDbSend()

  Abs:  

  Args:
       Type:        int
       Name:        asyncFuncIx
       Access:      R
       Description: Index of cycling function to be run

       Type:        slcReqType_te
       Name:        requestType
       Access:      R
       Description: tells if vms message or cstrAsync msg recvd.

       Type:        epicsBoolean
       Name:        dblputDone
       Access:      R
       Description: tells us if dblput was done before we got here.

  Rem: 

  side:

  Ret: -1 (error)
        0 (success)

==============================================================================*/
int slcAsyncMeterDbSend(const int asyncFuncIx,
                        const slcReqType_te requestType_e,
                        const epicsBoolean dblputDone)    
{
  int            retStat = 0;
  vmsstat_t      iss = ASYNC_OKOK;
  epicsTimeStamp currTimeStamp_s;                               /* current time */
  epicsBoolean   dbUpdateNeeded = epicsFalse;                   /* local flag   */
  double         timeDiff;                                   /* time difference */
  unsigned       long tmpInt = 0;                               /* temporary    */


  /*--------------------------------- code ------------------------------------*/

  DEBUGPRINT(DP_DEBUG, slcFlag, ("\nCSTR_A: Entered slcAsyncMeterDbSend!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"));

  /*********
  slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "Func IX = %d entering slcAsyncMeterDbSend.\n",asyncFuncIx);
  **********/

  /* Ignore an invalid function index */
  if (asyncFuncIx < 0) return retStat;
  
  if (!slcAsyncReady()) return -1;

  /*
  slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "Func IX = %d entering slcAsyncMeterDbSend.\n",asyncFuncIx);
  */

  if (epicsTimeGetCurrent(&currTimeStamp_s)==epicsTimeERROR) {
      DEBUGPRINT(DP_ERROR, slcFlag, ("CSTR_H: Error fetching epics time in MeterDbSend\n"));
      slcCmlogLogMsg(MSG_BADTIME, "BADTIME", "Unable to get Epics TimeOfDay at location MeterDbSend\n");
      return -1;
  }

  if (slcAsyncLockStats() != 0)  return -1;

  /* Check to see if dbupdate needs to be done.  
   * This could be done with one long OR statement. But, syntax is long for ease of debugging 
   */

  /* Calculate interval since begin meter time.
   * If the interval has finished, reset meter counter and start time 
   */

  if ((timeDiff = epicsTimeDiffInSeconds(&currTimeStamp_s, &slcJobFunc_as[asyncFuncIx].meterStartTime_s))
	    >= (double)slcJobFunc_as[asyncFuncIx].mtrlSec) { 
    slcJobFunc_as[asyncFuncIx].meterStartTime_s = currTimeStamp_s;
    slcJobFunc_as[asyncFuncIx].meterCount = 0;
    DEBUGPRINT(DP_INFO, slcFlag, ("CSTR_H: Meter interval restarted \n"));
    /*
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "Meter interval Reset in slcAsyncMeterDbSend.\n");
    */
  }

  /* Increment current meter count.  It really counts number of times we come into this 
   * routine (even if we decide not to do a dbupdate once we're in here).
   */
  slcJobFunc_as[asyncFuncIx].meterCount++;

  /* 
   * Do an update if last dbupdate failed 
   */
  if (!SUCCESS(slcJobFunc_as[asyncFuncIx].lastDbupdateStatus)) {
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: MeterDbSend updating due to lastDbupdateStatus bad \n"));
    dbUpdateNeeded = epicsTrue;
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "INFO: Updating because Last Update Failed in slcAsyncMeterDbSend.\n");
  }
  /* 
   * Do an update if no successful dbupdate has happened in last MAXT seconds. 
   */
  if ((timeDiff = epicsTimeDiffInSeconds(&currTimeStamp_s, &slcJobFunc_as[asyncFuncIx].updateTimeStamp_s))
	    >= (double)slcJobFunc_as[asyncFuncIx].maxtSec) {   
                                               /* maxtSec is short.  Convert to double with cast */
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: MeterDbSend updating due to no updates in MAXT seconds\n"));
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: MeterDbSend MAXT is %d \n",slcJobFunc_as[asyncFuncIx].maxtSec)); 
    dbUpdateNeeded = epicsTrue;
    /*
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "Updating because No updates in MAXT seconds.\n");
    */
  }
  /*
   * Do an update if this is not an ASYNC request (ie it is a VAX request).
   */
  if (requestType_e != slcAsync) {
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: MeterDbSend updating because not ASYNC function \n"));
    dbUpdateNeeded = epicsTrue;
    /* 
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "Updating because not Async Request.\n");
    */
  }

  /*
   *  Do an update if dblputDone is true and the number of successful database updates is less 
   *  than the allowed number (MTRC) in the last MTRL seconds.
   *  At cstrHdlr init. time, this test will be true and we'll do dbupdate.  
   */

  if ((dblputDone == epicsTrue) &&
      (slcJobFunc_as[asyncFuncIx].meterCount <=  slcJobFunc_as[asyncFuncIx].mtrcSec)) {
    DEBUGPRINT(DP_DEBUG, slcFlag, 
	       ("CSTR_A: MeterDbSend indicating DBUPDATE due to DBLPUTDONE and meterCount < mtrcSec\n"));
    dbUpdateNeeded = epicsTrue;
    /*
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "Updating because num successful updates < MTRC in MRTL.\n");
    */
  }      

  /* If bit set in FMSK, then do dbUpdate.
   * Are bits in FMSK same as index in slcAsyncJobFunc?  I think so because same as CMSK and CNAM.
   *
   * Note: On RMX, they got FMSK fresh from the DB here. We don't need to do that.
   * We got it in slcAsyncUpdSt2() which was called before we got here.
   */

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: MeterDbSend FMSK = %x\n", (unsigned int)tmpInt));

  tmpInt = asyncFmsk>>(asyncFuncIx);            /* shift bit in question into zeroTH */
  if ((tmpInt & BIT0_MASK) != 0) {
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: MeterDbSend updating due to BIT SET IN FMSK\n"));
    DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: MeterDbSend FMSK = %x\n", (unsigned int)tmpInt));
    dbUpdateNeeded = epicsTrue;
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "INFO: Updating because FMASK Seen as set..\n");
  }

  slcAsyncUnlockStats();

  /* Do the dbupdate if we passed one of the tests above.
   * Update DB stats in global only if we call dbupdate.
   */

  if (dbUpdateNeeded == epicsTrue) {        /* one of the criteria above was met. do dbupdate */
    DEBUGPRINT(DP_INFO, slcFlag, ("CSTR_A: slcAsyncMeterDbSend doing DBUPDATE for job id %d !\n", 
				  slcJobFunc_as[asyncFuncIx].slcJobId_e ))
    iss = dbupdate(slcJobFunc_as[asyncFuncIx].slcJobId_e);
    if (!SUCCESS(iss)) {
      DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_H: Error in dbupdate in slcAsyncMeterDbSend \n"));
    }
    retStat = slcAsyncSetDbStats(asyncFuncIx,    
                                iss,              /* db status              */
                                &currTimeStamp_s);
  }
  else {                   /* LEAVE THIS HERE. IT's NOT FOR DEBUG. WE'RE TELLING THAT WE'RE METERING */
    slcCmlogLogMsg(OP_MSGTXT, "MSGTXT",  "METERED DBUPDATE of func %s in slcAsyncMeterDbSend.\n", 
		   slcJobFunc_as[asyncFuncIx].funcName_c);
  }

  return retStat;

}  /* end MeterDbSend */


/*==============================================================================

  Name: slcAsyncSetDbStats()

  Abs:  Set database update statistics in job global for the cycling function as 
        indicated by input arguments.

  Args: 
       Type:        int
       Name:        asyncFuncIx
       Access:      R
       Description: the index in slcJobFunc_as  for the async function for which 
                    Stats are being set

       Type:        vmsstat_t
       Name:        dbUpdateStatus
       Access:      R
       Description: status code indicating the status of the last dbupdate() call.

       Type:        epicsTimeStamp *
       Name:        currentTime_ps
       Access:      R
       Description: current time of day.

  Rem: 

   This function is only called by slcAsyncMeterDbsend.  So, it is a static function.


  side:

  Ret: -1 (error)   
        0 (success). 

  ==============================================================================*/
LOCAL int slcAsyncSetDbStats(const int asyncFuncIx,
                             const vmsstat_t dbUpdateStatus,
                             const epicsTimeStamp * const currentTime_ps)
{

  /*---------------------------------- code -----------------------------------*/


  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncSetDbStats!\n"));

  if (slcAsyncLockStats() != 0)  return -1;
    
  /*
   * Increment either the database-update-failed counter or database-update-success counter for 
   * this job/function depending on dbUpdateStatus. ALSO
   * Set the last-database-update timestamp to current time input parameter 
   * if the dbUpdataStatus is good. 
   */
  if (SUCCESS(dbUpdateStatus)) {
    slcJobFunc_as[asyncFuncIx].numDbUpdates ++;
    slcJobFunc_as[asyncFuncIx].updateTimeStamp_s = *currentTime_ps;
  }
  else
    slcJobFunc_as[asyncFuncIx].numDbUpdateFail ++;

  /*
   * Set the last-database-update status to input parameter.
   */
  slcJobFunc_as[asyncFuncIx].lastDbupdateStatus = dbUpdateStatus;

  if (slcAsyncUnlockStats() != 0) return -1; 

  return 0;

}


/*==============================================================================

  Name: slcAsyncLockStats()

  Abs:  Lock Statistics for thread safe access protection.

  Args: none

  Rem:  This public function calls epicsMutexLock(id) to take out mutex on Global Area.   

        Note: That call blocks until the lock is available or an error occurs.  

  side:

  Ret: -1 (error)
        0 (success)

  ==============================================================================*/
int slcAsyncLockStats(void)    
{

  int retStat = 0;

  epicsMutexLockStatus eLockStat_e;                  /* epics status */

  /*-------------------------------- code -----------------------------------*/

#ifdef TESTREADYLOCKS
  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncLockStats!\n"));
#endif
  
  if (!slcAsyncReady()) {
    slcCmlogLogMsg(OP_LCKBLKTIMEOUT, "LCKBLKTIMEOUT",  "Can't get Lock because Async Global not ready.\n");
    retStat = -1;
    goto egress;
  }

  if (asyncRWMutex){
	eLockStat_e = epicsMutexLock(asyncRWMutex);
	if (eLockStat_e != epicsMutexLockOK) {
	  if (asyncRWMutex) {
		epicsMutexUnlock(asyncRWMutex);                   /* UNLOCK if lock failed */
		slcCmlogLogMsg(OP_LCKBLKTIMEOUT, 
					   "LCKBLKTIMEOUT","Get LOCK failed.Status = %d \n", eLockStat_e);
	  }
      retStat = -1;
    }
  }

 egress:

  return retStat;

}


/*==============================================================================

  Name: slcAsyncUnlockStats()

  Abs:  Unlock Statistics for thread safe access protection

  Args: none

  Rem: This public function calls epicsMutexUnlock(id) to release mutex on Global Area.   

        Note: That call blocks until the lock is released or an error occurs.  

  side:

  Ret: -1 (error)
        0 (success)

  ==============================================================================*/
int slcAsyncUnlockStats(void)    
{
#ifdef TESTREADYLOCKS
  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncUnlockStats!\n"));
#endif
  
  if (!slcAsyncReady()) {
    slcCmlogLogMsg(OP_LCKBLKFATAL, "LCKBLKFATAL",  "Can't release Lock because Async Global not ready.\n");
    return -1;
  }
  if (asyncRWMutex)
	epicsMutexUnlock(asyncRWMutex);

  /* TODO. Could call epicsMutexShow() here and put out an error message if it's not unlocked */

  return 0;

}


/*==============================================================================

  Name: slcAsyncSetFcnStats()

  Abs:  This public function is for use by all async threads.  It sets statistics in 
        the Global Area for the specified cycling function.

  Args: 
       Type:        int
       Name:        asyncFuncIx
       Access:      R
       Description: Index to slcJobFunc_as indicating which async function this request is for.

       Type:        epicsTimeStamp
       Name:        startTime_ps
       Access:      R
       Description: Passed by the caller so that the ELPS (elapsed) time can be calculated 
                    (current time - start_time (which is the time right before the function was executed)). 
                    startTime is the time right before function began and is set by caller 
                    (used as start time in calculating func execution elapsed time (ELPS)).

       Type:        slcReqType_te  
       Name:        requestType
       Access:      R
       Description: Indicates slcAsync or slcMsg.
It indicates if this is as a result of a vms message request or an async request.
It's an enum with two possible values: slcAsync or slcMsg. 

  Rem: 

  side:

  Ret: -1 (error)
        0 (success)

  ==============================================================================*/
int slcAsyncSetFcnStats(const int asyncFuncIx, 
			const epicsTimeStamp * const startTime_ps, 
                        const slcReqType_te requestType)    

{
  epicsTimeStamp currTimeStamp_s;                              /* current time */

  /*-------------------------------- code -------------------------------------*/

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncSetFcnStats for Index %d \n", asyncFuncIx));
  
  /* Ignore an invalid function index */
  if (asyncFuncIx < 0) return 0;

  if (!slcAsyncReady()) return -1;

  /* Get current time for use in ELPS calculation */

  if (epicsTimeGetCurrent(&currTimeStamp_s)==epicsTimeERROR) {
      DEBUGPRINT(DP_ERROR, slcFlag, ("CSTR_A: Error fetching epics time in SetFcnStats\n"));
      slcCmlogLogMsg(MSG_BADTIME, "BADTIME", "Unable to get Epics TimeOfDay at location SetFcnStats\n");
      return -1;
  }

  if (slcAsyncLockStats() != 0)  return -1;

  /* ELPS calculation here: converting double to float. OK? */

  slcJobFunc_as[asyncFuncIx].betweenCycle = epicsTimeDiffInSeconds(&currTimeStamp_s, startTime_ps); 
  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: setFcnStats calculated ELPS to be %f \n",
				 slcJobFunc_as[asyncFuncIx].betweenCycle ));


  /*
   * Store current time into in Global Area for this function (set last execution time to current time).
   */

  slcJobFunc_as[asyncFuncIx].cycleTimeStamp_s = currTimeStamp_s;  

  /* 
   * Increment NRUNS (execution counter) in Global Area for this function.
   */

  slcJobFunc_as[asyncFuncIx].numFuncExe++;

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: setFcnStats just incremented numFuncExe to %d \n", (int)slcJobFunc_as[asyncFuncIx].numFuncExe));

  /*
   * If slcRequestType indicates the call is from a Handler as a result of VMS message (slcMsg), then, 
   * increment number of hdlr functions in last calculation interval (and calculate PVAX). Otherwise,
   * slcRequestType indicates the call is from a Handler as a result of Async call (slcAsync), so, 
   * increment number of async functions in last calculation interval.
   */
  if (requestType == slcMsg) 
    slcJobFunc_as[asyncFuncIx].numVmsMsg++;

  /*** NONONO. numFuncExe is the total number of executions. Not number of async executions. 
  else if (requestType == slcAsync)
    slcJobFunc_as[asyncFuncIx].numFuncExe++;
  ***/

  /* 
   * Call slcAsyncUnlockStats() to release the mutex on the Global Area. Return if error.
   */
  if (slcAsyncUnlockStats() != 0) return -1;  

  return 0;
  
} /* end slcAsyncSetFcnStats */

/*==============================================================================

  Name: slcAsyncGetThreadEnum()

  Abs:  Get Thread enum based on input cycling function index.

  Args: 
       Type:        int
       Name:        cyclIx2Exe
       Access:      R
       Description: Index of cycling functio

  Rem: Cycling func ix is index to slcJobFuncIx.
       Special case for mgnt job - all IOC_ASYNC messages are sent to
       mgntAsync, not mgntHdlr. 

  side:

  Ret: slcThread_te thread enum of cycling function.

  ==============================================================================*/

slcThread_te slcAsyncGetThreadEnum(const int cyclIx2Exe)
{
  if (slcJobFunc_as[cyclIx2Exe].slcJobId_e == mgntJob)
	return(slcJobs_as[slcJobFunc_as[cyclIx2Exe].slcJobId_e].async_e);
  else	
	return(slcJobs_as[slcJobFunc_as[cyclIx2Exe].slcJobId_e].hdlr_e);
}



/*==============================================================================

  Name: slcAsyncExit()

  Abs:  deallocate job global

  Args: none

  Rem: 

  side:

  Ret: none

  ==============================================================================*/
void slcAsyncExit(void)    
{
  epicsMutexId temp_ps; 

  DEBUGPRINT(DP_DEBUG, slcFlag, ("CSTR_A: Entered slcAsyncExit!\n"));

  /* Indicate that global no longer set up */
  slcAsyncSetAsyncExists(epicsFalse);

  /* Destroy mutex used for protecting Async Job Global */
  if (asyncRWMutex) {
	temp_ps = asyncRWMutex;
    asyncRWMutex = NULL;
	epicsMutexDestroy(temp_ps);
  }

  /* Free async job global */
  if (slcJobFunc_as)
	free(slcJobFunc_as);

  return;

}
