/*=============================================================================
 
  Name: slcCmlog.c
        slcCmlogLogMsg      - Log a Message
        slcCmlogVLogMsg     - Log a Message with Variable Vormat Arguments
        slcCmlogLogMsgFile  - Log a Message with File Name and Line Number

  Abs:  Cmlog utilities for the slc-aware threads 
        Refer to web page LCLS Controls | Facilities | SLC-Aware IOC | General Purpose Utilities

  Auth: 06-Feb-2005, Stephanie Allison
  Rev:  02 feb 2005, Debbie Rogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
 
=============================================================================*/

/* c includes */
#include    <stdio.h>      /* for debugPrint.h  */
#include    <stdarg.h>     /* va_list, va_start */
#include    <string.h>     /* strcpy, strncpy   */

/* site includes */
#include    "iocCmlog.h"   /* iocCmlogVLogMsg   */  
#include    "debugPrint.h" /* DEBUGPRINT        */

/* slc common includes */
#include    "slcUtil.h"    /* slcGetThreadDestinationByTid, V017 */
#include    "slcCmlog.h"   /* slcCmlogLogMsg    */

#ifdef  DEBUG_PRINT
  int slcCmlogFlag = DP_NONE;
#endif

/*=============================================================================

  Abs:  Log a message with slcCmlogVLogMsg

  Name: slcCmlogLogMsg

  Args: Type		    Name	     Access	   Description
	------------------- -------------    ---------- -----------------------
 	const unsigned long codeI	     read-only  Condition code
        const char * const  code_c	     read-only  String version of code
	const char * const  format_c 	     read-only  Format string
	...                 (none)           read-only  Variable arguments
 
  Rem:  Calls slcCmlogVLogMsg

  Ret:  int slcCmlogVLogMsg status code

=============================================================================*/
int slcCmlogLogMsg(const unsigned long codeI,
                   const char * const  code_c,
                   const char * const  format_c, ...)
{
  va_list format_args_p;

  /* get the optional argument list */
  va_start(format_args_p, format_c);
  return slcCmlogVLogMsg(codeI, code_c, 0, 0, format_c, format_args_p);
}

/*=============================================================================

  Abs:  Log a message with slcCmlogVLogMsg

  Name: slcCmlogLogMsgFile

  Args: Type		    Name	     Access	   Description
	------------------- -------------    ---------- -----------------------
 	const unsigned long codeI	     read-only  Condition code
        const char * const  code_c	     read-only  String version of code
	const char * const  filename_c	     read-only  File name (can be NULL)
	const unsigned long lineno	     read-only  File line number
	const char * const  format_c 	     read-only  Format string
	...                 (none)           read-only  Variable arguments
 
  Rem:  Calls slcCmlogVLogMsg

  Ret:  int slcCmlogVLogMsg status code

=============================================================================*/
int slcCmlogLogMsgFile(const unsigned long codeI,
                       const char * const  code_c,
                       const char * const  filename_c,
                       const unsigned long lineno,
                       const char * const  format_c, ...)
{
  va_list format_args_p;

  /* get the optional argument list */
  va_start(format_args_p, format_c);
  return slcCmlogVLogMsg(codeI, code_c, filename_c, lineno,
                         format_c, format_args_p);
}

/*=============================================================================

  Abs:  Log a message using iocCmlogVLogMsg

  Name: slcCmlogVLogMsg

  Args: Type		    Name	     Access	   Description
	------------------- -------------    ---------- -----------------------
 	const unsigned long codeI	     read-only  Condition code
        const char * const  code_c	     read-only  String version of code
	const char * const  filename_c	     read-only  File name (can be NULL)
	const unsigned long lineno	     read-only  File line number
	const char * const  format_c 	     read-only  Format string
	va_list             format_args_p    read-only  Variable arguments
 
  Rem:  looks up destination and calls iocCmlogVLogMsg

  Ret:  int iocCmlogVLogMsg status code

=============================================================================*/
int slcCmlogVLogMsg(const unsigned long codeI,
                    const char * const  code_c,
                    const char * const  filename_c,
                    const unsigned long lineno,
                    const char * const  format_c,
                    va_list             format_args_p)
{
  int  dest;
  char dest_c[5];

  /* get destination and convert it to a string, V017 is default */
  dest = slcGetThreadDestinationByTid(NULL);
  if (dest == 0) strcpy (dest_c, V017);
  else           strncpy(dest_c, (char *)&dest, 4);
  dest_c[4] = 0;
  
  DEBUGPRINT(DP_ERROR, slcCmlogFlag,
             ("Destination = %s, File name = %s, Line no = %ld\n",
              dest_c, filename_c?filename_c:"N/A", lineno));
  DEBUGPRINT(DP_ERROR, slcCmlogFlag,
             ("Condition Code = %X %s\n", (unsigned int)codeI,
              code_c?code_c:"N/A"));
  DEBUGPRINT(DP_ERROR, slcCmlogFlag,
             ("Format String = %s\n", format_c?format_c:"N/A"));

  return iocCmlogVLogMsg("SLC", codeI, IOC_CMLOG_USE_VMS_SEVERITY,
                         IOC_CMLOG_PRINTF, code_c, dest_c, filename_c, lineno,
                         format_c, format_args_p);
}
