/*=============================================================================
 
  Name: slcGlob.c
        slcInitGlobals
        slcThreadGetTid
		slcThreadGetMqId
		slcThreadGetDestination
		slcCpyIOCName           - copy the IOC name from env var into slcIOCName_a global
		slcGetIOCName           - return a ptr to the slcIOCName string
		slcSetProxyIPAddress    - copy the Proxy IP from env var into  slcProxyIPAddr_a global
		slcGetProxyIPAddress    - return a ptr to the slcproxyIPAddr string
		slcSetDbExists          - set the value of dbExists
		slcGetDbExists          - return the current value of dbExists
		slcDbDownloadEventDestroy

  Abs:  general purpose utilities for the slc-aware threads
        reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | General Purpose Utilities

  Auth: 09-SEP-2004, diane fairley (DFAIRLEY): created
  Rev:  02 feb 2005, Debbie Rogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
       27-apr-2007, dfairley added LOCA2 fixes
       25-apr-2006 added slcGetLOCA2Name, slcCpyLOCA2Name (dfairley)
       07-APR-2005, Ron MacKenzie (ronm)
           Removed asyncExists at Steph's behest.  It's in slcAsync.c now.
           Remove setting of slcJobs_as[].async_e because we don't have async jobs.
           Initialize 2 new elements of slcJob_ts which are new for async jobs.
       09-DEC-2004, diane fairley:
           added header, changed a lot to meet new coding standards
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include    <stdio.h>
  int slcGlobFlag = DP_ERROR;
#endif

/* c includes */
#include    <stdlib.h>   /* getenv */
#include    <string.h>   /* memcpy */

/* epics includes */
#include    <epicsEvent.h>  /* epicsEventId */
#include    <cantProceed.h> /* callocMustSucceed */
#include    <freeList.h>    /* freeListInitPvt */

/* slcIOC includes */
#include "slcStructs.h"    /* slcThread_ts, slcSocket_ts, etc. */
#include "tcpTypes.h"      /* MAX_SOCKETS */
#include "msgfunc.hc"      /* FUNC_TEST job code */
#include "msgTypes.h"      /* MAX_SMALL_MSG_SIZE, MAX_LARGE_MSG_SIZE, etc */
#include "slcAsync.h"      /* slcAsyncSetAsyncExists prototype */
#include "clustat.h"       /* CDSC FECC bit */

/* slc utility prototypes */
#include    "cvtMacr.h"   /* macro used for endian-ness */

/* function prototypes */
epicsThreadId       slcThreadGetTid(const slcThread_te thrd_e);
epicsMessageQueueId slcThreadGetMqId(const slcThread_te thrd_e);
unsigned long       slcThreadGetDestination(const slcThread_te thrd_e);
void          slcCpyIOCName(void);
char*         slcGetIOCName(void);
void          slcCpyProxyIPAddress(void);
char*         slcGetProxyIPAddress(void);
void          slcSetDbExists(const epicsBoolean val);
epicsBoolean  slcGetDbExists (void);
void          slcDbDownloadEventDestroy(void);
void          slcCpyLOCAName(void);
char*         slcGetLOCAName(void);
void          slcCpyLOCA2Name(void);
char*         slcGetLOCA2Name(void);

/* global variables */
slcThread_ts slcThreads_as[MAX_THREADS];
slcSocket_ts slcSockets_as[MAX_SOCKETS];
slcJob_ts    slcJobs_as[MAX_SLC_JOBS];
char         slcIocName_a[5];
char         slcLocaName_a[5];
char         slcLoca2Name_a[5];
char         slcProxyIPAddr_a[16];
epicsEventId dbDownloadEvent_ps = NULL;
epicsBoolean dbExists;
epicsBoolean slcIocIsLittleEndian;
void *       slcSmMsgPvt_p = NULL; /* pointer to small-message pool */
void *       slcLrgMsgPvt_p = NULL;/* pointer to large-message pool */

execDiags_ts  execDiagnostics_s = {{0,0}, 0, 0};
epicsBoolean feccIOC;              /* true indicates this is a FECC */

static epicsBoolean slcNotify = epicsFalse;

/*=============================================================================

  Name: slcInitGlobals

  Abs:  initialize the global variables and structures

  Args: none

  Rem:  this function initializes all the global vars 

  Side: 

  Ret: none

==============================================================================*/
/**procedure**/
void  slcInitGlobals(void)
{
  int cnt;

  /* initialize the dbDownloadEvent pointer */
  dbDownloadEvent_ps = NULL; /* DMF leave this in until deb is ready to remove */

  /* set the dbExists and asyncExists flags to false to begin with */
  slcSetDbExists(epicsFalse);
  slcAsyncSetAsyncExists(epicsFalse);

  /* test for endianness */
  slcIocIsLittleEndian=epicsFalse;
  if (ENDIAN_TEST_IS_LITTLE) slcIocIsLittleEndian=epicsTrue;

  /* initialize feccIOC to false */
  feccIOC = epicsFalse;

  /* initialize the threads structure */
  for (cnt=0; cnt<MAX_THREADS; cnt++)
  {
	slcThreads_as[cnt].tid_ps = NO_THREAD; /* no thread */
    slcThreads_as[cnt].active = epicsFalse;
    slcThreads_as[cnt].stop = epicsFalse;
    slcThreads_as[cnt].mqId_ps = NULL;
    slcThreads_as[cnt].destination = V017_UL;
  }

  /* initialize the sockets structure */
  for (cnt=0; cnt<MAX_SOCKETS; cnt++)
  {
	slcSockets_as[cnt].sd = INVALID_SOCKET;
  }

  /* initialize the slcJobs_as -> slc threads map */
  for (cnt=0; cnt<MAX_SLC_JOBS; cnt++) 
  {
    slcJobs_as[cnt].hdlr_e  = slcInvalid;
    slcJobs_as[cnt].async_e  = slcInvalid;
    slcJobs_as[cnt].dbupdate_ps = NULL;
    strcpy(slcJobs_as[cnt].jobName_a,"");
    slcJobs_as[cnt].frstCyclFuncIx = 0;  
    slcJobs_as[cnt].numCyclFunc = 0;  
  }

  /* initialize the message service memory pools */
  freeListInitPvt(&slcSmMsgPvt_p, MAX_SMALL_MSG_BYTES, SMALL_MSGPOOL_ELEMENTS);
  freeListInitPvt(&slcLrgMsgPvt_p, MAX_LARGE_MSG_BYTES, LARGE_MSGPOOL_ELEMENTS);

  /* init the execDiagnostics */
  execDiagnostics_s.numStarts = 0;
  execDiagnostics_s.numStops = 0;

  /* set the micro name strings and the proxy ip address string. 
     These values are retrieved from environment vars set in the
	 startup file for the slcIoc
  */
  slcCpyIOCName();
  slcCpyProxyIPAddress();
  slcCpyLOCAName();
  slcCpyLOCA2Name();

  /* this jobs array puts job information into an array slot based on the
	 slc control system function code for the job.  
  */
  /* init the MSG Service */
  slcJobs_as[(FUNC_MSG>>8)].hdlr_e = msgHdlr;
  strcpy(slcJobs_as[(FUNC_MSG>>8)].jobName_a,MSG_JOB_NAME);
  slcJobs_as[(FUNC_MSG>>8)].dbupdate_ps = NULL;

  /* init the TEST handler, cstr */
  slcJobs_as[(FUNC_TEST>>8)].hdlr_e = cstrHdlr;
  strcpy(slcJobs_as[(FUNC_TEST>>8)].jobName_a, CSTR_JOB_NAME);
  slcJobs_as[(FUNC_TEST>>8)].dbupdate_ps = 
	callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for cstrHdlr dbupdate_ps/n"); 
  slcJobs_as[(FUNC_TEST>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =  
	epicsMutexCreate();
  slcJobs_as[(FUNC_TEST>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =
	epicsEventCreate(epicsEventEmpty);

  /* init the DB Service */
  slcJobs_as[(FUNC_DB>>8)].hdlr_e = dbHdlr;
  strcpy(slcJobs_as[(FUNC_DB>>8)].jobName_a,DBS_JOB_NAME);
  slcJobs_as[(FUNC_DB>>8)].dbupdate_ps = NULL;

  /* init the ioc shell job */
  slcJobs_as[iocsJob].hdlr_e = slcInvalid;
  strcpy(slcJobs_as[iocsJob].jobName_a, IOCS_JOB_NAME);
  slcJobs_as[iocsJob].dbupdate_ps = 
	callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for ioc shell dbupdate_ps/n"); 
  slcJobs_as[iocsJob].dbupdate_ps->dbupdateHiLoMutex_ps =  
	epicsMutexCreate();
  slcJobs_as[iocsJob].dbupdate_ps->dbupdateHiLoAckEvent_ps =
	epicsEventCreate(epicsEventEmpty);

  /* init the bpm job */
  slcJobs_as[(FUNC_BPMO>>8)].hdlr_e = bsacHdlr;
  strcpy(slcJobs_as[(FUNC_BPMO>>8)].jobName_a, BPMO_JOB_NAME);
  slcJobs_as[(FUNC_BPMO>>8)].dbupdate_ps = 
	callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for bsacHdlr dbupdate_ps/n"); 
  slcJobs_as[(FUNC_BPMO>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =  
	epicsMutexCreate();
  slcJobs_as[(FUNC_BPMO>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =  
	epicsEventCreate(epicsEventEmpty);
  
  /* init the stat job */
  slcJobs_as[(FUNC_STAT>>8)].hdlr_e = statHdlr;
  strcpy(slcJobs_as[(FUNC_STAT>>8)].jobName_a, STAT_JOB_NAME);
  slcJobs_as[(FUNC_STAT>>8)].dbupdate_ps = 
	callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for statmain dbupdate_ps/n"); 
  slcJobs_as[(FUNC_STAT>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =  
	epicsMutexCreate();
  slcJobs_as[(FUNC_STAT>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =  
	epicsEventCreate(epicsEventEmpty);
  
  /* init the magnet hdlr job */
  slcJobs_as[(FUNC_MGNT>>8)].hdlr_e = mgntHdlr;
  slcJobs_as[(FUNC_MGNT>>8)].async_e = mgntAsync;
  strcpy(slcJobs_as[(FUNC_MGNT>>8)].jobName_a, MGNT_JOB_NAME);
  slcJobs_as[(FUNC_MGNT>>8)].dbupdate_ps = 
	callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for mgntHdlr dbupdate_ps/n"); 
  slcJobs_as[(FUNC_MGNT>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =  
	epicsMutexCreate();
  slcJobs_as[(FUNC_MGNT>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =  
	epicsEventCreate(epicsEventEmpty);

             /* to support FECC */
  /* time job */
  slcJobs_as[(FUNC_TIMING>>8)].hdlr_e = timeHdlr;
  strcpy(slcJobs_as[(FUNC_TIMING>>8)].jobName_a, TIME_JOB_NAME);
  slcJobs_as[(FUNC_TIMING>>8)].dbupdate_ps =
        callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for statmain dbupdate_ps/n");
  slcJobs_as[(FUNC_TIMING>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =
        epicsMutexCreate();
  slcJobs_as[(FUNC_TIMING>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =
        epicsEventCreate(epicsEventEmpty);
  /* klys job */
  slcJobs_as[(FUNC_KLYS>>8)].hdlr_e = klysHdlr;
  strcpy(slcJobs_as[(FUNC_KLYS>>8)].jobName_a, KLYS_JOB_NAME);
  slcJobs_as[(FUNC_KLYS>>8)].dbupdate_ps =
        callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for statmain dbupdate_ps/n");
  slcJobs_as[(FUNC_KLYS>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =
        epicsMutexCreate();
  slcJobs_as[(FUNC_KLYS>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =
        epicsEventCreate(epicsEventEmpty);
  /* camv job */
  slcJobs_as[(FUNC_CAMV>>8)].hdlr_e = camvHdlr;
  strcpy(slcJobs_as[(FUNC_CAMV>>8)].jobName_a, CAMV_JOB_NAME);
  /* mpg job */
  slcJobs_as[(FUNC_MPG>>8)].hdlr_e = mpgHdlr;
  strcpy(slcJobs_as[(FUNC_MPG>>8)].jobName_a, MPG_JOB_NAME);
  /* crat job */
  slcJobs_as[(FUNC_CRAT>>8)].hdlr_e = cratHdlr;
  strcpy(slcJobs_as[(FUNC_CRAT>>8)].jobName_a, CRAT_JOB_NAME);
  slcJobs_as[(FUNC_CRAT>>8)].dbupdate_ps =
        callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for statmain dbupdate_ps/n");
  slcJobs_as[(FUNC_CRAT>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =
        epicsMutexCreate();
  slcJobs_as[(FUNC_CRAT>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =
        epicsEventCreate(epicsEventEmpty);
  /* fbck job */
  slcJobs_as[(FUNC_FBCK>>8)].hdlr_e = fbckHdlr;
  strcpy(slcJobs_as[(FUNC_FBCK>>8)].jobName_a, FBCK_JOB_NAME);
  slcJobs_as[(FUNC_FBCK>>8)].dbupdate_ps =
        callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for statmain dbupdate_ps/n");
  slcJobs_as[(FUNC_FBCK>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =
        epicsMutexCreate();
  slcJobs_as[(FUNC_FBCK>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =
        epicsEventCreate(epicsEventEmpty);
  /* anlg job */
  slcJobs_as[(FUNC_ANLG>>8)].hdlr_e = anlgHdlr;
  strcpy(slcJobs_as[(FUNC_ANLG>>8)].jobName_a, ANLG_JOB_NAME);
  slcJobs_as[(FUNC_ANLG>>8)].dbupdate_ps =
        callocMustSucceed( 1, sizeof(dbupdate_ts),"no memory for statmain dbupdate_ps/n");
  slcJobs_as[(FUNC_ANLG>>8)].dbupdate_ps->dbupdateHiLoMutex_ps =
        epicsMutexCreate();
  slcJobs_as[(FUNC_ANLG>>8)].dbupdate_ps->dbupdateHiLoAckEvent_ps =
        epicsEventCreate(epicsEventEmpty);
  
  /* micro job */
  slcJobs_as[(FUNC_MICRO>>8)].hdlr_e = microHdlr;
  strcpy(slcJobs_as[(FUNC_MICRO>>8)].jobName_a, MICRO_JOB_NAME);
  /* bcom job */
  slcJobs_as[(FUNC_BCOM>>8)].hdlr_e = bcomHdlr;
  strcpy(slcJobs_as[(FUNC_BCOM>>8)].jobName_a, BCOM_JOB_NAME);
            /* end to support FECC */

}

/*===========================================================

  Name: slcCpyProxyIPString()

  Abs:  set the proxy IP address from the environment var

  Args: none

  Rem:  copy the SLC_PROXY_IP environment variable into the 
        global character array

  Side: none

  Ret: none

===========================================================*/
void slcCpyProxyIPAddress(void)
{
  memset( &slcProxyIPAddr_a, 0, 16);
  memcpy( &slcProxyIPAddr_a, getenv("SLC_PROXY_IP"), 16);
}

/*===========================================================

  Name:  slcGetProxyIPString()

  Abs:   return the proxy ip address string

  Args:  none

  Rem:  get a pointer to the Proxy_IP_ADDRESS string

  Side: none

  Ret:  char * str 

===========================================================*/
char * slcGetProxyIPAddress(void)
{
  return (char *)&slcProxyIPAddr_a;
}


/*===========================================================

  Name: slcCpyIOCName

  Abs:  set the slc microname from the environment var

  Args: none

  Rem: copy the IOC_NAME environment variable into the 
       global character array

  Side: none

  Ret: none

===========================================================*/
void slcCpyIOCName(void)
{
  memset( &slcIocName_a, 0, 5);
  memcpy(&slcIocName_a, getenv("IOC_NAME"), 5);
}

/*===========================================================

  Name: slcGetSlcIocName

  Abs:  get the slc-aware Ioc name

  Args: none

  Rem:  get a pointer to the IOC_NAME string

  Side: none

  Ret:  char * str

===========================================================*/
char * slcGetIOCName(void)
{
  return (char *)&slcIocName_a;
}

/*===========================================================

  Name:  slcGetDbExists

  Abs:   return the value of the dbExists flag

  Args:  none

  Rem:  returns the value of the dbExists boolean flag

  Side: none

  Ret:  returns epicsTrue if dbExists, epicsFalse if not

===========================================================*/
epicsBoolean slcGetDbExists(void)
{
  return dbExists;
}

/*===========================================================

  Name:  slcSetDbExists

  Abs:   set the dbExists flag

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        epicsBoolean        val         read only   value to set the dbExists global

  Rem:  sets the dbExists flag according to boolean val passed in

  Side: sets the dbExists global

  Ret:  none 

===========================================================*/
void slcSetDbExists(const epicsBoolean val)
{
  dbExists = val;
}

/*===========================================================

  Name:  slcGetNotify

  Abs:   return the value of the slcNotify flag

  Args:  none

  Rem:  returns the value of the slcNotify boolean flag

  Side: none

  Ret:  returns epicsTrue if slcNotify, epicsFalse if not

===========================================================*/
epicsBoolean slcGetNotify(void)
{
  return slcNotify;
}

/*===========================================================

  Name:  slcSetNotify

  Abs:   set the slcNotify flag

  Args: Type		        Name	    Access	    Description
	------------------- -----------	---------- ----------------------------
        epicsBoolean        val         read only   value to set slcNotify

  Rem:  sets the slcNotify flag according to boolean val passed in

  Side: sets the slcNotify file-scope flag

  Ret:  none 

===========================================================*/
void slcSetNotify(const epicsBoolean val)
{
  slcNotify = val;
}

/*===========================================================

  Name:  slcThreadGetTid

  Abs:   return the epicsThreadId for the given thread enumeration

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcThread_te        thrd_e      read only   thread enumeration 

  Rem:  checks for a valid thrd_e, returns the assigned tid_ps or NULL

  Side: 

  Ret:  epicsThreadId or NULL if thrd_e is invalid

===========================================================*/
epicsThreadId slcThreadGetTid(const slcThread_te thrd_e)
{
  if (thrd_e < slcEnd)
	return slcThreads_as[thrd_e].tid_ps;
  else
	return NULL;
}

/*===========================================================

  Name:  slcThreadGetMqId

  Abs:  return the epicsMessageQueueId assigned to the given thread

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcThread_te        thrd_e      read only   thread enumeration 

  Rem:   checks for a valid thrd_e, returns the assigned mqId_ps or NULL

  Side: 

  Ret:  epicsMessageQueueId or NULL if thrd_e is invalid 

===========================================================*/
epicsMessageQueueId slcThreadGetMqId(const slcThread_te thrd_e)
{
  if (thrd_e < slcEnd)
	return slcThreads_as[thrd_e].mqId_ps;
  else
	return NULL;
}

/*===========================================================

  Name:  slcThreadGetDestination

  Abs:   returns the destination element of the slcThreads_as struct

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcThread_te        thrd_e      read only   thread enumeration 

  Rem:   checks for a valid thrd_e, returns the assigned destination or 0

  Side: 

  Ret:  unsigned long slcThreads_as[thrd_e]destination or 0 if thrd_e is invalid

===========================================================*/
unsigned long slcThreadGetDestination(const slcThread_te thrd_e)
{
  if (thrd_e < slcEnd)
	return slcThreads_as[thrd_e].destination;
  else
	return 0;
}

/*===========================================================

  Name:  slcDbDownloadEventDestroy

  Abs:   destroy the dbDownloadEvent  

  Args:  none

  Rem:   if the dbDownloadEvent exists, destroy it

  Side: sets the dbDownloadEvent_ps to NULL

  Ret:  none

===========================================================*/
void slcDbDownloadEventDestroy(void)
{
  if (dbDownloadEvent_ps !=NULL) 
  {
	epicsEventDestroy(dbDownloadEvent_ps);
	dbDownloadEvent_ps = NULL;
  }
}

/*===========================================================

  Name: slcCpyLOCAName

  Abs:  set the slcIOC location name from the environment var

  Args: none

  Rem: copy the LOCA_NAME environment variable into the 
       global character array

  Side: none

  Ret: none

===========================================================*/
void slcCpyLOCAName(void)
{
  char * tmpLoca;

  memset( &slcLocaName_a, 0, 5);
  tmpLoca = getenv("LOCA_NAME");
  if (tmpLoca != NULL)
	memcpy(&slcLocaName_a, tmpLoca, 5);
  else
	memcpy(&slcLocaName_a, getenv("IOC_NAME"), 5);
}

/*===========================================================

  Name: slcGetSlcLOCAName

  Abs:  get the slc-aware Ioc location name for epics

  Args: none

  Rem:  get a pointer to the LOCA_NAME string

  Side: none

  Ret:  char * str

===========================================================*/
char * slcGetLOCAName(void)
{
  return (char *)&slcLocaName_a;
}


/*===========================================================

  Name: slcCpyLOCA2Name

  Abs:  set the slcIOC second location name from the environment var

  Args: none

  Rem: copy the LOCA2_NAME environment variable into the 
       global character array

  Side: none

  Ret: none

===========================================================*/
void slcCpyLOCA2Name(void)
{
  char * tmpLoca;

  memset( &slcLoca2Name_a, 0, 5);
  tmpLoca = getenv("LOCA2_NAME");
  if (tmpLoca != NULL)
	memcpy(&slcLoca2Name_a, tmpLoca, 5);
}

/*===========================================================

  Name: slcGetSlcLOCA2Name

  Abs:  get the slc-aware Ioc second location name for epics

  Args: none

  Rem:  get a pointer to the LOCA2_NAME string

  Side: none

  Ret:  char * str

===========================================================*/
char * slcGetLOCA2Name(void)
{
  char tmpLoca2_a[5];

  memset( &tmpLoca2_a, 0, 5);
  if (strcmp((const char *)&tmpLoca2_a, (const char *)&slcLoca2Name_a)==0)
	return NULL;
  else
	return (char *)&slcLoca2Name_a;
}

/*===========================================================

  Name: slcGetSlcFeccIOC

  Abs:  get the CDSC bit and check to see if this is a FECC.
        Return true if this is a FECC IOC.

  Args: none

  Rem: 

  Side: none

  Ret: boolean
===========================================================*/
epicsBoolean slcInitFeccIOC(void)
{
  unsigned short cdsc = 0;

  /* get CSTR CDSC to determine if this is a FECC ioc */
  if (!(slcAsyncGetCDSC (&cdsc))){ 
	DEBUGPRINT(DP_ERROR,slcGlobFlag, ("EXEC: Error reading CSTR CDSC\n"));
	/* other error processing goes here ... */
	return(epicsFalse);
  }
  DEBUGPRINT(DP_INFO,slcGlobFlag, ("EXEC: CSTR CDSC: %x\n", cdsc ));
  if ((cdsc&CDSC_FECC)>0) feccIOC = epicsTrue;
  return(epicsTrue);  
}


/*===========================================================

  Name: slcGetFeccIOC

  Abs:  get the value of the feccIOC global boolean which indicates
        whether this ioc is a FECC ioc

  Args: none

  Rem:  get a pointer to the LOCA_NAME string

  Side: none

  Ret:  boolean value of feccIOC

===========================================================*/
epicsBoolean slcGetFeccIOC(void)
{
  return(feccIOC);
}
