/*=============================================================================
 
  Name: slcUtil.c
        slcGetThreadEnumById        - return thread enumeration given thread id
		slcGetThreadEnumByName      - return thread enumeration given thread name
        slcGetThreadDestinationById - return the destination value for the given thread id
        slcGetJobEnumByName         - return the slcJobs_as enum given the job name
        slcGetJobEnumByThreadEnum   - return the slcJobs_as enum given the thread enum
        slcGetJobEnum               - return the slcJobs_as enum for the current job
        slcPrintMsgheader           - print msgheader

  Abs:  general purpose utilities for the slc-aware threads
        reference document:
		    LCLS Controls | Facilities | SLC-Aware IOC | General Purpose Utilities

  Auth: 09-SEP-2004, diane fairley (DFAIRLEY): created
  Rev:  02 feb 2005, Debbie Rogind
        28 Jul 2005, Kristi Luchini added slcPrintMsgheader        

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  
       28-Jul-2008, K. Luchini (LUCHINI):
           Added header file msgQUtil.h for messageQRelease()
       14-Mar-2004, Ron MacKenzie:
           Change strncmp in slcGetJobEnumByName() to strcmp so we can support
           3 character job names, not just 4 character ones.             
       08-Mar-2004, Ron MacKenzie:
           Change slcGetJobIndexByName() to slcGetJobEnumByName() and have it
           return an enum instead of an int.
       09-DEC-2004, diane fairley:
           added header, changed a lot to meet new coding standards
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int slcUtilFlag = DP_ERROR;
#endif

/* c includes */
#include    <stdlib.h>   /* getenv */
#include    <string.h>   /* memcpy */

/* slc IOC includes */
#include    "micrdef.hc"   /* MICR_OKOK */
#include    "slcGlob.h"    /* slcThreads_as, slcIocName_a, slcProxyIPAddrs_a */

/* function prototypes */
#include    "msgQUtil.h"   /* for messageQRelease() */


/*===========================================================

  Name: slcGetThreadEnumByTid

  Abs:  get the thread enumeration for the thread indicated by tid_ps

  Args: Type		    Name	Access	    Description
	------------------- -----------	---------- ----------------------------
  Args: const epicsThreadId tid_ps      read only   ptr to the epics thread id of the thread 

  Rem:  returns index into slcThreads_as array based on thread id

  Side: if the input poiner is NULL use epicsThreadGetIdSelf

  Ret:  slcThread_te - the enumeration of the thread

===========================================================*/
slcThread_te slcGetThreadEnumByTid(const epicsThreadId tid_ps)
{
  unsigned short cnt;
  epicsThreadId t_ps;

  t_ps = tid_ps;

  if (t_ps==NULL)
  {
	t_ps = epicsThreadGetIdSelf();
  }
	
  for (cnt=0; cnt < slcEnd; cnt++)
  {
	/* search for the thread id */
	if (slcThreads_as[cnt].tid_ps!=NO_THREAD)
	{
	  if (slcThreads_as[cnt].tid_ps == t_ps)
	  {
		return cnt;
	  }
	}
  }

  return slcInvalid;   /* error condition */
}

/*===========================================================

  Name: slcGetThreadDestination

  Abs:  get the thread destination


  Args:none 

  Rem:  returns the destination field of the slcThreads_ts for this thread

  Side: uses epicsThreadGetIdSelf if input id==NULL

  Ret:  returns the value of the destination field

===========================================================*/
unsigned long slcGetThreadDestination(void)
{
  slcThread_te thrd_e;
  unsigned long dest;

  thrd_e = slcGetThreadEnumByTid(NULL);
  if (thrd_e != slcInvalid)
    dest = slcThreadGetDestination(thrd_e);
  else 
	dest = 0;

  return dest;
}

/*===========================================================

  Name: slcGetThreadDestination

  Abs:  get the thread destination

  Args: Type		    Name	Access	    Description
	------------------- -----------	---------- ----------------------------
        const epicsThreadId tid_ps      read only   ptr to the epics thread id of the thread 

  Rem:  returns the destination field of the slcThreads_ts for the given thread id

  Side: uses epicsThreadGetIdSelf if input id==NULL

  Ret:  returns the value of the destination field

===========================================================*/
unsigned long slcGetThreadDestinationByTid(const epicsThreadId tid_ps)
{
  slcThread_te thrd_e;
  unsigned long dest;

  thrd_e = slcGetThreadEnumByTid(tid_ps);
  if (thrd_e != slcInvalid)
    dest = slcThreadGetDestination(thrd_e);
  else 
	dest = 0;

  return dest;
}
/*===========================================================

  Name: slcGetThreadEnumByName

  Abs:  get the thread enumeration for the thread indicated by name

  Args: Type		    Name	Access	    Description
	------------------- -----------	---------- ----------------------------
        const char *        name_c      read only  ptr to the epics name of the thread

  Rem:  return index into slcThreads_as array based on thread name from epicsThreadGetName

  Side: if the input ptr is NULL, use epicsThreadGetNameSelf

  Ret:  slcThread_te - the enumeration of the thread

===========================================================*/
slcThread_te slcGetThreadEnumByName(const char* name_c)
{
  unsigned short i;
  char           tName[MAX_THREADNAME_SIZE];
  const char * inName_c;

  
  if (name_c == NULL)
	inName_c = epicsThreadGetNameSelf();
  else
	inName_c = name_c;

  for (i=0; i < slcEnd; i++)
  {
    /* search for the name passed in  */
    epicsThreadGetName(slcThreads_as[i].tid_ps, (char *)&tName, MAX_THREADNAME_SIZE );
	if (strcmp((char *)&tName, inName_c)==0)
    {
	  return i;
	}
  }
  return slcInvalid;   /* error condition */
}

/*===========================================================

  Name:  slcGetJobEnumByName

  Abs:   return the enum (index) to slcJobs_as for the job name passed in

  Args: Type		     Name	 Access	    Description
	------------------- -----------	---------- ----------------------------
        char *              name_c      read only  ptr to the job name array

  Rem:  return the enum to slcJobs_as for the job name passed in 

  Side: return invalidJob (enum element) if no match is found

  Ret:  slcJob_te - the enum (index) into the slcJobs_as global

===========================================================*/
slcJob_te    slcGetJobEnumByName(const char * const name_pa)
{
  unsigned short i;

  for (i=0; i < invalidJob; i++)
  {
    /* search for the name passed in  */
    if (strcmp((char *)slcJobs_as[i].jobName_a, name_pa)==0)
    {
	  return i;
    }
  }
  return invalidJob;   /* error condition */
}

/*===========================================================

  Name:  slcGetJobEnumByThreadEnum

  Abs:   return the enum (index) to slcJobs_as for the job name passed in

  Args: Type		    Name	Access	    Description
	------------------- -----------	---------- ----------------------------
        slcThread_te        thread_e    read only  thread enumeration

  Rem:  return the enum to slcJobs_as for the thread enumeration passed in 

  Side: return invalidJob (enum element) if no match is found

  Ret:  slcJob_te - the enum (index) into the slcJobs_as global

===========================================================*/
slcJob_te     slcGetJobEnumByThreadEnum(const slcThread_te thread_e)
{
{
  unsigned short i;

  for (i=0; i < invalidJob; i++)
  {
    /* search for the name passed in  */
    if (slcJobs_as[i].hdlr_e == thread_e)
    {
	  return i;
    }
  }
  return invalidJob;   /* error condition */
}
}

/*===========================================================

  Name:  slcGetJobEnum

  Abs:   return the enum (index) to slcJobs_as for the job name passed in

  Args: none

  Rem:  return the enum to slcJobs_as for the calling thread 

  Side: return invalidJob (enum element) if no match is found

  Ret:  slcJob_te - the enum (index) into the slcJobs_as global

===========================================================*/
slcJob_te     slcGetJobEnum(void)
{
  epicsThreadId t_ps;
  slcThread_te  thread_e;

  t_ps = epicsThreadGetIdSelf();
  if ((thread_e = slcGetThreadEnumByTid(t_ps))!=slcInvalid){
	return  slcGetJobEnumByThreadEnum(thread_e);
  }
  else return invalidJob;
 
}

/*=============================================================================

  Name: slcStopThread

  Abs:  Terminate thread by setting the active flag to false

  Args: Type	        Name	        Access	     Description
	------          -----------    ----------   ----------------------------
	slcThread_te    thread_e       read-only    Index to facility

  Rem: The purpose of this function is to set the thread active flag
       false. This which will break the thread out of it's WHILE loop 
       and cause it to begin the termination process.

  Side: The thread active flag is set to false.

  Ret:  None

===============================================================================*/
void slcStopThread(slcThread_te thread_e)
{
  if (thread_e && (thread_e<slcEnd)) {
     slcThreads_as[thread_e].active = epicsFalse;
  }
  return;
}


/*==============================================================================

  Name: slcCleanupThread

  Abs:  Cleanup the globals for thread before shutdown

  Args: Type	        Name	           Access      Description
	------          -----------        ----------  ----------------------------
	slcThread_te    thread_e           read-only   Index to facility
        int (*)()       cleanupRoutine_pf  read-only   Ptr to function
 
  Rem: The purpose of this function is to clean up the slcThreads_as
       struct for the specified thread and release any resources. 

  Side: Thread active and task id are set cleared.

  Ret:  None

==============================================================================*/
void slcCleanupThread(slcThread_te thread_e,int (*cleanupRoutine_pf)() )
{
  int status;

  if (thread_e && (thread_e<slcEnd)) {
     if (slcThreads_as[thread_e].mqId_ps) 
        messageQRelease(thread_e);
     if (cleanupRoutine_pf) 
        status = (*cleanupRoutine_pf)();
     slcThreads_as[thread_e].tid_ps = NO_THREAD;
  }
  return;
}

#ifdef DEBUG_PRINT
/*=============================================================================

  Name slcPrintMsgheader

  Abs:  The purpose of this function is to print the msgheader structure.

  Args: Type	            Name        Access	   Description
        ------------------- -----------	---------- ----------------------------
        msgheader_ts        msgheader   read-only  message header structure
        int                 debugFlag           

  Rem:  a debug routine to print interesting things in the msgheader

  Side: The message header structure is used for all facilities.
 
  Ret:  None
   
==============================================================================*/
void slcPrintMsgheader(const msgheader_ts *msgheader, int debugFlag )
{
  DEBUGPRINT(DP_DEBUG, debugFlag, ("slcPrintMsgheader :the msg header is filled with:\n"));
  DEBUGPRINT(DP_DEBUG, debugFlag, ("\tthe source: %c%c%c%c\n",msgheader->source[0],
		msgheader->source[1],msgheader->source[2],
		msgheader->source[3]));
  DEBUGPRINT(DP_DEBUG, debugFlag, ("\tthe dest: %c%c%c%c\n",msgheader->dest[0],
		msgheader->dest[1],msgheader->dest[2],
		msgheader->dest[3]));
  DEBUGPRINT(DP_DEBUG, debugFlag, ("\tthe timestamp[0]:  0x%.8lx\n",msgheader->timestamp[0]));
  DEBUGPRINT(DP_DEBUG, debugFlag, ("\tthe timestamp[1]:  0x%.8lx\n",msgheader->timestamp[1]));
  DEBUGPRINT(DP_DEBUG, debugFlag, ("\tthe func: %x\n",msgheader->func));
  DEBUGPRINT(DP_DEBUG, debugFlag, ("\tthe datalen: %u\n",msgheader->datalen));
  return;
}
#endif
