/*=============================================================================
 
  Name: tcpUtil.c
        tcpGetConnection
		tcpGetNonblockingConnection
		tcpGetBuffer
		tcpSendBuffer
		tcpSockConnect
		tcpSockNonblockingConnect
		tcpSockClose
		tcpSockCheck
		tcpSockReadyForRead

  Abs:  utilities for making socket connections, registering with
        the proxy server, and receiving messages from sockets

  Auth: 18 AUG 2004,  drogind created: extracted common tcp-releated routines 
                      in msgRecv.c and dbRecv.c  

  Rev:  02 feb 2005, Debbie Rogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        11-feb-2005 dfairley added utilities to support nonblocking sockets
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
#include <stdio.h>
  int tcpUtilFlag = DP_ERROR;
#endif

/* c includes */
#include <string.h>        /* memccpy */
#include <fcntl.h>

/* shared slc micro includes */
#include "micrdef.hc"      /* MICR_EXNET_ error codes */

/* slc includes */
#include    "slcCmlog.h"   /* slcCmlogLogMsg */
#include    "slcGlob.h"    /* slcGetProxyIPAddress */

/* function prototypes */
#include    "tcpUtil.h"    /* socket stuff */

/* local function prototypes */
LOCAL vmsstat_t tcpRead(const slcSocket_ts* sock_ps,
						const int           bytesReqd, 
						int                 *bytesRcvd_p,
						void                *msg_p);
/*=============================================================================

  Name: tcpRead

  Abs:  read a buffer of data from the TCP socket

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcSocket_ts        sock_ps     read only   pointer to the socket struct to read from
        int                 bytesReqd   read only   the number of bytes of data requested
        int *               bytesRcvd_p  read/write  the number of bytes actually received
        void *              msg_p        read/write  a pointer to the message buffer

  Rem:  reads the requested number of bytes into the msg buffer, used only by tcpGetBuffer
        this routine is private to this module
  Side: 

  Ret:  vsstat_t the vms status code. returns several 
        possible errors: MICR_EXNET_EOFERR-when 0 bytes read,
						 MICR_EXNET_RECVFAIL-reports any op sys specific read errors

==============================================================================*/
/* procedure */
LOCAL vmsstat_t tcpRead(const slcSocket_ts* sock_ps,
						const int           bytesReqd, 
						int                 *bytesRcvd_p,
						void                *msg_p)
{
   int            bytes_got;
   int            bytes_left;
   vmsstat_t      iss = MICR_OKOK;

   /* Let's get all the bytes. */
   for (bytes_left = bytesReqd;  bytes_left > 0;  )
  {
	bytes_got = recv(sock_ps->sd,(char *)msg_p + (bytesReqd - bytes_left), bytes_left,0);
	if (bytes_got > 0)
    {
	  bytes_left -= bytes_got;
	}
	else if (bytes_got == 0)
    {
	  /* End of file flag - this is usually a socket disconnect ?  */
	  iss = MICR_EXNET_EOFERR;
      slcCmlogLogMsg(iss, "EXNET_EOFERR", "tcpRead recv = 0 bytes.\n"); 
	  break;
	}
	else /* bytes_got < 0 */
    {
	  /* can check for other errors herer like Normal conn lost connections. */
	  /* Nice error report due to normal disconnect. */
	  iss = MICR_EXNET_RECVFAIL;
      slcCmlogLogMsg(iss, "EXNET_RECVFAIL", "tcpRead recv<0 bytes.\n");
	  break;
    }
  }
 
  *bytesRcvd_p = bytesReqd - bytes_left;

  return iss;

}


/*=============================================================================

  Name: tcpFwdRegister

  Abs:  Tell the forward server (A.K.A. proxy server) who we really are, so he can match up our
        connection with a subset of his stream with the Alpha.

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcSocket_ts*       sock_ps     read only   pointer to the socket struct to use
        unsigned short      port        read only   'TAG' port to register with the proxy

  Rem:  Called only from getConnection.
        This routine is used only by getConnection and is private to this module

  Side: There is no parallel 'unregister' message for the slcIOC (or for the micros)

  IMPORTANT NOTE: the 'port' argument is not really the port number at all, it is
                  a shared flag between the Alpha and micro that indicates which
				  connection on the slcIoc has sent, or should receive, the messages.  
				  The connection in turn indicates the micro 'job' responsible for the 
				  messages on that connection. For example, the 'port' 6060 indicates the
				  message service socket, handled by the msgRecvThread. It is NOT the
				  actual port number used for that connection. 

				  The actual port number for the connection is assigned by the system
				  in the getConnection routine.  The proxy server will store both the 
				  'real port' number and the 'tag port' number (this port argument) when
				  it receives this register_port message. The proxy will use the 
				  'real port' number to communicate and will use the 'tag port' number
				  to identify the slcIoc job.  There must have been assigned ports in
				  the micros sometime in ancient history, and the Alpha, Proxy and slcIocs
				  must continue to use this  'tag port' number as a flag.

  Ret:  vmsstat_t:  MICR_OKOK or MICR_EXNET_FWDREGFAIL.

==============================================================================*/
/**procedure**/
LOCAL vmsstat_t tcpFwdRegister(const slcSocket_ts* sock_ps, const unsigned short port)
{
  fwd_hdr_ts     fwd_header_s;
  vmsstat_t      iss = MICR_OKOK;
 
  /*----------------------------------------------------------------*/

  memset(&fwd_header_s, 0, sizeof(fwd_hdr_ts)); /* zero all */
  fwd_header_s.ip_port_u.s.port = htons(port);  /* give'em the 'port' flag */
  fwd_header_s.cmd = PX_REGISTER_PORT_FUNC; /* register using port value */
  fwd_header_s.crc = 0x55; /* default value for now */
  fwd_header_s.len = 0;    /* no user data for this message */
  
  if ((send(sock_ps->sd, (char *) &fwd_header_s, sizeof(fwd_hdr_ts),0)) <= 0)
  {
	iss = MICR_EXNET_FWREGFAIL;
	slcCmlogLogMsg(iss, "EXNET_FWREGFAIL","fwdRegister send failed.\n"); 
  }
  return iss;
}
 

/*=============================================================================

  Name: tcpGetConnection

  Abs:  get a tcp/ip connection to the proxy server

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        unsigned short      proxyPort   read only  The proxy port number
        unsigned short      srvcPort    read only  the "tag" port number that 
		                                           indicates which slc service is listening
												   to this socket.  ie; 6060=msg service, 
												   6070=db 6090=ctl The proxy uses this 'tag' 
												   port number to identify the correct
												   connection to send responses to...
        slcSocket_ts*      sock_ps      read/write socket structure to hold the returned sd

  Rem:  Create a socket and make a connection to the Proxy server
        This routine is called only by tcpSockConnect and tcpGetNonblockingConnection, it
		is private to this module

  Side:  

  IMPORTANT NOTE: the 'srvcPort' argument is not really the socket port number at all, it is
                  a shared flag between the Alpha and micro that indicates which
				  connection on the slcIoc has sent, or should receive, the messages.  
				  The connection in turn indicates the micro 'job' responsible for the 
				  messages on that connection. For example, the 'port' 6060 indicates the
				  message service socket, handled by the msgRecvThread. It is NOT the
				  actual port number used for that connection. 

				  The actual port number for the connection is assigned by the system
				  in thsi getConnection routine.  The proxy server will store both the 
				  'real port' number and the 'tag port' number (this port argument) when
				  it receives the register_port message sent at the end of the getConnection.
				  The proxy will use the  'real port' number to communicate and will use 
				  the 'tag port' number to identify the slcIoc job.  There must have been 
				  assigned ports in the micros sometime in ancient history, and the
				  Alpha, Proxy and slcIocs must continue to use this  'tag port'
				  number as a flag.

  Ret:  vmsstat_t iss the SLC return code

==============================================================================*/
/* procedure */
vmsstat_t tcpGetConnection(const unsigned short proxyPort,  
						   const unsigned short srvcPort,
						   slcSocket_ts* sock_ps)
{
  struct sockaddr_in   srvrName;        /* server socket name (addr) */
  struct sockaddr      addr;            /* a temp socket structure for getting this port # */
  unsigned int         addrlen;         /* length of this socket info structure */
  int                  rv;     
  struct sockaddr_in   *nptr, myName;   /* ptr to get this port number */
  unsigned short       myPort;          /* this IOC's port number */
  int                  flag = 1;        /* for setsockopt */
  vmsstat_t            iss = MICR_OKOK; /* return status, vms style */


  /* initialize sd to be invalid */
  sock_ps->sd = INVALID_SOCKET;

  /*set up the client socket. */
  sock_ps->sd = epicsSocketCreate(AF_INET,SOCK_STREAM,0);
  if(sock_ps->sd == INVALID_SOCKET)
  {
    iss = MICR_EXNET_SOCKFAIL;
    slcCmlogLogMsg(iss, "EXNET_SOCKFAIL", "%s socket create failed.\n"); 
	return(iss);
  }
  
  /* set up the proxy server information */
  memset (&srvrName, 0, sizeof(srvrName));
  srvrName.sin_port = htons(proxyPort);
  srvrName.sin_family = AF_INET;              /* Internet domain */
  srvrName.sin_addr.s_addr = inet_addr(slcGetProxyIPAddress()); 

  /* More debug code: Verify the contents of structure srvrName
      prior to trying to connect to the (remote) server.    */

  DEBUGPRINT(DP_INFO,tcpUtilFlag,
			 ("TCP: Information provided in client's connect request to Proxy\n"));
  DEBUGPRINT(DP_INFO,tcpUtilFlag,
			 ("\tTCP :Proxy Port number supplied is %u\n", ntohs ( srvrName.sin_port )) );


  /* set SO_REUSEADDR just in case it helps for this O.S.  */
  if ( (rv = setsockopt(sock_ps->sd, SOL_SOCKET, SO_REUSEADDR, (char *)&flag, sizeof(flag)))!=0)
  {
    iss = MICR_EXNET_SOCKFAIL;
    slcCmlogLogMsg(iss, "EXNET_SOCKFAIL", "setsockopt REUSEADDR failed.\n"); 
    epicsSocketDestroy(sock_ps->sd);
	sock_ps->sd=INVALID_SOCKET;
	return(iss);   
  }

  /* THIS is for DEBUGGING Purposes 
	 A bind is done here before the connect, 
	 so that we can then find the port # with getsockname
  */
  myName.sin_family = AF_INET;
  myName.sin_port = 0; /* set to zero so OS assigns the port num */
  myName.sin_addr.s_addr = INADDR_ANY;
  rv = bind(sock_ps->sd,(struct sockaddr *)&myName, sizeof(myName));
  if(rv < 0)
  {
    iss = MICR_EXNET_SOCKFAIL;
    slcCmlogLogMsg(iss, "EXNET_SOCKFAIL", "bind socket failed.\n"); 
    epicsSocketDestroy(sock_ps->sd);
	sock_ps->sd=INVALID_SOCKET;
	return(iss);
  }

  /* for DEBUGGING now,  after bind get the IOC socket information */
  addrlen = sizeof(struct sockaddr);
  if ( ( rv = getsockname(sock_ps->sd, &addr, &addrlen ) ) < 0 )
  {
    iss = MICR_EXNET_SOCKFAIL;
    slcCmlogLogMsg(iss, "EXNET_SOCKFAIL", "getsockname failed.\n"); 
    epicsSocketDestroy(sock_ps->sd);
	sock_ps->sd=INVALID_SOCKET;
	return(iss);
  }

  /*DEBUG advertise the IOC socket information and register with the proxy server */
  nptr = (struct sockaddr_in *) &addr; 
  myPort = ntohs ( nptr -> sin_port); 
  /* DEBUG stmt to make sure the user knows the port number of this IOC */
  DEBUGPRINT(DP_INFO,tcpUtilFlag,("TCP: %s has port number: %u\n",
								  epicsThreadGetNameSelf(), myPort));

  /* Establish socket connection with Proxy server PX01.  */
  if ( ( connect ( sock_ps->sd, (struct sockaddr *)&srvrName, sizeof(srvrName) ) ) < 0 ) {
    iss = MICR_EXNET_CONNFAIL;
    slcCmlogLogMsg(iss, "EXNET_CONNFAIL", "socket connect (to proxy) failed.\n"); 
	epicsSocketDestroy(sock_ps->sd);
	sock_ps->sd = INVALID_SOCKET;
	return(iss);
  }

  /* register this connection with the proxy - give the proxy the 
	 'tag' port number (porttag) so that the VMS jobs can send to 
	 the right connection */
  if (!SUCCESS(iss=tcpFwdRegister(sock_ps, srvcPort)))
  {
	epicsSocketDestroy(sock_ps->sd);
	sock_ps->sd = INVALID_SOCKET;
  }

  return(iss);
}

/*=============================================================================

  Name: tcpGetNonblockingConnection

  Abs:  get a tcp/ip connection to the proxy server then set it as nonblocking

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        unsigned short      proxyPort   read only  The proxy port number
        unsigned short      srvcPort    read only  the "tag" port number that 
		                                           indicates which slc service is listening
												   to this socket.  ie; 6060=msg service, 
												   6070=db 6090=ctl The proxy uses this 'tag' 
												   port number to identify the correct
												   connection to send responses to...
        slcSocket_ts*      sock_ps      read/write socket structure to hold the returned sd

  Rem:  Create a socket and make a connection to the Proxy server
        This routine is called only by tcpSockNonblockingConnection and is private to
		this module.

  Side:  the socket is set to non-blocking

  IMPORTANT NOTE: the 'srvcPort' argument is not really the socket port number at all, it is
                  a shared flag between the Alpha and micro that indicates which
				  connection on the slcIoc has sent, or should receive, the messages.  
				  The connection in turn indicates the micro 'job' responsible for the 
				  messages on that connection. For example, the 'port' 6060 indicates the
				  message service socket, handled by the msgRecvThread. It is NOT the
				  actual port number used for that connection. 

				  The actual port number for the connection is assigned by the system
				  in thsi getConnection routine.  The proxy server will store both the 
				  'real port' number and the 'tag port' number (this port argument) when
				  it receives the register_port message sent at the end of the getConnection.
				  The proxy will use the  'real port' number to communicate and will use 
				  the 'tag port' number to identify the slcIoc job.  There must have been 
				  assigned ports in the micros sometime in ancient history, and the
				  Alpha, Proxy and slcIocs must continue to use this  'tag port'
				  number as a flag.

  Ret:  vmsstat_t iss the SLC return code

==============================================================================*/
/* procedure */
vmsstat_t tcpGetNonblockingConnection(const unsigned short proxyPort,  
									  const unsigned short srvcPort,
									  slcSocket_ts* sock_ps)
{
  vmsstat_t iss;
  int block = 1;

  iss = tcpGetConnection(proxyPort,srvcPort,sock_ps);
  if (SUCCESS(iss)) socket_ioctl(sock_ps->sd, FIONBIO, &block);  
  return(iss);
}


/*=============================================================================

  Name: tcpGetBuffer

  Abs:  read a buffer of data from the TCP socket - used by threads recieving at sockets

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcSocket_ts*       sock_ps     read only  pointer to the socket struct to read from
        int                 bytesReqd   read only  the number of bytes of data requested
        int *               bytesRcvd_p read/write pointere to the number of bytes actually received
        fwd_hdr_ts*         fwd_ps      read/write a pointer to the forward header
        void *              buf_p       read/write a pointer to the message buffer

  Rem:  First reads the forward header and checks for problems.
        checks for: any read error, correct CRC, zero length buffer, and
        the wrong number of data bytes. These are all possible problems 
		in communication with proxy.  
        If the forward header passes all tests, it then reads in the
		message header and data.

  Side: 

  Ret:  vsstat_t the vms status code
        Possible return codes:  
		EXNET_RECVFAIL, EXNET_FWDCRC, EXNET_FWDBCMAX, EXNET_FWDSHT

==============================================================================*/
/* procedure */
vmsstat_t tcpGetBuffer(const slcSocket_ts* sock_ps, 
					   const int    bytesReqd, 
					   int*         bytesRcvd_p,
					   fwd_hdr_ts*  fwd_ps, 
					   void*        buf_p)
{
  vmsstat_t iss;

  if (!SUCCESS(iss = tcpRead(sock_ps,sizeof(fwd_hdr_ts), bytesRcvd_p, fwd_ps ) ))
  { 
    iss = MICR_EXNET_RECVFAIL;
    slcCmlogLogMsg(iss, "EXNET_RECVFAIL", "getBuffer recv fwd_hdr failed.\n"); 
    *bytesRcvd_p = 0;
    return(iss);
  }

  /* verify incoming packet CRC */
  if (fwd_ps->crc != 0x55)
  {
    iss = MICR_EXNET_FWDCRC;
    slcCmlogLogMsg(iss, "EXNET_FWDCRC", "getBuffer received fwd_hdr containing bad crc.\n"); 
    *bytesRcvd_p = 0;
    return iss;
  }
  
   
  /* Handle zero length data sections for things such as keepalive messages. */
  if (fwd_ps->len == 0)
  {
    iss = MICR_OKOK;
    DEBUGPRINT(DP_DEBUG,tcpUtilFlag,("TCP:data len = 0\n"));
    *bytesRcvd_p = 0;
    return iss;
  }
     
   /* Verify the incoming buffer is not bigger than the buffer provided. */
   /* If it's too big then there is a system error on the sending end so */
   /* we should close the connection and start over.                     */
  if ( (ntohl(fwd_ps->len) > bytesReqd) || (buf_p == NULL ) )
  {
    iss = MICR_EXNET_FWDBCMAX;
    slcCmlogLogMsg(iss, "EXNET_FWDBCMAX", 
				   "getBuffer received fwd_hdr containing bytecount .gt. max. requested\n"); 
    *bytesRcvd_p = 0;
    return iss;
  }


  /* Now get the rest of the message into the caller's area.  The wait value */
  /* here COULD be short since all the data should be on its way, something  */
  /* like one second should be enough.  But, we set it to 10 seconds just in */
  /* case.                                                                   */
  if (!SUCCESS(iss = tcpRead(sock_ps,ntohl(fwd_ps->len), bytesRcvd_p, buf_p)))
  {
    *bytesRcvd_p = 0;
    iss = MICR_EXNET_RECVFAIL;
    slcCmlogLogMsg(iss, "EXNET_RECVFAIL", "getBuffer recv data failed.\n"); 
    return iss;
  }


  /* DO NOT RETURN SHORT PACKETS!  Most application code barfs so check it */
  /* here.  If the packet is short and we have some error, simply set the  */
  /* length to zero and set an error so the app will drop the packet,      */
  /* report error, and try again.                                          */
  /* Obvious problem here, when a Micro packet gets lost there is no way   */
  /* to tell the host directly so we rely on internal micro code to detect */
  /* the problem.  Actually, this connection is probably stuffed so we     */
  /* should re-connect and move on.                                        */
 
  if (*bytesRcvd_p == ntohl(fwd_ps->len))
  {
    DEBUGPRINT(DP_DEBUG, tcpUtilFlag,("TCP:got data portion\n"));
    iss = MICR_OKOK;
  }
  else
  {
    iss = MICR_EXNET_FWDBCSHT;
    slcCmlogLogMsg(iss, " EXNET_FWDBCSHT", 
				   "getBuffer received fwd_hdr containing bad bytecount.\n"); 
    *bytesRcvd_p = 0;
  }

  return (iss);
}

/*=============================================================================

  Name: tcpSendBuffer

  Abs:  send a buffer of data out the TCP socket - used by threads sending on sockets

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcSocket_ts*       sock_ps     read only  pointer to the socket struct to send to
        int                 msgBytes    read only  the number of bytes of data to be sent
        void *              msg_p       read only  a pointer to the message buffer

  Rem:  send msgBytes bytes to the given socket. report errors

  Side: 

  Ret:  vsstat_t the vms status code
        Possible return codes:  
		EXNET_SENDFAIL, MICR_OKOK

==============================================================================*/
vmsstat_t tcpSendBuffer(const slcSocket_ts* sock_ps, const int msgBytes, const void * msg_p)
{
  int cnt;
  SOCKET sd;
  vmsstat_t iss = MICR_EXNET_SENDFAIL;

  /* send the message */
  DEBUGPRINT(DP_DEBUG,tcpUtilFlag,("socket desc sd=%d\n", sock_ps->sd));
  sd = sock_ps->sd;
  cnt = send(sd, (char *)msg_p, msgBytes, 0);

  if (cnt>0)
  {
	iss = MICR_OKOK;
  }
  else
  {
    slcCmlogLogMsg(iss, "EXNET_SENDFAIL", "sendBuffer send data failed.\n"); 
  }
  return iss;
}



/*===========================================================

  Name: tcpSockConnect

  Abs:  get a socket connection 
        Used by Recv threads to create and connect to a blocking socket

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcThread_te        thrd_e      read only   enumeration of thread whose stop flag to check
        unsigned short      proxyport   read only   proxy port identifier
        unsigned short      hostport    read only   host service port identifier
        float               waitDeltaTime read only time to sleep between retries to connect
		unsigned long*      elapsedCount_p read/write counter for number of sleeps before connecting
        slcSocket_ts*       sock_ps     read/write  pointer to the socket struct to connect

  Rem:  Loop till a good connection is made, or the thread is stopped.  This routine uses a
        temporary socket structure to mess with the socket, until its done trying. Then it
		will put the results into the global structure pointed to by sock_ps.
        
  Side: starts with the current socket descriptor. Does nothing
        if the current sd is valid
        slcSocket_as[].sd is set by this routine, equals INVALID_SOCKET
         if this routine cannot get a connection

  Ret: vmsstat_t slc return code
        Possible return codes:  
		EXNET_CONNFAIL, MICR_OKOK
===========================================================*/
vmsstat_t tcpSockConnect(const slcThread_te thrd_e,
                         const unsigned short proxyport, 
						 const unsigned short hostport,
						 const float waitDeltaTime,
						 unsigned long *elapsedCount_p,
						 slcSocket_ts* sock_ps)
{
  slcSocket_ts   sock;                /* temporary socket structure */
  vmsstat_t      iss = MICR_OKOK;     /* SLC return code */
  unsigned long  count = 0;

  /* loop until there is a connection made */
  sock.sd = sock_ps->sd;
  do {
	if (sock.sd == INVALID_SOCKET) 
		iss = tcpGetConnection(proxyport, hostport, &sock);
	if (sock.sd == INVALID_SOCKET) {
	  epicsThreadSleep(waitDeltaTime);
	  count++;
	}
  } while ( (sock.sd == INVALID_SOCKET) && (!slcThreads_as[thrd_e].stop) );

  /* save the sd in the global structure */
  *elapsedCount_p = count;
  sock_ps->sd = sock.sd;
  if (sock_ps->sd==INVALID_SOCKET) iss=MICR_EXNET_CONNFAIL;
  return iss;
}

/*===========================================================

  Name: tcpSockNonblockingConnect

  Abs:  get a nonblocking socket connection 
        Used by threads to create and connect to a special purpose non-blocking socket

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcThread_te        thrd_e      read only   enumeration of thread whose stop flag to check
        unsigned short      proxyport   read only   proxy port identifier
        unsigned short      hostport    read only   host service port identifier
        float               waitDeltaTime read only time to sleep between retries to connect
        slcSocket_ts*       sock_ps     read/write  pointer to the socket struct to connect

  Rem:  Loop till a good connection is made, or the thread is stopped.  This routine uses a
        temporary socket structure to mess with the socket, until its done trying. Then it
		will put the results into the global structure pointed to by sock_ps.
        
  Side: starts with the current socket descriptor. Does nothing
        if the current sd is valid
        slcSocket_as[].sd is set by this routine, equals INVALID_SOCKET
         if this routine cannot get a connection

  Ret: vmsstat_t slc return code
        Possible return codes:  
		EXNET_CONNFAIL, MICR_OKOK
===========================================================*/
vmsstat_t tcpSockNonblockingConnect(const slcThread_te thrd_e,
									const unsigned short proxyport, 
									const unsigned short hostport,
									const float waitDeltaTime,
									unsigned long *elapsedCount_p,
									slcSocket_ts* sock_ps)
{
  slcSocket_ts   sock;                /* temporary socket structure */
  vmsstat_t      iss = MICR_OKOK;     /* SLC return code */
  unsigned long  count = 0;

  /* loop until there is a connection made */
  sock.sd = sock_ps->sd;
  do {
	if (sock.sd == INVALID_SOCKET) 
		iss = tcpGetNonblockingConnection(proxyport, hostport, &sock);
	if (sock.sd == INVALID_SOCKET) { 
	  epicsThreadSleep(waitDeltaTime);
	  count++;
	}
  } while ( (sock.sd == INVALID_SOCKET) && (!slcThreads_as[thrd_e].stop) );

  /* save the sd in the global structure */
  *elapsedCount_p = count;
  sock_ps->sd = sock.sd;
  if (sock_ps->sd==INVALID_SOCKET) iss=MICR_EXNET_CONNFAIL;
  return iss;
}

/*===========================================================

  Name: tcpSockClose

  Abs:  close the given socket
        Used by Recv threads to close a socket (or by a thread controlling a 
		special purpose non-blocking socket)

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcSocket_ts*       sock_ps     read/write  pointer to socket to close

  Rem:  If the socket descriptor is valid, destroy the socket.
        Be sure to set the slcSockets_as[].sd=INVALID_SOCKET
        before destroying it - so that other threads won't try
        to destroy it at the same time.

  Side: sets slcSockets_as[].sd = INVALID_SOCKET

  Ret: none

===========================================================*/
void tcpSockClose(slcSocket_ts* sock_ps)
{
  SOCKET sd;
  int status; 
  vmsstat_t iss = MICR_OKOK; 
 
  if (sock_ps->sd != INVALID_SOCKET) 
  {
	sd = sock_ps->sd;
    /*
	  timing issue here, set the global socket invalid first to keep
      other threads from attempting to use it (or close it) before this
      routine is finished.
	*/
	sock_ps->sd = INVALID_SOCKET;


    /* find out what is required to interrupt the socket */
    switch ( epicsSocketSystemCallInterruptMechanismQuery()  ) {

	case esscimqi_socketCloseRequired:
	  /*
	   on winsock and probably vxWorks shutdown() does not
	   unblock a thread in recv() so we use close()
	  */
	  DEBUGPRINT(DP_INFO, tcpUtilFlag, 
				 ("TCP:CloseReqd destroying socket %d!\n", sd));
	  epicsSocketDestroy (sd);
	  break;

	case esscimqi_socketBothShutdownRequired:
	  /* this case is needed by RTEMS, Linux - all posix sockets os */
	  DEBUGPRINT(DP_INFO, tcpUtilFlag, 
				 ("TCP:ShutdownReqd destroying socket %d!\n", sd));
	  status = shutdown ( sd, SHUT_RDWR );
	  if (status) {
		iss = MICR_EXNET_SOCKFAIL;
		slcCmlogLogMsg(MICR_EXNET_SOCKFAIL, "EXNET_SOCKFAIL",
					   "%s error when closing socket.\n", 
					   epicsThreadGetNameSelf()); 
	  }
	  epicsSocketDestroy(sd);
	  break;

	case esscimqi_socketSigAlarmRequired:
	  /* this case not implemented because only HPUX needs it */
	  /* simply report that a sigAlarm is needed and try to close socket */
	  DEBUGPRINT(DP_INFO, tcpUtilFlag, 
				 ("TCP:SigAlarmReqd destroying socket %d!\n", sd));
	  epicsSocketDestroy(sd);
	  break;

	default:
	  /* unknown what this operating system needs, try both shutdown and destroy */
	  DEBUGPRINT(DP_INFO, tcpUtilFlag, 
				 ("TCP:unknown socket intr. destroying socket %d!\n", sd));
	  status = shutdown ( sd, SHUT_RDWR );
	  if (status) {
		iss = MICR_EXNET_SOCKFAIL;
		slcCmlogLogMsg(MICR_EXNET_SOCKFAIL, "EXNET_SOCKFAIL",
					   "%s error when closing socket.\n", 
					   epicsThreadGetNameSelf()); 
	  }
	  epicsSocketDestroy(sd);
	  break;
	};
  }
}

/*===========================================================

  Name: tcpSockCheck

  Abs: check for a valid socket connection 
       Used by Send threads (that do not control a socket) to wait for a valid socket
	   before using it.

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcSocket_ts*       sock_ps     read only   pointer to socket to check
        slcThread_te        recv_e      read only   enum of recv thread that must be active
        slcThread_te        send_e      read only   enum of send thread that may be stopped
        float               waitDeltaTime  read only sleep this long between checks
		unsigned long*      elapsedCount_p read/write count the number of sleeps - for diagnos.

  Rem: *elapsedCount will be set to zero to start, then incremented with each sleep  

  Ret: epicsTrue if socket is valid, else false
       *elapsedCount - the number of times the sleep was called, so caller can calculate
	   how long this call waited to get a good connection

===========================================================*/
epicsBoolean tcpSockCheck(const slcSocket_ts* sock_ps,
				  const slcThread_te recv_e,
				  const slcThread_te send_e,
				  const float waitDeltaTime,
				  unsigned long *elapsedCount_p)
{
  unsigned long count = 0;

  /* check the connection with the server, 
   * loop until there is a connection made, or the 
   * receive thread becomes inactive (this means it
   * will never get a socket connection)
   */
  while ( (sock_ps->sd == INVALID_SOCKET) && 
		  (slcThreads_as[recv_e].active) &&
		  (!slcThreads_as[send_e].stop) )
  {
	epicsThreadSleep(waitDeltaTime);
	++count;
  } 
  *elapsedCount_p = count;
  if (sock_ps->sd != INVALID_SOCKET) 
	return (epicsTrue);
  else return (epicsFalse);								  
}

/*===========================================================

  Name: tcpSockReadyForRead

  Abs: use select statement to wait at a socket for a message\
       Used by a thread controlling a special-purpose non-blocking socket to check
	   that the socket has a message waiting to be read.

  Args: Type		        Name	    Access	    Description
	    ------------------- -----------	---------- ----------------------------
        slcSocket_ts*       sock_ps     read only   pointer to socket to check
		unsigned short      waittime    read only   timeout value in seconds

  Rem:  

  Ret: vmsstat MICR_OKOK if the socket has a message waiting
       vmsstat MICR_EXNET_RECVTMOUT if the call timed out without getting a msg
	   vmsstat MICR_EXNET RECVFAIL if there was an error on the select call (-1)

===========================================================*/
vmsstat_t  tcpSockReadyForRead(const slcSocket_ts* sock_ps, unsigned short waitTime)
{
  int numread;
  fd_set readset;
  struct timeval t;
  vmsstat_t iss = MICR_OKOK;

  t.tv_sec = waitTime;
  t.tv_usec = 0;
  /* put the socket into a readset */
  FD_ZERO(&readset);
  FD_SET(sock_ps->sd, &readset);
	
  /* now use select to see if it's got something */
  if ( (numread=select((sock_ps->sd)+1, &readset, (fd_set *) 0, (fd_set *) 0, &t))<=0)
  {
	DEBUGPRINT(DP_INFO,tcpUtilFlag,("TCP:number of sockets ready to read = %d\n", numread) );
	if (numread<0) {
	  iss = MICR_EXNET_RECVFAIL;
	  slcCmlogLogMsg(iss, "EXNET_RECVFAIL", "getBuffer recv data failed.\n"); 
	  return iss;
	} else {
	  return MICR_EXNET_RECVTMOUT;
	}	  
  }
  /* 
	 the following debug statement will cause a warning on 
	 a couple of our target OSs. Please comment out when done
	 using it for test
  */
  /* DEBUGPRINT(DP_INFO,tcpUtilFlag,
	 ("TCP:ISSET = %d\n", FD_ISSET(sock_ps->sd,&readset)) );
  */

  if (  FD_ISSET(sock_ps->sd, &readset) )
	return MICR_OKOK;
  else
	return MICR_EXNET_RECVTMOUT;
}

