/*=============================================================================
 
  Name: timeUtil.c
        timeGetVMS
		timeEpicsToVMS
		timeDiffVMS10usec
		timeDiffVMS10msec
		timeVMSDouble
        timeEpicsFromVMS - for debug only, it doesn't work with DST


  Abs:  time conversion utility routines.  These routines convert
        times from vms format to host format and vice versa.  There is a built-in
        test for endian-ness on the host for all conversions.  
        refer to web page LCLS Controls | Facilities | SLC-Aware IOC | General Purpose Utilities

  Auth: 29-nov-2004, diane fairley (DFAIRLEY): moved time functions from cvtUtil
  Rev:   02 feb 2005, Debbie Rogind

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"	/* SLAC copyright comments */
 
/*-----------------------------------------------------------------------------
 
  Mod:  (newest to oldest)  
        DD-MMM-YYYY, My Name:
           Changed such and such to so and so. etc. etc.
        DD-MMM-YYYY, Your Name:
           More changes ... The ordering of the revision history 
           should be such that the NEWEST changes are at the HEAD of
           the list.
 
=============================================================================*/
#include "debugPrint.h"
#ifdef  DEBUG_PRINT
  int timeUtilFlag = DP_ERROR;
#endif
/* c includes */
#include <math.h>
#include <string.h>

/* epics includes */
#include <epicsTime.h>

/* slcIOC includes */
#include "slcGlob.h"
#include "cvtUtil.h"

/* function prototypes */
#include "timeUtil.h"
 
/* local definitions and typedefs */

  /* seconds from Nov. 17, 1858 til Jan.1, 1970 = 3506716800 */
#define NSECS100_1858_TO_1970 3.5067168e+16
  /* seconds from Nov. 17, 1858 til Jan.1, 1990 = 4137868800 */
#define NSECS100_1858_TO_1990  4.1378688e+16
  /* multiplier for hi word of vms timestamp; 2**32 */
#define MULT2EXP32  4294967296.0

 
typedef union { /* represent the vms timestamp in different formats */
  double vmstime;
  unsigned long int liTime[2];
} vmstime_tu;

/*=============================================================================

  Abs:  get the current time in VMS time format

  Name: timeGetVMS

  Args: dest_a - pointer to the destination8 byte  memory location, usually
                an array of two unsigned long integers;
        

  Scop: called by many procedures as a utility

  Rem:  this routine gets the current time in epics format, which is GMT. This time
        is then converted to local time, then to VMS format.  The VMS control system
		works with 'local' time.

  Side: 
  Ret:  an epics error code, epicsTimeOK or epicsTimeERROR

=======================================================================================*/
int timeGetVMS(vmsTimeStamp_ta dest_a)
{
  epicsTimeStamp eTime;
  
  /* get the current time */
  if (epicsTimeGetCurrent(&eTime)==epicsTimeERROR) return epicsTimeERROR;
  /* convert to time_t format so we can get the local time */
  return (timeEpicsToVMS(&eTime, dest_a));
}

/*=============================================================================

  Abs:  convert an epicsTimeStamp to VMS time format

  Name: timeEpicsToVMS

  Args: dest_a - pointer to the destination8 byte  memory location, usually
                 an array of two unsigned long integers;
        eTime_p  an epicsTimeStamp time value

  Scop: called by many procedures as a utility

  Rem:  this routine converts an epics timestamp, which is GMT, to VMS time, which
        is 'local' time. The epicsTimeStamp is converted to local time then to VMS format.
		The VMS control system	works with 'local' time.

  Side: 
  Ret:  an epics error code, epicsTimeOK or epicsTimeERROR

=======================================================================================*/
int timeEpicsToVMS(const epicsTimeStamp* eTime_p, vmsTimeStamp_ta dest_a)
{
#define SEC_PER_DAY     86400
#define SEC_PER_HOUR     3600
#define SEC_PER_MIN        60

  time_t tTime;
  struct tm gmTime, locTime;
  vmstime_tu vmst;
  unsigned long gmsec, locsec;
  double nanos100_in_sec, sec_val,  nsec100_val, temp;
  unsigned char * p;

  if (epicsTimeToTime_t(&tTime, eTime_p)==epicsTimeERROR) return epicsTimeERROR;
  /* now get the local time, VMS wants 'local' timestamps */
  if (epicsTime_localtime(&tTime, &locTime)==epicsTimeERROR) return epicsTimeERROR;
  /* get the GM time as well to correct for local offset */
  if (epicsTime_gmtime(&tTime, &gmTime)==epicsTimeERROR) return epicsTimeERROR;
  
  /* convert epics nsecs into 100 nanosec counts beyond the last second */
  nanos100_in_sec = ((double)eTime_p->nsec) / 100.0;
  /* now get the number of seconds since Jan 1, 1970
     tTime is secs since 1970  in gmtime
     subtract off the gmsec since 1/1/this year then 
     add on the locsec since 1/1/this year.
   */
  gmsec = (gmTime.tm_yday*SEC_PER_DAY) + (gmTime.tm_hour*SEC_PER_HOUR) + 
	      (gmTime.tm_min*SEC_PER_MIN) + gmTime.tm_sec;
  locsec = (locTime.tm_yday*SEC_PER_DAY) + (locTime.tm_hour*SEC_PER_HOUR) + 
	       (locTime.tm_min*SEC_PER_MIN) + locTime.tm_sec;
  sec_val = tTime - gmsec + locsec;
  /* now convert to 100nanosec counts since Nov 17, 1858 (vms time 0) */
  nsec100_val = (sec_val*1.0e+7)+nanos100_in_sec;
  nsec100_val = nsec100_val+NSECS100_1858_TO_1970;

  if (!slcIocIsLittleEndian) {
	/* now create the 2 32-bit hi and lo parts of the vms style timestamp */ 
	vmst.liTime[0] = (unsigned long int)( nsec100_val/MULT2EXP32);
	temp = (double)vmst.liTime[0];
	temp =  nsec100_val - (temp*MULT2EXP32);
	vmst.liTime[1] = (unsigned long int)temp;
	p = (unsigned char *)dest_a;
  } 
  else {
	/* now create the 2 32-bit hi and lo parts of the vms style timestamp */ 
	vmst.liTime[1] = (unsigned long int)( nsec100_val/MULT2EXP32);
	temp = (double)vmst.liTime[1];
	temp =  nsec100_val - (temp*MULT2EXP32);
	vmst.liTime[0] = (unsigned long int)temp;
	p = (unsigned char *)dest_a;
  }
  /* swap the 8 bytes little endian to big endian*/
  if (!slcIocIsLittleEndian) {
	cvtLSwapMove( &vmst.liTime[1],(void *)p);
	cvtLSwapMove( &vmst.liTime[0],(void *)(p+4));
  }
  else memcpy(dest_a,&vmst.liTime[0], 8);

  return epicsTimeOK;
}

/*=============================================================================

  Abs:  get the difference between two VMS format timestamps in 10 usecs units

  Name: timeDiffVMS10usec

  Args: ref_a - first  vms timestamp - the earlier measurement
        later_a - second vms timestamp - ref_a is subtracted from later_a

  Scop: called by many procedures as a utility

  Rem:  This returns a TRUNCATED long int value

  Side: 
  Ret:  an epics error code, epicsTimeOK or epicsTimeERROR

=======================================================================================*/
long int timeDiffVMS10usec(const vmsTimeStamp_ta ref_a, const vmsTimeStamp_ta later_a)
{
  double refTime, laterTime, nsec100_diff;

  refTime = timeVMSDouble(ref_a);
  laterTime = timeVMSDouble(later_a);

  nsec100_diff = laterTime-refTime;

  return((long int)(nsec100_diff*1.0e-2));
}

/*=============================================================================

  Abs:  get the difference between two VMS format timestamps in 10 msec units

  Name: timeDiffVMS10msec

  Args: ref_a - first  vms timestamp - the earlier measurement
        later_a - second vms timestamp - ref_a is subtracted from later_a

  Scop: called by many procedures as a utility

  Rem:  this returns a TRUNCATED long int value

  Side: 
  Ret:  an epics error code, epicsTimeOK or epicsTimeERROR

=======================================================================================*/
long int timeDiffVMS10msec(const vmsTimeStamp_ta ref_a, const vmsTimeStamp_ta later_a)
{
  double refTime, laterTime, nsec100_diff;

  refTime = timeVMSDouble(ref_a);
  laterTime = timeVMSDouble(later_a);

  nsec100_diff = laterTime-refTime;

  return((long int)(nsec100_diff*1.0e-5));
}


/*=============================================================================

  Abs:  get the vms timestamp in a double type variable

  Name: timeVMSDouble

  Args: src_a - the vms timestamp

  Scop: called by diffVMSTimes_10usec and diffVMSTimes_10msec

  Rem:  put the vms timestamp into a double type variable, swap bytes if reqd.

  Side: 
  Ret:  an epics error code, epicsTimeOK or epicsTimeERROR

=======================================================================================*/
double timeVMSDouble(const vmsTimeStamp_ta src_a)
{
  vmstime_tu vmst;
  unsigned char *p;
  double nsec100_val;

  p = (unsigned char *)src_a;

  /* swap the 8 bytes little endian to big endian and place in temp area*/
  if (!slcIocIsLittleEndian) {
	cvtLSwapMove( (void *)p, &vmst.liTime[1]);
	cvtLSwapMove( (void *)(p+4), &vmst.liTime[0]);
  }
  else memcpy(&vmst.liTime[0], src_a,8);

  /* calculate the number of 100nsec counts and store in a double */
  if (!slcIocIsLittleEndian) {
	nsec100_val = (((double)vmst.liTime[0]) * MULT2EXP32) + 
	  (((double)vmst.liTime[1]));
  }
  else {
	nsec100_val = (((double)vmst.liTime[1]) * MULT2EXP32) + 
	  (((double)vmst.liTime[0]));
  }
  return (nsec100_val);
}


/*=============================================================================

  Abs:  convert vms timestamp to an epics timestamp value

        !!!!!!!!!!!!!!!DEBUG ROUTINE ONLY!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        NOTE: this produces the WRONG VALUE - it produces a local time
		from the VMS timestamp, which epics will interpret as a GMTime.
        DONT USE THIS ROUTINE except for debugging purposes.

  Name: cvt_vmsTime_to_epicsTime

  Args: eTime_p - a pointer to an epicsTimeStamp structure
        src_a - a pointer to 64 bits of vms time (usually 2 longwords in an array)

  Scop: 

  Rem:  convert the 64-bit vms timestamp into an epicsTimeStamp. Uses a temporary var
        called vmsTime.  Swaps bytes for endian change if required.

  Side: 

  Ret:  none

=======================================================================================*/
void timeEpicsFromVMS(const vmsTimeStamp_ta src_a, epicsTimeStamp* eTime_p)
{
  vmstime_tu vmst;
  unsigned char *p;
  double nsec100_val,sec_val, secs_since_1990, nsecs, nanos_in_sec;

  p = (unsigned char *)src_a;

  /* swap the 8 bytes little endian to big endian and place in temp area*/
  if (!slcIocIsLittleEndian) {
	cvtLSwapMove( (void *)p, &vmst.liTime[1]);
	cvtLSwapMove( (void *)(p+4), &vmst.liTime[0]);
  }
  else memcpy(&vmst.liTime[0], src_a,8);

  /* calculate the number of 100nsec counts and store in a double */
  if (!slcIocIsLittleEndian) {
	nsec100_val = (((double)vmst.liTime[0]) * MULT2EXP32) + 
	  (((double)vmst.liTime[1]));
  }
  else {
	nsec100_val = (((double)vmst.liTime[1]) * MULT2EXP32) + 
	  (((double)vmst.liTime[0]));
  }

  /* subtract the offset from vms time 0 (nov.17,1858) to epics time 0 (1/1/1990) */
  nsec100_val =nsec100_val-NSECS100_1858_TO_1990; 
  sec_val = nsec100_val*1.0E-7;

  /* now get secs and nano parts for epicsTime */
  nsecs = modf(sec_val, &secs_since_1990);
  /* subtract off the seconds*1.e-7 - leaves the nanoseconds remainder */
  nanos_in_sec = (nsec100_val - (secs_since_1990*1.0e+7))*100; /* *100 for nano*/

  /* place in epicsTime */
  eTime_p->secPastEpoch = (long unsigned int)secs_since_1990;
  eTime_p->nsec = (long unsigned int)(nanos_in_sec);

}

